(* :Title: Make Calendar *)

(* :Author: Mark Fisher *)

(* :Summary:
MakeCalendarList[FirstYear, LastYear] 
is used to create DayCountList, a calendar-like nested list
that--using DayCount--can quickly find the number of business days
since January 1, 1900 for any date from {FirstYear, 1, 1} to
{LastYear, 12, 31}, where dates are given by {year, mo, da}.

Usage:	MakeCalendarList[1987, 2027] >> daycount.m

You may need to edit the filename to ensure "daycount.m" is written to
the proper subdirectory.

Holiday notes: There are two lists of holidays (DateHolidayList and
DoMHolidayList) and GoodFriday that are specified as options to
MakeCalendarList. These lists were made for the mid-1980s and
beyond. They may need to be modified for other eras. The option
GoodFridayQ indicates whether to make Good Friday a holiday. The
option ShiftDateHolidaysQ indicates whether to move date-holidays that
fall on a weekend to the nearest weekday. As an alternative, you may
simply pass a list of all holidays in the appropriate range of
years. Here is the usage: MakeCalendarList[1987, 40, holidaylist].
*)

(* :Context: FRB`MakeCalendar` *)

(* :Mathematica Version: 2.2 *)

(* :History: Version 1.0 by Mark Fisher, June 1995. *)

(* :Keywords: Calendar, Yield Curve *)

(* :Warnings: None *)

(* :Discussion:
After the number of days from {1900, 1, 1} are calculated, a
correction is necessary if the days following are holidays and/or
weekend days.  Most of the code is directed toward making a nested
list that contains the number of days to add to each day.
*)

BeginPackage["FRB`MakeCalendar`", "Miscellaneous`Calendar`"]

MakeCalendarList::usage = "MakeCalendarList[firstyear, lastyear, opts]
returns a calendar-like nested list that can be used to quickly find
the number of business days since January 1, 1900 for any date from
{FirstYear, 1, 1} to {LastYear, 12, 31}. It takes options
DateHolidayList, DoMHolidayList, GoodFridayQ, and ShiftDateHolidayQ.
Usage example:\n\t
MakeCalendarList[1987, 2027] >> daycount.m\n
 You may need to edit the filename to ensure daycount.m is written to
the proper subdirectory.\n 
Holiday notes: There are two lists of holidays (DateHolidayList and
DoMHolidayList) and GoodFriday that are specified as options to
MakeCalendarList. These lists were made for the mid-1980s and
beyond. They may need to be modified for other eras. The option
GoodFridayQ indicates whether to make Good Friday a holiday. The
option ShiftDateHolidaysQ indicates whether to move date-holidays that
fall on a weekend to the nearest weekday.  As an alternative, you may
simply pass a list (that you have created) of all holidays in the
appropriate range of years. For example, MakeCalendarList[1987, 2027,
holidaylist]."

MakeHolidayList::usage = "MakeHolidayList[firstyear, lastyear, opts] is
called by MakeCalendarList to specify the list of holidays."

DateHolidayList::usage = "DateHolidayList is an option for
MakeCalendarList. It is a list of {month, day} pairs. The
default is DateHolidayList -> {{1, 1}, {7, 4}, {11, 11}, {12, 25}}. By
default, if date holidays fall on a weekend, they are shifted to the
nearest weekday. This may be changed with the option
ShiftDateHolidayQ."

ShiftDateHolidaysQ::usage = "ShiftDateHolidaysQ is an option for
MakeCalendarList. It specifies whether to shift date holidays that
fall on a weekend to the nearest weekday. The default is
ShiftDateHolidayQ -> True."

DoMHolidayList::usage = "DoMHolidayList is an option for
MakeCalendarList. It is a list of triples that specify day-of-month
holidays. For example, the third Monday in February is denoted {3, 1,
2}. The default is DoMHolidayList -> {{3, 1, 1}, {3, 1, 2}, {-1, 1,
5}, {1, 1, 9}, {2, 1, 10}, {4, 4, 11}}."

GoodFridayQ::usage = "GoodFridayQ is an option for
MakeCalendarList. It specifies whether to include Good Friday as a
holiday. The default is GoodFridayQ -> True."

MergeCalendarLists::usage = "MergeCalendarLists[{cal1, cal2, ...}]
returns a single calendar list contructed from a list of two or more
calendar lists. MergeCalendarLists can be used to combine calendar
lists different holidays over different date ranges. The following
example illustrates how the date ranges of the lists are combined. Let
the first and last years of the calendar lists be given by {{first1,
last1}, {first2, last2}, {first3, last3}}, where first2 <= last1 + 1
and first3 <= last2 + 1. Then the merged calendar list will use the
following years from each calendar list: {{first1, last1}, {last1+1,
last2}, {last2+1, last3}}."

Begin["`Private`"]

(* need to account for possible holiday at beginning of cal2 *)
FixLastDay[{cal1_, cal2_}] := 
	Module[{year = Length[cal1]},
	ReplacePart[cal1, cal1[[year, 12, 31]] +  cal2[[year + 1, 1, 1]] - 
		DaysBetween[{1900, 1, 1}, {year + 1, 1, 1}], {year, 12, 31}]
	]

MergeCalendarLists[clist:{_List ..}] :=
	Fold[	Join[#1, Drop[#2, Length[#1]]]&, {}, 
			Append[FixLastDay /@ Partition[clist, 2, 1], Last[clist]]
		]

Options[MakeCalendarList] = {
		(* list of {month, day} *)
		DateHolidayList -> {{1, 1}, {7, 4}, {11, 11}, {12, 25}},
		(* list of {3rd, Monday, January} etc. *)
		DoMHolidayList -> {{3, 1, 1}, {3, 1, 2}, {-1, 1, 5}, {1, 1, 9}, 
		{2, 1, 10}, {4, 4, 11}},
		GoodFridayQ -> True,
		ShiftDateHolidaysQ -> True }

MakeCalendarList[firstyear_, lastyear_, opts___Rule] := 
	Module[{Base = firstyear - 1, max = lastyear - firstyear + 1, 
				holidaylist, holidays, wkend, between, temp},
		holidaylist = MakeHolidayList[firstyear, lastyear, opts];
		holidays = (# - {Base, 0, 0})& /@ holidaylist;
		wkend = WeekEnd[Base, max];
		MapThread[(wkend[[##]] = 1)&, Transpose[holidays]];
		between = DaysBetwixt[Base, max];
		temp = between + SumUp[wkend, 1];
		temp = Join[Table[{}, {Base}], temp]
		]

MakeCalendarList[firstyear_, lastyear_, holidaylist_List] := 
	Module[{Base = firstyear - 1, max = lastyear - firstyear + 1, 
				holidays, wkend, between, temp},
		wkend = WeekEnd[Base, max];
		If[holidaylist != {}, 
			holidays = (# - {Base, 0, 0})& /@ holidaylist;
			MapThread[(wkend[[##]] = 1)&, Transpose[holidays]]];
		between = DaysBetwixt[Base, max];
		temp = between + SumUp[wkend, 1];
		temp = Join[Table[{}, {Base}], temp]
		]

(* a useful utility for making calendar-like lists *)
NumberOfDays[year_, month_] := 
	{31, 
	If[Mod[year, 4] == 0 && Mod[year, 100] != 0 ||    (* all this just *)
		Mod[year, 400] == 0, 29, 28],                 (* for February  *)
	31, 30, 31, 30, 31, 31, 30, 31, 30, 31}[[month]];

(* make nested list of 0's (weekdays) and 1's (weekends) *)
WeekEnd[Base_, max_] := 
	Module[{counter = Switch[
		DayOfWeek[ {Base + 1, 1, 1}], 
		Monday, 1, Tuesday, 2, Wednesday, 3, Thursday, 4, Friday, 5,
		Saturday, 6,  Sunday, 7]},
	Table[WeekEndValue[counter++], 
		{i, max}, {j, 12}, {k, NumberOfDays[Base + i, j]}]]

(* 1 if weekend, 0 otherwise *)
WeekEndValue[n_] := If[Mod[Mod[n, 7], 6] == 0, 1, 0]

(*	HolidayList calls 3 separate routines--
	DateHolidays, DoMHolidays, and GoodFriday--
	and produces a list of dates that are used to change the values 
	of the week-end list from zero to one where there are holidays.  *)

MakeHolidayList[firstyear_, lastyear_, opts___Rule] :=
	Module[{
		Base = firstyear - 1, 
		max = lastyear - firstyear + 1,
		gfQ = GoodFridayQ /. {opts} /. Options[MakeCalendarList], 
		datelist = DateHolidayList /. {opts} /. Options[MakeCalendarList], 
		domlist = DoMHolidayList /. {opts} /. Options[MakeCalendarList],
		shift = ShiftDateHolidaysQ /. {opts} /. Options[MakeCalendarList]}, 
	Flatten[{	DateHolidays[Base, max, datelist, shift],
				DoMHolidays[Base, max, domlist],
				If[gfQ, GoodFriday[Base, max], {}, {}]	},	1 ]]

(* DateHolidays makes a list for each year from the MakeCalendarList
option DateHolidaysList. *)

DateHolidays[Base_, max_, DateHolidayList_List, 
		ShiftDateHolidaysQ_:True] := 
	If[ShiftDateHolidaysQ, 
	Switch[DayOfWeek[#], Saturday, DaysPlus[#, -1], Sunday, DaysPlus[#, 1], 
		Monday, #, Tuesday, #,  Wednesday, #, Thursday, #, Friday, #]&,
		##&, ##&] /@ 
	Flatten[
	Table[Join[{Base + i}, #]& /@ DateHolidayList, {i,max}], 
	1]

(* DoMHolidays makes a list for each year from the MakeCalendarList
option DoMHolidayList. *)

DoMHolidays[Base_, max_, DoMHolidayList_List] := 
	Flatten[Table[MapThread[
	{Base + i, #3, First[Position[ MakeMonth[Base + i, #3], #2][[#1]]]
		}&, Transpose[DoMHolidayList] ], {i, max} ], 1 ]

MakeMonth[year_, month_] :=
	Module[{counter = Switch[DayOfWeek[ {year, month, 1}], 
		Monday, 1, Tuesday, 2, Wednesday, 3, Thursday, 4, Friday, 5,
		Saturday, 6,  Sunday, 7]},
		Table[Mod[counter++, 7], {i, NumberOfDays[year, month]}]]

GoodFriday[Base_, max_] := 
	Table[DaysPlus[EasterSunday[Base + i], -2], {i, max}]

(* DaysBetwixt forms a parallel list with the number of days from
		{1900, 1, 1} *)

DaysBetwixt[Base_, max_] :=
	Module[{counter = DaysBetween[{1900, 1, 1}, {Base + 1, 1, 1}]},
		Table[counter++, {i, max}, {j, 12}, 
				{k, NumberOfDays[Base + i, j]}]]

(* SumUp constructs the list of days to add to the number of days
	since 1900 to correct for weekends and holidays.  It "integrates"
	the list backwards; hence, the 2 calls to Map[Reverse, _, {0, 2}].
	Ideally, initialize should be set to the number of days to add to
	last day of the last year.  I simply set it to 1. *)

SumUp[cal_, initialize_Integer] := 
	Module[{nextvalue = initialize},
	Map[Reverse, 
		Map[If[ # == 0, nextvalue = 0, nextvalue = # + nextvalue ]&,
			Map[Reverse, cal, {0, 2}], 
		{3}], 
	{0, 2}] ]

End[]
EndPackage[]
