(* :Title: TriangularPlot3D *)

(* :Author: Mark Fisher *)

(* :Summary:
This package creates and plots a triangular grid of 3-dimensional
points. This is especially useful for functions f[x,y] that are
defined only for y <= x. 
*)

(* :Context: FRB`TriangularPlot3D` *)

(* :Mathematica Version: 2.2 *)

(* :History: Version 1.0 by Mark Fisher, Jan 1995. *)

(* :Keywords: graphics, Plot3D, ListPlot3D *)

(* :Warning: none *)

(* :Limitation: none *)

(* :Discussion: no yet *)

BeginPackage["FRB`TriangularPlot3D`", "Utilities`FilterOptions`"]

TriangularPlot3D::usage = "TriangularPlot3D[f, {x, Lx, Ux}, {y, Ly, Uy}]
returns a 3-dimensional plot of the function f over a triangular region
of the x-y plane determined by the bounds for x and y and the options
SlopeSign and Side. For example, the defaults SlopeSign -> 1 and
Side -> -1, specify that the points to be ploted lie on or
below the line that connects {Lx, Ly} with {Ux, Uy}. The other option
for TriangularPlot3D is PlotPoints, whose default is PlotPoints -> 15.
TriangularPlot3D calls ListTriangularPlot3D."

ListTriangularPlot3D::usage = "ListTriangularPlot3D[trilist] plots a
triangular list of 3-dimensional numerical points in lists of
decreasing length. Some experimentation may be required to get the
desired result. ListTriangularPlot3D is called by TriangularPlot3D."

TriangularListQ::usage = "TriangularListQ[trilist] returns True if
trilist has the structure required for ListTriangularPlot3D, namely a
three-dimensional numerical points in lists of decreasing length."

SlopeSign::usage = "SlopeSign is an option for TriangularPlot3D that
specifies which way to cut the specified region into two triangles.
SlopeSign takes the values 1 and -1, indicating the slope of the
diagonal."

Side::usage = "Side is an option for TriangularPlot3D that specifies
which side of the cut (specified by SlopeSign) to use. Side takes the
values 1 and -1, signifying above and below the diagonal."

Begin["`Private`"]

Options[TriangularPlot3D] = {PlotPoints -> 15, SlopeSign -> 1, 
		Side -> -1}

SetAttributes[TriangularPlot3D, HoldAll]

TriangularPlot3D[fcn_, {x1_, Lx_, Ux_}, {y1_, Ly_, Uy_}, opts___Rule] :=
	Module[{f, x, y, plotpoints, slopesign, yside, slope, yeqn,
		Dx, Dy, tLy, tUy, tab},
	{plotpoints, slopesign, side} = 
		{PlotPoints, SlopeSign, Side} /. {opts} /. 
				Options[TriangularPlot3D];
	f = Function[{x1, y1}, fcn];
	slope = slopesign (Uy - Ly)/(Ux - Lx);
	yeqn = slope(x - Lx) + If[slopesign == 1, Ly, Uy];
	Dx = (Ux - Lx)/(plotpoints - 1);
	Dy = Abs[slope] Dx;
	If[side == -1, tLy = Ly; tUy = yeqn, tLy = yeqn; tUy = Uy];
	tab = Table[{x, y, f[x,y]}, {x, Lx, Ux, Dx}, {y, tLy, tUy, Dy}];
	If[side slopesign == -1, tab = Reverse[tab]];
	If[side == 1, tab = Reverse /@ tab];
	ListTriangularPlot3D[tab, opts]
]

ListTriangularPlot3D::nottri = "The argument is not a
triangular list of 3-dimensional numerical points.";

ListTriangularPlot3D[tab_, opts___Rule] :=
	If[TriangularListQ[tab],
	(*then*)
	Show[Graphics3D[
		Map[Polygon,
			(* rectangles *)
			{#[[{1,2,4,3}]]& /@
				(Flatten[#, 1]& /@
					Flatten[
						Transpose[{Drop[#[[1]], -1], #[[2]]}]& /@ 
							Partition[
								Partition[#, 2, 1]& /@ Drop[tab, -1], 
							2, 1], 
					1]), 
			(* triangles *)
			MapThread[Append,
				{Partition[Last /@ tab, 2, 1], 
				#[[-2]]& /@ Drop[tab, -1]}]}, 
		{2}]
	], 
		FilterOptions[Graphics3D, opts],
		BoxRatios -> {1,1,.4}, 
		PlotRange -> All, 
		Axes -> True],
	(*else*)
	Return[Message[ListTriangularPlot3D::nottri]]]

TriangularListQ[tab_] :=
	(Length /@ tab) == Reverse[Range[Length[tab]]] && 
		Union[Length /@ Flatten[tab, 1]] == {3} &&
		Union[NumberQ /@ Flatten[N[tab]]] == {True}


(* this is an extra semi-unrelated function *)
TriangularTranspose[tab_] := First /@ 
	NestList[Drop[Rest /@ #, -1]&, tab, Length[tab]-1]

End[]
EndPackage[]


(* :Examples:
TriangularPlot3D[Sin[x] Cos[y], {x, 0, 2Pi}, {y, 0, 2Pi}, 
		Side -> -1, SlopeSign -> 1];
TriangularPlot3D[Sin[x] Cos[y], {x, 0, 2Pi}, {y, 0, 2Pi}, 
		Side -> 1, SlopeSign -> 1];
TriangularPlot3D[Sin[x] Cos[y], {x, 0, 2Pi}, {y, 0, 2Pi}, 
		Side -> -1, SlopeSign -> -1];
TriangularPlot3D[Sin[x] Cos[y], {x, 0, 2Pi}, {y, 0, 2Pi}, 
		Side -> 1, SlopeSign -> -1];

(* the following duplicates the first example *)
trilist = Reverse @ Table[{x, y, Sin[x] Cos[y]}, 
		{x, 0, 2Pi, 2Pi/14}, {y, 0, x, 2Pi/14}];
ListTriangularPlot3D[trilist];
*)

