(************** Content-type: application/mathematica **************

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    772976,      40790]*)
(*NotebookOutlinePosition[    773856,      40822]*)
(*  CellTagsIndexPosition[    773812,      40818]*)
(*WindowFrame->Normal*)



Notebook[{


Cell[CellGroupData[{Cell["YieldCurve", "Title",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "A ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " Package for Fitting the Term Structure of Interest Rates with Smoothing \
Splines"
}], "Subtitle",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["Mark Fisher and David Zervos", "Subsubtitle",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
The term structure of interest rates\
\>", "Section",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "The term structure of interest rates occupies a central position in both \
macroeconomics and finance. The fundamental relationship is the ",
  StyleBox["discount function",
    FontWeight->"Bold"],
  ", $\\delta(t,\\tau)$, which gives the price at time $t$ of default-free \
zero-coupon bond that pays one unit at time $\\tau$. Hereafter, we assume \
that the current time is 0, suppress the first index, and write \
$\\delta(\\tau)$. Thus $\\delta(\\tau)$ is the present value of one unit to \
be delivered with certainty in $\\tau$ periods hence. We will find it \
convenient to look at the ",
  StyleBox["log of the discount function",
    FontWeight->"Bold"],
  ", $\\ell(\\tau) := \\log[\\delta(\\tau)]$."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Interest rates can be derived from bond prices. The ",
  StyleBox["zero-coupon (yield) curve",
    FontWeight->"Bold"],
  ", $z(\\tau) := -\\log[\\delta(\\tau)]/\\tau = -\\ell(\\tau)/\\tau$, gives \
the yield-to-maturity on a zero-coupon bond that matures at time $\\tau$. The \
",
  StyleBox["forward rate curve",
    FontWeight->"Bold"],
  ", $f(\\tau) := -d\\log[\\delta(\\tau)]/d\\tau$, gives the marginal return \
at maturity $\\tau$ of extending one's investment. We can write the forward \
curve in terms of the zero curve by first writing $\\delta(\\tau) = e^{-\\tau\
\\,z(\\tau)}$ and then applying the definition for forward rates: $f(\\tau) = \
z(\\tau ) + \\tau \\,z'(\\tau )$. Thus we see that the relationship between \
the zero and forward curves is that of average and marginal curves."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "The techniques embodied in the ",
  StyleBox["YieldCurve",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " package are designed to extract the term structure from a set of bonds \
(",
  StyleBox["i",
    FontSlant->"Italic"],
  ") that are default-free and (",
  StyleBox["ii",
    FontSlant->"Italic"],
  ") whose prices are determined by the present value of their stated \
payments. All U.S. Treasury securities probably satisfy the first criterion, \
but some do not reasonably satisfy the second criterion, such as callable \
bonds, ``flower'' bonds, and bonds ``on special'' in the repo market. "
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
Consider a set of $n$ bonds.  Let $p_i$ be the price of bond $i$, $c_{ij}$ be \
its $j$-th payment, paid at time $\\tau_{ij}$, and $m_i$ be the number of \
remaining payments.  Then\\footnote{$^1$}{``$^\\top$'' denotes transpose. In \
addition, the prices, $p_i$, include accrued interest.}\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$ p_i = \\sum_{j=1}^{m_i} c_{ij}\\,\\delta(\\tau_{ij}) + \\varepsilon_i \
={c_i}^\\top \\widetilde{\\delta}(\\tau_i)  +  \\varepsilon_i $$\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
where $c_i$ is the vector of payments for bond $i$, $\\tau_i$ is the vector \
of maturities of those payments, $\\varepsilon_i$ is a random error and \
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$\\widetilde{\\delta}(\\tau_i) := (\\delta(\\tau_{i1}), \\cdots, \
\\delta(\\tau_{im_{i}}))^\\top$$\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
is the $m_i\\times 1$ column vector that results from applying $\\delta$ to \
each element of $\\tau_i$. Thus, the present value of the payments is given \
by $\\pi_i := {c_i}^\\top \\widetilde{\\delta} (\\tau_i)$.\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["Fitting the term structure", "Section",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
Estimating the term structure from bond prices is not, however, a trivial \
matter. A variety of techniques have been proposed over the past twenth-five \
years. McCulloch (1971, 1975) was the pioneer in this field.  Essentially, \
McCulloch parameterized $\\delta(\\tau)$ as a cubic spline and estimated the \
spline coefficients with linear regression. Following McCulloch, Vasicek and \
Fong (1982), Shea (1984), Jordan (1984), Chambers, Carleton and Waldman \
(1984), and Coleman, Fisher, and Ibbotson (1992), among others, extended the \
spline-based estimation technique to explore tax-related effects on bond \
pricing, to consider different parameterizations of the splines, and to \
analyze potential sources of heteroskedasticity in the residuals. Other \
authors have pursued alternative estimation techniques based on parsimonious \
parameterizations of the discount function. For instance, Nelson and Seigel \
(1987) and Bliss (1993) consider a functional form with only four unknown \
parameters. (In contrast, for a sample of 150 securities, McCulloch would \
typically choose a spline with 18 parameters.) \
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Fisher, Nychka, and Zervos (1995) present an extension of the spline-based \
techniques (see their paper for the specifics.) In particular, they fit \
smoothing splines instead of regression splines. Smoothing splines have a \
penalty for excess ``roughness'' with a single parameter that controls the \
size of the penalty. An increase in the penalty reduces the effective number \
of parameters, so that a single value controls the entire parameterization of \
the spline. For regression splines, the number of parameters must be chosen \
in advance. By contrast, Fisher, Nychka, and Zervos use ",
  StyleBox["generalized cross validation",
    FontWeight->"Bold"],
  " to choose adaptively the roughness penalty---and hence the effective \
number of parameters. In other words, they let the data determine the \
appropriate number of parameters. In addition, they place the spline directly \
on the log of the discount function $\\ell(\\tau)$ and on the forward rate \
function $f(\\tau)$, as well as on the discount function $\\delta(\\tau)$. \
Based on their simulations and on their estimation results (using daily data \
on U.S. Treasury coupon securities from December of 1987 though September \
1994), they found that splining the forward rate function with a smoothing \
spline and choosing the effective number of parameters via generalized cross \
validation produced the most accurate and least biased results."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "This chapter describes the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " package ",
  StyleBox["YieldCurve",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " that implements the techniques described in Fisher, Nychka, and Zervos. \
",
  StyleBox["YieldCurve",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " contains commands to estimate the term structure of interest rates using \
regression splines and smoothing splines---with or without generalized cross \
validation. ",
  StyleBox["YieldCurve",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " also has functions to display and analyze results and produce reports. ",
  StyleBox["YieldCurve",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " calls a number of other Mathematica packages---both standard packages and \
other included packages."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["File names: Packages and Data", "Subsection",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "To install the packages, you should create a subdirectory, for example ",
  StyleBox["ycurve",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ".  The following files should be copied to this subdirectory: ",
  StyleBox["YieldCurve.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["ShowTime.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["PageLayout.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["BSplineBasis.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["CommaDelimited.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["TriangularPlot3D.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["MakeCalendar.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", and ",
  StyleBox["daycount.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ".\\footnote{$^2$}{If are using DOS or Windows, the files names are ",
  StyleBox["yieldcur.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["showtime.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["pagelayo.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["bsplineb.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["commadel.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["triangul.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["makecale.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["daycount.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ".} The first seven files contain ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " packages. The file ",
  StyleBox["daycount.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " contains information for determining the number of days until a payment \
is made. In addition, there are seven files with names of the form ",
  StyleBox["yyyymmdd.dat",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " that contain data on Treasury securities, as well as ",
  StyleBox["strips.dat",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["openmkt.dat",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". The final file in the distribution is ",
  StyleBox["iddrop.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", which contains a list of the CUSIPs of securities with special \
features."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Two other files will be created in your working directory as you use the \
main package: ",
  StyleBox["LStart.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["DLStart.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " If ",
  StyleBox["FunctionalForm -> LogDelta",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " or ",
  StyleBox["FunctionalForm -> DLogDelta",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["DiscountFunction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " will create the file ",
  StyleBox["LStart.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " or ",
  StyleBox["DLStart.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " in the default directory.  It is possible that bad starting values in ",
  StyleBox["LStart.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " or ",
  StyleBox["DLStart.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " could lead to difficultly in estimation. In any case, either file can \
always be deleted, causing internal defaults to be used."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  StyleBox["MakeCalendar.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " contains functions to create and combine calendar lists. You will only \
need to use it if you wish to extend the calendar before 1925 or after 2035 \
or if you wish to use bonds from other countries (which have different \
holidays). "
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Also included are the files ",
  StyleBox["CRSPDailyBonds.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["CRSPMonthlyBonds.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". These files contain packages that provide access to the Center for \
Research in Security Prices (CRSP) Daily and Monthly U.S. Government Bond \
Files. Each package contains (among other things) the function ",
  StyleBox["MassageCRSPData",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", which reads the (ASCII) data files and produces an ",
  StyleBox["MD",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object. Interested users should read the files for additional \
information."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["B-spline bases", "Section",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
A spline is a piecewise polynomial joined at so-called knot points. Let the \
order of the polynomial be given by $r$. Thus a cubic B-spline is order 3, \
while a step function is order 0. At each knot point, the polynomials that \
meet are restricted so that one additional independent parameter is added. \
For example, a step function has only one parameter, and thus there are no \
restrictions between adjacent step functions for an 0-order spline. By \
contrast, a cubic polynomial has four parameters; thus for a cubic spline, \
the level and first two derivatives of each cubic are restricted to be \
identical at the knot points. \
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
A numerically stable parameterization of a spline is provided by a B-spline \
basis. Let $\\{s_k\\}_{k = 1}^K$ denote the knot points, with $s_k < \
s_{k+1}$, $s_1 = 0$, and $s_K = M$, the maximum maturity of any bond in the \
sample.\\footnote{$^3$}{In all cases, we distribute the knot points according \
to the distribution of the final maturities of the bonds. For example, with \
three knot points, we place the single interior knot point, $s_2$, at the \
median maturity.} The knot points define $K-1$  intervals over the domain of \
the spline, $[0,\\,T]$. For the purpose of defining a B-spline basis, it is \
convenient to define an augmented set of knot points, \
$\\{d_k\\}_{k=1}^{K+2\\,r}$, where $d_1 = \\cdots = d_r = s_1$, $d_{K+r+1} = \
\\cdots = d_{K+2\\,r} = s_K$, and $d_{k+r} = s_k$ for $1 \\le k \\le K$. Then \
a B-spline basis is a vector of $\\kappa = K + r - 1$ order-$r$ B-splines \
defined over the domain.\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["Here is an example.", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell[TextData[StyleBox["SetDirectory[\"/tr/data1/m1mef00/math/ycurve\"]; (* \
your \n\t\t\tdirectory here *)\nNeeds[\"YieldCurve`\"]; (* loads BSplineBasis \
as well *)\nOff[ShowTime]; (* turn off timings *)",
  FontFamily->"courier"]], "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
knots = Range[0, 5];
basisknots = PadKnots[knots, 3]\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
{0, 0, 0, 0, 1, 2, 3, 4, 5, 5, 5, 5}\
\>", "\<\
{0, 0, 0, 0, 1, 2, 3, 4, 5, 5, 5, 5}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]]}, Open]],

Cell["\<\
A B-spline is defined by the following recursion, where $1\\le k \\le \\kappa \
$:\\footnote{$^4$}{For a more detailed  discussion of b-spline bases and \
their properties, see de Boor (1978).}\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
(* i-th spline, r-th order, evaluated at tau *)
phi[i_, r_, tau_] := 
  ((tau - k[i]) phi[i, r - 1, tau])/(k[i + r] - k[i]) + 
    ((k[i + r + 1] - tau) phi[i + 1, r - 1, tau])/
\t\t(k[i + r + 1] - k[i + 1])

(* terminal condition *)
phi[i_, 0, tau_] := d[i]

(* i-th spline, j-th region, r-th order, 
\tevaluated at tau *)
phi[i_, j_, r_, tau_] :=
\tphi[i, r, tau] /. {d[j] -> 1, d[_] -> 0}
\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "We can see that the fourth cubic B-spline is a function of five knot \
points beginning with ",
  StyleBox["k[4]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ":"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["Cases[phi[4, 3, tau], k[_], Infinity] //Union", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
{k[4], k[5], k[6], k[7], k[8]}\
\>", "\<\
{k[4], k[5], k[6], k[7], k[8]}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  "The functional form for the fourth cubic spline over the sixth region \
(between ",
  StyleBox["k[6]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["k[7]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ") is given by"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["phi[4, 6, 3, tau]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
((tau - k[4])*(-tau + k[7])^2)/
 
   ((-k[4] + k[7])*(-k[5] + k[7])*(-k[6] + k[7])) + 
 
  ((-tau + k[8])*(((tau - k[5])*(-tau + k[7]))/
 
        ((-k[5] + k[7])*(-k[6] + k[7])) + 
 
       ((tau - k[6])*(-tau + k[8]))/
 
        ((-k[6] + k[7])*(-k[6] + k[8]))))/(-k[5] + k[8])\
\>", "\<\
                                  2
        (tau - k[4]) (-tau + k[7])
-------------------------------------------- + 
(-k[4] + k[7]) (-k[5] + k[7]) (-k[6] + k[7])
 
                   (tau - k[5]) (-tau + k[7])
  ((-tau + k[8]) (----------------------------- + 
                  (-k[5] + k[7]) (-k[6] + k[7])
 
        (tau - k[6]) (-tau + k[8])
       -----------------------------)) / (-k[5] + k[8])
       (-k[6] + k[7]) (-k[6] + k[8])\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  "The function ",
  StyleBox["MakeBSpline",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " will produce a ",
  StyleBox["BSplineFunction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ":"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["?MakeBSpline", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
MakeBSpline[knots, d:0] takes a list of knot points and
   returns an InterpolatingFunction that represents a
   B-spline of order Length[knots] - 2. Valid orders are 0
   (step function), 1 (linear), 2 (quadratic), and 3
   (cubic), or higher. (Higher order splines use less
   efficient routines.) If the optional second argument is
   given as 1, then the derivative of the spline is
   returned. If the optional third argument is given as -1,
   then the integral of the spline is returned.
   Differentiation is supported to any order; integration is
   supported to order -2.\
\>", "Print",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell["\<\
We can construct a cubic BSplineFunction for the forth B-spline over the \
sixth region:\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["bs463 = MakeBSpline[Take[basisknots, {4, 8}]]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
BSplineFunction[{0, 4}, Which[#1 < 0, 0, #1 <= 4, 
 
    InterpolatingFunction[{0, 4}, 
 
      {{0, 0, {0, 0}, {0}}, 
 
       {1, 0, {1/6, 1/2, 1/3, 1/6}, {0, 0, 1}}, 
 
       {2, 1, {2/3, 0, -1/2, -1/2}, {0, 0, 1}}, 
 
       {3, 2, {1/6, -1/2, 0, 1/2}, {0, 0, 1}}, 
 
       {4, 3, {0, 0, 1/6, -1/6}, {0, 0, 1}}}][#1], True, 0]\\
 
     & ]\
\>", "\<\
BSplineFunction[{0, 4}, <>]\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["Plot[bs463[tau], {tau, 0, 4}];", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.238095 0.0147151 0.882906 [
[(1)] .2619 .01472 0 2 Msboxa
[(2)] .5 .01472 0 2 Msboxa
[(3)] .7381 .01472 0 2 Msboxa
[(4)] .97619 .01472 0 2 Msboxa
[(0.1)] .01131 .10301 1 0 Msboxa
[(0.2)] .01131 .1913 1 0 Msboxa
[(0.3)] .01131 .27959 1 0 Msboxa
[(0.4)] .01131 .36788 1 0 Msboxa
[(0.5)] .01131 .45617 1 0 Msboxa
[(0.6)] .01131 .54446 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
.002 w
.2619 .01472 m
.2619 .02097 L
s
P
[(1)] .2619 .01472 0 2 Mshowa
p
.002 w
.5 .01472 m
.5 .02097 L
s
P
[(2)] .5 .01472 0 2 Mshowa
p
.002 w
.7381 .01472 m
.7381 .02097 L
s
P
[(3)] .7381 .01472 0 2 Mshowa
p
.002 w
.97619 .01472 m
.97619 .02097 L
s
P
[(4)] .97619 .01472 0 2 Mshowa
p
.001 w
.07143 .01472 m
.07143 .01847 L
s
P
p
.001 w
.11905 .01472 m
.11905 .01847 L
s
P
p
.001 w
.16667 .01472 m
.16667 .01847 L
s
P
p
.001 w
.21429 .01472 m
.21429 .01847 L
s
P
p
.001 w
.30952 .01472 m
.30952 .01847 L
s
P
p
.001 w
.35714 .01472 m
.35714 .01847 L
s
P
p
.001 w
.40476 .01472 m
.40476 .01847 L
s
P
p
.001 w
.45238 .01472 m
.45238 .01847 L
s
P
p
.001 w
.54762 .01472 m
.54762 .01847 L
s
P
p
.001 w
.59524 .01472 m
.59524 .01847 L
s
P
p
.001 w
.64286 .01472 m
.64286 .01847 L
s
P
p
.001 w
.69048 .01472 m
.69048 .01847 L
s
P
p
.001 w
.78571 .01472 m
.78571 .01847 L
s
P
p
.001 w
.83333 .01472 m
.83333 .01847 L
s
P
p
.001 w
.88095 .01472 m
.88095 .01847 L
s
P
p
.001 w
.92857 .01472 m
.92857 .01847 L
s
P
p
.002 w
0 .01472 m
1 .01472 L
s
P
p
.002 w
.02381 .10301 m
.03006 .10301 L
s
P
[(0.1)] .01131 .10301 1 0 Mshowa
p
.002 w
.02381 .1913 m
.03006 .1913 L
s
P
[(0.2)] .01131 .1913 1 0 Mshowa
p
.002 w
.02381 .27959 m
.03006 .27959 L
s
P
[(0.3)] .01131 .27959 1 0 Mshowa
p
.002 w
.02381 .36788 m
.03006 .36788 L
s
P
[(0.4)] .01131 .36788 1 0 Mshowa
p
.002 w
.02381 .45617 m
.03006 .45617 L
s
P
[(0.5)] .01131 .45617 1 0 Mshowa
p
.002 w
.02381 .54446 m
.03006 .54446 L
s
P
[(0.6)] .01131 .54446 1 0 Mshowa
p
.001 w
.02381 .03237 m
.02756 .03237 L
s
P
p
.001 w
.02381 .05003 m
.02756 .05003 L
s
P
p
.001 w
.02381 .06769 m
.02756 .06769 L
s
P
p
.001 w
.02381 .08535 m
.02756 .08535 L
s
P
p
.001 w
.02381 .12066 m
.02756 .12066 L
s
P
p
.001 w
.02381 .13832 m
.02756 .13832 L
s
P
p
.001 w
.02381 .15598 m
.02756 .15598 L
s
P
p
.001 w
.02381 .17364 m
.02756 .17364 L
s
P
p
.001 w
.02381 .20895 m
.02756 .20895 L
s
P
p
.001 w
.02381 .22661 m
.02756 .22661 L
s
P
p
.001 w
.02381 .24427 m
.02756 .24427 L
s
P
p
.001 w
.02381 .26193 m
.02756 .26193 L
s
P
p
.001 w
.02381 .29724 m
.02756 .29724 L
s
P
p
.001 w
.02381 .3149 m
.02756 .3149 L
s
P
p
.001 w
.02381 .33256 m
.02756 .33256 L
s
P
p
.001 w
.02381 .35022 m
.02756 .35022 L
s
P
p
.001 w
.02381 .38554 m
.02756 .38554 L
s
P
p
.001 w
.02381 .40319 m
.02756 .40319 L
s
P
p
.001 w
.02381 .42085 m
.02756 .42085 L
s
P
p
.001 w
.02381 .43851 m
.02756 .43851 L
s
P
p
.001 w
.02381 .47383 m
.02756 .47383 L
s
P
p
.001 w
.02381 .49148 m
.02756 .49148 L
s
P
p
.001 w
.02381 .50914 m
.02756 .50914 L
s
P
p
.001 w
.02381 .5268 m
.02756 .5268 L
s
P
p
.001 w
.02381 .56212 m
.02756 .56212 L
s
P
p
.001 w
.02381 .57977 m
.02756 .57977 L
s
P
p
.001 w
.02381 .59743 m
.02756 .59743 L
s
P
p
.001 w
.02381 .61509 m
.02756 .61509 L
s
P
p
.002 w
.02381 0 m
.02381 .61803 L
s
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
.004 w
.02381 .01472 m
.02505 .01472 L
.02629 .01472 L
.02753 .01472 L
.02877 .01472 L
.03001 .01472 L
.03125 .01472 L
.03249 .01472 L
.03373 .01473 L
.03621 .01474 L
.03745 .01474 L
.03869 .01475 L
.04117 .01477 L
.04365 .0148 L
.04613 .01484 L
.04861 .01488 L
.05357 .015 L
.05853 .01517 L
.06349 .0154 L
.06845 .01569 L
.07341 .01605 L
.08333 .01701 L
.09325 .01837 L
.10317 .02017 L
.1131 .02248 L
.12302 .02536 L
.14286 .03311 L
.1627 .04392 L
.18254 .05832 L
.20238 .07679 L
.22222 .09987 L
.2619 .16187 L
.30159 .24566 L
.34127 .34172 L
.38095 .43777 L
.42063 .52157 L
.44048 .55503 L
.4504 .56899 L
.46032 .58084 L
.47024 .59039 L
.4752 .59424 L
.48016 .59744 L
.48512 .59998 L
.4876 .60099 L
.49008 .60182 L
.49256 .60247 L
.4938 .60273 L
.49504 .60294 L
.49628 .60311 L
.49752 .60322 L
Mistroke
.49876 .6033 L
.5 .60332 L
.50124 .6033 L
.50248 .60322 L
.50372 .60311 L
.50496 .60294 L
.5062 .60273 L
.50744 .60247 L
.50992 .60182 L
.51488 .59998 L
.51984 .59744 L
.52976 .59039 L
.53968 .58084 L
.55952 .55503 L
.57937 .52157 L
.61905 .43777 L
.65873 .34172 L
.69841 .24566 L
.7381 .16187 L
.75794 .12806 L
.77778 .09987 L
.79762 .07679 L
.81746 .05832 L
.8373 .04392 L
.85714 .03311 L
.86706 .02888 L
.87698 .02536 L
.8869 .02248 L
.89683 .02017 L
.90675 .01837 L
.91667 .01701 L
.92163 .01649 L
.92659 .01605 L
.93155 .01569 L
.93651 .0154 L
.94147 .01517 L
.94643 .015 L
.94891 .01494 L
.95139 .01488 L
.95387 .01484 L
.95635 .0148 L
.95883 .01477 L
.96131 .01475 L
.96379 .01474 L
.96503 .01473 L
.96627 .01473 L
.96751 .01472 L
.96875 .01472 L
.96999 .01472 L
.97123 .01472 L
Mistroke
.97247 .01472 L
.97371 .01472 L
.97495 .01472 L
.97619 .01472 L
Mfstroke
P
P
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell[TextData[{
  "The internal structure of a ",
  StyleBox["BSplineFunction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " can be seen to be an ",
  StyleBox["InterpolatingFunction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " wrapped in a ",
  StyleBox["Which",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ":"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["InputForm[bs463]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
InputForm[BSplineFunction[{0, 4}, 
 
   Which[#1 < 0, 0, #1 <= 4, 
 
     InterpolatingFunction[{0, 4}, 
 
       {{0, 0, {0, 0}, {0}}, 
 
        {1, 0, {1/6, 1/2, 1/3, 1/6}, {0, 0, 1}}, 
 
        {2, 1, {2/3, 0, -1/2, -1/2}, {0, 0, 1}}, 
 
        {3, 2, {1/6, -1/2, 0, 1/2}, {0, 0, 1}}, 
 
        {4, 3, {0, 0, 1/6, -1/6}, {0, 0, 1}}}][#1], True, 0]
 
      & ]]\
\>", "\<\
BSplineFunction[{0, 4}, Which[#1 < 0, 0, #1 <= 4, 
    InterpolatingFunction[{0, 4}, 
      {{0, 0, {0, 0}, {0}}, 
       {1, 0, {1/6, 1/2, 1/3, 1/6}, {0, 0, 1}}, 
       {2, 1, {2/3, 0, -1/2, -1/2}, {0, 0, 1}}, 
       {3, 2, {1/6, -1/2, 0, 1/2}, {0, 0, 1}}, 
       {4, 3, {0, 0, 1/6, -1/6}, {0, 0, 1}}}][#1], True, 0]\\
     & ]\
\>"], "Output",
  LineSpacing->{1, 0},
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  "The ",
  StyleBox["InterpolatingFunction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " contains information about the polynomials that make up the B-spline. We \
can see the polynomials with ",
  StyleBox["PlotIFPolynomials",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ":"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["PlotIFPolynomials[bs463];", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.238095 0.2354 0.220931 [
[(1)] .2619 .2354 0 2 Msboxa
[(2)] .5 .2354 0 2 Msboxa
[(3)] .7381 .2354 0 2 Msboxa
[(4)] .97619 .2354 0 2 Msboxa
[(-1)] .01131 .01447 1 0 Msboxa
[(-0.5)] .01131 .12493 1 0 Msboxa
[(0.5)] .01131 .34587 1 0 Msboxa
[(1)] .01131 .45633 1 0 Msboxa
[(1.5)] .01131 .5668 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
.002 w
.2619 .2354 m
.2619 .24165 L
s
P
[(1)] .2619 .2354 0 2 Mshowa
p
.002 w
.5 .2354 m
.5 .24165 L
s
P
[(2)] .5 .2354 0 2 Mshowa
p
.002 w
.7381 .2354 m
.7381 .24165 L
s
P
[(3)] .7381 .2354 0 2 Mshowa
p
.002 w
.97619 .2354 m
.97619 .24165 L
s
P
[(4)] .97619 .2354 0 2 Mshowa
p
.002 w
0 .2354 m
1 .2354 L
s
P
p
.002 w
.02381 .01447 m
.03006 .01447 L
s
P
[(-1)] .01131 .01447 1 0 Mshowa
p
.002 w
.02381 .12493 m
.03006 .12493 L
s
P
[(-0.5)] .01131 .12493 1 0 Mshowa
p
.002 w
.02381 .34587 m
.03006 .34587 L
s
P
[(0.5)] .01131 .34587 1 0 Mshowa
p
.002 w
.02381 .45633 m
.03006 .45633 L
s
P
[(1)] .01131 .45633 1 0 Mshowa
p
.002 w
.02381 .5668 m
.03006 .5668 L
s
P
[(1.5)] .01131 .5668 1 0 Mshowa
p
.001 w
.02381 .03656 m
.02756 .03656 L
s
P
p
.001 w
.02381 .05865 m
.02756 .05865 L
s
P
p
.001 w
.02381 .08075 m
.02756 .08075 L
s
P
p
.001 w
.02381 .10284 m
.02756 .10284 L
s
P
p
.001 w
.02381 .14703 m
.02756 .14703 L
s
P
p
.001 w
.02381 .16912 m
.02756 .16912 L
s
P
p
.001 w
.02381 .19121 m
.02756 .19121 L
s
P
p
.001 w
.02381 .21331 m
.02756 .21331 L
s
P
p
.001 w
.02381 .25749 m
.02756 .25749 L
s
P
p
.001 w
.02381 .27959 m
.02756 .27959 L
s
P
p
.001 w
.02381 .30168 m
.02756 .30168 L
s
P
p
.001 w
.02381 .32377 m
.02756 .32377 L
s
P
p
.001 w
.02381 .36796 m
.02756 .36796 L
s
P
p
.001 w
.02381 .39005 m
.02756 .39005 L
s
P
p
.001 w
.02381 .41214 m
.02756 .41214 L
s
P
p
.001 w
.02381 .43424 m
.02756 .43424 L
s
P
p
.001 w
.02381 .47842 m
.02756 .47842 L
s
P
p
.001 w
.02381 .50052 m
.02756 .50052 L
s
P
p
.001 w
.02381 .52261 m
.02756 .52261 L
s
P
p
.001 w
.02381 .5447 m
.02756 .5447 L
s
P
p
.001 w
.02381 .58889 m
.02756 .58889 L
s
P
p
.001 w
.02381 .61098 m
.02756 .61098 L
s
P
p
.002 w
.02381 0 m
.02381 .61803 L
s
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
p
.004 w
.02381 .2354 m
.02505 .2354 L
.02629 .2354 L
.02753 .2354 L
.02877 .2354 L
.03001 .2354 L
.03125 .2354 L
.03249 .2354 L
.03373 .2354 L
.03621 .23541 L
.03745 .23541 L
.03869 .23541 L
.04117 .23541 L
.04365 .23542 L
.04613 .23543 L
.04861 .23544 L
.05357 .23547 L
.05853 .23551 L
.06349 .23557 L
.06845 .23564 L
.07341 .23573 L
.08333 .23598 L
.09325 .23631 L
.10317 .23676 L
.1131 .23734 L
.12302 .23806 L
.14286 .24 L
.1627 .24271 L
.18254 .24631 L
.20238 .25093 L
.22222 .25671 L
.2619 .27222 L
.30159 .29387 L
.34127 .32268 L
.38095 .35967 L
.42063 .40587 L
.46032 .4623 L
.5 .52997 L
.53968 .60993 L
s
.54313 .61803 m
.53968 .60993 L
s
s
s
s
s
s
s
s
s
s
s
s
P
P
p
p
.004 w
.02381 .38269 m
.04365 .34887 L
.06349 .32081 L
.08333 .29811 L
.10317 .2804 L
.12302 .2673 L
.13294 .26235 L
.14286 .25841 L
.14782 .25681 L
.15278 .25543 L
.15774 .25429 L
.1627 .25336 L
.16518 .25298 L
.16766 .25266 L
.17014 .25238 L
.17262 .25216 L
.1751 .25198 L
.17634 .25192 L
.17758 .25186 L
.17882 .25182 L
.18006 .25179 L
.1813 .25177 L
.18254 .25177 L
.18378 .25177 L
.18502 .25179 L
.18626 .25182 L
.1875 .25186 L
.18998 .25198 L
.19246 .25214 L
.19494 .25235 L
.19742 .2526 L
.20238 .25324 L
.2123 .255 L
.22222 .25739 L
.24206 .26385 L
.2619 .27222 L
.30159 .29319 L
.34127 .31723 L
.38095 .34126 L
.42063 .36223 L
.44048 .37061 L
.4504 .3741 L
.46032 .37706 L
.47024 .37945 L
.4752 .38041 L
.48016 .38122 L
.48512 .38185 L
.4876 .3821 L
.49008 .38231 L
.49256 .38247 L
Mistroke
.4938 .38254 L
.49504 .38259 L
.49628 .38263 L
.49752 .38266 L
.49876 .38268 L
.5 .38269 L
.50124 .38268 L
.50248 .38266 L
.50372 .38263 L
.50496 .38259 L
.5062 .38254 L
.50744 .38247 L
.50992 .3823 L
.5124 .38207 L
.51488 .3818 L
.51984 .38109 L
.5248 .38017 L
.52976 .37902 L
.53968 .37604 L
.5496 .3721 L
.55952 .36715 L
.57937 .35405 L
.59921 .33634 L
.61905 .31365 L
.65873 .25177 L
.69841 .16534 L
.7381 .05129 L
Mfstroke
.75216 0 m
.7381 .05129 L
s
s
s
s
s
s
s
P
P
p
p
.004 w
s
s
s
s
s
s
.24784 0 m
.2619 .05129 L
s
.2619 .05129 m
.30159 .16534 L
.34127 .25177 L
.36111 .28558 L
.38095 .31365 L
.40079 .33634 L
.42063 .35405 L
.44048 .36715 L
.4504 .3721 L
.46032 .37604 L
.47024 .37902 L
.4752 .38017 L
.48016 .38109 L
.48512 .3818 L
.4876 .38207 L
.49008 .3823 L
.49256 .38247 L
.4938 .38254 L
.49504 .38259 L
.49628 .38263 L
.49752 .38266 L
.49876 .38268 L
.5 .38269 L
.50124 .38268 L
.50248 .38266 L
.50372 .38263 L
.50496 .38259 L
.5062 .38254 L
.50744 .38247 L
.50992 .38231 L
.51488 .38185 L
.51984 .38122 L
.52976 .37945 L
.53968 .37706 L
.55952 .37061 L
.57937 .36223 L
.61905 .34126 L
.65873 .31723 L
.69841 .29319 L
.7381 .27222 L
.75794 .26385 L
.76786 .26036 L
.77778 .25739 L
.7877 .255 L
.79266 .25404 L
.79762 .25324 L
.80258 .2526 L
.80506 .25235 L
.80754 .25214 L
.81002 .25198 L
Mistroke
.81126 .25191 L
.8125 .25186 L
.81374 .25182 L
.81498 .25179 L
.81622 .25177 L
.81746 .25177 L
.8187 .25177 L
.81994 .25179 L
.82118 .25182 L
.82242 .25186 L
.82366 .25192 L
.8249 .25198 L
.82738 .25216 L
.82986 .25238 L
.83234 .25266 L
.8373 .25336 L
.84226 .25429 L
.84722 .25543 L
.85714 .25841 L
.86706 .26235 L
.87698 .2673 L
.89683 .2804 L
.91667 .29811 L
.93651 .32081 L
.97619 .38269 L
Mfstroke
P
P
p
p
.004 w
s
s
s
s
s
s
s
s
s
s
s
.45687 .61803 m
.46032 .60993 L
s
.46032 .60993 m
.5 .52997 L
.53968 .4623 L
.57937 .40587 L
.61905 .35967 L
.65873 .32268 L
.69841 .29387 L
.7381 .27222 L
.75794 .26376 L
.77778 .25671 L
.79762 .25093 L
.81746 .24631 L
.8373 .24271 L
.85714 .24 L
.86706 .23895 L
.87698 .23806 L
.8869 .23734 L
.89683 .23676 L
.90675 .23631 L
.91667 .23598 L
.92163 .23584 L
.92659 .23573 L
.93155 .23564 L
.93651 .23557 L
.94147 .23551 L
.94643 .23547 L
.94891 .23546 L
.95139 .23544 L
.95387 .23543 L
.95635 .23542 L
.95883 .23541 L
.96131 .23541 L
.96379 .23541 L
.96503 .2354 L
.96627 .2354 L
.96751 .2354 L
.96875 .2354 L
.96999 .2354 L
.97123 .2354 L
.97247 .2354 L
.97371 .2354 L
.97495 .2354 L
.97619 .2354 L
s
P
P
P
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell[TextData[{
  "A ",
  StyleBox["B-spline basis",
    FontWeight->"Bold"],
  " is a row vector: "
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$\\phi^r(\\tau) := (\\phi_1^r(\\tau),\\cdots,\\phi_\\kappa^r(\\tau)). $$\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
For example, we can make a cubic B-spline basis and evaluate it at $\\tau = \
3/2$:\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
phi3 = MakeBSplineBasis[knots, 3]
Through[phi3[3/2]]\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
{BSplineFunction[{0, 1}, 
 
   Which[#1 < 0, 0, #1 <= 1, 
 
     InterpolatingFunction[{0, 1}, 
 
       {{0, 0, {1, 0}, {0}}, 
 
        {1, 0, {0, 0, 1, -1}, {0, 0, 1}}}][#1], True, 0] & ]
 
   , BSplineFunction[{0, 2}, 
 
   Which[#1 < 0, 0, #1 <= 2, 
 
     InterpolatingFunction[{0, 2}, 
 
       {{0, 0, {0, 0}, {0}}, 
 
        {1, 0, {1/4, -3/4, -1, 7/4}, {0, 0, 1}}, 
 
        {2, 1, {0, 0, 1/4, -1/4}, {0, 0, 1}}}][#1], True, 0]
 
      & ], BSplineFunction[{0, 3}, 
 
   Which[#1 < 0, 0, #1 <= 3, 
 
     InterpolatingFunction[{0, 3}, 
 
       {{0, 0, {0, 0}, {0}}, 
 
        {1, 0, {7/12, 1/4, -1/3, -11/12}, {0, 0, 1}}, 
 
        {2, 1, {1/6, -1/2, -1/12, 7/12}, {0, 0, 1}}, 
 
        {3, 2, {0, 0, 1/6, -1/6}, {0, 0, 1}}}][#1], True, 0]
 
      & ], BSplineFunction[{0, 4}, 
 
   Which[#1 < 0, 0, #1 <= 4, 
 
     InterpolatingFunction[{0, 4}, 
 
       {{0, 0, {0, 0}, {0}}, 
 
        {1, 0, {1/6, 1/2, 1/3, 1/6}, {0, 0, 1}}, 
 
        {2, 1, {2/3, 0, -1/2, -1/2}, {0, 0, 1}}, 
 
        {3, 2, {1/6, -1/2, 0, 1/2}, {0, 0, 1}}, 
 
        {4, 3, {0, 0, 1/6, -1/6}, {0, 0, 1}}}][#1], True, 0]
 
      & ], BSplineFunction[{1, 5}, 
 
   Which[#1 < 1, 0, #1 <= 5, 
 
     InterpolatingFunction[{1, 5}, 
 
       {{1, 1, {0, 0}, {0}}, 
 
        {2, 1, {1/6, 1/2, 1/3, 1/6}, {0, 0, 1}}, 
 
        {3, 2, {2/3, 0, -1/2, -1/2}, {0, 0, 1}}, 
 
        {4, 3, {1/6, -1/2, 0, 1/2}, {0, 0, 1}}, 
 
        {5, 4, {0, 0, 1/6, -1/6}, {0, 0, 1}}}][#1], True, 0]
 
      & ], BSplineFunction[{2, 5}, 
 
   Which[#1 < 2, 0, #1 <= 5, 
 
     InterpolatingFunction[{2, 5}, 
 
       {{2, 2, {0, 0}, {0}}, 
 
        {3, 2, {1/6, 1/2, 1/3, 1/6}, {0, 0, 1}}, 
 
        {4, 3, {7/12, -1/4, -2/3, -7/12}, {0, 0, 1}}, 
 
        {5, 4, {0, 0, 7/12, 11/12}, {0, 0, 1}}}][#1], True, 
 
     0] & ], BSplineFunction[{3, 5}, 
 
   Which[#1 < 3, 0, #1 <= 5, 
 
     InterpolatingFunction[{3, 5}, 
 
       {{3, 3, {0, 0}, {0}}, 
 
        {4, 3, {1/4, 3/4, 1/2, 1/4}, {0, 0, 1}}, 
 
        {5, 4, {0, -3, -11/4, -7/4}, {0, 0, 1}}}][#1], 
 
     True, 0] & ], BSplineFunction[{4, 5}, 
 
   Which[#1 < 4, 0, #1 <= 5, 
 
     InterpolatingFunction[{4, 5}, 
 
       {{4, 4, {0, 0}, {0}}, 
 
        {5, 4, {1, 3, 2, 1}, {0, 0, 1}}}][#1], True, 0] & ]}\
\>", "\<\
{BSplineFunction[{0, 1}, <>], BSplineFunction[{0, 2}, <>], 
 
  BSplineFunction[{0, 3}, <>], BSplineFunction[{0, 4}, <>], 
 
  BSplineFunction[{1, 5}, <>], BSplineFunction[{2, 5}, <>], 
 
  BSplineFunction[{3, 5}, <>], BSplineFunction[{4, 5}, <>]}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
{0, 1/32, 15/32, 23/48, 1/48, 0, 0, 0}\
\>", "\<\
    1   15  23  1
{0, --, --, --, --, 0, 0, 0}
    32  32  48  48\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell["\<\
Over any interval between adjacent knot points, $s_k$ and $s_{k+1}$, there \
are $r+1$ non-zero B-splines, with adjacent intervals sharing $r$. This gives \
$\\phi^r(\\tau)$ a quasi-orthogonal structure from which it gets its \
numerical stability. 

Here we plot $\\phi^r(\\tau)$ for $r$ from 0 to 3 and $\\tau$ from 0 to 5:\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
bases = Table[MakeBSplineBasis[knots, i], {i, 0, 3}];
Show[GraphicsArray[Partition[
\tPlotBSplineBasis[#, 
\t\tDisplayFunction -> Identity]& /@ bases, 2],
\t\tDisplayFunction -> $DisplayFunction]];\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.47619 0.0147151 0.47619 [
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
% Start of sub-graphic
p
0.0238095 0.0147151 0.477324 0.295003 MathSubStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.190476 0.0147151 0.588604 [
[(1)] .21429 .01472 0 2 Msboxa
[(2)] .40476 .01472 0 2 Msboxa
[(3)] .59524 .01472 0 2 Msboxa
[(4)] .78571 .01472 0 2 Msboxa
[(5)] .97619 .01472 0 2 Msboxa
[(0.2)] .01131 .13244 1 0 Msboxa
[(0.4)] .01131 .25016 1 0 Msboxa
[(0.6)] .01131 .36788 1 0 Msboxa
[(0.8)] .01131 .4856 1 0 Msboxa
[(1)] .01131 .60332 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
p
p
.002 w
.21429 .01472 m
.21429 .02097 L
s
P
[(1)] .21429 .01472 0 2 Mshowa
p
.002 w
.40476 .01472 m
.40476 .02097 L
s
P
[(2)] .40476 .01472 0 2 Mshowa
p
.002 w
.59524 .01472 m
.59524 .02097 L
s
P
[(3)] .59524 .01472 0 2 Mshowa
p
.002 w
.78571 .01472 m
.78571 .02097 L
s
P
[(4)] .78571 .01472 0 2 Mshowa
p
.002 w
.97619 .01472 m
.97619 .02097 L
s
P
[(5)] .97619 .01472 0 2 Mshowa
p
.001 w
.0619 .01472 m
.0619 .01847 L
s
P
p
.001 w
.1 .01472 m
.1 .01847 L
s
P
p
.001 w
.1381 .01472 m
.1381 .01847 L
s
P
p
.001 w
.17619 .01472 m
.17619 .01847 L
s
P
p
.001 w
.25238 .01472 m
.25238 .01847 L
s
P
p
.001 w
.29048 .01472 m
.29048 .01847 L
s
P
p
.001 w
.32857 .01472 m
.32857 .01847 L
s
P
p
.001 w
.36667 .01472 m
.36667 .01847 L
s
P
p
.001 w
.44286 .01472 m
.44286 .01847 L
s
P
p
.001 w
.48095 .01472 m
.48095 .01847 L
s
P
p
.001 w
.51905 .01472 m
.51905 .01847 L
s
P
p
.001 w
.55714 .01472 m
.55714 .01847 L
s
P
p
.001 w
.63333 .01472 m
.63333 .01847 L
s
P
p
.001 w
.67143 .01472 m
.67143 .01847 L
s
P
p
.001 w
.70952 .01472 m
.70952 .01847 L
s
P
p
.001 w
.74762 .01472 m
.74762 .01847 L
s
P
p
.001 w
.82381 .01472 m
.82381 .01847 L
s
P
p
.001 w
.8619 .01472 m
.8619 .01847 L
s
P
p
.001 w
.9 .01472 m
.9 .01847 L
s
P
p
.001 w
.9381 .01472 m
.9381 .01847 L
s
P
p
.002 w
0 .01472 m
1 .01472 L
s
P
p
.002 w
.02381 .13244 m
.03006 .13244 L
s
P
[(0.2)] .01131 .13244 1 0 Mshowa
p
.002 w
.02381 .25016 m
.03006 .25016 L
s
P
[(0.4)] .01131 .25016 1 0 Mshowa
p
.002 w
.02381 .36788 m
.03006 .36788 L
s
P
[(0.6)] .01131 .36788 1 0 Mshowa
p
.002 w
.02381 .4856 m
.03006 .4856 L
s
P
[(0.8)] .01131 .4856 1 0 Mshowa
p
.002 w
.02381 .60332 m
.03006 .60332 L
s
P
[(1)] .01131 .60332 1 0 Mshowa
p
.001 w
.02381 .03826 m
.02756 .03826 L
s
P
p
.001 w
.02381 .0618 m
.02756 .0618 L
s
P
p
.001 w
.02381 .08535 m
.02756 .08535 L
s
P
p
.001 w
.02381 .10889 m
.02756 .10889 L
s
P
p
.001 w
.02381 .15598 m
.02756 .15598 L
s
P
p
.001 w
.02381 .17952 m
.02756 .17952 L
s
P
p
.001 w
.02381 .20307 m
.02756 .20307 L
s
P
p
.001 w
.02381 .22661 m
.02756 .22661 L
s
P
p
.001 w
.02381 .2737 m
.02756 .2737 L
s
P
p
.001 w
.02381 .29724 m
.02756 .29724 L
s
P
p
.001 w
.02381 .32079 m
.02756 .32079 L
s
P
p
.001 w
.02381 .34433 m
.02756 .34433 L
s
P
p
.001 w
.02381 .39142 m
.02756 .39142 L
s
P
p
.001 w
.02381 .41497 m
.02756 .41497 L
s
P
p
.001 w
.02381 .43851 m
.02756 .43851 L
s
P
p
.001 w
.02381 .46205 m
.02756 .46205 L
s
P
p
.001 w
.02381 .50914 m
.02756 .50914 L
s
P
p
.001 w
.02381 .53269 m
.02756 .53269 L
s
P
p
.001 w
.02381 .55623 m
.02756 .55623 L
s
P
p
.001 w
.02381 .57977 m
.02756 .57977 L
s
P
p
.002 w
.02381 0 m
.02381 .61803 L
s
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
p
.004 w
.02381 .60332 m
.06349 .38361 L
.08333 .29292 L
.10317 .215 L
.12302 .14986 L
.14286 .09749 L
.15278 .07609 L
.1627 .05789 L
.17262 .04288 L
.18254 .03107 L
.1875 .02635 L
.19246 .02244 L
.19742 .01933 L
.20238 .01701 L
.20486 .01616 L
.2061 .0158 L
.20734 .0155 L
.20858 .01524 L
.20982 .01504 L
.21106 .01488 L
.2123 .01478 L
.21354 .01472 L
.21478 .01472 L
.21726 .01472 L
.21974 .01472 L
.22222 .01472 L
.23214 .01472 L
.24206 .01472 L
.2619 .01472 L
.30159 .01472 L
.34127 .01472 L
.38095 .01472 L
.42063 .01472 L
.46032 .01472 L
.5 .01472 L
.53968 .01472 L
.57937 .01472 L
.61905 .01472 L
.65873 .01472 L
.69841 .01472 L
.7381 .01472 L
.77778 .01472 L
.81746 .01472 L
.85714 .01472 L
.89683 .01472 L
.93651 .01472 L
.97619 .01472 L
s
P
P
p
p
.004 w
.02381 .01472 m
.04365 .12776 L
.06349 .22165 L
.08333 .29637 L
.09325 .32655 L
.10317 .35194 L
.1131 .37253 L
.11806 .38104 L
.12302 .38834 L
.12798 .39445 L
.13294 .39936 L
.13542 .40136 L
.1379 .40307 L
.14038 .40448 L
.14286 .40558 L
.1441 .40603 L
.14534 .40639 L
.14658 .40669 L
.14782 .4069 L
.14906 .40704 L
.1503 .40711 L
.15154 .4071 L
.15278 .40702 L
.15402 .40686 L
.15526 .40663 L
.15774 .40594 L
.15898 .40549 L
.16022 .40496 L
.1627 .40367 L
.16766 .4002 L
.17262 .39553 L
.18254 .38259 L
.19246 .36487 L
.20238 .34236 L
.22222 .285 L
.2619 .18026 L
.28175 .13747 L
.30159 .10106 L
.32143 .07105 L
.34127 .04742 L
.36111 .03017 L
.37103 .02394 L
.37599 .02143 L
.38095 .01931 L
.38591 .0176 L
.39087 .01628 L
.39335 .01577 L
.39583 .01536 L
.39831 .01505 L
.39955 .01494 L
Mistroke
.40079 .01484 L
.40203 .01478 L
.40327 .01473 L
.40451 .01472 L
.40575 .01472 L
.40699 .01472 L
.40823 .01472 L
.41071 .01472 L
.42063 .01472 L
.43056 .01472 L
.44048 .01472 L
.46032 .01472 L
.5 .01472 L
.53968 .01472 L
.57937 .01472 L
.61905 .01472 L
.65873 .01472 L
.69841 .01472 L
.7381 .01472 L
.77778 .01472 L
.81746 .01472 L
.85714 .01472 L
.89683 .01472 L
.93651 .01472 L
.97619 .01472 L
Mfstroke
P
P
p
p
.004 w
.02381 .01472 m
.02505 .01473 L
.02629 .01476 L
.02753 .01483 L
.02877 .01491 L
.03125 .01516 L
.03373 .01551 L
.03621 .01596 L
.03869 .01651 L
.04365 .01791 L
.04861 .0197 L
.05357 .0219 L
.06349 .02749 L
.07341 .03467 L
.08333 .04346 L
.10317 .06581 L
.12302 .09455 L
.14286 .12968 L
.18254 .21909 L
.22222 .33252 L
.24206 .38234 L
.25198 .40246 L
.2619 .41938 L
.27183 .43311 L
.28175 .44365 L
.28671 .44772 L
.29167 .45099 L
.29663 .45347 L
.29911 .45441 L
.30159 .45515 L
.30407 .45568 L
.30531 .45588 L
.30655 .45602 L
.30779 .45612 L
.30903 .45616 L
.31027 .45616 L
.31151 .4561 L
.31275 .456 L
.31399 .45584 L
.31523 .45564 L
.31647 .45539 L
.31895 .45473 L
.32143 .45387 L
.32639 .45155 L
.33135 .44844 L
.34127 .43982 L
.35119 .428 L
.36111 .41299 L
.38095 .3734 L
.42063 .26201 L
Mistroke
.46032 .16238 L
.48016 .12214 L
.5 .08829 L
.51984 .06083 L
.52976 .04949 L
.53968 .03975 L
.5496 .03161 L
.55952 .02506 L
.56448 .02239 L
.56944 .02011 L
.5744 .01824 L
.57937 .01676 L
.58185 .01617 L
.58433 .01568 L
.58681 .01529 L
.58805 .01513 L
.58929 .015 L
.59053 .0149 L
.59177 .01481 L
.59301 .01476 L
.59425 .01472 L
.59549 .01472 L
.59673 .01472 L
.59921 .01472 L
.61905 .01472 L
.65873 .01472 L
.69841 .01472 L
.7381 .01472 L
.77778 .01472 L
.81746 .01472 L
.85714 .01472 L
.89683 .01472 L
.93651 .01472 L
.97619 .01472 L
Mfstroke
P
P
p
p
.004 w
.02381 .01472 m
.06349 .01472 L
.10317 .01472 L
.14286 .01472 L
.1627 .01472 L
.18254 .01472 L
.20238 .01472 L
.20734 .01472 L
.20982 .01472 L
.21106 .01472 L
.2123 .01472 L
.21354 .01472 L
.21478 .01472 L
.21602 .01474 L
.21726 .01479 L
.2185 .01486 L
.21974 .01496 L
.22222 .01523 L
.2247 .0156 L
.22718 .01606 L
.23214 .0173 L
.2371 .01894 L
.24206 .02097 L
.25198 .02624 L
.2619 .03311 L
.28175 .05163 L
.30159 .07654 L
.34127 .14552 L
.38095 .24004 L
.42063 .35398 L
.44048 .39869 L
.4504 .41625 L
.46032 .43062 L
.47024 .4418 L
.4752 .44619 L
.48016 .44978 L
.48512 .45258 L
.4876 .45367 L
.49008 .45457 L
.49256 .45527 L
.4938 .45554 L
.49504 .45577 L
.49628 .45594 L
.49752 .45607 L
.49876 .45614 L
.5 .45617 L
.50124 .45614 L
.50248 .45607 L
.50372 .45594 L
.50496 .45577 L
Mistroke
.50744 .45527 L
.50992 .45457 L
.5124 .45367 L
.51488 .45258 L
.51984 .44978 L
.5248 .44619 L
.52976 .4418 L
.53968 .43062 L
.5496 .41625 L
.55952 .39869 L
.57937 .35398 L
.61905 .24004 L
.65873 .14552 L
.67857 .10783 L
.69841 .07654 L
.71825 .05163 L
.72817 .04157 L
.7381 .03311 L
.74802 .02624 L
.75298 .02341 L
.75794 .02097 L
.7629 .01894 L
.76786 .0173 L
.77034 .01663 L
.77282 .01606 L
.7753 .0156 L
.77778 .01523 L
.77902 .01508 L
.78026 .01496 L
.7815 .01486 L
.78274 .01479 L
.78398 .01474 L
.78522 .01472 L
.78646 .01472 L
.7877 .01472 L
.79018 .01472 L
.79266 .01472 L
.79762 .01472 L
.81746 .01472 L
.85714 .01472 L
.89683 .01472 L
.93651 .01472 L
.97619 .01472 L
Mfstroke
P
P
p
p
.004 w
.02381 .01472 m
.06349 .01472 L
.10317 .01472 L
.14286 .01472 L
.18254 .01472 L
.22222 .01472 L
.2619 .01472 L
.30159 .01472 L
.34127 .01472 L
.36111 .01472 L
.38095 .01472 L
.39087 .01472 L
.39583 .01472 L
.39831 .01472 L
.40079 .01472 L
.40203 .01472 L
.40327 .01472 L
.40451 .01472 L
.40575 .01472 L
.40699 .01476 L
.40823 .01481 L
.40947 .0149 L
.41071 .015 L
.41319 .01529 L
.41567 .01568 L
.42063 .01676 L
.4256 .01824 L
.43056 .02011 L
.44048 .02506 L
.4504 .03161 L
.46032 .03975 L
.48016 .06083 L
.5 .08829 L
.53968 .16238 L
.57937 .26201 L
.59921 .32102 L
.61905 .3734 L
.63889 .41299 L
.64881 .428 L
.65873 .43982 L
.66369 .44453 L
.66865 .44844 L
.67361 .45155 L
.67609 .45281 L
.67857 .45387 L
.68105 .45473 L
.68229 .45508 L
.68353 .45539 L
.68477 .45564 L
.68601 .45584 L
Mistroke
.68725 .456 L
.68849 .4561 L
.68973 .45616 L
.69097 .45616 L
.69221 .45612 L
.69345 .45602 L
.69469 .45588 L
.69593 .45568 L
.69717 .45544 L
.69841 .45515 L
.70337 .45347 L
.70585 .45233 L
.70833 .45099 L
.71825 .44365 L
.72321 .43878 L
.72817 .43311 L
.7381 .41938 L
.75794 .38234 L
.77778 .33252 L
.81746 .21909 L
.85714 .12968 L
.87698 .09455 L
.89683 .06581 L
.91667 .04346 L
.92659 .03467 L
.93651 .02749 L
.94643 .0219 L
.95139 .0197 L
.95635 .01791 L
.96131 .01651 L
.96379 .01596 L
.96627 .01551 L
.96875 .01516 L
.96999 .01503 L
.97123 .01491 L
.97247 .01483 L
.97371 .01476 L
.97495 .01473 L
.97619 .01472 L
Mfstroke
P
P
p
p
.004 w
.02381 .01472 m
.06349 .01472 L
.10317 .01472 L
.14286 .01472 L
.18254 .01472 L
.22222 .01472 L
.2619 .01472 L
.30159 .01472 L
.34127 .01472 L
.38095 .01472 L
.42063 .01472 L
.46032 .01472 L
.5 .01472 L
.53968 .01472 L
.55952 .01472 L
.56944 .01472 L
.57937 .01472 L
.58433 .01472 L
.58929 .01472 L
.59177 .01472 L
.59301 .01472 L
.59425 .01472 L
.59549 .01472 L
.59673 .01473 L
.59797 .01478 L
.59921 .01484 L
.60169 .01505 L
.60293 .01519 L
.60417 .01536 L
.60913 .01628 L
.61161 .01689 L
.61409 .0176 L
.61905 .01931 L
.62897 .02394 L
.63889 .03017 L
.65873 .04742 L
.67857 .07105 L
.69841 .10106 L
.7381 .18026 L
.77778 .285 L
.79762 .34236 L
.80754 .36487 L
.81746 .38259 L
.82242 .38966 L
.82738 .39553 L
.83234 .4002 L
.8373 .40367 L
.83978 .40496 L
.84102 .40549 L
.84226 .40594 L
Mistroke
.8435 .40633 L
.84474 .40663 L
.84598 .40686 L
.84722 .40702 L
.84846 .4071 L
.8497 .40711 L
.85094 .40704 L
.85218 .4069 L
.85342 .40669 L
.85466 .40639 L
.85714 .40558 L
.85962 .40448 L
.8621 .40307 L
.86706 .39936 L
.87202 .39445 L
.87698 .38834 L
.8869 .37253 L
.89683 .35194 L
.91667 .29637 L
.93651 .22165 L
.97619 .01472 L
Mfstroke
P
P
p
p
.004 w
.02381 .01472 m
.06349 .01472 L
.10317 .01472 L
.14286 .01472 L
.18254 .01472 L
.22222 .01472 L
.2619 .01472 L
.30159 .01472 L
.34127 .01472 L
.38095 .01472 L
.42063 .01472 L
.46032 .01472 L
.5 .01472 L
.53968 .01472 L
.57937 .01472 L
.61905 .01472 L
.65873 .01472 L
.69841 .01472 L
.7381 .01472 L
.75794 .01472 L
.76786 .01472 L
.77282 .01472 L
.77778 .01472 L
.78274 .01472 L
.78398 .01472 L
.78522 .01472 L
.78646 .01472 L
.7877 .01478 L
.78894 .01488 L
.79018 .01504 L
.79266 .0155 L
.7939 .0158 L
.79514 .01616 L
.79762 .01701 L
.80258 .01933 L
.80754 .02244 L
.81746 .03107 L
.82738 .04288 L
.8373 .05789 L
.85714 .09749 L
.87698 .14986 L
.89683 .215 L
.93651 .38361 L
.97619 .60332 L
s
P
P
P
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.522676 0.0147151 0.97619 0.295003 MathSubStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.190476 0.0147151 0.588604 [
[(1)] .21429 .01472 0 2 Msboxa
[(2)] .40476 .01472 0 2 Msboxa
[(3)] .59524 .01472 0 2 Msboxa
[(4)] .78571 .01472 0 2 Msboxa
[(5)] .97619 .01472 0 2 Msboxa
[(0.2)] .01131 .13244 1 0 Msboxa
[(0.4)] .01131 .25016 1 0 Msboxa
[(0.6)] .01131 .36788 1 0 Msboxa
[(0.8)] .01131 .4856 1 0 Msboxa
[(1)] .01131 .60332 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
p
p
.002 w
.21429 .01472 m
.21429 .02097 L
s
P
[(1)] .21429 .01472 0 2 Mshowa
p
.002 w
.40476 .01472 m
.40476 .02097 L
s
P
[(2)] .40476 .01472 0 2 Mshowa
p
.002 w
.59524 .01472 m
.59524 .02097 L
s
P
[(3)] .59524 .01472 0 2 Mshowa
p
.002 w
.78571 .01472 m
.78571 .02097 L
s
P
[(4)] .78571 .01472 0 2 Mshowa
p
.002 w
.97619 .01472 m
.97619 .02097 L
s
P
[(5)] .97619 .01472 0 2 Mshowa
p
.001 w
.0619 .01472 m
.0619 .01847 L
s
P
p
.001 w
.1 .01472 m
.1 .01847 L
s
P
p
.001 w
.1381 .01472 m
.1381 .01847 L
s
P
p
.001 w
.17619 .01472 m
.17619 .01847 L
s
P
p
.001 w
.25238 .01472 m
.25238 .01847 L
s
P
p
.001 w
.29048 .01472 m
.29048 .01847 L
s
P
p
.001 w
.32857 .01472 m
.32857 .01847 L
s
P
p
.001 w
.36667 .01472 m
.36667 .01847 L
s
P
p
.001 w
.44286 .01472 m
.44286 .01847 L
s
P
p
.001 w
.48095 .01472 m
.48095 .01847 L
s
P
p
.001 w
.51905 .01472 m
.51905 .01847 L
s
P
p
.001 w
.55714 .01472 m
.55714 .01847 L
s
P
p
.001 w
.63333 .01472 m
.63333 .01847 L
s
P
p
.001 w
.67143 .01472 m
.67143 .01847 L
s
P
p
.001 w
.70952 .01472 m
.70952 .01847 L
s
P
p
.001 w
.74762 .01472 m
.74762 .01847 L
s
P
p
.001 w
.82381 .01472 m
.82381 .01847 L
s
P
p
.001 w
.8619 .01472 m
.8619 .01847 L
s
P
p
.001 w
.9 .01472 m
.9 .01847 L
s
P
p
.001 w
.9381 .01472 m
.9381 .01847 L
s
P
p
.002 w
0 .01472 m
1 .01472 L
s
P
p
.002 w
.02381 .13244 m
.03006 .13244 L
s
P
[(0.2)] .01131 .13244 1 0 Mshowa
p
.002 w
.02381 .25016 m
.03006 .25016 L
s
P
[(0.4)] .01131 .25016 1 0 Mshowa
p
.002 w
.02381 .36788 m
.03006 .36788 L
s
P
[(0.6)] .01131 .36788 1 0 Mshowa
p
.002 w
.02381 .4856 m
.03006 .4856 L
s
P
[(0.8)] .01131 .4856 1 0 Mshowa
p
.002 w
.02381 .60332 m
.03006 .60332 L
s
P
[(1)] .01131 .60332 1 0 Mshowa
p
.001 w
.02381 .03826 m
.02756 .03826 L
s
P
p
.001 w
.02381 .0618 m
.02756 .0618 L
s
P
p
.001 w
.02381 .08535 m
.02756 .08535 L
s
P
p
.001 w
.02381 .10889 m
.02756 .10889 L
s
P
p
.001 w
.02381 .15598 m
.02756 .15598 L
s
P
p
.001 w
.02381 .17952 m
.02756 .17952 L
s
P
p
.001 w
.02381 .20307 m
.02756 .20307 L
s
P
p
.001 w
.02381 .22661 m
.02756 .22661 L
s
P
p
.001 w
.02381 .2737 m
.02756 .2737 L
s
P
p
.001 w
.02381 .29724 m
.02756 .29724 L
s
P
p
.001 w
.02381 .32079 m
.02756 .32079 L
s
P
p
.001 w
.02381 .34433 m
.02756 .34433 L
s
P
p
.001 w
.02381 .39142 m
.02756 .39142 L
s
P
p
.001 w
.02381 .41497 m
.02756 .41497 L
s
P
p
.001 w
.02381 .43851 m
.02756 .43851 L
s
P
p
.001 w
.02381 .46205 m
.02756 .46205 L
s
P
p
.001 w
.02381 .50914 m
.02756 .50914 L
s
P
p
.001 w
.02381 .53269 m
.02756 .53269 L
s
P
p
.001 w
.02381 .55623 m
.02756 .55623 L
s
P
p
.001 w
.02381 .57977 m
.02756 .57977 L
s
P
p
.002 w
.02381 0 m
.02381 .61803 L
s
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
p
.004 w
.02381 .60332 m
.04365 .43788 L
.06349 .30676 L
.08333 .20598 L
.09325 .16572 L
.10317 .13155 L
.1131 .10297 L
.12302 .07947 L
.13294 .06057 L
.14286 .04575 L
.15278 .03453 L
.15774 .03012 L
.1627 .02641 L
.17262 .02088 L
.17758 .01893 L
.18254 .01744 L
.1875 .01635 L
.18998 .01594 L
.19246 .0156 L
.19494 .01533 L
.19742 .01512 L
.1999 .01497 L
.20238 .01486 L
.20362 .01482 L
.20486 .01479 L
.2061 .01476 L
.20734 .01474 L
.20858 .01473 L
.20982 .01472 L
.21106 .01472 L
.2123 .01472 L
.21354 .01472 L
.21478 .01472 L
.21726 .01472 L
.21974 .01472 L
.22222 .01472 L
.23214 .01472 L
.24206 .01472 L
.2619 .01472 L
.30159 .01472 L
.34127 .01472 L
.38095 .01472 L
.42063 .01472 L
.46032 .01472 L
.5 .01472 L
.53968 .01472 L
.57937 .01472 L
.61905 .01472 L
.65873 .01472 L
.69841 .01472 L
Mistroke
.7381 .01472 L
.77778 .01472 L
.81746 .01472 L
.85714 .01472 L
.89683 .01472 L
.93651 .01472 L
.97619 .01472 L
Mfstroke
P
P
p
p
.004 w
.02381 .01472 m
.03373 .09964 L
.04365 .17108 L
.05357 .22989 L
.06349 .27694 L
.07341 .31313 L
.08333 .3393 L
.08829 .34891 L
.09325 .35635 L
.09821 .36172 L
.10069 .36366 L
.10193 .36446 L
.10317 .36514 L
.10441 .3657 L
.10565 .36614 L
.10689 .36648 L
.10813 .36671 L
.10938 .36682 L
.11062 .36683 L
.11186 .36674 L
.1131 .36654 L
.11434 .36624 L
.11558 .36584 L
.11806 .36474 L
.12054 .36327 L
.12302 .36143 L
.12798 .35671 L
.13294 .35068 L
.14286 .33517 L
.1627 .29334 L
.18254 .24294 L
.20238 .19093 L
.22222 .14423 L
.24206 .10642 L
.2619 .07679 L
.28175 .05435 L
.29167 .04552 L
.30159 .0381 L
.31151 .03198 L
.32143 .02704 L
.33135 .02314 L
.34127 .02017 L
.34623 .01898 L
.35119 .01799 L
.36111 .01649 L
.36607 .01595 L
.37103 .01553 L
.37599 .01522 L
.37847 .0151 L
.38095 .015 L
Mistroke
.38343 .01492 L
.38591 .01486 L
.38839 .01481 L
.39087 .01477 L
.39211 .01476 L
.39335 .01475 L
.39459 .01474 L
.39583 .01473 L
.39707 .01472 L
.39831 .01472 L
.39955 .01472 L
.40079 .01472 L
.40203 .01472 L
.40327 .01472 L
.40451 .01472 L
.40575 .01472 L
.40699 .01472 L
.40823 .01472 L
.41071 .01472 L
.42063 .01472 L
.43056 .01472 L
.44048 .01472 L
.46032 .01472 L
.5 .01472 L
.53968 .01472 L
.57937 .01472 L
.61905 .01472 L
.65873 .01472 L
.69841 .01472 L
.7381 .01472 L
.77778 .01472 L
.81746 .01472 L
.85714 .01472 L
.89683 .01472 L
.93651 .01472 L
.97619 .01472 L
Mfstroke
P
P
p
p
.004 w
.02381 .01472 m
.02505 .01475 L
.02629 .01486 L
.02753 .01505 L
.02877 .0153 L
.03125 .01603 L
.03373 .01703 L
.03869 .01985 L
.04365 .02369 L
.05357 .03421 L
.06349 .04816 L
.08333 .08447 L
.10317 .12897 L
.14286 .22787 L
.1627 .27496 L
.18254 .3156 L
.19246 .33236 L
.20238 .34613 L
.20734 .35175 L
.2123 .35646 L
.21726 .36019 L
.21974 .36169 L
.22222 .36295 L
.2247 .36397 L
.22718 .36476 L
.22842 .36508 L
.22966 .36533 L
.2309 .36553 L
.23214 .36568 L
.23338 .36577 L
.23462 .36581 L
.23586 .36579 L
.2371 .36573 L
.23834 .36561 L
.23958 .36544 L
.24082 .36522 L
.24206 .36494 L
.24702 .36337 L
.2495 .36229 L
.25198 .36103 L
.2619 .35424 L
.27183 .34484 L
.28175 .33315 L
.30159 .30401 L
.34127 .2309 L
.38095 .15353 L
.42063 .09028 L
.44048 .06733 L
.46032 .04958 L
.48016 .03635 L
Mistroke
.49008 .03122 L
.5 .02698 L
.50992 .02353 L
.51984 .0208 L
.52976 .0187 L
.53472 .01786 L
.53968 .01715 L
.54464 .01655 L
.5496 .01606 L
.55456 .01567 L
.55952 .01536 L
.56448 .01513 L
.56696 .01504 L
.56944 .01496 L
.5744 .01484 L
.57688 .0148 L
.57937 .01477 L
.58061 .01476 L
.58185 .01475 L
.58309 .01474 L
.58433 .01473 L
.58557 .01473 L
.58681 .01472 L
.58805 .01472 L
.58929 .01472 L
.59053 .01472 L
.59177 .01472 L
.59301 .01472 L
.59425 .01472 L
.59549 .01472 L
.59673 .01472 L
.59921 .01472 L
.60913 .01472 L
.61905 .01472 L
.65873 .01472 L
.69841 .01472 L
.7381 .01472 L
.77778 .01472 L
.81746 .01472 L
.85714 .01472 L
.89683 .01472 L
.93651 .01472 L
.97619 .01472 L
Mfstroke
P
P
p
p
.004 w
.02381 .01472 m
.02505 .01472 L
.02629 .01472 L
.02753 .01472 L
.02877 .01472 L
.03001 .01472 L
.03125 .01472 L
.03249 .01472 L
.03373 .01473 L
.03497 .01473 L
.03621 .01474 L
.03869 .01476 L
.04117 .01479 L
.04365 .01483 L
.04613 .01487 L
.04861 .01493 L
.05357 .01509 L
.05853 .01531 L
.06349 .0156 L
.06845 .01598 L
.07341 .01645 L
.08333 .01771 L
.09325 .01947 L
.10317 .02181 L
.1131 .02482 L
.12302 .02858 L
.14286 .03867 L
.1627 .05275 L
.18254 .07149 L
.20238 .09555 L
.22222 .12557 L
.2619 .20019 L
.30159 .28119 L
.32143 .3191 L
.34127 .35262 L
.36111 .37975 L
.37103 .39029 L
.37599 .3947 L
.38095 .3985 L
.38591 .40164 L
.39087 .4041 L
.39335 .40507 L
.39583 .40585 L
.39831 .40645 L
.39955 .40668 L
.40079 .40686 L
.40203 .407 L
.40327 .40708 L
.40451 .40712 L
.40575 .4071 L
Mistroke
.40699 .40704 L
.40823 .40692 L
.40947 .40676 L
.41071 .40655 L
.41319 .40599 L
.41567 .40524 L
.42063 .4032 L
.4256 .40046 L
.43056 .39705 L
.44048 .38836 L
.46032 .36435 L
.5 .29675 L
.53968 .21639 L
.57937 .13921 L
.59921 .10681 L
.61905 .08043 L
.63889 .05965 L
.65873 .04378 L
.66865 .03749 L
.67857 .03217 L
.68849 .02776 L
.69841 .02416 L
.70833 .02129 L
.71825 .01907 L
.72817 .01742 L
.73313 .01678 L
.7381 .01625 L
.74306 .01582 L
.74802 .01548 L
.75298 .01521 L
.75546 .01511 L
.75794 .01502 L
.76042 .01494 L
.7629 .01488 L
.76538 .01483 L
.76786 .0148 L
.77034 .01477 L
.77158 .01476 L
.77282 .01475 L
.7753 .01473 L
.77654 .01473 L
.77778 .01472 L
.77902 .01472 L
.78026 .01472 L
.7815 .01472 L
.78274 .01472 L
.78398 .01472 L
.78522 .01472 L
.78646 .01472 L
.7877 .01472 L
Mistroke
.79018 .01472 L
.79266 .01472 L
.79762 .01472 L
.80754 .01472 L
.81746 .01472 L
.85714 .01472 L
.89683 .01472 L
.93651 .01472 L
.97619 .01472 L
Mfstroke
P
P
p
p
.004 w
.02381 .01472 m
.06349 .01472 L
.10317 .01472 L
.14286 .01472 L
.1627 .01472 L
.18254 .01472 L
.20238 .01472 L
.20734 .01472 L
.20982 .01472 L
.21106 .01472 L
.2123 .01472 L
.21354 .01472 L
.21478 .01472 L
.21602 .01472 L
.21726 .01472 L
.2185 .01472 L
.21974 .01472 L
.22098 .01472 L
.22222 .01472 L
.22346 .01473 L
.2247 .01473 L
.22594 .01474 L
.22718 .01475 L
.22966 .01477 L
.23214 .0148 L
.23462 .01483 L
.2371 .01488 L
.23958 .01494 L
.24206 .01502 L
.24702 .01521 L
.25198 .01548 L
.25694 .01582 L
.2619 .01625 L
.27183 .01742 L
.28175 .01907 L
.29167 .02129 L
.30159 .02416 L
.31151 .02776 L
.32143 .03217 L
.34127 .04378 L
.36111 .05965 L
.38095 .08043 L
.42063 .13921 L
.46032 .21639 L
.5 .29675 L
.51984 .33315 L
.53968 .36435 L
.55952 .38836 L
.56944 .39705 L
.5744 .40046 L
Mistroke
.57937 .4032 L
.58185 .40431 L
.58433 .40524 L
.58681 .40599 L
.58805 .40629 L
.58929 .40655 L
.59053 .40676 L
.59177 .40692 L
.59301 .40704 L
.59425 .4071 L
.59549 .40712 L
.59673 .40708 L
.59797 .407 L
.59921 .40686 L
.60169 .40645 L
.60293 .40618 L
.60417 .40585 L
.60913 .4041 L
.61409 .40164 L
.61905 .3985 L
.62897 .39029 L
.63889 .37975 L
.65873 .35262 L
.69841 .28119 L
.7381 .20019 L
.77778 .12557 L
.79762 .09555 L
.81746 .07149 L
.8373 .05275 L
.85714 .03867 L
.86706 .03316 L
.87698 .02858 L
.8869 .02482 L
.89683 .02181 L
.90675 .01947 L
.91667 .01771 L
.92163 .01702 L
.92659 .01645 L
.93155 .01598 L
.93651 .0156 L
.94147 .01531 L
.94643 .01509 L
.94891 .015 L
.95139 .01493 L
.95387 .01487 L
.95635 .01483 L
.95883 .01479 L
.96131 .01476 L
.96255 .01475 L
.96379 .01474 L
Mistroke
.96503 .01473 L
.96627 .01473 L
.96751 .01472 L
.96875 .01472 L
.96999 .01472 L
.97123 .01472 L
.97247 .01472 L
.97371 .01472 L
.97495 .01472 L
.97619 .01472 L
Mfstroke
P
P
p
p
.004 w
.02381 .01472 m
.06349 .01472 L
.10317 .01472 L
.14286 .01472 L
.18254 .01472 L
.22222 .01472 L
.2619 .01472 L
.30159 .01472 L
.34127 .01472 L
.36111 .01472 L
.38095 .01472 L
.39087 .01472 L
.39583 .01472 L
.39831 .01472 L
.40079 .01472 L
.40203 .01472 L
.40327 .01472 L
.40451 .01472 L
.40575 .01472 L
.40699 .01472 L
.40823 .01472 L
.40947 .01472 L
.41071 .01472 L
.41195 .01472 L
.41319 .01472 L
.41443 .01473 L
.41567 .01473 L
.41691 .01474 L
.41815 .01475 L
.42063 .01477 L
.42312 .0148 L
.4256 .01484 L
.42808 .01489 L
.43056 .01496 L
.43552 .01513 L
.44048 .01536 L
.44544 .01567 L
.4504 .01606 L
.45536 .01655 L
.46032 .01715 L
.47024 .0187 L
.48016 .0208 L
.49008 .02353 L
.5 .02698 L
.51984 .03635 L
.53968 .04958 L
.55952 .06733 L
.57937 .09028 L
.61905 .15353 L
.65873 .2309 L
Mistroke
.67857 .26915 L
.69841 .30401 L
.71825 .33315 L
.72817 .34484 L
.7381 .35424 L
.74306 .35798 L
.74802 .36103 L
.75298 .36337 L
.75546 .36425 L
.75794 .36494 L
.75918 .36522 L
.76042 .36544 L
.76166 .36561 L
.7629 .36573 L
.76414 .36579 L
.76538 .36581 L
.76662 .36577 L
.76786 .36568 L
.7691 .36553 L
.77034 .36533 L
.77282 .36476 L
.7753 .36397 L
.77778 .36295 L
.78274 .36019 L
.7877 .35646 L
.79762 .34613 L
.80754 .33236 L
.81746 .3156 L
.85714 .22787 L
.89683 .12897 L
.91667 .08447 L
.92659 .06506 L
.93651 .04816 L
.94643 .03421 L
.95139 .02849 L
.95635 .02369 L
.96131 .01985 L
.96379 .01831 L
.96627 .01703 L
.96875 .01603 L
.96999 .01563 L
.97123 .0153 L
.97247 .01505 L
.97371 .01486 L
.97495 .01475 L
.97619 .01472 L
Mfstroke
P
P
p
p
.004 w
.02381 .01472 m
.06349 .01472 L
.10317 .01472 L
.14286 .01472 L
.18254 .01472 L
.22222 .01472 L
.2619 .01472 L
.30159 .01472 L
.34127 .01472 L
.38095 .01472 L
.42063 .01472 L
.46032 .01472 L
.5 .01472 L
.53968 .01472 L
.55952 .01472 L
.56944 .01472 L
.57937 .01472 L
.58433 .01472 L
.58929 .01472 L
.59177 .01472 L
.59301 .01472 L
.59425 .01472 L
.59549 .01472 L
.59673 .01472 L
.59797 .01472 L
.59921 .01472 L
.60045 .01472 L
.60169 .01472 L
.60293 .01472 L
.60417 .01473 L
.60541 .01474 L
.60665 .01475 L
.60913 .01477 L
.61037 .01479 L
.61161 .01481 L
.61409 .01486 L
.61657 .01492 L
.61905 .015 L
.62401 .01522 L
.62897 .01553 L
.63393 .01595 L
.63889 .01649 L
.64385 .01716 L
.64881 .01799 L
.65873 .02017 L
.66865 .02314 L
.67857 .02704 L
.68849 .03198 L
.69841 .0381 L
.71825 .05435 L
Mistroke
.7381 .07679 L
.75794 .10642 L
.77778 .14423 L
.81746 .24294 L
.8373 .29334 L
.84722 .31577 L
.85714 .33517 L
.86706 .35068 L
.87202 .35671 L
.8745 .35924 L
.87698 .36143 L
.87946 .36327 L
.88194 .36474 L
.88318 .36534 L
.88442 .36584 L
.88566 .36624 L
.8869 .36654 L
.88814 .36674 L
.88938 .36683 L
.89063 .36682 L
.89187 .36671 L
.89311 .36648 L
.89435 .36614 L
.89559 .3657 L
.89683 .36514 L
.89931 .36366 L
.90179 .36172 L
.90675 .35635 L
.91171 .34891 L
.91667 .3393 L
.92659 .31313 L
.93651 .27694 L
.94643 .22989 L
.95635 .17108 L
.97619 .01472 L
Mfstroke
P
P
p
p
.004 w
.02381 .01472 m
.06349 .01472 L
.10317 .01472 L
.14286 .01472 L
.18254 .01472 L
.22222 .01472 L
.2619 .01472 L
.30159 .01472 L
.34127 .01472 L
.38095 .01472 L
.42063 .01472 L
.46032 .01472 L
.5 .01472 L
.53968 .01472 L
.57937 .01472 L
.61905 .01472 L
.65873 .01472 L
.69841 .01472 L
.7381 .01472 L
.75794 .01472 L
.76786 .01472 L
.77282 .01472 L
.77778 .01472 L
.78274 .01472 L
.78398 .01472 L
.78522 .01472 L
.78646 .01472 L
.7877 .01472 L
.78894 .01472 L
.79018 .01472 L
.79142 .01473 L
.79266 .01474 L
.7939 .01476 L
.79514 .01479 L
.79638 .01482 L
.79762 .01486 L
.8001 .01497 L
.80258 .01512 L
.80506 .01533 L
.80754 .0156 L
.81002 .01594 L
.8125 .01635 L
.81746 .01744 L
.82242 .01893 L
.82738 .02088 L
.8373 .02641 L
.84226 .03012 L
.84722 .03453 L
.85714 .04575 L
.86706 .06057 L
Mistroke
.87698 .07947 L
.8869 .10297 L
.89683 .13155 L
.91667 .20598 L
.93651 .30676 L
.95635 .43788 L
.97619 .60332 L
Mfstroke
P
P
P
MathSubEnd
P
% End of sub-graphic
P
p
% Start of sub-graphic
p
0.0238095 0.323031 0.477324 0.603319 MathSubStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.190476 0.0147151 0.588604 [
[(1)] .21429 .01472 0 2 Msboxa
[(2)] .40476 .01472 0 2 Msboxa
[(3)] .59524 .01472 0 2 Msboxa
[(4)] .78571 .01472 0 2 Msboxa
[(5)] .97619 .01472 0 2 Msboxa
[(0.2)] .01131 .13244 1 0 Msboxa
[(0.4)] .01131 .25016 1 0 Msboxa
[(0.6)] .01131 .36788 1 0 Msboxa
[(0.8)] .01131 .4856 1 0 Msboxa
[(1)] .01131 .60332 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
p
p
.002 w
.21429 .01472 m
.21429 .02097 L
s
P
[(1)] .21429 .01472 0 2 Mshowa
p
.002 w
.40476 .01472 m
.40476 .02097 L
s
P
[(2)] .40476 .01472 0 2 Mshowa
p
.002 w
.59524 .01472 m
.59524 .02097 L
s
P
[(3)] .59524 .01472 0 2 Mshowa
p
.002 w
.78571 .01472 m
.78571 .02097 L
s
P
[(4)] .78571 .01472 0 2 Mshowa
p
.002 w
.97619 .01472 m
.97619 .02097 L
s
P
[(5)] .97619 .01472 0 2 Mshowa
p
.001 w
.0619 .01472 m
.0619 .01847 L
s
P
p
.001 w
.1 .01472 m
.1 .01847 L
s
P
p
.001 w
.1381 .01472 m
.1381 .01847 L
s
P
p
.001 w
.17619 .01472 m
.17619 .01847 L
s
P
p
.001 w
.25238 .01472 m
.25238 .01847 L
s
P
p
.001 w
.29048 .01472 m
.29048 .01847 L
s
P
p
.001 w
.32857 .01472 m
.32857 .01847 L
s
P
p
.001 w
.36667 .01472 m
.36667 .01847 L
s
P
p
.001 w
.44286 .01472 m
.44286 .01847 L
s
P
p
.001 w
.48095 .01472 m
.48095 .01847 L
s
P
p
.001 w
.51905 .01472 m
.51905 .01847 L
s
P
p
.001 w
.55714 .01472 m
.55714 .01847 L
s
P
p
.001 w
.63333 .01472 m
.63333 .01847 L
s
P
p
.001 w
.67143 .01472 m
.67143 .01847 L
s
P
p
.001 w
.70952 .01472 m
.70952 .01847 L
s
P
p
.001 w
.74762 .01472 m
.74762 .01847 L
s
P
p
.001 w
.82381 .01472 m
.82381 .01847 L
s
P
p
.001 w
.8619 .01472 m
.8619 .01847 L
s
P
p
.001 w
.9 .01472 m
.9 .01847 L
s
P
p
.001 w
.9381 .01472 m
.9381 .01847 L
s
P
p
.002 w
0 .01472 m
1 .01472 L
s
P
p
.002 w
.02381 .13244 m
.03006 .13244 L
s
P
[(0.2)] .01131 .13244 1 0 Mshowa
p
.002 w
.02381 .25016 m
.03006 .25016 L
s
P
[(0.4)] .01131 .25016 1 0 Mshowa
p
.002 w
.02381 .36788 m
.03006 .36788 L
s
P
[(0.6)] .01131 .36788 1 0 Mshowa
p
.002 w
.02381 .4856 m
.03006 .4856 L
s
P
[(0.8)] .01131 .4856 1 0 Mshowa
p
.002 w
.02381 .60332 m
.03006 .60332 L
s
P
[(1)] .01131 .60332 1 0 Mshowa
p
.001 w
.02381 .03826 m
.02756 .03826 L
s
P
p
.001 w
.02381 .0618 m
.02756 .0618 L
s
P
p
.001 w
.02381 .08535 m
.02756 .08535 L
s
P
p
.001 w
.02381 .10889 m
.02756 .10889 L
s
P
p
.001 w
.02381 .15598 m
.02756 .15598 L
s
P
p
.001 w
.02381 .17952 m
.02756 .17952 L
s
P
p
.001 w
.02381 .20307 m
.02756 .20307 L
s
P
p
.001 w
.02381 .22661 m
.02756 .22661 L
s
P
p
.001 w
.02381 .2737 m
.02756 .2737 L
s
P
p
.001 w
.02381 .29724 m
.02756 .29724 L
s
P
p
.001 w
.02381 .32079 m
.02756 .32079 L
s
P
p
.001 w
.02381 .34433 m
.02756 .34433 L
s
P
p
.001 w
.02381 .39142 m
.02756 .39142 L
s
P
p
.001 w
.02381 .41497 m
.02756 .41497 L
s
P
p
.001 w
.02381 .43851 m
.02756 .43851 L
s
P
p
.001 w
.02381 .46205 m
.02756 .46205 L
s
P
p
.001 w
.02381 .50914 m
.02756 .50914 L
s
P
p
.001 w
.02381 .53269 m
.02756 .53269 L
s
P
p
.001 w
.02381 .55623 m
.02756 .55623 L
s
P
p
.001 w
.02381 .57977 m
.02756 .57977 L
s
P
p
.002 w
.02381 0 m
.02381 .61803 L
s
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
p
.004 w
.02381 .60332 m
.06349 .60332 L
.10317 .60332 L
.14286 .60332 L
.1627 .60332 L
.18254 .60332 L
.20238 .60332 L
.20734 .60332 L
.20982 .60332 L
.21106 .60332 L
.2123 .60332 L
.21354 .60332 L
.21478 .01472 L
.21726 .01472 L
.22222 .01472 L
.2619 .01472 L
.30159 .01472 L
.34127 .01472 L
.38095 .01472 L
.42063 .01472 L
.46032 .01472 L
.5 .01472 L
.53968 .01472 L
.57937 .01472 L
.61905 .01472 L
.65873 .01472 L
.69841 .01472 L
.7381 .01472 L
.77778 .01472 L
.81746 .01472 L
.85714 .01472 L
.89683 .01472 L
.93651 .01472 L
.97619 .01472 L
s
P
P
p
p
.004 w
.02381 .01472 m
.06349 .01472 L
.10317 .01472 L
.14286 .01472 L
.1627 .01472 L
.18254 .01472 L
.20238 .01472 L
.20734 .01472 L
.20982 .01472 L
.21106 .01472 L
.2123 .01472 L
.21354 .01472 L
.21478 .60332 L
.21726 .60332 L
.22222 .60332 L
.2619 .60332 L
.30159 .60332 L
.34127 .60332 L
.36111 .60332 L
.38095 .60332 L
.39087 .60332 L
.39583 .60332 L
.39831 .60332 L
.40079 .60332 L
.40203 .60332 L
.40327 .60332 L
.40451 .60332 L
.40575 .01472 L
.40699 .01472 L
.40823 .01472 L
.41071 .01472 L
.42063 .01472 L
.46032 .01472 L
.5 .01472 L
.53968 .01472 L
.57937 .01472 L
.61905 .01472 L
.65873 .01472 L
.69841 .01472 L
.7381 .01472 L
.77778 .01472 L
.81746 .01472 L
.85714 .01472 L
.89683 .01472 L
.93651 .01472 L
.97619 .01472 L
s
P
P
p
p
.004 w
.02381 .01472 m
.06349 .01472 L
.10317 .01472 L
.14286 .01472 L
.18254 .01472 L
.22222 .01472 L
.2619 .01472 L
.30159 .01472 L
.34127 .01472 L
.36111 .01472 L
.38095 .01472 L
.39087 .01472 L
.39583 .01472 L
.39831 .01472 L
.40079 .01472 L
.40203 .01472 L
.40327 .01472 L
.40451 .01472 L
.40575 .60332 L
.40699 .60332 L
.40823 .60332 L
.41071 .60332 L
.42063 .60332 L
.46032 .60332 L
.5 .60332 L
.53968 .60332 L
.55952 .60332 L
.56944 .60332 L
.57937 .60332 L
.58433 .60332 L
.58929 .60332 L
.59177 .60332 L
.59301 .60332 L
.59425 .60332 L
.59549 .01472 L
.59673 .01472 L
.59921 .01472 L
.61905 .01472 L
.65873 .01472 L
.69841 .01472 L
.7381 .01472 L
.77778 .01472 L
.81746 .01472 L
.85714 .01472 L
.89683 .01472 L
.93651 .01472 L
.97619 .01472 L
s
P
P
p
p
.004 w
.02381 .01472 m
.06349 .01472 L
.10317 .01472 L
.14286 .01472 L
.18254 .01472 L
.22222 .01472 L
.2619 .01472 L
.30159 .01472 L
.34127 .01472 L
.38095 .01472 L
.42063 .01472 L
.46032 .01472 L
.5 .01472 L
.53968 .01472 L
.55952 .01472 L
.56944 .01472 L
.57937 .01472 L
.58433 .01472 L
.58929 .01472 L
.59177 .01472 L
.59301 .01472 L
.59425 .01472 L
.59549 .60332 L
.59673 .60332 L
.59921 .60332 L
.61905 .60332 L
.65873 .60332 L
.69841 .60332 L
.7381 .60332 L
.75794 .60332 L
.76786 .60332 L
.77282 .60332 L
.77778 .60332 L
.78274 .60332 L
.78398 .60332 L
.78522 .60332 L
.78646 .01472 L
.7877 .01472 L
.79018 .01472 L
.79266 .01472 L
.79762 .01472 L
.81746 .01472 L
.85714 .01472 L
.89683 .01472 L
.93651 .01472 L
.97619 .01472 L
s
P
P
p
p
.004 w
.02381 .01472 m
.06349 .01472 L
.10317 .01472 L
.14286 .01472 L
.18254 .01472 L
.22222 .01472 L
.2619 .01472 L
.30159 .01472 L
.34127 .01472 L
.38095 .01472 L
.42063 .01472 L
.46032 .01472 L
.5 .01472 L
.53968 .01472 L
.57937 .01472 L
.61905 .01472 L
.65873 .01472 L
.69841 .01472 L
.7381 .01472 L
.75794 .01472 L
.76786 .01472 L
.77282 .01472 L
.77778 .01472 L
.78274 .01472 L
.78398 .01472 L
.78522 .01472 L
.78646 .60332 L
.7877 .60332 L
.79018 .60332 L
.79266 .60332 L
.79762 .60332 L
.81746 .60332 L
.85714 .60332 L
.89683 .60332 L
.93651 .60332 L
.97619 .60332 L
s
P
P
P
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.522676 0.323031 0.97619 0.603319 MathSubStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.190476 0.0147151 0.588604 [
[(1)] .21429 .01472 0 2 Msboxa
[(2)] .40476 .01472 0 2 Msboxa
[(3)] .59524 .01472 0 2 Msboxa
[(4)] .78571 .01472 0 2 Msboxa
[(5)] .97619 .01472 0 2 Msboxa
[(0.2)] .01131 .13244 1 0 Msboxa
[(0.4)] .01131 .25016 1 0 Msboxa
[(0.6)] .01131 .36788 1 0 Msboxa
[(0.8)] .01131 .4856 1 0 Msboxa
[(1)] .01131 .60332 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
p
p
.002 w
.21429 .01472 m
.21429 .02097 L
s
P
[(1)] .21429 .01472 0 2 Mshowa
p
.002 w
.40476 .01472 m
.40476 .02097 L
s
P
[(2)] .40476 .01472 0 2 Mshowa
p
.002 w
.59524 .01472 m
.59524 .02097 L
s
P
[(3)] .59524 .01472 0 2 Mshowa
p
.002 w
.78571 .01472 m
.78571 .02097 L
s
P
[(4)] .78571 .01472 0 2 Mshowa
p
.002 w
.97619 .01472 m
.97619 .02097 L
s
P
[(5)] .97619 .01472 0 2 Mshowa
p
.001 w
.0619 .01472 m
.0619 .01847 L
s
P
p
.001 w
.1 .01472 m
.1 .01847 L
s
P
p
.001 w
.1381 .01472 m
.1381 .01847 L
s
P
p
.001 w
.17619 .01472 m
.17619 .01847 L
s
P
p
.001 w
.25238 .01472 m
.25238 .01847 L
s
P
p
.001 w
.29048 .01472 m
.29048 .01847 L
s
P
p
.001 w
.32857 .01472 m
.32857 .01847 L
s
P
p
.001 w
.36667 .01472 m
.36667 .01847 L
s
P
p
.001 w
.44286 .01472 m
.44286 .01847 L
s
P
p
.001 w
.48095 .01472 m
.48095 .01847 L
s
P
p
.001 w
.51905 .01472 m
.51905 .01847 L
s
P
p
.001 w
.55714 .01472 m
.55714 .01847 L
s
P
p
.001 w
.63333 .01472 m
.63333 .01847 L
s
P
p
.001 w
.67143 .01472 m
.67143 .01847 L
s
P
p
.001 w
.70952 .01472 m
.70952 .01847 L
s
P
p
.001 w
.74762 .01472 m
.74762 .01847 L
s
P
p
.001 w
.82381 .01472 m
.82381 .01847 L
s
P
p
.001 w
.8619 .01472 m
.8619 .01847 L
s
P
p
.001 w
.9 .01472 m
.9 .01847 L
s
P
p
.001 w
.9381 .01472 m
.9381 .01847 L
s
P
p
.002 w
0 .01472 m
1 .01472 L
s
P
p
.002 w
.02381 .13244 m
.03006 .13244 L
s
P
[(0.2)] .01131 .13244 1 0 Mshowa
p
.002 w
.02381 .25016 m
.03006 .25016 L
s
P
[(0.4)] .01131 .25016 1 0 Mshowa
p
.002 w
.02381 .36788 m
.03006 .36788 L
s
P
[(0.6)] .01131 .36788 1 0 Mshowa
p
.002 w
.02381 .4856 m
.03006 .4856 L
s
P
[(0.8)] .01131 .4856 1 0 Mshowa
p
.002 w
.02381 .60332 m
.03006 .60332 L
s
P
[(1)] .01131 .60332 1 0 Mshowa
p
.001 w
.02381 .03826 m
.02756 .03826 L
s
P
p
.001 w
.02381 .0618 m
.02756 .0618 L
s
P
p
.001 w
.02381 .08535 m
.02756 .08535 L
s
P
p
.001 w
.02381 .10889 m
.02756 .10889 L
s
P
p
.001 w
.02381 .15598 m
.02756 .15598 L
s
P
p
.001 w
.02381 .17952 m
.02756 .17952 L
s
P
p
.001 w
.02381 .20307 m
.02756 .20307 L
s
P
p
.001 w
.02381 .22661 m
.02756 .22661 L
s
P
p
.001 w
.02381 .2737 m
.02756 .2737 L
s
P
p
.001 w
.02381 .29724 m
.02756 .29724 L
s
P
p
.001 w
.02381 .32079 m
.02756 .32079 L
s
P
p
.001 w
.02381 .34433 m
.02756 .34433 L
s
P
p
.001 w
.02381 .39142 m
.02756 .39142 L
s
P
p
.001 w
.02381 .41497 m
.02756 .41497 L
s
P
p
.001 w
.02381 .43851 m
.02756 .43851 L
s
P
p
.001 w
.02381 .46205 m
.02756 .46205 L
s
P
p
.001 w
.02381 .50914 m
.02756 .50914 L
s
P
p
.001 w
.02381 .53269 m
.02756 .53269 L
s
P
p
.001 w
.02381 .55623 m
.02756 .55623 L
s
P
p
.001 w
.02381 .57977 m
.02756 .57977 L
s
P
p
.002 w
.02381 0 m
.02381 .61803 L
s
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
p
.004 w
.02381 .60332 m
.06349 .48069 L
.10317 .35807 L
.14286 .23544 L
.18254 .11282 L
.20238 .0515 L
.20734 .03617 L
.20982 .02851 L
.2123 .02085 L
.21354 .01701 L
.21478 .01472 L
.21726 .01472 L
.22222 .01472 L
.2619 .01472 L
.30159 .01472 L
.34127 .01472 L
.38095 .01472 L
.42063 .01472 L
.46032 .01472 L
.5 .01472 L
.53968 .01472 L
.57937 .01472 L
.61905 .01472 L
.65873 .01472 L
.69841 .01472 L
.7381 .01472 L
.77778 .01472 L
.81746 .01472 L
.85714 .01472 L
.89683 .01472 L
.93651 .01472 L
.97619 .01472 L
s
P
P
p
p
.004 w
.02381 .01472 m
.06349 .13734 L
.10317 .25997 L
.14286 .38259 L
.1627 .44391 L
.18254 .50522 L
.20238 .56653 L
.20734 .58186 L
.20982 .58952 L
.21106 .59336 L
.2123 .59719 L
.21354 .60102 L
.21478 .60179 L
.21726 .59412 L
.22222 .57879 L
.2619 .45617 L
.30159 .33354 L
.34127 .21092 L
.36111 .1496 L
.38095 .08829 L
.39087 .05763 L
.39583 .04231 L
.40079 .02698 L
.40203 .02315 L
.40327 .01931 L
.40451 .01548 L
.40575 .01472 L
.40699 .01472 L
.40823 .01472 L
.41071 .01472 L
.42063 .01472 L
.46032 .01472 L
.5 .01472 L
.53968 .01472 L
.57937 .01472 L
.61905 .01472 L
.65873 .01472 L
.69841 .01472 L
.7381 .01472 L
.77778 .01472 L
.81746 .01472 L
.85714 .01472 L
.89683 .01472 L
.93651 .01472 L
.97619 .01472 L
s
P
P
p
p
.004 w
.02381 .01472 m
.06349 .01472 L
.10317 .01472 L
.14286 .01472 L
.1627 .01472 L
.18254 .01472 L
.20238 .01472 L
.20734 .01472 L
.20982 .01472 L
.21106 .01472 L
.2123 .01472 L
.21354 .01472 L
.21478 .01625 L
.21726 .02391 L
.22222 .03924 L
.2619 .16187 L
.30159 .28449 L
.34127 .40712 L
.36111 .46843 L
.38095 .52974 L
.39087 .5604 L
.39583 .57573 L
.39831 .58339 L
.40079 .59106 L
.40203 .59489 L
.40327 .59872 L
.40451 .60255 L
.40575 .60025 L
.40699 .59642 L
.40823 .59259 L
.41071 .58493 L
.42063 .55427 L
.46032 .43164 L
.5 .30902 L
.53968 .18639 L
.55952 .12508 L
.57937 .06377 L
.58433 .04844 L
.58929 .03311 L
.59177 .02544 L
.59425 .01778 L
.59549 .01472 L
.59673 .01472 L
.59921 .01472 L
.61905 .01472 L
.65873 .01472 L
.69841 .01472 L
.7381 .01472 L
.77778 .01472 L
.81746 .01472 L
Mistroke
.85714 .01472 L
.89683 .01472 L
.93651 .01472 L
.97619 .01472 L
Mfstroke
P
P
p
p
.004 w
.02381 .01472 m
.06349 .01472 L
.10317 .01472 L
.14286 .01472 L
.18254 .01472 L
.22222 .01472 L
.2619 .01472 L
.30159 .01472 L
.34127 .01472 L
.36111 .01472 L
.38095 .01472 L
.39087 .01472 L
.39583 .01472 L
.39831 .01472 L
.40079 .01472 L
.40203 .01472 L
.40327 .01472 L
.40451 .01472 L
.40575 .01778 L
.40699 .02161 L
.40823 .02544 L
.41071 .03311 L
.42063 .06377 L
.46032 .18639 L
.5 .30902 L
.53968 .43164 L
.55952 .49296 L
.56944 .52361 L
.57937 .55427 L
.58433 .5696 L
.58929 .58493 L
.59177 .59259 L
.59301 .59642 L
.59425 .60025 L
.59549 .60255 L
.59673 .59872 L
.59921 .59106 L
.61905 .52974 L
.65873 .40712 L
.69841 .28449 L
.7381 .16187 L
.75794 .10055 L
.76786 .0699 L
.77778 .03924 L
.78274 .02391 L
.78398 .02008 L
.78522 .01625 L
.78646 .01472 L
.7877 .01472 L
.79018 .01472 L
Mistroke
.79266 .01472 L
.79762 .01472 L
.81746 .01472 L
.85714 .01472 L
.89683 .01472 L
.93651 .01472 L
.97619 .01472 L
Mfstroke
P
P
p
p
.004 w
.02381 .01472 m
.06349 .01472 L
.10317 .01472 L
.14286 .01472 L
.18254 .01472 L
.22222 .01472 L
.2619 .01472 L
.30159 .01472 L
.34127 .01472 L
.38095 .01472 L
.42063 .01472 L
.46032 .01472 L
.5 .01472 L
.53968 .01472 L
.55952 .01472 L
.56944 .01472 L
.57937 .01472 L
.58433 .01472 L
.58929 .01472 L
.59177 .01472 L
.59301 .01472 L
.59425 .01472 L
.59549 .01548 L
.59673 .01931 L
.59921 .02698 L
.61905 .08829 L
.65873 .21092 L
.69841 .33354 L
.7381 .45617 L
.75794 .51748 L
.76786 .54814 L
.77282 .56347 L
.77778 .57879 L
.78274 .59412 L
.78398 .59795 L
.78522 .60179 L
.78646 .60102 L
.7877 .59719 L
.79018 .58952 L
.79266 .58186 L
.79762 .56653 L
.81746 .50522 L
.85714 .38259 L
.89683 .25997 L
.93651 .13734 L
.97619 .01472 L
s
P
P
p
p
.004 w
.02381 .01472 m
.06349 .01472 L
.10317 .01472 L
.14286 .01472 L
.18254 .01472 L
.22222 .01472 L
.2619 .01472 L
.30159 .01472 L
.34127 .01472 L
.38095 .01472 L
.42063 .01472 L
.46032 .01472 L
.5 .01472 L
.53968 .01472 L
.57937 .01472 L
.61905 .01472 L
.65873 .01472 L
.69841 .01472 L
.7381 .01472 L
.75794 .01472 L
.76786 .01472 L
.77282 .01472 L
.77778 .01472 L
.78274 .01472 L
.78398 .01472 L
.78522 .01472 L
.7877 .02085 L
.79762 .0515 L
.81746 .11282 L
.85714 .23544 L
.89683 .35807 L
.93651 .48069 L
.97619 .60332 L
s
P
P
P
MathSubEnd
P
% End of sub-graphic
P
P
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell[TextData[{
  "Any $r$-order spline can be constructed from linear combinations of the \
B-splines, $\\phi^r(\\tau)\\,\\beta$, where \
$\\beta:=(\\beta_1,\\cdots,\\beta_\\kappa)^\\top$ is a vector of \
coefficients. Let's plot $\\phi^r(\\tau)\\,\\beta$ for the given bases, where \
the $\\beta$s are given by ",
  StyleBox["Range[r + 5]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ":"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell[TextData[StyleBox["betas = Table[Range[r + 5], {r, 0, 3}];\nsplines = \
MapThread[Through[#1[tau]] . #2 &, \n\t{bases, betas}];\n\
Show[GraphicsArray[Partition[\n\tPlot[#, {tau, 0, 5}, PlotRange -> {0, 8}, \n\
\t\tTicks -> {Automatic, {2,4,6,8}},\n\t\tDisplayFunction -> Identity]& /@ \
splines, 2]],\n\t\tDisplayFunction -> $DisplayFunction];",
  FontFamily->"courier"]], "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.47619 0.0147151 0.47619 [
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
% Start of sub-graphic
p
0.0238095 0.0147151 0.477324 0.295003 MathSubStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.190476 0 0.0772542 [
[(0)] .02381 0 0 2 Msboxa
[(1)] .21429 0 0 2 Msboxa
[(2)] .40476 0 0 2 Msboxa
[(3)] .59524 0 0 2 Msboxa
[(4)] .78571 0 0 2 Msboxa
[(5)] .97619 0 0 2 Msboxa
[(2)] .01131 .15451 1 0 Msboxa
[(4)] .01131 .30902 1 0 Msboxa
[(6)] .01131 .46353 1 0 Msboxa
[(8)] .01131 .61803 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
p
p
.002 w
.02381 0 m
.02381 .00625 L
s
P
[(0)] .02381 0 0 2 Mshowa
p
.002 w
.21429 0 m
.21429 .00625 L
s
P
[(1)] .21429 0 0 2 Mshowa
p
.002 w
.40476 0 m
.40476 .00625 L
s
P
[(2)] .40476 0 0 2 Mshowa
p
.002 w
.59524 0 m
.59524 .00625 L
s
P
[(3)] .59524 0 0 2 Mshowa
p
.002 w
.78571 0 m
.78571 .00625 L
s
P
[(4)] .78571 0 0 2 Mshowa
p
.002 w
.97619 0 m
.97619 .00625 L
s
P
[(5)] .97619 0 0 2 Mshowa
p
.001 w
.0619 0 m
.0619 .00375 L
s
P
p
.001 w
.1 0 m
.1 .00375 L
s
P
p
.001 w
.1381 0 m
.1381 .00375 L
s
P
p
.001 w
.17619 0 m
.17619 .00375 L
s
P
p
.001 w
.25238 0 m
.25238 .00375 L
s
P
p
.001 w
.29048 0 m
.29048 .00375 L
s
P
p
.001 w
.32857 0 m
.32857 .00375 L
s
P
p
.001 w
.36667 0 m
.36667 .00375 L
s
P
p
.001 w
.44286 0 m
.44286 .00375 L
s
P
p
.001 w
.48095 0 m
.48095 .00375 L
s
P
p
.001 w
.51905 0 m
.51905 .00375 L
s
P
p
.001 w
.55714 0 m
.55714 .00375 L
s
P
p
.001 w
.63333 0 m
.63333 .00375 L
s
P
p
.001 w
.67143 0 m
.67143 .00375 L
s
P
p
.001 w
.70952 0 m
.70952 .00375 L
s
P
p
.001 w
.74762 0 m
.74762 .00375 L
s
P
p
.001 w
.82381 0 m
.82381 .00375 L
s
P
p
.001 w
.8619 0 m
.8619 .00375 L
s
P
p
.001 w
.9 0 m
.9 .00375 L
s
P
p
.001 w
.9381 0 m
.9381 .00375 L
s
P
p
.002 w
0 0 m
1 0 L
s
P
p
.002 w
.02381 .15451 m
.03006 .15451 L
s
P
[(2)] .01131 .15451 1 0 Mshowa
p
.002 w
.02381 .30902 m
.03006 .30902 L
s
P
[(4)] .01131 .30902 1 0 Mshowa
p
.002 w
.02381 .46353 m
.03006 .46353 L
s
P
[(6)] .01131 .46353 1 0 Mshowa
p
.002 w
.02381 .61803 m
.03006 .61803 L
s
P
[(8)] .01131 .61803 1 0 Mshowa
p
.002 w
.02381 0 m
.02381 .61803 L
s
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
.004 w
.02381 .07725 m
.06349 .10777 L
.10317 .13493 L
.14286 .15873 L
.18254 .17919 L
.22222 .19635 L
.2619 .21245 L
.30159 .22854 L
.34127 .24464 L
.38095 .26073 L
.42063 .27683 L
.46032 .29292 L
.5 .30902 L
.53968 .32511 L
.57937 .34121 L
.61905 .3573 L
.65873 .3734 L
.69841 .38949 L
.7381 .40558 L
.77778 .42168 L
.81746 .43885 L
.85714 .4593 L
.89683 .48311 L
.93651 .51027 L
.97619 .54078 L
s
P
P
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.522676 0.0147151 0.97619 0.295003 MathSubStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.190476 0 0.0772542 [
[(0)] .02381 0 0 2 Msboxa
[(1)] .21429 0 0 2 Msboxa
[(2)] .40476 0 0 2 Msboxa
[(3)] .59524 0 0 2 Msboxa
[(4)] .78571 0 0 2 Msboxa
[(5)] .97619 0 0 2 Msboxa
[(2)] .01131 .15451 1 0 Msboxa
[(4)] .01131 .30902 1 0 Msboxa
[(6)] .01131 .46353 1 0 Msboxa
[(8)] .01131 .61803 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
p
p
.002 w
.02381 0 m
.02381 .00625 L
s
P
[(0)] .02381 0 0 2 Mshowa
p
.002 w
.21429 0 m
.21429 .00625 L
s
P
[(1)] .21429 0 0 2 Mshowa
p
.002 w
.40476 0 m
.40476 .00625 L
s
P
[(2)] .40476 0 0 2 Mshowa
p
.002 w
.59524 0 m
.59524 .00625 L
s
P
[(3)] .59524 0 0 2 Mshowa
p
.002 w
.78571 0 m
.78571 .00625 L
s
P
[(4)] .78571 0 0 2 Mshowa
p
.002 w
.97619 0 m
.97619 .00625 L
s
P
[(5)] .97619 0 0 2 Mshowa
p
.001 w
.0619 0 m
.0619 .00375 L
s
P
p
.001 w
.1 0 m
.1 .00375 L
s
P
p
.001 w
.1381 0 m
.1381 .00375 L
s
P
p
.001 w
.17619 0 m
.17619 .00375 L
s
P
p
.001 w
.25238 0 m
.25238 .00375 L
s
P
p
.001 w
.29048 0 m
.29048 .00375 L
s
P
p
.001 w
.32857 0 m
.32857 .00375 L
s
P
p
.001 w
.36667 0 m
.36667 .00375 L
s
P
p
.001 w
.44286 0 m
.44286 .00375 L
s
P
p
.001 w
.48095 0 m
.48095 .00375 L
s
P
p
.001 w
.51905 0 m
.51905 .00375 L
s
P
p
.001 w
.55714 0 m
.55714 .00375 L
s
P
p
.001 w
.63333 0 m
.63333 .00375 L
s
P
p
.001 w
.67143 0 m
.67143 .00375 L
s
P
p
.001 w
.70952 0 m
.70952 .00375 L
s
P
p
.001 w
.74762 0 m
.74762 .00375 L
s
P
p
.001 w
.82381 0 m
.82381 .00375 L
s
P
p
.001 w
.8619 0 m
.8619 .00375 L
s
P
p
.001 w
.9 0 m
.9 .00375 L
s
P
p
.001 w
.9381 0 m
.9381 .00375 L
s
P
p
.002 w
0 0 m
1 0 L
s
P
p
.002 w
.02381 .15451 m
.03006 .15451 L
s
P
[(2)] .01131 .15451 1 0 Mshowa
p
.002 w
.02381 .30902 m
.03006 .30902 L
s
P
[(4)] .01131 .30902 1 0 Mshowa
p
.002 w
.02381 .46353 m
.03006 .46353 L
s
P
[(6)] .01131 .46353 1 0 Mshowa
p
.002 w
.02381 .61803 m
.03006 .61803 L
s
P
[(8)] .01131 .61803 1 0 Mshowa
p
.002 w
.02381 0 m
.02381 .61803 L
s
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
.004 w
.02381 .07725 m
.06349 .1208 L
.10317 .15603 L
.14286 .1847 L
.18254 .20854 L
.22222 .22932 L
.2619 .24836 L
.30159 .26615 L
.34127 .28303 L
.38095 .29935 L
.42063 .31545 L
.46032 .33155 L
.5 .34764 L
.53968 .36374 L
.57937 .37983 L
.61905 .39594 L
.65873 .41226 L
.69841 .42914 L
.7381 .44693 L
.77778 .46597 L
.81746 .48674 L
.85714 .51059 L
.89683 .53926 L
.93651 .57449 L
.97619 .61803 L
s
P
P
MathSubEnd
P
% End of sub-graphic
P
p
% Start of sub-graphic
p
0.0238095 0.323031 0.477324 0.603319 MathSubStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.190476 0 0.0772542 [
[(0)] .02381 0 0 2 Msboxa
[(1)] .21429 0 0 2 Msboxa
[(2)] .40476 0 0 2 Msboxa
[(3)] .59524 0 0 2 Msboxa
[(4)] .78571 0 0 2 Msboxa
[(5)] .97619 0 0 2 Msboxa
[(2)] .01131 .15451 1 0 Msboxa
[(4)] .01131 .30902 1 0 Msboxa
[(6)] .01131 .46353 1 0 Msboxa
[(8)] .01131 .61803 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
p
p
.002 w
.02381 0 m
.02381 .00625 L
s
P
[(0)] .02381 0 0 2 Mshowa
p
.002 w
.21429 0 m
.21429 .00625 L
s
P
[(1)] .21429 0 0 2 Mshowa
p
.002 w
.40476 0 m
.40476 .00625 L
s
P
[(2)] .40476 0 0 2 Mshowa
p
.002 w
.59524 0 m
.59524 .00625 L
s
P
[(3)] .59524 0 0 2 Mshowa
p
.002 w
.78571 0 m
.78571 .00625 L
s
P
[(4)] .78571 0 0 2 Mshowa
p
.002 w
.97619 0 m
.97619 .00625 L
s
P
[(5)] .97619 0 0 2 Mshowa
p
.001 w
.0619 0 m
.0619 .00375 L
s
P
p
.001 w
.1 0 m
.1 .00375 L
s
P
p
.001 w
.1381 0 m
.1381 .00375 L
s
P
p
.001 w
.17619 0 m
.17619 .00375 L
s
P
p
.001 w
.25238 0 m
.25238 .00375 L
s
P
p
.001 w
.29048 0 m
.29048 .00375 L
s
P
p
.001 w
.32857 0 m
.32857 .00375 L
s
P
p
.001 w
.36667 0 m
.36667 .00375 L
s
P
p
.001 w
.44286 0 m
.44286 .00375 L
s
P
p
.001 w
.48095 0 m
.48095 .00375 L
s
P
p
.001 w
.51905 0 m
.51905 .00375 L
s
P
p
.001 w
.55714 0 m
.55714 .00375 L
s
P
p
.001 w
.63333 0 m
.63333 .00375 L
s
P
p
.001 w
.67143 0 m
.67143 .00375 L
s
P
p
.001 w
.70952 0 m
.70952 .00375 L
s
P
p
.001 w
.74762 0 m
.74762 .00375 L
s
P
p
.001 w
.82381 0 m
.82381 .00375 L
s
P
p
.001 w
.8619 0 m
.8619 .00375 L
s
P
p
.001 w
.9 0 m
.9 .00375 L
s
P
p
.001 w
.9381 0 m
.9381 .00375 L
s
P
p
.002 w
0 0 m
1 0 L
s
P
p
.002 w
.02381 .15451 m
.03006 .15451 L
s
P
[(2)] .01131 .15451 1 0 Mshowa
p
.002 w
.02381 .30902 m
.03006 .30902 L
s
P
[(4)] .01131 .30902 1 0 Mshowa
p
.002 w
.02381 .46353 m
.03006 .46353 L
s
P
[(6)] .01131 .46353 1 0 Mshowa
p
.002 w
.02381 .61803 m
.03006 .61803 L
s
P
[(8)] .01131 .61803 1 0 Mshowa
p
.002 w
.02381 0 m
.02381 .61803 L
s
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
.004 w
.02381 .07725 m
.06349 .07725 L
.10317 .07725 L
.14286 .07725 L
.1627 .07725 L
.18254 .07725 L
.20238 .07725 L
.20734 .07725 L
.20982 .07725 L
.21106 .07725 L
.2123 .07725 L
.21354 .07725 L
.21478 .15451 L
.21726 .15451 L
.22222 .15451 L
.2619 .15451 L
.30159 .15451 L
.34127 .15451 L
.36111 .15451 L
.38095 .15451 L
.39087 .15451 L
.39583 .15451 L
.39831 .15451 L
.40079 .15451 L
.40203 .15451 L
.40327 .15451 L
.40451 .15451 L
.40575 .23176 L
.40699 .23176 L
.40823 .23176 L
.41071 .23176 L
.42063 .23176 L
.46032 .23176 L
.5 .23176 L
.53968 .23176 L
.55952 .23176 L
.56944 .23176 L
.57937 .23176 L
.58433 .23176 L
.58929 .23176 L
.59177 .23176 L
.59301 .23176 L
.59425 .23176 L
.59549 .30902 L
.59673 .30902 L
.59921 .30902 L
.61905 .30902 L
.65873 .30902 L
.69841 .30902 L
.7381 .30902 L
Mistroke
.75794 .30902 L
.76786 .30902 L
.77282 .30902 L
.77778 .30902 L
.78274 .30902 L
.78398 .30902 L
.78522 .30902 L
.78646 .38627 L
.7877 .38627 L
.79018 .38627 L
.79266 .38627 L
.79762 .38627 L
.81746 .38627 L
.85714 .38627 L
.89683 .38627 L
.93651 .38627 L
.97619 .38627 L
Mfstroke
P
P
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.522676 0.323031 0.97619 0.603319 MathSubStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.190476 0 0.0772542 [
[(0)] .02381 0 0 2 Msboxa
[(1)] .21429 0 0 2 Msboxa
[(2)] .40476 0 0 2 Msboxa
[(3)] .59524 0 0 2 Msboxa
[(4)] .78571 0 0 2 Msboxa
[(5)] .97619 0 0 2 Msboxa
[(2)] .01131 .15451 1 0 Msboxa
[(4)] .01131 .30902 1 0 Msboxa
[(6)] .01131 .46353 1 0 Msboxa
[(8)] .01131 .61803 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
p
p
.002 w
.02381 0 m
.02381 .00625 L
s
P
[(0)] .02381 0 0 2 Mshowa
p
.002 w
.21429 0 m
.21429 .00625 L
s
P
[(1)] .21429 0 0 2 Mshowa
p
.002 w
.40476 0 m
.40476 .00625 L
s
P
[(2)] .40476 0 0 2 Mshowa
p
.002 w
.59524 0 m
.59524 .00625 L
s
P
[(3)] .59524 0 0 2 Mshowa
p
.002 w
.78571 0 m
.78571 .00625 L
s
P
[(4)] .78571 0 0 2 Mshowa
p
.002 w
.97619 0 m
.97619 .00625 L
s
P
[(5)] .97619 0 0 2 Mshowa
p
.001 w
.0619 0 m
.0619 .00375 L
s
P
p
.001 w
.1 0 m
.1 .00375 L
s
P
p
.001 w
.1381 0 m
.1381 .00375 L
s
P
p
.001 w
.17619 0 m
.17619 .00375 L
s
P
p
.001 w
.25238 0 m
.25238 .00375 L
s
P
p
.001 w
.29048 0 m
.29048 .00375 L
s
P
p
.001 w
.32857 0 m
.32857 .00375 L
s
P
p
.001 w
.36667 0 m
.36667 .00375 L
s
P
p
.001 w
.44286 0 m
.44286 .00375 L
s
P
p
.001 w
.48095 0 m
.48095 .00375 L
s
P
p
.001 w
.51905 0 m
.51905 .00375 L
s
P
p
.001 w
.55714 0 m
.55714 .00375 L
s
P
p
.001 w
.63333 0 m
.63333 .00375 L
s
P
p
.001 w
.67143 0 m
.67143 .00375 L
s
P
p
.001 w
.70952 0 m
.70952 .00375 L
s
P
p
.001 w
.74762 0 m
.74762 .00375 L
s
P
p
.001 w
.82381 0 m
.82381 .00375 L
s
P
p
.001 w
.8619 0 m
.8619 .00375 L
s
P
p
.001 w
.9 0 m
.9 .00375 L
s
P
p
.001 w
.9381 0 m
.9381 .00375 L
s
P
p
.002 w
0 0 m
1 0 L
s
P
p
.002 w
.02381 .15451 m
.03006 .15451 L
s
P
[(2)] .01131 .15451 1 0 Mshowa
p
.002 w
.02381 .30902 m
.03006 .30902 L
s
P
[(4)] .01131 .30902 1 0 Mshowa
p
.002 w
.02381 .46353 m
.03006 .46353 L
s
P
[(6)] .01131 .46353 1 0 Mshowa
p
.002 w
.02381 .61803 m
.03006 .61803 L
s
P
[(8)] .01131 .61803 1 0 Mshowa
p
.002 w
.02381 0 m
.02381 .61803 L
s
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
.004 w
.02381 .07725 m
.06349 .09335 L
.10317 .10944 L
.14286 .12554 L
.18254 .14163 L
.22222 .15773 L
.2619 .17382 L
.30159 .18992 L
.34127 .20601 L
.38095 .22211 L
.42063 .2382 L
.46032 .2543 L
.5 .27039 L
.53968 .28648 L
.57937 .30258 L
.61905 .31867 L
.65873 .33477 L
.69841 .35086 L
.7381 .36696 L
.77778 .38305 L
.81746 .39915 L
.85714 .41524 L
.89683 .43134 L
.93651 .44743 L
.97619 .46353 L
s
P
P
MathSubEnd
P
% End of sub-graphic
P
P
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell[CellGroupData[{Cell[TextData[{
  "As an alternative, we can construct the linear combination directly with \
",
  StyleBox["MakeFinalSpline",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ":"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
fsplines = MakeFinalSpline[knots, #]& /@ betas;
Plot[Evaluate[Through[fsplines[tau]]], {tau, 0, 5}];\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.190476 0.0147151 0.0735755 [
[(1)] .21429 .01472 0 2 Msboxa
[(2)] .40476 .01472 0 2 Msboxa
[(3)] .59524 .01472 0 2 Msboxa
[(4)] .78571 .01472 0 2 Msboxa
[(5)] .97619 .01472 0 2 Msboxa
[(2)] .01131 .16187 1 0 Msboxa
[(4)] .01131 .30902 1 0 Msboxa
[(6)] .01131 .45617 1 0 Msboxa
[(8)] .01131 .60332 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
.002 w
.21429 .01472 m
.21429 .02097 L
s
P
[(1)] .21429 .01472 0 2 Mshowa
p
.002 w
.40476 .01472 m
.40476 .02097 L
s
P
[(2)] .40476 .01472 0 2 Mshowa
p
.002 w
.59524 .01472 m
.59524 .02097 L
s
P
[(3)] .59524 .01472 0 2 Mshowa
p
.002 w
.78571 .01472 m
.78571 .02097 L
s
P
[(4)] .78571 .01472 0 2 Mshowa
p
.002 w
.97619 .01472 m
.97619 .02097 L
s
P
[(5)] .97619 .01472 0 2 Mshowa
p
.001 w
.0619 .01472 m
.0619 .01847 L
s
P
p
.001 w
.1 .01472 m
.1 .01847 L
s
P
p
.001 w
.1381 .01472 m
.1381 .01847 L
s
P
p
.001 w
.17619 .01472 m
.17619 .01847 L
s
P
p
.001 w
.25238 .01472 m
.25238 .01847 L
s
P
p
.001 w
.29048 .01472 m
.29048 .01847 L
s
P
p
.001 w
.32857 .01472 m
.32857 .01847 L
s
P
p
.001 w
.36667 .01472 m
.36667 .01847 L
s
P
p
.001 w
.44286 .01472 m
.44286 .01847 L
s
P
p
.001 w
.48095 .01472 m
.48095 .01847 L
s
P
p
.001 w
.51905 .01472 m
.51905 .01847 L
s
P
p
.001 w
.55714 .01472 m
.55714 .01847 L
s
P
p
.001 w
.63333 .01472 m
.63333 .01847 L
s
P
p
.001 w
.67143 .01472 m
.67143 .01847 L
s
P
p
.001 w
.70952 .01472 m
.70952 .01847 L
s
P
p
.001 w
.74762 .01472 m
.74762 .01847 L
s
P
p
.001 w
.82381 .01472 m
.82381 .01847 L
s
P
p
.001 w
.8619 .01472 m
.8619 .01847 L
s
P
p
.001 w
.9 .01472 m
.9 .01847 L
s
P
p
.001 w
.9381 .01472 m
.9381 .01847 L
s
P
p
.002 w
0 .01472 m
1 .01472 L
s
P
p
.002 w
.02381 .16187 m
.03006 .16187 L
s
P
[(2)] .01131 .16187 1 0 Mshowa
p
.002 w
.02381 .30902 m
.03006 .30902 L
s
P
[(4)] .01131 .30902 1 0 Mshowa
p
.002 w
.02381 .45617 m
.03006 .45617 L
s
P
[(6)] .01131 .45617 1 0 Mshowa
p
.002 w
.02381 .60332 m
.03006 .60332 L
s
P
[(8)] .01131 .60332 1 0 Mshowa
p
.001 w
.02381 .04415 m
.02756 .04415 L
s
P
p
.001 w
.02381 .07358 m
.02756 .07358 L
s
P
p
.001 w
.02381 .10301 m
.02756 .10301 L
s
P
p
.001 w
.02381 .13244 m
.02756 .13244 L
s
P
p
.001 w
.02381 .1913 m
.02756 .1913 L
s
P
p
.001 w
.02381 .22073 m
.02756 .22073 L
s
P
p
.001 w
.02381 .25016 m
.02756 .25016 L
s
P
p
.001 w
.02381 .27959 m
.02756 .27959 L
s
P
p
.001 w
.02381 .33845 m
.02756 .33845 L
s
P
p
.001 w
.02381 .36788 m
.02756 .36788 L
s
P
p
.001 w
.02381 .39731 m
.02756 .39731 L
s
P
p
.001 w
.02381 .42674 m
.02756 .42674 L
s
P
p
.001 w
.02381 .4856 m
.02756 .4856 L
s
P
p
.001 w
.02381 .51503 m
.02756 .51503 L
s
P
p
.001 w
.02381 .54446 m
.02756 .54446 L
s
P
p
.001 w
.02381 .57389 m
.02756 .57389 L
s
P
p
.002 w
.02381 0 m
.02381 .61803 L
s
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
p
.004 w
.02381 .08829 m
.06349 .08829 L
.10317 .08829 L
.14286 .08829 L
.1627 .08829 L
.18254 .08829 L
.20238 .08829 L
.20734 .08829 L
.20982 .08829 L
.21106 .08829 L
.2123 .08829 L
.21354 .08829 L
.21478 .16187 L
.21726 .16187 L
.22222 .16187 L
.2619 .16187 L
.30159 .16187 L
.34127 .16187 L
.36111 .16187 L
.38095 .16187 L
.39087 .16187 L
.39583 .16187 L
.39831 .16187 L
.40079 .16187 L
.40203 .16187 L
.40327 .16187 L
.40451 .16187 L
.40575 .23544 L
.40699 .23544 L
.40823 .23544 L
.41071 .23544 L
.42063 .23544 L
.46032 .23544 L
.5 .23544 L
.53968 .23544 L
.55952 .23544 L
.56944 .23544 L
.57937 .23544 L
.58433 .23544 L
.58929 .23544 L
.59177 .23544 L
.59301 .23544 L
.59425 .23544 L
.59549 .30902 L
.59673 .30902 L
.59921 .30902 L
.61905 .30902 L
.65873 .30902 L
.69841 .30902 L
.7381 .30902 L
Mistroke
.75794 .30902 L
.76786 .30902 L
.77282 .30902 L
.77778 .30902 L
.78274 .30902 L
.78398 .30902 L
.78522 .30902 L
.78646 .38259 L
.7877 .38259 L
.79018 .38259 L
.79266 .38259 L
.79762 .38259 L
.81746 .38259 L
.85714 .38259 L
.89683 .38259 L
.93651 .38259 L
.97619 .38259 L
Mfstroke
P
P
p
p
.004 w
.02381 .08829 m
.06349 .10362 L
.10317 .11895 L
.14286 .13428 L
.18254 .1496 L
.22222 .16493 L
.2619 .18026 L
.30159 .19559 L
.34127 .21092 L
.38095 .22624 L
.42063 .24157 L
.46032 .2569 L
.5 .27223 L
.53968 .28756 L
.57937 .30289 L
.61905 .31821 L
.65873 .33354 L
.69841 .34887 L
.7381 .3642 L
.77778 .37953 L
.81746 .39486 L
.85714 .41018 L
.89683 .42551 L
.93651 .44084 L
.97619 .45617 L
s
P
P
p
p
.004 w
.02381 .08829 m
.06349 .11735 L
.10317 .14322 L
.14286 .16589 L
.18254 .18537 L
.22222 .20172 L
.2619 .21705 L
.30159 .23238 L
.34127 .2477 L
.38095 .26303 L
.42063 .27836 L
.46032 .29369 L
.5 .30902 L
.53968 .32435 L
.57937 .33967 L
.61905 .355 L
.65873 .37033 L
.69841 .38566 L
.7381 .40099 L
.77778 .41631 L
.81746 .43266 L
.85714 .45214 L
.89683 .47482 L
.93651 .50068 L
.97619 .52974 L
s
P
P
p
p
.004 w
.02381 .08829 m
.06349 .12976 L
.10317 .16332 L
.14286 .19062 L
.18254 .21333 L
.22222 .23311 L
.2619 .25125 L
.30159 .26819 L
.34127 .28426 L
.38095 .29981 L
.42063 .31515 L
.46032 .33048 L
.5 .3458 L
.53968 .36113 L
.57937 .37646 L
.61905 .3918 L
.65873 .40734 L
.69841 .42342 L
.7381 .44036 L
.77778 .4585 L
.81746 .47828 L
.85714 .50099 L
.89683 .52829 L
.93651 .56185 L
.97619 .60332 L
s
P
P
P
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]]}, Open]],

Cell[CellGroupData[{Cell[TextData[{
  "As is stands, $\\phi^r(\\tau)$ is a vector-valued function of a scalar \
argument, $\\tau$. In what follows, it will prove useful to have notation for \
a B-spline basis as a function of vector-valued argument, $\\tau_i$. To that \
end, define $\\widetilde\\phi_k^r(\\tau_i) :=\n\
(\\phi_k^r(\\tau_{i1}),\\cdots, \\phi_k^r(\\tau_{im_i}))^\\top$, an $m_i \
\\times 1$ column vector, and $\\widetilde\\phi^r(\\tau_i) := \
(\\widetilde\\phi_1^r(\\tau_i),\\cdots, \
\\widetilde\\phi_\\kappa^r(\\tau_i)),$ an $m_i\\times \\kappa$ matrix.\n\nThe \
function ",
  StyleBox["SplineMatrix",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " returns the matrix $\\widetilde\\phi^r(\\tau_i)$. The quasi-orthogonal \
structure is evident in a plot:"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
taui = Table[i, {i, 0, 5, 1/3}];
phi3mat = SplineMatrix[knots, taui, 3];
ListPlot3D[phi3mat];\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .82055 
MathPictureStart
%% SurfaceGraphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0249355 0.99742 -0.0396341 0.99742 [
[(2)] .13095 .22742 1 .98897 Msboxa
[(4)] .29892 .16134 .89202 1 Msboxa
[(6)] .47898 .09053 .77288 1 Msboxa
[(8)] .6725 .0145 .65374 1 Msboxa
[(5)] .7739 .14743 -1 .36244 Msboxa
[(10)] .86455 .28624 -1 .3316 Msboxa
[(15)] .94334 .40691 -1 .30559 Msboxa
[(0)] .04785 .27682 1 -0.3906 Msboxa
[(0.25)] .03984 .32703 1 -0.38067 Msboxa
[(0.5)] .03155 .37899 1 -0.37036 Msboxa
[(0.75)] .02296 .4328 1 -0.35964 Msboxa
[(1)] .01407 .48855 1 -0.3485 Msboxa
[ 0 0 0 0 ]
[ 1 .82055 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
.002 w
.06024 .26735 m
.67932 .02494 L
s
P
p
.002 w
.13982 .23619 m
.14425 .24057 L
s
P
[(2)] .13095 .22742 1 .98897 Mshowa
p
.002 w
.30722 .17064 m
.31137 .17529 L
s
P
[(4)] .29892 .16134 .89202 1 Mshowa
p
.002 w
.48661 .1004 m
.49042 .10533 L
s
P
[(6)] .47898 .09053 .77288 1 Mshowa
p
.002 w
.67932 .02494 m
.68274 .03015 L
s
P
[(8)] .6725 .0145 .65374 1 Mshowa
p
.001 w
.17239 .22343 m
.17502 .22609 L
s
P
p
.001 w
.20542 .2105 m
.20801 .2132 L
s
P
p
.001 w
.23889 .1974 m
.24145 .20012 L
s
P
p
.001 w
.27282 .18411 m
.27534 .18687 L
s
P
p
.001 w
.3421 .15698 m
.34455 .15981 L
s
P
p
.001 w
.37746 .14313 m
.37988 .14599 L
s
P
p
.001 w
.41333 .12909 m
.4157 .13198 L
s
P
p
.001 w
.44971 .11485 m
.45204 .11777 L
s
P
p
.001 w
.52404 .08574 m
.52628 .08873 L
s
P
p
.001 w
.56201 .07087 m
.56421 .0739 L
s
P
p
.001 w
.60054 .05578 m
.60269 .05885 L
s
P
p
.001 w
.63964 .04047 m
.64174 .04357 L
s
P
p
.001 w
.10767 .24878 m
.11036 .25138 L
s
P
p
.001 w
.07595 .2612 m
.07867 .26377 L
s
P
P
p
p
.002 w
.67932 .02494 m
.94594 .43277 L
s
P
p
.002 w
.76218 .15168 m
.75632 .1538 L
s
P
[(5)] .7739 .14743 -1 .36244 Mshowa
p
.002 w
.85272 .29017 m
.8468 .29213 L
s
P
[(10)] .86455 .28624 -1 .3316 Mshowa
p
.002 w
.93142 .41055 m
.92546 .41238 L
s
P
[(15)] .94334 .40691 -1 .30559 Mshowa
p
.001 w
.78136 .18102 m
.77784 .18227 L
s
P
p
.001 w
.79998 .2095 m
.79645 .21074 L
s
P
p
.001 w
.81807 .23716 m
.81453 .23838 L
s
P
p
.001 w
.83564 .26404 m
.83209 .26524 L
s
P
p
.001 w
.86933 .31557 m
.86577 .31673 L
s
P
p
.001 w
.88548 .34029 m
.88192 .34143 L
s
P
p
.001 w
.90121 .36434 m
.89764 .36546 L
s
P
p
.001 w
.91651 .38775 m
.91294 .38886 L
s
P
p
.001 w
.74242 .12145 m
.73891 .12274 L
s
P
p
.001 w
.72204 .09027 m
.71854 .09159 L
s
P
p
.001 w
.70102 .05812 m
.69753 .05946 L
s
P
P
p
p
.002 w
.06024 .26735 m
.02494 .49015 L
s
P
p
.002 w
.05946 .27229 m
.06527 .27002 L
s
P
[(0)] .04785 .27682 1 -0.3906 Mshowa
p
.002 w
.05149 .32259 m
.05731 .32038 L
s
P
[(0.25)] .03984 .32703 1 -0.38067 Mshowa
p
.002 w
.04324 .37466 m
.04908 .37249 L
s
P
[(0.5)] .03155 .37899 1 -0.37036 Mshowa
p
.002 w
.03469 .42858 m
.04056 .42647 L
s
P
[(0.75)] .02296 .4328 1 -0.35964 Mshowa
p
.002 w
.02584 .48445 m
.03173 .4824 L
s
P
[(1)] .01407 .48855 1 -0.3485 Mshowa
p
.001 w
.05789 .28221 m
.06137 .28086 L
s
P
p
.001 w
.0563 .2922 m
.05979 .29085 L
s
P
p
.001 w
.05471 .30226 m
.0582 .30092 L
s
P
p
.001 w
.0531 .31239 m
.0566 .31106 L
s
P
p
.001 w
.04986 .33286 m
.05336 .33154 L
s
P
p
.001 w
.04822 .3432 m
.05172 .34189 L
s
P
p
.001 w
.04657 .35362 m
.05007 .3523 L
s
P
p
.001 w
.04491 .3641 m
.04842 .3628 L
s
P
p
.001 w
.04155 .38529 m
.04506 .384 L
s
P
p
.001 w
.03986 .396 m
.04337 .39471 L
s
P
p
.001 w
.03815 .40678 m
.04166 .4055 L
s
P
p
.001 w
.03643 .41764 m
.03994 .41637 L
s
P
p
.001 w
.03295 .43959 m
.03647 .43833 L
s
P
p
.001 w
.03119 .45069 m
.03471 .44943 L
s
P
p
.001 w
.02942 .46186 m
.03295 .46061 L
s
P
p
.001 w
.02763 .47311 m
.03116 .47187 L
s
P
P
0 0 m
1 0 L
1 .82055 L
0 .82055 L
closepath
clip
newpath
p
.002 w
.06024 .26735 m
.02494 .49015 L
s
.02494 .49015 m
.40296 .79562 L
s
.40296 .79562 m
.41001 .59401 L
s
.41001 .59401 m
.06024 .26735 L
s
.67932 .02494 m
.94594 .43277 L
s
.94594 .43277 m
.97506 .64585 L
s
.97506 .64585 m
.69286 .25814 L
s
.69286 .25814 m
.67932 .02494 L
s
.06024 .26735 m
.02494 .49015 L
s
.02494 .49015 m
.69286 .25814 L
s
.69286 .25814 m
.67932 .02494 L
s
.67932 .02494 m
.06024 .26735 L
s
.41001 .59401 m
.94594 .43277 L
s
.94594 .43277 m
.97506 .64585 L
s
.97506 .64585 m
.40296 .79562 L
s
.40296 .79562 m
.41001 .59401 L
s
P
p
.664 .741 .908 r
.0015 w
.39036 .58039 .40985 .59854 .4802 .57744 .46128 .55878 Metetra
.664 .741 .908 r
.46128 .55878 .4802 .57744 .55251 .55574 .5342 .53656 Metetra
.664 .741 .908 r
.5342 .53656 .55251 .55574 .62686 .53344 .60921 .51371 Metetra
.658 .739 .91 r
.60921 .51371 .62686 .53344 .70335 .51049 .68647 .49135 Metetra
.511 .692 .95 r
.68647 .49135 .70335 .51049 .78207 .48688 .76783 .49125 Metetra
.649 .107 0 r
.76783 .49125 .78207 .48688 .86311 .46256 .85936 .55235 Metetra
.424 .164 .413 r
.85936 .55235 .86311 .46256 .97432 .64043 .93974 .47315 Metetra
.664 .741 .908 r
.37041 .56181 .39036 .58039 .46128 .55878 .44191 .53968 Metetra
.664 .741 .908 r
.44191 .53968 .46128 .55878 .5342 .53656 .51545 .51691 Metetra
.664 .741 .908 r
.51545 .51691 .5342 .53656 .60921 .51371 .59112 .49348 Metetra
.614 .724 .924 r
.59112 .49348 .60921 .51371 .68647 .49135 .66946 .47872 Metetra
0 .347 .769 r
.66946 .47872 .68647 .49135 .76783 .49125 .75486 .52148 Metetra
.479 .069 .185 r
.75486 .52148 .76783 .49125 .85936 .55235 .8422 .52147 Metetra
.857 .744 .737 r
.8422 .52147 .85936 .55235 .93974 .47315 .91806 .39971 Metetra
.664 .741 .908 r
.34997 .54278 .37041 .56181 .44191 .53968 .42205 .5201 Metetra
.664 .741 .908 r
.42205 .5201 .44191 .53968 .51545 .51691 .49623 .49677 Metetra
.664 .741 .908 r
.49623 .49677 .51545 .51691 .59112 .49348 .57258 .47275 Metetra
.449 .653 .947 r
.57258 .47275 .59112 .49348 .66946 .47872 .65252 .47999 Metetra
.628 .071 0 r
.65252 .47999 .66946 .47872 .75486 .52148 .74005 .53824 Metetra
.547 .824 .473 r
.74005 .53824 .75486 .52148 .8422 .52147 .82038 .44515 Metetra
.692 .55 .688 r
.82038 .44515 .8422 .52147 .91806 .39971 .90181 .36917 Metetra
.664 .741 .908 r
.32903 .52328 .34997 .54278 .42205 .5201 .40171 .50004 Metetra
.664 .741 .908 r
.40171 .50004 .42205 .5201 .49623 .49677 .47652 .47612 Metetra
.658 .74 .91 r
.47652 .47612 .49623 .49677 .57258 .47275 .55356 .45265 Metetra
0 .375 .834 r
.55356 .45265 .57258 .47275 .65252 .47999 .63548 .49851 Metetra
.465 0 .018 r
.63548 .49851 .65252 .47999 .74005 .53824 .72155 .50936 Metetra
.861 .753 .743 r
.72155 .50936 .74005 .53824 .82038 .44515 .80037 .38738 Metetra
.662 .567 .737 r
.80037 .38738 .82038 .44515 .90181 .36917 .88607 .34515 Metetra
.664 .741 .908 r
.30756 .5033 .32903 .52328 .40171 .50004 .38085 .47947 Metetra
.664 .741 .908 r
.38085 .47947 .40171 .50004 .47652 .47612 .4563 .45494 Metetra
.612 .725 .926 r
.4563 .45494 .47652 .47612 .55356 .45265 .53407 .43912 Metetra
.595 .054 0 r
.53407 .43912 .55356 .45265 .63548 .49851 .61753 .51802 Metetra
.61753 .51802 .63548 .49851 .72155 .50936 .7009 .44949 Metetra
.748 .61 .705 r
.7009 .44949 .72155 .50936 .80037 .38738 .78211 .35086 Metetra
.66 .647 .827 r
.78211 .35086 .80037 .38738 .88607 .34515 .8699 .32046 Metetra
.664 .741 .908 r
.28556 .48281 .30756 .5033 .38085 .47947 .35945 .45838 Metetra
.664 .741 .908 r
.35945 .45838 .38085 .47947 .4563 .45494 .43557 .43321 Metetra
.437 .652 .95 r
.43557 .43321 .4563 .45494 .53407 .43912 .5139 .4396 Metetra
.708 .181 0 r
.5139 .4396 .53407 .43912 .61753 .51802 .59803 .51503 Metetra
.943 .918 .759 r
.59803 .51503 .61753 .51802 .7009 .44949 .67998 .38571 Metetra
.687 .57 .718 r
.67998 .38571 .7009 .44949 .78211 .35086 .76438 .32448 Metetra
.663 .725 .895 r
.76438 .32448 .78211 .35086 .8699 .32046 .85327 .29508 Metetra
.664 .741 .908 r
.263 .4618 .28556 .48281 .35945 .45838 .33751 .43674 Metetra
.658 .74 .91 r
.33751 .43674 .35945 .45838 .43557 .43321 .41424 .41209 Metetra
0 .339 .806 r
.41424 .41209 .43557 .43321 .5139 .4396 .49271 .45749 Metetra
.469 0 0 r
.49271 .45749 .5139 .4396 .59803 .51503 .57706 .47264 Metetra
.83 .693 .713 r
.57706 .47264 .59803 .51503 .67998 .38571 .65977 .33816 Metetra
.666 .605 .777 r
.65977 .33816 .67998 .38571 .76438 .32448 .74631 .29921 Metetra
.664 .741 .908 r
.74631 .29921 .76438 .32448 .85327 .29508 .83617 .26898 Metetra
.664 .741 .908 r
.23985 .44025 .263 .4618 .33751 .43674 .31499 .41453 Metetra
.611 .726 .927 r
.31499 .41453 .33751 .43674 .41424 .41209 .39203 .39757 Metetra
.641 .1 0 r
.39203 .39757 .41424 .41209 .49271 .45749 .47035 .47638 Metetra
.47035 .47638 .49271 .45749 .57706 .47264 .55552 .4066 Metetra
.741 .598 .698 r
.55552 .4066 .57706 .47264 .65977 .33816 .63989 .30453 Metetra
.662 .676 .853 r
.63989 .30453 .65977 .33816 .74631 .29921 .72774 .27323 Metetra
.664 .741 .908 r
.72774 .27323 .74631 .29921 .83617 .26898 .81857 .24213 Metetra
.664 .741 .908 r
.21611 .41814 .23985 .44025 .31499 .41453 .29187 .39174 Metetra
.425 .651 .953 r
.29187 .39174 .31499 .41453 .39203 .39757 .36834 .39718 Metetra
.714 .193 0 r
.36834 .39718 .39203 .39757 .47035 .47638 .4473 .47245 Metetra
.935 .899 .766 r
.4473 .47245 .47035 .47638 .55552 .4066 .53396 .34051 Metetra
.685 .568 .717 r
.53396 .34051 .55552 .4066 .63989 .30453 .6198 .27745 Metetra
.663 .731 .9 r
.6198 .27745 .63989 .30453 .72774 .27323 .70862 .2465 Metetra
.664 .741 .908 r
.70862 .2465 .72774 .27323 .81857 .24213 .80046 .21449 Metetra
.654 .739 .912 r
.19173 .39544 .21611 .41814 .29187 .39174 .26798 .37012 Metetra
0 .308 .773 r
.26798 .37012 .29187 .39174 .36834 .39718 .34263 .41374 Metetra
.441 0 0 r
.34263 .41374 .36834 .39718 .4473 .47245 .42456 .42859 Metetra
.828 .691 .713 r
.42456 .42859 .4473 .47245 .53396 .34051 .51231 .29144 Metetra
.665 .606 .779 r
.51231 .29144 .53396 .34051 .6198 .27745 .59918 .25085 Metetra
.664 .741 .908 r
.59918 .25085 .6198 .27745 .70862 .2465 .68895 .21899 Metetra
.664 .741 .908 r
.68895 .21899 .70862 .2465 .80046 .21449 .78182 .18603 Metetra
.577 .716 .937 r
.16669 .37213 .19173 .39544 .26798 .37012 .24239 .35878 Metetra
.657 .103 0 r
.24239 .35878 .26798 .37012 .34263 .41374 .31573 .42731 Metetra
.609 .939 .742 r
.31573 .42731 .34263 .41374 .42456 .42859 .40238 .36078 Metetra
.742 .599 .698 r
.40238 .36078 .42456 .42859 .51231 .29144 .49023 .2564 Metetra
.66 .677 .855 r
.49023 .2564 .51231 .29144 .59918 .25085 .57797 .22346 Metetra
.664 .741 .908 r
.57797 .22346 .59918 .25085 .68895 .21899 .6687 .19066 Metetra
.664 .741 .908 r
.6687 .19066 .68895 .21899 .78182 .18603 .76261 .15672 Metetra
.188 .531 .923 r
.14098 .34819 .16669 .37213 .24239 .35878 .21358 .36937 Metetra
.578 .056 0 r
.21358 .36937 .24239 .35878 .31573 .42731 .28979 .40968 Metetra
.88 .817 .783 r
.28979 .40968 .31573 .42731 .40238 .36078 .38018 .29291 Metetra
.684 .567 .717 r
.38018 .29291 .40238 .36078 .49023 .2564 .4675 .22792 Metetra
.663 .731 .9 r
.4675 .22792 .49023 .2564 .57797 .22346 .55612 .19527 Metetra
.664 .741 .908 r
.55612 .19527 .57797 .22346 .6687 .19066 .64784 .16149 Metetra
.664 .741 .908 r
.64784 .16149 .6687 .19066 .76261 .15672 .74282 .1265 Metetra
.649 .105 0 r
.11356 .33085 .14098 .34819 .21358 .36937 .18106 .39996 Metetra
.18106 .39996 .21358 .36937 .28979 .40968 .26699 .34421 Metetra
.744 .606 .705 r
.26699 .34421 .28979 .40968 .38018 .29291 .35697 .24223 Metetra
.663 .606 .781 r
.35697 .24223 .38018 .29291 .4675 .22792 .44408 .19986 Metetra
.664 .741 .908 r
.44408 .19986 .4675 .22792 .55612 .19527 .53362 .16624 Metetra
.664 .741 .908 r
.53362 .16624 .55612 .19527 .64784 .16149 .62634 .13142 Metetra
.664 .741 .908 r
.62634 .13142 .64784 .16149 .74282 .1265 .7224 .09535 Metetra
.523 .056 .088 r
.07861 .35778 .11356 .33085 .18106 .39996 .15222 .38431 Metetra
.883 .82 .781 r
.15222 .38431 .18106 .39996 .26699 .34421 .24555 .26344 Metetra
.653 .526 .697 r
.24555 .26344 .26699 .34421 .35697 .24223 .33248 .20566 Metetra
.659 .678 .857 r
.33248 .20566 .35697 .24223 .44408 .19986 .41996 .17096 Metetra
.664 .741 .908 r
.41996 .17096 .44408 .19986 .53362 .16624 .51044 .13632 Metetra
.664 .741 .908 r
.51044 .13632 .53362 .16624 .62634 .13142 .60418 .10042 Metetra
.664 .741 .908 r
.60418 .10042 .62634 .13142 .7224 .09535 .70135 .06322 Metetra
.02584 .48445 .07861 .35778 .15222 .38431 .13916 .24116 Metetra
.564 .401 .619 r
.13916 .24116 .15222 .38431 .24555 .26344 .22157 .20897 Metetra
.63 .579 .779 r
.22157 .20897 .24555 .26344 .33248 .20566 .30683 .17566 Metetra
.663 .731 .901 r
.30683 .17566 .33248 .20566 .41996 .17096 .3951 .14119 Metetra
.664 .741 .908 r
.3951 .14119 .41996 .17096 .51044 .13632 .48654 .10547 Metetra
.664 .741 .908 r
.48654 .10547 .51044 .13632 .60418 .10042 .58132 .06845 Metetra
.664 .741 .908 r
.58132 .06845 .60418 .10042 .70135 .06322 .67962 .03005 Metetra
P
p
.002 w
.67932 .02494 m
.94594 .43277 L
s
.94594 .43277 m
.97506 .64585 L
s
.97506 .64585 m
.69286 .25814 L
s
.69286 .25814 m
.67932 .02494 L
s
.06024 .26735 m
.02494 .49015 L
s
.02494 .49015 m
.69286 .25814 L
s
.69286 .25814 m
.67932 .02494 L
s
.67932 .02494 m
.06024 .26735 L
s
P
p
p
.002 w
.06024 .26735 m
.67932 .02494 L
s
P
p
.002 w
.13982 .23619 m
.14425 .24057 L
s
P
[(2)] .13095 .22742 1 .98897 Mshowa
p
.002 w
.30722 .17064 m
.31137 .17529 L
s
P
[(4)] .29892 .16134 .89202 1 Mshowa
p
.002 w
.48661 .1004 m
.49042 .10533 L
s
P
[(6)] .47898 .09053 .77288 1 Mshowa
p
.002 w
.67932 .02494 m
.68274 .03015 L
s
P
[(8)] .6725 .0145 .65374 1 Mshowa
p
.001 w
.17239 .22343 m
.17502 .22609 L
s
P
p
.001 w
.20542 .2105 m
.20801 .2132 L
s
P
p
.001 w
.23889 .1974 m
.24145 .20012 L
s
P
p
.001 w
.27282 .18411 m
.27534 .18687 L
s
P
p
.001 w
.3421 .15698 m
.34455 .15981 L
s
P
p
.001 w
.37746 .14313 m
.37988 .14599 L
s
P
p
.001 w
.41333 .12909 m
.4157 .13198 L
s
P
p
.001 w
.44971 .11485 m
.45204 .11777 L
s
P
p
.001 w
.52404 .08574 m
.52628 .08873 L
s
P
p
.001 w
.56201 .07087 m
.56421 .0739 L
s
P
p
.001 w
.60054 .05578 m
.60269 .05885 L
s
P
p
.001 w
.63964 .04047 m
.64174 .04357 L
s
P
p
.001 w
.10767 .24878 m
.11036 .25138 L
s
P
p
.001 w
.07595 .2612 m
.07867 .26377 L
s
P
P
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 231},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]]}, Open]]}, Open]],

Cell[CellGroupData[{Cell["Splining functional forms", "Section",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
Consider splining each of the three functional forms: let $\\delta_s(\\tau) \
:= \\phi^r(\\tau)\\,\\beta$, $\\ell_s(\\tau) := \\phi^r(\\tau)\\,\\beta$, and \
$f_s(\\tau) := \\phi^r(\\tau)\\,\\beta$. In the latter two cases, we can \
transform the splined functional form to produce a discount function that can \
be used to represent present values\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$\\delta_s^\\ell(\\tau) := \\exp(-\\ell_s(\\tau)) = \\exp(\\phi^r(\\tau)\\, \
\\beta)$$\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["and", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$\\delta_s^f(\\tau) = \\exp(-\\int_0^\\tau f_s(u)\\,du) = \\exp( \
-\\int_0^\\tau \\phi^r(u)\\,\\beta \\, du) = \\exp(-\\psi(\\tau)\\, \
\\beta),$$\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "where $\\psi(\\tau) := \\int_0^\\tau \\phi^r(u)\\,\\beta \\, du$ is the \
integral of a B-spline basis. \nWe replace $\\widetilde\\delta(\\tau_i)$ in \
the present value expression with the spline representation; where we had \
$\\pi_i = c_i^\\top\\widetilde\\delta(\\tau_i)$, we now have (",
  StyleBox["i",
    FontSlant->"Italic"],
  ") $\\pi_i^\\delta(\\beta) = \
c_i^\\top\\widetilde\\phi^r(\\tau_i)\\,\\beta$, (",
  StyleBox["ii",
    FontSlant->"Italic"],
  ") $\\pi_i^\\ell(\\beta) = c_i^\\top \
\\exp(-\\widetilde\\phi^r(\\tau_i)\\,\\beta)$, and (",
  StyleBox["iii",
    FontSlant->"Italic"],
  ") $\\pi_i^f(\\beta) = c_i^\\top \\exp(-\\widetilde \
\\psi^r(\\tau_i)\\,\\beta)$."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
In order to spline the forward-rate curve, we need the integrals B-splines, \
whic all of spline-making functions can produce. For example,\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
bs463i = MakeBSpline[Take[basisknots, {4, 8}], -1];
Plot[bs463i[x], {x, 0, 4}];\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.238095 0.0147151 0.588604 [
[(1)] .2619 .01472 0 2 Msboxa
[(2)] .5 .01472 0 2 Msboxa
[(3)] .7381 .01472 0 2 Msboxa
[(4)] .97619 .01472 0 2 Msboxa
[(0.2)] .01131 .13244 1 0 Msboxa
[(0.4)] .01131 .25016 1 0 Msboxa
[(0.6)] .01131 .36788 1 0 Msboxa
[(0.8)] .01131 .4856 1 0 Msboxa
[(1)] .01131 .60332 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
.002 w
.2619 .01472 m
.2619 .02097 L
s
P
[(1)] .2619 .01472 0 2 Mshowa
p
.002 w
.5 .01472 m
.5 .02097 L
s
P
[(2)] .5 .01472 0 2 Mshowa
p
.002 w
.7381 .01472 m
.7381 .02097 L
s
P
[(3)] .7381 .01472 0 2 Mshowa
p
.002 w
.97619 .01472 m
.97619 .02097 L
s
P
[(4)] .97619 .01472 0 2 Mshowa
p
.001 w
.07143 .01472 m
.07143 .01847 L
s
P
p
.001 w
.11905 .01472 m
.11905 .01847 L
s
P
p
.001 w
.16667 .01472 m
.16667 .01847 L
s
P
p
.001 w
.21429 .01472 m
.21429 .01847 L
s
P
p
.001 w
.30952 .01472 m
.30952 .01847 L
s
P
p
.001 w
.35714 .01472 m
.35714 .01847 L
s
P
p
.001 w
.40476 .01472 m
.40476 .01847 L
s
P
p
.001 w
.45238 .01472 m
.45238 .01847 L
s
P
p
.001 w
.54762 .01472 m
.54762 .01847 L
s
P
p
.001 w
.59524 .01472 m
.59524 .01847 L
s
P
p
.001 w
.64286 .01472 m
.64286 .01847 L
s
P
p
.001 w
.69048 .01472 m
.69048 .01847 L
s
P
p
.001 w
.78571 .01472 m
.78571 .01847 L
s
P
p
.001 w
.83333 .01472 m
.83333 .01847 L
s
P
p
.001 w
.88095 .01472 m
.88095 .01847 L
s
P
p
.001 w
.92857 .01472 m
.92857 .01847 L
s
P
p
.002 w
0 .01472 m
1 .01472 L
s
P
p
.002 w
.02381 .13244 m
.03006 .13244 L
s
P
[(0.2)] .01131 .13244 1 0 Mshowa
p
.002 w
.02381 .25016 m
.03006 .25016 L
s
P
[(0.4)] .01131 .25016 1 0 Mshowa
p
.002 w
.02381 .36788 m
.03006 .36788 L
s
P
[(0.6)] .01131 .36788 1 0 Mshowa
p
.002 w
.02381 .4856 m
.03006 .4856 L
s
P
[(0.8)] .01131 .4856 1 0 Mshowa
p
.002 w
.02381 .60332 m
.03006 .60332 L
s
P
[(1)] .01131 .60332 1 0 Mshowa
p
.001 w
.02381 .03826 m
.02756 .03826 L
s
P
p
.001 w
.02381 .0618 m
.02756 .0618 L
s
P
p
.001 w
.02381 .08535 m
.02756 .08535 L
s
P
p
.001 w
.02381 .10889 m
.02756 .10889 L
s
P
p
.001 w
.02381 .15598 m
.02756 .15598 L
s
P
p
.001 w
.02381 .17952 m
.02756 .17952 L
s
P
p
.001 w
.02381 .20307 m
.02756 .20307 L
s
P
p
.001 w
.02381 .22661 m
.02756 .22661 L
s
P
p
.001 w
.02381 .2737 m
.02756 .2737 L
s
P
p
.001 w
.02381 .29724 m
.02756 .29724 L
s
P
p
.001 w
.02381 .32079 m
.02756 .32079 L
s
P
p
.001 w
.02381 .34433 m
.02756 .34433 L
s
P
p
.001 w
.02381 .39142 m
.02756 .39142 L
s
P
p
.001 w
.02381 .41497 m
.02756 .41497 L
s
P
p
.001 w
.02381 .43851 m
.02756 .43851 L
s
P
p
.001 w
.02381 .46205 m
.02756 .46205 L
s
P
p
.001 w
.02381 .50914 m
.02756 .50914 L
s
P
p
.001 w
.02381 .53269 m
.02756 .53269 L
s
P
p
.001 w
.02381 .55623 m
.02756 .55623 L
s
P
p
.001 w
.02381 .57977 m
.02756 .57977 L
s
P
p
.002 w
.02381 0 m
.02381 .61803 L
s
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
.004 w
.02381 .01472 m
.02505 .01472 L
.02629 .01472 L
.02753 .01472 L
.02877 .01472 L
.03001 .01472 L
.03125 .01472 L
.03249 .01472 L
.03373 .01472 L
.03497 .01472 L
.03621 .01472 L
.03745 .01472 L
.03869 .01472 L
.04117 .01472 L
.04241 .01472 L
.04365 .01472 L
.04613 .01472 L
.04861 .01472 L
.05109 .01472 L
.05357 .01472 L
.05605 .01472 L
.05853 .01473 L
.06101 .01473 L
.06349 .01473 L
.06845 .01475 L
.07341 .01476 L
.07837 .01478 L
.08333 .01481 L
.08829 .01485 L
.09325 .01489 L
.09821 .01495 L
.10317 .01502 L
.1131 .0152 L
.12302 .01545 L
.13294 .0158 L
.14286 .01625 L
.15278 .01683 L
.1627 .01755 L
.17262 .01846 L
.18254 .01956 L
.20238 .02248 L
.22222 .02654 L
.24206 .03203 L
.2619 .03924 L
.28175 .04849 L
.30159 .06008 L
.34127 .09102 L
.38095 .13274 L
.42063 .18457 L
.46032 .24447 L
Mistroke
.5 .30902 L
.53968 .37357 L
.57937 .43346 L
.61905 .48529 L
.65873 .52702 L
.67857 .54383 L
.69841 .55796 L
.71825 .56954 L
.7381 .57879 L
.75794 .586 L
.77778 .59149 L
.79762 .59556 L
.80754 .59714 L
.81746 .59847 L
.8373 .60048 L
.84722 .60121 L
.85714 .60179 L
.86706 .60224 L
.87698 .60258 L
.8869 .60283 L
.89187 .60293 L
.89683 .60302 L
.90179 .60309 L
.90675 .60314 L
.91171 .60319 L
.91667 .60322 L
.92163 .60325 L
.92659 .60327 L
.93155 .60329 L
.93403 .60329 L
.93651 .6033 L
.94147 .60331 L
.94395 .60331 L
.94643 .60331 L
.94891 .60331 L
.95139 .60332 L
.95387 .60332 L
.95511 .60332 L
.95635 .60332 L
.95759 .60332 L
.95883 .60332 L
.96007 .60332 L
.96131 .60332 L
.96255 .60332 L
.96379 .60332 L
.96503 .60332 L
.96627 .60332 L
.96751 .60332 L
.96875 .60332 L
.96999 .60332 L
Mistroke
.97123 .60332 L
.97247 .60332 L
.97371 .60332 L
.97495 .60332 L
.97619 .60332 L
Mfstroke
P
P
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell[TextData[{
  "Mathematica automatically differentiates an ",
  StyleBox["InterpolatingFunction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[" ",
    FontWeight->"Bold"],
  "but will not integrate it. The integral of an",
  StyleBox[" ",
    FontWeight->"Bold"],
  StyleBox["InterpolatingFunction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is instead calculated by ",
  StyleBox["IntegrateIF",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", which returns another ",
  StyleBox["InterpolatingFunction,",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[" ",
    FontWeight->"Bold"],
  "which we can visually decompose:"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["PlotIFPolynomials[bs463i];", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.238095 0.235439 0.147155 [
[(1)] .2619 .23544 0 2 Msboxa
[(2)] .5 .23544 0 2 Msboxa
[(3)] .7381 .23544 0 2 Msboxa
[(4)] .97619 .23544 0 2 Msboxa
[(-1)] .01131 .08828 1 0 Msboxa
[(1)] .01131 .38259 1 0 Msboxa
[(2)] .01131 .52975 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
.002 w
.2619 .23544 m
.2619 .24169 L
s
P
[(1)] .2619 .23544 0 2 Mshowa
p
.002 w
.5 .23544 m
.5 .24169 L
s
P
[(2)] .5 .23544 0 2 Mshowa
p
.002 w
.7381 .23544 m
.7381 .24169 L
s
P
[(3)] .7381 .23544 0 2 Mshowa
p
.002 w
.97619 .23544 m
.97619 .24169 L
s
P
[(4)] .97619 .23544 0 2 Mshowa
p
.002 w
0 .23544 m
1 .23544 L
s
P
p
.002 w
.02381 .08828 m
.03006 .08828 L
s
P
[(-1)] .01131 .08828 1 0 Mshowa
p
.002 w
.02381 .38259 m
.03006 .38259 L
s
P
[(1)] .01131 .38259 1 0 Mshowa
p
.002 w
.02381 .52975 m
.03006 .52975 L
s
P
[(2)] .01131 .52975 1 0 Mshowa
p
.001 w
.02381 .11771 m
.02756 .11771 L
s
P
p
.001 w
.02381 .14715 m
.02756 .14715 L
s
P
p
.001 w
.02381 .17658 m
.02756 .17658 L
s
P
p
.001 w
.02381 .20601 m
.02756 .20601 L
s
P
p
.001 w
.02381 .26487 m
.02756 .26487 L
s
P
p
.001 w
.02381 .2943 m
.02756 .2943 L
s
P
p
.001 w
.02381 .32373 m
.02756 .32373 L
s
P
p
.001 w
.02381 .35316 m
.02756 .35316 L
s
P
p
.001 w
.02381 .41202 m
.02756 .41202 L
s
P
p
.001 w
.02381 .44146 m
.02756 .44146 L
s
P
p
.001 w
.02381 .47089 m
.02756 .47089 L
s
P
p
.001 w
.02381 .50032 m
.02756 .50032 L
s
P
p
.001 w
.02381 .05885 m
.02756 .05885 L
s
P
p
.001 w
.02381 .02942 m
.02756 .02942 L
s
P
p
.001 w
.02381 .55918 m
.02756 .55918 L
s
P
p
.001 w
.02381 .58861 m
.02756 .58861 L
s
P
p
.002 w
.02381 0 m
.02381 .61803 L
s
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
p
.004 w
.02381 .23544 m
.02505 .23544 L
.02629 .23544 L
.02753 .23544 L
.02877 .23544 L
.03001 .23544 L
.03125 .23544 L
.03249 .23544 L
.03373 .23544 L
.03497 .23544 L
.03621 .23544 L
.03745 .23544 L
.03869 .23544 L
.04117 .23544 L
.04241 .23544 L
.04365 .23544 L
.04613 .23544 L
.04861 .23544 L
.05109 .23544 L
.05357 .23544 L
.05605 .23544 L
.05853 .23544 L
.06101 .23544 L
.06349 .23544 L
.06845 .23545 L
.07341 .23545 L
.07837 .23546 L
.08333 .23546 L
.08829 .23547 L
.09325 .23548 L
.09821 .2355 L
.10317 .23551 L
.1131 .23556 L
.12302 .23562 L
.13294 .23571 L
.14286 .23582 L
.15278 .23597 L
.1627 .23615 L
.17262 .23637 L
.18254 .23665 L
.20238 .23738 L
.22222 .2384 L
.24206 .23977 L
.2619 .24157 L
.28175 .24388 L
.30159 .2468 L
.32143 .25041 L
.34127 .25482 L
.38095 .26648 L
.42063 .28275 L
Mistroke
.46032 .30471 L
.5 .33354 L
.53968 .37056 L
.57937 .41719 L
.61905 .47495 L
.65873 .54549 L
Mfstroke
.69256 .61803 m
.65873 .54549 L
s
s
s
s
s
s
s
s
s
P
P
p
p
.004 w
.02381 .21091 m
.04365 .21812 L
.06349 .22362 L
.08333 .2277 L
.10317 .23067 L
.12302 .23278 L
.14286 .23429 L
.1627 .23541 L
.18254 .23635 L
.20238 .23728 L
.22222 .23838 L
.24206 .23977 L
.2619 .24157 L
.28175 .24388 L
.30159 .24678 L
.34127 .25451 L
.38095 .26495 L
.42063 .2779 L
.46032 .29288 L
.5 .30902 L
.53968 .32513 L
.57937 .33967 L
.59921 .34579 L
.61905 .35079 L
.62897 .35278 L
.63889 .35438 L
.64385 .35502 L
.64881 .35555 L
.65129 .35577 L
.65377 .35596 L
.65625 .35612 L
.65873 .35625 L
.65997 .3563 L
.66121 .35635 L
.66245 .35639 L
.66369 .35641 L
.66493 .35643 L
.66617 .35644 L
.66741 .35645 L
.66865 .35644 L
.66989 .35643 L
.67113 .35641 L
.67237 .35637 L
.67361 .35633 L
.67609 .35622 L
.67857 .35608 L
.68105 .3559 L
.68353 .35568 L
.68849 .35512 L
.69345 .3544 L
Mistroke
.69841 .35351 L
.70833 .35121 L
.71825 .34817 L
.72817 .34434 L
.7381 .33967 L
.75794 .32759 L
.77778 .3115 L
.79762 .29093 L
.81746 .26541 L
.85714 .1975 L
.89683 .1035 L
Mfstroke
.92976 0 m
.89683 .1035 L
s
s
s
P
P
p
p
.004 w
s
s
.07023 .61803 m
.10317 .51453 L
s
.10317 .51453 m
.14286 .42053 L
.1627 .38359 L
.18254 .35262 L
.20238 .3271 L
.22222 .30653 L
.24206 .29044 L
.25198 .28392 L
.2619 .27836 L
.27183 .27369 L
.28175 .26986 L
.29167 .26682 L
.30159 .26452 L
.30655 .26363 L
.31151 .26291 L
.31647 .26236 L
.31895 .26213 L
.32143 .26195 L
.32391 .26181 L
.32515 .26175 L
.32639 .2617 L
.32763 .26166 L
.32887 .26163 L
.33011 .2616 L
.33135 .26159 L
.33259 .26158 L
.33383 .26159 L
.33507 .2616 L
.33631 .26162 L
.33879 .26168 L
.34003 .26173 L
.34127 .26178 L
.34623 .26207 L
.34871 .26226 L
.35119 .26248 L
.36111 .26366 L
.37103 .26526 L
.38095 .26725 L
.42063 .27836 L
.46032 .29291 L
.5 .30902 L
.53968 .32515 L
.57937 .34013 L
.61905 .35309 L
.65873 .36352 L
.67857 .36772 L
.69841 .37125 L
.71825 .37415 L
.7381 .37646 L
.75794 .37827 L
Mistroke
.77778 .37966 L
.81746 .38169 L
.8373 .38262 L
.85714 .38374 L
.87698 .38525 L
.89683 .38736 L
.91667 .39033 L
.93651 .39442 L
.95635 .39991 L
.97619 .40712 L
Mfstroke
P
P
p
p
.004 w
s
s
s
s
s
s
s
s
.30744 0 m
.34127 .07254 L
s
.34127 .07254 m
.38095 .14308 L
.42063 .20085 L
.46032 .24747 L
.5 .28449 L
.53968 .31333 L
.57937 .33528 L
.61905 .35155 L
.63889 .3579 L
.65873 .36322 L
.67857 .36762 L
.69841 .37123 L
.71825 .37415 L
.7381 .37646 L
.75794 .37826 L
.77778 .37964 L
.79762 .38065 L
.80754 .38105 L
.81746 .38138 L
.8373 .38188 L
.84722 .38207 L
.85714 .38221 L
.86706 .38232 L
.87698 .38241 L
.8869 .38247 L
.89187 .3825 L
.89683 .38252 L
.90179 .38254 L
.90675 .38255 L
.91171 .38256 L
.91667 .38257 L
.92163 .38258 L
.92659 .38258 L
.93155 .38259 L
.93403 .38259 L
.93651 .38259 L
.94147 .38259 L
.94395 .38259 L
.94643 .38259 L
.94891 .38259 L
.95139 .38259 L
.95387 .38259 L
.95511 .38259 L
.95635 .38259 L
.95759 .38259 L
.95883 .38259 L
.96007 .38259 L
.96131 .38259 L
.96255 .38259 L
.96379 .38259 L
Mistroke
.96503 .38259 L
.96627 .38259 L
.96751 .38259 L
.96875 .38259 L
.96999 .38259 L
.97123 .38259 L
.97247 .38259 L
.97371 .38259 L
.97495 .38259 L
.97619 .38259 L
Mfstroke
P
P
P
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell[TextData[{
  StyleBox["SplineMatrix",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " can return $\\widetilde \\psi^r(\\tau_i)$. The last argument indicates \
the order of differentiation (in this case integration):"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["phi3int = SplineMatrix[knots, {1, 3/2, 4.2}, 3, -1]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
{{1/4, 7/16, 13/48, 1/24, 0, 0, 0, 0}, 
 
  {1/4, 127/256, 141/256, 77/384, 1/384, 0, 0, 0}, 
 
  {1/4, 1/2, 3/4, 1, 0.982933333333333, 0.5878666666666668, 
 
   0.1288000000000001, 0.0004000000000000115}}\
\>", "\<\
  1  7   13  1
{{-, --, --, --, 0, 0, 0, 0}, 
  4  16  48  24
 
   1  127  141  77    1
  {-, ---, ---, ---, ---, 0, 0, 0}, 
   4  256  256  384  384
 
   1  1  3
  {-, -, -, 1, 0.982933, 0.587867, 0.1288, 0.0004}}
   4  2  4\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell["\<\
We see that the integral of a B-spline basis does not retain the \
quasi-orthogonality of the B-Spline basis.\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["Regression Splines", "Section",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
Since neither $\\pi_i^\\ell(\\beta)$ nor $\\pi_i^f(\\beta)$ is linear in \
$\\beta$, we must use some sort of nonlinear least squares to fit the spline \
when we choose either of these functional forms.\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
Let $P$ be an $n\\times 1$ vector of bond prices, $p_i$, and ${\\mit \
\\Pi}(\\beta)$ be the corresponding vector of present values of the bonds, \
$\\pi_i^j(\\beta)$. Then for a regression spline, we choose the $\\beta$ that \
minimizes the sum of squared residuals:\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$\\min_\\beta\\, [(P-{\\mit \\Pi}(\\beta))^\\top(P-{\\mit \\Pi}(\\beta))].$$\
\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
Using a standard technique (see Chow (1983)), we linearize ${\\mit \
\\Pi}(\\beta)$ around an initial guess $\\beta^0$,\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$${\\mit \\Pi}(\\beta) \\approx {\\mit \\Pi}(\\beta^0) + \\left.(\\beta - \
\\beta^0){\\partial{\\mit \\Pi} (\\beta) \\over \\partial \
\\beta^\\top}\\right|_{\\beta = \\beta^0},$$\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
and define $X(\\beta^0) := \\left. {\\partial {\\mit \
\\Pi}(\\beta)/\\partial\\beta^\\top} \\right|_{\\beta=\\beta^0}$ and \
$Y(\\beta^0) := P - {\\mit \\Pi}(\\beta^0) + \\beta^0\\, X(\\beta^0)$. \
Rearranging the minimization problem using these definitions yields\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$ \\min_{\\beta} \\left[ \\Big(Y(\\beta^0) - X(\\beta^0) \\beta \\Big)^\\top \
\\Big(Y(\\beta^0) - X(\\beta^0) \\beta \\Big)\\right].$$\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["The minimizer for the previous expression is ", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$\\beta^1 = \\Big(X(\\beta^0)^\\top X(\\beta^0)\\Big)^{-1} X \
(\\beta^0)^\\top Y(\\beta^0),$$\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
where $\\beta^1$ is an updated $\\beta^0$.  We can use $\\beta^1$ as the \
initial guess for the next iteration,  obtaining $\\beta^2$. We iterate until \
convergence. The solution is the fixed-point $\\beta^* = (X(\\beta^*)^\\top \
X(\\beta^*))^{-1} X(\\beta^*)^\\top Y(\\beta^*)$.\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "By contrast, splining the discount function is easier, since \
$\\pi_i^\\delta(\\beta)$ is linear in $\\beta$. The relationship between bond \
prices and present values can be written $p_i = x_i\\,\\beta + \
\\varepsilon_i$, where the ``independent variables'' are the bond payments \
evaluated according to the B-spline basis; ",
  StyleBox["i.e.",
    FontSlant->"Italic"],
  " $x_i := c_i^\\top\\widetilde \\phi^r(\\tau_i)$. Now let $X$ be the matrix \
of corresponding to the $x_i$. We can use ordinary least squares (OLS) to \
determine $\\beta^* = (X^\\top X)^{-1} X^\\top P$. This produces a regression \
spline of the discount function of the sort McCulloch used."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["Smoothing Splines", "Section",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
A smoothing spline is a regression spline with a penalty for roughness. With \
a smoothing spline, one can use a large number of knot points but penalize \
excess variability in the estimated discount function, which has the effect \
of reducing the effective number of parameters since the penalty forces \
implicit relationships between the parameters of the spline. The penalty we \
describe here applies only to a cubic spline. Let $h(\\tau)$ be the function \
being splined. The penalty is defined as\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$\\lambda \\int_0^T h''(\\tau)^2 d\\tau,$$\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
a constant times the integral of the squared second derivative of the \
function being splined.\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
The penalty can be written in terms of the B-spline basis as follows: \
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$\\lambda \\int_0^T \\left({\\partial^2 \\phi^3(\\tau)\\, \\beta \\over \
\\partial \\tau^2}\\right)^2 \\, d\\tau = \\lambda\\, \
\\beta^\\top\\left(\\int_0^T \\phi^3''(\\tau)^\\top\\phi^3''(\\tau) \
d\\tau\\right)\\beta =\\lambda \\,\\beta^\\top H\\beta .$$\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$H$ is a $\\kappa\\times\\kappa$ matrix that is band diagonal by the \
structure of a B-spline basis. Since any $\\beta$ that makes $h(\\tau)$ \
linear in $\\tau$ is not penalized, $H$ has two zero eigenvalues. Also note \
that $H$ is completely determined by the knot points. For example, we can \
construct $H$ for the knots given above:\\footnote{$^5$}{We thank Jon Faust \
and Ludger Hentschel for the code that constructs the penalty matrix.}\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
H = PenaltyMatrix[knots, 3];
NullSpace[H]\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
{{-14, -13, -11, -8, -5, -2, 0, 1}, 
 
  {15, 14, 12, 9, 6, 3, 1, 0}}\
\>", "\<\
{{-14, -13, -11, -8, -5, -2, 0, 1}, 
 
  {15, 14, 12, 9, 6, 3, 1, 0}}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell["We see the the null space of $H$ is indeed two-dimensional.", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
The minimization problem can be stated as follows for a given $\\lambda$: \
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$ \\min_{\\beta(\\lambda)} \\left[ \\Big(P - {\\mit \\Pi} (\\beta(\\lambda)) \
 \\Big)^\\top  \\Big(P - {\\mit \\Pi}(\\beta (\\lambda)) \\Big) + \\lambda\\, \
 \\beta(\\lambda)^\\top H \\beta (\\lambda) \\right].$$\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
In general, the minimizer is found by nonlinear least squares as described in \
the previous section, iterating on\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$\\beta^{i+1}(\\lambda) = \\Big(X(\\beta^i(\\lambda))^\\top \
X(\\beta^i(\\lambda)) + \\lambda H \\Big)^{-1}  X (\\beta^i (\\lambda))^\\top \
Y (\\beta^i (\\lambda))$$\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["until convergence:", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$\\beta^*(\\lambda) = \\Big(X(\\beta^*(\\lambda))^\\top \
X(\\beta^*(\\lambda)) + \\lambda H \\Big)^{-1} X(\\beta^*(\\lambda))^\\top Y(\
\\beta^*(\\lambda)).$$\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
(Note that when splining $\\delta(\\tau)$, $\\beta^*(\\lambda) = (X^\\top X + \
\\lambda H)^{-1} X^\\top P.$)

Formally, the previous expression is ridge regression estimator. By employing \
a roughness penalty, we can over-parameterize the spline, making $X(\\beta^*(\
\\lambda))^\\top X(\\beta^*(\\lambda))$ nearly singular, and use the penalty \
to reduce the effective number of parameters. The penalty thus ``solves'' the \
multicollinearity problem. The advantage of this technique is that the shape \
of the spline is controlled by a single parameter, $\\lambda$.

One common measure of the effective number of parameters is the trace of \
$A(\\lambda)$, denoted ${\\rm tr}(A(\\lambda))$, where\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$A(\\lambda) := X(\\beta^*(\\lambda))\\Big(X(\\beta^*(\\lambda))^\\top \
X(\\beta^*(\\lambda)) + \\lambda H \\Big)^{-1} X(\\beta^*(\\lambda))^\\top.$$\
\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
Note that $A(\\lambda)\\,Y(\\beta^*(\\lambda))$ is the vector of fitted $Y$ \
values which in the linear case is the vetcor of fitted prices. The extreme \
cases are ${\\rm tr}(A(0)) = \\kappa$ (with no penalty the number of \
effective parameters equals the number of B-splines), and ${\\rm \
tr}(A(\\infty)) = 2$ (with an infinite penalty the number of effective \
parameters equals 2).\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["Generalized cross validation", "Section",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
In this section, we present a technique for choosing the appropriate value \
for $\\lambda$. We choose the value of $\\lambda$ that minimizes the \
``generalized cross validation'' (GCV) value,\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$\\gamma(\\lambda) := {\\Big((I - A(\\lambda))Y \
(\\beta^*(\\lambda))\\Big)^\\top \\Big((I - A (\\lambda))) \
Y(\\beta^*(\\lambda)) \\Big) \\over \\Big(n - \\theta\\,{\\rm \
tr}(A(\\lambda))\\Big)^2}.$$\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "The numerator of the previous expression is the residual sum of squares. \
When $\\theta = 1$, the denominator is the squared effective degrees of \
freedom (the difference between the number of observations and the effective \
number of parameters).  The parameter $\\theta$ is called the cost. It \
controls the trade-off between goodness-of-fit and parsimony. In \
plain-vanilla GCV, $\\theta=1$.  However, $\\theta$ can be increased to \
reduce the signal extracted, thereby stiffening the spline. (In the code, \
$\\theta$ is called the ",
  StyleBox["TuningParameter",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", which is an option to ",
  StyleBox["DiscountFunction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ".)"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
When the discount function is splined directly, $A(\\lambda) = X(X^\\top X + \
\\lambda H)^{-1} X^\\top$ and there is a simplified expression for \
$\\gamma(\\lambda)$ that can be minimized directly. In general, however, a \
new $X(\\beta^*(\\lambda))$ matrix must be formed for each value of \
$\\lambda$. Thus for each value of $\\lambda$ we test, we must solve for \
$\\beta^*(\\lambda)$ and then calculate $\\gamma(\\lambda)$.  The overall \
solution is given by\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$\\beta^*(\\lambda^*) = \\Big (X (\\beta^* (\\lambda^*))^\\top X (\\beta^* (\
\\lambda^*)) + \\lambda H \\Big)^{-1} X (\\beta^*(\\lambda^*))^\\top \
Y(\\beta^*(\\lambda^*)),$$\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["where $\\lambda^*$ minimizes $\\gamma(\\lambda)$.", "Text",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["Implementing the Estimators", "Section",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "For smoothing splines, we need starting values for $\\lambda$. However, it \
is not easy to know in advance what a good starting value is.  At extreme \
values of $\\lambda$ (both large and small) the GCV function becomes flat \
(see below) and optimizers can get stuck at non-minimums. Experimenting with \
differnt values is important to find out where good starting values are for a \
given data set. We have found that starting near $10^{10}$ usually (but not \
always) converges to the true global minimu. In addition, one may need to \
``tune'' the ",
  StyleBox["FindMinimum",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " routine. For example, one may need to boost the accuracy goal; ",
  StyleBox["e.g.,",
    FontSlant->"Italic"],
  " ",
  StyleBox["AccuracyGoal -> 10",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "When splining either $\\ell(\\tau)$ or $f(\\tau)$, we need starting values \
for $\\beta$ as well. The fixed point problem will not converge with bad \
starting values. Fortunately, good starting values for $\\beta$ are easy to \
calculate. One of the properties of B-splines is that $\\sum_{k=1}^\\kappa \
\\phi(\\tau) = 1$. As a consequence, the coefficients, $\\beta$, track the \
value of the function, $\\phi(\\tau)\\,\\beta$. Thus any reasonable estimate \
of the function to be splined can be used to form starting values. For \
example, suppose a crude estimate of the function to be splined is $\\widehat \
h(\\tau)$.  For a cubic spline, let the starting value for $\\beta_k$ be \
$\\beta_{k0} = {1 \\over 3} \\sum_{i=k}^{k+2} \\widehat h(d_i)$. With these \
starting values, the fixed point problem converges rapidly. The internal \
default for the yield curve is a flat 5 percent curve. If ",
  StyleBox["DLStart.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " or ",
  StyleBox["LStart.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is used, the yield curve stored in there is used."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["Usage Overview", "Section",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
A typical session will include reading data, calculating when the coupon \
payments occur, estimating the yield curve. A session may include calculating \
yields and yield errors and producing graphs and reports.

We start by reading some data:\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["md = ReadData[\"19880719.dat\"]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
MD[{1988, 7, 19}, {{1988, 8, 4}, {1988, 9, 1}, 
 
   {1988, 9, 29}, {1988, 10, 27}, {1988, 11, 25}, 
 
   {1988, 12, 22}, {1988, 7, 21}, {1988, 7, 28}, 
 
   {1988, 8, 11}, {1988, 8, 18}, {1988, 8, 25}, 
 
   {1988, 9, 8}, {1988, 9, 15}, {1988, 9, 22}, 
 
   {1988, 10, 6}, {1988, 10, 13}, {1988, 11, 3}, 
 
   {1988, 11, 10}, {1988, 11, 17}, {1988, 12, 1}, 
 
   {1988, 12, 8}, {1988, 12, 15}, {1988, 12, 29}, 
 
   {1989, 1, 5}, {1989, 1, 12}, {1989, 2, 16}, 
 
   {1989, 3, 16}, {1989, 4, 13}, {1989, 5, 11}, 
 
   {1989, 6, 8}, {1989, 7, 6}, {1990, 2, 15}, 
 
   {1992, 8, 15}, {1993, 2, 15}, {1993, 11, 15}, 
 
   {1995, 2, 15}, {1998, 11, 15}, {1993, 2, 15}, 
 
   {1992, 2, 15}, {1994, 5, 15}, {1995, 2, 15}, 
 
   {1990, 5, 15}, {2000, 5, 15}, {1995, 8, 15}, 
 
   {1996, 8, 15}, {2002, 2, 15}, {1992, 8, 15}, 
 
   {2002, 11, 15}, {1993, 2, 15}, {1993, 8, 15}, 
 
   {2003, 8, 15}, {1993, 11, 15}, {2003, 11, 15}, 
 
   {1994, 2, 15}, {2004, 5, 15}, {1994, 8, 15}, 
 
   {1994, 11, 15}, {2004, 11, 15}, {1995, 2, 15}, 
 
   {2005, 2, 15}, {1995, 5, 15}, {2005, 5, 15}, 
 
   {1995, 5, 15}, {1995, 11, 15}, {2005, 11, 15}, 
 
   {2001, 2, 15}, {2001, 5, 15}, {2006, 5, 15}, 
 
   {2001, 8, 15}, {2001, 11, 15}, {2006, 11, 15}, 
 
   {2002, 2, 15}, {2002, 11, 15}, {2007, 11, 15}, 
 
   {2003, 2, 15}, {2003, 5, 15}, {2003, 8, 15}, 
 
   {2008, 8, 15}, {2003, 11, 15}, {2004, 5, 15}, 
 
   {2009, 5, 15}, {2004, 8, 15}, {2009, 8, 15}, 
 
   {2004, 11, 15}, {2009, 11, 15}, {2015, 2, 15}, 
 
   {2005, 5, 15}, {2005, 8, 15}, {2015, 8, 15}, 
 
   {2015, 11, 15}, {2006, 2, 15}, {2016, 2, 15}, 
 
   {2016, 5, 15}, {2016, 11, 15}, {2017, 5, 15}, 
 
   {2017, 8, 15}, {2018, 5, 15}, {1988, 11, 15}, 
 
   {1989, 5, 15}, {1989, 11, 15}, {1990, 8, 15}, 
 
   {1990, 11, 15}, {1991, 5, 15}, {1991, 8, 15}, 
 
   {1988, 10, 15}, {1991, 11, 15}, {1989, 1, 15}, 
 
   {1992, 2, 15}, {1989, 4, 15}, {1992, 5, 15}, 
 
   {1989, 7, 15}, {1989, 10, 15}, {1992, 11, 15}, 
 
   {1990, 1, 15}, {1993, 2, 15}, {1990, 4, 15}, 
 
   {1993, 5, 15}, {1990, 7, 15}, {1993, 8, 15}, 
 
   {1988, 11, 15}, {1990, 10, 15}, {1993, 11, 15}, 
 
   {1989, 2, 15}, {1991, 1, 15}, {1989, 5, 15}, 
 
   {1991, 4, 15}, {1994, 5, 15}, {1989, 8, 15}, 
 
   {1991, 7, 15}, {1994, 8, 15}, {1989, 11, 15}, 
 
   {1991, 10, 15}, {1988, 9, 30}, {1994, 11, 15}, 
 
   {1990, 2, 15}, {1988, 12, 31}, {1992, 1, 15}, 
 
   {1995, 2, 15}, {1990, 5, 15}, {1989, 3, 31}, 
 
   {1992, 4, 15}, {1995, 5, 15}, {1990, 8, 15}, 
 
   {1989, 6, 30}, {1992, 7, 15}, {1995, 8, 15}, 
 
   {1990, 11, 15}, {1989, 9, 30}, {1992, 10, 15}, 
 
   {1988, 11, 15}, {1995, 11, 15}, {1991, 2, 15}, 
 
   {1989, 12, 31}, {1993, 1, 15}, {1989, 2, 15}, 
 
   {1996, 2, 15}, {1991, 5, 15}, {1990, 3, 31}, 
 
   {1993, 4, 15}, {1989, 5, 15}, {1996, 5, 15}, 
 
   {1991, 8, 15}, {1990, 6, 30}, {1993, 7, 15}, 
 
   {1989, 8, 15}, {1988, 8, 31}, {1991, 11, 15}, 
 
   {1988, 9, 30}, {1990, 9, 30}, {1988, 10, 31}, 
 
   {1993, 10, 15}, {1989, 11, 15}, {1996, 11, 15}, 
 
   {1988, 11, 30}, {1992, 2, 15}, {1988, 12, 31}, 
 
   {1990, 12, 31}, {1994, 1, 15}, {1989, 1, 31}, 
 
   {1990, 2, 15}, {1989, 2, 28}, {1992, 5, 15}, 
 
   {1989, 3, 31}, {1991, 3, 31}, {1994, 4, 15}, 
 
   {1989, 4, 30}, {1990, 5, 15}, {1997, 5, 15}, 
 
   {1989, 5, 31}, {1992, 8, 15}, {1989, 6, 30}, 
 
   {1991, 6, 30}, {1994, 7, 15}, {1989, 7, 31}, 
 
   {1990, 8, 15}, {1997, 8, 15}, {1989, 8, 31}, 
 
   {1992, 11, 15}, {1989, 9, 30}, {1991, 9, 30}, 
 
   {1994, 10, 15}, {1989, 10, 31}, {1990, 11, 15}, 
 
   {1997, 11, 15}, {1989, 11, 30}, {1993, 2, 15}, 
 
   {1989, 12, 31}, {1991, 12, 31}, {1995, 1, 15}, 
 
   {1990, 1, 31}, {1991, 2, 15}, {1998, 2, 15}, 
 
   {1990, 2, 28}, {1993, 5, 15}, {1990, 3, 31}, 
 
   {1992, 3, 31}, {1995, 4, 15}, {1990, 4, 30}, 
 
   {1998, 5, 15}, {1990, 5, 31}, {1993, 8, 15}, 
 
   {1990, 6, 30}, {1992, 6, 30}, {1995, 7, 15}}, 
 
  {0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 
 
   0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 
 
   0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 32, 30, 30, 
 
   31, 40, 38, 20, 19, 20, 20, 15, 25, 20, 20, 25, 15, 25, 
 
   15, 15, 25, 15, 25, 15, 25, 15, 15, 25, 15, 25, 15, 25, 
 
   15, 15, 25, 20, 20, 25, 20, 20, 25, 20, 20, 25, 20, 20, 
 
   20, 25, 20, 20, 25, 20, 25, 20, 25, 30, 20, 20, 30, 30, 
 
   20, 30, 30, 30, 30, 30, 30, 10, 10, 10, 10, 10, 10, 10, 
 
   7, 10, 7, 10, 7, 10, 7, 7, 10, 7, 10, 7, 10, 7, 10, 5, 
 
   7, 10, 5, 7, 5, 7, 10, 5, 7, 10, 5, 7, 4, 10, 5, 4, 7, 
 
   10, 5, 4, 7, 10, 5, 4, 7, 10, 5, 4, 7, 3, 10, 5, 4, 7, 
 
   3, 10, 5, 4, 7, 3, 10, 5, 4, 7, 3, 2, 5, 2, 4, 2, 7, 3, 
 
   10, 2, 5, 2, 4, 7, 2, 3, 2, 5, 2, 4, 7, 2, 3, 10, 2, 5, 
 
   2, 4, 7, 2, 3, 10, 2, 5, 2, 4, 7, 2, 3, 10, 2, 5, 2, 4, 
 
   7, 2, 3, 10, 2, 5, 2, 4, 7, 2, 10, 2, 5, 2, 4, 7}, 
 
  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
 
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.5, 4.25, 4, 4.125, 
 
   3, 3.5, 6.75, 7.5, 8.5, 7.875, 8.25, 8.25, 8.375, 8, 
 
   7.625, 7.25, 7.875, 7.875, 8.625, 8.375, 8.625, 8.75, 9, 
 
   9.125, 8.75, 10.125, 10.375, 10.5, 11.75, 12.625, 10, 
 
   10.375, 11.5, 12.75, 11.75, 13.125, 13.875, 13.375, 
 
   15.75, 14, 14.25, 11.625, 10.375, 10.75, 10.75, 11.125, 
 
   12, 11.875, 12.375, 13.25, 13.75, 12.5, 11.625, 11.75, 
 
   11.25, 12, 10.75, 10.625, 9.875, 9.375, 9.25, 7.25, 7.5, 
 
   8.75, 8.875, 9.125, 8.75, 9.25, 10.75, 10.75, 13, 14.5, 
 
   14.875, 15.375, 14.25, 14.625, 14.625, 14.375, 13.75, 
 
   14.5, 11.875, 10.5, 10.5, 10.875, 10.5, 10.125, 10.75, 
 
   11.875, 11.75, 11.5, 11.75, 11.375, 11.75, 11.75, 
 
   12.375, 13.125, 13.875, 13.75, 12.625, 12.75, 12.25, 
 
   11.375, 11.625, 11, 10.625, 11.625, 11.25, 11.375, 
 
   11.25, 11.75, 11.25, 9.875, 9.625, 10.375, 10.5, 9.625, 
 
   9.375, 9.75, 8.625, 9.5, 9.125, 8.375, 8.75, 8, 8.875, 
 
   8.125, 7.25, 7.375, 6.875, 7.375, 7.5, 7.25, 7.25, 
 
   6.625, 6.125, 6.5, 6.375, 6.75, 6.375, 7.125, 6.375, 
 
   7.25, 6.25, 6.625, 6.25, 6.625, 7, 6.125, 6.5, 6.25, 
 
   6.625, 6.375, 6.75, 7, 7.125, 7.875, 8.5, 8, 8.25, 
 
   7.375, 7.875, 8, 7.625, 7.875, 8.625, 7.75, 8.375, 8.5, 
 
   9.125, 9.5, 7.875, 8, 8.875, 7.75, 8.25, 7.875, 8.25, 
 
   8.625, 7.375, 7.375, 8.125, 7.125, 7.625, 7.375, 7.875, 
 
   8.375, 7.625, 9, 8.125, 8.75, 8, 8.25, 8.875}, 
 
  {99.7321, 99.226, 98.6885, 98.1437, 97.5431, 96.999, 
 
   99.983, 99.8716, 99.626, 99.5215, 99.385, 99.1014, 
 
   98.9788, 98.8231, 98.5678, 98.4369, 97.9978, 97.8593, 
 
   97.7133, 97.4354, 97.3053, 97.1551, 96.886, 96.7186, 
 
   96.568, 95.8269, 95.2532, 94.6897, 94.1164, 93.5669, 
 
   92.9508, 94.344, 93.75, 94.156, 94, 94.063, 94.25, 
 
   93.25, 95, 95.313, 90.344, 100.281, 91.094, 93.813, 
 
   91.719, 85, 96.406, 86.969, 97.406, 100, 90.719, 99.844, 
 
   93.813, 101, 97.063, 100.156, 105.875, 107.188, 107.625, 
 
   118.719, 118.656, 104.25, 107.156, 113.5, 127.469, 
 
   118.5, 128.75, 137.375, 130.813, 149.5, 138.938, 
 
   138.188, 117.813, 107.688, 110.906, 110.969, 113.938, 
 
   122.344, 120.094, 124.438, 133.844, 136.25, 127.188, 
 
   118.531, 121.406, 118.438, 122, 111.438, 112.469, 
 
   105.063, 100.781, 99.094, 79.375, 81.844, 94.188, 
 
   95.625, 99.125, 100.344, 101.063, 103.25, 104.5, 
 
   109.531, 115.281, 117.25, 101.969, 116.375, 103.344, 
 
   118.5, 104.875, 116.438, 106.344, 104.344, 106.5, 
 
   103.188, 108.094, 103.531, 105.406, 104.375, 112.531, 
 
   101.313, 106.219, 112.375, 101.938, 107.313, 102.969, 
 
   109.375, 119.156, 105.969, 113.594, 117.313, 105.781, 
 
   110.313, 100.75, 112.875, 104.031, 101.219, 109.063, 
 
   111.25, 105.094, 102.219, 109.844, 111.438, 102.844, 
 
   101.5, 105.813, 107.75, 102.563, 101.406, 103.813, 
 
   100.313, 102.625, 101.563, 100.25, 100.313, 100.094, 
 
   99.188, 99.188, 98.469, 94.938, 99.156, 90.719, 97.344, 
 
   98.156, 94.156, 98.563, 99.875, 94.219, 99.813, 96.813, 
 
   99.688, 93.219, 97.813, 89.438, 99.531, 94.125, 99.344, 
 
   96.156, 92.281, 99.188, 97.531, 99.094, 93.656, 99.031, 
 
   95.969, 91.875, 99.406, 99.344, 96.281, 100.031, 98.75, 
 
   99.469, 98.469, 96.063, 99.625, 99.156, 96.906, 99.688, 
 
   99.031, 100.438, 101.688, 102.844, 99.719, 99.219, 
 
   98.313, 99.5, 98.438, 99.563, 99.219, 98.563, 98.844, 
 
   97.656, 93.594, 98.375, 95.719, 98.688, 97.813, 97.094, 
 
   98.969, 99.344, 99.719, 100.156, 99.469, 98.875, 99.625}\\
 
   , {\"912794PY4\", \"912794PZ1\", \"912794QA5\", \"912794QB3\"\\
 
    , \"912794QC1\", \"912794QD9\", \"912794QF4\", \"912794QG2\"\\
 
    , \"912794QH0\", \"912794QJ6\", \"912794QK3\", \"912794QL1\"\\
 
    , \"912794QM9\", \"912794QN7\", \"912794QP2\", \"912794QQ0\"\\
 
    , \"912794QS6\", \"912794QT4\", \"912794QU1\", \"912794QV9\"\\
 
    , \"912794QW7\", \"912794QX5\", \"912794QY3\", \"912794QZ0\"\\
 
    , \"912794RA4\", \"912794RG1\", \"912794RM8\", \"912794RS5\"\\
 
    , \"912794RX4\", \"912794SC9\", \"912794SH8\", \"912810BB3\"\\
 
    , \"912810BC1\", \"912810BD9\", \"912810BE7\", \"912810BF4\"\\
 
    , \"912810BG2\", \"912810BN7\", \"912810BQ0\", \"912810BR8\"\\
 
    , \"912810BS6\", \"912810BT4\", \"912810BU1\", \"912810BV9\"\\
 
    , \"912810BW7\", \"912810BX5\", \"912810BY3\", \"912810BZ0\"\\
 
    , \"912810CA4\", \"912810CB2\", \"912810CC0\", \"912810CD8\"\\
 
    , \"912810CE6\", \"912810CF3\", \"912810CG1\", \"912810CH9\"\\
 
    , \"912810CJ5\", \"912810CK2\", \"912810CL0\", \"912810CM8\"\\
 
    , \"912810CN6\", \"912810CP1\", \"912810CQ9\", \"912810CR7\"\\
 
    , \"912810CS5\", \"912810CT3\", \"912810CU0\", \"912810CV8\"\\
 
    , \"912810CW6\", \"912810CX4\", \"912810CY2\", \"912810CZ9\"\\
 
    , \"912810DA3\", \"912810DB1\", \"912810DC9\", \"912810DD7\"\\
 
    , \"912810DE5\", \"912810DF2\", \"912810DG0\", \"912810DH8\"\\
 
    , \"912810DJ4\", \"912810DK1\", \"912810DL9\", \"912810DM7\"\\
 
    , \"912810DN5\", \"912810DP0\", \"912810DQ8\", \"912810DR6\"\\
 
    , \"912810DS4\", \"912810DT2\", \"912810DU9\", \"912810DV7\"\\
 
    , \"912810DW5\", \"912810DX3\", \"912810DY1\", \"912810DZ8\"\\
 
    , \"912810EA2\", \"912827JE3\", \"912827JQ6\", \"912827KC5\"\\
 
    , \"912827KY7\", \"912827LF7\", \"912827LW0\", \"912827ME9\"\\
 
    , \"912827MK5\", \"912827MN9\", \"912827MT6\", \"912827MW9\"\\
 
    , \"912827NB4\", \"912827NE8\", \"912827NK4\", \"912827NS7\"\\
 
    , \"912827NV0\", \"912827PA4\", \"912827PD8\", \"912827PJ5\"\\
 
    , \"912827PM8\", \"912827PS5\", \"912827PV8\", \"912827PX4\"\\
 
    , \"912827QA3\", \"912827QD7\", \"912827QF2\", \"912827QJ4\"\\
 
    , \"912827QN5\", \"912827QR6\", \"912827QU9\", \"912827QW5\"\\
 
    , \"912827QZ8\", \"912827RC8\", \"912827RE4\", \"912827RG9\"\\
 
    , \"912827RH7\", \"912827RM6\", \"912827RP9\", \"912827RS3\"\\
 
    , \"912827RT1\", \"912827RW4\", \"912827RY0\", \"912827SA1\"\\
 
    , \"912827SB9\", \"912827SE3\", \"912827SG8\", \"912827SK9\"\\
 
    , \"912827SL7\", \"912827SP8\", \"912827SR4\", \"912827SU7\"\\
 
    , \"912827SV5\", \"912827SX1\", \"912827SY9\", \"912827SZ6\"\\
 
    , \"912827TB8\", \"912827TC6\", \"912827TE2\", \"912827TF9\"\\
 
    , \"912827TJ1\", \"912827TL6\", \"912827TM4\", \"912827TP7\"\\
 
    , \"912827TQ5\", \"912827TS1\", \"912827TU6\", \"912827TV4\"\\
 
    , \"912827TX0\", \"912827TY8\", \"912827TZ5\", \"912827UA8\"\\
 
    , \"912827UB6\", \"912827UC4\", \"912827UD2\", \"912827UE0\"\\
 
    , \"912827UF7\", \"912827UG5\", \"912827UH3\", \"912827UJ9\"\\
 
    , \"912827UK6\", \"912827UL4\", \"912827UM2\", \"912827UN0\"\\
 
    , \"912827UP5\", \"912827UQ3\", \"912827UR1\", \"912827US9\"\\
 
    , \"912827UT7\", \"912827UU4\", \"912827UV2\", \"912827UW0\"\\
 
    , \"912827UX8\", \"912827UY6\", \"912827UZ3\", \"912827VA7\"\\
 
    , \"912827VB5\", \"912827VC3\", \"912827VD1\", \"912827VE9\"\\
 
    , \"912827VF6\", \"912827VG4\", \"912827VH2\", \"912827VJ8\"\\
 
    , \"912827VK5\", \"912827VL3\", \"912827VM1\", \"912827VN9\"\\
 
    , \"912827VP4\", \"912827VQ2\", \"912827VR0\", \"912827VS8\"\\
 
    , \"912827VT6\", \"912827VU3\", \"912827VV1\", \"912827VW9\"\\
 
    , \"912827VX7\", \"912827VY5\", \"912827VZ2\", \"912827WA6\"\\
 
    , \"912827WB4\", \"912827WC2\", \"912827WE8\", \"912827WF5\"\\
 
    , \"912827WG3\", \"912827WH1\", \"912827WJ7\", \"912827WK4\"}, 
 
  {}]\
\>", "\<\
-MD[{1988, 7, 19}]-\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  StyleBox["ReadData",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " reads the data in ",
  StyleBox["19880719.dat",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and creates an ",
  StyleBox["MD",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object, which in this case is named ",
  StyleBox["md",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " (for future reference). The ",
  StyleBox["MD",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object contains information about the bonds found in the input file. ",
  StyleBox["ReadData",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " requires the data be in a particular format, which we can see as \
follows:"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["First @ ReadList[\"19880719.dat\", String]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
\"7/19/1988,912794PY4,8/4/1988,0.5,0,99.7321\"\
\>", "\<\
7/19/1988,912794PY4,8/4/1988,0.5,0,99.7321\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  "The order of fields is quote date, I.D. number, maturity date, original \
term to maturity (in years), coupon rate, and price per \\$100 of face value \
(not including accrued interest for coupon bonds).\n\nLet's see what options \
",
  StyleBox["ReadData ",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox["takes",
    FontFamily->"Charter"],
  StyleBox[":",
    FontFamily->"Courier",
    FontWeight->"Bold"]
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["Options[ReadData]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
{Description -> {}}\
\>", "\<\
{Description -> {}}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  "One can add a description (a string or list of strings) to help keep track \
of different ",
  StyleBox["MD",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " objects. "
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "A list of information the object contains can be had by entering ",
  StyleBox["MDSelectors",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". The selectors can be used to extract information the bonds included:"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["MDSelectors", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
{CouponRate, Description, IDNumber, MaturityDate, 
 
  NumberOfSecurities, Price, QuoteDate, SettlementDate, 
 
  Term}\
\>", "\<\
{CouponRate, Description, IDNumber, MaturityDate, 
 
  NumberOfSecurities, Price, QuoteDate, SettlementDate, 
 
  Term}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell["Here are two examples:", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
NumberOfSecurities[md]
CouponRate[md] //Short\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
224\
\>", "\<\
224\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
Short[\"<<>>\"]\
\>", "\<\
{0, 0, 0, 0, 0, 0, 0, 0, <<212>>, 8.75, 8, 8.25, 8.875}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  "The next step is to construct all the payments ($c_i$) and the number of \
days to each payments ($\\tau_i$) for each of the bonds in the sample. ",
  StyleBox["ConstructPayments",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " does this. First, let's see what the options are:"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["Options[ConstructPayments]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
{FullPrice -> False, PaymentsPerYear -> 2}\
\>", "\<\
{FullPrice -> False, PaymentsPerYear -> 2}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  "These are the appropriate options for our data: U.S Treasury coupon \
securities pay coupons every six months, and our coupon-bond price data does \
not include accrued interest (which therefore must be added by ",
  StyleBox["ConstructPayments",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ").  Neither of these options has any effect on bills or STRIPs. "
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["cp = ConstructPayments[md]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
CP[{1988, 7, 19}, {{100}, {100}, {100}, {100}, {100}, 
 
   {100}, {100}, {100}, {100}, {100}, {100}, {100}, {100}, 
 
   {100}, {100}, {100}, {100}, {100}, {100}, {100}, {100}, 
 
   {100}, {100}, {100}, {100}, {100}, {100}, {100}, {100}, 
 
   {100}, {100}, {1.75, 1.75, 1.75, 101.75}, 
 
   {2.125, 2.125, 2.125, 2.125, 2.125, 2.125, 2.125, 2.125, 
 
    102.125}, {2, 2, 2, 2, 2, 2, 2, 2, 2, 102}, 
 
   {2.0625, 2.0625, 2.0625, 2.0625, 2.0625, 2.0625, 2.0625, 
 
    2.0625, 2.0625, 2.0625, 102.0625}, 
 
   {3/2, 3/2, 3/2, 3/2, 3/2, 3/2, 3/2, 3/2, 3/2, 3/2, 3/2, 
 
    3/2, 3/2, 203/2}, {1.75, 1.75, 1.75, 1.75, 1.75, 1.75, 
 
    1.75, 1.75, 1.75, 1.75, 1.75, 1.75, 1.75, 1.75, 1.75, 
 
    1.75, 1.75, 1.75, 1.75, 1.75, 101.75}, 
 
   {3.375, 3.375, 3.375, 3.375, 3.375, 3.375, 3.375, 3.375, 
 
    3.375, 103.375}, {3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 
 
    3.75, 103.75}, {4.25, 4.25, 4.25, 4.25, 4.25, 4.25, 
 
    4.25, 4.25, 4.25, 4.25, 4.25, 104.25}, 
 
   {3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 
 
    3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 
 
    103.9375}, {4.125, 4.125, 4.125, 104.125}, 
 
   {4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 
 
    4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 
 
    4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 
 
    104.125}, {4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 4.1875, 104.1875}, 
 
   {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 104}, 
 
   {3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 
 
    3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 
 
    3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 
 
    3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 
 
    103.8125}, {3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 103.625}, 
 
   {3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 
 
    3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 
 
    3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 
 
    3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 
 
    103.9375}, {3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 
 
    3.9375, 3.9375, 3.9375, 3.9375, 103.9375}, 
 
   {4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 
 
    4.3125, 4.3125, 4.3125, 104.3125}, 
 
   {4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 4.1875, 104.1875}, 
 
   {4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 
 
    4.3125, 4.3125, 4.3125, 104.3125}, 
 
   {4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 104.375}, 
 
   {9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 
 
    209/2}, {4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 104.5625}, 
 
   {4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 104.375}, 
 
   {5.0625, 5.0625, 5.0625, 5.0625, 5.0625, 5.0625, 5.0625, 
 
    5.0625, 5.0625, 5.0625, 5.0625, 5.0625, 105.0625}, 
 
   {5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 105.1875}, 
 
   {5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 
 
    5.25, 5.25, 5.25, 5.25, 105.25}, 
 
   {5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 105.875}, {6.3125, 6.3125, 6.3125, 6.3125, 
 
    6.3125, 6.3125, 6.3125, 6.3125, 6.3125, 6.3125, 6.3125, 
 
    6.3125, 6.3125, 106.3125}, 
 
   {5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 
 
    5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 105}, 
 
   {5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    105.1875}, {5.75, 5.75, 5.75, 5.75, 5.75, 5.75, 5.75, 
 
    5.75, 5.75, 5.75, 5.75, 5.75, 5.75, 5.75, 105.75}, 
 
   {6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 
 
    6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 
 
    6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 
 
    6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 
 
    6.375, 6.375, 106.375}, 
 
   {5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 105.875}, {6.5625, 6.5625, 6.5625, 6.5625, 
 
    6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 
 
    6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 
 
    6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 
 
    106.5625}, {6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 
 
    6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 
 
    6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 
 
    6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 
 
    6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 
 
    6.9375, 6.9375, 106.9375}, 
 
   {6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 
 
    6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 
 
    6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 
 
    6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 106.6875}, 
 
   {7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 
 
    7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 
 
    7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 
 
    7.875, 7.875, 107.875}, 
 
   {7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
 
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
 
    107}, {7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 
 
    7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 
 
    7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 
 
    7.125, 7.125, 7.125, 7.125, 107.125}, 
 
   {5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    105.8125}, {5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 105.1875}, 
 
   {5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 105.375}, 
 
   {5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 105.375}, 
 
   {5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 
 
    5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 
 
    5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 
 
    5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 
 
    5.5625, 5.5625, 105.5625}, 
 
   {6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 
 
    6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 
 
    6, 6, 6, 6, 106}, {5.9375, 5.9375, 5.9375, 5.9375, 
 
    5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 
 
    5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 
 
    5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 
 
    5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 105.9375}, 
 
   {6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 
 
    6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 
 
    6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 
 
    6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 
 
    6.1875, 6.1875, 6.1875, 106.1875}, 
 
   {6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 
 
    6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 
 
    6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 
 
    6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 
 
    6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 
 
    6.625, 106.625}, {6.875, 6.875, 6.875, 6.875, 6.875, 
 
    6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 
 
    6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 
 
    6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 
 
    6.875, 6.875, 6.875, 106.875}, 
 
   {6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 
 
    6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 
 
    6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 
 
    6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 
 
    6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 106.25}, 
 
   {5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    5.8125, 5.8125, 5.8125, 5.8125, 105.8125}, 
 
   {5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 105.875}, 
 
   {5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    5.625, 5.625, 5.625, 5.625, 5.625, 105.625}, 
 
   {6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 
 
    6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 106}, 
 
   {5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 105.375}, 
 
   {5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 
 
    5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 
 
    5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 
 
    5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 
 
    5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 
 
    5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 
 
    5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 
 
    5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 105.3125}, 
 
   {4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 
 
    4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 
 
    4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 
 
    4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 
 
    4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 
 
    4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 
 
    4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 
 
    4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 104.9375}, 
 
   {4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 
 
    4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 
 
    4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 
 
    4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 
 
    4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 
 
    104.6875}, {4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 
 
    4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 
 
    4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 
 
    4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 
 
    4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 
 
    4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 
 
    4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 
 
    4.625, 104.625}, {3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 103.625}, 
 
   {3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 
 
    3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 
 
    3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 
 
    3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 
 
    3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 
 
    3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 
 
    3.75, 3.75, 103.75}, 
 
   {4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 104.375}, {4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 104.4375}, 
 
   {4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 104.5625}, {104.375}, 
 
   {4.625, 104.625}, {5.375, 5.375, 105.375}, 
 
   {5.375, 5.375, 5.375, 5.375, 105.375}, 
 
   {13/2, 13/2, 13/2, 13/2, 213/2}, 
 
   {7.25, 7.25, 7.25, 7.25, 7.25, 107.25}, 
 
   {7.4375, 7.4375, 7.4375, 7.4375, 7.4375, 7.4375, 
 
    107.4375}, {107.6875}, 
 
   {7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 107.125}, 
 
   {107.3125}, {7.3125, 7.3125, 7.3125, 7.3125, 7.3125, 
 
    7.3125, 7.3125, 107.3125}, {7.1875, 107.1875}, 
 
   {6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 
 
    106.875}, {7.25, 107.25}, {5.9375, 5.9375, 105.9375}, 
 
   {5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 
 
    105.25}, {5.25, 5.25, 105.25}, 
 
   {5.4375, 5.4375, 5.4375, 5.4375, 5.4375, 5.4375, 5.4375, 
 
    5.4375, 5.4375, 105.4375}, {5.25, 5.25, 5.25, 105.25}, 
 
   {5.0625, 5.0625, 5.0625, 5.0625, 5.0625, 5.0625, 5.0625, 
 
    5.0625, 5.0625, 105.0625}, 
 
   {5.375, 5.375, 5.375, 105.375}, 
 
   {5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 
 
    5.9375, 5.9375, 5.9375, 105.9375}, {105.875}, 
 
   {5.75, 5.75, 5.75, 5.75, 105.75}, 
 
   {5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 105.875}, {5.6875, 105.6875}, 
 
   {5.875, 5.875, 5.875, 5.875, 105.875}, {5.875, 105.875}, 
 
   {6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 106.1875}, 
 
   {6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 
 
    6.5625, 6.5625, 6.5625, 6.5625, 106.5625}, 
 
   {6.9375, 6.9375, 106.9375}, 
 
   {6.875, 6.875, 6.875, 6.875, 6.875, 106.875}, 
 
   {6.3125, 6.3125, 6.3125, 6.3125, 6.3125, 6.3125, 6.3125, 
 
    6.3125, 6.3125, 6.3125, 6.3125, 6.3125, 106.3125}, 
 
   {6.375, 6.375, 106.375}, 
 
   {6.125, 6.125, 6.125, 6.125, 6.125, 6.125, 106.125}, 
 
   {105.6875}, {5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    105.8125}, {11/2, 11/2, 11/2, 211/2}, {105.3125}, 
 
   {5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    105.8125}, {5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    105.625}, {5.6875, 5.6875, 5.6875, 105.6875}, 
 
   {5.625, 105.625}, {5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 105.875}, 
 
   {5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    5.625, 5.625, 5.625, 5.625, 5.625, 105.625}, 
 
   {4.9375, 4.9375, 4.9375, 4.9375, 104.9375}, 
 
   {4.8125, 104.8125}, {5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 105.1875}, 
 
   {5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 
 
    5.25, 5.25, 5.25, 5.25, 5.25, 105.25}, 
 
   {4.8125, 4.8125, 4.8125, 4.8125, 104.8125}, 
 
   {4.6875, 4.6875, 104.6875}, 
 
   {4.875, 4.875, 4.875, 4.875, 4.875, 4.875, 4.875, 4.875, 
 
    104.875}, {104.3125}, 
 
   {4.75, 4.75, 4.75, 4.75, 4.75, 4.75, 4.75, 4.75, 4.75, 
 
    4.75, 4.75, 4.75, 4.75, 4.75, 104.75}, 
 
   {4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 104.5625}, 
 
   {4.1875, 4.1875, 104.1875}, 
 
   {4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    104.375}, {4, 104}, {4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 104.4375}, 
 
   {4.0625, 4.0625, 4.0625, 4.0625, 4.0625, 104.0625}, 
 
   {3.625, 3.625, 3.625, 103.625}, 
 
   {3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 
 
    3.6875, 3.6875, 103.6875}, {3.4375, 103.4375}, 
 
   {3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 
 
    3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 
 
    3.6875, 103.6875}, {3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 
 
    103.75}, {3.625, 3.625, 3.625, 103.625}, 
 
   {3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 103.625}, {3.3125, 3.3125, 103.3125}, 
 
   {103.0625}, {3.25, 3.25, 3.25, 3.25, 3.25, 3.25, 
 
    103.25}, {103.1875}, 
 
   {3.375, 3.375, 3.375, 3.375, 103.375}, {103.1875}, 
 
   {3.5625, 3.5625, 3.5625, 3.5625, 3.5625, 3.5625, 3.5625, 
 
    3.5625, 3.5625, 3.5625, 103.5625}, 
 
   {3.1875, 3.1875, 103.1875}, 
 
   {3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    103.625}, {103.125}, 
 
   {3.3125, 3.3125, 3.3125, 3.3125, 3.3125, 3.3125, 3.3125, 
 
    103.3125}, {103.125}, 
 
   {3.3125, 3.3125, 3.3125, 3.3125, 103.3125}, 
 
   {7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 
 
    207/2}, {3.0625, 103.0625}, {3.25, 3.25, 3.25, 103.25}, 
 
   {3.125, 103.125}, {3.3125, 3.3125, 3.3125, 3.3125, 
 
    3.3125, 3.3125, 3.3125, 103.3125}, {3.1875, 103.1875}, 
 
   {3.375, 3.375, 3.375, 3.375, 3.375, 103.375}, 
 
   {7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 
 
    207/2}, {3.5625, 103.5625}, 
 
   {3.9375, 3.9375, 3.9375, 103.9375}, 
 
   {4.25, 4.25, 4.25, 4.25, 4.25, 4.25, 4.25, 4.25, 4.25, 
 
    4.25, 4.25, 4.25, 4.25, 4.25, 4.25, 4.25, 4.25, 104.25}\\
 
    , {4, 104}, {4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 
 
    4.125, 4.125, 104.125}, {3.6875, 103.6875}, 
 
   {3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 103.9375}, 
 
   {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 104}, 
 
   {3.8125, 3.8125, 103.8125}, 
 
   {3.9375, 3.9375, 3.9375, 3.9375, 103.9375}, 
 
   {4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 
 
    4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 
 
    4.3125, 4.3125, 4.3125, 4.3125, 104.3125}, 
 
   {3.875, 3.875, 103.875}, 
 
   {4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 104.1875}, {4.25, 4.25, 104.25}, 
 
   {4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    104.5625}, {4.75, 4.75, 4.75, 4.75, 4.75, 4.75, 4.75, 
 
    4.75, 4.75, 4.75, 4.75, 4.75, 104.75}, 
 
   {3.9375, 3.9375, 103.9375}, {4, 4, 4, 4, 104}, 
 
   {4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 104.4375}, 
 
   {3.875, 3.875, 103.875}, 
 
   {4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 
 
    4.125, 104.125}, {3.9375, 3.9375, 103.9375}, 
 
   {4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 104.125}, 
 
   {4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 
 
    4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 104.3125}, 
 
   {3.6875, 3.6875, 3.6875, 103.6875}, 
 
   {3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 103.6875}, 
 
   {4.0625, 4.0625, 4.0625, 4.0625, 4.0625, 4.0625, 4.0625, 
 
    4.0625, 4.0625, 4.0625, 4.0625, 4.0625, 4.0625, 4.0625, 
 
    4.0625, 4.0625, 4.0625, 4.0625, 4.0625, 104.0625}, 
 
   {3.5625, 3.5625, 3.5625, 103.5625}, 
 
   {3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 
 
    3.8125, 3.8125, 103.8125}, 
 
   {3.6875, 3.6875, 3.6875, 103.6875}, 
 
   {3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 
 
    103.9375}, {4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 104.1875}, {3.8125, 3.8125, 3.8125, 103.8125}, 
 
   {9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 
 
    9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 209/2}, 
 
   {4.0625, 4.0625, 4.0625, 104.0625}, 
 
   {4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 104.375}, {4, 4, 4, 104}, 
 
   {4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 
 
    104.125}, {4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 104.4375}}, {{15}, {43}, {71}, {99}, {128}, 
 
   {155}, {1}, {8}, {22}, {29}, {36}, {50}, {57}, {64}, 
 
   {78}, {85}, {106}, {113}, {120}, {134}, {141}, {148}, 
 
   {162}, {169}, {176}, {211}, {239}, {267}, {295}, {323}, 
 
   {351}, {26, 210, 391, 575}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944}, {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 
 
    1672, 1853, 2036, 2217, 2401}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771}, {26, 210, 391, 575, 756, 940, 1121, 1308, 
 
    1489, 1672}, {26, 210, 391, 575, 756, 940, 1121, 1308}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126}, {26, 210, 391, 575, 756, 940, 1121, 1308, 
 
    1489, 1672, 1853, 2036, 2217, 2401}, 
 
   {118, 299, 483, 664}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231}, {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 
 
    1672}, {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 
 
    1672, 1853}, {26, 210, 391, 575, 756, 940, 1121, 1308, 
 
    1489, 1672, 1853, 2036, 2217, 2401, 2582, 2766, 2948, 
 
    3135, 3313, 3499, 3680, 3863, 4044, 4227, 4409, 4593, 
 
    4774, 4958, 5139, 5326, 5504}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944}, {118, 299, 483, 664, 848, 1029, 1213, 1395, 
 
    1580, 1762, 1944, 2126, 2309, 2490, 2674, 2856, 3040, 
 
    3221, 3407, 3586, 3771, 3953, 4135, 4317, 4501, 4682, 
 
    4866, 5047, 5231, 5412, 5598}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036}, {118, 299, 483, 664, 848, 1029, 1213, 
 
    1395, 1580, 1762, 1944, 2126, 2309, 2490, 2674, 2856, 
 
    3040, 3221, 3407, 3586, 3771, 3953, 4135, 4317, 4501, 
 
    4682, 4866, 5047, 5231, 5412, 5598, 5780}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217}, {118, 299, 483, 664, 848, 1029, 
 
    1213, 1395, 1580, 1762, 1944, 2126, 2309}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871, 6054}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144, 6327}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144, 6327, 6508}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144, 6327, 6508, 6692}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231}, {118, 299, 483, 664, 848, 1029, 1213, 1395, 
 
    1580, 1762, 1944, 2126, 2309, 2490, 2674, 2856, 3040, 
 
    3221, 3407, 3586, 3771, 3953, 4135, 4317, 4501, 4682, 
 
    4866, 5047, 5231, 5412, 5598, 5780, 5962, 6144, 6327, 
 
    6508, 6692, 6873, 7057}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326}, {118, 299, 483, 664, 848, 1029, 1213, 
 
    1395, 1580, 1762, 1944, 2126, 2309, 2490, 2674, 2856, 
 
    3040, 3221, 3407, 3586, 3771, 3953, 4135, 4317, 4501, 
 
    4682, 4866, 5047, 5231, 5412}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504}, {26, 210, 391, 575, 756, 940, 1121, 
 
    1308, 1489, 1672, 1853, 2036, 2217, 2401, 2582, 2766, 
 
    2948, 3135, 3313, 3499, 3680, 3863, 4044, 4227, 4409, 
 
    4593, 4774, 4958, 5139, 5326, 5504, 5690, 5871, 6054, 
 
    6235, 6419, 6600, 6784, 6965, 7149, 7331}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598}, {118, 299, 483, 664, 848, 1029, 
 
    1213, 1395, 1580, 1762, 1944, 2126, 2309, 2490, 2674, 
 
    2856, 3040, 3221, 3407, 3586, 3771, 3953, 4135, 4317, 
 
    4501, 4682, 4866, 5047, 5231, 5412, 5598, 5780}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144, 6327, 6508, 6692, 
 
    6873, 7057, 7239, 7425, 7604}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871, 6054, 6235, 6419, 6600, 
 
    6784, 6965, 7149, 7331, 7517, 7698}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144, 6327, 6508, 6692, 
 
    6873, 7057, 7239, 7425, 7604, 7789}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871, 6054, 6235, 6419, 6600, 
 
    6784, 6965, 7149, 7331, 7517, 7698, 7881, 8062, 8245, 
 
    8426, 8610, 8792, 8976, 9157, 9344, 9522, 9708}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871, 6054, 6235}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871, 6054, 6235, 6419, 6600, 
 
    6784, 6965, 7149, 7331, 7517, 7698, 7881, 8062, 8245, 
 
    8426, 8610, 8792, 8976, 9157, 9344, 9522, 9708, 9889}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144, 6327, 6508, 6692, 
 
    6873, 7057, 7239, 7425, 7604, 7789, 7971, 8153, 8335, 
 
    8518, 8700, 8884, 9065, 9249, 9430, 9616, 9795, 9980}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871, 6054, 6235, 6419}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871, 6054, 6235, 6419, 6600, 
 
    6784, 6965, 7149, 7331, 7517, 7698, 7881, 8062, 8245, 
 
    8426, 8610, 8792, 8976, 9157, 9344, 9522, 9708, 9889, 
 
    10072}, {118, 299, 483, 664, 848, 1029, 1213, 1395, 
 
    1580, 1762, 1944, 2126, 2309, 2490, 2674, 2856, 3040, 
 
    3221, 3407, 3586, 3771, 3953, 4135, 4317, 4501, 4682, 
 
    4866, 5047, 5231, 5412, 5598, 5780, 5962, 6144, 6327, 
 
    6508, 6692, 6873, 7057, 7239, 7425, 7604, 7789, 7971, 
 
    8153, 8335, 8518, 8700, 8884, 9065, 9249, 9430, 9616, 
 
    9795, 9980, 10162}, {118, 299, 483, 664, 848, 1029, 
 
    1213, 1395, 1580, 1762, 1944, 2126, 2309, 2490, 2674, 
 
    2856, 3040, 3221, 3407, 3586, 3771, 3953, 4135, 4317, 
 
    4501, 4682, 4866, 5047, 5231, 5412, 5598, 5780, 5962, 
 
    6144, 6327, 6508, 6692, 6873, 7057, 7239, 7425, 7604, 
 
    7789, 7971, 8153, 8335, 8518, 8700, 8884, 9065, 9249, 
 
    9430, 9616, 9795, 9980, 10162, 10345}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144, 6327, 6508, 6692, 
 
    6873, 7057, 7239, 7425, 7604, 7789, 7971, 8153, 8335, 
 
    8518, 8700, 8884, 9065, 9249, 9430, 9616, 9795, 9980, 
 
    10162, 10345, 10526}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871, 6054, 6235, 6419, 6600, 
 
    6784, 6965, 7149, 7331, 7517, 7698, 7881, 8062, 8245, 
 
    8426, 8610, 8792, 8976, 9157, 9344, 9522, 9708, 9889, 
 
    10072, 10253, 10437, 10618}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144, 6327, 6508, 6692, 
 
    6873, 7057, 7239, 7425, 7604, 7789, 7971, 8153, 8335, 
 
    8518, 8700, 8884, 9065, 9249, 9430, 9616, 9795, 9980, 
 
    10162, 10345, 10526, 10710, 10891}, {118}, {118, 299}, 
 
   {118, 299, 483}, {26, 210, 391, 575, 756}, 
 
   {118, 299, 483, 664, 848}, 
 
   {118, 299, 483, 664, 848, 1029}, 
 
   {26, 210, 391, 575, 756, 940, 1121}, {89}, 
 
   {118, 299, 483, 664, 848, 1029, 1213}, {181}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308}, {89, 271}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395}, {181, 362}, 
 
   {89, 271, 453}, {118, 299, 483, 664, 848, 1029, 1213, 
 
    1395, 1580}, {181, 362, 545}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672}, 
 
   {89, 271, 453, 635}, {118, 299, 483, 664, 848, 1029, 
 
    1213, 1395, 1580, 1762}, {181, 362, 545, 726}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853}, {118}, {89, 271, 453, 635, 817}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944}, {26, 210}, {181, 362, 545, 726, 909}, 
 
   {118, 299}, {89, 271, 453, 635, 817, 999}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126}, {26, 210, 391}, 
 
   {181, 362, 545, 726, 909, 1090}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217}, {118, 299, 483}, 
 
   {89, 271, 453, 635, 817, 999, 1182}, {72}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309}, {26, 210, 391, 575}, {167}, 
 
   {181, 362, 545, 726, 909, 1090, 1274}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401}, {118, 299, 483, 664}, 
 
   {72, 254}, {89, 271, 453, 635, 817, 999, 1182, 1365}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490}, {26, 210, 391, 575, 756}, 
 
   {167, 345}, {181, 362, 545, 726, 909, 1090, 1274, 1456}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582}, 
 
   {118, 299, 483, 664, 848}, {72, 254, 439}, 
 
   {89, 271, 453, 635, 817, 999, 1182, 1365, 1548}, {118}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674}, 
 
   {26, 210, 391, 575, 756, 940}, {167, 345, 531}, 
 
   {181, 362, 545, 726, 909, 1090, 1274, 1456, 1640}, 
 
   {26, 210}, {26, 210, 391, 575, 756, 940, 1121, 1308, 
 
    1489, 1672, 1853, 2036, 2217, 2401, 2582, 2766}, 
 
   {118, 299, 483, 664, 848, 1029}, {72, 254, 439, 621}, 
 
   {89, 271, 453, 635, 817, 999, 1182, 1365, 1548, 1730}, 
 
   {118, 299}, {118, 299, 483, 664, 848, 1029, 1213, 1395, 
 
    1580, 1762, 1944, 2126, 2309, 2490, 2674, 2856}, 
 
   {26, 210, 391, 575, 756, 940, 1121}, 
 
   {167, 345, 531, 712}, 
 
   {181, 362, 545, 726, 909, 1090, 1274, 1456, 1640, 1821}, 
 
   {26, 210, 391}, {42}, 
 
   {118, 299, 483, 664, 848, 1029, 1213}, {72}, 
 
   {72, 254, 439, 621, 803}, {103}, 
 
   {89, 271, 453, 635, 817, 999, 1182, 1365, 1548, 1730, 
 
    1913}, {118, 299, 483}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040}, {133}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308}, {167}, 
 
   {167, 345, 531, 712, 894}, 
 
   {181, 362, 545, 726, 909, 1090, 1274, 1456, 1640, 1821, 
 
    2008}, {12, 195}, {26, 210, 391, 575}, {42, 223}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395}, {72, 254}, 
 
   {72, 254, 439, 621, 803, 985}, 
 
   {89, 271, 453, 635, 817, 999, 1182, 1365, 1548, 1730, 
 
    1913, 2095}, {103, 285}, {118, 299, 483, 664}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221}, 
 
   {133, 315}, {26, 210, 391, 575, 756, 940, 1121, 1308, 
 
    1489}, {167, 345}, {167, 345, 531, 712, 894, 1076}, 
 
   {181, 362, 545, 726, 909, 1090, 1274, 1456, 1640, 1821, 
 
    2008, 2186}, {12, 195, 376}, {26, 210, 391, 575, 756}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313}, 
 
   {42, 223, 407}, {118, 299, 483, 664, 848, 1029, 1213, 
 
    1395, 1580}, {72, 254, 439}, 
 
   {72, 254, 439, 621, 803, 985, 1167}, 
 
   {89, 271, 453, 635, 817, 999, 1182, 1365, 1548, 1730, 
 
    1913, 2095, 2280}, {103, 285, 468}, 
 
   {118, 299, 483, 664, 848}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407}, 
 
   {133, 315, 498}, {26, 210, 391, 575, 756, 940, 1121, 
 
    1308, 1489, 1672}, {167, 345, 531}, 
 
   {167, 345, 531, 712, 894, 1076, 1259}, 
 
   {181, 362, 545, 726, 909, 1090, 1274, 1456, 1640, 1821, 
 
    2008, 2186, 2372}, {12, 195, 376, 560}, 
 
   {26, 210, 391, 575, 756, 940}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499}, {42, 223, 407, 588}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762}, 
 
   {72, 254, 439, 621}, {72, 254, 439, 621, 803, 985, 1167, 
 
    1350}, {89, 271, 453, 635, 817, 999, 1182, 1365, 1548, 
 
    1730, 1913, 2095, 2280, 2462}, {103, 285, 468, 649}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586}, {133, 315, 498, 680}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853}, {167, 345, 531, 712}, 
 
   {167, 345, 531, 712, 894, 1076, 1259, 1441}, 
 
   {181, 362, 545, 726, 909, 1090, 1274, 1456, 1640, 1821, 
 
    2008, 2186, 2372, 2553}}, 
 
  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
 
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 155, 155, 155, 65, 
 
   155, 65, 155, 155, 65, 155, 65, 65, 155, 155, 155, 155, 
 
   65, 155, 155, 155, 65, 65, 155, 65, 155, 65, 65, 155, 
 
   155, 65, 65, 65, 65, 65, 155, 65, 65, 155, 65, 65, 155, 
 
   65, 65, 155, 65, 155, 155, 65, 65, 65, 155, 155, 65, 65, 
 
   155, 65, 155, 155, 65, 155, 155, 65, 65, 65, 155, 65, 
 
   65, 65, 65, 155, 65, 65, 155, 96, 65, 5, 155, 96, 65, 5, 
 
   96, 65, 5, 155, 96, 65, 5, 155, 65, 96, 65, 155, 5, 65, 
 
   96, 65, 155, 5, 155, 65, 96, 111, 65, 155, 20, 5, 155, 
 
   65, 111, 96, 65, 155, 20, 5, 155, 65, 111, 96, 65, 65, 
 
   155, 20, 5, 155, 155, 65, 111, 96, 65, 65, 155, 20, 5, 
 
   155, 142, 65, 111, 111, 79, 96, 65, 65, 50, 155, 20, 20, 
 
   5, 170, 155, 142, 65, 111, 111, 96, 79, 65, 65, 50, 155, 
 
   20, 20, 5, 170, 155, 155, 142, 65, 111, 111, 96, 79, 65, 
 
   65, 50, 155, 20, 20, 5, 170, 155, 155, 142, 65, 111, 
 
   111, 96, 79, 65, 50, 155, 20, 20, 5}, 
 
  {2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 
 
   2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 
 
   2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 
 
   2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 
 
   2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 
 
   2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 
 
   2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 
 
   2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 
 
   2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 
 
   2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 
 
   2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 
 
   2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2}, False, 
 
  Hold[md]]\
\>", "\<\
-CP[{1988, 7, 19}]-\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  StyleBox["ConstructPayments",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " calculates the number of days to each coupon payment for each bond in the \
sample, using information in ",
  StyleBox["daycount.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". It takes an ",
  StyleBox["MD",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object as an argument and returns a ",
  StyleBox["CP",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object that contains a reference to the ",
  StyleBox["MD",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object used to create it. Thus, ",
  StyleBox["CouponRate[cp]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " will return the same list of coupon rates for the bonds in the sample. ",
  StyleBox["CPSelectors",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is a list of information stored in ",
  StyleBox["CP",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " objects."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["CPSelectors", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
{AccruedDays, AccruedInterest, DaysInPeriod, 
 
  DaysToPayments, FullPriceQ, LastPayment, NextPayment, 
 
  Payments, PaymentsPerYear, QuoteDate, RemainingPayments}\
\>", "\<\
{AccruedDays, AccruedInterest, DaysInPeriod, 
 
  DaysToPayments, FullPriceQ, LastPayment, NextPayment, 
 
  Payments, PaymentsPerYear, QuoteDate, RemainingPayments}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["\<\
DaysToPayments[cp] //Shallow
Payments[cp] //Shallow
CouponRate[cp] //Shallow\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
Shallow[{{15}, {43}, {71}, {99}, {128}, {155}, {1}, {8}, 
 
   {22}, {29}, {36}, {50}, {57}, {64}, {78}, {85}, {106}, 
 
   {113}, {120}, {134}, {141}, {148}, {162}, {169}, {176}, 
 
   {211}, {239}, {267}, {295}, {323}, {351}, 
 
   {26, 210, 391, 575}, {26, 210, 391, 575, 756, 940, 1121, 
 
    1308, 1489}, {26, 210, 391, 575, 756, 940, 1121, 1308, 
 
    1489, 1672}, {118, 299, 483, 664, 848, 1029, 1213, 
 
    1395, 1580, 1762, 1944}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771}, {26, 210, 391, 575, 756, 940, 1121, 1308, 
 
    1489, 1672}, {26, 210, 391, 575, 756, 940, 1121, 1308}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126}, {26, 210, 391, 575, 756, 940, 1121, 1308, 
 
    1489, 1672, 1853, 2036, 2217, 2401}, 
 
   {118, 299, 483, 664}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231}, {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 
 
    1672}, {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 
 
    1672, 1853}, {26, 210, 391, 575, 756, 940, 1121, 1308, 
 
    1489, 1672, 1853, 2036, 2217, 2401, 2582, 2766, 2948, 
 
    3135, 3313, 3499, 3680, 3863, 4044, 4227, 4409, 4593, 
 
    4774, 4958, 5139, 5326, 5504}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944}, {118, 299, 483, 664, 848, 1029, 1213, 1395, 
 
    1580, 1762, 1944, 2126, 2309, 2490, 2674, 2856, 3040, 
 
    3221, 3407, 3586, 3771, 3953, 4135, 4317, 4501, 4682, 
 
    4866, 5047, 5231, 5412, 5598}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036}, {118, 299, 483, 664, 848, 1029, 1213, 
 
    1395, 1580, 1762, 1944, 2126, 2309, 2490, 2674, 2856, 
 
    3040, 3221, 3407, 3586, 3771, 3953, 4135, 4317, 4501, 
 
    4682, 4866, 5047, 5231, 5412, 5598, 5780}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217}, {118, 299, 483, 664, 848, 1029, 
 
    1213, 1395, 1580, 1762, 1944, 2126, 2309}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871, 6054}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144, 6327}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144, 6327, 6508}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144, 6327, 6508, 6692}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231}, {118, 299, 483, 664, 848, 1029, 1213, 1395, 
 
    1580, 1762, 1944, 2126, 2309, 2490, 2674, 2856, 3040, 
 
    3221, 3407, 3586, 3771, 3953, 4135, 4317, 4501, 4682, 
 
    4866, 5047, 5231, 5412, 5598, 5780, 5962, 6144, 6327, 
 
    6508, 6692, 6873, 7057}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326}, {118, 299, 483, 664, 848, 1029, 1213, 
 
    1395, 1580, 1762, 1944, 2126, 2309, 2490, 2674, 2856, 
 
    3040, 3221, 3407, 3586, 3771, 3953, 4135, 4317, 4501, 
 
    4682, 4866, 5047, 5231, 5412}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504}, {26, 210, 391, 575, 756, 940, 1121, 
 
    1308, 1489, 1672, 1853, 2036, 2217, 2401, 2582, 2766, 
 
    2948, 3135, 3313, 3499, 3680, 3863, 4044, 4227, 4409, 
 
    4593, 4774, 4958, 5139, 5326, 5504, 5690, 5871, 6054, 
 
    6235, 6419, 6600, 6784, 6965, 7149, 7331}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598}, {118, 299, 483, 664, 848, 1029, 
 
    1213, 1395, 1580, 1762, 1944, 2126, 2309, 2490, 2674, 
 
    2856, 3040, 3221, 3407, 3586, 3771, 3953, 4135, 4317, 
 
    4501, 4682, 4866, 5047, 5231, 5412, 5598, 5780}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144, 6327, 6508, 6692, 
 
    6873, 7057, 7239, 7425, 7604}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871, 6054, 6235, 6419, 6600, 
 
    6784, 6965, 7149, 7331, 7517, 7698}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144, 6327, 6508, 6692, 
 
    6873, 7057, 7239, 7425, 7604, 7789}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871, 6054, 6235, 6419, 6600, 
 
    6784, 6965, 7149, 7331, 7517, 7698, 7881, 8062, 8245, 
 
    8426, 8610, 8792, 8976, 9157, 9344, 9522, 9708}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871, 6054, 6235}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871, 6054, 6235, 6419, 6600, 
 
    6784, 6965, 7149, 7331, 7517, 7698, 7881, 8062, 8245, 
 
    8426, 8610, 8792, 8976, 9157, 9344, 9522, 9708, 9889}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144, 6327, 6508, 6692, 
 
    6873, 7057, 7239, 7425, 7604, 7789, 7971, 8153, 8335, 
 
    8518, 8700, 8884, 9065, 9249, 9430, 9616, 9795, 9980}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871, 6054, 6235, 6419}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871, 6054, 6235, 6419, 6600, 
 
    6784, 6965, 7149, 7331, 7517, 7698, 7881, 8062, 8245, 
 
    8426, 8610, 8792, 8976, 9157, 9344, 9522, 9708, 9889, 
 
    10072}, {118, 299, 483, 664, 848, 1029, 1213, 1395, 
 
    1580, 1762, 1944, 2126, 2309, 2490, 2674, 2856, 3040, 
 
    3221, 3407, 3586, 3771, 3953, 4135, 4317, 4501, 4682, 
 
    4866, 5047, 5231, 5412, 5598, 5780, 5962, 6144, 6327, 
 
    6508, 6692, 6873, 7057, 7239, 7425, 7604, 7789, 7971, 
 
    8153, 8335, 8518, 8700, 8884, 9065, 9249, 9430, 9616, 
 
    9795, 9980, 10162}, {118, 299, 483, 664, 848, 1029, 
 
    1213, 1395, 1580, 1762, 1944, 2126, 2309, 2490, 2674, 
 
    2856, 3040, 3221, 3407, 3586, 3771, 3953, 4135, 4317, 
 
    4501, 4682, 4866, 5047, 5231, 5412, 5598, 5780, 5962, 
 
    6144, 6327, 6508, 6692, 6873, 7057, 7239, 7425, 7604, 
 
    7789, 7971, 8153, 8335, 8518, 8700, 8884, 9065, 9249, 
 
    9430, 9616, 9795, 9980, 10162, 10345}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144, 6327, 6508, 6692, 
 
    6873, 7057, 7239, 7425, 7604, 7789, 7971, 8153, 8335, 
 
    8518, 8700, 8884, 9065, 9249, 9430, 9616, 9795, 9980, 
 
    10162, 10345, 10526}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
 
    5139, 5326, 5504, 5690, 5871, 6054, 6235, 6419, 6600, 
 
    6784, 6965, 7149, 7331, 7517, 7698, 7881, 8062, 8245, 
 
    8426, 8610, 8792, 8976, 9157, 9344, 9522, 9708, 9889, 
 
    10072, 10253, 10437, 10618}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586, 3771, 3953, 4135, 4317, 4501, 4682, 4866, 5047, 
 
    5231, 5412, 5598, 5780, 5962, 6144, 6327, 6508, 6692, 
 
    6873, 7057, 7239, 7425, 7604, 7789, 7971, 8153, 8335, 
 
    8518, 8700, 8884, 9065, 9249, 9430, 9616, 9795, 9980, 
 
    10162, 10345, 10526, 10710, 10891}, {118}, {118, 299}, 
 
   {118, 299, 483}, {26, 210, 391, 575, 756}, 
 
   {118, 299, 483, 664, 848}, 
 
   {118, 299, 483, 664, 848, 1029}, 
 
   {26, 210, 391, 575, 756, 940, 1121}, {89}, 
 
   {118, 299, 483, 664, 848, 1029, 1213}, {181}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308}, {89, 271}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395}, {181, 362}, 
 
   {89, 271, 453}, {118, 299, 483, 664, 848, 1029, 1213, 
 
    1395, 1580}, {181, 362, 545}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672}, 
 
   {89, 271, 453, 635}, {118, 299, 483, 664, 848, 1029, 
 
    1213, 1395, 1580, 1762}, {181, 362, 545, 726}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853}, {118}, {89, 271, 453, 635, 817}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944}, {26, 210}, {181, 362, 545, 726, 909}, 
 
   {118, 299}, {89, 271, 453, 635, 817, 999}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126}, {26, 210, 391}, 
 
   {181, 362, 545, 726, 909, 1090}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217}, {118, 299, 483}, 
 
   {89, 271, 453, 635, 817, 999, 1182}, {72}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309}, {26, 210, 391, 575}, {167}, 
 
   {181, 362, 545, 726, 909, 1090, 1274}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401}, {118, 299, 483, 664}, 
 
   {72, 254}, {89, 271, 453, 635, 817, 999, 1182, 1365}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490}, {26, 210, 391, 575, 756}, 
 
   {167, 345}, {181, 362, 545, 726, 909, 1090, 1274, 1456}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582}, 
 
   {118, 299, 483, 664, 848}, {72, 254, 439}, 
 
   {89, 271, 453, 635, 817, 999, 1182, 1365, 1548}, {118}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674}, 
 
   {26, 210, 391, 575, 756, 940}, {167, 345, 531}, 
 
   {181, 362, 545, 726, 909, 1090, 1274, 1456, 1640}, 
 
   {26, 210}, {26, 210, 391, 575, 756, 940, 1121, 1308, 
 
    1489, 1672, 1853, 2036, 2217, 2401, 2582, 2766}, 
 
   {118, 299, 483, 664, 848, 1029}, {72, 254, 439, 621}, 
 
   {89, 271, 453, 635, 817, 999, 1182, 1365, 1548, 1730}, 
 
   {118, 299}, {118, 299, 483, 664, 848, 1029, 1213, 1395, 
 
    1580, 1762, 1944, 2126, 2309, 2490, 2674, 2856}, 
 
   {26, 210, 391, 575, 756, 940, 1121}, 
 
   {167, 345, 531, 712}, 
 
   {181, 362, 545, 726, 909, 1090, 1274, 1456, 1640, 1821}, 
 
   {26, 210, 391}, {42}, 
 
   {118, 299, 483, 664, 848, 1029, 1213}, {72}, 
 
   {72, 254, 439, 621, 803}, {103}, 
 
   {89, 271, 453, 635, 817, 999, 1182, 1365, 1548, 1730, 
 
    1913}, {118, 299, 483}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040}, {133}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308}, {167}, 
 
   {167, 345, 531, 712, 894}, 
 
   {181, 362, 545, 726, 909, 1090, 1274, 1456, 1640, 1821, 
 
    2008}, {12, 195}, {26, 210, 391, 575}, {42, 223}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395}, {72, 254}, 
 
   {72, 254, 439, 621, 803, 985}, 
 
   {89, 271, 453, 635, 817, 999, 1182, 1365, 1548, 1730, 
 
    1913, 2095}, {103, 285}, {118, 299, 483, 664}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221}, 
 
   {133, 315}, {26, 210, 391, 575, 756, 940, 1121, 1308, 
 
    1489}, {167, 345}, {167, 345, 531, 712, 894, 1076}, 
 
   {181, 362, 545, 726, 909, 1090, 1274, 1456, 1640, 1821, 
 
    2008, 2186}, {12, 195, 376}, {26, 210, 391, 575, 756}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313}, 
 
   {42, 223, 407}, {118, 299, 483, 664, 848, 1029, 1213, 
 
    1395, 1580}, {72, 254, 439}, 
 
   {72, 254, 439, 621, 803, 985, 1167}, 
 
   {89, 271, 453, 635, 817, 999, 1182, 1365, 1548, 1730, 
 
    1913, 2095, 2280}, {103, 285, 468}, 
 
   {118, 299, 483, 664, 848}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407}, 
 
   {133, 315, 498}, {26, 210, 391, 575, 756, 940, 1121, 
 
    1308, 1489, 1672}, {167, 345, 531}, 
 
   {167, 345, 531, 712, 894, 1076, 1259}, 
 
   {181, 362, 545, 726, 909, 1090, 1274, 1456, 1640, 1821, 
 
    2008, 2186, 2372}, {12, 195, 376, 560}, 
 
   {26, 210, 391, 575, 756, 940}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
 
    3499}, {42, 223, 407, 588}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762}, 
 
   {72, 254, 439, 621}, {72, 254, 439, 621, 803, 985, 1167, 
 
    1350}, {89, 271, 453, 635, 817, 999, 1182, 1365, 1548, 
 
    1730, 1913, 2095, 2280, 2462}, {103, 285, 468, 649}, 
 
   {118, 299, 483, 664, 848, 1029, 1213, 1395, 1580, 1762, 
 
    1944, 2126, 2309, 2490, 2674, 2856, 3040, 3221, 3407, 
 
    3586}, {133, 315, 498, 680}, 
 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
 
    1853}, {167, 345, 531, 712}, 
 
   {167, 345, 531, 712, 894, 1076, 1259, 1441}, 
 
   {181, 362, 545, 726, 909, 1090, 1274, 1456, 1640, 1821, 
 
    2008, 2186, 2372, 2553}}]\
\>", "\<\
{{15}, {43}, {71}, {99}, {128}, {155}, {1}, {8}, {22}, 
 
  {29}, <<214>>}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
Shallow[{{100}, {100}, {100}, {100}, {100}, {100}, {100}, 
 
   {100}, {100}, {100}, {100}, {100}, {100}, {100}, {100}, 
 
   {100}, {100}, {100}, {100}, {100}, {100}, {100}, {100}, 
 
   {100}, {100}, {100}, {100}, {100}, {100}, {100}, {100}, 
 
   {1.75, 1.75, 1.75, 101.75}, 
 
   {2.125, 2.125, 2.125, 2.125, 2.125, 2.125, 2.125, 2.125, 
 
    102.125}, {2, 2, 2, 2, 2, 2, 2, 2, 2, 102}, 
 
   {2.0625, 2.0625, 2.0625, 2.0625, 2.0625, 2.0625, 2.0625, 
 
    2.0625, 2.0625, 2.0625, 102.0625}, 
 
   {3/2, 3/2, 3/2, 3/2, 3/2, 3/2, 3/2, 3/2, 3/2, 3/2, 3/2, 
 
    3/2, 3/2, 203/2}, {1.75, 1.75, 1.75, 1.75, 1.75, 1.75, 
 
    1.75, 1.75, 1.75, 1.75, 1.75, 1.75, 1.75, 1.75, 1.75, 
 
    1.75, 1.75, 1.75, 1.75, 1.75, 101.75}, 
 
   {3.375, 3.375, 3.375, 3.375, 3.375, 3.375, 3.375, 3.375, 
 
    3.375, 103.375}, {3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 
 
    3.75, 103.75}, {4.25, 4.25, 4.25, 4.25, 4.25, 4.25, 
 
    4.25, 4.25, 4.25, 4.25, 4.25, 104.25}, 
 
   {3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 
 
    3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 
 
    103.9375}, {4.125, 4.125, 4.125, 104.125}, 
 
   {4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 
 
    4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 
 
    4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 
 
    104.125}, {4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 4.1875, 104.1875}, 
 
   {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 104}, 
 
   {3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 
 
    3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 
 
    3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 
 
    3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 
 
    103.8125}, {3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 103.625}, 
 
   {3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 
 
    3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 
 
    3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 
 
    3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 
 
    103.9375}, {3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 
 
    3.9375, 3.9375, 3.9375, 3.9375, 103.9375}, 
 
   {4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 
 
    4.3125, 4.3125, 4.3125, 104.3125}, 
 
   {4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 4.1875, 104.1875}, 
 
   {4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 
 
    4.3125, 4.3125, 4.3125, 104.3125}, 
 
   {4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 104.375}, 
 
   {9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 
 
    209/2}, {4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 104.5625}, 
 
   {4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 104.375}, 
 
   {5.0625, 5.0625, 5.0625, 5.0625, 5.0625, 5.0625, 5.0625, 
 
    5.0625, 5.0625, 5.0625, 5.0625, 5.0625, 105.0625}, 
 
   {5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 105.1875}, 
 
   {5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 
 
    5.25, 5.25, 5.25, 5.25, 105.25}, 
 
   {5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 105.875}, {6.3125, 6.3125, 6.3125, 6.3125, 
 
    6.3125, 6.3125, 6.3125, 6.3125, 6.3125, 6.3125, 6.3125, 
 
    6.3125, 6.3125, 106.3125}, 
 
   {5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 
 
    5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 105}, 
 
   {5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    105.1875}, {5.75, 5.75, 5.75, 5.75, 5.75, 5.75, 5.75, 
 
    5.75, 5.75, 5.75, 5.75, 5.75, 5.75, 5.75, 105.75}, 
 
   {6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 
 
    6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 
 
    6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 
 
    6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 6.375, 
 
    6.375, 6.375, 106.375}, 
 
   {5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 105.875}, {6.5625, 6.5625, 6.5625, 6.5625, 
 
    6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 
 
    6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 
 
    6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 
 
    106.5625}, {6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 
 
    6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 
 
    6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 
 
    6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 
 
    6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 6.9375, 
 
    6.9375, 6.9375, 106.9375}, 
 
   {6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 
 
    6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 
 
    6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 
 
    6.6875, 6.6875, 6.6875, 6.6875, 6.6875, 106.6875}, 
 
   {7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 
 
    7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 
 
    7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 7.875, 
 
    7.875, 7.875, 107.875}, 
 
   {7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
 
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
 
    107}, {7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 
 
    7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 
 
    7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 
 
    7.125, 7.125, 7.125, 7.125, 107.125}, 
 
   {5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    105.8125}, {5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 5.1875, 5.1875, 105.1875}, 
 
   {5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 105.375}, 
 
   {5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 105.375}, 
 
   {5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 
 
    5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 
 
    5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 
 
    5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 5.5625, 
 
    5.5625, 5.5625, 105.5625}, 
 
   {6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 
 
    6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 
 
    6, 6, 6, 6, 106}, {5.9375, 5.9375, 5.9375, 5.9375, 
 
    5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 
 
    5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 
 
    5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 
 
    5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 105.9375}, 
 
   {6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 
 
    6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 
 
    6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 
 
    6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 
 
    6.1875, 6.1875, 6.1875, 106.1875}, 
 
   {6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 
 
    6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 
 
    6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 
 
    6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 
 
    6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 6.625, 
 
    6.625, 106.625}, {6.875, 6.875, 6.875, 6.875, 6.875, 
 
    6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 
 
    6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 
 
    6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 
 
    6.875, 6.875, 6.875, 106.875}, 
 
   {6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 
 
    6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 
 
    6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 
 
    6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 
 
    6.25, 6.25, 6.25, 6.25, 6.25, 6.25, 106.25}, 
 
   {5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    5.8125, 5.8125, 5.8125, 5.8125, 105.8125}, 
 
   {5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 105.875}, 
 
   {5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    5.625, 5.625, 5.625, 5.625, 5.625, 105.625}, 
 
   {6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 
 
    6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 106}, 
 
   {5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 5.375, 
 
    5.375, 5.375, 105.375}, 
 
   {5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 
 
    5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 
 
    5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 
 
    5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 
 
    5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 
 
    5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 
 
    5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 
 
    5.3125, 5.3125, 5.3125, 5.3125, 5.3125, 105.3125}, 
 
   {4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 
 
    4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 
 
    4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 
 
    4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 
 
    4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 
 
    4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 
 
    4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 
 
    4.9375, 4.9375, 4.9375, 4.9375, 4.9375, 104.9375}, 
 
   {4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 
 
    4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 
 
    4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 
 
    4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 
 
    4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 4.6875, 
 
    104.6875}, {4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 
 
    4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 
 
    4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 
 
    4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 
 
    4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 
 
    4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 
 
    4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 4.625, 
 
    4.625, 104.625}, {3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 103.625}, 
 
   {3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 
 
    3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 
 
    3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 
 
    3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 
 
    3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 
 
    3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 
 
    3.75, 3.75, 103.75}, 
 
   {4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 104.375}, {4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 104.4375}, 
 
   {4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    4.5625, 4.5625, 4.5625, 104.5625}, {104.375}, 
 
   {4.625, 104.625}, {5.375, 5.375, 105.375}, 
 
   {5.375, 5.375, 5.375, 5.375, 105.375}, 
 
   {13/2, 13/2, 13/2, 13/2, 213/2}, 
 
   {7.25, 7.25, 7.25, 7.25, 7.25, 107.25}, 
 
   {7.4375, 7.4375, 7.4375, 7.4375, 7.4375, 7.4375, 
 
    107.4375}, {107.6875}, 
 
   {7.125, 7.125, 7.125, 7.125, 7.125, 7.125, 107.125}, 
 
   {107.3125}, {7.3125, 7.3125, 7.3125, 7.3125, 7.3125, 
 
    7.3125, 7.3125, 107.3125}, {7.1875, 107.1875}, 
 
   {6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 6.875, 
 
    106.875}, {7.25, 107.25}, {5.9375, 5.9375, 105.9375}, 
 
   {5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 
 
    105.25}, {5.25, 5.25, 105.25}, 
 
   {5.4375, 5.4375, 5.4375, 5.4375, 5.4375, 5.4375, 5.4375, 
 
    5.4375, 5.4375, 105.4375}, {5.25, 5.25, 5.25, 105.25}, 
 
   {5.0625, 5.0625, 5.0625, 5.0625, 5.0625, 5.0625, 5.0625, 
 
    5.0625, 5.0625, 105.0625}, 
 
   {5.375, 5.375, 5.375, 105.375}, 
 
   {5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 5.9375, 
 
    5.9375, 5.9375, 5.9375, 105.9375}, {105.875}, 
 
   {5.75, 5.75, 5.75, 5.75, 105.75}, 
 
   {5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 105.875}, {5.6875, 105.6875}, 
 
   {5.875, 5.875, 5.875, 5.875, 105.875}, {5.875, 105.875}, 
 
   {6.1875, 6.1875, 6.1875, 6.1875, 6.1875, 106.1875}, 
 
   {6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 6.5625, 
 
    6.5625, 6.5625, 6.5625, 6.5625, 106.5625}, 
 
   {6.9375, 6.9375, 106.9375}, 
 
   {6.875, 6.875, 6.875, 6.875, 6.875, 106.875}, 
 
   {6.3125, 6.3125, 6.3125, 6.3125, 6.3125, 6.3125, 6.3125, 
 
    6.3125, 6.3125, 6.3125, 6.3125, 6.3125, 106.3125}, 
 
   {6.375, 6.375, 106.375}, 
 
   {6.125, 6.125, 6.125, 6.125, 6.125, 6.125, 106.125}, 
 
   {105.6875}, {5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    105.8125}, {11/2, 11/2, 11/2, 211/2}, {105.3125}, 
 
   {5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 5.8125, 
 
    105.8125}, {5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    105.625}, {5.6875, 5.6875, 5.6875, 105.6875}, 
 
   {5.625, 105.625}, {5.875, 5.875, 5.875, 5.875, 5.875, 
 
    5.875, 5.875, 105.875}, 
 
   {5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 5.625, 
 
    5.625, 5.625, 5.625, 5.625, 5.625, 105.625}, 
 
   {4.9375, 4.9375, 4.9375, 4.9375, 104.9375}, 
 
   {4.8125, 104.8125}, {5.1875, 5.1875, 5.1875, 5.1875, 
 
    5.1875, 5.1875, 5.1875, 105.1875}, 
 
   {5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 5.25, 
 
    5.25, 5.25, 5.25, 5.25, 5.25, 105.25}, 
 
   {4.8125, 4.8125, 4.8125, 4.8125, 104.8125}, 
 
   {4.6875, 4.6875, 104.6875}, 
 
   {4.875, 4.875, 4.875, 4.875, 4.875, 4.875, 4.875, 4.875, 
 
    104.875}, {104.3125}, 
 
   {4.75, 4.75, 4.75, 4.75, 4.75, 4.75, 4.75, 4.75, 4.75, 
 
    4.75, 4.75, 4.75, 4.75, 4.75, 104.75}, 
 
   {4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 104.5625}, 
 
   {4.1875, 4.1875, 104.1875}, 
 
   {4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    104.375}, {4, 104}, {4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 104.4375}, 
 
   {4.0625, 4.0625, 4.0625, 4.0625, 4.0625, 104.0625}, 
 
   {3.625, 3.625, 3.625, 103.625}, 
 
   {3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 
 
    3.6875, 3.6875, 103.6875}, {3.4375, 103.4375}, 
 
   {3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 
 
    3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 
 
    3.6875, 103.6875}, {3.75, 3.75, 3.75, 3.75, 3.75, 3.75, 
 
    103.75}, {3.625, 3.625, 3.625, 103.625}, 
 
   {3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 103.625}, {3.3125, 3.3125, 103.3125}, 
 
   {103.0625}, {3.25, 3.25, 3.25, 3.25, 3.25, 3.25, 
 
    103.25}, {103.1875}, 
 
   {3.375, 3.375, 3.375, 3.375, 103.375}, {103.1875}, 
 
   {3.5625, 3.5625, 3.5625, 3.5625, 3.5625, 3.5625, 3.5625, 
 
    3.5625, 3.5625, 3.5625, 103.5625}, 
 
   {3.1875, 3.1875, 103.1875}, 
 
   {3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 3.625, 
 
    103.625}, {103.125}, 
 
   {3.3125, 3.3125, 3.3125, 3.3125, 3.3125, 3.3125, 3.3125, 
 
    103.3125}, {103.125}, 
 
   {3.3125, 3.3125, 3.3125, 3.3125, 103.3125}, 
 
   {7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 
 
    207/2}, {3.0625, 103.0625}, {3.25, 3.25, 3.25, 103.25}, 
 
   {3.125, 103.125}, {3.3125, 3.3125, 3.3125, 3.3125, 
 
    3.3125, 3.3125, 3.3125, 103.3125}, {3.1875, 103.1875}, 
 
   {3.375, 3.375, 3.375, 3.375, 3.375, 103.375}, 
 
   {7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 7/2, 
 
    207/2}, {3.5625, 103.5625}, 
 
   {3.9375, 3.9375, 3.9375, 103.9375}, 
 
   {4.25, 4.25, 4.25, 4.25, 4.25, 4.25, 4.25, 4.25, 4.25, 
 
    4.25, 4.25, 4.25, 4.25, 4.25, 4.25, 4.25, 4.25, 104.25}\\
 
    , {4, 104}, {4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 
 
    4.125, 4.125, 104.125}, {3.6875, 103.6875}, 
 
   {3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 103.9375}, 
 
   {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 104}, 
 
   {3.8125, 3.8125, 103.8125}, 
 
   {3.9375, 3.9375, 3.9375, 3.9375, 103.9375}, 
 
   {4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 
 
    4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 
 
    4.3125, 4.3125, 4.3125, 4.3125, 104.3125}, 
 
   {3.875, 3.875, 103.875}, 
 
   {4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 104.1875}, {4.25, 4.25, 104.25}, 
 
   {4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 4.5625, 
 
    104.5625}, {4.75, 4.75, 4.75, 4.75, 4.75, 4.75, 4.75, 
 
    4.75, 4.75, 4.75, 4.75, 4.75, 104.75}, 
 
   {3.9375, 3.9375, 103.9375}, {4, 4, 4, 4, 104}, 
 
   {4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 104.4375}, 
 
   {3.875, 3.875, 103.875}, 
 
   {4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 
 
    4.125, 104.125}, {3.9375, 3.9375, 103.9375}, 
 
   {4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 104.125}, 
 
   {4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 
 
    4.3125, 4.3125, 4.3125, 4.3125, 4.3125, 104.3125}, 
 
   {3.6875, 3.6875, 3.6875, 103.6875}, 
 
   {3.6875, 3.6875, 3.6875, 3.6875, 3.6875, 103.6875}, 
 
   {4.0625, 4.0625, 4.0625, 4.0625, 4.0625, 4.0625, 4.0625, 
 
    4.0625, 4.0625, 4.0625, 4.0625, 4.0625, 4.0625, 4.0625, 
 
    4.0625, 4.0625, 4.0625, 4.0625, 4.0625, 104.0625}, 
 
   {3.5625, 3.5625, 3.5625, 103.5625}, 
 
   {3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 3.8125, 
 
    3.8125, 3.8125, 103.8125}, 
 
   {3.6875, 3.6875, 3.6875, 103.6875}, 
 
   {3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 3.9375, 
 
    103.9375}, {4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 4.1875, 
 
    4.1875, 104.1875}, {3.8125, 3.8125, 3.8125, 103.8125}, 
 
   {9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 
 
    9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 9/2, 209/2}, 
 
   {4.0625, 4.0625, 4.0625, 104.0625}, 
 
   {4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 4.375, 
 
    4.375, 4.375, 104.375}, {4, 4, 4, 104}, 
 
   {4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 4.125, 
 
    104.125}, {4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
    4.4375, 104.4375}}]\
\>", "\<\
{{100}, {100}, {100}, {100}, {100}, {100}, {100}, {100}, 
 
  {100}, {100}, <<214>>}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
Shallow[{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
 
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.5, 4.25, 4, 
 
   4.125, 3, 3.5, 6.75, 7.5, 8.5, 7.875, 8.25, 8.25, 8.375, 
 
   8, 7.625, 7.25, 7.875, 7.875, 8.625, 8.375, 8.625, 8.75, 
 
   9, 9.125, 8.75, 10.125, 10.375, 10.5, 11.75, 12.625, 10, 
 
   10.375, 11.5, 12.75, 11.75, 13.125, 13.875, 13.375, 
 
   15.75, 14, 14.25, 11.625, 10.375, 10.75, 10.75, 11.125, 
 
   12, 11.875, 12.375, 13.25, 13.75, 12.5, 11.625, 11.75, 
 
   11.25, 12, 10.75, 10.625, 9.875, 9.375, 9.25, 7.25, 7.5, 
 
   8.75, 8.875, 9.125, 8.75, 9.25, 10.75, 10.75, 13, 14.5, 
 
   14.875, 15.375, 14.25, 14.625, 14.625, 14.375, 13.75, 
 
   14.5, 11.875, 10.5, 10.5, 10.875, 10.5, 10.125, 10.75, 
 
   11.875, 11.75, 11.5, 11.75, 11.375, 11.75, 11.75, 
 
   12.375, 13.125, 13.875, 13.75, 12.625, 12.75, 12.25, 
 
   11.375, 11.625, 11, 10.625, 11.625, 11.25, 11.375, 
 
   11.25, 11.75, 11.25, 9.875, 9.625, 10.375, 10.5, 9.625, 
 
   9.375, 9.75, 8.625, 9.5, 9.125, 8.375, 8.75, 8, 8.875, 
 
   8.125, 7.25, 7.375, 6.875, 7.375, 7.5, 7.25, 7.25, 
 
   6.625, 6.125, 6.5, 6.375, 6.75, 6.375, 7.125, 6.375, 
 
   7.25, 6.25, 6.625, 6.25, 6.625, 7, 6.125, 6.5, 6.25, 
 
   6.625, 6.375, 6.75, 7, 7.125, 7.875, 8.5, 8, 8.25, 
 
   7.375, 7.875, 8, 7.625, 7.875, 8.625, 7.75, 8.375, 8.5, 
 
   9.125, 9.5, 7.875, 8, 8.875, 7.75, 8.25, 7.875, 8.25, 
 
   8.625, 7.375, 7.375, 8.125, 7.125, 7.625, 7.375, 7.875, 
 
   8.375, 7.625, 9, 8.125, 8.75, 8, 8.25, 8.875}]\
\>", "\<\
{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, <<214>>}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell["\<\
We see that each of the first 10 securities is a Treasury bill with one \
remaining payment of 100.\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "We are now ready to fit the term structure. ",
  StyleBox["DiscountFunction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is the main estimation function. It takes a ",
  StyleBox["CP",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object as an argument and returns a ",
  StyleBox["DF",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object. The default functional form is ",
  StyleBox["DLogDelta",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", which indicates that the forward rates should be splined. The default ",
  StyleBox["LambdaValue",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is ",
  StyleBox["Automatic",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", which indicates the generalized cross validation should be used to \
determine the value of $\\lambda$, the weight on the penalty. The ",
  StyleBox["DF",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object contains the results of the estimation, including for example the \
estimated spline (",
  StyleBox["FinalSpline[df]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ") and the difference between the actual and fitted bond prices (",
  StyleBox["PriceError[df]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ")."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "First look at the options available for ",
  StyleBox["DiscountFunction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ":"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["Options[DiscountFunction]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
{BetaConvergenceTest -> 
 
   (Max[Abs[(#2 - #1)/#2]] < 0.0001 & ), 
 
  Bounds -> {0., 1.*10^30}, DropByTerm -> 0, 
 
  FunctionalForm -> DLogDelta, IDNumberDropList -> {}, 
 
  InitialCurve -> Automatic, Knots -> Automatic, 
 
  LambdaValue -> Automatic, Lambda0 -> Automatic, 
 
  MaximumMaturity -> 30, ObservationWeights -> Automatic, 
 
  OvernightForwardRate -> Automatic, Restriction -> True, 
 
  ShowLambdaProgress -> True, SplineOrder -> 3, 
 
  TuningParameter -> 2}\
\>", "\<\
                                 #2 - #1
{BetaConvergenceTest -> (Max[Abs[-------]] < 0.0001 & ), 
                                   #2
 
                      30
  Bounds -> {0., 1. 10  }, DropByTerm -> 0, 
 
  FunctionalForm -> DLogDelta, IDNumberDropList -> {}, 
 
  InitialCurve -> Automatic, Knots -> Automatic, 
 
  LambdaValue -> Automatic, Lambda0 -> Automatic, 
 
  MaximumMaturity -> 30, ObservationWeights -> Automatic, 
 
  OvernightForwardRate -> Automatic, Restriction -> True, 
 
  ShowLambdaProgress -> True, SplineOrder -> 3, 
 
  TuningParameter -> 2}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  "In addition to the options listed, one can pass options for ",
  StyleBox["FindMinimum",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " (such as ",
  StyleBox["AccuracyGoal -> 10",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ")."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["It's always a good idea to read the usage statement.", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["?DiscountFunction", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
DiscountFunction[cp] estimates a discount function for a
   given day.  Run MassageData and ConstructPayments first
   to create the CP object. The default options are Knots ->
   (number of securities)/3, DropByTerm -> 0 (drop the 0
   most recently issued securities of each term),
   MaximumMaturity -> 30. The default FunctionalForm is
   DLogDelta, which uses a nonlinear routine to estimate the
   forward rate function; other functional forms are
   LogDelta and Delta. If DLogDelta or LogDelta is used,
   DiscountFunction takes options to set the starting values
   (Lambda0, Bounds, and options to tune FindMinimum) and
   looks for a file named \"LStart.m\" or \"DLStart.m\" to find
   starting values for beta0 and lambda0.  If the
   appropriate file cannot be opened, internal default
   values are used.\
\>", "Print",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["\<\
df1 = DiscountFunction[cp, FunctionalForm -> Delta,
\tLambdaValue -> 10^10]\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
DF[{1988, 7, 19}, InterpolatingFunction[{0, 10958}, 
 
   {{0, 0, {1, 0}, {0}}, 
 
    {15, 0, {0.997136753264484, -0.0001911119010071377, 
 
      -(1.525235373969461*10^-8), 
 
      -(4.945649619425313*10^-10)}, {0, 0, 15}}, 
 
    {29, 15, {0.994455423515109, -0.0001920294669717979, 
 
      -(3.613667464165587*10^-8), 
 
      -(4.809230882970708*10^-10)}, {0, 0, 14}}, 
 
    {42, 29, {0.991950629575721, -0.0001934131319119361, 
 
      -(5.666731045685442*10^-8), 
 
      -(5.30681252898861*10^-10)}, {0, 0, 13}}, 
 
    {50, 42, {0.990399080805426, -0.0001944959957027453, 
 
      -(6.904992698626711*10^-8), 
 
      -(3.427350151732553*10^-10)}, {0, 0, 8}}, 
 
    {64, 50, {0.98766153978868, -0.000196618855455008, 
 
      -(7.715795726217943*10^-8), 
 
      -(1.91648219787499*10^-10)}, {0, 0, 14}}, 
 
    {78, 64, {0.98489331303928, -0.0001988359609371992, 
 
      -(7.896277408665551*10^-8), 3.136636614731838*10^-11}\\
 
      , {0, 0, 14}}, {89, 78, 
 
     {0.98269696925952, -0.0002004671670564686, 
 
      -(7.26864286057695*10^-8), 2.653280160536767*10^-10}, 
 
     {0, 0, 11}}, {103, 89, 
 
     {0.979877547200048, -0.0002022507862879247, 
 
      -(6.167830897450201*10^-8), 2.889111233162191*10^-10}\\
 
      , {0, 0, 14}}, {113, 103, 
 
     {0.977849745155008, -0.0002032627054035924, 
 
      -(4.825008996317494*10^-8), 4.691731640417111*10^-10}\\
 
      , {0, 0, 10}}, {120, 113, 
 
     {0.97642484782695, -0.000203839964138211, 
 
      -(4.045756956858957*10^-8), 2.214849113962189*10^-10}\\
 
      , {0, 0, 7}}, {134, 120, 
 
     {0.97356199486079, -0.0002052438640193013, 
 
      -(5.388331689079227*10^-8), 
 
      -(5.348622033259436*10^-10)}, {0, 0, 14}}, 
 
    {148, 134, {0.970675839901628, -0.0002071148494964883, 
 
      -(6.863741728770955*10^-8), 
 
      -(2.595010554533857*10^-10)}, {0, 0, 14}}, 
 
    {162, 148, {0.96776215627344, -0.000209119292426882, 
 
      -(7.135951933169347*10^-8), 3.253259748076187*10^-11}\\
 
      , {0, 0, 14}}, {169, 162, 
 
     {0.966294931566166, -0.0002100756755374286, 
 
      -(6.744941810873302*10^-8), 2.467603474361268*10^-10}\\
 
      , {0, 0, 7}}, {181, 169, 
 
     {0.963765340067427, -0.0002114578584616785, 
 
      -(5.488057501029192*10^-8), 4.517300277973546*10^-10}\\
 
      , {0, 0, 12}}, {210, 181, 
 
     {0.957602336202621, -0.0002132020723966885, 
 
      -(2.361026717903999*10^-8), 4.456818533004951*10^-10}\\
 
      , {0, 0, 29}}, {239, 210, 
 
     {0.951410873537634, -0.0002137821136542727, 
 
      -(9.77245063888844*10^-9), 1.574211715653979*10^-11}, 
 
     {0, 0, 29}}, {267, 239, 
 
     {0.94541972557898, -0.0002140836365284345, 
 
      -(4.073806304252517*10^-9), 9.36093381679287*10^-11}, 
 
     {0, 0, 28}}, {285, 267, 
 
     {0.941567977524919, -0.0002137645865464103, 
 
      1.230091427922183*10^-8, 3.820460865058548*10^-10}, 
 
     {0, 0, 18}}, {299, 285, 
 
     {0.938580398052152, -0.0002129349166279139, 
 
      3.311551007791862*10^-8, 4.977773657410357*10^-10}, 
 
     {0, 0, 14}}, {323, 299, 
 
     {0.933484913817754, -0.0002120277214807467, 
 
      1.183840079987721*10^-8, -(5.884581749532306*10^-10)}\\
 
      , {0, 0, 24}}, {362, 323, 
 
     {0.925204837149732, -0.0002127844322811645, 
 
      -(1.217369555773236*10^-8), 
 
      -(1.267833353617836*10^-10)}, {0, 0, 39}}, 
 
    {391, 362, {0.919030143920371, -0.000212696074868771, 
 
      7.737286761544017*10^-9, 4.28543662040369*10^-10}, 
 
     {0, 0, 29}}, {439, 391, 
 
     {0.908891435266451, -0.0002092450636745138, 
 
      4.120902671121521*10^-8, 2.192080599043111*10^-10}, 
 
     {0, 0, 48}}, {468, 439, 
 
     {0.902874112359699, -0.0002054917520538553, 
 
      6.903935456655388*10^-8, 2.984196720589672*10^-10}, 
 
     {0, 0, 29}}, {498, 468, 
 
     {0.896771883337335, -0.0002015702038803421, 
 
      6.124767328157361*10^-8, -(2.740975295763581*10^-10)}\\
 
      , {0, 0, 30}}, {560, 498, 
 
     {0.884311397739626, -0.0002030738490475248, 
 
      -(3.384314340204647*10^-8), 
 
      -(7.005475064180912*10^-10)}, {0, 0, 62}}, 
 
    {588, 560, {0.878559483910796, -0.0002079650250789711, 
 
      -(9.06975425773383*10^-8), 
 
      -(2.396509603120658*10^-10)}, {0, 0, 28}}, 
 
    {635, 588, {0.868646599835331, -0.0002122290660898102, 
 
      -(2.801359472888333*10^-8), 7.382359014595473*10^-10}\\
 
      , {0, 0, 47}}, {664, 635, 
 
     {0.862505715286365, -0.0002109996079654765, 
 
      2.603557427700025*10^-8, 3.336565799432183*10^-10}, 
 
     {0, 0, 29}}, {712, 664, 
 
     {0.852447992209391, -0.0002083226339045316, 
 
      2.527632359224703*10^-8, -(1.087009531636727*10^-10)}\\
 
      , {0, 0, 48}}, {756, 712, 
 
     {0.843330558309179, -0.0002058805344848924, 
 
      3.031528041143414*10^-8, 1.165523019457371*10^-10}, 
 
     {0, 0, 44}}, {848, 756, 
 
     {0.824782457922585, -0.0001963291009787491, 
 
      5.739875904415*10^-8, 1.193216160184649*10^-10}, 
 
     {0, 0, 92}}, {909, 848, 
 
     {0.812997511358164, -0.0001911002912084068, 
 
      3.435334606501861*10^-8, -(2.78877062727763*10^-10)}, 
 
     {0, 0, 61}}, {985, 909, 
 
     {0.7985442716882978, -0.0001896400163686629, 
 
      7.028813339288115*10^-9, -(6.784889407410921*10^-11)}\\
 
      , {0, 0, 76}}, {1029, 985, 
 
     {0.7902028806412329, -0.0001895335561022497, 
 
      9.8893520965637*10^-10, -(1.0038206592707*10^-11)}, 
 
     {0, 0, 44}}, {1090, 1029, 
 
     {0.7785705306446583, -0.0001930490570135829, 
 
      -(3.860319302713529*10^-8), 
 
      -(3.209053044819451*10^-10)}, {0, 0, 61}}, 
 
    {1182, 1090, {0.7605189130980225, 
 
      -0.0001971891742532763, -(1.060804402949659*10^-8), 
 
      2.585346335380403*10^-10}, {0, 0, 92}}, 
 
    {1259, 1182, {0.7457709514864344, 
 
      -0.0001812321983921507, 1.337632544092757*10^-7, 
 
      7.83026702294301*10^-10}, {0, 0, 77}}, 
 
    {1308, 1259, {0.737469815691123, 
 
      -0.0001552771662240845, 2.884442525328205*10^-7, 
 
      9.63142265785176*10^-10}, {0, 0, 49}}, 
 
    {1365, 1308, {0.7299178275453904, 
 
      -0.0001060501065962534, 4.638756755143434*10^-7, 
 
      1.124889929456463*10^-9}, {0, 0, 57}}, 
 
    {1441, 1365, {0.7230704525207801, 
 
      -0.0000983184807135007, -(1.081768541578516*10^-7), 
 
      -(4.185337208231676*10^-9)}, {0, 0, 76}}, 
 
    {1548, 1441, {0.7090377956838268, 
 
      -0.0001511919297377756, -(1.87342094941801*10^-7), 
 
      1.116450406736792*10^-9}, {0, 0, 107}}, 
 
    {1640, 1548, {0.6937972013986043, 
 
      -0.0001883469061988007, -(2.466116593888419*10^-7), 
 
      -(9.71357380260143*10^-10)}, {0, 0, 92}}, 
 
    {1730, 1640, {0.6751502294126883, 
 
      -0.0002146340320127185, -(8.27272710158804*10^-8), 
 
      1.406940374204845*10^-9}, {0, 0, 90}}, 
 
    {1821, 1730, {0.6584590796601514, 
 
      -0.0001249842801264193, 6.421422848729224*10^-7, 
 
      3.287060012145*10^-9}, {0, 0, 91}}, 
 
    {1913, 1821, {0.649433492075028, 
 
      -0.0001309404350224741, -(3.569154580510666*10^-7), 
 
      -(7.055327195810657*10^-9)}, {0, 0, 92}}, 
 
    {2008, 1913, {0.6294476877808847, 
 
      -0.0002736792636691301, -(6.663408038142987*10^-7), 
 
      1.787709243428167*10^-9}, {0, 0, 95}}, 
 
    {2126, 2008, {0.5974506463973095, 
 
      -0.0002075375814027339, 5.391846292769729*10^-7, 
 
      4.388529893921961*10^-9}, {0, 0, 118}}, 
 
    {2217, 2126, {0.5855770221469041, 
 
      -0.00007255283456094731, 6.365555253422518*10^-7, 
 
      -(2.310305667129192*10^-9)}, {0, 0, 91}}, 
 
    {2309, 2217, {0.5818885036816115, 
 
      -0.000014393336206538, 2.793397370381752*10^-7, 
 
      -(7.987933293224281*10^-10)}, {0, 0, 92}}, 
 
    {2401, 2309, {0.5762735060387168, 
 
      -0.0001732493458798832, -(1.219747421792829*10^-6), 
 
      -(7.747816155072707*10^-9)}, {0, 0, 92}}, 
 
    {2490, 2401, {0.5522337269630455, 
 
      -0.0002918343007871369, -(2.440946464314911*10^-7), 
 
      9.48568461588792*10^-9}, {0, 0, 89}}, 
 
    {2674, 2490, {0.5131816352640289, 
 
      -0.0001634744411491121, 2.650281937494081*10^-7, 
 
      -(9.10606224546392*10^-10)}, {0, 0, 184}}, 
 
    {2856, 2674, {0.4876090795038187, 
 
      -0.000112317512115627, 1.548958022933822*10^-7, 
 
      1.577449281882044*10^-10}, {0, 0, 182}}, 
 
    {3040, 2856, {0.4618349089814889, 
 
      -0.0002293794066708402, -(4.85339092187639*10^-7), 
 
      -(1.817783889704572*10^-9)}, {0, 0, 184}}, 
 
    {3313, 3040, {0.4141542964699432, 
 
      0.0001586045310043961, 1.220728165891744*10^-6, 
 
      3.737251820119092*10^-9}, {0, 0, 273}}, 
 
    {3499, 3313, {0.4838711546982038, 
 
      0.0003904307815078612, 8.39191563244773*10^-8, 
 
      -(5.798598807687582*10^-9)}, {0, 0, 186}}, 
 
    {4317, 3499, {0.3693345647025215, 
 
      -0.0003873230650020456, -(3.023261416060268*10^-7), 
 
      4.231626407697932*10^-10}, {0, 0, 818}}, 
 
    {4682, 4317, {0.2649545660699382, 
 
      -0.0000992662949444825, 5.115241206819081*10^-7, 
 
      6.406893453948555*10^-10}, {0, 0, 365}}, 
 
    {4866, 4682, {0.2603882152724313, 
 
      -0.00001306791645170593, 6.385438830319585*10^-8, 
 
      -(1.851960172412584*10^-9)}, {0, 0, 184}}, 
 
    {5231, 4866, {0.2370765908668375, 
 
      -0.00006439576602276809, -(1.447402459873114*10^-9), 
 
      3.773409328230795*10^-10}, {0, 0, 365}}, 
 
    {5412, 5231, {0.2273198391337867, 
 
      -0.00005758963395857263, -(2.035871961938877*10^-8), 
 
      -(4.327092752484413*10^-10)}, {0, 0, 181}}, 
 
    {5598, 5412, {0.2148642141298121, 
 
      -0.00006736359725659486, -(2.139093703089188*10^-9), 
 
      2.595161417641646*10^-10}, {0, 0, 186}}, 
 
    {5962, 5598, {0.1942505191118603, 
 
      -0.00005195754704818219, 1.283923961790151*10^-8, 
 
      -(4.573031462519894*10^-11)}, {0, 0, 364}}, 
 
    {6144, 5962, {0.1856533313049291, 
 
      -0.00003710399255824092, 5.56774894738378*10^-8, 
 
      1.634178142294315*10^-10}, {0, 0, 182}}, 
 
    {6327, 6144, {0.1788428839362492, 
 
      -0.00005307046735303491, -(8.6638841318807*10^-8), 
 
      -(4.701048442142008*10^-10)}, {0, 0, 183}}, 
 
    {6508, 6327, {0.1662328774898671, 
 
      -0.00007161178396831048, -(1.073613295937524*10^-8), 
 
      4.47325676382947*10^-10}, {0, 0, 181}}, 
 
    {7057, 6508, {0.1415518654864529, 
 
      -0.0000302015182566245, 2.687575184066182*10^-8, 
 
      -(3.948457433996338*10^-11)}, {0, 0, 549}}, 
 
    {7698, 7057, {0.1243036394027507, 
 
      -0.00002365427041402241, 5.076503290037461*10^-9, 
 
      -(9.5333258958351*10^-14)}, {0, 0, 641}}, 
 
    {9708, 7698, {0.0853528219819148, 
 
      -0.00002090795085006354, -(7.609127961663857*10^-10), 
 
      -(1.436892404779023*10^-12)}, {0, 0, 2010}}, 
 
    {9980, 9708, {0.0801657668151943, 
 
      -0.00001440171948378013, 1.716300101552073*10^-8, 
 
      3.825747710531821*10^-11}, {0, 0, 272}}, 
 
    {10162, 9980, {0.07781416789167448, 
 
      -0.0000149767449823449, -(1.129599877028574*10^-8), 
 
      -(1.067720702155308*10^-10)}, {0, 0, 182}}, 
 
    {10526, 10162, {0.07229967695707826, 
 
      -(4.310431484597094*10^-6), 2.977821122300201*10^-8, 
 
      8.31136356765311*10^-11}, {0, 0, 364}}, 
 
    {10958, 10526, {0.07779984484213509, 
 
      0.00002088283305088896, 1.886796978377486*10^-8, 
 
      -(4.764306114060667*10^-11)}, {0, 0, 432}}}], 
 
  1.*10^10, 31.81202656402795, Delta, 
 
  {0.01842467355157851, 0.05028474650929127, 
 
   0.06036936514267666, 0.07513902819643193, 
 
   0.06396357948167975, 0.07674941373379341, 
 
   0.002076651748112113, 0.02424046632638976, 
 
   0.0462305869765487, 0.07595764848913689, 
 
   0.07410653876597451, 0.06149191945738153, 
 
   0.07539746983809437, 0.05694602113206316, 
 
   0.07846869607203644, 0.0871295338374409, 
 
   0.07077111931198488, 0.07432548449919807, 
 
   0.07081521730505358, 0.07920051392100902, 
 
   0.0930808394206508, 0.0875160098372589, 
 
   0.1097843726559802, 0.0891068433834192, 
 
   0.0858663601884899, 0.08798765395278, 
 
   0.1121126462366107, 0.1477274421019814, 
 
   0.1731252788796382, 0.2184086182246432, 
 
   0.1964436654278501, 1.150315653025757, 
 
   7.604986054798673, 10.48488762513718, 12.15455376031697, 
 
   21.584925563785, 18.08056669957446, -0.6812909230907565, 
 
   -1.460669112323458, -2.824464281817356, 
 
   -6.266500663772945, 0.3249694558785876, 
 
   -7.010275207332584, -2.839161888492953, 
 
   -1.649410742414915, -0.89506667882894, 
 
   0.0992722467570815, -0.3152641871066351, 
 
   -0.722636692820359, -0.5243923578023697, 
 
   -0.4397621021363563, -0.982570331099552, 
 
   -0.2486489798126286, -0.4730184002587663, 
 
   0.06167161833980117, 0.804883196560908, 
 
   -0.991784021122825, -0.3557156240933779, 
 
   -1.979884017073331, -0.5821744284344704, 
 
   -0.7505246279949489, -0.1772515584945324, 
 
   -0.811943095614453, 0.7671809053288002, 
 
   -1.029357774420589, 0.879405240727436, 
 
   1.206573720878225, -0.7756626627326284, 
 
   0.899513638236982, 0.1530576709328955, 
 
   -0.2893364653031938, 0.004123486634256324, 
 
   0.2541619555491508, 0.3563376493822972, 
 
   0.3214397388902057, 0.2280602268904346, 
 
   0.1159817009945527, -0.003483201662234592, 
 
   0.1126268891761555, 0.148891886166524, 
 
   -0.419438304059554, 0.2305375632278412, 
 
   -0.3458075921106456, 0.3705779044719577, 
 
   0.6294341036454512, -0.04043080585272208, 
 
   0.4003647559051018, 0.3887066353080968, 
 
   0.03232216983106184, -0.06346365936478549, 
 
   1.767770116757873, 0.0541149869806219, 
 
   0.01564046342741676, 0.06290417715797503, 
 
   -0.07162357622672744, -0.0422435743427343, 
 
   0.04199267731901557, -0.05891690713227149, 
 
   -0.01106429625512817, 0.04651060126086293, 
 
   0.0650478880959895, 0.0951873583594818, 
 
   0.4636857192246993, 0.4323378844187431, 
 
   0.1340093120546584, 0.3755604542615742, 
 
   0.1165045246594332, 0.4613277216259633, 
 
   0.2962066574627897, -0.4028987262163355, 
 
   0.3233749733207247, 0.005461968006940765, 
 
   -1.306419899839071, 0.0192393210381141, 
 
   -1.227558745432674, -0.0442052594317488, 
 
   -0.882406446683021, 0.07698150516804958, 
 
   -1.18374531915498, -0.02237857741835114, 
 
   0.04604053402908903, -1.632906505694336, 
 
   -0.0573254611772569, 0.0929593226723711, 
 
   -0.05534118572538205, 0.02551206077775703, 
 
   0.1285535599885605, 0.0847722884538911, 
 
   0.05202081856238295, -0.1037718355992041, 
 
   0.1174807976417895, 0.1430188386356264, 
 
   -0.01690431471568843, -1.203329747487515, 
 
   -0.0247735540027918, -0.01997245821168292, 
 
   0.2638136750121873, -2.067564975159186, 
 
   -0.05349229241114984, -0.06970886695481227, 
 
   -0.1615647469482155, -0.97828035820686, 
 
   0.0452487481762347, 0.01067010473897767, 
 
   -0.86087985157414, 0.01571027981728434, 
 
   0.1257864442567467, 0.03270062143575103, 
 
   -1.260902603276605, -0.0510643375370421, 
 
   0.5729905693754631, 0.1086185909866231, 
 
   0.01754026523549612, -1.218310000893922, 
 
   -0.0513531567299026, 0.7801214769899901, 
 
   0.1090334524078997, 0.0399317523698528, 
 
   -0.913270938622219, -0.06530125125840414, 
 
   0.7745479327757181, 0.1420440232153624, 
 
   0.110668125028937, -0.7167108717823539, 
 
   -0.01572885991075168, 0.005539326308777959, 
 
   0.2759166783835525, -0.005089019392272576, 
 
   0.1343255753126442, -0.03952923686975396, 
 
   -1.413653240802191, -0.00917420332243069, 
 
   1.094839611422728, -0.03491508311722669, 
 
   0.3142611010511729, -0.01424101296822755, 
 
   0.1695100601798458, -0.892714352958876, 
 
   -0.02899468335367317, -0.007520029785652583, 
 
   -0.06089516591123357, -0.3169924468504206, 
 
   -0.02137069741881703, 0.1152971283845829, 
 
   0.1129301540645855, -0.04958754421463141, 
 
   0.01094486567336617, 3.524002281709471, 
 
   -0.03002676596203457, -0.943965689256785, 
 
   -0.00817650132584902, 0.149998154482347, 
 
   0.3319471527679667, -0.01902446175132866, 
 
   0.0971364283597183, 2.633353956685028, 
 
   0.02333348680805614, -1.207062691837052, 
 
   0.04322183428152471, 0.2280929902034359, 
 
   -0.6311124680854335, 0.01996480839382286, 
 
   0.1514823004294925, -1.515866047089801, 
 
   0.00938847208500704, -1.089751949396899, 
 
   0.000223622633626519, 0.4113086517503746, 
 
   -1.707238389529806, -0.0263619179655592, 
 
   0.1802223518850496, -6.551871974344735, 
 
   -0.05884091048218921, -0.829913538122426, 
 
   0.06466693948976854, 0.0978544125766092, 
 
   -1.19016879162055, -0.01200000584540817, 
 
   -10.21720167575844, 0.02664207357886994, 
 
   -0.87571362554668, 0.0994750918532077, 
 
   -0.6453070683324143, 0.1365284523310493}, 
 
  {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 
 
   17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 
 
   31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 
 
   45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 
 
   59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 
 
   73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 
 
   87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 
 
   101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 
 
   112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 
 
   123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 
 
   134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 
 
   145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 
 
   156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 
 
   167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 
 
   178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 
 
   189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 
 
   200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 
 
   211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 
 
   222, 223, 224}, 7.530029069597913, 0.825380445338856, 
 
  {0.006708538277442905, 0.01822060922760724, 
 
   0.02176331379909124, 0.02701240827182584, 
 
   0.02297223899820656, 0.0275510619621934, 
 
   0.000756769518582951, 0.00883148862838116, 
 
   0.01681729466845157, 0.02759881505438859, 
 
   0.02689036635158406, 0.02225000899354962, 
 
   0.02724463954712539, 0.02055170604061749, 
 
   0.02826179219151831, 0.03135709119920683, 
 
   0.02543297270834561, 0.026703185430057, 
 
   0.02543710049346579, 0.02844165621183149, 
 
   0.03342244999065274, 0.03142043820449283, 
 
   0.03940332090358046, 0.0319760086386387, 
 
   0.03080707924662311, 0.03153110881530892, 
 
   0.04013779764811759, 0.05283845379321779, 
 
   0.06186995523259386, 0.07799226609631165, 
 
   0.07008368183172045, 0.4090993701842099, 
 
   2.662477281373051, 3.679369673658229, 4.232102578363416, 
 
   7.565315349124307, 5.565076460105609, 
 
   -0.2388106689780588, -0.5122877481136918, 
 
   -0.974841547313816, -2.192106229919894, 
 
   0.1152634746626367, -1.976453549022247, 
 
   -0.981306439162123, -0.5344497399378985, 
 
   -0.2933974376110134, 0.03471133239750939, 
 
   -0.1026955013185971, -0.2531796798003088, 
 
   -0.1828090475291465, -0.1482897665842114, 
 
   -0.3413670210369768, -0.0830928909555277, 
 
   -0.1634983936710362, 0.02080955934770178, 
 
   0.2785351822220748, -0.3446786628410084, 
 
   -0.1194113852158606, -0.6917324553535316, 
 
   -0.1962989617172348, -0.2612563419672709, 
 
   -0.0590967382596332, -0.2829569269794298, 
 
   0.2587434752758347, -0.3352795321614395, 
 
   0.2776045088173667, 0.3868159084314066, 
 
   -0.2465752698304024, 0.2971829054714786, 
 
   0.05006901373014366, -0.0881233807402883, 
 
   0.001343608238343144, 0.0823135815333364, 
 
   0.1058622086386895, 0.1066085443370816, 
 
   0.07580567542732429, 0.03900604359223441, 
 
   -0.001079926623055573, 0.03752516659246521, 
 
   0.0500761937319928, -0.1365055020449445, 
 
   0.07710623398355116, -0.1095879714823678, 
 
   0.1242387735255789, 0.1917394258095982, 
 
   -0.01114142083889274, 0.1332159090532702, 
 
   0.1289226570499122, 0.01031602640899676, 
 
   -0.02039296564537234, 0.5798031925065174, 
 
   0.01753139820427243, 0.004938621096619227, 
 
   0.02000583287408947, -0.02222396995572601, 
 
   -0.0131171331488489, 0.01105293667966819, 
 
   -0.02113747679684766, -0.003946078095551003, 
 
   0.01650315686805556, 0.02294396408662781, 
 
   0.03337409918412662, 0.1616946123373634, 
 
   0.1507286282821136, 0.04811341948931254, 
 
   0.1311343977263408, 0.04169528057826106, 
 
   0.1612926331118855, 0.10562665903742, 
 
   -0.140857563913936, 0.1149272200559877, 
 
   0.001937695774756816, -0.4564283882394649, 
 
   0.006829816546984519, -0.4294869148509813, 
 
   -0.01568390825293812, -0.3082637405023893, 
 
   0.02720315088865145, -0.4119855048483979, 
 
   -0.00802513386462662, 0.01618265706382708, 
 
   -0.5663352772875701, -0.02050186910656429, 
 
   0.03253831911627161, -0.01972676820219562, 
 
   0.00890852605102775, 0.04423471796393105, 
 
   0.03010204473680375, 0.018140913226686, 
 
   -0.03582163985928765, 0.04166220125883568, 
 
   0.04994120980832299, -0.006085469423199711, 
 
   -0.4178312822169214, -0.00879312677934999, 
 
   -0.007155718985319206, 0.0923342532434317, 
 
   -0.7220925688410524, -0.01895642442458183, 
 
   -0.02488961025182832, -0.05655914039805607, 
 
   -0.3407781077645247, 0.01596491614356593, 
 
   0.003799218828726969, -0.3007539457307736, 
 
   0.005423730927466854, 0.04416110103695653, 
 
   0.01161182910319807, -0.4401165552287813, 
 
   -0.01832056687984391, 0.1935602031171954, 
 
   0.03803612865420818, 0.006230511928403092, 
 
   -0.4264758500935657, -0.01837730652893412, 
 
   0.2600264572162181, 0.03813613551849553, 
 
   0.01418486098424744, -0.3197179952757978, 
 
   -0.02330139770043085, 0.253601574237227, 
 
   0.04969551017961125, 0.03918004256757905, 
 
   -0.2500888552589847, -0.005595390247043662, 
 
   0.00200689410514776, 0.0966809487716699, 
 
   -0.00183305615174304, 0.04733470328027378, 
 
   -0.01419509052344809, -0.4921858320884875, 
 
   -0.003259048861676481, 0.3492199355059927, 
 
   -0.01252705572307252, 0.1102565660449228, 
 
   -0.005105768798705375, 0.05948767738037454, 
 
   -0.3091941585593169, -0.0103856600368088, 
 
   -0.00267238069543673, -0.02178832227602695, 
 
   -0.1111694629265804, -0.007637810452200929, 
 
   0.04037207435919017, 0.03902400237750114, 
 
   -0.01770115065805611, 0.003882438029843885, 
 
   1.133025076796895, -0.01070681766981016, 
 
   -0.3299197754812053, -0.002912869647019059, 
 
   0.05245886023215144, 0.1148301947246799, 
 
   -0.006769265137416749, 0.03429428948733162, 
 
   0.846466935883099, 0.008295207605056, 
 
   -0.4221384215938806, 0.01535134677661914, 
 
   0.07975500431581027, -0.2190506334134537, 
 
   0.007089395048547386, 0.05321518485404808, 
 
   -0.4991438936211693, 0.003334325721050224, 
 
   -0.3817368895652962, 0.00007944389201517586, 
 
   0.1441405767934442, -0.5963842846805347, 
 
   -0.00936580424770133, 0.06315686801667209, 
 
   -2.147900017956749, -0.02090705784599681, 
 
   -0.2902592512325492, 0.0229707226020717, 
 
   0.03431740470976494, -0.4162567064325933, 
 
   -0.004259145422314207, -3.249205607001987, 
 
   0.00944467747211632, -0.3052650334843236, 
 
   0.03520966011588241, -0.2257527796465446, 
 
   0.04736469380355012}, 
 
  {0, 15, 29, 42, 50, 64, 78, 89, 103, 113, 120, 134, 148, 
 
   162, 169, 181, 210, 239, 267, 285, 299, 323, 362, 391, 
 
   439, 468, 498, 560, 588, 635, 664, 712, 756, 848, 909, 
 
   985, 1029, 1090, 1182, 1259, 1308, 1365, 1441, 1548, 
 
   1640, 1730, 1821, 1913, 2008, 2126, 2217, 2309, 2401, 
 
   2490, 2674, 2856, 3040, 3313, 3499, 4317, 4682, 4866, 
 
   5231, 5412, 5598, 5962, 6144, 6327, 6508, 7057, 7698, 
 
   9708, 9980, 10162, 10526, 10958}, 
 
  {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
 
    0, 0, 0, 0, 0, 0}, {0, 1.974256698742945*10^-8, 
 
    5.790225434983105*10^-8, 1.081650847336188*10^-7, 
 
    1.431726935826242*10^-7, 1.713057947746285*10^-7, 
 
    1.92362370130312*10^-7, 2.038009940934937*10^-7, 
 
    2.057003239615876*10^-7, 1.999353024565615*10^-7, 
 
    1.898206101977503*10^-7, 1.769165102985022*10^-7, 
 
    1.591917518254693*10^-7, 1.350721202423076*10^-7, 
 
    1.14174620148994*10^-7, 9.48580434867827*10^-8, 
 
    6.775870168557942*10^-8, 3.307297528013832*10^-8, 
 
    2.562518296272531*10^-9, -(1.39620420026048*10^-8), 
 
    -(2.163454337735367*10^-8), -(2.580501894319802*10^-8), 
 
    -(2.741239285351231*10^-8), -(2.396845770092741*10^-8), 
 
    -(1.742719982918614*10^-8), -(1.209400889885328*10^-8), 
 
    -(8.51771400974781*10^-9), -(5.721800371106298*10^-9), 
 
    -(4.137442549833056*10^-9), -(3.629022422148471*10^-9), 
 
    -(4.23816933661434*10^-9), -(5.08312190870567*10^-9), 
 
    -(5.32926243321008*10^-9), -(4.768057617605571*10^-9), 
 
    -(6.085105766893234*10^-9), -(5.741597800532415*10^-9), 
 
    -(5.510366782292367*10^-9), -(6.159463133553308*10^-9), 
 
    -(5.541770132883992*10^-9), -(5.545958186597578*10^-9), 
 
    -(5.731044583630531*10^-9), -(4.524596156830485*10^-9), 
 
    -(4.80594233517838*10^-9), -(4.497018868360711*10^-9), 
 
    -(4.122473624319884*10^-9), -(3.44070030334208*10^-9), 
 
    -(3.817514814264071*10^-9), -(3.445821279506184*10^-9), 
 
    -(4.015669980739577*10^-9), -(4.179134461709379*10^-9), 
 
    -(4.60444395547132*10^-9), -(3.409763763944332*10^-9), 
 
    -(4.4122713071158*10^-9), -(3.180402340354411*10^-9), 
 
    -(4.409814491223005*10^-9), -(3.586416064873145*10^-9), 
 
    -(2.635146392040295*10^-9), -(3.900579197536165*10^-9), 
 
    -(2.842259168034537*10^-9), -(2.410162726676321*10^-9), 
 
    -(2.718556092930482*10^-9), -(3.222387607620941*10^-9), 
 
    -(1.901643842673703*10^-9), -(3.835843469536818*10^-9), 
 
    -(5.63828311700804*10^-10), -(3.566156206980295*10^-9), 
 
    -(1.527051099484731*10^-9), -(2.340022411927482*10^-9), 
 
    -(6.733568908418403*10^-10), 
 
    -(4.354582870336678*10^-9), 1.476877320988501*10^-10, 
 
    -(2.892408524733532*10^-9), 1.724679090287907*10^-9, 
 
    -(1.026432799014884*10^-10), 
 
    -(1.537496082757789*10^-9), 
 
    -(2.687372290758657*10^-10), 3.294965543306838*10^-10, 
 
    -(2.008461437593873*10^-9)}, 
 
   {0, 5.790225434983105*10^-8, 1.698226739252314*10^-7, 
 
    3.172548686404961*10^-7, 4.199429163192892*10^-7, 
 
    5.024691279095071*10^-7, 5.642397527362418*10^-7, 
 
    5.977993882784638*10^-7, 6.033770120362987*10^-7, 
 
    5.864714541551822*10^-7, 5.568055776525414*10^-7, 
 
    5.18956896312343*10^-7, 4.669672686352613*10^-7, 
 
    3.962187583632127*10^-7, 3.349206120653458*10^-7, 
 
    2.782593450640336*10^-7, 1.987684329539591*10^-7, 
 
    9.70237229703132*10^-8, 7.523358119338517*10^-9, 
 
    -(4.095237660195538*10^-8), -(6.346091561991553*10^-8), 
 
    -(7.569613180315001*10^-8), -(8.04122877409989*10^-8), 
 
    -(7.030965433180487*10^-8), -(5.111833667828157*10^-8), 
 
    -(3.547440512906392*10^-8), -(2.498580486998744*10^-8), 
 
    -(1.678604840258178*10^-8), -(1.213801288408694*10^-8), 
 
    -(1.064480517395976*10^-8), -(1.243368324881083*10^-8), 
 
    -(1.49122384988307*10^-8), -(1.562975007845081*10^-8), 
 
    -(1.397582965822735*10^-8), -(1.785621999662008*10^-8), 
 
    -(1.684079842108678*10^-8), -(1.616350456380061*10^-8), 
 
    -(1.807215188350662*10^-8), -(1.624985278638608*10^-8), 
 
    -(1.627011453401329*10^-8), -(1.681315268674972*10^-8), 
 
    -(1.326557710380297*10^-8), -(1.410098450885462*10^-8), 
 
    -(1.319113248179715*10^-8), -(1.209677201631329*10^-8), 
 
    -(1.008808193977428*10^-8), -(1.12004353131244*10^-8), 
 
    -(1.010209405619913*10^-8), -(1.178158198832475*10^-8), 
 
    -(1.226013620583173*10^-8), -(1.351226230342545*10^-8), 
 
    -(9.99304070066876*10^-9), -(1.29488823223312*10^-8), 
 
    -(9.32216118987096*10^-9), -(1.294268558277814*10^-8), 
 
    -(1.051682808035198*10^-8), -(7.724126632332628*10^-9), 
 
    -(1.14514320804234*10^-8), -(8.33254828449815*10^-9), 
 
    -(7.072255467948505*10^-9), -(7.977848968822839*10^-9), 
 
    -(9.45250733440602*10^-9), -(5.570218609612553*10^-9), 
 
    -(1.127009323601047*10^-8), -(1.637985700244695*10^-9), 
 
    -(1.047526155762535*10^-8), -(4.469375021229952*10^-9), 
 
    -(6.870099128631982*10^-9), -(1.962645466297517*10^-9), 
 
    -(1.278878554022708*10^-8), 4.443346234738987*10^-10, 
 
    -(8.49643178258203*10^-9), 5.056755452912071*10^-9, 
 
    -(2.96654038785526*10^-10), -(4.515321744180361*10^-9), 
 
    -(7.857774685343624*10^-10), 9.81062812481666*10^-10, 
 
    -(5.901519775951777*10^-9)}, 
 
   {0, 1.081650847336188*10^-7, 3.172548686404961*10^-7, 
 
    5.942125322511739*10^-7, 7.887958139632595*10^-7, 
 
    9.46675901399663*10^-7, 1.066667981950436*10^-6, 
 
    1.134694740180473*10^-6, 1.150151986236316*10^-6, 
 
    1.122482806386808*10^-6, 1.069717810058977*10^-6, 
 
    1.001000344860966*10^-6, 9.05144040894917*10^-7, 
 
    7.730165312164474*10^-7, 6.574766235734355*10^-7, 
 
    5.500590162150319*10^-7, 3.985994250063221*10^-7, 
 
    2.026921339535032*10^-7, 2.808459462823384*10^-8, 
 
    -(6.82318927472708*10^-8), -(1.141059708639423*10^-7), 
 
    -(1.399947255962143*10^-7), -(1.520752228628168*10^-7), 
 
    -(1.354509041279703*10^-7), -(1.004917263267112*10^-7), 
 
    -(7.024027812163053*10^-8), -(4.941738102616253*10^-8), 
 
    -(3.312303605499072*10^-8), -(2.434028348230885*10^-8), 
 
    -(2.112223686527234*10^-8), -(2.387681817265897*10^-8), 
 
    -(2.8305545063857*10^-8), -(3.009289406864969*10^-8), 
 
    -(2.78290960358033*10^-8), -(3.388246095598884*10^-8), 
 
    -(3.279157291815471*10^-8), -(3.137181619076497*10^-8), 
 
    -(3.452326910043713*10^-8), -(3.198423329651045*10^-8), 
 
    -(3.123172868070505*10^-8), -(3.225199583577881*10^-8), 
 
    -(2.625447778308287*10^-8), -(2.689651558058567*10^-8), 
 
    -(2.553119063341111*10^-8), -(2.294479685849557*10^-8), 
 
    -(1.989567157074632*10^-8), -(2.13406171217486*10^-8), 
 
    -(2.000226509437109*10^-8), -(2.251331019938175*10^-8), 
 
    -(2.361764323209937*10^-8), -(2.553684495014275*10^-8), 
 
    -(2.014959816609806*10^-8), -(2.436127637146531*10^-8), 
 
    -(1.869123541195808*10^-8), -(2.430948260810508*10^-8), 
 
    -(2.063652594479067*10^-8), -(1.544122791967036*10^-8), 
 
    -(2.117761972057673*10^-8), -(1.648682160121366*10^-8), 
 
    -(1.345263907878196*10^-8), -(1.53010465138941*10^-8), 
 
    -(1.808838500400501*10^-8), -(1.137849574555956*10^-8), 
 
    -(1.998762664218068*10^-8), -(4.647012995247085*10^-9), 
 
    -(1.886119697805682*10^-8), -(9.61871648696807*10^-9), 
 
    -(1.265298506897008*10^-8), -(4.992998768903547*10^-9), 
 
    -(2.339298001559469*10^-8), 
 
    -(1.168555509818656*10^-10), -(1.53285166369225*10^-8), 
 
    1.058640662354996*10^-8, -(9.99162626950936*10^-10), 
 
    -(8.36648469557216*10^-9), -(1.784805600536449*10^-9), 
 
    3.53116090844005*10^-10, -(1.058873759403372*10^-8)}, 
 
   {0, 1.431726935826241*10^-7, 4.199429163192892*10^-7, 
 
    7.887958139632595*10^-7, 1.050827482137725*10^-6, 
 
    1.266161017954039*10^-6, 1.43312899086048*10^-6, 
 
    1.532942279210409*10^-6, 1.562879557674109*10^-6, 
 
    1.533819151764737*10^-6, 1.469278577548369*10^-6, 
 
    1.382434002395921*10^-6, 1.25842207467118*10^-6, 
 
    1.084219559841433*10^-6, 9.29842350857128*10^-7, 
 
    7.851295317775341*10^-7, 5.796210786041289*10^-7, 
 
    3.098455447427066*10^-7, 6.493128942322435*10^-8, 
 
    -(7.350869069435985*10^-8), -(1.415895448738479*10^-7), 
 
    -(1.817283611764963*10^-7), -(2.040967159580842*10^-7), 
 
    -(1.865378373534128*10^-7), -(1.422227920770766*10^-7), 
 
    -(1.005268109265712*10^-7), -(7.074976965826751*10^-8), 
 
    -(4.730750290027402*10^-8), -(3.529903057245132*10^-8), 
 
    -(3.025287806502034*10^-8), -(3.286961071105057*10^-8), 
 
    -(3.836851764303345*10^-8), -(4.152060920127855*10^-8), 
 
    -(3.998752395987646*10^-8), -(4.586412567577156*10^-8), 
 
    -(4.585162273092413*10^-8), -(4.371927329293079*10^-8), 
 
    -(4.714797948744948*10^-8), -(4.532627936610804*10^-8), 
 
    -(4.294207451065836*10^-8), -(4.427685040371832*10^-8), 
 
    -(3.746068226265776*10^-8), -(3.668139013774824*10^-8), 
 
    -(3.544738750595458*10^-8), -(3.106305012213553*10^-8), 
 
    -(2.827254429621926*10^-8), -(2.906404393003053*10^-8), 
 
    -(2.855472899384885*10^-8), -(3.076216954913499*10^-8), 
 
    -(3.259712883077089*10^-8), -(3.439030087266095*10^-8), 
 
    -(2.937474603329236*10^-8), -(3.26195003235589*10^-8), 
 
    -(2.707911292387611*10^-8), -(3.24779447431661*10^-8), 
 
    -(2.913540311351178*10^-8), -(2.229386157543206*10^-8), 
 
    -(2.769113005069086*10^-8), -(2.351001009423045*10^-8), 
 
    -(1.826519935291676*10^-8), -(2.096527470760029*10^-8), 
 
    -(2.476996022450871*10^-8), -(1.685275853623912*10^-8), 
 
    -(2.458968800563984*10^-8), -(9.01202807960311*10^-9), 
 
    -(2.372953928697428*10^-8), -(1.501898784369628*10^-8), 
 
    -(1.645414201717354*10^-8), -(9.00060890613554*10^-9), 
 
    -(3.008074819136022*10^-8), -(1.873105423440043*10^-9), 
 
    -(1.932532737200777*10^-8), 1.596854503947591*10^-8, 
 
    -(2.121616711416534*10^-9), -(1.094005535295263*10^-8), 
 
    -(2.928822866045674*10^-9), -(2.161429949914268*10^-9), 
 
    -(1.322595838421134*10^-8)}, 
 
   {0, 1.713057947746285*10^-7, 5.024691279095071*10^-7, 
 
    9.46675901399663*10^-7, 1.266161017954039*10^-6, 
 
    1.532693820700302*10^-6, 1.744238665026024*10^-6, 
 
    1.878312954510029*10^-6, 1.928764752344239*10^-6, 
 
    1.906011133539275*10^-6, 1.837492132055448*10^-6, 
 
    1.740563059877515*10^-6, 1.597436238358609*10^-6, 
 
    1.391108462515377*10^-6, 1.204997121794796*10^-6, 
 
    1.028646625542286*10^-6, 7.758897680212657*10^-7, 
 
    4.377789390882309*10^-7, 1.234460744960321*10^-7, 
 
    -(5.956292594092063*10^-8), -(1.528336017132812*10^-7), 
 
    -(2.103325091646378*10^-7), -(2.47416486931664*10^-7), 
 
    -(2.336277609570643*10^-7), -(1.839896415098104*10^-7), 
 
    -(1.321334684908753*10^-7), -(9.33714591177502*10^-8), 
 
    -(6.24071120290576*10^-8), -(4.697956511015192*10^-8), 
 
    -(3.969686230837857*10^-8), -(4.151744618676145*10^-8), 
 
    -(4.757909231742433*10^-8), -(5.222734780282005*10^-8), 
 
    -(5.208445929919533*10^-8), -(5.68870413463808*10^-8), 
 
    -(5.851220006942917*10^-8), -(5.570701782125278*10^-8), 
 
    -(5.89660489835532*10^-8), -(5.846581506776989*10^-8), 
 
    -(5.407893839184017*10^-8), -(5.558719976681676*10^-8), 
 
    -(4.860998119334499*10^-8), -(4.586713765901526*10^-8), 
 
    -(4.49608844936067*10^-8), -(3.857027904450543*10^-8), 
 
    -(3.655795494388609*10^-8), -(3.626887332926177*10^-8), 
 
    -(3.704426222954988*10^-8), -(3.847988286330941*10^-8), 
 
    -(4.115124123210867*10^-8), -(4.246187341937099*10^-8), 
 
    -(3.875546588653289*10^-8), -(4.00664185345595*10^-8), 
 
    -(3.557699774561529*10^-8), -(3.983342255867242*10^-8), 
 
    -(3.745823375528632*10^-8), -(2.919913558958434*10^-8), 
 
    -(3.324733429897261*10^-8), -(3.044736318257989*10^-8), 
 
    -(2.271418300233856*10^-8), -(2.628143322896053*10^-8), 
 
    -(3.100772302256317*10^-8), -(2.250295834637397*10^-8), 
 
    -(2.77490484922984*10^-8), -(1.418562362396597*10^-8), 
 
    -(2.741547163040504*10^-8), -(2.082408559144381*10^-8), 
 
    -(1.964843760753753*10^-8), -(1.365038183590744*10^-8), 
 
    -(3.550287878501503*10^-8), -(4.237330766814127*10^-9), 
 
    -(2.235567818262032*10^-8), 2.158022265707819*10^-8, 
 
    -(3.48081064924346*10^-9), -(1.312827732988997*10^-8), 
 
    -(4.195770077858092*10^-9), -(5.606384821920222*10^-9), 
 
    -(1.515569703434508*10^-8)}, 
 
   {0, 1.923623701303121*10^-7, 5.642397527362418*10^-7, 
 
    1.066667981950436*10^-6, 1.43312899086048*10^-6, 
 
    1.744238665026024*10^-6, 1.997893202310524*10^-6, 
 
    2.169250213042888*10^-6, 2.247300212564896*10^-6, 
 
    2.239820136513385*10^-6, 2.176370445687314*10^-6, 
 
    2.07866321958726*10^-6, 1.92683997082753*10^-6, 
 
    1.699756045691659*10^-6, 1.489921815980251*10^-6, 
 
    1.288219028087343*10^-6, 9.9561872750808*10^-7, 
 
    5.946579659885649*10^-7, 2.105222387567115*10^-7, 
 
    -(2.111367634640088*10^-8), -(1.438743104477522*10^-7), 
 
    -(2.230109429629168*10^-7), -(2.806575564863543*10^-7), 
 
    -(2.76561906377002*10^-7), -(2.265284468604032*10^-7), 
 
    -(1.661364185678615*10^-7), -(1.183309360677975*10^-7), 
 
    -(7.924458588967962*10^-8), -(5.987806838999355*10^-8), 
 
    -(4.976044271863997*10^-8), -(5.009379530717214*10^-8), 
 
    -(5.614781277909712*10^-8), -(6.236302710297609*10^-8), 
 
    -(6.424699453866526*10^-8), -(6.723141620728144*10^-8), 
 
    -(7.097604744441418*10^-8), -(6.760443231065294*10^-8), 
 
    -(7.025616236179171*10^-8), -(7.158490601631251*10^-8), 
 
    -(6.494825768023115*10^-8), -(6.640965427395079*10^-8), 
 
    -(5.987331324963073*10^-8), -(5.472142111747576*10^-8), 
 
    -(5.423841233168889*10^-8), -(4.5693101380638*10^-8), 
 
    -(4.488201255097964*10^-8), -(4.316065473703667*10^-8), 
 
    -(4.558959014326987*10^-8), -(4.585255612788354*10^-8), 
 
    -(4.944658505216836*10^-8), -(4.996075051824817*10^-8), 
 
    -(4.839235036614515*10^-8), -(4.692465412658587*10^-8), 
 
    -(4.42966760636417*10^-8), -(4.659883128536383*10^-8), 
 
    -(4.572154013729798*10^-8), -(3.625417430641506*10^-8), 
 
    -(3.802170393123864*10^-8), -(3.738608462136461*10^-8), 
 
    -(2.689349204412517*10^-8), -(3.129846826978524*10^-8), 
 
    -(3.695267238746195*10^-8), -(2.841335111487812*10^-8), 
 
    -(2.967860293589272*10^-8), -(2.013862353537027*10^-8), 
 
    -(3.008971017418594*10^-8), -(2.703340343874144*10^-8), 
 
    -(2.235109644522529*10^-8), -(1.892478011349424*10^-8), 
 
    -(3.979655791592824*10^-8), -(7.202088473703139*10^-9), 
 
    -(2.452692833169979*10^-8), 2.726214227638183*10^-8, 
 
    -(5.061913525516503*10^-9), -(1.495135983398125*10^-8), 
 
    -(5.574968380385002*10^-9), -(9.8701823810109*10^-9), 
 
    -(1.642941894313754*10^-8)}, 
 
   {0, 2.038009940934937*10^-7, 5.977993882784638*10^-7, 
 
    1.134694740180473*10^-6, 1.532942279210409*10^-6, 
 
    1.878312954510029*10^-6, 2.169250213042888*10^-6, 
 
    2.381319316837644*10^-6, 2.497150251072199*10^-6, 
 
    2.518562241006106*10^-6, 2.474323552084904*10^-6, 
 
    2.390592654068495*10^-6, 2.246835589552657*10^-6, 
 
    2.017525011249072*10^-6, 1.797125734887183*10^-6, 
 
    1.580407690287832*10^-6, 1.260126114279432*10^-6, 
 
    8.0511967781728*10^-7, 3.49553009665104*10^-7, 
 
    6.144390720314658*10^-8, -(9.87939748565501*10^-8), 
 
    -(2.07396642055293*10^-7), -(2.961554126141283*10^-7), 
 
    -(3.122799532994247*10^-7), -(2.706557925639326*10^-7), 
 
    -(2.051106093539786*10^-7), -(1.484076559608463*10^-7), 
 
    -(1.000512193568613*10^-7), -(7.531155467438853*10^-8), 
 
    -(6.12320814797442*10^-8), -(5.918580465868489*10^-8), 
 
    -(6.428344834012428*10^-8), -(7.188607090844199*10^-8), 
 
    -(7.649160228861773*10^-8), -(7.725115940141297*10^-8), 
 
    -(8.33788200683942*10^-8), -(7.984530395216663*10^-8), 
 
    -(8.13557361988266*10^-8), -(8.47874007162076*10^-8), 
 
    -(7.61124586133378*10^-8), -(7.693121048645344*10^-8), 
 
    -(7.144377779305639*10^-8), -(6.372410446174476*10^-8), 
 
    -(6.339288393721221*10^-8), -(5.277969010640423*10^-8), 
 
    -(5.338585623811945*10^-8), -(5.007510602759031*10^-8), 
 
    -(5.429871795779303*10^-8), -(5.312407957708445*10^-8), 
 
    -(5.76416822677237*10^-8), -(5.712928772027166*10^-8), 
 
    -(5.835290852765298*10^-8), -(5.350228742307169*10^-8), 
 
    -(5.335697574262253*10^-8), -(5.307374223392982*10^-8), 
 
    -(5.399543016442032*10^-8), -(4.35746698765268*10^-8), 
 
    -(4.21832404059767*10^-8), -(4.437903633428851*10^-8), 
 
    -(3.087894113408058*10^-8), -(3.589940479961809*10^-8), 
 
    -(4.281336841925944*10^-8), -(3.473803737641089*10^-8), 
 
    -(3.063460375283811*10^-8), -(2.680857420865342*10^-8), 
 
    -(3.19102435375746*10^-8), -(3.35539830387038*10^-8), 
 
    -(2.469784643182754*10^-8), -(2.477110973724076*10^-8), 
 
    -(4.292034026707287*10^-8), -(1.082224778719527*10^-8), 
 
    -(2.585718991438004*10^-8), 3.23026194366776*10^-8, 
 
    -(6.829500873293694*10^-9), -(1.629673646008201*10^-8), 
 
    -(7.01772227015196*10^-9), -(1.468274331190586*10^-8), 
 
    -(1.696921609062695*10^-8)}, 
 
   {0, 2.057003239615876*10^-7, 6.033770120362987*10^-7, 
 
    1.150151986236316*10^-6, 1.562879557674109*10^-6, 
 
    1.928764752344239*10^-6, 2.247300212564895*10^-6, 
 
    2.497150251072199*10^-6, 2.654817323570228*10^-6, 
 
    2.714129163880934*10^-6, 2.700361372377339*10^-6, 
 
    2.64343897227077*10^-6, 2.523711168207372*10^-6, 
 
    2.311663375040008*10^-6, 2.095902002132794*10^-6, 
 
    1.877045712823597*10^-6, 1.545574518844304*10^-6, 
 
    1.052806786399001*10^-6, 5.325346594525769*10^-7, 
 
    1.857412674328407*10^-7, -(1.668538458910561*10^-8), 
 
    -(1.602630361916417*10^-7), -(2.886212941329039*10^-7), 
 
    -(3.348188525303087*10^-7), -(3.110629945922256*10^-7), 
 
    -(2.454525077707882*10^-7), -(1.81525822698533*10^-7), 
 
    -(1.238158230384438*10^-7), -(9.23552377389825*10^-8), 
 
    -(7.337672992596535*10^-8), -(6.838313587488645*10^-8), 
 
    -(7.160746613411134*10^-8), -(7.999450311932879*10^-8), 
 
    -(8.74946765052857*10^-8), -(8.64659741404698*10^-8), 
 
    -(9.45542310742616*10^-8), -(9.14908067571961*10^-8), 
 
    -(9.16038805379973*10^-8), -(9.67469319836494*10^-8), 
 
    -(8.69418108784553*10^-8), -(8.64034380512675*10^-8), 
 
    -(8.21970888574653*10^-8), -(7.242530515620686*10^-8), 
 
    -(7.159577749164738*10^-8), -(5.960785406356504*10^-8), 
 
    -(6.12361121276736*10^-8), -(5.670477694258587*10^-8), 
 
    -(6.229236734616367*10^-8), -(5.98858332802401*10^-8), 
 
    -(6.50331154077453*10^-8), -(6.363891961424258*10^-8), 
 
    -(6.752593702115793*10^-8), -(5.962596764763496*10^-8), 
 
    -(6.177230266422306*10^-8), -(5.905681070744816*10^-8), 
 
    -(6.144555449445771*10^-8), -(5.042687416258817*10^-8), 
 
    -(4.569997700047674*10^-8), -(5.074122107547561*10^-8), 
 
    -(3.438725145529036*10^-8), -(3.953243451273528*10^-8), 
 
    -(4.828879442163192*10^-8), -(4.085602881976401*10^-8), 
 
    -(3.11985505331302*10^-8), -(3.31378475366299*10^-8), 
 
    -(3.318963640400118*10^-8), -(3.944567938445837*10^-8), 
 
    -(2.674464908684277*10^-8), -(3.026878775680355*10^-8), 
 
    -(4.49027519343357*10^-8), -(1.453819338284717*10^-8), 
 
    -(2.653759446903545*10^-8), 3.539217983782867*10^-8, 
 
    -(8.47906649633486*10^-9), -(1.702411858838022*10^-8), 
 
    -(8.2796195686578*10^-9), -(1.894541386815513*10^-8), 
 
    -(1.686632383117327*10^-8)}, 
 
   {0, 1.999353024565615*10^-7, 5.864714541551824*10^-7, 
 
    1.122482806386808*10^-6, 1.533819151764737*10^-6, 
 
    1.906011133539275*10^-6, 2.239820136513385*10^-6, 
 
    2.518562241006106*10^-6, 2.714129163880934*10^-6, 
 
    2.812526529604946*10^-6, 2.834007176126236*10^-6, 
 
    2.810932079419631*10^-6, 2.725885083079207*10^-6, 
 
    2.54646129584452*10^-6, 2.34892761156199*10^-6, 
 
    2.140462609580715*10^-6, 1.815216571524912*10^-6, 
 
    1.305993353860091*10^-6, 7.367071803870427*10^-7, 
 
    3.369878136945456*10^-7, 9.32335931416058*10^-8, 
 
    -(8.62809660661678*10^-8), -(2.576956674727014*10^-7), 
 
    -(3.402166142984366*10^-7), -(3.419592172117485*10^-7), 
 
    -(2.819363522024742*10^-7), -(2.139269970684055*10^-7), 
 
    -(1.482536138916604*10^-7), -(1.093807968762114*10^-7), 
 
    -(8.50357784551321*10^-8), -(7.695967903180846*10^-8), 
 
    -(7.770933553116394*10^-8), -(8.60216064183056*10^-8), 
 
    -(9.60022004948785*10^-8), -(9.42785526118063*10^-8), 
 
    -(1.033506745604018*10^-7), -(1.013945473438562*10^-7), 
 
    -(1.002454338412732*10^-7), -(1.061724985339318*10^-7), 
 
    -(9.65039130619925*10^-8), -(9.40751198562644*10^-8), 
 
    -(9.09558282246434*10^-8), -(8.01170072942147*10^-8), 
 
    -(7.802105978279672*10^-8), -(6.583808669288512*10^-8), 
 
    -(6.754915026027138*10^-8), -(6.258474308487909*10^-8), 
 
    -(6.866654511162434*10^-8), -(6.564804683612212*10^-8), 
 
    -(7.086916429042839*10^-8), -(6.909074577883729*10^-8), 
 
    -(7.481439015890447*10^-8), -(6.500755283104011*10^-8), 
 
    -(6.851947650853355*10^-8), -(6.423347464041952*10^-8), 
 
    -(6.724690573693195*10^-8), -(5.604579785048708*10^-8), 
 
    -(4.848445596835506*10^-8), -(5.580553318971369*10^-8), 
 
    -(3.712122535905194*10^-8), -(4.177702952336113*10^-8), 
 
    -(5.299085967262234*10^-8), -(4.606380720085176*10^-8), 
 
    -(3.184650798234507*10^-8), -(3.811616787145697*10^-8), 
 
    -(3.420032668053102*10^-8), -(4.386077742130106*10^-8), 
 
    -(2.849609049257666*10^-8), -(3.453752247036542*10^-8), 
 
    -(4.590003136759485*10^-8), -(1.773668581868073*10^-8), 
 
    -(2.68132275700462*10^-8), 3.577520631985778*10^-8, 
 
    -(9.73164628017527*10^-9), -(1.710739817790177*10^-8), 
 
    -(9.1560022910811*10^-9), -(2.173960822757029*10^-8), 
 
    -(1.631437753402486*10^-8)}, 
 
   {0, 1.898206101977503*10^-7, 5.568055776525414*10^-7, 
 
    1.069717810058977*10^-6, 1.469278577548369*10^-6, 
 
    1.837492132055448*10^-6, 2.176370445687314*10^-6, 
 
    2.474323552084904*10^-6, 2.70036137237734*10^-6, 
 
    2.834007176126236*10^-6, 2.890421838416218*10^-6, 
 
    2.903089816008844*10^-6, 2.857808912328383*10^-6, 
 
    2.720619938353862*10^-6, 2.551096259561088*10^-6, 
 
    2.362723438152533*10^-6, 2.058002492755472*10^-6, 
 
    1.551965134438907*10^-6, 9.50636032538451*10^-7, 
 
    5.062166649767097*10^-7, 2.240829604424609*10^-7, 
 
    9.55651398660046*10^-9, -(2.060055796095636*10^-7), 
 
    -(3.289788086966379*10^-7), -(3.622779642788418*10^-7), 
 
    -(3.131039548376201*10^-7), -(2.445787953256419*10^-7), 
 
    -(1.729824870080831*10^-7), -(1.263012385508843*10^-7), 
 
    -(9.6224594624662*10^-8), -(8.50496362664551*10^-8), 
 
    -(8.30545415213588*10^-8), -(9.04487363247029*10^-8), 
 
    -(1.021027395814451*10^-7), -(1.00989613876364*10^-7), 
 
    -(1.101323876004882*10^-7), -(1.095586025987162*10^-7), 
 
    -(1.076291168578256*10^-7), -(1.134062404079249*10^-7), 
 
    -(1.047993941612378*10^-7), -(1.004112026658456*10^-7), 
 
    -(9.78853160346159*10^-8), -(8.68688749975746*10^-8), 
 
    -(8.30291258780359*10^-8), -(7.150002924549501*10^-8), 
 
    -(7.254999477569607*10^-8), -(6.780569425926304*10^-8), 
 
    -(7.359658495356426*10^-8), -(7.058426216983932*10^-8), 
 
    -(7.536364515001025*10^-8), -(7.380535549643122*10^-8), 
 
    -(8.02752947195176*10^-8), -(6.987724262464301*10^-8), 
 
    -(7.365444749856314*10^-8), -(6.880821528517495*10^-8), 
 
    -(7.172602177404191*10^-8), -(6.032900273416845*10^-8), 
 
    -(5.090045225379568*10^-8), -(5.968235163844139*10^-8), 
 
    -(3.921575545908501*10^-8), -(4.279640877358136*10^-8), 
 
    -(5.713062633885659*10^-8), -(5.028654737874657*10^-8), 
 
    -(3.313210716184129*10^-8), -(4.146500535173989*10^-8), 
 
    -(3.542071621117806*10^-8), -(4.670822931026561*10^-8), 
 
    -(3.019682020553994*10^-8), -(3.733586961165136*10^-8), 
 
    -(4.65307471936087*10^-8), -(2.014250121807781*10^-8), 
 
    -(2.71407874619386*10^-8), 3.358945090832437*10^-8, 
 
    -(1.050996052645889*10^-8), -(1.674963243273219*10^-8), 
 
    -(9.62109265579735*10^-9), -(2.275424608720494*10^-8), 
 
    -(1.565225676982838*10^-8)}, 
 
   {0, 1.769165102985022*10^-7, 5.189568963123429*10^-7, 
 
    1.001000344860966*10^-6, 1.382434002395921*10^-6, 
 
    1.740563059877516*10^-6, 2.07866321958726*10^-6, 
 
    2.390592654068495*10^-6, 2.643438972270771*10^-6, 
 
    2.81093207941963*10^-6, 2.903089816008844*10^-6, 
 
    2.953867335787931*10^-6, 2.953138245355356*10^-6, 
 
    2.866348252486331*10^-6, 2.732716061684996*10^-6, 
 
    2.57199608647914*10^-6, 2.298617834765811*10^-6, 
 
    1.809594957933191*10^-6, 1.186295429282158*10^-6, 
 
    7.002675154770428*10^-7, 3.793794755820889*10^-7, 
 
    1.281853297598152*10^-7, -(1.352899548668782*10^-7), 
 
    -(3.045201693664007*10^-7), -(3.760751353407548*10^-7), 
 
    -(3.425022982759574*10^-7), -(2.762656241874129*10^-7), 
 
    -(2.000423660989005*10^-7), -(1.446926295687053*10^-7), 
 
    -(1.081390057023594*10^-7), -(9.36040486920507*10^-8), 
 
    -(8.8545643792973*10^-8), -(9.43217059459145*10^-8), 
 
    -(1.070674556118995*10^-7), -(1.07734309054095*10^-7), 
 
    -(1.162018663510194*10^-7), -(1.172178900774857*10^-7), 
 
    -(1.149562051089487*10^-7), -(1.198169636890104*10^-7), 
 
    -(1.12986997468375*10^-7), -(1.065584838103867*10^-7), 
 
    -(1.041713933135656*10^-7), -(9.36332061583895*10^-8), 
 
    -(8.7589411694146*10^-8), -(7.733106953051471*10^-8), 
 
    -(7.712278334923671*10^-8), -(7.308873411576612*10^-8), 
 
    -(7.797980001902093*10^-8), -(7.548094841244824*10^-8), 
 
    -(7.937627889707948*10^-8), -(7.860131549779008*10^-8), 
 
    -(8.49086777867603*10^-8), -(7.496991659388888*10^-8), 
 
    -(7.807363648881691*10^-8), -(7.350309129174433*10^-8), 
 
    -(7.579313937525509*10^-8), -(6.397930022070171*10^-8), 
 
    -(5.345941473788187*10^-8), -(6.310120441993237*10^-8), 
 
    -(4.110556780119132*10^-8), -(4.312527228508736*10^-8), 
 
    -(6.133816632683705*10^-8), -(5.410568166893968*10^-8), 
 
    -(3.531363860560897*10^-8), -(4.375755210218726*10^-8), 
 
    -(3.718212930567669*10^-8), -(4.861820744861364*10^-8), 
 
    -(3.214950926001896*10^-8), -(3.916708443820996*10^-8), 
 
    -(4.733476444281074*10^-8), -(2.200956600615755*10^-8), 
 
    -(2.781460972507579*10^-8), 2.948823110577266*10^-8, 
 
    -(1.09673409638459*10^-8), -(1.617413412237667*10^-8), 
 
    -(9.81629428214929*10^-9), -(2.239542331329443*10^-8), 
 
    -(1.506692196012973*10^-8)}, 
 
   {0, 1.591917518254693*10^-7, 4.669672686352613*10^-7, 
 
    9.05144040894917*10^-7, 1.25842207467118*10^-6, 
 
    1.597436238358609*10^-6, 1.92683997082753*10^-6, 
 
    2.246835589552658*10^-6, 2.523711168207372*10^-6, 
 
    2.725885083079207*10^-6, 2.857808912328383*10^-6, 
 
    2.953138245355356*10^-6, 3.007506846512297*10^-6, 
 
    2.987305228543768*10^-6, 2.904297860420356*10^-6, 
 
    2.784877677647986*10^-6, 2.561759920810676*10^-6, 
 
    2.11212465108421*10^-6, 1.480040009798045*10^-6, 
 
    9.53067704113124*10^-7, 5.889774910359446*10^-7, 
 
    2.948532797104597*10^-7, -(2.727358531817848*10^-8), 
 
    -(2.56608803093143*10^-7), -(3.809316639777465*10^-7), 
 
    -(3.721162555931579*10^-7), -(3.129484279220895*10^-7), 
 
    -(2.338419738433015*10^-7), -(1.676608247507364*10^-7), 
 
    -(1.22754006418259*10^-7), -(1.040407149149554*10^-7), 
 
    -(9.5118048659168*10^-8), -(9.80512234738713*10^-8), 
 
    -(1.110593736461831*10^-7), -(1.154654920989794*10^-7), 
 
    -(1.221020295963531*10^-7), -(1.25128867765397*10^-7), 
 
    -(1.232295938331869*10^-7), -(1.259456910159142*10^-7), 
 
    -(1.220732898744292*10^-7), -(1.133365393631536*10^-7), 
 
    -(1.10369063214074*10^-7), -(1.013028230064587*10^-7), 
 
    -(9.21763970616989*10^-8), -(8.41627091032478*10^-8), 
 
    -(8.17086671481852*10^-8), -(7.915975687712268*10^-8), 
 
    -(8.21752894211263*10^-8), -(8.09968595536204*10^-8), 
 
    -(8.32681328701335*10^-8), -(8.42188139423207*10^-8), 
 
    -(8.89570097584949*10^-8), -(8.1101302081343*10^-8), 
 
    -(8.20349518629615*10^-8), -(7.903453172937125*10^-8), 
 
    -(7.986510696621908*10^-8), -(6.716274692907769*10^-8), 
 
    -(5.665585117666943*10^-8), -(6.63281491615801*10^-8), 
 
    -(4.29848311501212*10^-8), -(4.257707301395387*10^-8), 
 
    -(6.622715865269814*10^-8), -(5.779749998096227*10^-8), 
 
    -(3.921634634198638*10^-8), -(4.474419925591656*10^-8), 
 
    -(4.008618176702033*10^-8), -(4.940662197793002*10^-8), 
 
    -(3.478036586163778*10^-8), -(3.977560401995727*10^-8), 
 
    -(4.86903949220579*10^-8), -(2.32521014910253*10^-8), 
 
    -(2.921359331163462*10^-8), 2.222522882861147*10^-8, 
 
    -(1.100109093158392*10^-8), -(1.528968178741602*10^-8), 
 
    -(9.64527479014406*10^-9), -(1.994776764518278*10^-8), 
 
    -(1.46330078662128*10^-8)}, 
 
   {0, 1.350721202423076*10^-7, 3.962187583632127*10^-7, 
 
    7.730165312164474*10^-7, 1.084219559841433*10^-6, 
 
    1.391108462515377*10^-6, 1.699756045691659*10^-6, 
 
    2.017525011249072*10^-6, 2.311663375040008*10^-6, 
 
    2.54646129584452*10^-6, 2.720619938353862*10^-6, 
 
    2.866348252486331*10^-6, 2.987305228543768*10^-6, 
 
    3.054433868443386*10^-6, 3.04303345003393*10^-6, 
 
    2.98563216883559*10^-6, 2.843125855126533*10^-6, 
 
    2.473246800904947*10^-6, 1.862106850289337*10^-6, 
 
    1.302307209219965*10^-6, 8.92023955444878*10^-7, 
 
    5.476653995094184*10^-7, 1.514869472386781*10^-7, 
 
    -(1.608267788412785*10^-7), -(3.637226811168325*10^-7), 
 
    -(3.972047832752153*10^-7), -(3.552637178102914*10^-7), 
 
    -(2.784620370655342*10^-7), -(1.991758172858602*10^-7), 
 
    -(1.429229734047493*10^-7), -(1.183087010550078*10^-7), 
 
    -(1.039569848554798*10^-7), -(1.022001039792669*10^-7), 
 
    -(1.138669960494903*10^-7), -(1.242797238848276*10^-7), 
 
    -(1.281060514997828*10^-7), -(1.336149068324294*10^-7), 
 
    -(1.329794061405687*10^-7), -(1.322212370179876*10^-7), 
 
    -(1.324673168331098*10^-7), -(1.212875317467026*10^-7), 
 
    -(1.167770377471652*10^-7), -(1.102556560506956*10^-7), 
 
    -(9.72233119750451*10^-8), -(9.23556504690662*10^-8), 
 
    -(8.65960480178581*10^-8), -(8.64111596941446*10^-8), 
 
    -(8.64197428937983*10^-8), -(8.74320573839941*10^-8), 
 
    -(8.73062060535485*10^-8), -(9.09929808672642*10^-8), 
 
    -(9.25440790402921*10^-8), -(8.8670536968969*10^-8), 
 
    -(8.5739870046192*10^-8), -(8.57325169177285*10^-8), 
 
    -(8.40951811745269*10^-8), -(6.994581999220701*10^-8), 
 
    -(6.07785908559691*10^-8), -(6.944276528716609*10^-8), 
 
    -(4.495128329960369*10^-8), -(4.089817821473185*10^-8), 
 
    -(7.21501102377663*10^-8), -(6.16068072828516*10^-8), 
 
    -(4.526324193762844*10^-8), -(4.430176324550299*10^-8), 
 
    -(4.438592149140064*10^-8), -(4.889549255251452*10^-8), 
 
    -(3.831098715896963*10^-8), -(3.904359392549597*10^-8), 
 
    -(5.05158855950326*10^-8), -(2.394721371767877*10^-8), 
 
    -(3.136925056443141*10^-8), 1.054242094058786*10^-8, 
 
    -(1.054639841724202*10^-8), -(1.388750614700677*10^-8), 
 
    -(9.0204549388697*10^-9), -(1.492272974672625*10^-8), 
 
    -(1.420748798090778*10^-8)}, 
 
   {0, 1.14174620148994*10^-7, 3.349206120653458*10^-7, 
 
    6.574766235734355*10^-7, 9.29842350857128*10^-7, 
 
    1.204997121794796*10^-6, 1.489921815980251*10^-6, 
 
    1.797125734887183*10^-6, 2.095902002132794*10^-6, 
 
    2.34892761156199*10^-6, 2.551096259561088*10^-6, 
 
    2.732716061684996*10^-6, 2.904297860420356*10^-6, 
 
    3.04303345003393*10^-6, 3.094707415324556*10^-6, 
 
    3.09503250741104*10^-6, 3.031812087218518*10^-6, 
 
    2.753295129401754*10^-6, 2.189419267853208*10^-6, 
 
    1.621389389911801*10^-6, 1.181754749434096*10^-6, 
 
    8.00324018718746*10^-7, 3.433708295342628*10^-7, 
 
    -(4.465458351972423*10^-8), -(3.248511374642991*10^-7), 
 
    -(4.047617271362935*10^-7), -(3.857649089880777*10^-7), 
 
    -(3.174632702203613*10^-7), -(2.287612990258921*10^-7), 
 
    -(1.624673364979992*10^-7), -(1.320755919167988*10^-7), 
 
    -(1.12402552226596*10^-7), -(1.059290353113011*10^-7), 
 
    -(1.150785373369691*10^-7), -(1.309156272798569*10^-7), 
 
    -(1.325564984552842*10^-7), -(1.399281969109744*10^-7), 
 
    -(1.408406385403926*10^-7), -(1.370372840567114*10^-7), 
 
    -(1.404959642023066*10^-7), -(1.278300783104343*10^-7), 
 
    -(1.215105516764577*10^-7), -(1.172910378560302*10^-7), 
 
    -(1.013709114255419*10^-7), -(9.88880280362429*10^-8), 
 
    -(9.03363446439898*10^-8), -(9.22493822030549*10^-8), 
 
    -(8.9580388182971*10^-8), -(9.254106387003*10^-8), 
 
    -(9.04390697706851*10^-8), -(9.65228173313442*10^-8), 
 
    -(9.49047995901021*10^-8), -(9.49009528118706*10^-8), 
 
    -(8.83179317115911*10^-8), -(9.12154490106254*10^-8), 
 
    -(8.72452769683872*10^-8), -(7.17022335072749*10^-8), 
 
    -(6.437971472998136*10^-8), -(7.161113912557737*10^-8), 
 
    -(4.646903337258295*10^-8), -(3.913885534487071*10^-8), 
 
    -(7.699047313051695*10^-8), -(6.441833682581275*10^-8), 
 
    -(5.087650677470155*10^-8), -(4.316617002975516*10^-8), 
 
    -(4.83160922418509*10^-8), -(4.779898243329025*10^-8), 
 
    -(4.138139672669893*10^-8), -(3.777585683166805*10^-8), 
 
    -(5.213457611723204*10^-8), -(2.41402898232021*10^-8), 
 
    -(3.336474154761526*10^-8), 
 
    -(2.922573486493186*10^-10), -(9.94044143899656*10^-9), 
 
    -(1.262220104932115*10^-8), -(8.31581488118928*10^-9), 
 
    -(9.86751640082998*10^-9), -(1.394001591972748*10^-8)}, 
 
   {0, 9.48580434867827*10^-8, 2.782593450640335*10^-7, 
 
    5.500590162150317*10^-7, 7.851295317775341*10^-7, 
 
    1.028646625542286*10^-6, 1.288219028087343*10^-6, 
 
    1.580407690287831*10^-6, 1.877045712823597*10^-6, 
 
    2.140462609580715*10^-6, 2.362723438152533*10^-6, 
 
    2.57199608647914*10^-6, 2.784877677647986*10^-6, 
 
    2.98563216883559*10^-6, 3.09503250741104*10^-6, 
 
    3.150595922684149*10^-6, 3.166528853113335*10^-6, 
 
    2.988886547403946*10^-6, 2.492106374917842*10^-6, 
 
    1.933378024942794*10^-6, 1.475666905892901*10^-6, 
 
    1.065434392510226*10^-6, 5.55005363790091*10^-7, 
 
    9.32292618816176*10^-8, -(2.670553108623558*10^-7), 
 
    -(3.988703111901088*10^-7), -(4.084341154081072*10^-7), 
 
    -(3.540975664789292*10^-7), -(2.58819267904818*10^-7), 
 
    -(1.831002737752142*10^-7), -(1.466344833878921*10^-7), 
 
    -(1.213504779877749*10^-7), -(1.10073965382436*10^-7), 
 
    -(1.158167448651323*10^-7), -(1.364383019968298*10^-7), 
 
    -(1.365954926720553*10^-7), -(1.453723728490535*10^-7), 
 
    -(1.479769488217106*10^-7), -(1.41634970172499*10^-7), 
 
    -(1.47444009321251*10^-7), -(1.339854197668788*10^-7), 
 
    -(1.257046131219431*10^-7), -(1.234544988821826*10^-7), 
 
    -(1.054765651332993*10^-7), -(1.046161313844963*10^-7), 
 
    -(9.37429438707804*10^-8), -(9.74656670757493*10^-8), 
 
    -(9.24913665843834*10^-8), -(9.70949485453*10^-8), 
 
    -(9.34269668715785*10^-8), -(1.01550695589836*10^-7), 
 
    -(9.69855524616892*10^-8), -(1.005326119768692*10^-7), 
 
    -(9.06245124016539*10^-8), -(9.62147602652539*10^-8), 
 
    -(9.00752325973973*10^-8), -(7.315562586228951*10^-8), 
 
    -(6.783411815293197*10^-8), -(7.349659629537624*10^-8), 
 
    -(4.789230662371842*10^-8), -(3.757979961180338*10^-8), 
 
    -(8.13764725130586*10^-8), -(6.690216147011557*10^-8), 
 
    -(5.617590462533555*10^-8), -(4.196018272266125*10^-8), 
 
    -(5.203576535875661*10^-8), -(4.667646949641552*10^-8), 
 
    -(4.423428662646855*10^-8), -(3.65014847977024*10^-8), 
 
    -(5.370940710920616*10^-8), -(2.422582735029197*10^-8), 
 
    -(3.526472819994382*10^-8), -(1.025336942691501*10^-8), 
 
    -(9.33891184168877*10^-9), -(1.150851226155672*10^-8), 
 
    -(7.640938554646195*10^-9), -(5.109889755525128*10^-9), 
 
    -(1.377428223971716*10^-8)}, 
 
   {0, 6.775870168557942*10^-8, 1.987684329539591*10^-7, 
 
    3.985994250063221*10^-7, 5.796210786041289*10^-7, 
 
    7.758897680212654*10^-7, 9.9561872750808*10^-7, 
 
    1.260126114279432*10^-6, 1.545574518844304*10^-6, 
 
    1.815216571524912*10^-6, 2.058002492755472*10^-6, 
 
    2.298617834765811*10^-6, 2.561759920810676*10^-6, 
 
    2.843125855126533*10^-6, 3.031812087218518*10^-6, 
 
    3.166528853113335*10^-6, 3.300366510295305*10^-6, 
 
    3.28595729611042*10^-6, 2.916234893512386*10^-6, 
 
    2.395229738082984*10^-6, 1.925719093824145*10^-6, 
 
    1.483473703056115*10^-6, 9.02534168149793*10^-7, 
 
    3.322407265551447*10^-7, -(1.528396467926018*10^-7), 
 
    -(3.707282260834468*10^-7), -(4.31379704033641*10^-7), 
 
    -(4.059779288060268*10^-7), -(3.050423043022572*10^-7), 
 
    -(2.160911001339343*10^-7), -(1.700183692151862*10^-7), 
 
    -(1.358086922624347*10^-7), -(1.172150479092326*10^-7), 
 
    -(1.16649687244838*10^-7), -(1.433689802341507*10^-7), 
 
    -(1.424058756842643*10^-7), -(1.526315098617255*10^-7), 
 
    -(1.579202751115315*10^-7), -(1.486162061378992*10^-7), 
 
    -(1.566202070578436*10^-7), -(1.429313865112477*10^-7), 
 
    -(1.315607394537311*10^-7), -(1.316876970307955*10^-7), 
 
    -(1.118252262956589*10^-7), -(1.122429273796497*10^-7), 
 
    -(9.86281247138852*10^-8), -(1.045641386924158*10^-7), 
 
    -(9.67673574233811*10^-8), -(1.032987304606699*10^-7), 
 
    -(9.79331526239624*10^-8), -(1.085392649635501*10^-7), 
 
    -(1.000201063770703*10^-7), -(1.082699157968301*10^-7), 
 
    -(9.39573403548668*10^-8), -(1.031925473655627*10^-7), 
 
    -(9.4090488209824*10^-8), -(7.511285114831856*10^-8), 
 
    -(7.287826722445807*10^-8), -(7.61228241215647*10^-8), 
 
    -(4.99865764505865*10^-8), -(3.564195320459041*10^-8), 
 
    -(8.74516352180387*10^-8), -(7.035716740933135*10^-8), 
 
    -(6.361117995260525*10^-8), -(4.030092995147423*10^-8), 
 
    -(5.729251333193274*10^-8), -(4.517658124479033*10^-8), 
 
    -(4.823308555896118*10^-8), -(3.479478946874299*10^-8), 
 
    -(5.602420207571012*10^-8), -(2.43450505558557*10^-8), 
 
    -(3.795806633432503*10^-8), -(2.377396828687091*10^-8), 
 
    -(8.510942218991*10^-9), -(1.007161979346849*10^-8), 
 
    -(6.722514278265335*10^-9), 1.397072742633965*10^-9, 
 
    -(1.364382625604009*10^-8)}, 
 
   {0, 3.307297528013832*10^-8, 9.70237229703131*10^-8, 
 
    2.026921339535032*10^-7, 3.098455447427066*10^-7, 
 
    4.377789390882309*10^-7, 5.946579659885646*10^-7, 
 
    8.0511967781728*10^-7, 1.052806786399001*10^-6, 
 
    1.305993353860091*10^-6, 1.551965134438907*10^-6, 
 
    1.809594957933191*10^-6, 2.11212465108421*10^-6, 
 
    2.473246800904947*10^-6, 2.753295129401754*10^-6, 
 
    2.988886547403946*10^-6, 3.28595729611042*10^-6, 
 
    3.538783906999407*10^-6, 3.438021988759259*10^-6, 
 
    3.052786728577567*10^-6, 2.620126739808713*10^-6, 
 
    2.171507497737229*10^-6, 1.523658033029619*10^-6, 
 
    8.03655217526066*10^-7, 1.195091812797894*10^-7, 
 
    -(2.545266907494066*10^-7), -(4.209161579128902*10^-7), 
 
    -(4.684507551195664*10^-7), -(3.770176210852314*10^-7), 
 
    -(2.735272247747094*10^-7), -(2.116690432295959*10^-7), 
 
    -(1.621274932731344*10^-7), -(1.3224817136813*10^-7), 
 
    -(1.189151523455072*10^-7), -(1.495022297227705*10^-7), 
 
    -(1.513547949761417*10^-7), -(1.613052763470061*10^-7), 
 
    -(1.704831991584144*10^-7), -(1.606954315009463*10^-7), 
 
    -(1.666977091137193*10^-7), -(1.557087652206243*10^-7), 
 
    -(1.400736918953763*10^-7), -(1.408604482328297*10^-7), 
 
    -(1.223647708415603*10^-7), -(1.205819862792159*10^-7), 
 
    -(1.062154087723972*10^-7), -(1.126658180670135*10^-7), 
 
    -(1.038825180746865*10^-7), -(1.104502690558479*10^-7), 
 
    -(1.055273084059487*10^-7), -(1.171554154211274*10^-7), 
 
    -(1.056048125193652*10^-7), -(1.172790220620746*10^-7), 
 
    -(9.94531937583402*10^-8), -(1.119974656368134*10^-7), 
 
    -(1.000872370247201*10^-7), -(7.795017399290877*10^-8), 
 
    -(7.978524394722092*10^-8), -(8.00886153073743*10^-8), 
 
    -(5.315872203752078*10^-8), -(3.460970855137762*10^-8), 
 
    -(9.4992111998666*10^-8), -(7.531865032088009*10^-8), 
 
    -(7.18305484123527*10^-8), -(3.977307496604137*10^-8), 
 
    -(6.330475802750208*10^-8), -(4.475316166204059*10^-8), 
 
    -(5.298073640475108*10^-8), -(3.416658422628116*10^-8), 
 
    -(5.891568475077173*10^-8), -(2.528660268402052*10^-8), 
 
    -(4.09456417035613*10^-8), -(3.74973031233175*10^-8), 
 
    -(7.946244860327884*10^-9), -(8.77113062086052*10^-9), 
 
    -(5.989278493797947*10^-9), 7.428961333319201*10^-9, 
 
    -(1.354348017198769*10^-8)}, 
 
   {0, 2.562518296272515*10^-9, 7.523358119338443*10^-9, 
 
    2.808459462823384*10^-8, 6.493128942322425*10^-8, 
 
    1.23446074496032*10^-7, 2.105222387567115*10^-7, 
 
    3.49553009665104*10^-7, 5.325346594525771*10^-7, 
 
    7.367071803870427*10^-7, 9.50636032538451*10^-7, 
 
    1.186295429282158*10^-6, 1.480040009798044*10^-6, 
 
    1.862106850289337*10^-6, 2.189419267853208*10^-6, 
 
    2.492106374917842*10^-6, 2.916234893512386*10^-6, 
 
    3.438021988759259*10^-6, 3.736848167508659*10^-6, 
 
    3.651587861289036*10^-6, 3.379762284411729*10^-6, 
 
    3.024637372473875*10^-6, 2.409971572690433*10^-6, 
 
    1.582085609640315*10^-6, 6.721578420072026*10^-7, 
 
    7.212700253397318*10^-8, -(2.776212114867225*10^-7), 
 
    -(4.783676343736723*10^-7), -(4.508729857742645*10^-7), 
 
    -(3.562848385723471*10^-7), -(2.775564572611861*10^-7), 
 
    -(2.069755143024951*10^-7), -(1.628682570604396*10^-7), 
 
    -(1.325816019087651*10^-7), -(1.499150125443925*10^-7), 
 
    -(1.627180394928486*10^-7), -(1.704517234800536*10^-7), 
 
    -(1.803697886922686*10^-7), -(1.794996491238401*10^-7), 
 
    -(1.748607677072298*10^-7), -(1.68451593416529*10^-7), 
 
    -(1.541363345449522*10^-7), -(1.470793078889626*10^-7), 
 
    -(1.353383831202363*10^-7), -(1.252518018091098*10^-7), 
 
    -(1.176771482816669*10^-7), -(1.18214603567099*10^-7), 
 
    -(1.15754925376607*10^-7), -(1.160442932053524*10^-7), 
 
    -(1.151496040666845*10^-7), -(1.22507569544264*10^-7), 
 
    -(1.188007289960845*10^-7), -(1.216577438696685*10^-7), 
 
    -(1.111464110404158*10^-7), -(1.165822946528358*10^-7), 
 
    -(1.095581989369646*10^-7), -(8.55710706301346*10^-8), 
 
    -(8.19116851682599*10^-8), -(8.78707903358196*10^-8), 
 
    -(5.635515749899386*10^-8), -(3.688510998476857*10^-8), 
 
    -(1.010819239277128*10^-7), -(8.44599859080408*10^-8), 
 
    -(6.780474375678207*10^-8), -(5.06928572838411*10^-8), 
 
    -(6.073779589942797*10^-8), -(5.326757455281991*10^-8), 
 
    -(5.368583819214089*10^-8), -(4.336260507367561*10^-8), 
 
    -(5.585097929066013*10^-8), -(3.27131765949526*10^-8), 
 
    -(3.785425466068357*10^-8), -(3.655936253940959*10^-8), 
 
    -(1.080020490883275*10^-8), -(7.422261065559681*10^-9), 
 
    -(7.811873307862785*10^-9), -(2.83764873518872*10^-10), 
 
    -(1.001297507308931*10^-8)}, 
 
   {0, -(1.39620420026048*10^-8), 
 
    -(4.095237660195539*10^-8), -(6.823189274727094*10^-8), 
 
    -(7.350869069435985*10^-8), -(5.956292594092094*10^-8), 
 
    -(2.111367634640091*10^-8), 6.144390720314661*10^-8, 
 
    1.857412674328408*10^-7, 3.369878136945456*10^-7, 
 
    5.062166649767095*10^-7, 7.002675154770427*10^-7, 
 
    9.53067704113124*10^-7, 1.302307209219965*10^-6, 
 
    1.621389389911801*10^-6, 1.933378024942794*10^-6, 
 
    2.395229738082984*10^-6, 3.052786728577567*10^-6, 
 
    3.651587861289036*10^-6, 3.872399000098046*10^-6, 
 
    3.818197307992753*10^-6, 3.629825799230801*10^-6, 
 
    3.16750610609914*10^-6, 2.36679438936608*10^-6, 
 
    1.338253079084069*10^-6, 5.443535562476747*10^-7, 
 
    1.646666360870188*10^-9, -(3.854511436957456*10^-7), 
 
    -(4.699025486287759*10^-7), -(4.241166801807376*10^-7), 
 
    -(3.457959855485298*10^-7), -(2.616040235911168*10^-7), 
 
    -(2.036337347132782*10^-7), -(1.545187807061602*10^-7), 
 
    -(1.506554898151077*10^-7), -(1.708267455345145*10^-7), 
 
    -(1.779112298274836*10^-7), -(1.86321510586079*10^-7), 
 
    -(1.948640938831003*10^-7), -(1.822432053211412*10^-7), 
 
    -(1.773633281990102*10^-7), -(1.669459795196313*10^-7), 
 
    -(1.521336597472171*10^-7), -(1.450927949722586*10^-7), 
 
    -(1.281301905251002*10^-7), -(1.274446806971736*10^-7), 
 
    -(1.221156326375346*10^-7), -(1.260961707497639*10^-7), 
 
    -(1.203171167276006*10^-7), -(1.233144927723073*10^-7), 
 
    -(1.257528299147402*10^-7), -(1.314795434186014*10^-7), 
 
    -(1.237389071454665*10^-7), -(1.224474344352663*10^-7), 
 
    -(1.189323330751122*10^-7), -(1.178934155224383*10^-7), 
 
    -(9.37925173855457*10^-8), -(8.20531291602829*10^-8), 
 
    -(9.52368456326146*10^-8), -(5.917509089486396*10^-8), 
 
    -(4.085609583470684*10^-8), -(1.050006600146956*10^-7), 
 
    -(9.27265993474348*10^-8), -(6.059919350390075*10^-8), 
 
    -(6.310290117087535*10^-8), -(5.608960703673456*10^-8), 
 
    -(6.323064677254733*10^-8), -(5.28704019846571*10^-8), 
 
    -(5.405653286524608*10^-8), -(5.219617056341636*10^-8), 
 
    -(4.049553886412586*10^-8), -(3.372287700176914*10^-8), 
 
    -(2.856949266382935*10^-8), -(1.42284871226768*10^-8), 
 
    -(7.004195252384937*10^-9), -(1.022740009260851*10^-8), 
 
    -(1.160437360257291*10^-8), -(6.712031721770716*10^-9)}\\
 
    , {0, -(2.163454337735367*10^-8), 
 
    -(6.346091561991553*10^-8), -(1.141059708639424*10^-7), 
 
    -(1.415895448738479*10^-7), -(1.528336017132813*10^-7), 
 
    -(1.438743104477522*10^-7), -(9.87939748565501*10^-8), 
 
    -(1.668538458910561*10^-8), 9.32335931416058*10^-8, 
 
    2.240829604424609*10^-7, 3.793794755820889*10^-7, 
 
    5.889774910359446*10^-7, 8.92023955444878*10^-7, 
 
    1.181754749434096*10^-6, 1.475666905892901*10^-6, 
 
    1.925719093824145*10^-6, 2.620126739808713*10^-6, 
 
    3.379762284411729*10^-6, 3.818197307992753*10^-6, 
 
    3.956185101470004*10^-6, 3.938493456162459*10^-6, 
 
    3.674504836060503*10^-6, 2.998813970052337*10^-6, 
 
    1.965660581837623*10^-6, 1.047420929374141*10^-6, 
 
    3.44594728245643*10^-7, -(2.177220940499579*10^-7), 
 
    -(4.319202525838912*10^-7), -(4.616561629878598*10^-7), 
 
    -(4.022837797677418*10^-7), -(3.163889159609809*10^-7), 
 
    -(2.472738689398016*10^-7), -(1.797089140906567*10^-7), 
 
    -(1.558723935595409*10^-7), -(1.75053102909392*10^-7), 
 
    -(1.837410487234073*10^-7), -(1.906888168716559*10^-7), 
 
    -(2.039968522438576*10^-7), -(1.902731708222356*10^-7), 
 
    -(1.83475990353985*10^-7), -(1.760238779416832*10^-7), 
 
    -(1.579860511224797*10^-7), -(1.508991772492007*10^-7), 
 
    -(1.322433586261645*10^-7), -(1.338246536837516*10^-7), 
 
    -(1.263930201789782*10^-7), -(1.328214784034435*10^-7), 
 
    -(1.246108770767536*10^-7), -(1.290426327456096*10^-7), 
 
    -(1.291439292760648*10^-7), -(1.39960621409864*10^-7), 
 
    -(1.267748174260436*10^-7), -(1.301816442500402*10^-7), 
 
    -(1.219397876820381*10^-7), -(1.23334185733825*10^-7), 
 
    -(1.002242362272027*10^-7), -(8.28932590558158*10^-8), 
 
    -(1.004805929566053*10^-7), -(6.154306536241338*10^-8), 
 
    -(4.384118210661783*10^-8), -(1.08285586575396*10^-7), 
 
    -(9.85284180507389*10^-8), -(5.703260252552877*10^-8), 
 
    -(7.102972492603387*10^-8), -(5.398112602277007*10^-8), 
 
    -(6.97344319484467*10^-8), -(5.294076369073835*10^-8), 
 
    -(6.096189214691707*10^-8), -(5.057450663912502*10^-8), 
 
    -(4.552346099387526*10^-8), -(3.159964626462507*10^-8), 
 
    -(2.399602939800868*10^-8), -(1.638571595844432*10^-8), 
 
    -(6.899569090893843*10^-9), -(1.175364413585277*10^-8), 
 
    -(1.846500075664753*10^-8), -(4.8815047927685*10^-9)}, 
 
   {0, -(2.580501894319802*10^-8), 
 
    -(7.569613180315001*10^-8), -(1.399947255962142*10^-7), 
 
    -(1.817283611764961*10^-7), -(2.103325091646376*10^-7), 
 
    -(2.230109429629167*10^-7), -(2.073966420552931*10^-7), 
 
    -(1.602630361916417*10^-7), -(8.62809660661682*10^-8), 
 
    9.55651398660053*10^-9, 1.28185329759815*10^-7, 
 
    2.948532797104595*10^-7, 5.476653995094184*10^-7, 
 
    8.00324018718746*10^-7, 1.065434392510226*10^-6, 
 
    1.483473703056115*10^-6, 2.171507497737229*10^-6, 
 
    3.024637372473875*10^-6, 3.629825799230801*10^-6, 
 
    3.938493456162459*10^-6, 4.087688177253721*10^-6, 
 
    4.044397067732005*10^-6, 3.552965381499987*10^-6, 
 
    2.587971272213051*10^-6, 1.589775128250572*10^-6, 
 
    7.448199873774453*10^-7, 9.535319897661*10^-9, 
 
    -(3.50877578800777*10^-7), -(4.77451330780437*10^-7), 
 
    -(4.513868262364972*10^-7), -(3.737653160367798*10^-7), 
 
    -(2.957432665660325*10^-7), -(2.087056792263997*10^-7), 
 
    -(1.654879481445826*10^-7), -(1.77799387219513*10^-7), 
 
    -(1.885696642810701*10^-7), -(1.950765064904711*10^-7), 
 
    -(2.100826473813799*10^-7), -(1.989189512880837*10^-7), 
 
    -(1.891919839439126*10^-7), -(1.832555141392253*10^-7), 
 
    -(1.647275708118605*10^-7), -(1.552083072973341*10^-7), 
 
    -(1.376952758120291*10^-7), -(1.3868465552991*10^-7), 
 
    -(1.314102307087529*10^-7), -(1.377428551211537*10^-7), 
 
    -(1.29480064873806*10^-7), -(1.33652183790814*10^-7), 
 
    -(1.335047695078338*10^-7), -(1.459415867075073*10^-7), 
 
    -(1.312247321542408*10^-7), -(1.357779901264547*10^-7), 
 
    -(1.260701783651497*10^-7), -(1.276469303937555*10^-7), 
 
    -(1.052406003189934*10^-7), -(8.49017567804884*10^-8), 
 
    -(1.046098784494435*10^-7), -(6.379470239933854*10^-8), 
 
    -(4.591793643958646*10^-8), -(1.119005387236541*10^-7), 
 
    -(1.028342326332272*10^-7), -(5.709622796673044*10^-8), 
 
    -(7.5432864403502*10^-8), -(5.447729217904913*10^-8), 
 
    -(7.359656005869666*10^-8), -(5.417879674918649*10^-8), 
 
    -(6.479454197071339*10^-8), -(5.111381535942382*10^-8), 
 
    -(4.83572361732352*10^-8), -(3.154155411082922*10^-8), 
 
    -(2.301142415878518*10^-8), -(1.75143557088448*10^-8), 
 
    -(7.007478961270424*10^-9), -(1.255165212176501*10^-8), 
 
    -(2.129118428460117*10^-8), -(4.328665235494034*10^-9)}\\
 
    , {0, -(2.741239285351232*10^-8), 
 
    -(8.04122877409989*10^-8), -(1.520752228628169*10^-7), 
 
    -(2.040967159580842*10^-7), -(2.47416486931664*10^-7), 
 
    -(2.806575564863543*10^-7), -(2.961554126141283*10^-7), 
 
    -(2.886212941329042*10^-7), -(2.576956674727011*10^-7), 
 
    -(2.060055796095637*10^-7), -(1.352899548668782*10^-7), 
 
    -(2.727358531817848*10^-8), 1.51486947238678*10^-7, 
 
    3.433708295342628*10^-7, 5.550053637900911*10^-7, 
 
    9.02534168149794*10^-7, 1.523658033029619*10^-6, 
 
    2.409971572690433*10^-6, 3.16750610609914*10^-6, 
 
    3.674504836060503*10^-6, 4.044397067732005*10^-6, 
 
    4.341793728747376*10^-6, 4.20949929142001*10^-6, 
 
    3.468290416297641*10^-6, 2.437923883279844*10^-6, 
 
    1.424289301093149*10^-6, 4.470014851478887*10^-7, 
 
    -(1.520679962337401*10^-7), -(4.566471715942924*10^-7), 
 
    -(5.056356884306864*10^-7), -(4.603983776989487*10^-7), 
 
    -(3.749356290806553*10^-7), -(2.58044977153957*10^-7), 
 
    -(1.888402718221578*10^-7), -(1.801021695107042*10^-7), 
 
    -(1.934950143567467*10^-7), -(2.022783449973151*10^-7), 
 
    -(2.139893464000442*10^-7), -(2.123686736023032*10^-7), 
 
    -(1.978704788821142*10^-7), -(1.905937832810888*10^-7), 
 
    -(1.762299117803099*10^-7), -(1.595305798595489*10^-7), 
 
    -(1.480313044417445*10^-7), -(1.435127762066011*10^-7), 
 
    -(1.401488489366911*10^-7), -(1.419319769519385*10^-7), 
 
    -(1.378726682703845*10^-7), -(1.385709771486115*10^-7), 
 
    -(1.423827075672162*10^-7), -(1.499246167184674*10^-7), 
 
    -(1.408089305879602*10^-7), -(1.397990775232083*10^-7), 
 
    -(1.346132127394328*10^-7), -(1.326573323521791*10^-7), 
 
    -(1.091812887179048*10^-7), -(9.07322165703013*10^-8), 
 
    -(1.08631495842743*10^-7), -(6.694936898342628*10^-8), 
 
    -(4.711361567224383*10^-8), -(1.183030950220164*10^-7), 
 
    -(1.059777231434997*10^-7), -(6.542886818801826*10^-8), 
 
    -(7.425844638014477*10^-8), -(6.122665261825665*10^-8), 
 
    -(7.371008404422632*10^-8), -(5.875549267977947*10^-8), 
 
    -(6.367941565950986*10^-8), -(5.729539113262885*10^-8), 
 
    -(4.755300265205343*10^-8), -(3.643463960462872*10^-8), 
 
    -(2.888442966214932*10^-8), -(1.694746986259094*10^-8), 
 
    -(7.841320949774391*10^-9), -(1.218899574984197*10^-8), 
 
    -(1.659963345755695*10^-8), -(6.615002453873266*10^-9)}\\
 
    , {0, -(2.396845770092741*10^-8), 
 
    -(7.030965433180487*10^-8), -(1.354509041279704*10^-7), 
 
    -(1.865378373534128*10^-7), -(2.336277609570641*10^-7), 
 
    -(2.765619063770018*10^-7), -(3.122799532994247*10^-7), 
 
    -(3.348188525303088*10^-7), -(3.402166142984369*10^-7), 
 
    -(3.289788086966379*10^-7), -(3.045201693664007*10^-7), 
 
    -(2.566088030931431*10^-7), -(1.608267788412785*10^-7), 
 
    -(4.465458351972434*10^-8), 9.32292618816176*10^-8, 
 
    3.322407265551448*10^-7, 8.03655217526067*10^-7, 
 
    1.582085609640315*10^-6, 2.36679438936608*10^-6, 
 
    2.998813970052337*10^-6, 3.552965381499987*10^-6, 
 
    4.20949929142001*10^-6, 4.605283572697228*10^-6, 
 
    4.374895331960004*10^-6, 3.525317738783518*10^-6, 
 
    2.439070473403637*10^-6, 1.230908474338207*10^-6, 
 
    2.985382843611058*10^-7, -(3.0221143347463*10^-7), 
 
    -(5.059292247531733*10^-7), -(5.562648807321777*10^-7), 
 
    -(4.897391129118994*10^-7), -(3.434237737446498*10^-7), 
 
    -(2.358414370977044*10^-7), -(1.86193478052114*10^-7), 
 
    -(1.966784906320862*10^-7), -(2.1049603858554*10^-7), 
 
    -(2.144430646090627*10^-7), -(2.267767567431749*10^-7), 
 
    -(2.096476873011436*10^-7), -(1.961632784170039*10^-7), 
 
    -(1.902910972590317*10^-7), -(1.652379088477125*10^-7), 
 
    -(1.612239405956974*10^-7), -(1.472791320667466*10^-7), 
 
    -(1.515046759949108*10^-7), -(1.445600001734374*10^-7), 
 
    -(1.486845120333576*10^-7), -(1.436715378948794*10^-7), 
 
    -(1.550688116767284*10^-7), -(1.503292105111529*10^-7), 
 
    -(1.546906802911191*10^-7), -(1.411655603458695*10^-7), 
 
    -(1.469206674046436*10^-7), -(1.373927029573082*10^-7), 
 
    -(1.105866928732504*10^-7), -(1.010087152326053*10^-7), 
 
    -(1.112860362541109*10^-7), -(7.090501259153898*10^-8), 
 
    -(4.725860585194964*10^-8), -(1.268031653267196*10^-7), 
 
    -(1.06861821181529*10^-7), -(8.32250884740313*10^-8), 
 
    -(6.570509688282913*10^-8), -(7.501039565244883*10^-8), 
 
    -(6.859876055269969*10^-8), -(6.684049993527358*10^-8), 
 
    -(5.624724582699294*10^-8), -(6.946575171089933*10^-8), 
 
    -(4.217409725504436*10^-8), -(4.669245289650802*10^-8), 
 
    -(4.285591610398696*10^-8), -(1.417658564531041*10^-8), 
 
    -(9.50677044580229*10^-9), -(1.029581512191913*10^-8), 
 
    -(2.857882117906212*10^-9), -(1.218815199320223*10^-8)}\\
 
    , {0, -(1.742719982918614*10^-8), 
 
    -(5.11183366782816*10^-8), -(1.004917263267112*10^-7), 
 
    -(1.422227920770767*10^-7), -(1.839896415098104*10^-7), 
 
    -(2.265284468604031*10^-7), -(2.706557925639323*10^-7), 
 
    -(3.110629945922256*10^-7), -(3.419592172117485*10^-7), 
 
    -(3.622779642788415*10^-7), -(3.760751353407548*10^-7), 
 
    -(3.809316639777465*10^-7), -(3.637226811168325*10^-7), 
 
    -(3.248511374642995*10^-7), -(2.67055310862356*10^-7), 
 
    -(1.528396467926018*10^-7), 1.195091812797894*10^-7, 
 
    6.721578420072022*10^-7, 1.338253079084069*10^-6, 
 
    1.965660581837623*10^-6, 2.587971272213051*10^-6, 
 
    3.468290416297641*10^-6, 4.374895331960004*10^-6, 
 
    4.922194825676337*10^-6, 4.63564941607864*10^-6, 
 
    3.763532447975852*10^-6, 2.501633550739523*10^-6, 
 
    1.191793647563649*10^-6, 1.588233554564633*10^-7, 
 
    -(3.221882315184104*10^-7), -(5.901571319277997*10^-7), 
 
    -(6.235482478636623*10^-7), -(4.921502992256741*10^-7), 
 
    -(3.206327041219605*10^-7), -(2.136012228796034*10^-7), 
 
    -(2.010377634019215*10^-7), -(2.151328457801576*10^-7), 
 
    -(2.178578423894468*10^-7), -(2.336973503489246*10^-7), 
 
    -(2.235663981567001*10^-7), -(2.031271316395216*10^-7), 
 
    -(2.002212521680222*10^-7), -(1.768986463264552*10^-7), 
 
    -(1.700883697939414*10^-7), -(1.534369924934287*10^-7), 
 
    -(1.597755221549836*10^-7), -(1.504038821732998*10^-7), 
 
    -(1.568870201175065*10^-7), -(1.525964935277752*10^-7), 
 
    -(1.662922822599763*10^-7), -(1.538558416984144*10^-7), 
 
    -(1.657218096824716*10^-7), -(1.447298429018368*10^-7), 
 
    -(1.583960671694182*10^-7), -(1.436145981912624*10^-7), 
 
    -(1.121113517647253*10^-7), -(1.121499577093622*10^-7), 
 
    -(1.147137792186392*10^-7), -(7.577670325201871*10^-8), 
 
    -(5.063557697635596*10^-8), -(1.341776361149469*10^-7), 
 
    -(1.084287897601104*10^-7), -(9.79651641482066*10^-8), 
 
    -(6.029589452983375*10^-8), -(8.70397305651506*10^-8), 
 
    -(6.645772537970957*10^-8), -(7.387427153383241*10^-8), 
 
    -(5.199974567527751*10^-8), -(8.13081906825617*10^-8), 
 
    -(3.831638076968228*10^-8), -(5.579801290950978*10^-8), 
 
    -(4.853679593053196*10^-8), -(1.241188833230183*10^-8), 
 
    -(1.243305542437569*10^-8), -(9.32738298896672*10^-9), 
 
    5.821843744207314*10^-9, -(1.797453208947519*10^-8)}, 
 
   {0, -(1.209400889885329*10^-8), 
 
    -(3.547440512906396*10^-8), -(7.024027812163049*10^-8), 
 
    -(1.005268109265713*10^-7), -(1.321334684908752*10^-7), 
 
    -(1.661364185678616*10^-7), -(2.051106093539784*10^-7), 
 
    -(2.454525077707882*10^-7), -(2.819363522024742*10^-7), 
 
    -(3.131039548376201*10^-7), -(3.425022982759574*10^-7), 
 
    -(3.721162555931579*10^-7), -(3.972047832752149*10^-7), 
 
    -(4.047617271362939*10^-7), -(3.988703111901088*10^-7), 
 
    -(3.707282260834468*10^-7), -(2.545266907494061*10^-7), 
 
    7.212700253397336*10^-8, 5.443535562476746*10^-7, 
 
    1.047420929374141*10^-6, 1.589775128250572*10^-6, 
 
    2.437923883279844*10^-6, 3.525317738783518*10^-6, 
 
    4.63564941607864*10^-6, 5.083268461467814*10^-6, 
 
    4.745687698472688*10^-6, 3.781480274971613*10^-6, 
 
    2.303271799851558*10^-6, 8.98905938555498*10^-7, 
 
    1.104919988123994*10^-7, -(4.528491248574952*10^-7), 
 
    -(6.756936168216697*10^-7), -(6.593348965027155*10^-7), 
 
    -(4.245198374048124*10^-7), -(2.662175550458312*10^-7), 
 
    -(2.139391050292805*10^-7), -(2.114746643776313*10^-7), 
 
    -(2.244111551776717*10^-7), -(2.289398813264182*10^-7), 
 
    -(2.290202023197313*10^-7), -(2.145158025142555*10^-7), 
 
    -(1.995674846503112*10^-7), -(1.870140784529804*10^-7), 
 
    -(1.677374838024992*10^-7), -(1.625109638506113*10^-7), 
 
    -(1.590749934070594*10^-7), -(1.605339851174842*10^-7), 
 
    -(1.574996507675295*10^-7), -(1.59766992334653*10^-7), 
 
    -(1.658607415281403*10^-7), -(1.66837351636552*10^-7), 
 
    -(1.630565978229192*10^-7), -(1.557087135956657*10^-7), 
 
    -(1.567311566572465*10^-7), -(1.517748633304545*10^-7), 
 
    -(1.194051660249659*10^-7), -(1.098138905394264*10^-7), 
 
    -(1.217282716013453*10^-7), -(7.772697583064227*10^-8), 
 
    -(5.605938510427331*10^-8), -(1.354209474086687*10^-7), 
 
    -(1.15763303686979*10^-7), -(8.58800615516989*10^-8), 
 
    -(7.5125491921868*10^-8), -(7.904953198514159*10^-8), 
 
    -(7.854996708212708*10^-8), -(7.082786247639526*10^-8), 
 
    -(6.482689448819193*10^-8), -(7.633342986940738*10^-8), 
 
    -(4.680001279443333*10^-8), -(4.998087793139588*10^-8), 
 
    -(3.181012007033265*10^-8), -(1.672101604300977*10^-8), 
 
    -(1.28512345923724*10^-8), -(1.257361819135307*10^-8), 
 
    -(1.039711513745989*10^-8), -(1.451830450749739*10^-8)}\\
 
    , {0, -(8.51771400974781*10^-9), 
 
    -(2.498580486998736*10^-8), -(4.941738102616256*10^-8), 
 
    -(7.074976965826743*10^-8), -(9.33714591177501*10^-8), 
 
    -(1.183309360677975*10^-7), -(1.484076559608462*10^-7), 
 
    -(1.815258226985328*10^-7), -(2.139269970684058*10^-7), 
 
    -(2.445787953256419*10^-7), -(2.762656241874129*10^-7), 
 
    -(3.129484279220893*10^-7), -(3.552637178102913*10^-7), 
 
    -(3.857649089880777*10^-7), -(4.08434115408107*10^-7), 
 
    -(4.313797040336415*10^-7), -(4.209161579128901*10^-7), 
 
    -(2.776212114867225*10^-7), 1.646666360870234*10^-9, 
 
    3.44594728245643*10^-7, 7.448199873774449*10^-7, 
 
    1.424289301093149*10^-6, 2.439070473403638*10^-6, 
 
    3.763532447975851*10^-6, 4.745687698472688*10^-6, 
 
    5.06689612106174*10^-6, 4.733296231177098*10^-6, 
 
    3.487832451286493*10^-6, 1.940111869569384*10^-6, 
 
    8.84556652969639*10^-7, -(1.939318749513744*10^-8), 
 
    -(5.265938575502737*10^-7), -(7.741383942224627*10^-7), 
 
    -(5.568792936955264*10^-7), -(3.510481308217209*10^-7), 
 
    -(2.517117218286045*10^-7), -(2.15065535619237*10^-7), 
 
    -(2.233880550561191*10^-7), -(2.239571020074632*10^-7), 
 
    -(2.261819556111504*10^-7), -(2.214701317776424*10^-7), 
 
    -(2.002378134117148*10^-7), -(1.906131820000792*10^-7), 
 
    -(1.661826722792256*10^-7), -(1.675372372254516*10^-7), 
 
    -(1.585139256882146*10^-7), -(1.665512953118111*10^-7), 
 
    -(1.57372370525587*10^-7), -(1.635492556550961*10^-7), 
 
    -(1.633401091950128*10^-7), -(1.763430380761484*10^-7), 
 
    -(1.595639517001423*10^-7), -(1.641479485069339*10^-7), 
 
    -(1.538119805103844*10^-7), -(1.556736612270081*10^-7), 
 
    -(1.268850336941128*10^-7), -(1.054956912376992*10^-7), 
 
    -(1.266667648619785*10^-7), -(7.870785437219375*10^-8), 
 
    -(6.026370229678335*10^-8), -(1.347670274884664*10^-7), 
 
    -(1.220091807596249*10^-7), -(7.256408943604632*10^-8), 
 
    -(8.89638138236661*10^-8), -(6.966143187378066*10^-8), 
 
    -(8.91417199692626*10^-8), -(6.677334331588208*10^-8), 
 
    -(7.697228869539279*10^-8), -(6.876340557095367*10^-8), 
 
    -(5.536108438927798*10^-8), -(4.248668696801332*10^-8), 
 
    -(1.751836063513988*10^-8), -(2.074484806988369*10^-8), 
 
    -(1.23443300542525*10^-8), -(1.545923029117818*10^-8), 
 
    -(2.606093886815372*10^-8), -(9.94264996531597*10^-9)}, 
 
   {0, -(5.721800371106298*10^-9), 
 
    -(1.678604840258178*10^-8), -(3.312303605499061*10^-8), 
 
    -(4.730750290027386*10^-8), -(6.240711202905759*10^-8), 
 
    -(7.924458588967973*10^-8), -(1.000512193568614*10^-7), 
 
    -(1.238158230384438*10^-7), -(1.482536138916604*10^-7), 
 
    -(1.729824870080831*10^-7), -(2.000423660989004*10^-7), 
 
    -(2.338419738433015*10^-7), -(2.784620370655342*10^-7), 
 
    -(3.174632702203613*10^-7), -(3.540975664789291*10^-7), 
 
    -(4.05977928806027*10^-7), -(4.684507551195661*10^-7), 
 
    -(4.783676343736725*10^-7), -(3.854511436957456*10^-7), 
 
    -(2.177220940499581*10^-7), 9.53531989766082*10^-9, 
 
    4.470014851478887*10^-7, 1.230908474338207*10^-6, 
 
    2.501633550739523*10^-6, 3.781480274971613*10^-6, 
 
    4.733296231177099*10^-6, 5.253527361786824*10^-6, 
 
    4.684785581472294*10^-6, 3.284733716644289*10^-6, 
 
    2.041686800007216*10^-6, 7.71732913257293*10^-7, 
 
    -(1.119184614211175*10^-7), -(7.904165322514003*10^-7), 
 
    -(7.164602936220943*10^-7), -(4.734527274482215*10^-7), 
 
    -(3.201183743671547*10^-7), -(2.36878097443178*10^-7), 
 
    -(2.146114443692931*10^-7), -(2.207209828670367*10^-7), 
 
    -(2.194020002766499*10^-7), -(2.212159437185895*10^-7), 
 
    -(2.052298901679141*10^-7), -(1.881522159252389*10^-7), 
 
    -(1.709755779686664*10^-7), -(1.675418144726423*10^-7), 
 
    -(1.615644612892956*10^-7), -(1.663073619818002*10^-7), 
 
    -(1.600537808515299*10^-7), -(1.630082284693394*10^-7), 
 
    -(1.647546920008866*10^-7), -(1.772866261588248*10^-7), 
 
    -(1.619117637939742*10^-7), -(1.652049088495244*10^-7), 
 
    -(1.553664694759286*10^-7), -(1.557522510062852*10^-7), 
 
    -(1.298420177731243*10^-7), -(1.050154870825629*10^-7), 
 
    -(1.280661684978376*10^-7), -(7.912984952173998*10^-8), 
 
    -(6.031233437389381*10^-8), -(1.359904333278468*10^-7), 
 
    -(1.236191152346351*10^-7), -(7.210371050328344*10^-8), 
 
    -(9.06476392968575*10^-8), -(6.933386754921662*10^-8), 
 
    -(9.03855056322994*10^-8), -(6.699408641375722*10^-8), 
 
    -(7.832326523567344*10^-8), -(6.818369646230762*10^-8), 
 
    -(5.657532618673731*10^-8), -(4.201000502533015*10^-8), 
 
    -(1.828107539713625*10^-8), -(2.116517018005931*10^-8), 
 
    -(1.193396114734357*10^-8), -(1.569437422630029*10^-8), 
 
    -(2.693966060786506*10^-8), -(9.23804161507576*10^-9)}, 
 
   {0, -(4.137442549833068*10^-9), 
 
    -(1.213801288408694*10^-8), -(2.43402834823089*10^-8), 
 
    -(3.529903057245129*10^-8), -(4.69795651101519*10^-8), 
 
    -(5.98780683899935*10^-8), -(7.531155467438845*10^-8), 
 
    -(9.23552377389825*10^-8), -(1.093807968762113*10^-7), 
 
    -(1.263012385508842*10^-7), -(1.446926295687052*10^-7), 
 
    -(1.676608247507366*10^-7), -(1.991758172858602*10^-7), 
 
    -(2.287612990258918*10^-7), -(2.58819267904818*10^-7), 
 
    -(3.050423043022575*10^-7), -(3.770176210852314*10^-7), 
 
    -(4.508729857742643*10^-7), -(4.699025486287759*10^-7), 
 
    -(4.319202525838912*10^-7), -(3.50877578800777*10^-7), 
 
    -(1.520679962337399*10^-7), 2.985382843611054*10^-7, 
 
    1.191793647563649*10^-6, 2.303271799851558*10^-6, 
 
    3.487832451286493*10^-6, 4.684785581472294*10^-6, 
 
    5.2230256157321*10^-6, 4.570817014613091*10^-6, 
 
    3.481459903039558*10^-6, 2.036538868518975*10^-6, 
 
    7.794204395088201*10^-7, -(4.93766907421539*10^-7), 
 
    -(8.16111135869694*10^-7), -(6.305395286369585*10^-7), 
 
    -(4.325818136096266*10^-7), -(3.0004289947371*10^-7), 
 
    -(2.125828074375758*10^-7), -(2.172112585754838*10^-7), 
 
    -(2.141129983731667*10^-7), -(2.097016065249083*10^-7), 
 
    -(2.120627948481978*10^-7), -(1.85581124955581*10^-7), 
 
    -(1.808980167769927*10^-7), -(1.626600473610054*10^-7), 
 
    -(1.683644748314891*10^-7), -(1.595260055866427*10^-7), 
 
    -(1.658797400161971*10^-7), -(1.599033957925263*10^-7), 
 
    -(1.736531479295758*10^-7), -(1.657391629549823*10^-7), 
 
    -(1.729978997833806*10^-7), -(1.555389028154143*10^-7), 
 
    -(1.647957369971706*10^-7), -(1.528719926773609*10^-7), 
 
    -(1.226345735325798*10^-7), -(1.145279592146049*10^-7), 
 
    -(1.237131682749499*10^-7), -(7.982074916346789*10^-8), 
 
    -(5.53051091887569*10^-8), -(1.409407456988138*10^-7), 
 
    -(1.169389647329332*10^-7), -(9.60047705673303*10^-8), 
 
    -(7.060114663895945*10^-8), -(8.67366322961196*10^-8), 
 
    -(7.567102324068169*10^-8), -(7.556907470874014*10^-8), 
 
    -(6.072016580905581*10^-8), -(8.26028640148245*10^-8), 
 
    -(4.417986569534742*10^-8), -(5.551328579941255*10^-8), 
 
    -(4.201375420551261*10^-8), -(1.518925948565161*10^-8), 
 
    -(1.316868733998202*10^-8), -(1.14316559234922*10^-8), 
 
    -(2.129248393776133*10^-9), -(1.70504412383335*10^-8)}, 
 
   {0, -(3.629022422148475*10^-9), 
 
    -(1.064480517395976*10^-8), -(2.112223686527233*10^-8), 
 
    -(3.02528780650204*10^-8), -(3.969686230837859*10^-8), 
 
    -(4.976044271864004*10^-8), -(6.123208147974423*10^-8), 
 
    -(7.337672992596535*10^-8), -(8.50357784551321*10^-8), 
 
    -(9.6224594624662*10^-8), -(1.081390057023592*10^-7), 
 
    -(1.227540064182589*10^-7), -(1.429229734047493*10^-7), 
 
    -(1.62467336497999*10^-7), -(1.831002737752142*10^-7), 
 
    -(2.160911001339345*10^-7), -(2.735272247747094*10^-7), 
 
    -(3.562848385723471*10^-7), -(4.241166801807376*10^-7), 
 
    -(4.616561629878601*10^-7), -(4.77451330780437*10^-7), 
 
    -(4.566471715942928*10^-7), -(3.0221143347463*10^-7), 
 
    1.588233554564634*10^-7, 8.98905938555497*10^-7, 
 
    1.940111869569384*10^-6, 3.284733716644289*10^-6, 
 
    4.570817014613091*10^-6, 5.065821168932468*10^-6, 
 
    4.722703518103559*10^-6, 3.680649251714787*10^-6, 
 
    2.341510735565596*10^-6, 4.924533301000129*10^-7, 
 
    -(5.648605425984672*10^-7), -(7.516768282651519*10^-7), 
 
    -(5.951586732390603*10^-7), -(4.276222017059762*10^-7), 
 
    -(2.775666931897438*10^-7), -(2.198472308538508*10^-7), 
 
    -(2.141689225008934*10^-7), -(2.061511377776233*10^-7), 
 
    -(2.08474110502439*10^-7), -(1.945869165007755*10^-7), 
 
    -(1.827585842139026*10^-7), -(1.683100230968761*10^-7), 
 
    -(1.701520877667771*10^-7), -(1.6514045504713*10^-7), 
 
    -(1.677174881791239*10^-7), -(1.65834538094328*10^-7), 
 
    -(1.774205936513165*10^-7), -(1.702080754596355*10^-7), 
 
    -(1.758989755348161*10^-7), -(1.589732348273151*10^-7), 
 
    -(1.689037245277025*10^-7), -(1.572908061848103*10^-7), 
 
    -(1.233186834339806*10^-7), -(1.189659412029915*10^-7), 
 
    -(1.263094050690474*10^-7), -(8.19079772085155*10^-8), 
 
    -(5.679044499071755*10^-8), -(1.443179898089194*10^-7), 
 
    -(1.192769027802469*10^-7), -(9.93038472689415*10^-8), 
 
    -(7.140594472470605*10^-8), -(8.96115205124623*10^-8), 
 
    -(7.693217712223566*10^-8), -(7.770393332506674*10^-8), 
 
    -(6.14729761091068*10^-8), -(8.54446606166698*10^-8), 
 
    -(4.460772924802109*10^-8), -(5.751725949239396*10^-8), 
 
    -(4.309302078947967*10^-8), -(1.53079029742069*10^-8), 
 
    -(1.378954669475401*10^-8), -(1.156855469525527*10^-8), 
 
    -(1.385389313590076*10^-9), -(1.802592605766932*10^-8)}\\
 
    , {0, -(4.238169336614342*10^-9), 
 
    -(1.243368324881083*10^-8), -(2.387681817265895*10^-8), 
 
    -(3.286961071105057*10^-8), -(4.151744618676145*10^-8), 
 
    -(5.009379530717202*10^-8), -(5.918580465868501*10^-8), 
 
    -(6.838313587488658*10^-8), -(7.695967903180837*10^-8), 
 
    -(8.50496362664551*10^-8), -(9.36040486920506*10^-8), 
 
    -(1.040407149149555*10^-7), -(1.183087010550079*10^-7), 
 
    -(1.320755919167986*10^-7), -(1.466344833878921*10^-7), 
 
    -(1.70018369215186*10^-7), -(2.116690432295961*10^-7), 
 
    -(2.775564572611862*10^-7), -(3.457959855485297*10^-7), 
 
    -(4.022837797677419*10^-7), -(4.51386826236497*10^-7), 
 
    -(5.056356884306866*10^-7), -(5.059292247531731*10^-7), 
 
    -(3.2218823151841*10^-7), 1.104919988123998*10^-7, 
 
    8.84556652969639*10^-7, 2.041686800007216*10^-6, 
 
    3.481459903039558*10^-6, 4.722703518103559*10^-6, 
 
    5.157319564385792*10^-6, 4.802233868906062*10^-6, 
 
    3.727328263394327*10^-6, 1.67662864966244*10^-6, 
 
    -(4.564478733506953*10^-8), -(7.285908061031688*10^-7), 
 
    -(7.100270872876722*10^-7), -(5.538165926301247*10^-7), 
 
    -(3.738429556955112*10^-7), -(2.423783797234565*10^-7), 
 
    -(2.197539637745958*10^-7), -(2.131282064988776*10^-7), 
 
    -(2.039061050514145*10^-7), -(2.03282610478661*10^-7), 
 
    -(1.823740944999894*10^-7), -(1.795877084359878*10^-7), 
 
    -(1.714280458709316*10^-7), -(1.770116236173233*10^-7), 
 
    -(1.696962633433407*10^-7), -(1.739583467113017*10^-7), 
 
    -(1.776277390443428*10^-7), -(1.854893743436481*10^-7), 
 
    -(1.741937753118119*10^-7), -(1.7228431945926*10^-7), 
 
    -(1.677612590308941*10^-7), -(1.665091153203175*10^-7), 
 
    -(1.32698978763037*10^-7), -(1.160382469070118*10^-7), 
 
    -(1.346702450640327*10^-7), -(8.40925372351243*10^-8), 
 
    -(6.147261414488461*10^-8), -(1.46758030415983*10^-7), 
 
    -(1.28997993804179*10^-7), -(8.54614547638972*10^-8), 
 
    -(8.8930290998489*10^-8), -(8.00635959293244*10^-8), 
 
    -(9.05752512685451*10^-8), -(7.441893763625938*10^-8), 
 
    -(7.649395778755332*10^-8), -(7.783265059766143*10^-8), 
 
    -(5.543644823328727*10^-8), -(4.979550650344395*10^-8), 
 
    -(2.904216345239295*10^-8), -(2.027645861747476*10^-8), 
 
    -(1.288418012890588*10^-8), -(1.507566992583518*10^-8), 
 
    -(1.905070579038487*10^-8), -(1.256094109879312*10^-8)}\\
 
    , {0, -(5.08312190870567*10^-9), 
 
    -(1.491223849883073*10^-8), -(2.830554506385708*10^-8), 
 
    -(3.83685176430335*10^-8), -(4.757909231742433*10^-8), 
 
    -(5.6147812779097*10^-8), -(6.428344834012434*10^-8), 
 
    -(7.160746613411134*10^-8), -(7.770933553116402*10^-8), 
 
    -(8.30545415213588*10^-8), -(8.85456437929728*10^-8), 
 
    -(9.5118048659168*10^-8), -(1.039569848554796*10^-7), 
 
    -(1.12402552226596*10^-7), -(1.213504779877747*10^-7), 
 
    -(1.358086922624347*10^-7), -(1.621274932731345*10^-7), 
 
    -(2.069755143024952*10^-7), -(2.616040235911168*10^-7), 
 
    -(3.163889159609809*10^-7), -(3.737653160367798*10^-7), 
 
    -(4.603983776989485*10^-7), -(5.562648807321777*10^-7), 
 
    -(5.90157131927799*10^-7), -(4.528491248574952*10^-7), 
 
    -(1.939318749513771*10^-8), 7.71732913257293*10^-7, 
 
    2.036538868518975*10^-6, 3.680649251714787*10^-6, 
 
    4.802233868906062*10^-6, 5.436436220080269*10^-6, 
 
    5.140787007552835*10^-6, 3.489681427967563*10^-6, 
 
    1.11673597836931*10^-6, -(3.677571841452582*10^-7), 
 
    -(7.255587741640691*10^-7), -(7.062724399081564*10^-7), 
 
    -(5.383187651735978*10^-7), -(3.255621413553397*10^-7), 
 
    -(2.482212097180217*10^-7), -(2.285478508470983*10^-7), 
 
    -(2.105073648786027*10^-7), -(2.104209508901212*10^-7), 
 
    -(1.942833400966805*10^-7), -(1.93940055314787*10^-7), 
 
    -(1.838453139713187*10^-7), -(1.914656288945571*10^-7), 
 
    -(1.812800232760157*10^-7), -(1.866781125899658*10^-7), 
 
    -(1.87345255978832*10^-7), -(2.031866990853662*10^-7), 
 
    -(1.839219440274958*10^-7), -(1.887435724776156*10^-7), 
 
    -(1.769018666006836*10^-7), -(1.782138836493505*10^-7), 
 
    -(1.4705025765886*10^-7), -(1.198129251688356*10^-7), 
 
    -(1.462422974725217*10^-7), -(8.98328826236956*10^-8), 
 
    -(6.698816407747234*10^-8), -(1.557782206589275*10^-7), 
 
    -(1.419817971375109*10^-7), -(8.15462872000382*10^-8), 
 
    -(1.04052861232625*10^-7), -(7.817927026220675*10^-8), 
 
    -(1.030018295622428*10^-7), -(7.626699279429295*10^-8), 
 
    -(8.96354739002012*10^-8), -(7.577181814547398*10^-8), 
 
    -(6.549111129909337*10^-8), -(4.67735023760793*10^-8), 
 
    -(2.500253630048991*10^-8), -(2.424211113556466*10^-8), 
 
    -(1.226050584027717*10^-8), -(1.776657195271909*10^-8), 
 
    -(3.018548931077307*10^-8), -(9.0258304860968*10^-9)}, 
 
   {0, -(5.329262433210081*10^-9), 
 
    -(1.562975007845082*10^-8), -(3.009289406864978*10^-8), 
 
    -(4.152060920127864*10^-8), -(5.222734780282013*10^-8), 
 
    -(6.236302710297601*10^-8), -(7.188607090844207*10^-8), 
 
    -(7.999450311932888*10^-8), -(8.60216064183057*10^-8), 
 
    -(9.04487363247029*10^-8), -(9.43217059459144*10^-8), 
 
    -(9.80512234738715*10^-8), -(1.022001039792671*10^-7), 
 
    -(1.05929035311301*10^-7), -(1.10073965382436*10^-7), 
 
    -(1.172150479092328*10^-7), -(1.3224817136813*10^-7), 
 
    -(1.628682570604392*10^-7), -(2.036337347132781*10^-7), 
 
    -(2.472738689398016*10^-7), -(2.957432665660329*10^-7), 
 
    -(3.749356290806554*10^-7), -(4.897391129118994*10^-7), 
 
    -(6.235482478636625*10^-7), -(6.756936168216697*10^-7), 
 
    -(5.26593857550274*10^-7), -(1.119184614211173*10^-7), 
 
    7.794204395088203*10^-7, 2.341510735565596*10^-6, 
 
    3.727328263394327*10^-6, 5.140787007552835*10^-6, 
 
    5.924596087445636*10^-6, 5.404386738085724*10^-6, 
 
    2.793582412493876*10^-6, 4.344327688558016*10^-7, 
 
    -(5.145991548466783*10^-7), -(8.05533777059624*10^-7), 
 
    -(7.338268151295342*10^-7), -(4.702575901672066*10^-7), 
 
    -(3.125969203614938*10^-7), -(2.523840886339765*10^-7), 
 
    -(2.310919695554586*10^-7), -(2.152140108016305*10^-7), 
 
    -(2.19863164088456*10^-7), -(2.077782483604495*10^-7), 
 
    -(2.084147044445017*10^-7), -(2.033661990880305*10^-7), 
 
    -(2.043287867315299*10^-7), -(2.0039760996309*10^-7), 
 
    -(2.123111050016728*10^-7), -(2.132420713647326*10^-7), 
 
    -(2.107490988112002*10^-7), -(1.992150257210505*10^-7), 
 
    -(2.007401042283277*10^-7), -(1.923865278882284*10^-7), 
 
    -(1.565490367584842*10^-7), -(1.371671409598333*10^-7), 
 
    -(1.568937381162469*10^-7), -(9.85788438184012*10^-8), 
 
    -(6.927772317902452*10^-8), -(1.742051984091466*10^-7), 
 
    -(1.502785327131533*10^-7), -(1.070209056829102*10^-7), 
 
    -(9.8820783628458*10^-8), -(9.84888785588772*10^-8), 
 
    -(1.016700851329016*10^-7), -(8.98274588736113*10^-8), 
 
    -(8.47670825497198*10^-8), -(9.37671563637565*10^-8), 
 
    -(6.226270381676252*10^-8), -(6.137443934975049*10^-8), 
 
    -(4.516614657625845*10^-8), -(2.205009078700323*10^-8), 
 
    -(1.42871014016338*10^-8), -(1.625038436844849*10^-8), 
 
    -(1.435578530057413*10^-8), -(1.571887279193951*10^-8)}\\
 
    , {0, -(4.768057617605578*10^-9), 
 
    -(1.397582965822737*10^-8), -(2.782909603580335*10^-8), 
 
    -(3.998752395987651*10^-8), -(5.208445929919547*10^-8), 
 
    -(6.424699453866526*10^-8), -(7.649160228861795*10^-8), 
 
    -(8.74946765052859*10^-8), -(9.60022004948785*10^-8), 
 
    -(1.021027395814454*10^-7), -(1.070674556118995*10^-7), 
 
    -(1.110593736461831*10^-7), -(1.138669960494905*10^-7), 
 
    -(1.150785373369691*10^-7), -(1.158167448651324*10^-7), 
 
    -(1.166496872448379*10^-7), -(1.189151523455074*10^-7), 
 
    -(1.325816019087651*10^-7), -(1.545187807061602*10^-7), 
 
    -(1.797089140906568*10^-7), -(2.087056792263998*10^-7), 
 
    -(2.58044977153957*10^-7), -(3.434237737446494*10^-7), 
 
    -(4.921502992256743*10^-7), -(6.593348965027155*10^-7), 
 
    -(7.741383942224628*10^-7), -(7.904165322514003*10^-7), 
 
    -(4.937669074215392*10^-7), 4.924533301000129*10^-7, 
 
    1.67662864966244*10^-6, 3.489681427967563*10^-6, 
 
    5.404386738085724*10^-6, 7.267221282565322*10^-6, 
 
    5.902350478694963*10^-6, 2.613909835115851*10^-6, 
 
    5.387761088950079*10^-7, -(5.852234769556138*10^-7), 
 
    -(9.95108665967913*10^-7), -(7.671848405307224*10^-7), 
 
    -(4.970095348983925*10^-7), -(3.361589283643437*10^-7), 
 
    -(2.732476943418819*10^-7), -(2.403650746314169*10^-7), 
 
    -(2.519483936622159*10^-7), -(2.385323458465076*10^-7), 
 
    -(2.477993639223221*10^-7), -(2.345380999900741*10^-7), 
 
    -(2.418324674472296*10^-7), -(2.362263680353529*10^-7), 
 
    -(2.565729341947633*10^-7), -(2.389301794802356*10^-7), 
 
    -(2.550999498608971*10^-7), -(2.241345324003131*10^-7), 
 
    -(2.443924474945811*10^-7), -(2.23503406894951*10^-7), 
 
    -(1.734746024318274*10^-7), -(1.735042496485408*10^-7), 
 
    -(1.78357992414288*10^-7), -(1.174292067833467*10^-7), 
 
    -(7.982435614725672*10^-8), -(2.074038685756721*10^-7), 
 
    -(1.676820945775324*10^-7), -(1.505217123999902*10^-7), 
 
    -(9.44938674054751*10^-8), -(1.343232083588898*10^-7), 
 
    -(1.043619429685759*10^-7), -(1.139959842792162*10^-7), 
 
    -(8.14503795388394*10^-8), -(1.271156050804137*10^-7), 
 
    -(5.921989382156166*10^-8), -(8.68544126556656*10^-8), 
 
    -(6.966577360231772*10^-8), -(1.964226748379224*10^-8), 
 
    -(2.031708480502115*10^-8), -(1.494849496993374*10^-8), 
 
    6.835032293788763*10^-9, -(2.867929458199419*10^-8)}, 
 
   {0, -(6.08510576689322*10^-9), -(1.785621999662*10^-8), 
 
    -(3.388246095598873*10^-8), -(4.58641256757715*10^-8), 
 
    -(5.688704134638078*10^-8), -(6.723141620728132*10^-8), 
 
    -(7.725115940141309*10^-8), -(8.64659741404699*10^-8), 
 
    -(9.42785526118063*10^-8), -(1.009896138763641*10^-7), 
 
    -(1.077343090540948*10^-7), -(1.154654920989795*10^-7), 
 
    -(1.242797238848274*10^-7), -(1.309156272798566*10^-7), 
 
    -(1.364383019968299*10^-7), -(1.433689802341507*10^-7), 
 
    -(1.495022297227705*10^-7), -(1.499150125443925*10^-7), 
 
    -(1.506554898151077*10^-7), -(1.558723935595409*10^-7), 
 
    -(1.654879481445825*10^-7), -(1.888402718221575*10^-7), 
 
    -(2.358414370977043*10^-7), -(3.206327041219605*10^-7), 
 
    -(4.245198374048126*10^-7), -(5.568792936955266*10^-7), 
 
    -(7.164602936220943*10^-7), -(8.16111135869694*10^-7), 
 
    -(5.648605425984672*10^-7), -(4.564478733506934*10^-8), 
 
    1.11673597836931*10^-6, 2.793582412493876*10^-6, 
 
    5.902350478694963*10^-6, 8.42398383108571*10^-6, 
 
    6.253213641371792*10^-6, 3.19984476870899*10^-6, 
 
    7.906286992699206*10^-7, -(8.55835821777291*10^-7), 
 
    -(1.129696070767825*10^-6), -(8.18559671944408*10^-7), 
 
    -(5.640825388770017*10^-7), -(3.590433729879424*10^-7), 
 
    -(2.758013895971882*10^-7), -(2.5565791386464*10^-7), 
 
    -(2.888410590749492*10^-7), -(2.780492198002283*10^-7), 
 
    -(2.960699826225716*10^-7), -(2.729169941895416*10^-7), 
 
    -(2.847856215553393*10^-7), -(2.805002312786409*10^-7), 
 
    -(3.120302335002137*10^-7), -(2.741345765172922*10^-7), 
 
    -(2.900187895062367*10^-7), -(2.643023496300446*10^-7), 
 
    -(2.724283952678612*10^-7), -(2.245519924105496*10^-7), 
 
    -(1.790149210871343*10^-7), -(2.230930935001778*10^-7), 
 
    -(1.360827291316954*10^-7), -(1.029736920123437*10^-7), 
 
    -(2.347617489760268*10^-7), -(2.176849702006146*10^-7), 
 
    -(1.160486142413768*10^-7), -(1.64263233576736*10^-7), 
 
    -(1.129270144058327*10^-7), -(1.608104696404858*10^-7), 
 
    -(1.129476520924256*10^-7), -(1.416413015685238*10^-7), 
 
    -(1.098381121520442*10^-7), -(1.035305030127313*10^-7), 
 
    -(6.633400211752782*10^-8), -(3.136211138545398*10^-8), 
 
    -(3.866505479541319*10^-8), -(1.793501138723367*10^-8), 
 
    -(2.824953160396859*10^-8), -(5.354221083865473*10^-8), 
 
    -(1.088219674392196*10^-8)}, 
 
   {0, -(5.741597800532432*10^-9), 
 
    -(1.684079842108683*10^-8), -(3.279157291815477*10^-8), 
 
    -(4.585162273092417*10^-8), -(5.851220006942933*10^-8), 
 
    -(7.097604744441418*10^-8), -(8.33788200683943*10^-8), 
 
    -(9.45542310742615*10^-8), -(1.033506745604019*10^-7), 
 
    -(1.101323876004882*10^-7), -(1.162018663510193*10^-7), 
 
    -(1.221020295963532*10^-7), -(1.281060514997827*10^-7), 
 
    -(1.32556498455284*10^-7), -(1.365954926720554*10^-7), 
 
    -(1.424058756842643*10^-7), -(1.513547949761419*10^-7), 
 
    -(1.627180394928486*10^-7), -(1.708267455345145*10^-7), 
 
    -(1.750531029093918*10^-7), -(1.777993872195132*10^-7), 
 
    -(1.801021695107042*10^-7), -(1.861934780521144*10^-7), 
 
    -(2.136012228796034*10^-7), -(2.662175550458314*10^-7), 
 
    -(3.510481308217211*10^-7), -(4.734527274482214*10^-7), 
 
    -(6.305395286369582*10^-7), -(7.516768282651515*10^-7), 
 
    -(7.285908061031689*10^-7), -(3.677571841452588*10^-7), 
 
    4.344327688558016*10^-7, 2.613909835115852*10^-6, 
 
    6.253213641371792*10^-6, 8.53696398105452*10^-6, 
 
    7.094460252904648*10^-6, 4.282418229433734*10^-6, 
 
    1.186354387780433*10^-6, -(8.44741511528371*10^-7), 
 
    -(1.085933168665957*10^-6), -(9.00919732757324*10^-7), 
 
    -(6.483689226003665*10^-7), -(3.634883691410912*10^-7), 
 
    -(2.976190774995794*10^-7), -(2.971015313869566*10^-7), 
 
    -(3.287342481485653*10^-7), -(3.190063530618705*10^-7), 
 
    -(3.279393197177699*10^-7), -(3.172137801612066*10^-7), 
 
    -(3.408719286012482*10^-7), -(3.288224527388812*10^-7), 
 
    -(3.386626376709552*10^-7), -(3.084910041022428*10^-7), 
 
    -(3.232462443460155*10^-7), -(3.019570222288664*10^-7), 
 
    -(2.431038139217236*10^-7), -(2.250230571924871*10^-7), 
 
    -(2.445496847429906*10^-7), -(1.580511515106558*10^-7), 
 
    -(1.134130361193122*10^-7), -(2.756814888590369*10^-7), 
 
    -(2.302037146665937*10^-7), -(1.85432172240421*10^-7), 
 
    -(1.423483952307003*10^-7), -(1.689626839194572*10^-7), 
 
    -(1.526134740391119*10^-7), -(1.474434644699313*10^-7), 
 
    -(1.229610106134588*10^-7), -(1.635923047191937*10^-7), 
 
    -(8.79638877120275*10^-8), -(1.08755566044309*10^-7), 
 
    -(6.988759365388593*10^-8), -(3.106559043960485*10^-8), 
 
    -(2.823368713055019*10^-8), -(2.372826813161321*10^-8), 
 
    -(1.057005520755021*10^-8), -(3.47832318175061*10^-8)}, 
 
   {0, -(5.510366782292376*10^-9), 
 
    -(1.616350456380056*10^-8), -(3.137181619076479*10^-8), 
 
    -(4.371927329293081*10^-8), -(5.570701782125278*10^-8), 
 
    -(6.760443231065292*10^-8), -(7.984530395216667*10^-8), 
 
    -(9.14908067571958*10^-8), -(1.013945473438561*10^-7), 
 
    -(1.095586025987162*10^-7), -(1.172178900774858*10^-7), 
 
    -(1.251288677653968*10^-7), -(1.336149068324295*10^-7), 
 
    -(1.399281969109745*10^-7), -(1.453723728490535*10^-7), 
 
    -(1.526315098617256*10^-7), -(1.613052763470062*10^-7), 
 
    -(1.704517234800535*10^-7), -(1.779112298274835*10^-7), 
 
    -(1.837410487234074*10^-7), -(1.885696642810701*10^-7), 
 
    -(1.934950143567468*10^-7), -(1.966784906320862*10^-7), 
 
    -(2.010377634019213*10^-7), -(2.139391050292806*10^-7), 
 
    -(2.51711721828605*10^-7), -(3.201183743671547*10^-7), 
 
    -(4.325818136096268*10^-7), -(5.951586732390603*10^-7), 
 
    -(7.10027087287672*10^-7), -(7.255587741640691*10^-7), 
 
    -(5.145991548466783*10^-7), 5.387761088950076*10^-7, 
 
    3.19984476870899*10^-6, 7.094460252904648*10^-6, 
 
    8.50632012323074*10^-6, 7.280684446249794*10^-6, 
 
    4.191846221186798*10^-6, 5.851086099288978*10^-7, 
 
    -(8.06069833526571*10^-7), -(1.074220810993617*10^-6), 
 
    -(9.35131642662578*10^-7), -(5.633910228064539*10^-7), 
 
    -(3.278294057698167*10^-7), -(3.085423225272858*10^-7), 
 
    -(3.299289468066868*10^-7), -(3.433217652415522*10^-7), 
 
    -(3.423814328950036*10^-7), -(3.436055781589563*10^-7), 
 
    -(3.575166770716569*10^-7), -(3.563607877990407*10^-7), 
 
    -(3.51873988894427*10^-7), -(3.328346393878664*10^-7), 
 
    -(3.386814459185284*10^-7), -(3.255652022543047*10^-7), 
 
    -(2.573083667176109*10^-7), -(2.37514532038957*10^-7), 
 
    -(2.617485858980324*10^-7), -(1.676839632905893*10^-7), 
 
    -(1.212198043277191*10^-7), -(2.918454054373767*10^-7), 
 
    -(2.481913713730511*10^-7), -(1.875959538944976*10^-7), 
 
    -(1.595988028247503*10^-7), -(1.723984378079144*10^-7), 
 
    -(1.678985695442544*10^-7), -(1.534526530438109*10^-7), 
 
    -(1.378045483718074*10^-7), -(1.669371758695892*10^-7), 
 
    -(9.90867178450034*10^-8), -(1.096062852715182*10^-7), 
 
    -(6.841769718365249*10^-8), -(3.540887865396968*10^-8), 
 
    -(2.854276149897663*10^-8), -(2.6767161977013*10^-8), 
 
    -(2.03592668807492*10^-8), -(3.285289543721508*10^-8)}, 
 
   {0, -(6.159463133553294*10^-9), 
 
    -(1.807215188350661*10^-8), -(3.452326910043713*10^-8), 
 
    -(4.714797948744954*10^-8), -(5.896604898355326*10^-8), 
 
    -(7.025616236179168*10^-8), -(8.13557361988265*10^-8), 
 
    -(9.16038805379971*10^-8), -(1.002454338412731*10^-7), 
 
    -(1.076291168578254*10^-7), -(1.149562051089488*10^-7), 
 
    -(1.232295938331867*10^-7), -(1.329794061405688*10^-7), 
 
    -(1.408406385403926*10^-7), -(1.479769488217105*10^-7), 
 
    -(1.579202751115313*10^-7), -(1.704831991584144*10^-7), 
 
    -(1.803697886922686*10^-7), -(1.86321510586079*10^-7), 
 
    -(1.906888168716557*10^-7), -(1.950765064904709*10^-7), 
 
    -(2.022783449973148*10^-7), -(2.104960385855399*10^-7), 
 
    -(2.151328457801576*10^-7), -(2.114746643776313*10^-7), 
 
    -(2.150655356192368*10^-7), -(2.36878097443178*10^-7), 
 
    -(3.0004289947371*10^-7), -(4.276222017059762*10^-7), 
 
    -(5.538165926301249*10^-7), -(7.062724399081562*10^-7), 
 
    -(8.05533777059624*10^-7), -(5.852234769556143*10^-7), 
 
    7.906286992699206*10^-7, 4.282418229433734*10^-6, 
 
    7.280684446249795*10^-6, 8.45433379163366*10^-6, 
 
    7.238035971342117*10^-6, 3.219931021634589*10^-6, 
 
    4.328235772730728*10^-7, -(7.486179004364415*10^-7), 
 
    -(1.098864179390423*10^-6), -(8.5206993184351*10^-7), 
 
    -(4.433088283204016*10^-7), -(3.300595557611081*10^-7), 
 
    -(3.14528674172029*10^-7), -(3.471936986344369*10^-7), 
 
    -(3.437552321110823*10^-7), -(3.537454121357463*10^-7), 
 
    -(3.557524634584527*10^-7), -(3.785891905804048*10^-7), 
 
    -(3.487422086015563*10^-7), -(3.529765375978194*10^-7), 
 
    -(3.351235406851303*10^-7), -(3.346696990918525*10^-7), 
 
    -(2.749782478181868*10^-7), -(2.28658278105994*10^-7), 
 
    -(2.735838381284359*10^-7), -(1.695694623309151*10^-7), 
 
    -(1.24293116854383*10^-7), -(2.947539703344859*10^-7), 
 
    -(2.658256324301662*10^-7), -(1.607593578829562*10^-7), 
 
    -(1.899112570577249*10^-7), -(1.521993419963533*10^-7), 
 
    -(1.890993648658927*10^-7), -(1.461776780365045*10^-7), 
 
    -(1.637115846849438*10^-7), -(1.455223634358154*10^-7), 
 
    -(1.203053036348899*10^-7), -(9.11658428846732*10^-8), 
 
    -(5.696833806405654*10^-8), -(4.378688674977104*10^-8), 
 
    -(2.276750322834264*10^-8), -(3.198430449084763*10^-8), 
 
    -(4.898950351621113*10^-8), -(1.84802082296734*10^-8)}, 
 
   {0, -(5.541770132883985*10^-9), 
 
    -(1.624985278638602*10^-8), -(3.198423329651041*10^-8), 
 
    -(4.532627936610803*10^-8), -(5.846581506776989*10^-8), 
 
    -(7.158490601631251*10^-8), -(8.47874007162073*10^-8), 
 
    -(9.67469319836494*10^-8), -(1.061724985339319*10^-7), 
 
    -(1.13406240407925*10^-7), -(1.198169636890105*10^-7), 
 
    -(1.259456910159141*10^-7), -(1.322212370179874*10^-7), 
 
    -(1.370372840567115*10^-7), -(1.416349701724989*10^-7), 
 
    -(1.486162061378991*10^-7), -(1.606954315009467*10^-7), 
 
    -(1.794996491238403*10^-7), -(1.948640938831003*10^-7), 
 
    -(2.039968522438576*10^-7), -(2.100826473813797*10^-7), 
 
    -(2.139893464000442*10^-7), -(2.144430646090625*10^-7), 
 
    -(2.178578423894468*10^-7), -(2.244111551776715*10^-7), 
 
    -(2.233880550561189*10^-7), -(2.146114443692933*10^-7), 
 
    -(2.125828074375755*10^-7), -(2.775666931897437*10^-7), 
 
    -(3.738429556955107*10^-7), -(5.383187651735975*10^-7), 
 
    -(7.338268151295342*10^-7), -(9.95108665967913*10^-7), 
 
    -(8.55835821777291*10^-7), 1.186354387780433*10^-6, 
 
    4.191846221186798*10^-6, 7.238035971342117*10^-6, 
 
    9.39411836235973*10^-6, 6.847168388878581*10^-6, 
 
    2.772181534972986*10^-6, 3.632082701740904*10^-7, 
 
    -(9.54933438151278*10^-7), -(1.183319514150261*10^-6), 
 
    -(6.962136722014842*10^-7), -(3.66240216608963*10^-7), 
 
    -(3.030652039208277*10^-7), -(3.083876153295639*10^-7), 
 
    -(3.476746675419069*10^-7), -(3.462732551731704*10^-7), 
 
    -(3.73580281620387*10^-7), -(3.536323360278165*10^-7), 
 
    -(3.685459802786802*10^-7), -(3.305283055756033*10^-7), 
 
    -(3.520180557982683*10^-7), -(3.261425940162642*10^-7), 
 
    -(2.601986256825258*10^-7), -(2.458610376166337*10^-7), 
 
    -(2.632336488795999*10^-7), -(1.710434872145207*10^-7), 
 
    -(1.224270930424415*10^-7), -(2.987145834717495*10^-7), 
 
    -(2.462057126465107*10^-7), -(2.066103016740815*10^-7), 
 
    -(1.48434096398204*10^-7), -(1.874724772892076*10^-7), 
 
    -(1.615068132292033*10^-7), -(1.614901482595509*10^-7), 
 
    -(1.281680561107421*10^-7), -(1.822476824451949*10^-7), 
 
    -(9.10117305320339*10^-8), -(1.220166023127589*10^-7), 
 
    -(7.754265249873529*10^-8), -(3.205777675509*10^-8), 
 
    -(3.193785587410307*10^-8), -(2.475652383677672*10^-8), 
 
    -(5.778897437351631*10^-9), -(4.083349764388115*10^-8)}\\
 
    , {0, -(5.545958186597568*10^-9), 
 
    -(1.627011453401327*10^-8), -(3.123172868070502*10^-8), 
 
    -(4.294207451065839*10^-8), -(5.407893839184009*10^-8), 
 
    -(6.494825768023129*10^-8), -(7.611245861333749*10^-8), 
 
    -(8.69418108784552*10^-8), -(9.65039130619925*10^-8), 
 
    -(1.047993941612378*10^-7), -(1.129869974683751*10^-7), 
 
    -(1.220732898744292*10^-7), -(1.324673168331097*10^-7), 
 
    -(1.404959642023065*10^-7), -(1.474440093212508*10^-7), 
 
    -(1.566202070578436*10^-7), -(1.666977091137193*10^-7), 
 
    -(1.748607677072298*10^-7), -(1.822432053211413*10^-7), 
 
    -(1.902731708222357*10^-7), -(1.989189512880837*10^-7), 
 
    -(2.123686736023033*10^-7), -(2.267767567431751*10^-7), 
 
    -(2.336973503489245*10^-7), -(2.289398813264178*10^-7), 
 
    -(2.239571020074633*10^-7), -(2.207209828670368*10^-7), 
 
    -(2.172112585754833*10^-7), -(2.198472308538508*10^-7), 
 
    -(2.423783797234562*10^-7), -(3.255621413553397*10^-7), 
 
    -(4.70257590167206*10^-7), -(7.671848405307219*10^-7), 
 
    -(1.129696070767825*10^-6), -(8.4474151152837*10^-7), 
 
    5.851086099288983*10^-7, 3.219931021634589*10^-6, 
 
    6.847168388878581*10^-6, 9.68037136972661*10^-6, 
 
    6.857561747028293*10^-6, 3.421949538653052*10^-6, 
 
    5.929639172406707*10^-7, -(1.283314060182401*10^-6), 
 
    -(1.135179916419093*10^-6), -(5.636984066270118*10^-7), 
 
    -(2.783770665545455*10^-7), -(2.761290376525689*10^-7), 
 
    -(3.002329766872288*10^-7), -(3.39826738408467*10^-7), 
 
    -(3.458235768184787*10^-7), -(3.613127902438343*10^-7), 
 
    -(3.336211935937511*10^-7), -(3.33738211668709*10^-7), 
 
    -(3.221422843263719*10^-7), -(3.211949174421461*10^-7), 
 
    -(2.563329318171252*10^-7), -(2.237466182833433*10^-7), 
 
    -(2.590056044665424*10^-7), -(1.632177911209536*10^-7), 
 
    -(1.246230034742901*10^-7), -(2.793400084009142*10^-7), 
 
    -(2.467731273075943*10^-7), -(1.619149855105241*10^-7), 
 
    -(1.730373488124538*10^-7), -(1.532989304925298*10^-7), 
 
    -(1.772357378536822*10^-7), -(1.419634173179055*10^-7), 
 
    -(1.498645178365225*10^-7), -(1.522789399978753*10^-7), 
 
    -(1.065464117585222*10^-7), -(9.61525415112255*10^-8), 
 
    -(3.990574871379152*10^-8), -(3.984593305109287*10^-8), 
 
    -(2.834710050197531*10^-8), -(3.01598681869155*10^-8), 
 
    -(4.300817646481617*10^-8), -(2.682431096823051*10^-8)}\\
 
    , {0, -(5.731044583630531*10^-9), 
 
    -(1.681315268674976*10^-8), -(3.22519958357788*10^-8), 
 
    -(4.427685040371824*10^-8), -(5.55871997668166*10^-8), 
 
    -(6.640965427395079*10^-8), -(7.693121048645346*10^-8), 
 
    -(8.64034380512674*10^-8), -(9.40751198562644*10^-8), 
 
    -(1.004112026658456*10^-7), -(1.065584838103868*10^-7), 
 
    -(1.133365393631536*10^-7), -(1.212875317467025*10^-7), 
 
    -(1.278300783104342*10^-7), -(1.339854197668789*10^-7), 
 
    -(1.429313865112477*10^-7), -(1.557087652206244*10^-7), 
 
    -(1.68451593416529*10^-7), -(1.773633281990103*10^-7), 
 
    -(1.83475990353985*10^-7), -(1.891919839439126*10^-7), 
 
    -(1.978704788821143*10^-7), -(2.096476873011435*10^-7), 
 
    -(2.235663981567002*10^-7), -(2.290202023197311*10^-7), 
 
    -(2.261819556111502*10^-7), -(2.1940200027665*10^-7), 
 
    -(2.141129983731665*10^-7), -(2.141689225008935*10^-7), 
 
    -(2.197539637745956*10^-7), -(2.482212097180217*10^-7), 
 
    -(3.125969203614938*10^-7), -(4.970095348983922*10^-7), 
 
    -(8.18559671944408*10^-7), -(1.085933168665957*10^-6), 
 
    -(8.06069833526571*10^-7), 4.328235772730727*10^-7, 
 
    2.772181534972986*10^-6, 6.857561747028293*10^-6, 
 
    8.35928592072641*10^-6, 6.869422002561199*10^-6, 
 
    3.977491071252746*10^-6, 2.72251829707355*10^-7, 
 
    -(1.234839548890774*10^-6), -(8.58745300319882*10^-7), 
 
    -(4.239959193921866*10^-7), -(2.754645543896497*10^-7), 
 
    -(2.694836105383279*10^-7), -(3.036008415683409*10^-7), 
 
    -(3.305880059400153*10^-7), -(3.466470632526143*10^-7), 
 
    -(3.265535678380393*10^-7), -(3.204381358448283*10^-7), 
 
    -(3.103101110287911*10^-7), -(3.053858880760629*10^-7), 
 
    -(2.517517081899538*10^-7), -(2.106954601932016*10^-7), 
 
    -(2.501432215522565*10^-7), -(1.555406195858032*10^-7), 
 
    -(1.14718391278386*10^-7), -(2.707455861160469*10^-7), 
 
    -(2.40227228778491*10^-7), -(1.538752854611234*10^-7), 
 
    -(1.683052497428396*10^-7), -(1.449746689799552*10^-7), 
 
    -(1.70222876449337*10^-7), -(1.362117274877735*10^-7), 
 
    -(1.449263042083163*10^-7), -(1.409298419281286*10^-7), 
 
    -(1.051613964927939*10^-7), -(8.93706317618305*10^-8), 
 
    -(5.039368007329305*10^-8), -(3.860680159860718*10^-8), 
 
    -(2.339380273964129*10^-8), -(2.862171244208666*10^-8), 
 
    -(3.967586031160088*10^-8), -(2.152662025398511*10^-8)}\\
 
    , {0, -(4.524596156830477*10^-9), 
 
    -(1.326557710380296*10^-8), -(2.625447778308279*10^-8), 
 
    -(3.746068226265766*10^-8), -(4.860998119334481*10^-8), 
 
    -(5.987331324963061*10^-8), -(7.144377779305643*10^-8), 
 
    -(8.21970888574653*10^-8), -(9.09558282246434*10^-8), 
 
    -(9.78853160346159*10^-8), -(1.041713933135657*10^-7), 
 
    -(1.103690632140741*10^-7), -(1.167770377471653*10^-7), 
 
    -(1.215105516764576*10^-7), -(1.25704613121943*10^-7), 
 
    -(1.315607394537311*10^-7), -(1.400736918953763*10^-7), 
 
    -(1.541363345449522*10^-7), -(1.669459795196315*10^-7), 
 
    -(1.760238779416832*10^-7), -(1.832555141392253*10^-7), 
 
    -(1.905937832810887*10^-7), -(1.961632784170042*10^-7), 
 
    -(2.031271316395218*10^-7), -(2.145158025142555*10^-7), 
 
    -(2.214701317776424*10^-7), -(2.212159437185898*10^-7), 
 
    -(2.097016065249088*10^-7), -(2.061511377776233*10^-7), 
 
    -(2.131282064988776*10^-7), -(2.285478508470986*10^-7), 
 
    -(2.523840886339761*10^-7), -(3.361589283643434*10^-7), 
 
    -(5.640825388770017*10^-7), -(9.00919732757325*10^-7), 
 
    -(1.074220810993617*10^-6), -(7.486179004364418*10^-7), 
 
    3.632082701740904*10^-7, 3.421949538653052*10^-6, 
 
    6.869422002561199*10^-6, 8.00369390952975*10^-6, 
 
    6.806978316641203*10^-6, 2.958684745973675*10^-6, 
 
    -(3.583258738117883*10^-7), -(9.50028661744531*10^-7), 
 
    -(6.809041970712993*10^-7), -(3.819302196133099*10^-7), 
 
    -(2.793662785213233*10^-7), -(2.787763474200999*10^-7), 
 
    -(3.296240885814097*10^-7), -(3.282105139141089*10^-7), 
 
    -(3.387072100042694*10^-7), -(3.058185753041431*10^-7), 
 
    -(3.224019124628486*10^-7), -(3.010991608751698*10^-7), 
 
    -(2.38033013640931*10^-7), -(2.259707737274805*10^-7), 
 
    -(2.428126942980421*10^-7), -(1.567677084192465*10^-7), 
 
    -(1.104907876309241*10^-7), -(2.754525721341649*10^-7), 
 
    -(2.285417158717119*10^-7), -(1.871711108730949*10^-7), 
 
    -(1.392058048970206*10^-7), -(1.697966650162161*10^-7), 
 
    -(1.497162572939948*10^-7), -(1.476995504865708*10^-7), 
 
    -(1.198393415197345*10^-7), -(1.636580410972445*10^-7), 
 
    -(8.61637651767408*10^-8), -(1.095313155710146*10^-7), 
 
    -(7.53591667024745*10^-8), -(3.014110517652568*10^-8), 
 
    -(2.735136279761661*10^-8), -(2.294212695870294*10^-8), 
 
    -(6.592325154460984*10^-9), -(3.476403160635559*10^-8)}\\
 
    , {0, -(4.805942335178372*10^-9), 
 
    -(1.41009845088546*10^-8), -(2.689651558058566*10^-8), 
 
    -(3.668139013774819*10^-8), -(4.586713765901521*10^-8), 
 
    -(5.472142111747572*10^-8), -(6.372410446174462*10^-8), 
 
    -(7.242530515620681*10^-8), -(8.01170072942146*10^-8), 
 
    -(8.68688749975745*10^-8), -(9.36332061583892*10^-8), 
 
    -(1.013028230064586*10^-7), -(1.102556560506955*10^-7), 
 
    -(1.172910378560302*10^-7), -(1.234544988821824*10^-7), 
 
    -(1.316876970307954*10^-7), -(1.408604482328295*10^-7), 
 
    -(1.470793078889625*10^-7), -(1.521336597472172*10^-7), 
 
    -(1.579860511224798*10^-7), -(1.647275708118605*10^-7), 
 
    -(1.7622991178031*10^-7), -(1.902910972590318*10^-7), 
 
    -(2.002212521680224*10^-7), -(1.995674846503111*10^-7), 
 
    -(2.00237813411715*10^-7), -(2.052298901679141*10^-7), 
 
    -(2.120627948481977*10^-7), -(2.08474110502439*10^-7), 
 
    -(2.039061050514144*10^-7), -(2.105073648786026*10^-7), 
 
    -(2.310919695554586*10^-7), -(2.732476943418819*10^-7), 
 
    -(3.590433729879422*10^-7), -(6.483689226003665*10^-7), 
 
    -(9.35131642662578*10^-7), -(1.098864179390423*10^-6), 
 
    -(9.54933438151278*10^-7), 5.929639172406707*10^-7, 
 
    3.977491071252746*10^-6, 6.806978316641203*10^-6, 
 
    8.43386316161277*10^-6, 6.460064809259363*10^-6, 
 
    1.697817611406231*10^-6, -(6.722842594594036*10^-7), 
 
    -(9.8761048475955*10^-7), -(6.335660426484492*10^-7), 
 
    -(3.369238233080422*10^-7), -(2.706526811196264*10^-7), 
 
    -(3.035308441383807*10^-7), -(3.512796313558372*10^-7), 
 
    -(3.282236752947037*10^-7), -(3.335032607031602*10^-7), 
 
    -(3.138421851949515*10^-7), -(3.156076239430183*10^-7), 
 
    -(2.55822014967849*10^-7), -(2.149582457907682*10^-7), 
 
    -(2.571185855792702*10^-7), -(1.584890024802249*10^-7), 
 
    -(1.138792537741249*10^-7), -(2.779033313439776*10^-7), 
 
    -(2.506459598085512*10^-7), -(1.506256617056217*10^-7), 
 
    -(1.786257861818419*10^-7), -(1.421562174251338*10^-7), 
 
    -(1.770482373840418*10^-7), -(1.372843939174686*10^-7), 
 
    -(1.534841061715084*10^-7), -(1.345694160720605*10^-7), 
 
    -(1.137388976410642*10^-7), -(8.46291922776023*10^-8), 
 
    -(5.959569964781318*10^-8), -(4.108636646349377*10^-8), 
 
    -(1.960040456801111*10^-8), -(2.97483569668397*10^-8), 
 
    -(4.448979728865822*10^-8), -(1.56104380385618*10^-8)}, 
 
   {0, -(4.497018868360721*10^-9), 
 
    -(1.319113248179718*10^-8), -(2.553119063341111*10^-8), 
 
    -(3.544738750595461*10^-8), -(4.496088449360664*10^-8), 
 
    -(5.423841233168889*10^-8), -(6.339288393721216*10^-8), 
 
    -(7.159577749164742*10^-8), -(7.802105978279662*10^-8), 
 
    -(8.30291258780358*10^-8), -(8.7589411694146*10^-8), 
 
    -(9.21763970616989*10^-8), -(9.72233119750451*10^-8), 
 
    -(1.013709114255419*10^-7), -(1.054765651332994*10^-7), 
 
    -(1.118252262956588*10^-7), -(1.223647708415604*10^-7), 
 
    -(1.353383831202362*10^-7), -(1.450927949722587*10^-7), 
 
    -(1.508991772492009*10^-7), -(1.552083072973337*10^-7), 
 
    -(1.59530579859549*10^-7), -(1.652379088477124*10^-7), 
 
    -(1.768986463264551*10^-7), -(1.870140784529806*10^-7), 
 
    -(1.906131820000792*10^-7), -(1.881522159252389*10^-7), 
 
    -(1.85581124955581*10^-7), -(1.945869165007757*10^-7), 
 
    -(2.032826104786607*10^-7), -(2.10420950890121*10^-7), 
 
    -(2.152140108016304*10^-7), -(2.40365074631417*10^-7), 
 
    -(2.758013895971884*10^-7), -(3.634883691410912*10^-7), 
 
    -(5.633910228064539*10^-7), -(8.5206993184351*10^-7), 
 
    -(1.18331951415026*10^-6), -(1.283314060182401*10^-6), 
 
    2.722518297073555*10^-7, 2.958684745973675*10^-6, 
 
    6.460064809259363*10^-6, 0.00001015053265421256, 
 
    6.55709801460879*10^-6, 9.18412761083542*10^-7, 
 
    -(1.19159854293993*10^-6), -(1.105360385937734*10^-6), 
 
    -(5.932441530746927*10^-7), -(2.951613644694832*10^-7), 
 
    -(3.090555063681543*10^-7), -(3.359775683375509*10^-7), 
 
    -(3.658971093048775*10^-7), -(3.404723487767666*10^-7), 
 
    -(3.527620728139943*10^-7), -(3.286284620882613*10^-7), 
 
    -(2.666938410697595*10^-7), -(2.430045366371892*10^-7), 
 
    -(2.678405559895626*10^-7), -(1.717041957029819*10^-7), 
 
    -(1.226880095729792*10^-7), -(3.006542839911442*10^-7), 
 
    -(2.530067087054039*10^-7), -(1.978117457365532*10^-7), 
 
    -(1.590588456587737*10^-7), -(1.807201915324575*10^-7), 
 
    -(1.687753131941376*10^-7), -(1.593418384378631*10^-7), 
 
    -(1.370770086201787*10^-7), -(1.745748410540254*10^-7), 
 
    -(9.86423602425514*10^-8), -(1.156430971447437*10^-7), 
 
    -(7.607217960948125*10^-8), -(3.493581277549876*10^-8), 
 
    -(2.939234693630184*10^-8), -(2.645867346840895*10^-8), 
 
    -(1.490985639528836*10^-8), -(3.540933096234092*10^-8)}\\
 
    , {0, -(4.122473624319882*10^-9), 
 
    -(1.209677201631323*10^-8), -(2.294479685849551*10^-8), 
 
    -(3.106305012213553*10^-8), -(3.857027904450537*10^-8), 
 
    -(4.569310138063785*10^-8), -(5.277969010640398*10^-8), 
 
    -(5.960785406356493*10^-8), -(6.583808669288505*10^-8), 
 
    -(7.150002924549489*10^-8), -(7.733106953051465*10^-8), 
 
    -(8.41627091032478*10^-8), -(9.23556504690663*10^-8), 
 
    -(9.88880280362426*10^-8), -(1.046161313844963*10^-7), 
 
    -(1.122429273796496*10^-7), -(1.205819862792157*10^-7), 
 
    -(1.252518018091098*10^-7), -(1.281301905251002*10^-7), 
 
    -(1.322433586261646*10^-7), -(1.376952758120289*10^-7), 
 
    -(1.480313044417445*10^-7), -(1.612239405956972*10^-7), 
 
    -(1.700883697939415*10^-7), -(1.677374838024994*10^-7), 
 
    -(1.661826722792256*10^-7), -(1.709755779686665*10^-7), 
 
    -(1.808980167769927*10^-7), -(1.827585842139026*10^-7), 
 
    -(1.823740944999893*10^-7), -(1.942833400966807*10^-7), 
 
    -(2.198631640884562*10^-7), -(2.519483936622159*10^-7), 
 
    -(2.5565791386464*10^-7), -(2.976190774995791*10^-7), 
 
    -(3.278294057698166*10^-7), -(4.433088283204013*10^-7), 
 
    -(6.962136722014839*10^-7), -(1.135179916419093*10^-6), 
 
    -(1.234839548890774*10^-6), -(3.583258738117883*10^-7), 
 
    1.697817611406231*10^-6, 6.55709801460879*10^-6, 
 
    0.00001100208901612718, 5.895108105808122*10^-6, 
 
    4.99025072892526*10^-7, -(1.37773562638597*10^-6), 
 
    -(1.19066631950765*10^-6), -(5.830149731792068*10^-7), 
 
    -(3.241099979501162*10^-7), -(3.287951382509045*10^-7), 
 
    -(3.380487651343049*10^-7), -(3.674340325669969*10^-7), 
 
    -(3.565670249480713*10^-7), -(3.50410367113439*10^-7), 
 
    -(2.815466165863407*10^-7), -(2.444797800142924*10^-7), 
 
    -(2.828910000206763*10^-7), -(1.782015539554426*10^-7), 
 
    -(1.291072515023705*10^-7), -(3.085510184465407*10^-7), 
 
    -(2.755314394663086*10^-7), -(1.761196025540955*10^-7), 
 
    -(1.913237385747155*10^-7), -(1.647442797113663*10^-7), 
 
    -(1.919344586910829*10^-7), -(1.55449091738335*10^-7), 
 
    -(1.65549630573171*10^-7), -(1.551674933661553*10^-7), 
 
    -(1.223151688514962*10^-7), -(9.84445719627373*10^-8), 
 
    -(6.900794130214989*10^-8), -(4.359006258914437*10^-8), 
 
    -(2.395127867197946*10^-8), -(3.176422965921186*10^-8), 
 
    -(4.296725269815757*10^-8), -(2.133100926806493*10^-8)}\\
 
    , {0, -(3.440700303342075*10^-9), 
 
    -(1.008808193977432*10^-8), -(1.989567157074632*10^-8), 
 
    -(2.827254429621915*10^-8), -(3.655795494388602*10^-8), 
 
    -(4.488201255097958*10^-8), -(5.338585623811929*10^-8), 
 
    -(6.123611212767348*10^-8), -(6.754915026027155*10^-8), 
 
    -(7.254999477569612*10^-8), -(7.712278334923692*10^-8), 
 
    -(8.17086671481852*10^-8), -(8.65960480178581*10^-8), 
 
    -(9.03363446439897*10^-8), -(9.37429438707803*10^-8), 
 
    -(9.86281247138853*10^-8), -(1.062154087723972*10^-7), 
 
    -(1.176771482816667*10^-7), -(1.274446806971738*10^-7), 
 
    -(1.338246536837517*10^-7), -(1.386846555299099*10^-7), 
 
    -(1.435127762066011*10^-7), -(1.472791320667465*10^-7), 
 
    -(1.534369924934285*10^-7), -(1.625109638506114*10^-7), 
 
    -(1.675372372254515*10^-7), -(1.675418144726422*10^-7), 
 
    -(1.626600473610053*10^-7), -(1.683100230968762*10^-7), 
 
    -(1.795877084359877*10^-7), -(1.93940055314787*10^-7), 
 
    -(2.077782483604497*10^-7), -(2.385323458465076*10^-7), 
 
    -(2.888410590749494*10^-7), -(2.971015313869564*10^-7), 
 
    -(3.08542322527286*10^-7), -(3.300595557611079*10^-7), 
 
    -(3.66240216608963*10^-7), -(5.636984066270116*10^-7), 
 
    -(8.58745300319882*10^-7), -(9.50028661744531*10^-7), 
 
    -(6.722842594594027*10^-7), 9.18412761083542*10^-7, 
 
    5.895108105808121*10^-6, 9.7320832325488*10^-6, 
 
    5.974139106017076*10^-6, 8.148027771594*10^-7, 
 
    -(1.212823629945577*10^-6), -(1.171488768846034*10^-6), 
 
    -(5.927326655425614*10^-7), -(3.401525554160695*10^-7), 
 
    -(3.164227294929703*10^-7), -(3.189276169387334*10^-7), 
 
    -(3.585521541448824*10^-7), -(3.336252319245024*10^-7), 
 
    -(2.676469242325803*10^-7), -(2.422180913011381*10^-7), 
 
    -(2.696629036785921*10^-7), -(1.713904347386594*10^-7), 
 
    -(1.199888125178399*10^-7), -(3.030404431757334*10^-7), 
 
    -(2.557584689261769*10^-7), -(1.967328839126861*10^-7), 
 
    -(1.613594837014835*10^-7), -(1.794552257110856*10^-7), 
 
    -(1.697925432244555*10^-7), -(1.595076342770065*10^-7), 
 
    -(1.384271798938635*10^-7), -(1.719217632659089*10^-7), 
 
    -(1.007418952165636*10^-7), -(1.141433027388483*10^-7), 
 
    -(8.23763886126414*10^-8), -(3.542501110951868*10^-8), 
 
    -(2.725226813051841*10^-8), -(2.649353330584173*10^-8), 
 
    -(1.464547543367879*10^-8), -(3.294569769976221*10^-8)}\\
 
    , {0, -(3.817514814264063*10^-9), 
 
    -(1.12004353131244*10^-8), -(2.134061712174851*10^-8), 
 
    -(2.906404393003048*10^-8), -(3.626887332926176*10^-8), 
 
    -(4.316065473703667*10^-8), -(5.007510602759009*10^-8), 
 
    -(5.670477694258608*10^-8), -(6.258474308487897*10^-8), 
 
    -(6.780569425926304*10^-8), -(7.308873411576603*10^-8), 
 
    -(7.915975687712257*10^-8), -(8.64111596941446*10^-8), 
 
    -(9.22493822030547*10^-8), -(9.74656670757492*10^-8), 
 
    -(1.045641386924158*10^-7), -(1.126658180670136*10^-7), 
 
    -(1.18214603567099*10^-7), -(1.221156326375346*10^-7), 
 
    -(1.263930201789783*10^-7), -(1.314102307087528*10^-7), 
 
    -(1.401488489366913*10^-7), -(1.515046759949109*10^-7), 
 
    -(1.597755221549839*10^-7), -(1.590749934070597*10^-7), 
 
    -(1.585139256882146*10^-7), -(1.615644612892957*10^-7), 
 
    -(1.683644748314891*10^-7), -(1.701520877667771*10^-7), 
 
    -(1.714280458709314*10^-7), -(1.838453139713185*10^-7), 
 
    -(2.084147044445015*10^-7), -(2.477993639223221*10^-7), 
 
    -(2.780492198002285*10^-7), -(3.287342481485652*10^-7), 
 
    -(3.29928946806687*10^-7), -(3.145286741720294*10^-7), 
 
    -(3.030652039208275*10^-7), -(2.783770665545459*10^-7), 
 
    -(4.239959193921866*10^-7), -(6.809041970712988*10^-7), 
 
    -(9.8761048475955*10^-7), -(1.19159854293993*10^-6), 
 
    4.990250728925258*10^-7, 5.974139106017077*10^-6, 
 
    9.78759628611444*10^-6, 6.503833810076314*10^-6, 
 
    1.264516501928496*10^-6, -(1.372874822132618*10^-6), 
 
    -(1.216047923582963*10^-6), -(6.751569658649683*10^-7), 
 
    -(3.571617344275094*10^-7), -(3.109450193784204*10^-7), 
 
    -(3.288991835049242*10^-7), -(3.578677819242431*10^-7), 
 
    -(2.811805756238464*10^-7), -(2.38230429628384*10^-7), 
 
    -(2.817440219729752*10^-7), -(1.750764905944501*10^-7), 
 
    -(1.266333151894219*10^-7), -(3.054021573535815*10^-7), 
 
    -(2.737353285116519*10^-7), -(1.692059059671848*10^-7), 
 
    -(1.933443109506047*10^-7), -(1.593635763911205*10^-7), 
 
    -(1.930691134045152*10^-7), -(1.521233297205363*10^-7), 
 
    -(1.664159900806077*10^-7), -(1.5186017097806*10^-7), 
 
    -(1.225762759788003*10^-7), -(9.58794750151326*10^-8), 
 
    -(6.363597457920511*10^-8), -(4.436880241810064*10^-8), 
 
    -(2.322023190321223*10^-8), -(3.236387772311347*10^-8), 
 
    -(4.666546694616686*10^-8), -(1.976414513131132*10^-8)}\\
 
    , {0, -(3.445821279506183*10^-9), 
 
    -(1.010209405619913*10^-8), -(2.000226509437112*10^-8), 
 
    -(2.855472899384885*10^-8), -(3.704426222954982*10^-8), 
 
    -(4.558959014326992*10^-8), -(5.429871795779309*10^-8), 
 
    -(6.229236734616388*10^-8), -(6.866654511162431*10^-8), 
 
    -(7.359658495356415*10^-8), -(7.797980001902103*10^-8), 
 
    -(8.21752894211263*10^-8), -(8.64197428937982*10^-8), 
 
    -(8.9580388182971*10^-8), -(9.24913665843833*10^-8), 
 
    -(9.67673574233812*10^-8), -(1.038825180746865*10^-7), 
 
    -(1.157549253766071*10^-7), -(1.260961707497637*10^-7), 
 
    -(1.328214784034435*10^-7), -(1.377428551211538*10^-7), 
 
    -(1.419319769519388*10^-7), -(1.445600001734375*10^-7), 
 
    -(1.504038821733001*10^-7), -(1.605339851174843*10^-7), 
 
    -(1.665512953118111*10^-7), -(1.663073619818*10^-7), 
 
    -(1.595260055866423*10^-7), -(1.6514045504713*10^-7), 
 
    -(1.770116236173233*10^-7), -(1.914656288945567*10^-7), 
 
    -(2.033661990880305*10^-7), -(2.345380999900743*10^-7), 
 
    -(2.960699826225716*10^-7), -(3.190063530618703*10^-7), 
 
    -(3.433217652415522*10^-7), -(3.471936986344369*10^-7), 
 
    -(3.083876153295644*10^-7), -(2.761290376525689*10^-7), 
 
    -(2.754645543896501*10^-7), -(3.819302196133099*10^-7), 
 
    -(6.335660426484492*10^-7), -(1.105360385937734*10^-6), 
 
    -(1.37773562638597*10^-6), 8.148027771594*10^-7, 
 
    6.503833810076314*10^-6, 0.0000108423698617454, 
 
    6.784137843696515*10^-6, 7.293590662913592*10^-7, 
 
    -(1.35256570851228*10^-6), -(1.17916825319167*10^-6), 
 
    -(6.745535699223899*10^-7), -(3.672829378391604*10^-7), 
 
    -(3.321969366682213*10^-7), -(3.853891930790079*10^-7), 
 
    -(3.114837641524281*10^-7), -(2.805803898642802*10^-7), 
 
    -(3.073987927459647*10^-7), -(1.980808596133191*10^-7), 
 
    -(1.414413859646342*10^-7), -(3.454928238435786*10^-7), 
 
    -(2.913096325294177*10^-7), -(2.270241104388385*10^-7), 
 
    -(1.834159618190007*10^-7), -(2.074351157150902*10^-7), 
 
    -(1.943342525189166*10^-7), -(1.83066974118394*10^-7), 
 
    -(1.583046567131321*10^-7), -(1.998929546594135*10^-7), 
 
    -(1.140877331659982*10^-7), -(1.322543125875897*10^-7), 
 
    -(8.75333325616547*10^-8), -(4.030083413686513*10^-8), 
 
    -(3.367008182237863*10^-8), -(3.047511649255298*10^-8), 
 
    -(1.780170985504024*10^-8), -(4.031034882536682*10^-8)}\\
 
    , {0, -(4.015669980739583*10^-9), 
 
    -(1.17815819883247*10^-8), -(2.251331019938177*10^-8), 
 
    -(3.076216954913502*10^-8), -(3.847988286330938*10^-8), 
 
    -(4.585255612788349*10^-8), -(5.312407957708435*10^-8), 
 
    -(5.988583328024026*10^-8), -(6.564804683612197*10^-8), 
 
    -(7.058426216983928*10^-8), -(7.548094841244837*10^-8), 
 
    -(8.09968595536203*10^-8), -(8.74320573839941*10^-8), 
 
    -(9.254106387003*10^-8), -(9.70949485453001*10^-8), 
 
    -(1.032987304606699*10^-7), -(1.10450269055848*10^-7), 
 
    -(1.160442932053525*10^-7), -(1.203171167276005*10^-7), 
 
    -(1.246108770767539*10^-7), -(1.29480064873806*10^-7), 
 
    -(1.378726682703844*10^-7), -(1.486845120333575*10^-7), 
 
    -(1.568870201175064*10^-7), -(1.574996507675294*10^-7), 
 
    -(1.573723705255869*10^-7), -(1.600537808515297*10^-7), 
 
    -(1.65879740016197*10^-7), -(1.677174881791239*10^-7), 
 
    -(1.696962633433407*10^-7), -(1.812800232760155*10^-7), 
 
    -(2.043287867315298*10^-7), -(2.418324674472298*10^-7), 
 
    -(2.729169941895418*10^-7), -(3.279393197177699*10^-7), 
 
    -(3.423814328950035*10^-7), -(3.437552321110823*10^-7), 
 
    -(3.476746675419069*10^-7), -(3.002329766872291*10^-7), 
 
    -(2.694836105383279*10^-7), -(2.793662785213233*10^-7), 
 
    -(3.369238233080423*10^-7), -(5.932441530746927*10^-7), 
 
    -(1.19066631950765*10^-6), -(1.212823629945577*10^-6), 
 
    1.264516501928496*10^-6, 6.784137843696517*10^-6, 
 
    0.00001118671585054617, 7.374159600396961*10^-6, 
 
    1.009232969351431*10^-6, -(1.192842942200849*10^-6), 
 
    -(1.211942860422428*10^-6), -(7.07194224841241*10^-7), 
 
    -(3.544274639840212*10^-7), -(4.208907378111605*10^-7), 
 
    -(3.72957967147084*10^-7), -(3.226297432400776*10^-7), 
 
    -(3.743650749767711*10^-7), -(2.313056171252219*10^-7), 
 
    -(1.669017731416445*10^-7), -(4.067745428109814*10^-7), 
 
    -(3.612593559385401*10^-7), -(2.285710981789309*10^-7), 
 
    -(2.530983973615404*10^-7), -(2.148557461927817*10^-7), 
 
    -(2.544504243601786*10^-7), -(2.034228099023183*10^-7), 
 
    -(2.169965003800045*10^-7), -(2.07026683822051*10^-7), 
 
    -(1.58929262720976*10^-7), -(1.317172677577579*10^-7), 
 
    -(8.42540599023639*10^-8), -(5.796323445778447*10^-8), 
 
    -(3.20250632366224*10^-8), -(4.254735241492619*10^-8), 
 
    -(5.783095254401382*10^-8), -(2.908255889588458*10^-8)}\\
 
    , {0, -(4.179134461709377*10^-9), 
 
    -(1.226013620583169*10^-8), -(2.361764323209939*10^-8), 
 
    -(3.259712883077093*10^-8), -(4.115124123210865*10^-8), 
 
    -(4.944658505216823*10^-8), -(5.764168226772385*10^-8), 
 
    -(6.503311540774525*10^-8), -(7.086916429042801*10^-8), 
 
    -(7.536364515001025*10^-8), -(7.937627889707933*10^-8), 
 
    -(8.32681328701335*10^-8), -(8.73062060535484*10^-8), 
 
    -(9.0439069770685*10^-8), -(9.34269668715785*10^-8), 
 
    -(9.79331526239623*10^-8), -(1.055273084059486*10^-7), 
 
    -(1.151496040666844*10^-7), -(1.233144927723071*10^-7), 
 
    -(1.290426327456097*10^-7), -(1.336521837908141*10^-7), 
 
    -(1.385709771486113*10^-7), -(1.436715378948793*10^-7), 
 
    -(1.525964935277751*10^-7), -(1.59766992334653*10^-7), 
 
    -(1.635492556550959*10^-7), -(1.630082284693394*10^-7), 
 
    -(1.599033957925263*10^-7), -(1.65834538094328*10^-7), 
 
    -(1.739583467113018*10^-7), -(1.866781125899656*10^-7), 
 
    -(2.003976099630901*10^-7), -(2.362263680353527*10^-7), 
 
    -(2.847856215553395*10^-7), -(3.172137801612068*10^-7), 
 
    -(3.436055781589569*10^-7), -(3.537454121357463*10^-7), 
 
    -(3.462732551731704*10^-7), -(3.398267384084672*10^-7), 
 
    -(3.03600841568341*10^-7), -(2.787763474200995*10^-7), 
 
    -(2.706526811196255*10^-7), -(2.951613644694829*10^-7), 
 
    -(5.83014973179207*10^-7), -(1.171488768846034*10^-6), 
 
    -(1.372874822132617*10^-6), 7.293590662913592*10^-7, 
 
    7.374159600396961*10^-6, 0.00001343165055458706, 
 
    8.17472540549097*10^-6, 1.412612053420099*10^-6, 
 
    -(1.29891237661823*10^-6), -(1.356417315911301*10^-6), 
 
    -(6.719533860284721*10^-7), -(3.86030649271765*10^-7), 
 
    -(4.199975392970165*10^-7), -(4.218278424443151*10^-7), 
 
    -(4.419326438425014*10^-7), -(2.854866325377563*10^-7), 
 
    -(2.101489743033463*10^-7), -(4.959325018296127*10^-7), 
 
    -(4.14579222808493*10^-7), -(3.360742223639969*10^-7), 
 
    -(2.562696216284815*10^-7), -(3.066345774515069*10^-7), 
 
    -(2.758154388015326*10^-7), -(2.664868539123887*10^-7), 
 
    -(2.225321695394128*10^-7), -(2.97632533074289*10^-7), 
 
    -(1.582189224028596*10^-7), -(1.972427668430987*10^-7), 
 
    -(1.19314137201369*10^-7), -(5.598179994947988*10^-8), 
 
    -(5.333294964368513*10^-8), -(4.305627944026614*10^-8), 
 
    -(2.083425527821124*10^-8), -(6.50019890969244*10^-8)}, 
 
   {0, -(4.60444395547131*10^-9), 
 
    -(1.351226230342543*10^-8), -(2.553684495014268*10^-8), 
 
    -(3.439030087266093*10^-8), -(4.246187341937081*10^-8), 
 
    -(4.996075051824803*10^-8), -(5.712928772027172*10^-8), 
 
    -(6.363891961424249*10^-8), -(6.9090745778837*10^-8), 
 
    -(7.380535549643104*10^-8), -(7.860131549778981*10^-8), 
 
    -(8.42188139423206*10^-8), -(9.09929808672642*10^-8), 
 
    -(9.65228173313441*10^-8), -(1.01550695589836*10^-7), 
 
    -(1.085392649635498*10^-7), -(1.171554154211273*10^-7), 
 
    -(1.22507569544264*10^-7), -(1.257528299147402*10^-7), 
 
    -(1.291439292760649*10^-7), -(1.335047695078337*10^-7), 
 
    -(1.42382707567216*10^-7), -(1.550688116767284*10^-7), 
 
    -(1.662922822599762*10^-7), -(1.658607415281399*10^-7), 
 
    -(1.633401091950126*10^-7), -(1.647546920008863*10^-7), 
 
    -(1.736531479295757*10^-7), -(1.774205936513164*10^-7), 
 
    -(1.776277390443428*10^-7), -(1.87345255978832*10^-7), 
 
    -(2.12311105001673*10^-7), -(2.565729341947636*10^-7), 
 
    -(2.805002312786408*10^-7), -(3.40871928601248*10^-7), 
 
    -(3.575166770716569*10^-7), -(3.55752463458453*10^-7), 
 
    -(3.735802816203869*10^-7), -(3.458235768184786*10^-7), 
 
    -(3.305880059400149*10^-7), -(3.296240885814095*10^-7), 
 
    -(3.035308441383807*10^-7), -(3.090555063681537*10^-7), 
 
    -(3.241099979501162*10^-7), -(5.927326655425614*10^-7), 
 
    -(1.216047923582961*10^-6), -(1.35256570851228*10^-6), 
 
    1.009232969351431*10^-6, 8.17472540549097*10^-6, 
 
    0.00001394035931947262, 8.02875317516178*10^-6, 
 
    1.548873665250658*10^-6, -(1.388109897479967*10^-6), 
 
    -(1.496687773517155*10^-6), -(4.110012218972639*10^-7), 
 
    -(4.022099034443957*10^-7), -(4.010746791025836*10^-7), 
 
    -(4.840392780213371*10^-7), -(2.925824899234903*10^-7), 
 
    -(2.223571832232618*10^-7), -(5.147837890457624*10^-7), 
 
    -(4.614316427923702*10^-7), -(2.872093924626741*10^-7), 
 
    -(3.269317203311501*10^-7), -(2.7160126125203*10^-7), 
 
    -(3.283664472073473*10^-7), -(2.576360810077074*10^-7), 
 
    -(2.822435798328462*10^-7), -(2.623423276121058*10^-7), 
 
    -(2.056365566353527*10^-7), -(1.649808101462358*10^-7), 
 
    -(9.38068489782451*10^-8), -(7.52631392702829*10^-8), 
 
    -(4.332789562695044*10^-8), -(5.551681948736454*10^-8), 
 
    -(8.23351846915165*10^-8), -(3.751142895051797*10^-8)}, 
 
   {0, -(3.409763763944324*10^-9), 
 
    -(9.99304070066875*10^-9), -(2.014959816609805*10^-8), 
 
    -(2.937474603329224*10^-8), -(3.875546588653275*10^-8), 
 
    -(4.839235036614511*10^-8), -(5.835290852765286*10^-8), 
 
    -(6.752593702115775*10^-8), -(7.481439015890433*10^-8), 
 
    -(8.02752947195177*10^-8), -(8.49086777867602*10^-8), 
 
    -(8.89570097584949*10^-8), -(9.2544079040292*10^-8), 
 
    -(9.4904799590102*10^-8), -(9.69855524616892*10^-8), 
 
    -(1.0002010637707*10^-7), -(1.056048125193653*10^-7), 
 
    -(1.188007289960845*10^-7), -(1.314795434186014*10^-7), 
 
    -(1.39960621409864*10^-7), -(1.459415867075075*10^-7), 
 
    -(1.499246167184676*10^-7), -(1.503292105111527*10^-7), 
 
    -(1.538558416984146*10^-7), -(1.668373516365523*10^-7), 
 
    -(1.763430380761483*10^-7), -(1.772866261588243*10^-7), 
 
    -(1.657391629549823*10^-7), -(1.702080754596352*10^-7), 
 
    -(1.854893743436478*10^-7), -(2.031866990853662*10^-7), 
 
    -(2.132420713647324*10^-7), -(2.389301794802356*10^-7), 
 
    -(3.120302335002134*10^-7), -(3.288224527388812*10^-7), 
 
    -(3.563607877990405*10^-7), -(3.785891905804054*10^-7), 
 
    -(3.536323360278165*10^-7), -(3.613127902438347*10^-7), 
 
    -(3.466470632526144*10^-7), -(3.282105139141087*10^-7), 
 
    -(3.512796313558375*10^-7), -(3.359775683375515*10^-7), 
 
    -(3.287951382509046*10^-7), -(3.4015255541607*10^-7), 
 
    -(6.751569658649684*10^-7), -(1.179168253191669*10^-6), 
 
    -(1.192842942200849*10^-6), 1.412612053420099*10^-6, 
 
    8.02875317516178*10^-6, 0.0000123131512394248, 
 
    7.830363396459626*10^-6, 1.443698641090513*10^-6, 
 
    -(2.385994808493703*10^-6), -(9.93286496638023*10^-7), 
 
    -(2.748838201963877*10^-7), -(3.098038226720268*10^-7), 
 
    -(4.099229416599219*10^-7), -(2.619824884845863*10^-7), 
 
    -(1.864142040915248*10^-7), -(4.582523193579013*10^-7), 
 
    -(3.784595060241209*10^-7), -(3.159680317043783*10^-7), 
 
    -(2.285300617742869*10^-7), -(2.864622848500539*10^-7), 
 
    -(2.477450857236335*10^-7), -(2.473271703095153*10^-7), 
 
    -(1.976083225760976*10^-7), -(2.768387411788144*10^-7), 
 
    -(1.411070690118318*10^-7), -(1.852482327574357*10^-7), 
 
    -(1.219387373156528*10^-7), -(4.933694755993802*10^-8), 
 
    -(4.79176083108783*10^-8), -(3.788639913400953*10^-8), 
 
    -(9.30385890243997*10^-9), -(6.105726685923293*10^-8)}, 
 
   {0, -(4.412271307115807*10^-9), 
 
    -(1.294888232233123*10^-8), -(2.436127637146535*10^-8), 
 
    -(3.261950032355887*10^-8), -(4.006641853455942*10^-8), 
 
    -(4.692465412658592*10^-8), -(5.350228742307149*10^-8), 
 
    -(5.962596764763476*10^-8), -(6.500755283104011*10^-8), 
 
    -(6.9877242624643*10^-8), -(7.49699165938889*10^-8), 
 
    -(8.1101302081343*10^-8), -(8.86705369689689*10^-8), 
 
    -(9.49009528118706*10^-8), -(1.005326119768692*10^-7), 
 
    -(1.082699157968299*10^-7), -(1.172790220620745*10^-7), 
 
    -(1.216577438696687*10^-7), -(1.237389071454668*10^-7), 
 
    -(1.267748174260436*10^-7), -(1.312247321542411*10^-7), 
 
    -(1.408089305879604*10^-7), -(1.546906802911192*10^-7), 
 
    -(1.657218096824718*10^-7), -(1.630565978229193*10^-7), 
 
    -(1.595639517001421*10^-7), -(1.619117637939741*10^-7), 
 
    -(1.729978997833808*10^-7), -(1.75898975534816*10^-7), 
 
    -(1.741937753118119*10^-7), -(1.839219440274959*10^-7), 
 
    -(2.107490988112002*10^-7), -(2.550999498608974*10^-7), 
 
    -(2.741345765172924*10^-7), -(3.386626376709551*10^-7), 
 
    -(3.518739888944275*10^-7), -(3.487422086015563*10^-7), 
 
    -(3.685459802786806*10^-7), -(3.336211935937511*10^-7), 
 
    -(3.265535678380394*10^-7), -(3.387072100042693*10^-7), 
 
    -(3.282236752947037*10^-7), -(3.658971093048771*10^-7), 
 
    -(3.380487651343043*10^-7), -(3.1642272949297*10^-7), 
 
    -(3.571617344275095*10^-7), -(6.745535699223904*10^-7), 
 
    -(1.211942860422428*10^-6), -(1.29891237661823*10^-6), 
 
    1.548873665250657*10^-6, 7.830363396459626*10^-6, 
 
    0.0000113237083278236, 6.499897214617477*10^-6, 
 
    -(1.256262436036551*10^-6), -(2.189091830357159*10^-6), 
 
    -(4.302254637642396*10^-7), -(5.907662131812871*10^-8), 
 
    -(3.179467302447796*10^-7), -(2.138050505239164*10^-7), 
 
    -(1.420568714647558*10^-7), -(3.432831852112627*10^-7), 
 
    -(3.166153794997153*10^-7), -(1.718923514448734*10^-7), 
 
    -(2.368724060980897*10^-7), -(1.660546749756647*10^-7), 
 
    -(2.317012583985661*10^-7), -(1.654163649248945*10^-7), 
 
    -(2.047861153742521*10^-7), -(1.588273024079524*10^-7), 
 
    -(1.507185735345442*10^-7), -(9.62626180469516*10^-8), 
 
    -(5.297654038570508*10^-8), -(5.549944287767063*10^-8), 
 
    -(2.514525800170957*10^-8), -(4.031295679489192*10^-8), 
 
    -(7.463864427867362*10^-8), -(1.524208954210165*10^-8)}\\
 
    , {0, -(3.180402340354415*10^-9), 
 
    -(9.32216118987099*10^-9), -(1.869123541195809*10^-8), 
 
    -(2.707911292387617*10^-8), -(3.557699774561537*10^-8), 
 
    -(4.42966760636417*10^-8), -(5.335697574262245*10^-8), 
 
    -(6.1772302664223*10^-8), -(6.851947650853365*10^-8), 
 
    -(7.365444749856314*10^-8), -(7.807363648881694*10^-8), 
 
    -(8.20349518629617*10^-8), -(8.57398700461919*10^-8), 
 
    -(8.83179317115911*10^-8), -(9.06245124016539*10^-8), 
 
    -(9.39573403548668*10^-8), -(9.94531937583404*10^-8), 
 
    -(1.111464110404161*10^-7), -(1.224474344352664*10^-7), 
 
    -(1.3018164425004*10^-7), -(1.357779901264549*10^-7), 
 
    -(1.397990775232083*10^-7), -(1.411655603458695*10^-7), 
 
    -(1.447298429018368*10^-7), -(1.557087135956658*10^-7), 
 
    -(1.641479485069337*10^-7), -(1.652049088495244*10^-7), 
 
    -(1.555389028154145*10^-7), -(1.58973234827315*10^-7), 
 
    -(1.7228431945926*10^-7), -(1.887435724776158*10^-7), 
 
    -(1.992150257210506*10^-7), -(2.241345324003127*10^-7), 
 
    -(2.900187895062367*10^-7), -(3.08491004102243*10^-7), 
 
    -(3.328346393878664*10^-7), -(3.529765375978191*10^-7), 
 
    -(3.305283055756036*10^-7), -(3.337382116687089*10^-7), 
 
    -(3.204381358448281*10^-7), -(3.058185753041432*10^-7), 
 
    -(3.335032607031602*10^-7), -(3.404723487767664*10^-7), 
 
    -(3.674340325669967*10^-7), -(3.189276169387334*10^-7), 
 
    -(3.1094501937842*10^-7), -(3.672829378391604*10^-7), 
 
    -(7.071942248412418*10^-7), -(1.356417315911301*10^-6), 
 
    -(1.388109897479967*10^-6), 1.443698641090513*10^-6, 
 
    6.499897214617478*10^-6, 0.00001023721693780772, 
 
    4.079322261894493*10^-6, -(2.801800168229852*10^-6), 
 
    -(9.21556015063827*10^-7), -(1.036892612315617*10^-7), 
 
    -(1.737638424609187*10^-7), -(2.099132665364467*10^-7), 
 
    -(7.619303160525912*10^-8), -(2.797063225354428*10^-7), 
 
    -(2.259651981491158*10^-7), -(1.892019941833665*10^-7), 
 
    -(1.335562161401487*10^-7), -(1.685136758299943*10^-7), 
 
    -(1.41954252421184*10^-7), -(1.474544091805032*10^-7), 
 
    -(1.140338950575277*10^-7), -(1.554398757112701*10^-7), 
 
    -(8.58684814582817*10^-8), -(1.061010036293638*10^-7), 
 
    -(1.023807640695364*10^-7), -(2.818521203770309*10^-8), 
 
    -(2.075898418252149*10^-8), -(2.048405306022348*10^-8), 
 
    3.980955528559554*10^-9, -(2.923427057269897*10^-8)}, 
 
   {0, -(4.409814491222999*10^-9), 
 
    -(1.294268558277809*10^-8), -(2.430948260810499*10^-8), 
 
    -(3.247794474316606*10^-8), -(3.983342255867244*10^-8), 
 
    -(4.659883128536369*10^-8), -(5.30737422339299*10^-8), 
 
    -(5.905681070744801*10^-8), -(6.423347464041943*10^-8), 
 
    -(6.880821528517459*10^-8), -(7.350309129174442*10^-8), 
 
    -(7.90345317293712*10^-8), -(8.57325169177284*10^-8), 
 
    -(9.12154490106252*10^-8), -(9.62147602652539*10^-8), 
 
    -(1.031925473655629*10^-7), -(1.119974656368132*10^-7), 
 
    -(1.165822946528356*10^-7), -(1.189323330751123*10^-7), 
 
    -(1.219397876820381*10^-7), -(1.260701783651501*10^-7), 
 
    -(1.346132127394328*10^-7), -(1.469206674046437*10^-7), 
 
    -(1.583960671694182*10^-7), -(1.567311566572466*10^-7), 
 
    -(1.538119805103844*10^-7), -(1.553664694759286*10^-7), 
 
    -(1.647957369971705*10^-7), -(1.689037245277025*10^-7), 
 
    -(1.677612590308941*10^-7), -(1.769018666006837*10^-7), 
 
    -(2.007401042283277*10^-7), -(2.443924474945806*10^-7), 
 
    -(2.643023496300443*10^-7), -(3.232462443460152*10^-7), 
 
    -(3.386814459185286*10^-7), -(3.351235406851297*10^-7), 
 
    -(3.520180557982683*10^-7), -(3.221422843263717*10^-7), 
 
    -(3.103101110287914*10^-7), -(3.224019124628486*10^-7), 
 
    -(3.138421851949513*10^-7), -(3.527620728139945*10^-7), 
 
    -(3.565670249480713*10^-7), -(3.585521541448824*10^-7), 
 
    -(3.288991835049242*10^-7), -(3.321969366682215*10^-7), 
 
    -(3.544274639840214*10^-7), -(6.719533860284726*10^-7), 
 
    -(1.496687773517155*10^-6), -(2.385994808493703*10^-6), 
 
    -(1.256262436036551*10^-6), 4.079322261894493*10^-6, 
 
    0.00001529612295724869, 2.971486114417432*10^-6, 
 
    -(3.197505231347628*10^-6), -(2.880906811590446*10^-7), 
 
    -(2.657999996558022*10^-7), -(1.827819317241432*10^-7), 
 
    -(1.763305321819113*10^-7), -(3.240137900223422*10^-7), 
 
    -(3.005693601055025*10^-7), -(1.834527474716913*10^-7), 
 
    -(2.128130422851579*10^-7), -(1.770776702913468*10^-7), 
 
    -(2.164200125529496*10^-7), -(1.666340849882515*10^-7), 
 
    -(1.82319229565955*10^-7), -(1.80149301681049*10^-7), 
 
    -(1.287276288225251*10^-7), -(1.130806058254524*10^-7), 
 
    -(3.853526063011092*10^-8), -(4.938871065134931*10^-8), 
 
    -(3.358263846809037*10^-8), -(3.748646341579199*10^-8), 
 
    -(5.772897067739004*10^-8), -(3.06246128909832*10^-8)}, 
 
   {0, -(3.586416064873128*10^-9), 
 
    -(1.051682808035197*10^-8), -(2.063652594479058*10^-8), 
 
    -(2.913540311351184*10^-8), -(3.745823375528623*10^-8), 
 
    -(4.572154013729808*10^-8), -(5.399543016442055*10^-8), 
 
    -(6.144555449445771*10^-8), -(6.724690573693183*10^-8), 
 
    -(7.172602177404146*10^-8), -(7.579313937525498*10^-8), 
 
    -(7.986510696621895*10^-8), -(8.4095181174527*10^-8), 
 
    -(8.7245276968387*10^-8), -(9.00752325973974*10^-8), 
 
    -(9.40904882098242*10^-8), -(1.000872370247202*10^-7), 
 
    -(1.095581989369645*10^-7), -(1.178934155224384*10^-7), 
 
    -(1.233341857338251*10^-7), -(1.276469303937555*10^-7), 
 
    -(1.326573323521789*10^-7), -(1.373927029573086*10^-7), 
 
    -(1.436145981912622*10^-7), -(1.517748633304543*10^-7), 
 
    -(1.556736612270081*10^-7), -(1.557522510062852*10^-7), 
 
    -(1.528719926773609*10^-7), -(1.572908061848105*10^-7), 
 
    -(1.66509115320318*10^-7), -(1.782138836493506*10^-7), 
 
    -(1.923865278882287*10^-7), -(2.235034068949511*10^-7), 
 
    -(2.724283952678614*10^-7), -(3.01957022228867*10^-7), 
 
    -(3.25565202254304*10^-7), -(3.346696990918535*10^-7), 
 
    -(3.261425940162644*10^-7), -(3.211949174421461*10^-7), 
 
    -(3.053858880760631*10^-7), -(3.010991608751698*10^-7), 
 
    -(3.156076239430185*10^-7), -(3.286284620882613*10^-7), 
 
    -(3.504103671134389*10^-7), -(3.336252319245024*10^-7), 
 
    -(3.578677819242429*10^-7), -(3.853891930790079*10^-7), 
 
    -(4.208907378111606*10^-7), -(3.860306492717646*10^-7), 
 
    -(4.110012218972622*10^-7), -(9.93286496638023*10^-7), 
 
    -(2.189091830357161*10^-6), -(2.801800168229851*10^-6), 
 
    2.971486114417432*10^-6, 0.00002906186587202033, 
 
    4.837910194074058*10^-6, -(7.234346857571281*10^-6), 
 
    -(6.598624296933134*10^-7), -(3.975979329380816*10^-7), 
 
    -(4.060465291063421*10^-7), -(9.24596486279641*10^-7), 
 
    -(7.711118333541918*10^-7), -(5.825147835476883*10^-7), 
 
    -(5.093178299723873*10^-7), -(5.319882998164948*10^-7), 
 
    -(5.323128399512132*10^-7), -(4.769463060478277*10^-7), 
 
    -(4.427015127470641*10^-7), -(5.100769305158538*10^-7), 
 
    -(3.203693258856568*10^-7), -(3.326886704699796*10^-7), 
 
    -(2.123995533187535*10^-7), -(1.139990789662503*10^-7), 
 
    -(8.63385312392247*10^-8), -(8.54986954785368*10^-8), 
 
    -(7.579938056424471*10^-8), -(9.55810047776078*10^-8)}, 
 
   {0, -(2.635146392040281*10^-9), 
 
    -(7.724126632332613*10^-9), -(1.544122791967052*10^-8), 
 
    -(2.229386157543201*10^-8), -(2.91991355895844*10^-8), 
 
    -(3.625417430641495*10^-8), -(4.357466987652651*10^-8), 
 
    -(5.042687416258845*10^-8), -(5.604579785048687*10^-8), 
 
    -(6.032900273416839*10^-8), -(6.397930022070174*10^-8), 
 
    -(6.716274692907797*10^-8), -(6.994581999220712*10^-8), 
 
    -(7.170223350727517*10^-8), -(7.315562586228957*10^-8), 
 
    -(7.511285114831869*10^-8), -(7.795017399290869*10^-8), 
 
    -(8.55710706301347*10^-8), -(9.37925173855458*10^-8), 
 
    -(1.002242362272025*10^-7), -(1.052406003189935*10^-7), 
 
    -(1.091812887179047*10^-7), -(1.105866928732504*10^-7), 
 
    -(1.121113517647252*10^-7), -(1.19405166024966*10^-7), 
 
    -(1.268850336941127*10^-7), -(1.298420177731245*10^-7), 
 
    -(1.226345735325798*10^-7), -(1.233186834339803*10^-7), 
 
    -(1.32698978763037*10^-7), -(1.470502576588601*10^-7), 
 
    -(1.565490367584841*10^-7), -(1.73474602431827*10^-7), 
 
    -(2.245519924105498*10^-7), -(2.431038139217242*10^-7), 
 
    -(2.573083667176113*10^-7), -(2.749782478181867*10^-7), 
 
    -(2.601986256825259*10^-7), -(2.56332931817126*10^-7), 
 
    -(2.517517081899542*10^-7), -(2.380330136409308*10^-7), 
 
    -(2.558220149678495*10^-7), -(2.666938410697597*10^-7), 
 
    -(2.815466165863405*10^-7), -(2.676469242325808*10^-7), 
 
    -(2.811805756238463*10^-7), -(3.114837641524279*10^-7), 
 
    -(3.72957967147084*10^-7), -(4.199975392970163*10^-7), 
 
    -(4.022099034443954*10^-7), -(2.748838201963876*10^-7), 
 
    -(4.302254637642398*10^-7), -(9.21556015063827*10^-7), 
 
    -(3.197505231347628*10^-6), 4.837910194074056*10^-6, 
 
    0.00003778135926131592, 5.139643150391083*10^-6, 
 
    -(6.472588892470393*10^-6), 2.035533423893113*10^-6, 
 
    -(2.016354777894955*10^-6), -(1.819537520403365*10^-6), 
 
    -(1.893824369251217*10^-6), -(1.210209847714436*10^-6), 
 
    -(1.198262536902132*10^-6), -(1.15003323885716*10^-6), 
 
    -(1.226661965517549*10^-6), -(1.056381904363022*10^-6), 
 
    -(1.017316971563093*10^-6), -(1.121850622709638*10^-6), 
 
    -(7.377955405590216*10^-7), -(7.30073309052515*10^-7), 
 
    -(4.679328133218747*10^-7), -(2.673898991756125*10^-7), 
 
    -(1.803251261915436*10^-7), -(1.992524261003282*10^-7), 
 
    -(1.992965835053939*10^-7), -(1.938641556185336*10^-7)}\\
 
    , {0, -(3.900579197536158*10^-9), 
 
    -(1.145143208042345*10^-8), -(2.117761972057683*10^-8), 
 
    -(2.769113005069087*10^-8), -(3.324733429897271*10^-8), 
 
    -(3.802170393123856*10^-8), -(4.218324040597646*10^-8), 
 
    -(4.569997700047657*10^-8), -(4.848445596835474*10^-8), 
 
    -(5.090045225379605*10^-8), -(5.345941473788213*10^-8), 
 
    -(5.665585117666964*10^-8), -(6.077859085596931*10^-8), 
 
    -(6.437971472998144*10^-8), -(6.783411815293271*10^-8), 
 
    -(7.287826722445807*10^-8), -(7.978524394722082*10^-8), 
 
    -(8.191168516826*10^-8), -(8.2053129160283*10^-8), 
 
    -(8.28932590558156*10^-8), -(8.49017567804884*10^-8), 
 
    -(9.0732216570301*10^-8), -(1.010087152326051*10^-7), 
 
    -(1.121499577093622*10^-7), -(1.098138905394264*10^-7), 
 
    -(1.054956912376995*10^-7), -(1.050154870825633*10^-7), 
 
    -(1.145279592146042*10^-7), -(1.189659412029915*10^-7), 
 
    -(1.160382469070117*10^-7), -(1.19812925168836*10^-7), 
 
    -(1.371671409598337*10^-7), -(1.73504249648541*10^-7), 
 
    -(1.79014921087135*10^-7), -(2.250230571924873*10^-7), 
 
    -(2.375145320389566*10^-7), -(2.286582781059944*10^-7), 
 
    -(2.458610376166328*10^-7), -(2.237466182833438*10^-7), 
 
    -(2.106954601932019*10^-7), -(2.2597077372748*10^-7), 
 
    -(2.149582457907684*10^-7), -(2.430045366371887*10^-7), 
 
    -(2.444797800142929*10^-7), -(2.422180913011381*10^-7), 
 
    -(2.38230429628384*10^-7), -(2.805803898642802*10^-7), 
 
    -(3.226297432400779*10^-7), -(4.218278424443155*10^-7), 
 
    -(4.01074679102584*10^-7), -(3.098038226720267*10^-7), 
 
    -(5.907662131812934*10^-8), -(1.03689261231563*10^-7), 
 
    -(2.880906811590446*10^-7), -(7.234346857571279*10^-6), 
 
    5.139643150391083*10^-6, 0.00005827269922411533, 
 
    5.803979025844014*10^-7, -(9.73823798537009*10^-6), 
 
    3.394821139495004*10^-6, -(3.815722596321645*10^-6), 
 
    -(1.926896882192026*10^-6), -(1.909286232511357*10^-6), 
 
    -(1.562245902470106*10^-6), -(1.647352860116489*10^-6), 
 
    -(1.658749858139103*10^-6), -(1.461313051195665*10^-6), 
 
    -(1.407110183324828*10^-6), -(1.553980823027994*10^-6), 
 
    -(1.0139165131982*10^-6), -(9.99469806137737*10^-7), 
 
    -(5.696336928303078*10^-7), -(3.611768013674121*10^-7), 
 
    -(2.812060482707791*10^-7), -(2.719704055303392*10^-7), 
 
    -(2.948422769080806*10^-7), -(2.936009749634173*10^-7)}\\
 
    , {0, -(2.842259168034555*10^-9), 
 
    -(8.33254828449812*10^-9), -(1.648682160121353*10^-8), 
 
    -(2.351001009423026*10^-8), -(3.044736318257984*10^-8), 
 
    -(3.73860846213647*10^-8), -(4.437903633428819*10^-8), 
 
    -(5.074122107547547*10^-8), -(5.58055331897137*10^-8), 
 
    -(5.968235163844166*10^-8), -(6.310120441993189*10^-8), 
 
    -(6.632814916157999*10^-8), -(6.944276528716597*10^-8), 
 
    -(7.161113912557746*10^-8), -(7.349659629537581*10^-8), 
 
    -(7.612282412156417*10^-8), -(8.0088615307374*10^-8), 
 
    -(8.78707903358197*10^-8), -(9.52368456326145*10^-8), 
 
    -(1.00480592956605*10^-7), -(1.046098784494434*10^-7), 
 
    -(1.086314958427428*10^-7), -(1.112860362541111*10^-7), 
 
    -(1.147137792186392*10^-7), -(1.21728271601345*10^-7), 
 
    -(1.266667648619785*10^-7), -(1.280661684978376*10^-7), 
 
    -(1.237131682749499*10^-7), -(1.263094050690477*10^-7), 
 
    -(1.346702450640328*10^-7), -(1.462422974725212*10^-7), 
 
    -(1.568937381162465*10^-7), -(1.78357992414288*10^-7), 
 
    -(2.230930935001773*10^-7), -(2.445496847429903*10^-7), 
 
    -(2.617485858980327*10^-7), -(2.735838381284359*10^-7), 
 
    -(2.632336488796001*10^-7), -(2.590056044665431*10^-7), 
 
    -(2.501432215522575*10^-7), -(2.428126942980416*10^-7), 
 
    -(2.571185855792708*10^-7), -(2.678405559895624*10^-7), 
 
    -(2.828910000206763*10^-7), -(2.696629036785923*10^-7), 
 
    -(2.817440219729754*10^-7), -(3.073987927459647*10^-7), 
 
    -(3.743650749767712*10^-7), -(4.419326438425014*10^-7), 
 
    -(4.840392780213367*10^-7), -(4.099229416599225*10^-7), 
 
    -(3.1794673024478*10^-7), -(1.737638424609187*10^-7), 
 
    -(2.657999996558014*10^-7), -(6.598624296933126*10^-7), 
 
    -(6.472588892470391*10^-6), 5.803979025844014*10^-7, 
 
    0.00005396351782360809, -0.00004513248391366806, 
 
    0.00001187369922066629, -(1.926472498237998*10^-6), 
 
    -(1.309231227146898*10^-6), -(3.511799098583476*10^-7), 
 
    -(7.69227933047527*10^-7), -(5.612538433498511*10^-7), 
 
    -(6.999447033442799*10^-7), -(5.757775648174333*10^-7), 
 
    -(5.605980776129904*10^-7), -(6.017658857062072*10^-7), 
 
    -(4.137398864116518*10^-7), -(3.949081546872495*10^-7), 
 
    -(3.136118604568383*10^-7), -(1.458454035459329*10^-7), 
 
    -(8.62387713818952*10^-8), -(1.064838810375575*10^-7), 
 
    -(9.38532132221357*10^-8), -(9.44347485631656*10^-8)}, 
 
   {0, -(2.410162726676253*10^-9), 
 
    -(7.072255467948383*10^-9), -(1.3452639078782*10^-8), 
 
    -(1.826519935291641*10^-8), -(2.271418300233905*10^-8), 
 
    -(2.689349204412533*10^-8), -(3.087894113408041*10^-8), 
 
    -(3.438725145529001*10^-8), -(3.712122535905155*10^-8), 
 
    -(3.921575545908448*10^-8), -(4.110556780119148*10^-8), 
 
    -(4.298483115012047*10^-8), -(4.495128329960327*10^-8), 
 
    -(4.646903337258279*10^-8), -(4.789230662371842*10^-8), 
 
    -(4.998657645058639*10^-8), -(5.315872203752057*10^-8), 
 
    -(5.635515749899352*10^-8), -(5.917509089486382*10^-8), 
 
    -(6.154306536241322*10^-8), -(6.379470239933859*10^-8), 
 
    -(6.694936898342638*10^-8), -(7.09050125915392*10^-8), 
 
    -(7.577670325201911*10^-8), -(7.772697583064219*10^-8), 
 
    -(7.870785437219416*10^-8), -(7.912984952173998*10^-8), 
 
    -(7.982074916346811*10^-8), -(8.19079772085162*10^-8), 
 
    -(8.40925372351242*10^-8), -(8.98328826236966*10^-8), 
 
    -(9.85788438184015*10^-8), -(1.174292067833467*10^-7), 
 
    -(1.36082729131695*10^-7), -(1.580511515106559*10^-7), 
 
    -(1.676839632905887*10^-7), -(1.695694623309154*10^-7), 
 
    -(1.710434872145198*10^-7), -(1.632177911209532*10^-7), 
 
    -(1.555406195858028*10^-7), -(1.567677084192466*10^-7), 
 
    -(1.584890024802242*10^-7), -(1.717041957029817*10^-7), 
 
    -(1.782015539554421*10^-7), -(1.713904347386594*10^-7), 
 
    -(1.750764905944486*10^-7), -(1.980808596133185*10^-7), 
 
    -(2.313056171252217*10^-7), -(2.85486632537756*10^-7), 
 
    -(2.925824899234934*10^-7), -(2.619824884845871*10^-7), 
 
    -(2.138050505239159*10^-7), -(2.099132665364454*10^-7), 
 
    -(1.827819317241428*10^-7), -(3.975979329380834*10^-7), 
 
    2.035533423893117*10^-6, -(9.73823798537009*10^-6), 
 
    -0.00004513248391366806, 0.0001412296315726365, 
 
    -0.00007591982679979541, 4.430027331483273*10^-6, 
 
    -(4.650330492716112*10^-6), -(3.607843674136792*10^-6), 
 
    -(2.215498437296669*10^-6), -(2.724902314757305*10^-6), 
 
    -(2.552514896138419*10^-6), -(2.309327900844601*10^-6), 
 
    -(2.181443055775835*10^-6), -(2.457532111328919*10^-6), 
 
    -(1.572329555485928*10^-6), -(1.597239550538876*10^-6), 
 
    -(9.28501031165304*10^-7), -(5.79846257865845*10^-7), 
 
    -(4.054262292241343*10^-7), -(4.356544392580536*10^-7), 
 
    -(4.470108672437969*10^-7), -(4.373436434591693*10^-7)}\\
 
    , {0, -(2.718556092930532*10^-9), 
 
    -(7.977848968822922*10^-9), -(1.530104651389432*10^-8), 
 
    -(2.096527470759996*10^-8), -(2.628143322896026*10^-8), 
 
    -(3.129846826978487*10^-8), -(3.589940479961834*10^-8), 
 
    -(3.953243451273594*10^-8), -(4.177702952336063*10^-8), 
 
    -(4.279640877358073*10^-8), -(4.312527228508821*10^-8), 
 
    -(4.25770730139544*10^-8), -(4.089817821473219*10^-8), 
 
    -(3.913885534487036*10^-8), -(3.757979961180338*10^-8), 
 
    -(3.564195320459031*10^-8), -(3.460970855137788*10^-8), 
 
    -(3.688510998476908*10^-8), -(4.085609583470662*10^-8), 
 
    -(4.384118210661825*10^-8), -(4.591793643958637*10^-8), 
 
    -(4.711361567224416*10^-8), -(4.725860585194952*10^-8), 
 
    -(5.063557697635575*10^-8), -(5.605938510427351*10^-8), 
 
    -(6.026370229678307*10^-8), -(6.031233437389359*10^-8), 
 
    -(5.530510918875742*10^-8), -(5.679044499071765*10^-8), 
 
    -(6.147261414488403*10^-8), -(6.698816407747116*10^-8), 
 
    -(6.927772317902431*10^-8), -(7.982435614725653*10^-8), 
 
    -(1.029736920123443*10^-7), -(1.134130361193117*10^-7), 
 
    -(1.212198043277193*10^-7), -(1.242931168543818*10^-7), 
 
    -(1.224270930424415*10^-7), -(1.246230034742901*10^-7), 
 
    -(1.147183912783876*10^-7), -(1.104907876309244*10^-7), 
 
    -(1.138792537741257*10^-7), -(1.22688009572979*10^-7), 
 
    -(1.291072515023708*10^-7), -(1.19988812517839*10^-7), 
 
    -(1.266333151894223*10^-7), -(1.414413859646342*10^-7), 
 
    -(1.66901773141644*10^-7), -(2.101489743033451*10^-7), 
 
    -(2.223571832232612*10^-7), -(1.864142040915235*10^-7), 
 
    -(1.420568714647559*10^-7), -(7.619303160525912*10^-8), 
 
    -(1.763305321819107*10^-7), -(4.060465291063421*10^-7), 
 
    -(2.016354777894959*10^-6), 3.394821139495*10^-6, 
 
    0.00001187369922066628, -0.00007591982679979541, 
 
    0.0003046077698538288, -0.00007602862271921247, 
 
    -0.0000110241953458989, -0.00001152466227292391, 
 
    -0.00001474616510366143, -0.00001263074546230968, 
 
    -0.00001428238310731983, -0.00001206959879814426, 
 
    -0.00001175508519362088, -0.00001279348116503695, 
 
    -(8.53933708266778*10^-6), -(8.28369365385109*10^-6), 
 
    -(5.4098446240247*10^-6), -(3.019968780388129*10^-6), 
 
    -(2.156822606465896*10^-6), -(2.24829336108642*10^-6), 
 
    -(2.282144156829794*10^-6), -(2.279055263932307*10^-6)}\\
 
    , {0, -(3.222387607620929*10^-9), 
 
    -(9.45250733440602*10^-9), -(1.808838500400495*10^-8), 
 
    -(2.476996022450857*10^-8), -(3.100772302256351*10^-8), 
 
    -(3.695267238746206*10^-8), -(4.281336841925978*10^-8), 
 
    -(4.828879442163203*10^-8), -(5.299085967262231*10^-8), 
 
    -(5.713062633885668*10^-8), -(6.133816632683687*10^-8), 
 
    -(6.622715865269831*10^-8), -(7.21501102377667*10^-8), 
 
    -(7.699047313051737*10^-8), -(8.1376472513059*10^-8), 
 
    -(8.74516352180388*10^-8), -(9.49921119986661*10^-8), 
 
    -(1.010819239277127*10^-7), -(1.050006600146956*10^-7), 
 
    -(1.082855865753964*10^-7), -(1.119005387236542*10^-7), 
 
    -(1.183030950220162*10^-7), -(1.268031653267194*10^-7), 
 
    -(1.341776361149468*10^-7), -(1.354209474086687*10^-7), 
 
    -(1.34767027488466*10^-7), -(1.35990433327847*10^-7), 
 
    -(1.409407456988138*10^-7), -(1.443179898089197*10^-7), 
 
    -(1.467580304159833*10^-7), -(1.557782206589277*10^-7), 
 
    -(1.742051984091469*10^-7), -(2.074038685756723*10^-7), 
 
    -(2.347617489760271*10^-7), -(2.756814888590371*10^-7), 
 
    -(2.918454054373768*10^-7), -(2.947539703344864*10^-7), 
 
    -(2.987145834717495*10^-7), -(2.793400084009141*10^-7), 
 
    -(2.707455861160471*10^-7), -(2.75452572134165*10^-7), 
 
    -(2.779033313439773*10^-7), -(3.006542839911442*10^-7), 
 
    -(3.085510184465413*10^-7), -(3.030404431757341*10^-7), 
 
    -(3.054021573535816*10^-7), -(3.454928238435792*10^-7), 
 
    -(4.067745428109808*10^-7), -(4.959325018296126*10^-7), 
 
    -(5.147837890457632*10^-7), -(4.582523193579018*10^-7), 
 
    -(3.432831852112625*10^-7), -(2.79706322535442*10^-7), 
 
    -(3.240137900223414*10^-7), -(9.24596486279642*10^-7), 
 
    -(1.819537520403365*10^-6), -(3.815722596321645*10^-6), 
 
    -(1.926472498237997*10^-6), 4.430027331483278*10^-6, 
 
    -0.00007602862271921247, 0.00007425216140885265, 
 
    -0.00002207359748530572, -(9.53934210307726*10^-7), 
 
    2.565174632197383*10^-6, 1.920724497480855*10^-7, 
 
    1.501227700193056*10^-6, 8.5948743881765*10^-7, 
 
    1.006351069447263*10^-6, 9.89321780056912*10^-7, 
 
    7.218552426684334*10^-7, 6.352586441593156*10^-7, 
 
    4.730517481469582*10^-7, 2.347296428175014*10^-7, 
 
    1.87769181349065*10^-7, 1.732192446473082*10^-7, 
 
    1.817717369558916*10^-7, 1.890231639065387*10^-7}, 
 
   {0, -(1.901643842673706*10^-9), 
 
    -(5.57021860961255*10^-9), -(1.137849574555958*10^-8), 
 
    -(1.68527585362392*10^-8), -(2.250295834637384*10^-8), 
 
    -(2.841335111487816*10^-8), -(3.473803737641072*10^-8), 
 
    -(4.085602881976388*10^-8), -(4.606380720085217*10^-8), 
 
    -(5.028654737874669*10^-8), -(5.41056816689399*10^-8), 
 
    -(5.779749998096239*10^-8), -(6.160680728285155*10^-8), 
 
    -(6.441833682581315*10^-8), -(6.690216147011565*10^-8), 
 
    -(7.035716740933151*10^-8), -(7.531865032088025*10^-8), 
 
    -(8.44599859080408*10^-8), -(9.27265993474351*10^-8), 
 
    -(9.85284180507386*10^-8), -(1.028342326332273*10^-7), 
 
    -(1.059777231434998*10^-7), -(1.068618211815291*10^-7), 
 
    -(1.084287897601107*10^-7), -(1.157633036869787*10^-7), 
 
    -(1.220091807596255*10^-7), -(1.236191152346351*10^-7), 
 
    -(1.169389647329333*10^-7), -(1.192769027802467*10^-7), 
 
    -(1.289979938041792*10^-7), -(1.419817971375115*10^-7), 
 
    -(1.502785327131535*10^-7), -(1.676820945775326*10^-7), 
 
    -(2.176849702006146*10^-7), -(2.302037146665936*10^-7), 
 
    -(2.481913713730513*10^-7), -(2.65825632430166*10^-7), 
 
    -(2.462057126465112*10^-7), -(2.467731273075943*10^-7), 
 
    -(2.40227228778491*10^-7), -(2.285417158717119*10^-7), 
 
    -(2.506459598085509*10^-7), -(2.53006708705403*10^-7), 
 
    -(2.755314394663083*10^-7), -(2.557584689261778*10^-7), 
 
    -(2.737353285116518*10^-7), -(2.913096325294173*10^-7), 
 
    -(3.612593559385401*10^-7), -(4.145792228084927*10^-7), 
 
    -(4.614316427923702*10^-7), -(3.784595060241211*10^-7), 
 
    -(3.166153794997149*10^-7), -(2.259651981491161*10^-7), 
 
    -(3.005693601055028*10^-7), -(7.711118333541917*10^-7), 
 
    -(1.893824369251215*10^-6), -(1.926896882192025*10^-6), 
 
    -(1.309231227146897*10^-6), -(4.650330492716114*10^-6), 
 
    -0.0000110241953458989, -0.00002207359748530572, 
 
    0.00006238582782595561, -(7.359653383865207*10^-6), 
 
    -(1.730682985394962*10^-6), 3.084143462223514*10^-6, 
 
    -(3.191868955161663*10^-7), 9.56330734167325*10^-7, 
 
    4.700072924037844*10^-7, 7.158907202507148*10^-7, 
 
    3.924245672872407*10^-7, 4.544858230687061*10^-7, 
 
    2.447768858759589*10^-7, 1.533606875872844*10^-7, 
 
    1.041813581183553*10^-7, 1.146663340518964*10^-7, 
 
    1.138370441373517*10^-7, 1.14130385364114*10^-7}, 
 
   {0, -(3.835843469536813*10^-9), 
 
    -(1.127009323601047*10^-8), -(1.998762664218066*10^-8), 
 
    -(2.458968800563986*10^-8), -(2.774904849229843*10^-8), 
 
    -(2.967860293589279*10^-8), -(3.063460375283813*10^-8), 
 
    -(3.119855053313008*10^-8), -(3.184650798234524*10^-8), 
 
    -(3.313210716184135*10^-8), -(3.531363860560948*10^-8), 
 
    -(3.921634634198639*10^-8), -(4.526324193762838*10^-8), 
 
    -(5.087650677470178*10^-8), -(5.617590462533552*10^-8), 
 
    -(6.361117995260547*10^-8), -(7.183054841235265*10^-8), 
 
    -(6.780474375678164*10^-8), -(6.059919350390089*10^-8), 
 
    -(5.703260252552855*10^-8), -(5.709622796673052*10^-8), 
 
    -(6.542886818801841*10^-8), -(8.32250884740312*10^-8), 
 
    -(9.79651641482064*10^-8), -(8.58800615516991*10^-8), 
 
    -(7.256408943604625*10^-8), -(7.21037105032832*10^-8), 
 
    -(9.60047705673302*10^-8), -(9.93038472689414*10^-8), 
 
    -(8.54614547638971*10^-8), -(8.15462872000382*10^-8), 
 
    -(1.070209056829106*10^-7), -(1.505217123999899*10^-7), 
 
    -(1.160486142413767*10^-7), -(1.854321722404213*10^-7), 
 
    -(1.875959538944974*10^-7), -(1.607593578829564*10^-7), 
 
    -(2.066103016740813*10^-7), -(1.61914985510524*10^-7), 
 
    -(1.538752854611231*10^-7), -(1.871711108730949*10^-7), 
 
    -(1.506256617056217*10^-7), -(1.97811745736553*10^-7), 
 
    -(1.761196025540953*10^-7), -(1.967328839126862*10^-7), 
 
    -(1.69205905967185*10^-7), -(2.270241104388384*10^-7), 
 
    -(2.285710981789309*10^-7), -(3.360742223639963*10^-7), 
 
    -(2.872093924626742*10^-7), -(3.159680317043783*10^-7), 
 
    -(1.718923514448739*10^-7), -(1.892019941833673*10^-7), 
 
    -(1.834527474716911*10^-7), -(5.82514783547688*10^-7), 
 
    -(1.210209847714437*10^-6), -(1.909286232511357*10^-6), 
 
    -(3.511799098583468*10^-7), -(3.607843674136796*10^-6), 
 
    -0.00001152466227292391, -(9.53934210307722*10^-7), 
 
    -(7.359653383865207*10^-6), 0.00006671796274250817, 
 
    -(6.187548118306147*10^-6), -(7.257201523654871*10^-6), 
 
    2.652687610772778*10^-6, -(1.97282452181752*10^-6), 
 
    -(3.260491352577135*10^-7), -(1.020981239861505*10^-6), 
 
    -(4.29150565197203*10^-7), -(6.09329662519577*10^-7), 
 
    -(2.346347590460168*10^-7), -(1.790161346584075*10^-7), 
 
    -(1.497173771637991*10^-7), -(1.343273343018287*10^-7), 
 
    -(1.512120183748218*10^-7), -(1.556530121054009*10^-7)}\\
 
    , {0, -(5.638283117007999*10^-10), 
 
    -(1.637985700244649*10^-9), -(4.647012995247066*10^-9), 
 
    -(9.01202807960312*10^-9), -(1.418562362396597*10^-8), 
 
    -(2.013862353537024*10^-8), -(2.680857420865351*10^-8), 
 
    -(3.31378475366298*10^-8), -(3.811616787145708*10^-8), 
 
    -(4.146500535173978*10^-8), -(4.375755210218726*10^-8), 
 
    -(4.474419925591659*10^-8), -(4.430176324550268*10^-8), 
 
    -(4.316617002975503*10^-8), -(4.196018272266101*10^-8), 
 
    -(4.030092995147445*10^-8), -(3.977307496604104*10^-8), 
 
    -(5.069285728384095*10^-8), -(6.31029011708752*10^-8), 
 
    -(7.102972492603382*10^-8), -(7.543286440350211*10^-8), 
 
    -(7.425844638014466*10^-8), -(6.570509688282873*10^-8), 
 
    -(6.029589452983356*10^-8), -(7.512549192186771*10^-8), 
 
    -(8.89638138236662*10^-8), -(9.06476392968571*10^-8), 
 
    -(7.060114663895939*10^-8), -(7.140594472470618*10^-8), 
 
    -(8.89302909984889*10^-8), -(1.040528612326248*10^-7), 
 
    -(9.88207836284579*10^-8), -(9.44938674054753*10^-8), 
 
    -(1.64263233576736*10^-7), -(1.423483952307005*10^-7), 
 
    -(1.595988028247504*10^-7), -(1.899112570577248*10^-7), 
 
    -(1.484340963982038*10^-7), -(1.730373488124537*10^-7), 
 
    -(1.683052497428397*10^-7), -(1.392058048970207*10^-7), 
 
    -(1.78625786181842*10^-7), -(1.590588456587739*10^-7), 
 
    -(1.913237385747156*10^-7), -(1.613594837014832*10^-7), 
 
    -(1.93344310950605*10^-7), -(1.834159618190008*10^-7), 
 
    -(2.530983973615407*10^-7), -(2.562696216284813*10^-7), 
 
    -(3.269317203311499*10^-7), -(2.285300617742866*10^-7), 
 
    -(2.368724060980894*10^-7), -(1.33556216140149*10^-7), 
 
    -(2.12813042285158*10^-7), -(5.093178299723879*10^-7), 
 
    -(1.198262536902133*10^-6), -(1.562245902470106*10^-6), 
 
    -(7.692279330475245*10^-7), -(2.21549843729667*10^-6), 
 
    -0.00001474616510366143, 2.565174632197383*10^-6, 
 
    -(1.730682985394963*10^-6), -(6.187548118306148*10^-6), 
 
    0.00003557172213978838, -(8.17998948576365*10^-6), 
 
    -(1.271978552213111*10^-6), 1.797747471618992*10^-6, 
 
    -(3.502165141750868*10^-7), 5.724643738507438*10^-7, 
 
    8.10779730298904*10^-8, 2.890151712312315*10^-7, 
 
    3.297934383901659*10^-8, 5.698889134666442*10^-8, 
 
    6.1834489048338*10^-8, 4.216794537410616*10^-8, 
 
    5.598315134879192*10^-8, 6.166910977460067*10^-8}, 
 
   {0, -(3.566156206980301*10^-9), 
 
    -(1.047526155762537*10^-8), -(1.886119697805685*10^-8), 
 
    -(2.372953928697433*10^-8), -(2.741547163040479*10^-8), 
 
    -(3.008971017418597*10^-8), -(3.19102435375748*10^-8), 
 
    -(3.318963640400108*10^-8), -(3.420032668053101*10^-8), 
 
    -(3.542071621117827*10^-8), -(3.718212930567628*10^-8), 
 
    -(4.008618176702031*10^-8), -(4.438592149140053*10^-8), 
 
    -(4.831609224185113*10^-8), -(5.203576535875678*10^-8), 
 
    -(5.729251333193252*10^-8), -(6.330475802750233*10^-8), 
 
    -(6.073779589942835*10^-8), -(5.608960703673449*10^-8), 
 
    -(5.398112602277029*10^-8), -(5.447729217904924*10^-8), 
 
    -(6.122665261825653*10^-8), -(7.501039565244904*10^-8), 
 
    -(8.70397305651507*10^-8), -(7.904953198514181*10^-8), 
 
    -(6.966143187378071*10^-8), -(6.933386754921679*10^-8), 
 
    -(8.67366322961196*10^-8), -(8.96115205124621*10^-8), 
 
    -(8.00635959293247*10^-8), -(7.817927026220686*10^-8), 
 
    -(9.84888785588768*10^-8), -(1.343232083588901*10^-7), 
 
    -(1.129270144058329*10^-7), -(1.689626839194563*10^-7), 
 
    -(1.723984378079141*10^-7), -(1.521993419963533*10^-7), 
 
    -(1.874724772892076*10^-7), -(1.532989304925298*10^-7), 
 
    -(1.449746689799553*10^-7), -(1.697966650162162*10^-7), 
 
    -(1.421562174251334*10^-7), -(1.807201915324574*10^-7), 
 
    -(1.647442797113662*10^-7), -(1.794552257110858*10^-7), 
 
    -(1.593635763911201*10^-7), -(2.074351157150905*10^-7), 
 
    -(2.148557461927818*10^-7), -(3.066345774515069*10^-7), 
 
    -(2.7160126125203*10^-7), -(2.864622848500538*10^-7), 
 
    -(1.660546749756647*10^-7), -(1.685136758299943*10^-7), 
 
    -(1.770776702913468*10^-7), -(5.319882998164948*10^-7), 
 
    -(1.15003323885716*10^-6), -(1.647352860116488*10^-6), 
 
    -(5.612538433498511*10^-7), -(2.724902314757304*10^-6), 
 
    -0.00001263074546230969, 1.920724497480854*10^-7, 
 
    3.084143462223514*10^-6, -(7.257201523654871*10^-6), 
 
    -(8.17998948576365*10^-6), 0.00005586128061355873, 
 
    -0.00001283202009715354, -(1.652253879619294*10^-6), 
 
    2.085524737009556*10^-6, -(6.236391991465592*10^-7), 
 
    3.4839047257717*10^-7, -(1.33625883301209*10^-7), 
 
    1.71672428797903*10^-7, 5.185159072235438*10^-8, 
 
    3.477403039441834*10^-8, 4.237977678006178*10^-8, 
 
    4.133215050348558*10^-8, 3.48649709165579*10^-8}, 
 
   {0, -(1.527051099484717*10^-9), 
 
    -(4.469375021229927*10^-9), -(9.61871648696794*10^-9), 
 
    -(1.501898784369629*10^-8), -(2.082408559144386*10^-8), 
 
    -(2.70334034387413*10^-8), -(3.355398303870367*10^-8), 
 
    -(3.944567938445814*10^-8), -(4.386077742130098*10^-8), 
 
    -(4.670822931026565*10^-8), -(4.861820744861383*10^-8), 
 
    -(4.940662197792997*10^-8), -(4.889549255251464*10^-8), 
 
    -(4.779898243329004*10^-8), -(4.667646949641561*10^-8), 
 
    -(4.517658124479019*10^-8), -(4.475316166204062*10^-8), 
 
    -(5.326757455281991*10^-8), -(6.323064677254733*10^-8), 
 
    -(6.973443194844667*10^-8), -(7.359656005869688*10^-8), 
 
    -(7.371008404422616*10^-8), -(6.859876055269966*10^-8), 
 
    -(6.645772537970972*10^-8), -(7.854996708212675*10^-8), 
 
    -(8.91417199692628*10^-8), -(9.03855056322993*10^-8), 
 
    -(7.567102324068115*10^-8), -(7.693217712223558*10^-8), 
 
    -(9.05752512685454*10^-8), -(1.030018295622428*10^-7), 
 
    -(1.016700851329019*10^-7), -(1.043619429685761*10^-7), 
 
    -(1.608104696404856*10^-7), -(1.526134740391121*10^-7), 
 
    -(1.678985695442546*10^-7), -(1.890993648658931*10^-7), 
 
    -(1.61506813229203*10^-7), -(1.772357378536821*10^-7), 
 
    -(1.702228764493369*10^-7), -(1.497162572939948*10^-7), 
 
    -(1.770482373840418*10^-7), -(1.687753131941376*10^-7), 
 
    -(1.919344586910829*10^-7), -(1.697925432244558*10^-7), 
 
    -(1.930691134045157*10^-7), -(1.943342525189159*10^-7), 
 
    -(2.544504243601784*10^-7), -(2.758154388015326*10^-7), 
 
    -(3.283664472073478*10^-7), -(2.477450857236337*10^-7), 
 
    -(2.317012583985661*10^-7), -(1.41954252421184*10^-7), 
 
    -(2.164200125529489*10^-7), -(5.323128399512135*10^-7), 
 
    -(1.226661965517549*10^-6), -(1.658749858139103*10^-6), 
 
    -(6.999447033442796*10^-7), -(2.552514896138416*10^-6), 
 
    -0.00001428238310731983, 1.501227700193058*10^-6, 
 
    -(3.191868955161655*10^-7), 2.652687610772776*10^-6, 
 
    -(1.271978552213111*10^-6), -0.00001283202009715354, 
 
    0.00005074210670853745, -(7.716806306480685*10^-6), 
 
    -(5.628695302057194*10^-6), 3.856725204672204*10^-6, 
 
    -(9.40778940442083*10^-7), 1.420045121481131*10^-6, 
 
    -(2.650024958912167*10^-7), 2.022959187451548*10^-8, 
 
    -(1.621917241104406*10^-8), -(2.303453393348859*10^-9), 
 
    -(8.22792887606147*10^-9), -(4.356984616218519*10^-9)}, 
 
   {0, -(2.340022411927473*10^-9), 
 
    -(6.870099128631959*10^-9), -(1.265298506897009*10^-8), 
 
    -(1.645414201717339*10^-8), -(1.964843760753741*10^-8), 
 
    -(2.235109644522516*10^-8), -(2.469784643182746*10^-8), 
 
    -(2.674464908684281*10^-8), -(2.849609049257644*10^-8), 
 
    -(3.019682020553976*10^-8), -(3.214950926001903*10^-8), 
 
    -(3.478036586163763*10^-8), -(3.831098715896946*10^-8), 
 
    -(4.138139672669886*10^-8), -(4.423428662646844*10^-8), 
 
    -(4.823308555896097*10^-8), -(5.298073640475087*10^-8), 
 
    -(5.368583819214088*10^-8), -(5.28704019846572*10^-8), 
 
    -(5.294076369073837*10^-8), -(5.417879674918643*10^-8), 
 
    -(5.875549267977974*10^-8), -(6.68404999352738*10^-8), 
 
    -(7.387427153383258*10^-8), -(7.082786247639528*10^-8), 
 
    -(6.677334331588228*10^-8), -(6.699408641375717*10^-8), 
 
    -(7.556907470874003*10^-8), -(7.770393332506665*10^-8), 
 
    -(7.441893763625931*10^-8), -(7.626699279429279*10^-8), 
 
    -(8.98274588736113*10^-8), -(1.139959842792162*10^-7), 
 
    -(1.129476520924254*10^-7), -(1.474434644699314*10^-7), 
 
    -(1.534526530438108*10^-7), -(1.46177678036504*10^-7), 
 
    -(1.614901482595508*10^-7), -(1.419634173179052*10^-7), 
 
    -(1.362117274877734*10^-7), -(1.476995504865705*10^-7), 
 
    -(1.372843939174685*10^-7), -(1.593418384378633*10^-7), 
 
    -(1.554490917383351*10^-7), -(1.595076342770064*10^-7), 
 
    -(1.521233297205366*10^-7), -(1.830669741183938*10^-7), 
 
    -(2.034228099023182*10^-7), -(2.664868539123889*10^-7), 
 
    -(2.576360810077074*10^-7), -(2.473271703095158*10^-7), 
 
    -(1.654163649248948*10^-7), -(1.474544091805036*10^-7), 
 
    -(1.666340849882513*10^-7), -(4.769463060478266*10^-7), 
 
    -(1.056381904363022*10^-6), -(1.461313051195665*10^-6), 
 
    -(5.757775648174331*10^-7), -(2.309327900844601*10^-6), 
 
    -0.00001206959879814427, 8.59487438817649*10^-7, 
 
    9.56330734167325*10^-7, -(1.972824521817521*10^-6), 
 
    1.797747471618993*10^-6, -(1.652253879619294*10^-6), 
 
    -(7.716806306480685*10^-6), 0.00003494566848390568, 
 
    4.341101474286297*10^-6, -(7.346527042592549*10^-6), 
 
    3.161677402839951*10^-6, -(3.123307348156089*10^-6), 
 
    1.163289631540898*10^-6, 1.828598605031059*10^-7, 
 
    2.254241582660482*10^-7, 1.836236945144417*10^-7, 
 
    2.12919024629093*10^-7, 1.969398467331119*10^-7}, 
 
   {0, -(6.733568908418411*10^-10), 
 
    -(1.962645466297524*10^-9), -(4.992998768903581*10^-9), 
 
    -(9.00060890613555*10^-9), -(1.365038183590745*10^-8), 
 
    -(1.89247801134943*10^-8), -(2.477110973724082*10^-8), 
 
    -(3.026878775680342*10^-8), -(3.453752247036534*10^-8), 
 
    -(3.733586961165126*10^-8), -(3.91670844382097*10^-8), 
 
    -(3.977560401995723*10^-8), -(3.9043593925496*10^-8), 
 
    -(3.777585683166807*10^-8), -(3.650148479770232*10^-8), 
 
    -(3.479478946874301*10^-8), -(3.416658422628107*10^-8), 
 
    -(4.336260507367557*10^-8), -(5.405653286524606*10^-8), 
 
    -(6.096189214691682*10^-8), -(6.479454197071343*10^-8), 
 
    -(6.36794156595097*10^-8), -(5.624724582699269*10^-8), 
 
    -(5.199974567527735*10^-8), -(6.482689448819161*10^-8), 
 
    -(7.697228869539263*10^-8), -(7.832326523567313*10^-8), 
 
    -(6.072016580905571*10^-8), -(6.147297610910627*10^-8), 
 
    -(7.649395778755263*10^-8), -(8.9635473900201*10^-8), 
 
    -(8.47670825497194*10^-8), -(8.14503795388392*10^-8), 
 
    -(1.416413015685238*10^-7), -(1.229610106134584*10^-7), 
 
    -(1.378045483718075*10^-7), -(1.637115846849435*10^-7), 
 
    -(1.281680561107422*10^-7), -(1.498645178365229*10^-7), 
 
    -(1.449263042083165*10^-7), -(1.198393415197342*10^-7), 
 
    -(1.534841061715085*10^-7), -(1.370770086201788*10^-7), 
 
    -(1.65549630573171*10^-7), -(1.384271798938637*10^-7), 
 
    -(1.664159900806076*10^-7), -(1.58304656713132*10^-7), 
 
    -(2.169965003800043*10^-7), -(2.225321695394131*10^-7), 
 
    -(2.822435798328464*10^-7), -(1.976083225760983*10^-7), 
 
    -(2.047861153742522*10^-7), -(1.140338950575277*10^-7), 
 
    -(1.823192295659549*10^-7), -(4.427015127470645*10^-7), 
 
    -(1.017316971563093*10^-6), -(1.407110183324828*10^-6), 
 
    -(5.605980776129915*10^-7), -(2.181443055775834*10^-6), 
 
    -0.00001175508519362088, 1.006351069447261*10^-6, 
 
    4.700072924037869*10^-7, -(3.260491352577128*10^-7), 
 
    -(3.502165141750871*10^-7), 2.085524737009557*10^-6, 
 
    -(5.628695302057194*10^-6), 4.341101474286297*10^-6, 
 
    0.00004180730005920021, -0.00001552460707104961, 
 
    2.831907578869124*10^-6, -(3.267121542890661*10^-6), 
 
    7.623962295609444*10^-8, 3.205626965321792*10^-7, 
 
    4.303130421917063*10^-8, 2.292304086659369*10^-7, 
 
    1.309287936356734*10^-7, 1.741606555079467*10^-7}, 
 
   {0, -(4.354582870336688*10^-9), 
 
    -(1.278878554022715*10^-8), -(2.339298001559446*10^-8), 
 
    -(3.008074819136029*10^-8), -(3.550287878501502*10^-8), 
 
    -(3.979655791592793*10^-8), -(4.292034026707274*10^-8), 
 
    -(4.490275193433562*10^-8), -(4.590003136759516*10^-8), 
 
    -(4.65307471936085*10^-8), -(4.733476444281086*10^-8), 
 
    -(4.869039492205786*10^-8), -(5.051588559503269*10^-8), 
 
    -(5.213457611723199*10^-8), -(5.370940710920593*10^-8), 
 
    -(5.60242020757102*10^-8), -(5.89156847507719*10^-8), 
 
    -(5.585097929066018*10^-8), -(5.219617056341667*10^-8), 
 
    -(5.057450663912471*10^-8), -(5.111381535942378*10^-8), 
 
    -(5.729539113262885*10^-8), -(6.946575171089999*10^-8), 
 
    -(8.13081906825616*10^-8), -(7.633342986940724*10^-8), 
 
    -(6.876340557095398*10^-8), -(6.818369646230769*10^-8), 
 
    -(8.26028640148243*10^-8), -(8.54446606166701*10^-8), 
 
    -(7.783265059766175*10^-8), -(7.577181814547421*10^-8), 
 
    -(9.3767156363757*10^-8), -(1.271156050804139*10^-7), 
 
    -(1.098381121520436*10^-7), -(1.63592304719194*10^-7), 
 
    -(1.669371758695888*10^-7), -(1.455223634358154*10^-7), 
 
    -(1.822476824451952*10^-7), -(1.522789399978758*10^-7), 
 
    -(1.409298419281296*10^-7), -(1.636580410972445*10^-7), 
 
    -(1.345694160720606*10^-7), -(1.74574841054025*10^-7), 
 
    -(1.551674933661549*10^-7), -(1.719217632659089*10^-7), 
 
    -(1.51860170978059*10^-7), -(1.99892954659414*10^-7), 
 
    -(2.070266838220509*10^-7), -(2.976325330742891*10^-7), 
 
    -(2.623423276121054*10^-7), -(2.768387411788145*10^-7), 
 
    -(1.58827302407953*10^-7), -(1.554398757112696*10^-7), 
 
    -(1.8014930168105*10^-7), -(5.10076930515854*10^-7), 
 
    -(1.121850622709638*10^-6), -(1.553980823027993*10^-6), 
 
    -(6.017658857062063*10^-7), -(2.457532111328921*10^-6), 
 
    -0.00001279348116503694, 9.89321780056912*10^-7, 
 
    7.158907202507155*10^-7, -(1.020981239861507*10^-6), 
 
    5.72464373850748*10^-7, -(6.236391991465613*10^-7), 
 
    3.856725204672202*10^-6, -(7.346527042592545*10^-6), 
 
    -0.00001552460707104961, 0.0001059320524435124, 
 
    -0.00005103761052426239, 0.00005194669026232917, 
 
    -0.00001491723257283342, -(4.222296693233151*10^-6), 
 
    -(3.028821955731338*10^-6), -(3.693567008847999*10^-6), 
 
    -(3.413676634887162*10^-6), -(3.501873612210662*10^-6)}\\
 
    , {0, 1.47687732098807*10^-10, 
 
    4.443346234738293*10^-10, -(1.16855550982144*10^-10), 
 
    -(1.87310542343991*10^-9), -(4.237330766814173*10^-9), 
 
    -(7.202088473703428*10^-9), -(1.082224778719521*10^-8), 
 
    -(1.453819338284773*10^-8), -(1.773668581868143*10^-8), 
 
    -(2.014250121807799*10^-8), -(2.200956600615845*10^-8), 
 
    -(2.325210149102667*10^-8), -(2.394721371767952*10^-8), 
 
    -(2.41402898232027*10^-8), -(2.422582735029224*10^-8), 
 
    -(2.434505055585554*10^-8), -(2.528660268402052*10^-8), 
 
    -(3.271317659495282*10^-8), -(4.049553886412572*10^-8), 
 
    -(4.552346099387476*10^-8), -(4.835723617323523*10^-8), 
 
    -(4.755300265205348*10^-8), -(4.217409725504407*10^-8), 
 
    -(3.831638076968163*10^-8), -(4.680001279443281*10^-8), 
 
    -(5.536108438927778*10^-8), -(5.657532618673731*10^-8), 
 
    -(4.417986569534752*10^-8), -(4.46077292480216*10^-8), 
 
    -(5.543644823328754*10^-8), -(6.549111129909379*10^-8), 
 
    -(6.22627038167626*10^-8), -(5.921989382156177*10^-8), 
 
    -(1.035305030127318*10^-7), -(8.79638877120281*10^-8), 
 
    -(9.90867178450037*10^-8), -(1.203053036348902*10^-7), 
 
    -(9.10117305320333*10^-8), -(1.065464117585209*10^-7), 
 
    -(1.051613964927937*10^-7), -(8.61637651767408*10^-8), 
 
    -(1.137388976410633*10^-7), -(9.86423602425518*10^-8), 
 
    -(1.223151688514958*10^-7), -(1.007418952165642*10^-7), 
 
    -(1.225762759788004*10^-7), -(1.140877331659978*10^-7), 
 
    -(1.589292627209762*10^-7), -(1.582189224028586*10^-7), 
 
    -(2.056365566353526*10^-7), -(1.411070690118307*10^-7), 
 
    -(1.507185735345448*10^-7), -(8.58684814582833*10^-8), 
 
    -(1.287276288225251*10^-7), -(3.203693258856565*10^-7), 
 
    -(7.377955405590218*10^-7), -(1.0139165131982*10^-6), 
 
    -(4.1373988641165*10^-7), -(1.572329555485931*10^-6), 
 
    -(8.53933708266778*10^-6), 7.218552426684291*10^-7, 
 
    3.924245672872417*10^-7, -(4.291505651972009*10^-7), 
 
    8.10779730298941*10^-8, 3.483904725771672*10^-7, 
 
    -(9.40778940442088*10^-7), 3.161677402839951*10^-6, 
 
    2.831907578869125*10^-6, -0.00005103761052426244, 
 
    0.0001600367597882788, -0.0002213504946293396, 
 
    0.0000548792861209399, 5.866750147910382*10^-6, 
 
    6.121657214603022*10^-6, 7.507592533907682*10^-6, 
 
    6.687123769100997*10^-6, 6.905002227212391*10^-6}, 
 
   {0, -(2.892408524733451*10^-9), 
 
    -(8.49643178258169*10^-9), -(1.532851663692248*10^-8), 
 
    -(1.932532737200796*10^-8), -(2.23556781826202*10^-8), 
 
    -(2.452692833169947*10^-8), -(2.585718991438001*10^-8), 
 
    -(2.6537594469035*10^-8), -(2.681322757004517*10^-8), 
 
    -(2.714078746193794*10^-8), -(2.781460972507477*10^-8), 
 
    -(2.921359331163344*10^-8), -(3.136925056443124*10^-8), 
 
    -(3.336474154761442*10^-8), -(3.526472819994223*10^-8), 
 
    -(3.795806633432417*10^-8), -(4.09456417035606*10^-8), 
 
    -(3.785425466068331*10^-8), -(3.372287700176872*10^-8), 
 
    -(3.159964626462485*10^-8), -(3.154155411082943*10^-8), 
 
    -(3.643463960462787*10^-8), -(4.669245289650855*10^-8), 
 
    -(5.579801290951061*10^-8), -(4.998087793139553*10^-8), 
 
    -(4.248668696801322*10^-8), -(4.20100050253294*10^-8), 
 
    -(5.551328579941255*10^-8), -(5.751725949239345*10^-8), 
 
    -(4.979550650344368*10^-8), -(4.677350237607972*10^-8), 
 
    -(6.137443934975045*10^-8), -(8.68544126556653*10^-8), 
 
    -(6.633400211752695*10^-8), -(1.087555660443095*10^-7), 
 
    -(1.096062852715176*10^-7), -(9.11658428846728*10^-8), 
 
    -(1.220166023127581*10^-7), -(9.61525415112251*10^-8), 
 
    -(8.93706317618284*10^-8), -(1.095313155710157*10^-7), 
 
    -(8.46291922776024*10^-8), -(1.156430971447431*10^-7), 
 
    -(9.84445719627375*10^-8), -(1.141433027388506*10^-7), 
 
    -(9.58794750151349*10^-8), -(1.32254312587589*10^-7), 
 
    -(1.317172677577576*10^-7), -(1.972427668430993*10^-7), 
 
    -(1.649808101462352*10^-7), -(1.852482327574364*10^-7), 
 
    -(9.62626180469521*10^-8), -(1.061010036293648*10^-7), 
 
    -(1.130806058254533*10^-7), -(3.326886704699802*10^-7), 
 
    -(7.300733090525134*10^-7), -(9.99469806137736*10^-7), 
 
    -(3.949081546872515*10^-7), -(1.597239550538867*10^-6), 
 
    -(8.28369365385108*10^-6), 6.352586441593163*10^-7, 
 
    4.544858230687044*10^-7, -(6.093296625195756*10^-7), 
 
    2.890151712312341*10^-7, -(1.336258833012125*10^-7), 
 
    1.420045121481131*10^-6, -(3.123307348156083*10^-6), 
 
    -(3.267121542890664*10^-6), 0.00005194669026232918, 
 
    -0.0002213504946293396, 0.0004364889512837173, 
 
    -0.0001982859082216262, 2.039525844438954*10^-6, 
 
    -(9.807852663197*10^-6), -0.00001367345508240122, 
 
    -0.00001025164299156477, -0.0000112354678324738}, 
 
   {0, 1.724679090287908*10^-9, 5.056755452912009*10^-9, 
 
    1.058640662354987*10^-8, 1.596854503947596*10^-8, 
 
    2.158022265707833*10^-8, 2.72621422763818*10^-8, 
 
    3.230261943667751*10^-8, 3.539217983782836*10^-8, 
 
    3.577520631985745*10^-8, 3.358945090832421*10^-8, 
 
    2.948823110577274*10^-8, 2.222522882861046*10^-8, 
 
    1.054242094058806*10^-8, -(2.922573486491069*10^-10), 
 
    -(1.025336942691615*10^-8), -(2.377396828687109*10^-8), 
 
    -(3.749730312331767*10^-8), -(3.655936253941034*10^-8), 
 
    -(2.856949266382973*10^-8), -(2.399602939800843*10^-8), 
 
    -(2.301142415878554*10^-8), -(2.888442966214911*10^-8), 
 
    -(4.28559161039872*10^-8), -(4.853679593053229*10^-8), 
 
    -(3.181012007033274*10^-8), -(1.751836063513947*10^-8), 
 
    -(1.828107539713625*10^-8), -(4.201375420551262*10^-8), 
 
    -(4.309302078947935*10^-8), -(2.904216345239322*10^-8), 
 
    -(2.500253630048955*10^-8), -(4.516614657625912*10^-8), 
 
    -(6.966577360231762*10^-8), -(3.136211138545448*10^-8), 
 
    -(6.988759365388713*10^-8), -(6.841769718365364*10^-8), 
 
    -(5.696833806405584*10^-8), -(7.754265249873577*10^-8), 
 
    -(3.99057487137914*10^-8), -(5.039368007329326*10^-8), 
 
    -(7.535916670247463*10^-8), -(5.959569964781302*10^-8), 
 
    -(7.607217960948088*10^-8), -(6.900794130214884*10^-8), 
 
    -(8.23763886126404*10^-8), -(6.363597457920502*10^-8), 
 
    -(8.75333325616552*10^-8), -(8.42540599023644*10^-8), 
 
    -(1.193141372013698*10^-7), -(9.38068489782444*10^-8), 
 
    -(1.219387373156544*10^-7), -(5.297654038570512*10^-8), 
 
    -(1.02380764069536*10^-7), -(3.853526063011092*10^-8), 
 
    -(2.123995533187546*10^-7), -(4.679328133218742*10^-7), 
 
    -(5.696336928303062*10^-7), -(3.136118604568385*10^-7), 
 
    -(9.28501031165308*10^-7), -(5.409844624024701*10^-6), 
 
    4.730517481469584*10^-7, 2.44776885875956*10^-7, 
 
    -(2.346347590460142*10^-7), 3.297934383901512*10^-8, 
 
    1.71672428797903*10^-7, -(2.650024958912154*10^-7), 
 
    1.163289631540896*10^-6, 7.623962295609698*10^-8, 
 
    -0.00001491723257283342, 0.0000548792861209399, 
 
    -0.0001982859082216262, 0.000912207996326339, 
 
    -0.0002341164021813022, -0.0000809035636635244, 
 
    -0.00006951782845310012, -0.0000923097819374186, 
 
    -0.0000846877464918167}, 
 
   {0, -(1.026432799014765*10^-10), 
 
    -(2.966540387855038*10^-10), 
 
    -(9.99162626950924*10^-10), -(2.121616711416478*10^-9), 
 
    -(3.480810649243496*10^-9), -(5.061913525516593*10^-9), 
 
    -(6.829500873293754*10^-9), -(8.47906649633454*10^-9), 
 
    -(9.73164628017534*10^-9), -(1.050996052645881*10^-8), 
 
    -(1.096734096384595*10^-8), -(1.100109093158381*10^-8), 
 
    -(1.054639841724202*10^-8), -(9.94044143899661*10^-9), 
 
    -(9.33891184168867*10^-9), -(8.51094221899108*10^-9), 
 
    -(7.946244860327817*10^-9), -(1.080020490883265*10^-8), 
 
    -(1.422848712267696*10^-8), -(1.638571595844446*10^-8), 
 
    -(1.751435570884486*10^-8), -(1.694746986259109*10^-8), 
 
    -(1.417658564531028*10^-8), -(1.241188833230183*10^-8), 
 
    -(1.672101604300985*10^-8), -(2.074484806988366*10^-8), 
 
    -(2.116517018005939*10^-8), -(1.51892594856514*10^-8), 
 
    -(1.530790297420693*10^-8), -(2.02764586174748*10^-8), 
 
    -(2.42421111355644*10^-8), -(2.205009078700333*10^-8), 
 
    -(1.964226748379219*10^-8), -(3.866505479541335*10^-8), 
 
    -(3.10655904396049*10^-8), -(3.540887865396968*10^-8), 
 
    -(4.378688674977125*10^-8), -(3.205777675508968*10^-8), 
 
    -(3.984593305109303*10^-8), -(3.860680159860718*10^-8), 
 
    -(3.014110517652552*10^-8), -(4.108636646349403*10^-8), 
 
    -(3.493581277549866*10^-8), -(4.359006258914458*10^-8), 
 
    -(3.54250111095189*10^-8), -(4.436880241810085*10^-8), 
 
    -(4.030083413686513*10^-8), -(5.796323445778445*10^-8), 
 
    -(5.598179994947988*10^-8), -(7.526313927028259*10^-8), 
 
    -(4.933694755993792*10^-8), -(5.549944287767063*10^-8), 
 
    -(2.818521203770309*10^-8), -(4.93887106513491*10^-8), 
 
    -(1.139990789662504*10^-7), -(2.673898991756125*10^-7), 
 
    -(3.611768013674122*10^-7), -(1.458454035459324*10^-7), 
 
    -(5.798462578658442*10^-7), -(3.019968780388131*10^-6), 
 
    2.347296428175007*10^-7, 1.533606875872844*10^-7, 
 
    -(1.79016134658408*10^-7), 5.698889134666368*10^-8, 
 
    5.185159072235438*10^-8, 2.022959187451484*10^-8, 
 
    1.828598605031054*10^-7, 3.205626965321792*10^-7, 
 
    -(4.222296693233151*10^-6), 5.866750147910383*10^-6, 
 
    2.039525844438947*10^-6, -0.0002341164021813022, 
 
    0.0000992608043002526, 0.0000107497443876033, 
 
    0.00001915856422114942, 0.00002689328170671521, 
 
    0.00002283610632132155}, 
 
   {0, -(1.537496082757785*10^-9), 
 
    -(4.515321744180375*10^-9), -(8.36648469557223*10^-9), 
 
    -(1.094005535295264*10^-8), -(1.312827732988998*10^-8), 
 
    -(1.495135983398124*10^-8), -(1.629673646008209*10^-8), 
 
    -(1.702411858838024*10^-8), -(1.710739817790186*10^-8), 
 
    -(1.67496324327322*10^-8), -(1.617413412237687*10^-8), 
 
    -(1.528968178741581*10^-8), -(1.388750614700702*10^-8), 
 
    -(1.262220104932117*10^-8), -(1.150851226155655*10^-8), 
 
    -(1.007161979346836*10^-8), -(8.77113062086041*10^-9), 
 
    -(7.422261065559788*10^-9), -(7.004195252384882*10^-9), 
 
    -(6.899569090893922*10^-9), -(7.007478961270208*10^-9), 
 
    -(7.841320949774469*10^-9), -(9.50677044580219*10^-9), 
 
    -(1.243305542437579*10^-8), -(1.285123459237233*10^-8), 
 
    -(1.23443300542525*10^-8), -(1.193396114734347*10^-8), 
 
    -(1.316868733998223*10^-8), -(1.378954669475412*10^-8), 
 
    -(1.288418012890605*10^-8), -(1.226050584027733*10^-8), 
 
    -(1.428710140163374*10^-8), -(2.03170848050211*10^-8), 
 
    -(1.793501138723399*10^-8), -(2.82336871305504*10^-8), 
 
    -(2.854276149897657*10^-8), -(2.276750322834259*10^-8), 
 
    -(3.193785587410328*10^-8), -(2.834710050197553*10^-8), 
 
    -(2.339380273964177*10^-8), -(2.735136279761642*10^-8), 
 
    -(1.960040456801132*10^-8), -(2.939234693630163*10^-8), 
 
    -(2.395127867197946*10^-8), -(2.725226813051854*10^-8), 
 
    -(2.322023190321254*10^-8), -(3.367008182237863*10^-8), 
 
    -(3.202506323662293*10^-8), -(5.333294964368492*10^-8), 
 
    -(4.332789562695064*10^-8), -(4.79176083108785*10^-8), 
 
    -(2.514525800170904*10^-8), -(2.07589841825217*10^-8), 
 
    -(3.358263846809037*10^-8), -(8.63385312392255*10^-8), 
 
    -(1.803251261915437*10^-7), -(2.812060482707791*10^-7), 
 
    -(8.62387713818947*10^-8), -(4.05426229224135*10^-7), 
 
    -(2.156822606465895*10^-6), 1.87769181349065*10^-7, 
 
    1.041813581183559*10^-7, -(1.497173771638007*10^-7), 
 
    6.183448904833737*10^-8, 3.477403039441866*10^-8, 
 
    -(1.621917241104395*10^-8), 2.254241582660485*10^-7, 
 
    4.303130421917031*10^-8, -(3.028821955731337*10^-6), 
 
    6.121657214603018*10^-6, -(9.807852663197*10^-6), 
 
    -0.0000809035636635244, 0.0000107497443876033, 
 
    0.0000885799184531811, -0.00001797151230180032, 
 
    5.706480613597183*10^-7, 0.00001307717269254168}, 
 
   {0, -(2.687372290758677*10^-10), 
 
    -(7.857774685343343*10^-10), 
 
    -(1.784805600536393*10^-9), -(2.92882286604562*10^-9), 
 
    -(4.19577007785798*10^-9), -(5.574968380384974*10^-9), 
 
    -(7.017722270151932*10^-9), -(8.27961956865766*10^-9), 
 
    -(9.15600229108087*10^-9), -(9.62109265579715*10^-9), 
 
    -(9.81629428214947*10^-9), -(9.64527479014372*10^-9), 
 
    -(9.02045493886959*10^-9), -(8.31581488118938*10^-9), 
 
    -(7.64093855464572*10^-9), -(6.722514278265122*10^-9), 
 
    -(5.989278493797788*10^-9), -(7.811873307862758*10^-9), 
 
    -(1.022740009260848*10^-8), -(1.175364413585271*10^-8), 
 
    -(1.25516521217649*10^-8), -(1.218899574984208*10^-8), 
 
    -(1.029581512191945*10^-8), -(9.3273829889666*10^-9), 
 
    -(1.257361819135333*10^-8), -(1.545923029117855*10^-8), 
 
    -(1.569437422630007*10^-8), -(1.14316559234924*10^-8), 
 
    -(1.156855469525527*10^-8), -(1.507566992583518*10^-8), 
 
    -(1.776657195271897*10^-8), -(1.625038436844874*10^-8), 
 
    -(1.494849496993362*10^-8), -(2.824953160396842*10^-8), 
 
    -(2.372826813161372*10^-8), -(2.676716197701278*10^-8), 
 
    -(3.198430449084761*10^-8), -(2.47565238367764*10^-8), 
 
    -(3.015986818691569*10^-8), -(2.862171244208644*10^-8), 
 
    -(2.294212695870294*10^-8), -(2.974835696683978*10^-8), 
 
    -(2.645867346840905*10^-8), -(3.176422965921215*10^-8), 
 
    -(2.649353330584234*10^-8), -(3.236387772311347*10^-8), 
 
    -(3.047511649255263*10^-8), -(4.254735241492637*10^-8), 
 
    -(4.305627944026592*10^-8), -(5.551681948736411*10^-8), 
 
    -(3.788639913400918*10^-8), -(4.031295679489179*10^-8), 
 
    -(2.048405306022358*10^-8), -(3.748646341579225*10^-8), 
 
    -(8.54986954785372*10^-8), -(1.99252426100328*10^-7), 
 
    -(2.719704055303392*10^-7), -(1.064838810375568*10^-7), 
 
    -(4.356544392580521*10^-7), -(2.248293361086422*10^-6), 
 
    1.732192446473056*10^-7, 1.146663340518968*10^-7, 
 
    -(1.343273343018286*10^-7), 4.216794537410648*10^-8, 
 
    4.237977678006061*10^-8, -(2.30345339334833*10^-9), 
 
    1.836236945144418*10^-7, 2.292304086659376*10^-7, 
 
    -(3.693567008847998*10^-6), 7.507592533907684*10^-6, 
 
    -0.00001367345508240122, -0.00006951782845310012, 
 
    0.00001915856422114941, -0.00001797151230180033, 
 
    0.0001205320869350265, 0.00001398911048187089, 
 
    -0.00003802999192802926}, 
 
   {0, 3.294965543306885*10^-10, 9.81062812481653*10^-10, 
 
    3.531160908439582*10^-10, -(2.161429949914294*10^-9), 
 
    -(5.606384821920282*10^-9), -(9.87018238101096*10^-9), 
 
    -(1.468274331190588*10^-8), -(1.894541386815515*10^-8), 
 
    -(2.173960822757025*10^-8), -(2.275424608720482*10^-8), 
 
    -(2.239542331329452*10^-8), -(1.994776764518305*10^-8), 
 
    -(1.492272974672625*10^-8), -(9.8675164008303*10^-9), 
 
    -(5.109889755525062*10^-9), 1.397072742633869*10^-9, 
 
    7.428961333318802*10^-9, -(2.837648735189283*10^-10), 
 
    -(1.160437360257295*10^-8), -(1.846500075664716*10^-8), 
 
    -(2.129118428460101*10^-8), -(1.659963345755656*10^-8), 
 
    -(2.857882117906034*10^-9), 5.821843744207526*10^-9, 
 
    -(1.03971151374599*10^-8), -(2.606093886815352*10^-8), 
 
    -(2.693966060786517*10^-8), -(2.129248393776292*10^-9), 
 
    -(1.385389313589666*10^-9), -(1.905070579038467*10^-8), 
 
    -(3.018548931077275*10^-8), -(1.435578530057381*10^-8), 
 
    6.835032293788711*10^-9, -(5.354221083865442*10^-8), 
 
    -(1.057005520754968*10^-8), -(2.035926688074888*10^-8), 
 
    -(4.898950351621072*10^-8), -(5.778897437350797*10^-9), 
 
    -(4.300817646481596*10^-8), -(3.967586031160027*10^-8), 
 
    -(6.592325154460559*10^-9), -(4.448979728865822*10^-8), 
 
    -(1.490985639528825*10^-8), -(4.29672526981577*10^-8), 
 
    -(1.464547543367903*10^-8), -(4.666546694616698*10^-8), 
 
    -(1.780170985504005*10^-8), -(5.783095254401425*10^-8), 
 
    -(2.083425527821084*10^-8), -(8.23351846915167*10^-8), 
 
    -(9.30385890243987*10^-9), -(7.46386442786741*10^-8), 
 
    3.98095552855913*10^-9, -(5.772897067739025*10^-8), 
 
    -(7.579938056424396*10^-8), -(1.99296583505394*10^-7), 
 
    -(2.948422769080809*10^-7), -(9.38532132221375*10^-8), 
 
    -(4.470108672437964*10^-7), -(2.282144156829793*10^-6), 
 
    1.817717369558918*10^-7, 1.138370441373519*10^-7, 
 
    -(1.512120183748209*10^-7), 5.598315134879213*10^-8, 
 
    4.133215050348516*10^-8, -(8.22792887606168*10^-9), 
 
    2.129190246290924*10^-7, 1.30928793635672*10^-7, 
 
    -(3.413676634887163*10^-6), 6.687123769100997*10^-6, 
 
    -0.00001025164299156477, -0.0000923097819374186, 
 
    0.0000268932817067152, 5.706480613597164*10^-7, 
 
    0.00001398911048187089, 0.0000661421392646412, 
 
    0.00007782792857994535}, 
 
   {0, -(2.008461437593879*10^-9), 
 
    -(5.901519775951762*10^-9), -(1.058873759403375*10^-8), 
 
    -(1.322595838421122*10^-8), -(1.515569703434513*10^-8), 
 
    -(1.642941894313745*10^-8), -(1.696921609062701*10^-8), 
 
    -(1.686632383117323*10^-8), -(1.631437753402474*10^-8), 
 
    -(1.565225676982833*10^-8), -(1.506692196012985*10^-8), 
 
    -(1.463300786621275*10^-8), -(1.42074879809082*10^-8), 
 
    -(1.394001591972737*10^-8), -(1.377428223971706*10^-8), 
 
    -(1.364382625604024*10^-8), -(1.354348017198819*10^-8), 
 
    -(1.001297507308952*10^-8), -(6.712031721770819*10^-9), 
 
    -(4.88150479276821*10^-9), -(4.328665235493901*10^-9), 
 
    -(6.615002453873266*10^-9), -(1.218815199320192*10^-8), 
 
    -(1.797453208947519*10^-8), -(1.451830450749755*10^-8), 
 
    -(9.94264996531587*10^-9), -(9.2380416150757*10^-9), 
 
    -(1.705044123833377*10^-8), -(1.802592605766947*10^-8), 
 
    -(1.256094109879301*10^-8), -(9.0258304860968*10^-9), 
 
    -(1.571887279193929*10^-8), -(2.867929458199483*10^-8), 
 
    -(1.088219674392282*10^-8), -(3.478323181750631*10^-8), 
 
    -(3.285289543721603*10^-8), -(1.848020822967307*10^-8), 
 
    -(4.083349764388115*10^-8), -(2.68243109682303*10^-8), 
 
    -(2.152662025398511*10^-8), -(3.476403160635516*10^-8), 
 
    -(1.561043803856191*10^-8), -(3.540933096234188*10^-8), 
 
    -(2.133100926806501*10^-8), -(3.294569769976298*10^-8), 
 
    -(1.976414513131164*10^-8), -(4.031034882536664*10^-8), 
 
    -(2.908255889588412*10^-8), -(6.500198909692461*10^-8), 
 
    -(3.751142895051723*10^-8), -(6.105726685923184*10^-8), 
 
    -(1.524208954210096*10^-8), -(2.923427057269897*10^-8), 
 
    -(3.062461289098357*10^-8), -(9.55810047776065*10^-8), 
 
    -(1.938641556185349*10^-7), -(2.936009749634176*10^-7), 
 
    -(9.44347485631664*10^-8), -(4.373436434591705*10^-7), 
 
    -(2.279055263932308*10^-6), 1.890231639065378*10^-7, 
 
    1.141303853641125*10^-7, -(1.556530121054003*10^-7), 
 
    6.166910977460118*10^-8, 3.486497091655748*10^-8, 
 
    -(4.356984616217778*10^-9), 1.969398467331115*10^-7, 
 
    1.741606555079475*10^-7, -(3.501873612210661*10^-6), 
 
    6.905002227212391*10^-6, -0.0000112354678324738, 
 
    -0.0000846877464918167, 0.00002283610632132155, 
 
    0.00001307717269254168, -0.00003802999192802925, 
 
    0.00007782792857994535, 0.0001441901820677384}}, 
 
  {1, 0.999046171962443, 0.997202044189458, 
 
   0.994522034093038, 0.992275188422688, 0.990012769041486, 
 
   0.987666756069459, 0.985096179880658, 0.982500083507245, 
 
   0.980149904480917, 0.978054024222106, 0.975950492211684, 
 
   0.973566004458122, 0.970680561569856, 0.968252417488493, 
 
   0.965946645658949, 0.962572816207979, 0.957605331702613, 
 
   0.9514846557537, 0.946133581761874, 0.941851386043165, 
 
   0.937866125544688, 0.932425488004115, 0.925920572169274, 
 
   0.917673712194962, 0.910192650812753, 0.902783083986748, 
 
   0.894588925819228, 0.886657619037551, 0.877285101626835, 
 
   0.86991693769838, 0.861152814246513, 0.852711634412064, 
 
   0.839988644551185, 0.826683282611567, 0.812015210970613, 
 
   0.800565011542073, 0.7891283675466344, 
 
   0.7766845228135053, 0.7614737433440416, 
 
   0.7472183938516067, 0.7367432640617278, 
 
   0.7284837529545244, 0.7232099503879037, 
 
   0.7100164984783785, 0.6948500612486456, 
 
   0.6749588449352819, 0.6557906620823331, 
 
   0.6522333811716634, 0.6292047665767445, 
 
   0.595535017776172, 0.5843631272470333, 
 
   0.5813077300968556, 0.5817213189872604, 
 
   0.5397163908004369, 0.512202518890555, 
 
   0.4854846755159279, 0.4687569074578323, 
 
   0.3716236343990294, 0.6165650651766813, 
 
   0.4234591456251031, 0.2542571544834555, 
 
   0.2657987929780235, 0.2380250402356247, 
 
   0.2283312332532089, 0.2098262250125013, 
 
   0.197486670065887, 0.1846926356668804, 
 
   0.1807846919428682, 0.1551222765118656, 
 
   0.1400158600791512, 0.111355445787778, 
 
   0.0981305007254373, 0.0801428925669899, 
 
   0.07758414626212024, 0.06905535882870848, 
 
   0.07479271688280709, 0.07779984484213509}, True, 0, 
 
  True, 2, Automatic, 3, 12.60353615113717, Hold[cp]]\
\>", "\<\
-DF[{1988, 7, 19}]-\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  "Setting ",
  StyleBox["LambdaValue -> 10^10 ",
    FontFamily->"courier",
    FontWeight->"Bold"],
  "fixes the value of $\\lambda$ without minimizing the GCV function. The ",
  StyleBox["DFSelectors",
    FontFamily->"courier",
    FontWeight->"Bold"],
  " can be used to extract information from the ",
  StyleBox["DF",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object:"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["DFSelectors", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
{AverageAbsoluteError, BetaHat, BetaHatCovarianceMatrix, 
 
  Delta, DropByTerm, EffectiveParameters, FinalSpline, 
 
  FixedLambdaQ, ForwardCurve, GCVMinimizedQ, GCVValue, 
 
  KeepList, Knots, Lambda, MaximumMaturity, 
 
  NumberOfObservations, ObservationWeights, PredictedPrice, 
 
  PriceError, QuoteDate, ResidualVariance, RestrictionQ, 
 
  SemiAnnualForwardCurve, SemiAnnualZeroCurve, 
 
  Significance, SplineOrder, TuningParameter, ZeroCurve}\
\>", "\<\
{AverageAbsoluteError, BetaHat, BetaHatCovarianceMatrix, 
 
  Delta, DropByTerm, EffectiveParameters, FinalSpline, 
 
  FixedLambdaQ, ForwardCurve, GCVMinimizedQ, GCVValue, 
 
  KeepList, Knots, Lambda, MaximumMaturity, 
 
  NumberOfObservations, ObservationWeights, PredictedPrice, 
 
  PriceError, QuoteDate, ResidualVariance, RestrictionQ, 
 
  SemiAnnualForwardCurve, SemiAnnualZeroCurve, 
 
  Significance, SplineOrder, TuningParameter, ZeroCurve}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell["For example, we can find the average absolute pricing error:", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["AverageAbsoluteError[df1]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
0.825380445338856\
\>", "\<\
0.82538\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell["\<\
This is an error of 82.5 basis points per \\$100 of face value. This is not \
very good, and we will see why in a moment.\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
There are other functions that take DF objects as arguments as well. For \
example,\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["DFSummaryStatistics[df1]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
TableForm[{PaddedForm[224, {3, 0}, NumberPoint -> \"\"], 
 
   StringForm[\"`1` `2`\", 
 
    PaddedForm[29.81793292265571, {5, 2}], \"years\"], 
 
   PaddedForm[0, {3, 0}, NumberPoint -> \"\"], Delta, True, 
 
   False, PaddedForm[2, {3, 0}, NumberPoint -> \"\"], 
 
   PaddedForm[74, {3, 0}, NumberPoint -> \"\"], 
 
   PaddedForm[AccountingForm[31.81202656402795, 
 
     NumberSigns -> {\"-\", \"\"}, NumberPadding -> {\" \", \"0\"}]\\
 
     , {4, 1}], PaddedForm[1.*10^10, {5, 2}, 
 
    NumberFormat -> (#1\"e\"#3 & )], 
 
   PaddedForm[AccountingForm[7.530029069597913, 
 
     NumberSigns -> {\"-\", \"\"}, NumberPadding -> {\" \", \"0\"}]\\
 
     , {8, 5}], PaddedForm[AccountingForm[0.8253804453388\\
 
      56, NumberSigns -> {\"-\", \"\"}, 
 
     NumberPadding -> {\" \", \"0\"}], {7, 4}]}, 
 
  TableSpacing -> {0}, TableHeadings -> 
 
   {{\"no. of obs.........\", \"longest payment....\", \"no.\\
 
      dropped........\", \"functional form....\", \"restriction\\
 
      ........\", \"minimize GCV.......\", \"tuning\\
 
      parameter...\", \"no. of knots.......\", \"no. of\\
 
      parameters..\", \"lambda.............\", \"residual\\
 
      variance..\", \"avg. abs. error....\"}}]\
\>", "\<\

no. of obs.........   224
longest payment....    29.82 years
no. dropped........     0
functional form....  Delta
restriction........  True
minimize GCV.......  False
tuning parameter...     2
no. of knots.......    74
no. of parameters..    31.8
lambda.............     1.00e10
residual variance..     7.53003
avg. abs. error....     0.8254\
\>"], "Output",
  LineSpacing->{1, 0},
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  "We see that the number of knot points (chosen automatically) is 74, but \
with the the penalty we used, $\\lambda = 10^{10}$, the effective number of \
parameters is only 31.8. Here are the forward rate (dashing) and zero-coupon \
rate curves (If you have a color screen, use can use the option ",
  StyleBox["Color -> True",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " in ShowGraph):\\footnote{$^6$}{Note that both ShowGraph and \
DFSummaryStatistics can be combined with DFSummaryPage, which is especially \
useful for sending reports to a printer.}"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["ShowGraph[df1, PlotStyle -> {Dashing[{.02, .01}],{}}];", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0237226 0.0317475 0.293809 0.00780523 [
[(0)] .02372 0 0 2 0 Minner Mrotsboxa
[(5)] .18246 0 0 2 0 Minner Mrotsboxa
[(10)] .3412 0 0 2 0 Minner Mrotsboxa
[(15)] .49993 0 0 2 0 Minner Mrotsboxa
[(20)] .65867 0 0 2 0 Minner Mrotsboxa
[(25)] .81741 0 0 2 0 Minner Mrotsboxa
[(30)] .97615 0 0 2 0 Minner Mrotsboxa
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotsboxa
[(-30)] -0.0125 .05965 1 0 0 Minner Mrotsboxa
[(-20)] -0.0125 .1377 1 0 0 Minner Mrotsboxa
[(-10)] -0.0125 .21576 1 0 0 Minner Mrotsboxa
[(0)] -0.0125 .29381 1 0 0 Minner Mrotsboxa
[(10)] -0.0125 .37186 1 0 0 Minner Mrotsboxa
[(20)] -0.0125 .44991 1 0 0 Minner Mrotsboxa
[(30)] -0.0125 .52797 1 0 0 Minner Mrotsboxa
[(40)] -0.0125 .60602 1 0 0 Minner Mrotsboxa
[(percent)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotsboxa
[(7/19/1988, 74 knots,  31.8 parms, Delta)] .5 .61803 0 -2 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
[ .002 .01 ] 0 setdash
.002 w
.02372 0 m
.02372 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.18246 0 m
.18246 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.3412 0 m
.3412 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.49993 0 m
.49993 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.65867 0 m
.65867 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.81741 0 m
.81741 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.97615 0 m
.97615 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .05965 m
1 .05965 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .1377 m
1 .1377 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .21576 m
1 .21576 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .29381 m
1 .29381 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .37186 m
1 .37186 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .44991 m
1 .44991 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .52797 m
1 .52797 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .60602 m
1 .60602 L
s
P
p
.002 w
.02372 0 m
.02372 .00625 L
s
P
[(0)] .02372 0 0 2 0 Minner Mrotshowa
p
.002 w
.18246 0 m
.18246 .00625 L
s
P
[(5)] .18246 0 0 2 0 Minner Mrotshowa
p
.002 w
.3412 0 m
.3412 .00625 L
s
P
[(10)] .3412 0 0 2 0 Minner Mrotshowa
p
.002 w
.49993 0 m
.49993 .00625 L
s
P
[(15)] .49993 0 0 2 0 Minner Mrotshowa
p
.002 w
.65867 0 m
.65867 .00625 L
s
P
[(20)] .65867 0 0 2 0 Minner Mrotshowa
p
.002 w
.81741 0 m
.81741 .00625 L
s
P
[(25)] .81741 0 0 2 0 Minner Mrotshowa
p
.002 w
.97615 0 m
.97615 .00625 L
s
P
[(30)] .97615 0 0 2 0 Minner Mrotshowa
p
.001 w
.05547 0 m
.05547 .00375 L
s
P
p
.001 w
.08722 0 m
.08722 .00375 L
s
P
p
.001 w
.11897 0 m
.11897 .00375 L
s
P
p
.001 w
.15071 0 m
.15071 .00375 L
s
P
p
.001 w
.21421 0 m
.21421 .00375 L
s
P
p
.001 w
.24595 0 m
.24595 .00375 L
s
P
p
.001 w
.2777 0 m
.2777 .00375 L
s
P
p
.001 w
.30945 0 m
.30945 .00375 L
s
P
p
.001 w
.37294 0 m
.37294 .00375 L
s
P
p
.001 w
.40469 0 m
.40469 .00375 L
s
P
p
.001 w
.43644 0 m
.43644 .00375 L
s
P
p
.001 w
.46819 0 m
.46819 .00375 L
s
P
p
.001 w
.53168 0 m
.53168 .00375 L
s
P
p
.001 w
.56343 0 m
.56343 .00375 L
s
P
p
.001 w
.59518 0 m
.59518 .00375 L
s
P
p
.001 w
.62692 0 m
.62692 .00375 L
s
P
p
.001 w
.69042 0 m
.69042 .00375 L
s
P
p
.001 w
.72217 0 m
.72217 .00375 L
s
P
p
.001 w
.75391 0 m
.75391 .00375 L
s
P
p
.001 w
.78566 0 m
.78566 .00375 L
s
P
p
.001 w
.84916 0 m
.84916 .00375 L
s
P
p
.001 w
.8809 0 m
.8809 .00375 L
s
P
p
.001 w
.91265 0 m
.91265 .00375 L
s
P
p
.001 w
.9444 0 m
.9444 .00375 L
s
P
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotshowa
p
.002 w
0 0 m
1 0 L
s
P
p
.002 w
0 .05965 m
.00625 .05965 L
s
P
[(-30)] -0.0125 .05965 1 0 0 Minner Mrotshowa
p
.002 w
0 .1377 m
.00625 .1377 L
s
P
[(-20)] -0.0125 .1377 1 0 0 Minner Mrotshowa
p
.002 w
0 .21576 m
.00625 .21576 L
s
P
[(-10)] -0.0125 .21576 1 0 0 Minner Mrotshowa
p
.002 w
0 .29381 m
.00625 .29381 L
s
P
[(0)] -0.0125 .29381 1 0 0 Minner Mrotshowa
p
.002 w
0 .37186 m
.00625 .37186 L
s
P
[(10)] -0.0125 .37186 1 0 0 Minner Mrotshowa
p
.002 w
0 .44991 m
.00625 .44991 L
s
P
[(20)] -0.0125 .44991 1 0 0 Minner Mrotshowa
p
.002 w
0 .52797 m
.00625 .52797 L
s
P
[(30)] -0.0125 .52797 1 0 0 Minner Mrotshowa
p
.002 w
0 .60602 m
.00625 .60602 L
s
P
[(40)] -0.0125 .60602 1 0 0 Minner Mrotshowa
p
.001 w
0 .07526 m
.00375 .07526 L
s
P
p
.001 w
0 .09087 m
.00375 .09087 L
s
P
p
.001 w
0 .10648 m
.00375 .10648 L
s
P
p
.001 w
0 .12209 m
.00375 .12209 L
s
P
p
.001 w
0 .15332 m
.00375 .15332 L
s
P
p
.001 w
0 .16893 m
.00375 .16893 L
s
P
p
.001 w
0 .18454 m
.00375 .18454 L
s
P
p
.001 w
0 .20015 m
.00375 .20015 L
s
P
p
.001 w
0 .23137 m
.00375 .23137 L
s
P
p
.001 w
0 .24698 m
.00375 .24698 L
s
P
p
.001 w
0 .26259 m
.00375 .26259 L
s
P
p
.001 w
0 .2782 m
.00375 .2782 L
s
P
p
.001 w
0 .30942 m
.00375 .30942 L
s
P
p
.001 w
0 .32503 m
.00375 .32503 L
s
P
p
.001 w
0 .34064 m
.00375 .34064 L
s
P
p
.001 w
0 .35625 m
.00375 .35625 L
s
P
p
.001 w
0 .38747 m
.00375 .38747 L
s
P
p
.001 w
0 .40308 m
.00375 .40308 L
s
P
p
.001 w
0 .41869 m
.00375 .41869 L
s
P
p
.001 w
0 .4343 m
.00375 .4343 L
s
P
p
.001 w
0 .46552 m
.00375 .46552 L
s
P
p
.001 w
0 .48113 m
.00375 .48113 L
s
P
p
.001 w
0 .49675 m
.00375 .49675 L
s
P
p
.001 w
0 .51236 m
.00375 .51236 L
s
P
p
.001 w
0 .54358 m
.00375 .54358 L
s
P
p
.001 w
0 .55919 m
.00375 .55919 L
s
P
p
.001 w
0 .5748 m
.00375 .5748 L
s
P
p
.001 w
0 .59041 m
.00375 .59041 L
s
P
p
.001 w
0 .04404 m
.00375 .04404 L
s
P
p
.001 w
0 .02843 m
.00375 .02843 L
s
P
p
.001 w
0 .01282 m
.00375 .01282 L
s
P
[(percent)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotshowa
p
.002 w
0 0 m
0 .61803 L
s
P
P
p
[(7/19/1988, 74 knots,  31.8 parms, Delta)] .5 .61803 0 -2 Mshowa
P
p
P
p
p
[ .02 .01 ] 0 setdash
p
.004 w
.02381 .3482 m
.03373 .35315 L
.03621 .3544 L
.03869 .35595 L
.03993 .35658 L
.04117 .35706 L
.04365 .35767 L
.04489 .35796 L
.04613 .35822 L
.04737 .35843 L
.04861 .35854 L
.04985 .35847 L
.05109 .35847 L
.05357 .35894 L
.05481 .35927 L
.05605 .35959 L
.05729 .35977 L
.05853 .3598 L
.05977 .35974 L
.06101 .35959 L
.06349 .35902 L
.06473 .35857 L
.06597 .35816 L
.06721 .35785 L
.06845 .35775 L
.06969 .35793 L
.07093 .35838 L
.07341 .36007 L
.07465 .36113 L
.07589 .36216 L
.07713 .3629 L
.07837 .36335 L
.07961 .36354 L
.08085 .36358 L
.08333 .36346 L
.08457 .36345 L
.08581 .36349 L
.08705 .36351 L
.08829 .36348 L
.08953 .3634 L
.09077 .36327 L
.09325 .36285 L
.09573 .36223 L
.09697 .36183 L
.09821 .36141 L
.09945 .3611 L
.10069 .3609 L
.10193 .36081 L
.10317 .36084 L
.10441 .36092 L
Mistroke
.10565 .36102 L
.10689 .36116 L
.10813 .36132 L
.11062 .36173 L
.11186 .36195 L
.1131 .36218 L
.11434 .36247 L
.11558 .36291 L
.11682 .36349 L
.11806 .36422 L
.12054 .36578 L
.12178 .3664 L
.12302 .36691 L
.12426 .36731 L
.1255 .36759 L
.12674 .36775 L
.12798 .36762 L
.12922 .36713 L
.13046 .36627 L
.1317 .36503 L
.13294 .36341 L
.13542 .35888 L
.13666 .35589 L
.1379 .35242 L
.14038 .3438 L
.14162 .33864 L
.14286 .3331 L
.1441 .32906 L
.14534 .32701 L
.14658 .32696 L
.14782 .32892 L
.1503 .33748 L
.15278 .34511 L
.15402 .34815 L
.15526 .35067 L
.15774 .35413 L
.15898 .35522 L
.16022 .35671 L
.16146 .35869 L
.1627 .36117 L
.16518 .36767 L
.16766 .37553 L
.1689 .37869 L
.17014 .38118 L
.17138 .38297 L
.17262 .38407 L
.17386 .38445 L
.1751 .38378 L
.17634 .38141 L
.17758 .37729 L
Mistroke
.17882 .37138 L
.18006 .36367 L
.1813 .35415 L
.18254 .3433 L
.18378 .33523 L
.18502 .33087 L
.18626 .33024 L
.1875 .33338 L
.19246 .37496 L
.19494 .39551 L
.19742 .41272 L
.19866 .41998 L
.1999 .42524 L
.20114 .42814 L
.20238 .4286 L
.20362 .42655 L
.20486 .42194 L
.2061 .41469 L
.20734 .40476 L
.20982 .37883 L
.2123 .35592 L
.21478 .33806 L
.21726 .32527 L
.21974 .31506 L
.22222 .3067 L
.22346 .30322 L
.2247 .30033 L
.22594 .30082 L
.22718 .30596 L
.23214 .37352 L
.23338 .39942 L
.23462 .42018 L
.23586 .43548 L
.2371 .44522 L
.23834 .44924 L
.23958 .44736 L
.24206 .43308 L
.24454 .41981 L
.24702 .4084 L
.2495 .399 L
.25074 .39509 L
.25198 .39173 L
.25322 .38893 L
.25446 .38671 L
.25694 .38387 L
.2619 .37799 L
.26687 .37 L
.26935 .36518 L
.27059 .36256 L
.27183 .3598 L
Mistroke
.27307 .35745 L
.27431 .35638 L
.27555 .35662 L
.27679 .35816 L
.27803 .36105 L
.27927 .36529 L
.28051 .3709 L
.28175 .37791 L
.28423 .39626 L
.28671 .42065 L
.28919 .44946 L
.29043 .46113 L
.29167 .47018 L
.29291 .47653 L
.29415 .4801 L
.29539 .48077 L
.29663 .47843 L
.29787 .47297 L
.29911 .46429 L
.30035 .45229 L
.30159 .43688 L
.31151 .19092 L
.31399 .11384 L
.31523 .0838 L
.31647 .0594 L
.31771 .0405 L
.31895 .02691 L
.32019 .0184 L
.32143 .01472 L
.32267 .01561 L
.32391 .02084 L
.32515 .03017 L
.32639 .04338 L
.33135 .1149 L
.34127 .23173 L
.36111 .39685 L
.37103 .46127 L
.38095 .51734 L
.39087 .5642 L
.39583 .58302 L
.39831 .59092 L
.39955 .59441 L
.40079 .5974 L
.40203 .59982 L
.40327 .60164 L
.40451 .60282 L
.40575 .60332 L
.40699 .60309 L
.40823 .6021 L
.40947 .60029 L
Mistroke
.41071 .59761 L
.41195 .59403 L
.41319 .58948 L
.41567 .57732 L
.41815 .56075 L
.42063 .53944 L
.4256 .48156 L
.43056 .40288 L
.43304 .36212 L
.43428 .34508 L
.43552 .33038 L
.43676 .31808 L
.438 .3082 L
.43924 .30077 L
.44048 .29581 L
.44172 .29332 L
.44296 .29331 L
.4442 .29577 L
.44544 .30071 L
.4504 .33196 L
.45536 .35722 L
.45784 .36708 L
.46032 .37507 L
.46156 .37835 L
.4628 .38114 L
.46404 .38345 L
.46528 .38525 L
.46652 .38655 L
.46776 .38733 L
.469 .38757 L
.47024 .38729 L
.47148 .38645 L
.47272 .38505 L
.47396 .38309 L
.4752 .38054 L
.47768 .37369 L
.48016 .36563 L
.4814 .3624 L
.48264 .35977 L
.48388 .35777 L
.48512 .35639 L
.48636 .35565 L
.4876 .35555 L
.48884 .35611 L
.49008 .35733 L
.49132 .35922 L
.49256 .36178 L
.49504 .36871 L
.49752 .37505 L
.49876 .37766 L
Mistroke
.5 .3799 L
.50124 .38177 L
.50248 .38324 L
.50372 .38432 L
.50496 .38501 L
.5062 .38528 L
.50744 .38514 L
.50868 .38457 L
.50992 .38358 L
.51488 .3786 L
.51984 .37454 L
.5248 .37151 L
.52728 .37041 L
.52852 .36996 L
.52976 .36959 L
.531 .36929 L
.53224 .36907 L
.53348 .36892 L
.53472 .36886 L
.53596 .36887 L
.5372 .36896 L
.53968 .36938 L
.54092 .36972 L
.54216 .37014 L
.5434 .37039 L
.54464 .37035 L
.54588 .37001 L
.54712 .36937 L
.54836 .36842 L
.5496 .36715 L
.55208 .36367 L
.55456 .3589 L
.55704 .35279 L
.55828 .34934 L
.55952 .34656 L
.56076 .34463 L
.562 .34358 L
.56324 .34342 L
.56448 .34415 L
.56572 .34579 L
.56696 .34833 L
.5682 .3518 L
.56944 .3562 L
.57192 .36787 L
.5744 .38317 L
.57688 .39713 L
.57813 .40292 L
.57937 .40792 L
.58061 .4121 L
.58185 .41545 L
Mistroke
.58309 .41795 L
.58433 .41957 L
.58557 .42029 L
.58681 .4201 L
.58805 .41896 L
.58929 .41686 L
.59177 .41155 L
.59425 .40638 L
.59921 .3965 L
.60913 .37913 L
.61409 .37191 L
.61905 .36581 L
.62401 .36094 L
.62649 .35899 L
.62897 .35739 L
.63145 .35613 L
.63269 .35564 L
.63393 .35524 L
.63517 .35493 L
.63641 .35472 L
.63889 .35447 L
.65873 .35244 L
.69841 .34727 L
.71825 .34486 L
.72817 .34392 L
.73313 .34353 L
.7381 .34319 L
.74306 .34291 L
.74802 .34269 L
.7505 .34261 L
.75298 .34254 L
.75422 .34251 L
.75546 .34248 L
.7567 .34246 L
.75794 .34244 L
.75918 .34243 L
.76042 .34242 L
.76166 .34242 L
.7629 .34241 L
.76414 .34242 L
.76538 .34243 L
.76662 .34244 L
.76786 .34246 L
.77034 .34251 L
.77282 .34257 L
.7753 .34266 L
.77778 .34277 L
.78274 .34304 L
.7877 .34339 L
.79762 .34435 L
Mistroke
.80754 .34568 L
.81746 .3474 L
.82738 .34956 L
.8373 .35219 L
.84722 .35534 L
.85714 .35906 L
.8621 .36115 L
.86458 .36226 L
.86706 .36342 L
.8683 .36398 L
.86954 .3644 L
.87078 .36467 L
.87202 .36478 L
.87326 .36473 L
.8745 .36453 L
.87574 .36416 L
.87698 .36362 L
.87822 .36292 L
.87946 .36204 L
.88194 .35978 L
.88442 .35681 L
.8869 .35312 L
.88938 .34869 L
.89187 .34363 L
.89311 .34143 L
.89435 .33968 L
.89559 .33839 L
.89683 .33756 L
.89807 .33719 L
.89931 .3373 L
.90055 .33787 L
.90179 .33892 L
.90303 .34046 L
.90427 .34248 L
.90675 .348 L
.90923 .35415 L
.91047 .35672 L
.91171 .35893 L
.91295 .36079 L
.91419 .36228 L
.91543 .36341 L
.91667 .36417 L
.91791 .36455 L
.91915 .36455 L
.92039 .36416 L
.92163 .36338 L
.92287 .3622 L
.92411 .36062 L
.92535 .35863 L
.92659 .35623 L
Mistroke
.93651 .32178 L
.97619 .21729 L
Mfstroke
P
P
p
p
.004 w
.02381 .3482 m
.03373 .35037 L
.03869 .35173 L
.04365 .35303 L
.04861 .35406 L
.05357 .35481 L
.05853 .35548 L
.06101 .35576 L
.06225 .35588 L
.06349 .35598 L
.06473 .35607 L
.06597 .35614 L
.06845 .35623 L
.06969 .35628 L
.07093 .35632 L
.07217 .35639 L
.07341 .35647 L
.07589 .35669 L
.07837 .35697 L
.08333 .35751 L
.08829 .35797 L
.09077 .35817 L
.09325 .35835 L
.09573 .35849 L
.09821 .3586 L
.10069 .35868 L
.10317 .35875 L
.10813 .35888 L
.1131 .35904 L
.11558 .35914 L
.11806 .35925 L
.12302 .35957 L
.12798 .35995 L
.12922 .36004 L
.13046 .36012 L
.1317 .36018 L
.13294 .36023 L
.13418 .36025 L
.13542 .36025 L
.13666 .36022 L
.1379 .36015 L
.14286 .35946 L
.14534 .35884 L
.14782 .35821 L
.14906 .35794 L
.1503 .35772 L
.15154 .35754 L
.15278 .35741 L
.15402 .3573 L
.15526 .35723 L
Mistroke
.1565 .35718 L
.15774 .35714 L
.15898 .35712 L
.16022 .35711 L
.16146 .35712 L
.1627 .35714 L
.16394 .35719 L
.16518 .35727 L
.16642 .35737 L
.16766 .35751 L
.17262 .35829 L
.1751 .35871 L
.17634 .35891 L
.17758 .35907 L
.17882 .3592 L
.18006 .35926 L
.1813 .35926 L
.18254 .35918 L
.18502 .35882 L
.18626 .3586 L
.1875 .3584 L
.18874 .35823 L
.18998 .35814 L
.19122 .35813 L
.19246 .35821 L
.1937 .35838 L
.19494 .35861 L
.19742 .35926 L
.20238 .36105 L
.20486 .36194 L
.2061 .36233 L
.20734 .36265 L
.20858 .36289 L
.20982 .36304 L
.21106 .3631 L
.2123 .36309 L
.21354 .36301 L
.21478 .36288 L
.21726 .36247 L
.21974 .36194 L
.22222 .3613 L
.2247 .36058 L
.22718 .35986 L
.22842 .35956 L
.22966 .35934 L
.2309 .35922 L
.23214 .35923 L
.23338 .35939 L
.23462 .35969 L
.2371 .36056 L
Mistroke
.23958 .36157 L
.24206 .36247 L
.24454 .36318 L
.24702 .36375 L
.2495 .36418 L
.25198 .36452 L
.25694 .365 L
.25942 .36519 L
.2619 .36534 L
.26438 .36545 L
.26563 .36549 L
.26687 .36552 L
.26811 .36553 L
.26935 .36554 L
.27059 .36553 L
.27183 .36551 L
.27431 .36543 L
.27555 .36539 L
.27679 .36535 L
.27803 .36532 L
.27927 .36531 L
.28051 .36532 L
.28175 .36536 L
.28299 .36544 L
.28423 .36556 L
.28547 .36574 L
.28671 .36596 L
.29167 .36748 L
.29663 .36949 L
.29911 .37042 L
.30035 .37082 L
.30159 .37115 L
.30283 .3714 L
.30407 .37156 L
.30531 .37161 L
.30655 .37153 L
.30779 .37133 L
.30903 .37097 L
.31027 .37046 L
.31151 .36978 L
.32143 .35992 L
.32639 .35441 L
.33135 .34997 L
.33383 .34822 L
.33631 .34675 L
.33879 .34553 L
.34127 .34454 L
.34375 .34376 L
.34499 .34345 L
.34623 .34318 L
Mistroke
.34747 .34296 L
.34871 .34278 L
.34995 .34264 L
.35119 .34254 L
.35243 .34248 L
.35367 .34246 L
.35491 .34248 L
.35615 .34253 L
.35739 .34262 L
.35863 .34274 L
.36111 .34307 L
.36359 .34352 L
.36607 .34409 L
.37103 .34555 L
.38095 .34956 L
.40079 .36073 L
.41071 .3669 L
.41567 .36971 L
.42063 .37209 L
.42312 .37305 L
.4256 .37382 L
.42684 .37412 L
.42808 .37437 L
.42932 .37456 L
.43056 .37467 L
.44048 .37377 L
.44544 .37284 L
.44792 .37247 L
.44916 .37232 L
.4504 .37219 L
.45164 .37208 L
.45288 .372 L
.45412 .37193 L
.45536 .37188 L
.4566 .37184 L
.45784 .37182 L
.45908 .37182 L
.46032 .37182 L
.47024 .37209 L
.47272 .37217 L
.47396 .3722 L
.4752 .37223 L
.47644 .37224 L
.47768 .37225 L
.47892 .37225 L
.48016 .37224 L
.48264 .37219 L
.48388 .37215 L
.48512 .37211 L
.4876 .37202 L
Mistroke
.49008 .37194 L
.49132 .3719 L
.49256 .37187 L
.4938 .37185 L
.49504 .37184 L
.49628 .37183 L
.49752 .37184 L
.49876 .37185 L
.5 .37187 L
.50992 .37211 L
.5124 .37216 L
.51364 .37218 L
.51488 .3722 L
.51612 .37222 L
.51736 .37223 L
.5186 .37224 L
.51984 .37225 L
.52108 .37225 L
.52232 .37225 L
.52356 .37225 L
.5248 .37225 L
.52604 .37225 L
.52728 .37225 L
.52976 .37223 L
.53472 .3722 L
.53968 .37217 L
.54464 .37215 L
.54588 .37215 L
.54712 .37214 L
.54836 .37214 L
.5496 .37213 L
.55084 .37211 L
.55208 .37209 L
.55456 .37204 L
.5558 .37201 L
.55704 .37197 L
.55952 .37187 L
.56448 .37161 L
.56696 .37149 L
.5682 .37144 L
.56944 .37141 L
.57068 .37138 L
.57192 .37136 L
.57316 .37136 L
.5744 .37138 L
.57564 .37141 L
.57688 .37146 L
.57813 .37153 L
.57937 .3716 L
.58929 .3724 L
Mistroke
.59425 .37274 L
.59673 .37288 L
.59921 .37299 L
.60169 .37308 L
.60417 .37315 L
.60665 .3732 L
.60789 .37322 L
.60913 .37324 L
.61037 .37325 L
.61161 .37325 L
.61285 .37326 L
.61409 .37326 L
.61533 .37325 L
.61657 .37324 L
.61781 .37323 L
.61905 .37322 L
.62153 .37318 L
.62401 .37314 L
.62897 .37302 L
.63889 .37274 L
.65873 .37214 L
.69841 .37083 L
.7381 .3694 L
.77778 .36799 L
.79762 .36736 L
.81746 .36682 L
.82738 .36659 L
.8373 .3664 L
.84722 .36625 L
.85714 .36614 L
.85962 .36612 L
.8621 .3661 L
.86458 .36609 L
.86582 .36608 L
.86706 .36608 L
.8683 .36608 L
.86954 .36607 L
.87202 .36607 L
.8745 .36607 L
.87574 .36606 L
.87698 .36606 L
.87822 .36606 L
.87946 .36605 L
.8807 .36604 L
.88194 .36604 L
.88318 .36603 L
.88442 .36601 L
.8869 .36598 L
.88938 .36594 L
.89187 .36588 L
Mistroke
.89683 .36574 L
.90179 .36558 L
.90427 .36551 L
.90551 .36548 L
.90675 .36545 L
.90799 .36543 L
.90923 .36541 L
.91047 .36539 L
.91171 .36538 L
.91295 .36538 L
.91419 .36537 L
.91543 .36537 L
.91667 .36537 L
.91791 .36536 L
.91915 .36536 L
.92039 .36536 L
.92163 .36536 L
.92287 .36536 L
.92411 .36535 L
.92535 .36534 L
.92659 .36533 L
.92783 .36532 L
.92907 .3653 L
.93155 .36525 L
.93279 .36521 L
.93403 .36517 L
.93651 .36507 L
.93899 .36494 L
.94147 .36477 L
.94643 .36433 L
.95635 .36317 L
.97619 .36024 L
Mfstroke
P
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell["\<\
This looks bad---the term structure is over-parameterized. Let's examine the \
generalized cross validation function:\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["Plot[GCVFunction[10^x], {x, 0, 20}];", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.047619 0.0147151 0.00246876 [
[(5)] .2619 .01472 0 2 Msboxa
[(10)] .5 .01472 0 2 Msboxa
[(15)] .7381 .01472 0 2 Msboxa
[(20)] .97619 .01472 0 2 Msboxa
[(50)] .01131 .13815 1 0 Msboxa
[(100)] .01131 .26159 1 0 Msboxa
[(150)] .01131 .38503 1 0 Msboxa
[(200)] .01131 .50847 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
.002 w
.2619 .01472 m
.2619 .02097 L
s
P
[(5)] .2619 .01472 0 2 Mshowa
p
.002 w
.5 .01472 m
.5 .02097 L
s
P
[(10)] .5 .01472 0 2 Mshowa
p
.002 w
.7381 .01472 m
.7381 .02097 L
s
P
[(15)] .7381 .01472 0 2 Mshowa
p
.002 w
.97619 .01472 m
.97619 .02097 L
s
P
[(20)] .97619 .01472 0 2 Mshowa
p
.001 w
.07143 .01472 m
.07143 .01847 L
s
P
p
.001 w
.11905 .01472 m
.11905 .01847 L
s
P
p
.001 w
.16667 .01472 m
.16667 .01847 L
s
P
p
.001 w
.21429 .01472 m
.21429 .01847 L
s
P
p
.001 w
.30952 .01472 m
.30952 .01847 L
s
P
p
.001 w
.35714 .01472 m
.35714 .01847 L
s
P
p
.001 w
.40476 .01472 m
.40476 .01847 L
s
P
p
.001 w
.45238 .01472 m
.45238 .01847 L
s
P
p
.001 w
.54762 .01472 m
.54762 .01847 L
s
P
p
.001 w
.59524 .01472 m
.59524 .01847 L
s
P
p
.001 w
.64286 .01472 m
.64286 .01847 L
s
P
p
.001 w
.69048 .01472 m
.69048 .01847 L
s
P
p
.001 w
.78571 .01472 m
.78571 .01847 L
s
P
p
.001 w
.83333 .01472 m
.83333 .01847 L
s
P
p
.001 w
.88095 .01472 m
.88095 .01847 L
s
P
p
.001 w
.92857 .01472 m
.92857 .01847 L
s
P
p
.002 w
0 .01472 m
1 .01472 L
s
P
p
.002 w
.02381 .13815 m
.03006 .13815 L
s
P
[(50)] .01131 .13815 1 0 Mshowa
p
.002 w
.02381 .26159 m
.03006 .26159 L
s
P
[(100)] .01131 .26159 1 0 Mshowa
p
.002 w
.02381 .38503 m
.03006 .38503 L
s
P
[(150)] .01131 .38503 1 0 Mshowa
p
.002 w
.02381 .50847 m
.03006 .50847 L
s
P
[(200)] .01131 .50847 1 0 Mshowa
p
.001 w
.02381 .0394 m
.02756 .0394 L
s
P
p
.001 w
.02381 .06409 m
.02756 .06409 L
s
P
p
.001 w
.02381 .08878 m
.02756 .08878 L
s
P
p
.001 w
.02381 .11347 m
.02756 .11347 L
s
P
p
.001 w
.02381 .16284 m
.02756 .16284 L
s
P
p
.001 w
.02381 .18753 m
.02756 .18753 L
s
P
p
.001 w
.02381 .21222 m
.02756 .21222 L
s
P
p
.001 w
.02381 .2369 m
.02756 .2369 L
s
P
p
.001 w
.02381 .28628 m
.02756 .28628 L
s
P
p
.001 w
.02381 .31097 m
.02756 .31097 L
s
P
p
.001 w
.02381 .33565 m
.02756 .33565 L
s
P
p
.001 w
.02381 .36034 m
.02756 .36034 L
s
P
p
.001 w
.02381 .40972 m
.02756 .40972 L
s
P
p
.001 w
.02381 .4344 m
.02756 .4344 L
s
P
p
.001 w
.02381 .45909 m
.02756 .45909 L
s
P
p
.001 w
.02381 .48378 m
.02756 .48378 L
s
P
p
.001 w
.02381 .53315 m
.02756 .53315 L
s
P
p
.001 w
.02381 .55784 m
.02756 .55784 L
s
P
p
.001 w
.02381 .58253 m
.02756 .58253 L
s
P
p
.001 w
.02381 .60722 m
.02756 .60722 L
s
P
p
.002 w
.02381 0 m
.02381 .61803 L
s
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
.004 w
.02381 .16812 m
.02877 .16812 L
.03373 .16812 L
.03869 .16812 L
.04365 .16812 L
.04861 .16812 L
.05357 .16812 L
.05853 .16812 L
.06349 .16812 L
.06845 .16812 L
.07341 .16811 L
.07837 .16811 L
.08333 .16811 L
.08829 .16811 L
.09325 .16811 L
.09821 .16811 L
.10317 .16811 L
.10813 .16811 L
.1131 .16811 L
.11806 .1681 L
.12302 .1681 L
.12798 .1681 L
.13294 .16809 L
.1379 .16808 L
.14286 .16808 L
.14782 .16806 L
.15278 .16805 L
.15774 .16803 L
.1627 .16801 L
.16766 .16798 L
.17262 .16794 L
.17758 .1679 L
.18254 .16784 L
.1875 .16777 L
.19246 .16767 L
.19742 .16755 L
.20238 .16741 L
.20734 .16722 L
.2123 .16698 L
.21726 .16669 L
.22222 .16632 L
.23214 .1653 L
.2371 .16462 L
.24206 .16378 L
.25198 .16159 L
.2619 .15857 L
.27183 .15465 L
.28175 .14986 L
.30159 .13829 L
.34127 .11248 L
Mistroke
.38095 .08874 L
.42063 .06888 L
.44048 .06089 L
.46032 .0545 L
.48016 .04957 L
.5 .04583 L
.50992 .04438 L
.51984 .04318 L
.52976 .0422 L
.53968 .0414 L
.5496 .04078 L
.55952 .0403 L
.56944 .03995 L
.5744 .03982 L
.57937 .0397 L
.58929 .03953 L
.59921 .0394 L
.61905 .03922 L
.63889 .03902 L
.65873 .03876 L
.66865 .03863 L
.67361 .03857 L
.67857 .03852 L
.68105 .0385 L
.68353 .03848 L
.68601 .03846 L
.68849 .03845 L
.68973 .03845 L
.69097 .03844 L
.69221 .03844 L
.69345 .03844 L
.69469 .03844 L
.69593 .03844 L
.69717 .03845 L
.69841 .03845 L
.69965 .03846 L
.70089 .03847 L
.70337 .03849 L
.70585 .03853 L
.70833 .03857 L
.71081 .03863 L
.71329 .03871 L
.71577 .03881 L
.71825 .03894 L
.72073 .03909 L
.72321 .03929 L
.72569 .03953 L
.72817 .03983 L
.73313 .04066 L
.73562 .04123 L
Mistroke
.7381 .04193 L
.74058 .04279 L
.74306 .04384 L
.74554 .04513 L
.74802 .0467 L
.75298 .05089 L
.75546 .05365 L
.75794 .05694 L
.7629 .06549 L
.76786 .07726 L
.77282 .09297 L
.77778 .11324 L
.7877 .16848 L
.79762 .24052 L
.81746 .39556 L
.82738 .4586 L
.8373 .50639 L
.84226 .52487 L
.84722 .54019 L
.85218 .55275 L
.85714 .56297 L
.86706 .57787 L
.87202 .58318 L
.87698 .58741 L
.88194 .59077 L
.8869 .59344 L
.89187 .59555 L
.89683 .59722 L
.90179 .59854 L
.90675 .59958 L
.91171 .60041 L
.91667 .60105 L
.92163 .60156 L
.92659 .60197 L
.93155 .60228 L
.93651 .60253 L
.94147 .60273 L
.94643 .60288 L
.95139 .603 L
.95635 .6031 L
.96131 .60317 L
.96627 .60323 L
.97619 .60332 L
Mfstroke
P
P
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell["\<\
Notice the fuction is flat at both ends. An adaptive optimizer (such as \
FindMinimum) can get stuck in these flat regions. It looks like the minimum \
is between $10^{10}$ and $10^{16}$. \
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["Plot[GCVFunction[10^x], {x, 10, 16}];", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-1.56349 0.15873 -0.969222 0.10239 [
[(11)] .18254 .05468 0 2 Msboxa
[(12)] .34127 .05468 0 2 Msboxa
[(13)] .5 .05468 0 2 Msboxa
[(14)] .65873 .05468 0 2 Msboxa
[(15)] .81746 .05468 0 2 Msboxa
[(16)] .97619 .05468 0 2 Msboxa
[(11)] .01131 .15707 1 0 Msboxa
[(12)] .01131 .25946 1 0 Msboxa
[(13)] .01131 .36185 1 0 Msboxa
[(14)] .01131 .46424 1 0 Msboxa
[(15)] .01131 .56663 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
.002 w
.18254 .05468 m
.18254 .06093 L
s
P
[(11)] .18254 .05468 0 2 Mshowa
p
.002 w
.34127 .05468 m
.34127 .06093 L
s
P
[(12)] .34127 .05468 0 2 Mshowa
p
.002 w
.5 .05468 m
.5 .06093 L
s
P
[(13)] .5 .05468 0 2 Mshowa
p
.002 w
.65873 .05468 m
.65873 .06093 L
s
P
[(14)] .65873 .05468 0 2 Mshowa
p
.002 w
.81746 .05468 m
.81746 .06093 L
s
P
[(15)] .81746 .05468 0 2 Mshowa
p
.002 w
.97619 .05468 m
.97619 .06093 L
s
P
[(16)] .97619 .05468 0 2 Mshowa
p
.001 w
.05556 .05468 m
.05556 .05843 L
s
P
p
.001 w
.0873 .05468 m
.0873 .05843 L
s
P
p
.001 w
.11905 .05468 m
.11905 .05843 L
s
P
p
.001 w
.15079 .05468 m
.15079 .05843 L
s
P
p
.001 w
.21429 .05468 m
.21429 .05843 L
s
P
p
.001 w
.24603 .05468 m
.24603 .05843 L
s
P
p
.001 w
.27778 .05468 m
.27778 .05843 L
s
P
p
.001 w
.30952 .05468 m
.30952 .05843 L
s
P
p
.001 w
.37302 .05468 m
.37302 .05843 L
s
P
p
.001 w
.40476 .05468 m
.40476 .05843 L
s
P
p
.001 w
.43651 .05468 m
.43651 .05843 L
s
P
p
.001 w
.46825 .05468 m
.46825 .05843 L
s
P
p
.001 w
.53175 .05468 m
.53175 .05843 L
s
P
p
.001 w
.56349 .05468 m
.56349 .05843 L
s
P
p
.001 w
.59524 .05468 m
.59524 .05843 L
s
P
p
.001 w
.62698 .05468 m
.62698 .05843 L
s
P
p
.001 w
.69048 .05468 m
.69048 .05843 L
s
P
p
.001 w
.72222 .05468 m
.72222 .05843 L
s
P
p
.001 w
.75397 .05468 m
.75397 .05843 L
s
P
p
.001 w
.78571 .05468 m
.78571 .05843 L
s
P
p
.001 w
.84921 .05468 m
.84921 .05843 L
s
P
p
.001 w
.88095 .05468 m
.88095 .05843 L
s
P
p
.001 w
.9127 .05468 m
.9127 .05843 L
s
P
p
.001 w
.94444 .05468 m
.94444 .05843 L
s
P
p
.002 w
0 .05468 m
1 .05468 L
s
P
p
.002 w
.02381 .15707 m
.03006 .15707 L
s
P
[(11)] .01131 .15707 1 0 Mshowa
p
.002 w
.02381 .25946 m
.03006 .25946 L
s
P
[(12)] .01131 .25946 1 0 Mshowa
p
.002 w
.02381 .36185 m
.03006 .36185 L
s
P
[(13)] .01131 .36185 1 0 Mshowa
p
.002 w
.02381 .46424 m
.03006 .46424 L
s
P
[(14)] .01131 .46424 1 0 Mshowa
p
.002 w
.02381 .56663 m
.03006 .56663 L
s
P
[(15)] .01131 .56663 1 0 Mshowa
p
.001 w
.02381 .07515 m
.02756 .07515 L
s
P
p
.001 w
.02381 .09563 m
.02756 .09563 L
s
P
p
.001 w
.02381 .11611 m
.02756 .11611 L
s
P
p
.001 w
.02381 .13659 m
.02756 .13659 L
s
P
p
.001 w
.02381 .17754 m
.02756 .17754 L
s
P
p
.001 w
.02381 .19802 m
.02756 .19802 L
s
P
p
.001 w
.02381 .2185 m
.02756 .2185 L
s
P
p
.001 w
.02381 .23898 m
.02756 .23898 L
s
P
p
.001 w
.02381 .27993 m
.02756 .27993 L
s
P
p
.001 w
.02381 .30041 m
.02756 .30041 L
s
P
p
.001 w
.02381 .32089 m
.02756 .32089 L
s
P
p
.001 w
.02381 .34137 m
.02756 .34137 L
s
P
p
.001 w
.02381 .38232 m
.02756 .38232 L
s
P
p
.001 w
.02381 .4028 m
.02756 .4028 L
s
P
p
.001 w
.02381 .42328 m
.02756 .42328 L
s
P
p
.001 w
.02381 .44376 m
.02756 .44376 L
s
P
p
.001 w
.02381 .48471 m
.02756 .48471 L
s
P
p
.001 w
.02381 .50519 m
.02756 .50519 L
s
P
p
.001 w
.02381 .52567 m
.02756 .52567 L
s
P
p
.001 w
.02381 .54615 m
.02756 .54615 L
s
P
p
.001 w
.02381 .0342 m
.02756 .0342 L
s
P
p
.001 w
.02381 .01372 m
.02756 .01372 L
s
P
p
.001 w
.02381 .5871 m
.02756 .5871 L
s
P
p
.001 w
.02381 .60758 m
.02756 .60758 L
s
P
p
.002 w
.02381 0 m
.02381 .61803 L
s
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
.004 w
.02381 .32125 m
.06349 .25022 L
.10317 .19389 L
.14286 .14996 L
.18254 .11643 L
.20238 .10313 L
.22222 .09194 L
.24206 .08266 L
.2619 .07507 L
.28175 .06893 L
.30159 .06397 L
.34127 .05665 L
.38095 .05143 L
.42063 .04699 L
.46032 .04229 L
.5 .03671 L
.53968 .03031 L
.57937 .02374 L
.59921 .02072 L
.61905 .01809 L
.62897 .017 L
.63889 .01608 L
.64385 .01569 L
.64881 .01536 L
.65377 .0151 L
.65625 .01499 L
.65873 .0149 L
.66121 .01482 L
.66245 .01479 L
.66369 .01477 L
.66493 .01475 L
.66617 .01473 L
.66741 .01472 L
.66865 .01472 L
.66989 .01472 L
.67113 .01472 L
.67237 .01473 L
.67361 .01475 L
.67485 .01477 L
.67609 .01479 L
.67857 .01487 L
.68105 .01496 L
.68353 .01508 L
.68849 .01541 L
.69345 .01585 L
.69841 .01641 L
.70337 .01712 L
.70833 .01798 L
.71825 .02024 L
.72817 .02335 L
Mistroke
.7381 .02753 L
.74802 .03307 L
.75794 .04035 L
.76786 .04984 L
.77778 .06218 L
.7877 .07816 L
.79762 .0988 L
.80754 .12539 L
.81746 .15954 L
.82738 .20325 L
.8373 .25897 L
.84722 .32969 L
.85714 .41899 L
.86706 .53109 L
Mfstroke
.87323 .61803 m
.86706 .53109 L
s
s
s
s
s
s
s
s
P
P
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell["\<\
The minimum appears to be close to $10^{14}$. Let's use that as a starting \
value to applay GCV. Note that a pair of starting values is required.\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell[TextData[StyleBox["df2 = DiscountFunction[cp, FunctionalForm -> Delta,\n\
\tLambda0 -> {10^14, 1.1 10^14}];\n{Lambda[df2], EffectiveParameters[df2], \n\
\tAverageAbsoluteError[df2]}",
  FontFamily->"courier"]], "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
{1.167087528370603*10^14, 4.145258889441611, 
 
  0.919242029791923}\
\>", "\<\
           14
{1.16709 10  , 4.14526, 0.919242}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell["\<\
We see that $\\lambda^*$ is close to our guess, and that the number of \
parameters has fallen to 4.1 while the average absolute error has risen a bit \
to 92 basis points. Let's look at the yield curves:\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["ShowGraph[df2, PlotStyle -> {Dashing[{.02, .01}],{}}];", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0237226 0.0317475 -1.00509 0.145813 [
[(0)] .02372 0 0 2 0 Minner Mrotsboxa
[(5)] .18246 0 0 2 0 Minner Mrotsboxa
[(10)] .3412 0 0 2 0 Minner Mrotsboxa
[(15)] .49993 0 0 2 0 Minner Mrotsboxa
[(20)] .65867 0 0 2 0 Minner Mrotsboxa
[(25)] .81741 0 0 2 0 Minner Mrotsboxa
[(30)] .97615 0 0 2 0 Minner Mrotsboxa
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotsboxa
[(7)] -0.0125 .0156 1 0 0 Minner Mrotsboxa
[(8)] -0.0125 .16141 1 0 0 Minner Mrotsboxa
[(9)] -0.0125 .30722 1 0 0 Minner Mrotsboxa
[(10)] -0.0125 .45304 1 0 0 Minner Mrotsboxa
[(11)] -0.0125 .59885 1 0 0 Minner Mrotsboxa
[(percent)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotsboxa
[(7/19/1988, 74 knots,   4.1 parms, Delta)] .5 .61803 0 -2 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
[ .002 .01 ] 0 setdash
.002 w
.02372 0 m
.02372 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.18246 0 m
.18246 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.3412 0 m
.3412 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.49993 0 m
.49993 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.65867 0 m
.65867 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.81741 0 m
.81741 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.97615 0 m
.97615 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .0156 m
1 .0156 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .16141 m
1 .16141 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .30722 m
1 .30722 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .45304 m
1 .45304 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .59885 m
1 .59885 L
s
P
p
.002 w
.02372 0 m
.02372 .00625 L
s
P
[(0)] .02372 0 0 2 0 Minner Mrotshowa
p
.002 w
.18246 0 m
.18246 .00625 L
s
P
[(5)] .18246 0 0 2 0 Minner Mrotshowa
p
.002 w
.3412 0 m
.3412 .00625 L
s
P
[(10)] .3412 0 0 2 0 Minner Mrotshowa
p
.002 w
.49993 0 m
.49993 .00625 L
s
P
[(15)] .49993 0 0 2 0 Minner Mrotshowa
p
.002 w
.65867 0 m
.65867 .00625 L
s
P
[(20)] .65867 0 0 2 0 Minner Mrotshowa
p
.002 w
.81741 0 m
.81741 .00625 L
s
P
[(25)] .81741 0 0 2 0 Minner Mrotshowa
p
.002 w
.97615 0 m
.97615 .00625 L
s
P
[(30)] .97615 0 0 2 0 Minner Mrotshowa
p
.001 w
.05547 0 m
.05547 .00375 L
s
P
p
.001 w
.08722 0 m
.08722 .00375 L
s
P
p
.001 w
.11897 0 m
.11897 .00375 L
s
P
p
.001 w
.15071 0 m
.15071 .00375 L
s
P
p
.001 w
.21421 0 m
.21421 .00375 L
s
P
p
.001 w
.24595 0 m
.24595 .00375 L
s
P
p
.001 w
.2777 0 m
.2777 .00375 L
s
P
p
.001 w
.30945 0 m
.30945 .00375 L
s
P
p
.001 w
.37294 0 m
.37294 .00375 L
s
P
p
.001 w
.40469 0 m
.40469 .00375 L
s
P
p
.001 w
.43644 0 m
.43644 .00375 L
s
P
p
.001 w
.46819 0 m
.46819 .00375 L
s
P
p
.001 w
.53168 0 m
.53168 .00375 L
s
P
p
.001 w
.56343 0 m
.56343 .00375 L
s
P
p
.001 w
.59518 0 m
.59518 .00375 L
s
P
p
.001 w
.62692 0 m
.62692 .00375 L
s
P
p
.001 w
.69042 0 m
.69042 .00375 L
s
P
p
.001 w
.72217 0 m
.72217 .00375 L
s
P
p
.001 w
.75391 0 m
.75391 .00375 L
s
P
p
.001 w
.78566 0 m
.78566 .00375 L
s
P
p
.001 w
.84916 0 m
.84916 .00375 L
s
P
p
.001 w
.8809 0 m
.8809 .00375 L
s
P
p
.001 w
.91265 0 m
.91265 .00375 L
s
P
p
.001 w
.9444 0 m
.9444 .00375 L
s
P
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotshowa
p
.002 w
0 0 m
1 0 L
s
P
p
.002 w
0 .0156 m
.00625 .0156 L
s
P
[(7)] -0.0125 .0156 1 0 0 Minner Mrotshowa
p
.002 w
0 .16141 m
.00625 .16141 L
s
P
[(8)] -0.0125 .16141 1 0 0 Minner Mrotshowa
p
.002 w
0 .30722 m
.00625 .30722 L
s
P
[(9)] -0.0125 .30722 1 0 0 Minner Mrotshowa
p
.002 w
0 .45304 m
.00625 .45304 L
s
P
[(10)] -0.0125 .45304 1 0 0 Minner Mrotshowa
p
.002 w
0 .59885 m
.00625 .59885 L
s
P
[(11)] -0.0125 .59885 1 0 0 Minner Mrotshowa
p
.001 w
0 .04476 m
.00375 .04476 L
s
P
p
.001 w
0 .07392 m
.00375 .07392 L
s
P
p
.001 w
0 .10309 m
.00375 .10309 L
s
P
p
.001 w
0 .13225 m
.00375 .13225 L
s
P
p
.001 w
0 .19057 m
.00375 .19057 L
s
P
p
.001 w
0 .21974 m
.00375 .21974 L
s
P
p
.001 w
0 .2489 m
.00375 .2489 L
s
P
p
.001 w
0 .27806 m
.00375 .27806 L
s
P
p
.001 w
0 .33639 m
.00375 .33639 L
s
P
p
.001 w
0 .36555 m
.00375 .36555 L
s
P
p
.001 w
0 .39471 m
.00375 .39471 L
s
P
p
.001 w
0 .42387 m
.00375 .42387 L
s
P
p
.001 w
0 .4822 m
.00375 .4822 L
s
P
p
.001 w
0 .51136 m
.00375 .51136 L
s
P
p
.001 w
0 .54052 m
.00375 .54052 L
s
P
p
.001 w
0 .56969 m
.00375 .56969 L
s
P
[(percent)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotshowa
p
.002 w
0 0 m
0 .61803 L
s
P
P
p
[(7/19/1988, 74 knots,   4.1 parms, Delta)] .5 .61803 0 -2 Mshowa
P
p
P
p
p
[ .02 .01 ] 0 setdash
p
.004 w
.02381 .08113 m
.06349 .17068 L
.10317 .235 L
.14286 .28152 L
.18254 .32354 L
.20238 .34194 L
.22222 .35662 L
.2619 .3797 L
.28175 .39005 L
.29167 .39636 L
.30159 .40357 L
.31151 .41176 L
.31647 .41669 L
.32143 .42288 L
.33135 .43896 L
.34127 .45619 L
.38095 .5236 L
.40079 .55549 L
.41071 .5695 L
.42063 .5809 L
.43056 .58933 L
.44048 .59538 L
.44544 .59783 L
.4504 .59987 L
.45288 .60072 L
.45536 .60146 L
.45784 .60208 L
.46032 .60258 L
.46156 .60278 L
.4628 .60295 L
.46404 .60309 L
.46528 .6032 L
.46652 .60327 L
.46776 .60331 L
.469 .60332 L
.47024 .60329 L
.47148 .60323 L
.47272 .60313 L
.47396 .603 L
.4752 .60283 L
.47768 .60238 L
.48016 .6018 L
.48512 .60032 L
.49008 .59841 L
.5 .5933 L
.50992 .58639 L
.51984 .57789 L
.53968 .55681 L
.57937 .50011 L
.61905 .42624 L
Mistroke
.65873 .33862 L
.69841 .24936 L
.7381 .16303 L
.77778 .08629 L
.79762 .05545 L
.80754 .04269 L
.81746 .03201 L
.82738 .02366 L
.83234 .02044 L
.8373 .0179 L
.83978 .0169 L
.84226 .01608 L
.84474 .01545 L
.84598 .0152 L
.84722 .01501 L
.84846 .01486 L
.8497 .01476 L
.85094 .01472 L
.85218 .01472 L
.85342 .01477 L
.85466 .01488 L
.8559 .01504 L
.85714 .01525 L
.85962 .01584 L
.8621 .01665 L
.86706 .01894 L
.87698 .02535 L
.8869 .03344 L
.89683 .04351 L
.90675 .05651 L
.91667 .0723 L
.93651 .11153 L
.97619 .22123 L
Mfstroke
P
P
p
p
.004 w
.02381 .08102 m
.06349 .12786 L
.10317 .16629 L
.14286 .19724 L
.18254 .22362 L
.22222 .24713 L
.2619 .26744 L
.30159 .28503 L
.34127 .30261 L
.38095 .32344 L
.42063 .34653 L
.46032 .36905 L
.5 .38833 L
.51984 .39624 L
.53968 .40284 L
.55952 .40807 L
.56944 .41016 L
.57937 .41191 L
.58929 .41331 L
.59425 .41388 L
.59921 .41436 L
.60417 .41476 L
.60913 .41507 L
.61409 .41529 L
.61657 .41537 L
.61781 .4154 L
.61905 .41542 L
.62029 .41544 L
.62153 .41546 L
.62277 .41547 L
.62401 .41547 L
.62525 .41547 L
.62649 .41546 L
.62773 .41545 L
.62897 .41543 L
.63145 .41538 L
.63393 .4153 L
.63641 .41521 L
.63889 .41509 L
.64385 .41479 L
.64881 .41441 L
.65873 .4134 L
.66865 .41208 L
.67857 .41045 L
.69841 .40636 L
.7381 .39522 L
.77778 .38092 L
.81746 .36471 L
.85714 .34829 L
.89683 .33366 L
Mistroke
.91667 .32752 L
.93651 .32238 L
.95635 .3184 L
.96627 .31689 L
.97619 .31573 L
Mfstroke
P
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell["\<\
These curves are much more well-behaved. But recall that the average pricing \
error is large (for Treasury bonds). Let's look at the pricing errors:\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["MaturityPlot[PriceError[df2]];", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0237221 0.0319428 0.101612 0.0160239 [
[(0)] .02372 0 0 2 0 Minner Mrotsboxa
[(5)] .18344 0 0 2 0 Minner Mrotsboxa
[(10)] .34315 0 0 2 0 Minner Mrotsboxa
[(15)] .50286 0 0 2 0 Minner Mrotsboxa
[(20)] .66258 0 0 2 0 Minner Mrotsboxa
[(25)] .82229 0 0 2 0 Minner Mrotsboxa
[(30)] .98201 0 0 2 0 Minner Mrotsboxa
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotsboxa
[(-5)] -0.0125 .02149 1 0 0 Minner Mrotsboxa
[(0)] -0.0125 .10161 1 0 0 Minner Mrotsboxa
[(5)] -0.0125 .18173 1 0 0 Minner Mrotsboxa
[(10)] -0.0125 .26185 1 0 0 Minner Mrotsboxa
[(15)] -0.0125 .34197 1 0 0 Minner Mrotsboxa
[(20)] -0.0125 .42209 1 0 0 Minner Mrotsboxa
[(25)] -0.0125 .50221 1 0 0 Minner Mrotsboxa
[(30)] -0.0125 .58233 1 0 0 Minner Mrotsboxa
[(PriceError)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotsboxa
[(7/19/1988)] .5 .61803 0 -2 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
.002 w
.02372 0 m
.02372 .00625 L
s
P
[(0)] .02372 0 0 2 0 Minner Mrotshowa
p
.002 w
.18344 0 m
.18344 .00625 L
s
P
[(5)] .18344 0 0 2 0 Minner Mrotshowa
p
.002 w
.34315 0 m
.34315 .00625 L
s
P
[(10)] .34315 0 0 2 0 Minner Mrotshowa
p
.002 w
.50286 0 m
.50286 .00625 L
s
P
[(15)] .50286 0 0 2 0 Minner Mrotshowa
p
.002 w
.66258 0 m
.66258 .00625 L
s
P
[(20)] .66258 0 0 2 0 Minner Mrotshowa
p
.002 w
.82229 0 m
.82229 .00625 L
s
P
[(25)] .82229 0 0 2 0 Minner Mrotshowa
p
.002 w
.98201 0 m
.98201 .00625 L
s
P
[(30)] .98201 0 0 2 0 Minner Mrotshowa
p
.001 w
.05566 0 m
.05566 .00375 L
s
P
p
.001 w
.08761 0 m
.08761 .00375 L
s
P
p
.001 w
.11955 0 m
.11955 .00375 L
s
P
p
.001 w
.15149 0 m
.15149 .00375 L
s
P
p
.001 w
.21538 0 m
.21538 .00375 L
s
P
p
.001 w
.24732 0 m
.24732 .00375 L
s
P
p
.001 w
.27926 0 m
.27926 .00375 L
s
P
p
.001 w
.31121 0 m
.31121 .00375 L
s
P
p
.001 w
.37509 0 m
.37509 .00375 L
s
P
p
.001 w
.40704 0 m
.40704 .00375 L
s
P
p
.001 w
.43898 0 m
.43898 .00375 L
s
P
p
.001 w
.47092 0 m
.47092 .00375 L
s
P
p
.001 w
.53481 0 m
.53481 .00375 L
s
P
p
.001 w
.56675 0 m
.56675 .00375 L
s
P
p
.001 w
.59869 0 m
.59869 .00375 L
s
P
p
.001 w
.63064 0 m
.63064 .00375 L
s
P
p
.001 w
.69452 0 m
.69452 .00375 L
s
P
p
.001 w
.72646 0 m
.72646 .00375 L
s
P
p
.001 w
.75841 0 m
.75841 .00375 L
s
P
p
.001 w
.79035 0 m
.79035 .00375 L
s
P
p
.001 w
.85423 0 m
.85423 .00375 L
s
P
p
.001 w
.88618 0 m
.88618 .00375 L
s
P
p
.001 w
.91812 0 m
.91812 .00375 L
s
P
p
.001 w
.95006 0 m
.95006 .00375 L
s
P
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotshowa
p
.002 w
0 0 m
1 0 L
s
P
p
.002 w
0 .02149 m
.00625 .02149 L
s
P
[(-5)] -0.0125 .02149 1 0 0 Minner Mrotshowa
p
.002 w
0 .10161 m
.00625 .10161 L
s
P
[(0)] -0.0125 .10161 1 0 0 Minner Mrotshowa
p
.002 w
0 .18173 m
.00625 .18173 L
s
P
[(5)] -0.0125 .18173 1 0 0 Minner Mrotshowa
p
.002 w
0 .26185 m
.00625 .26185 L
s
P
[(10)] -0.0125 .26185 1 0 0 Minner Mrotshowa
p
.002 w
0 .34197 m
.00625 .34197 L
s
P
[(15)] -0.0125 .34197 1 0 0 Minner Mrotshowa
p
.002 w
0 .42209 m
.00625 .42209 L
s
P
[(20)] -0.0125 .42209 1 0 0 Minner Mrotshowa
p
.002 w
0 .50221 m
.00625 .50221 L
s
P
[(25)] -0.0125 .50221 1 0 0 Minner Mrotshowa
p
.002 w
0 .58233 m
.00625 .58233 L
s
P
[(30)] -0.0125 .58233 1 0 0 Minner Mrotshowa
p
.001 w
0 .03752 m
.00375 .03752 L
s
P
p
.001 w
0 .05354 m
.00375 .05354 L
s
P
p
.001 w
0 .06956 m
.00375 .06956 L
s
P
p
.001 w
0 .08559 m
.00375 .08559 L
s
P
p
.001 w
0 .11764 m
.00375 .11764 L
s
P
p
.001 w
0 .13366 m
.00375 .13366 L
s
P
p
.001 w
0 .14968 m
.00375 .14968 L
s
P
p
.001 w
0 .16571 m
.00375 .16571 L
s
P
p
.001 w
0 .19776 m
.00375 .19776 L
s
P
p
.001 w
0 .21378 m
.00375 .21378 L
s
P
p
.001 w
0 .2298 m
.00375 .2298 L
s
P
p
.001 w
0 .24583 m
.00375 .24583 L
s
P
p
.001 w
0 .27787 m
.00375 .27787 L
s
P
p
.001 w
0 .2939 m
.00375 .2939 L
s
P
p
.001 w
0 .30992 m
.00375 .30992 L
s
P
p
.001 w
0 .32595 m
.00375 .32595 L
s
P
p
.001 w
0 .35799 m
.00375 .35799 L
s
P
p
.001 w
0 .37402 m
.00375 .37402 L
s
P
p
.001 w
0 .39004 m
.00375 .39004 L
s
P
p
.001 w
0 .40607 m
.00375 .40607 L
s
P
p
.001 w
0 .43811 m
.00375 .43811 L
s
P
p
.001 w
0 .45414 m
.00375 .45414 L
s
P
p
.001 w
0 .47016 m
.00375 .47016 L
s
P
p
.001 w
0 .48618 m
.00375 .48618 L
s
P
p
.001 w
0 .51823 m
.00375 .51823 L
s
P
p
.001 w
0 .53426 m
.00375 .53426 L
s
P
p
.001 w
0 .55028 m
.00375 .55028 L
s
P
p
.001 w
0 .5663 m
.00375 .5663 L
s
P
p
.001 w
0 .00547 m
.00375 .00547 L
s
P
p
.001 w
0 .59835 m
.00375 .59835 L
s
P
p
.001 w
0 .61438 m
.00375 .61438 L
s
P
[(PriceError)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotshowa
p
.002 w
0 0 m
0 .61803 L
s
P
P
p
[(7/19/1988)] .5 .61803 0 -2 Mshowa
P
p
P
p
.008 w
.02381 .10167 Mdot
.02442 .10217 Mdot
.02503 .10222 Mdot
.02565 .10281 Mdot
.02626 .10342 Mdot
.02687 .10352 Mdot
.0274 .10255 Mdot
.02748 .10326 Mdot
.02809 .10355 Mdot
.02871 .10387 Mdot
.02932 .10366 Mdot
.02993 .10379 Mdot
.03002 .10263 Mdot
.03002 .10279 Mdot
.03054 .10414 Mdot
.03116 .10433 Mdot
.03151 .10521 Mdot
.03238 .10421 Mdot
.03273 .10242 Mdot
.03299 .10415 Mdot
.0336 .10422 Mdot
.03404 .10215 Mdot
.03404 .10227 Mdot
.03404 .10275 Mdot
.03422 .10416 Mdot
.03492 .10405 Mdot
.03535 .1025 Mdot
.03544 .10428 Mdot
.03605 .10448 Mdot
.03667 .10435 Mdot
.03728 .10414 Mdot
.03789 .10461 Mdot
.03833 .10254 Mdot
.03833 .10261 Mdot
.0385 .10421 Mdot
.03911 .10408 Mdot
.03955 .10458 Mdot
.04078 .10201 Mdot
.04209 .10135 Mdot
.04209 .10143 Mdot
.04217 .1036 Mdot
.04322 .10105 Mdot
.04462 .10352 Mdot
.04594 .1004 Mdot
.04594 .10115 Mdot
.04707 .10359 Mdot
.04742 .10596 Mdot
.04865 .10012 Mdot
.04952 .10349 Mdot
.04987 .09961 Mdot
.04987 .09978 Mdot
.04987 .10049 Mdot
.05127 .09991 Mdot
.05197 .10376 Mdot
.05389 .09975 Mdot
.05389 .10004 Mdot
.05442 .10294 Mdot
.05538 .10472 Mdot
.05661 .099 Mdot
.05792 .09877 Mdot
.05792 .10033 Mdot
.05932 .0991 Mdot
.06211 .0987 Mdot
.06211 .09888 Mdot
.06334 .09803 Mdot
.06465 .09815 Mdot
.06596 .09759 Mdot
.06596 .09841 Mdot
.06596 .09951 Mdot
.06727 .09779 Mdot
.07016 .09754 Mdot
.07016 .09781 Mdot
.07138 .09772 Mdot
.0727 .09698 Mdot
.07401 .09675 Mdot
.07401 .09715 Mdot
.07401 .11579 Mdot
.07515 .09612 Mdot
.07803 .09698 Mdot
.07803 .09737 Mdot
.07926 .09514 Mdot
.08048 .0954 Mdot
.08179 .09418 Mdot
.08179 .09537 Mdot
.08179 .10039 Mdot
.08319 .09522 Mdot
.08599 .09565 Mdot
.08599 .09588 Mdot
.08721 .09481 Mdot
.08984 .09371 Mdot
.08984 .09397 Mdot
.08984 .09468 Mdot
.09395 .0945 Mdot
.09517 .09248 Mdot
.09788 .09274 Mdot
.09788 .09355 Mdot
.09788 .09411 Mdot
.10191 .09429 Mdot
.10322 .09246 Mdot
.10593 .09292 Mdot
.10593 .09427 Mdot
.10986 .09316 Mdot
.11109 .09092 Mdot
.11371 .09266 Mdot
.11371 .09745 Mdot
.11782 .09298 Mdot
.11905 .0903 Mdot
.12176 .09219 Mdot
.12176 .09562 Mdot
.12578 .09219 Mdot
.12709 .08984 Mdot
.1298 .09231 Mdot
.1298 .09233 Mdot
.13383 .09347 Mdot
.13514 .09037 Mdot
.13811 .06439 Mdot
.13811 .09303 Mdot
.13811 .09356 Mdot
.14179 .09189 Mdot
.1431 .08839 Mdot
.14572 .08816 Mdot
.14572 .09112 Mdot
.14974 .09215 Mdot
.15106 .09 Mdot
.15394 .09195 Mdot
.15394 .10886 Mdot
.15394 .22971 Mdot
.1591 .08949 Mdot
.1619 .08964 Mdot
.1619 .09159 Mdot
.16715 .09137 Mdot
.16995 .08945 Mdot
.16995 .09204 Mdot
.16995 .09798 Mdot
.16995 .09881 Mdot
.16995 .27814 Mdot
.17502 .09075 Mdot
.17782 .08858 Mdot
.17782 .08981 Mdot
.18298 .0914 Mdot
.18578 .08663 Mdot
.18578 .09194 Mdot
.18578 .09759 Mdot
.19102 .09002 Mdot
.19373 .08608 Mdot
.19373 .0968 Mdot
.19373 .30774 Mdot
.19933 .09052 Mdot
.20178 .09124 Mdot
.20694 .0886 Mdot
.20965 .03709 Mdot
.20965 .08356 Mdot
.2149 .08771 Mdot
.21761 .08302 Mdot
.21761 .09836 Mdot
.22312 .08559 Mdot
.22565 .0821 Mdot
.22565 .08575 Mdot
.23116 .08654 Mdot
.2337 .01472 Mdot
.2337 .08157 Mdot
.2337 .08307 Mdot
.2337 .46162 Mdot
.23904 .0848 Mdot
.24148 .08035 Mdot
.24148 .08319 Mdot
.24148 .08374 Mdot
.24699 .08682 Mdot
.24953 .03571 Mdot
.24953 .08098 Mdot
.25758 .0831 Mdot
.25758 .08557 Mdot
.26562 .08368 Mdot
.27349 .08422 Mdot
.28154 .04775 Mdot
.28958 .08502 Mdot
.30541 .08214 Mdot
.31346 .08257 Mdot
.32168 .0826 Mdot
.32973 .08622 Mdot
.33733 .08951 Mdot
.35351 .60332 Mdot
.40126 .063 Mdot
.4254 .09424 Mdot
.43318 .09088 Mdot
.44123 .09162 Mdot
.44928 .09569 Mdot
.45732 .0682 Mdot
.45732 .09654 Mdot
.4812 .07919 Mdot
.4812 .09521 Mdot
.48951 .09741 Mdot
.49703 .09803 Mdot
.50507 .08406 Mdot
.50507 .09757 Mdot
.51329 .08709 Mdot
.51329 .09784 Mdot
.52921 .09342 Mdot
.52921 .09961 Mdot
.53717 .10401 Mdot
.54513 .09038 Mdot
.54513 .10374 Mdot
.55317 .08965 Mdot
.56104 .09519 Mdot
.56104 .10704 Mdot
.569 .10844 Mdot
.57705 .08951 Mdot
.58509 .12806 Mdot
.59288 .08842 Mdot
.60897 .09128 Mdot
.64089 .0991 Mdot
.66485 .09768 Mdot
.68873 .09409 Mdot
.69695 .09589 Mdot
.70491 .11156 Mdot
.87273 .10788 Mdot
.88856 .10739 Mdot
.89652 .10504 Mdot
.90457 .10719 Mdot
.91244 .10553 Mdot
.92844 .10571 Mdot
.94427 .1051 Mdot
.95232 .1083 Mdot
.97619 .12357 Mdot
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell[TextData[{
  "There are a few enormous outliers. What is wrong with these securities? \
The answer is that we have included some callable bonds and some flower bonds \
in the sample. The prices of these bonds reflect features that are not \
captured in the present value function directly. Including them in the \
estimation will affect the results in two ways. First, their inclusion will \
affect the value of $\\beta^* (\\lambda^*)$, the estimated spline \
coefficients, conditional on the value of $\\lambda$. But perhaps more \
importantly, their inclusion will affect the value of $\\lambda^*$ chosen via \
GCV. This is because minimizing the GCV function is a signal-extraction \
technique: When there is more noise in the data, a larger penalty will be \
chosen for a smoother curve at the expense of the sum-of-squares fit. \n\nThe \
solution to this problem is to remove the securities with special features \
from the estimation. We can do that with ",
  StyleBox["IDNumberDropList",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", an option to ",
  StyleBox["DiscountFunction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". The file ",
  StyleBox["iddrop.m",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " contains a list of CUSIPs of the securities with special features."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
iddrop = << iddrop.m;
df3 = DiscountFunction[cp, FunctionalForm -> Delta,
\tLambdaValue -> 10^10, IDNumberDropList -> iddrop];\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["Plot[GCVFunction[10^x],{x, 8, 14}];", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-1.24603 0.15873 -0.303212 10.0675 [
[(9)] .18254 -0.00119 0 2 Msboxa
[(10)] .34127 -0.00119 0 2 Msboxa
[(11)] .5 -0.00119 0 2 Msboxa
[(12)] .65873 -0.00119 0 2 Msboxa
[(13)] .81746 -0.00119 0 2 Msboxa
[(14)] .97619 -0.00119 0 2 Msboxa
[(0.04)] .01131 .09949 1 0 Msboxa
[(0.05)] .01131 .20016 1 0 Msboxa
[(0.06)] .01131 .30084 1 0 Msboxa
[(0.07)] .01131 .40151 1 0 Msboxa
[(0.08)] .01131 .50219 1 0 Msboxa
[(0.09)] .01131 .60286 1 0 Msboxa
[ -0.001 -0.00219 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
.002 w
.18254 -0.00119 m
.18254 .00506 L
s
P
[(9)] .18254 -0.00119 0 2 Mshowa
p
.002 w
.34127 -0.00119 m
.34127 .00506 L
s
P
[(10)] .34127 -0.00119 0 2 Mshowa
p
.002 w
.5 -0.00119 m
.5 .00506 L
s
P
[(11)] .5 -0.00119 0 2 Mshowa
p
.002 w
.65873 -0.00119 m
.65873 .00506 L
s
P
[(12)] .65873 -0.00119 0 2 Mshowa
p
.002 w
.81746 -0.00119 m
.81746 .00506 L
s
P
[(13)] .81746 -0.00119 0 2 Mshowa
p
.002 w
.97619 -0.00119 m
.97619 .00506 L
s
P
[(14)] .97619 -0.00119 0 2 Mshowa
p
.001 w
.05556 -0.00119 m
.05556 .00256 L
s
P
p
.001 w
.0873 -0.00119 m
.0873 .00256 L
s
P
p
.001 w
.11905 -0.00119 m
.11905 .00256 L
s
P
p
.001 w
.15079 -0.00119 m
.15079 .00256 L
s
P
p
.001 w
.21429 -0.00119 m
.21429 .00256 L
s
P
p
.001 w
.24603 -0.00119 m
.24603 .00256 L
s
P
p
.001 w
.27778 -0.00119 m
.27778 .00256 L
s
P
p
.001 w
.30952 -0.00119 m
.30952 .00256 L
s
P
p
.001 w
.37302 -0.00119 m
.37302 .00256 L
s
P
p
.001 w
.40476 -0.00119 m
.40476 .00256 L
s
P
p
.001 w
.43651 -0.00119 m
.43651 .00256 L
s
P
p
.001 w
.46825 -0.00119 m
.46825 .00256 L
s
P
p
.001 w
.53175 -0.00119 m
.53175 .00256 L
s
P
p
.001 w
.56349 -0.00119 m
.56349 .00256 L
s
P
p
.001 w
.59524 -0.00119 m
.59524 .00256 L
s
P
p
.001 w
.62698 -0.00119 m
.62698 .00256 L
s
P
p
.001 w
.69048 -0.00119 m
.69048 .00256 L
s
P
p
.001 w
.72222 -0.00119 m
.72222 .00256 L
s
P
p
.001 w
.75397 -0.00119 m
.75397 .00256 L
s
P
p
.001 w
.78571 -0.00119 m
.78571 .00256 L
s
P
p
.001 w
.84921 -0.00119 m
.84921 .00256 L
s
P
p
.001 w
.88095 -0.00119 m
.88095 .00256 L
s
P
p
.001 w
.9127 -0.00119 m
.9127 .00256 L
s
P
p
.001 w
.94444 -0.00119 m
.94444 .00256 L
s
P
p
.002 w
0 -0.00119 m
1 -0.00119 L
s
P
p
.002 w
.02381 .09949 m
.03006 .09949 L
s
P
[(0.04)] .01131 .09949 1 0 Mshowa
p
.002 w
.02381 .20016 m
.03006 .20016 L
s
P
[(0.05)] .01131 .20016 1 0 Mshowa
p
.002 w
.02381 .30084 m
.03006 .30084 L
s
P
[(0.06)] .01131 .30084 1 0 Mshowa
p
.002 w
.02381 .40151 m
.03006 .40151 L
s
P
[(0.07)] .01131 .40151 1 0 Mshowa
p
.002 w
.02381 .50219 m
.03006 .50219 L
s
P
[(0.08)] .01131 .50219 1 0 Mshowa
p
.002 w
.02381 .60286 m
.03006 .60286 L
s
P
[(0.09)] .01131 .60286 1 0 Mshowa
p
.001 w
.02381 .01895 m
.02756 .01895 L
s
P
p
.001 w
.02381 .03908 m
.02756 .03908 L
s
P
p
.001 w
.02381 .05922 m
.02756 .05922 L
s
P
p
.001 w
.02381 .07935 m
.02756 .07935 L
s
P
p
.001 w
.02381 .11962 m
.02756 .11962 L
s
P
p
.001 w
.02381 .13976 m
.02756 .13976 L
s
P
p
.001 w
.02381 .15989 m
.02756 .15989 L
s
P
p
.001 w
.02381 .18003 m
.02756 .18003 L
s
P
p
.001 w
.02381 .2203 m
.02756 .2203 L
s
P
p
.001 w
.02381 .24043 m
.02756 .24043 L
s
P
p
.001 w
.02381 .26057 m
.02756 .26057 L
s
P
p
.001 w
.02381 .2807 m
.02756 .2807 L
s
P
p
.001 w
.02381 .32097 m
.02756 .32097 L
s
P
p
.001 w
.02381 .34111 m
.02756 .34111 L
s
P
p
.001 w
.02381 .36124 m
.02756 .36124 L
s
P
p
.001 w
.02381 .38138 m
.02756 .38138 L
s
P
p
.001 w
.02381 .42165 m
.02756 .42165 L
s
P
p
.001 w
.02381 .44178 m
.02756 .44178 L
s
P
p
.001 w
.02381 .46192 m
.02756 .46192 L
s
P
p
.001 w
.02381 .48205 m
.02756 .48205 L
s
P
p
.001 w
.02381 .52232 m
.02756 .52232 L
s
P
p
.001 w
.02381 .54246 m
.02756 .54246 L
s
P
p
.001 w
.02381 .56259 m
.02756 .56259 L
s
P
p
.001 w
.02381 .58272 m
.02756 .58272 L
s
P
p
.002 w
.02381 0 m
.02381 .61803 L
s
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
p
p
.004 w
.02381 .36639 m
.06349 .30686 L
.10317 .25389 L
.14286 .20745 L
.18254 .16708 L
.22222 .13204 L
.2619 .10173 L
.30159 .07598 L
.34127 .05499 L
.36111 .04638 L
.38095 .03902 L
.40079 .03288 L
.42063 .02785 L
.44048 .0238 L
.46032 .02058 L
.48016 .01809 L
.49008 .01709 L
.5 .01625 L
.50992 .01558 L
.51488 .01531 L
.51984 .01509 L
.5248 .01492 L
.52728 .01485 L
.52976 .0148 L
.531 .01477 L
.53224 .01476 L
.53348 .01474 L
.53472 .01473 L
.53596 .01472 L
.5372 .01472 L
.53844 .01472 L
.53968 .01472 L
.54092 .01472 L
.54216 .01473 L
.5434 .01475 L
.54464 .01477 L
.54712 .01481 L
.5496 .01488 L
.55456 .01506 L
.55952 .01531 L
.56448 .01564 L
.56944 .01605 L
.57937 .01712 L
.58929 .01857 L
.59921 .02044 L
.61905 .02556 L
.63889 .03275 L
.65873 .04217 L
.69841 .06727 L
.7381 .0977 L
Mistroke
.77778 .12847 L
.81746 .15742 L
.8373 .1723 L
.85714 .18914 L
.87698 .21012 L
.8869 .22318 L
.89683 .23869 L
.90675 .25742 L
.91667 .2803 L
.92659 .30855 L
.93651 .34372 L
.94643 .38782 L
.95635 .4434 L
.96627 .5138 L
.97619 .60332 L
Mfstroke
P
P
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell["\<\
Now the minimum appears to be between $10^{11}$ and $10^{12}$.\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
df4 = DiscountFunction[cp, FunctionalForm -> Delta,
\tLambda0 -> {10^11.5, 1.1 10^11.5},
\tIDNumberDropList -> iddrop];
DFSummaryStatistics[df4]
ShowGraph[df4, PlotStyle -> {Dashing[{.02, .01}],{}}];\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
TableForm[{PaddedForm[196, {3, 0}, NumberPoint -> \"\"], 
 
   StringForm[\"`1` `2`\", 
 
    PaddedForm[29.81793292265571, {5, 2}], \"years\"], 
 
   PaddedForm[0, {3, 0}, NumberPoint -> \"\"], Delta, True, 
 
   True, PaddedForm[2, {3, 0}, NumberPoint -> \"\"], 
 
   PaddedForm[65, {3, 0}, NumberPoint -> \"\"], 
 
   PaddedForm[AccountingForm[15.37050066522184, 
 
     NumberSigns -> {\"-\", \"\"}, NumberPadding -> {\" \", \"0\"}]\\
 
     , {4, 1}], PaddedForm[1.737711397395383*10^11, {5, 2}, 
 
    NumberFormat -> (#1\"e\"#3 & )], 
 
   PaddedForm[AccountingForm[0.02436085555230557, 
 
     NumberSigns -> {\"-\", \"\"}, NumberPadding -> {\" \", \"0\"}]\\
 
     , {8, 5}], PaddedForm[AccountingForm[0.0989824198044\\
 
      96, NumberSigns -> {\"-\", \"\"}, 
 
     NumberPadding -> {\" \", \"0\"}], {7, 4}]}, 
 
  TableSpacing -> {0}, TableHeadings -> 
 
   {{\"no. of obs.........\", \"longest payment....\", \"no.\\
 
      dropped........\", \"functional form....\", \"restriction\\
 
      ........\", \"minimize GCV.......\", \"tuning\\
 
      parameter...\", \"no. of knots.......\", \"no. of\\
 
      parameters..\", \"lambda.............\", \"residual\\
 
      variance..\", \"avg. abs. error....\"}}]\
\>", "\<\

no. of obs.........   196
longest payment....    29.82 years
no. dropped........     0
functional form....  Delta
restriction........  True
minimize GCV.......  True
tuning parameter...     2
no. of knots.......    65
no. of parameters..    15.4
lambda.............     1.74e11
residual variance..     0.02436
avg. abs. error....     0.0990\
\>"], "Output",
  LineSpacing->{1, 0},
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0237226 0.0317475 0.126588 0.042754 [
[(0)] .02372 0 0 2 0 Minner Mrotsboxa
[(5)] .18246 0 0 2 0 Minner Mrotsboxa
[(10)] .3412 0 0 2 0 Minner Mrotsboxa
[(15)] .49993 0 0 2 0 Minner Mrotsboxa
[(20)] .65867 0 0 2 0 Minner Mrotsboxa
[(25)] .81741 0 0 2 0 Minner Mrotsboxa
[(30)] .97615 0 0 2 0 Minner Mrotsboxa
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotsboxa
[(-2)] -0.0125 .04108 1 0 0 Minner Mrotsboxa
[(0)] -0.0125 .12659 1 0 0 Minner Mrotsboxa
[(2)] -0.0125 .2121 1 0 0 Minner Mrotsboxa
[(4)] -0.0125 .2976 1 0 0 Minner Mrotsboxa
[(6)] -0.0125 .38311 1 0 0 Minner Mrotsboxa
[(8)] -0.0125 .46862 1 0 0 Minner Mrotsboxa
[(10)] -0.0125 .55413 1 0 0 Minner Mrotsboxa
[(percent)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotsboxa
[(7/19/1988, 65 knots,  15.4 parms, Delta)] .5 .61803 0 -2 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
[ .002 .01 ] 0 setdash
.002 w
.02372 0 m
.02372 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.18246 0 m
.18246 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.3412 0 m
.3412 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.49993 0 m
.49993 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.65867 0 m
.65867 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.81741 0 m
.81741 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.97615 0 m
.97615 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .04108 m
1 .04108 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .12659 m
1 .12659 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .2121 m
1 .2121 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .2976 m
1 .2976 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .38311 m
1 .38311 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .46862 m
1 .46862 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .55413 m
1 .55413 L
s
P
p
.002 w
.02372 0 m
.02372 .00625 L
s
P
[(0)] .02372 0 0 2 0 Minner Mrotshowa
p
.002 w
.18246 0 m
.18246 .00625 L
s
P
[(5)] .18246 0 0 2 0 Minner Mrotshowa
p
.002 w
.3412 0 m
.3412 .00625 L
s
P
[(10)] .3412 0 0 2 0 Minner Mrotshowa
p
.002 w
.49993 0 m
.49993 .00625 L
s
P
[(15)] .49993 0 0 2 0 Minner Mrotshowa
p
.002 w
.65867 0 m
.65867 .00625 L
s
P
[(20)] .65867 0 0 2 0 Minner Mrotshowa
p
.002 w
.81741 0 m
.81741 .00625 L
s
P
[(25)] .81741 0 0 2 0 Minner Mrotshowa
p
.002 w
.97615 0 m
.97615 .00625 L
s
P
[(30)] .97615 0 0 2 0 Minner Mrotshowa
p
.001 w
.05547 0 m
.05547 .00375 L
s
P
p
.001 w
.08722 0 m
.08722 .00375 L
s
P
p
.001 w
.11897 0 m
.11897 .00375 L
s
P
p
.001 w
.15071 0 m
.15071 .00375 L
s
P
p
.001 w
.21421 0 m
.21421 .00375 L
s
P
p
.001 w
.24595 0 m
.24595 .00375 L
s
P
p
.001 w
.2777 0 m
.2777 .00375 L
s
P
p
.001 w
.30945 0 m
.30945 .00375 L
s
P
p
.001 w
.37294 0 m
.37294 .00375 L
s
P
p
.001 w
.40469 0 m
.40469 .00375 L
s
P
p
.001 w
.43644 0 m
.43644 .00375 L
s
P
p
.001 w
.46819 0 m
.46819 .00375 L
s
P
p
.001 w
.53168 0 m
.53168 .00375 L
s
P
p
.001 w
.56343 0 m
.56343 .00375 L
s
P
p
.001 w
.59518 0 m
.59518 .00375 L
s
P
p
.001 w
.62692 0 m
.62692 .00375 L
s
P
p
.001 w
.69042 0 m
.69042 .00375 L
s
P
p
.001 w
.72217 0 m
.72217 .00375 L
s
P
p
.001 w
.75391 0 m
.75391 .00375 L
s
P
p
.001 w
.78566 0 m
.78566 .00375 L
s
P
p
.001 w
.84916 0 m
.84916 .00375 L
s
P
p
.001 w
.8809 0 m
.8809 .00375 L
s
P
p
.001 w
.91265 0 m
.91265 .00375 L
s
P
p
.001 w
.9444 0 m
.9444 .00375 L
s
P
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotshowa
p
.002 w
0 0 m
1 0 L
s
P
p
.002 w
0 .04108 m
.00625 .04108 L
s
P
[(-2)] -0.0125 .04108 1 0 0 Minner Mrotshowa
p
.002 w
0 .12659 m
.00625 .12659 L
s
P
[(0)] -0.0125 .12659 1 0 0 Minner Mrotshowa
p
.002 w
0 .2121 m
.00625 .2121 L
s
P
[(2)] -0.0125 .2121 1 0 0 Minner Mrotshowa
p
.002 w
0 .2976 m
.00625 .2976 L
s
P
[(4)] -0.0125 .2976 1 0 0 Minner Mrotshowa
p
.002 w
0 .38311 m
.00625 .38311 L
s
P
[(6)] -0.0125 .38311 1 0 0 Minner Mrotshowa
p
.002 w
0 .46862 m
.00625 .46862 L
s
P
[(8)] -0.0125 .46862 1 0 0 Minner Mrotshowa
p
.002 w
0 .55413 m
.00625 .55413 L
s
P
[(10)] -0.0125 .55413 1 0 0 Minner Mrotshowa
p
.001 w
0 .05818 m
.00375 .05818 L
s
P
p
.001 w
0 .07528 m
.00375 .07528 L
s
P
p
.001 w
0 .09238 m
.00375 .09238 L
s
P
p
.001 w
0 .10949 m
.00375 .10949 L
s
P
p
.001 w
0 .14369 m
.00375 .14369 L
s
P
p
.001 w
0 .16079 m
.00375 .16079 L
s
P
p
.001 w
0 .17789 m
.00375 .17789 L
s
P
p
.001 w
0 .19499 m
.00375 .19499 L
s
P
p
.001 w
0 .2292 m
.00375 .2292 L
s
P
p
.001 w
0 .2463 m
.00375 .2463 L
s
P
p
.001 w
0 .2634 m
.00375 .2634 L
s
P
p
.001 w
0 .2805 m
.00375 .2805 L
s
P
p
.001 w
0 .31471 m
.00375 .31471 L
s
P
p
.001 w
0 .33181 m
.00375 .33181 L
s
P
p
.001 w
0 .34891 m
.00375 .34891 L
s
P
p
.001 w
0 .36601 m
.00375 .36601 L
s
P
p
.001 w
0 .40021 m
.00375 .40021 L
s
P
p
.001 w
0 .41732 m
.00375 .41732 L
s
P
p
.001 w
0 .43442 m
.00375 .43442 L
s
P
p
.001 w
0 .45152 m
.00375 .45152 L
s
P
p
.001 w
0 .48572 m
.00375 .48572 L
s
P
p
.001 w
0 .50282 m
.00375 .50282 L
s
P
p
.001 w
0 .51992 m
.00375 .51992 L
s
P
p
.001 w
0 .53703 m
.00375 .53703 L
s
P
p
.001 w
0 .02398 m
.00375 .02398 L
s
P
p
.001 w
0 .00688 m
.00375 .00688 L
s
P
p
.001 w
0 .57123 m
.00375 .57123 L
s
P
p
.001 w
0 .58833 m
.00375 .58833 L
s
P
p
.001 w
0 .60543 m
.00375 .60543 L
s
P
[(percent)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotshowa
p
.002 w
0 0 m
0 .61803 L
s
P
P
p
[(7/19/1988, 65 knots,  15.4 parms, Delta)] .5 .61803 0 -2 Mshowa
P
p
P
p
p
[ .02 .01 ] 0 setdash
p
.004 w
.02381 .4334 m
.03373 .44826 L
.04365 .4675 L
.04861 .47511 L
.05357 .48125 L
.05605 .48392 L
.05853 .48618 L
.06101 .48802 L
.06349 .48948 L
.06845 .49171 L
.07341 .49347 L
.07837 .49484 L
.08085 .49538 L
.08333 .49585 L
.08581 .49624 L
.08829 .49652 L
.08953 .49662 L
.09077 .49669 L
.09201 .49673 L
.09325 .49673 L
.09573 .49667 L
.09697 .49664 L
.09821 .49664 L
.09945 .49665 L
.10069 .49668 L
.10193 .49673 L
.10317 .49681 L
.10565 .49707 L
.10813 .49742 L
.1131 .49836 L
.11806 .49955 L
.12054 .50014 L
.12302 .50062 L
.1255 .50097 L
.12798 .50125 L
.13294 .50166 L
.13418 .50175 L
.13542 .50187 L
.13666 .50203 L
.1379 .5022 L
.13914 .50232 L
.14038 .50237 L
.14162 .50233 L
.14286 .50224 L
.1441 .50225 L
.14534 .50239 L
.14658 .50268 L
.14782 .50311 L
.15278 .50603 L
.1627 .5128 L
Mistroke
.18254 .52606 L
.19246 .53314 L
.20238 .53942 L
.20734 .54192 L
.2123 .54417 L
.21478 .54518 L
.21726 .54599 L
.2185 .54631 L
.21974 .54657 L
.22098 .54677 L
.22222 .5469 L
.22346 .54695 L
.2247 .54689 L
.22594 .54674 L
.22718 .54648 L
.23214 .54456 L
.2371 .54228 L
.23958 .5411 L
.24206 .53998 L
.24454 .5391 L
.24578 .53877 L
.24702 .53849 L
.24826 .53829 L
.2495 .53813 L
.25074 .538 L
.25198 .53789 L
.25322 .53781 L
.25446 .53776 L
.2557 .53773 L
.25694 .53773 L
.25818 .53776 L
.25942 .53781 L
.2619 .53802 L
.26438 .53834 L
.26687 .53866 L
.26811 .53879 L
.26935 .5389 L
.27059 .53899 L
.27183 .53906 L
.27307 .53911 L
.27431 .53914 L
.27555 .53915 L
.27679 .53913 L
.27803 .53909 L
.27927 .53903 L
.28051 .53895 L
.28175 .53884 L
.28423 .53855 L
.28671 .53817 L
.29167 .5371 L
Mistroke
.29663 .53561 L
.30159 .53368 L
.31151 .5291 L
.31647 .52691 L
.31895 .52586 L
.32019 .52535 L
.32143 .5249 L
.32267 .52454 L
.32391 .52427 L
.32515 .5241 L
.32639 .52402 L
.32763 .52403 L
.32887 .52414 L
.33011 .52435 L
.33135 .52466 L
.33259 .52506 L
.33383 .52558 L
.33631 .52688 L
.34127 .52964 L
.38095 .55014 L
.40079 .55884 L
.42063 .56599 L
.43056 .56884 L
.43304 .56956 L
.43552 .57049 L
.44048 .57295 L
.44544 .57628 L
.4504 .58053 L
.45536 .58575 L
.45784 .58827 L
.46032 .59043 L
.4628 .59221 L
.46528 .5936 L
.46652 .59414 L
.46776 .59458 L
.469 .59492 L
.47024 .59515 L
.47148 .59527 L
.47272 .59527 L
.47396 .59517 L
.4752 .59495 L
.47644 .59462 L
.47768 .59416 L
.48016 .5929 L
.48264 .59113 L
.48512 .58886 L
.49008 .58299 L
.5 .56813 L
.53968 .48582 L
.55952 .4255 L
Mistroke
.562 .41822 L
.56448 .41209 L
.56696 .40715 L
.5682 .40514 L
.56944 .40343 L
.57068 .40204 L
.57192 .40096 L
.57316 .40019 L
.5744 .39975 L
.57564 .39963 L
.57688 .39984 L
.57813 .40038 L
.57937 .40125 L
.61905 .45318 L
.65873 .50302 L
.69841 .54811 L
.71825 .56761 L
.7381 .58402 L
.74802 .59073 L
.75794 .59621 L
.7629 .59843 L
.76786 .60027 L
.77282 .60171 L
.7753 .60226 L
.77778 .60271 L
.77902 .60288 L
.78026 .60303 L
.7815 .60315 L
.78274 .60324 L
.78398 .60329 L
.78522 .60332 L
.78646 .60331 L
.7877 .60327 L
.78894 .6032 L
.79018 .60309 L
.79266 .60277 L
.7939 .60256 L
.79514 .60231 L
.79762 .6017 L
.80258 .60003 L
.80754 .59772 L
.8125 .59473 L
.81746 .59102 L
.82738 .58129 L
.8373 .56819 L
.84722 .55142 L
.85714 .53065 L
.87698 .47608 L
.89683 .40642 L
.93651 .22177 L
Mistroke
.97619 .01472 L
Mfstroke
P
P
p
p
.004 w
.02381 .43337 m
.04365 .4489 L
.05357 .45752 L
.06349 .4646 L
.07341 .47 L
.08333 .47412 L
.09325 .47731 L
.10317 .47973 L
.12302 .48348 L
.14286 .48651 L
.18254 .49316 L
.22222 .50221 L
.23214 .5043 L
.24206 .50603 L
.2619 .50871 L
.28175 .51103 L
.29167 .51203 L
.30159 .51287 L
.31151 .51351 L
.31647 .51375 L
.32143 .51396 L
.33135 .51429 L
.34127 .51468 L
.35119 .51522 L
.36111 .51588 L
.38095 .51751 L
.42063 .52162 L
.46032 .52648 L
.48016 .52941 L
.49008 .53067 L
.49504 .53118 L
.5 .53161 L
.50248 .53179 L
.50496 .53194 L
.50744 .53207 L
.50992 .53217 L
.51116 .53221 L
.5124 .53225 L
.51364 .53228 L
.51488 .5323 L
.51612 .53232 L
.51736 .53233 L
.5186 .53233 L
.51984 .53233 L
.52108 .53232 L
.52232 .53231 L
.52356 .53229 L
.5248 .53226 L
.52728 .53218 L
.52976 .53208 L
Mistroke
.53472 .5318 L
.53968 .53142 L
.54464 .53092 L
.5496 .53031 L
.55952 .52867 L
.57937 .52432 L
.58929 .52227 L
.59921 .52051 L
.60913 .51904 L
.61905 .51784 L
.62897 .51688 L
.63889 .51616 L
.64385 .51589 L
.64881 .51567 L
.65377 .51549 L
.65625 .51543 L
.65873 .51537 L
.66121 .51533 L
.66369 .5153 L
.66493 .51529 L
.66617 .51528 L
.66741 .51528 L
.66865 .51528 L
.66989 .51528 L
.67113 .51528 L
.67237 .51529 L
.67361 .51529 L
.67609 .51532 L
.67857 .51536 L
.68105 .5154 L
.68353 .51546 L
.68849 .5156 L
.69841 .516 L
.70833 .51654 L
.71825 .51721 L
.7381 .51884 L
.77778 .52286 L
.79762 .52492 L
.81746 .52673 L
.82738 .52746 L
.83234 .52777 L
.8373 .52804 L
.84226 .52826 L
.84474 .52835 L
.84722 .52843 L
.8497 .52849 L
.85218 .52854 L
.85342 .52855 L
.85466 .52857 L
.8559 .52858 L
Mistroke
.85714 .52858 L
.85838 .52858 L
.85962 .52858 L
.86086 .52857 L
.8621 .52856 L
.86458 .52852 L
.86582 .5285 L
.86706 .52846 L
.87202 .52829 L
.8745 .52817 L
.87698 .52803 L
.88194 .52768 L
.8869 .52724 L
.89683 .52606 L
.90675 .52459 L
.91667 .52295 L
.92659 .52091 L
.93651 .51816 L
.94643 .51442 L
.95635 .50987 L
.97619 .49975 L
Mfstroke
P
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell["\<\
The number of effective parameters is 15.4 and the average absolute error is \
down to about 10 basis points. This is quite an improvement from the previous \
fit. \
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "It turns out that there is another class of securities that tends to have \
another additional feature that is not captured in the present value of the \
the stated payments. The one or two most-recently-issued securities of a \
given original term-to-maturity are often ``on special'' in the repo market. \
As such, an additional stream of payments can be obtained in the financing \
market by the holder of the security. We can screen out the most recently \
issued securities by term with another option to ",
  StyleBox["DiscountFunction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ":"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
df5 = DiscountFunction[cp, FunctionalForm -> Delta,
\tDropByTerm -> 2, 
\tLambda0 -> {10^11.5, 1.1 10^11.5},
\tIDNumberDropList -> iddrop];
DFSummaryStatistics[df5]
ShowGraph[df5, PlotStyle -> {Dashing[{.02, .01}],{}}];\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
TableForm[{PaddedForm[197, {3, 0}, NumberPoint -> \"\"], 
 
   StringForm[\"`1` `2`\", 
 
    PaddedForm[29.81793292265571, {5, 2}], \"years\"], 
 
   PaddedForm[2, {3, 0}, NumberPoint -> \"\"], Delta, True, 
 
   True, PaddedForm[2, {3, 0}, NumberPoint -> \"\"], 
 
   PaddedForm[65, {3, 0}, NumberPoint -> \"\"], 
 
   PaddedForm[AccountingForm[5.074809230968149, 
 
     NumberSigns -> {\"-\", \"\"}, NumberPadding -> {\" \", \"0\"}]\\
 
     , {4, 1}], PaddedForm[3.340558776236393*10^13, {5, 2}, 
 
    NumberFormat -> (#1\"e\"#3 & )], 
 
   PaddedForm[AccountingForm[1.369620172852173, 
 
     NumberSigns -> {\"-\", \"\"}, NumberPadding -> {\" \", \"0\"}]\\
 
     , {8, 5}], PaddedForm[AccountingForm[0.41491575139059\\
 
      12, NumberSigns -> {\"-\", \"\"}, 
 
     NumberPadding -> {\" \", \"0\"}], {7, 4}]}, 
 
  TableSpacing -> {0}, TableHeadings -> 
 
   {{\"no. of obs.........\", \"longest payment....\", \"no.\\
 
      dropped........\", \"functional form....\", \"restriction\\
 
      ........\", \"minimize GCV.......\", \"tuning\\
 
      parameter...\", \"no. of knots.......\", \"no. of\\
 
      parameters..\", \"lambda.............\", \"residual\\
 
      variance..\", \"avg. abs. error....\"}}]\
\>", "\<\

no. of obs.........   197
longest payment....    29.82 years
no. dropped........     2
functional form....  Delta
restriction........  True
minimize GCV.......  True
tuning parameter...     2
no. of knots.......    65
no. of parameters..     5.1
lambda.............     3.34e13
residual variance..     1.36962
avg. abs. error....     0.4149\
\>"], "Output",
  LineSpacing->{1, 0},
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0237226 0.0317475 -1.43823 0.201029 [
[(0)] .02372 0 0 2 0 Minner Mrotsboxa
[(5)] .18246 0 0 2 0 Minner Mrotsboxa
[(10)] .3412 0 0 2 0 Minner Mrotsboxa
[(15)] .49993 0 0 2 0 Minner Mrotsboxa
[(20)] .65867 0 0 2 0 Minner Mrotsboxa
[(25)] .81741 0 0 2 0 Minner Mrotsboxa
[(30)] .97615 0 0 2 0 Minner Mrotsboxa
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotsboxa
[(7.5)] -0.0125 .06949 1 0 0 Minner Mrotsboxa
[(8)] -0.0125 .17001 1 0 0 Minner Mrotsboxa
[(8.5)] -0.0125 .27052 1 0 0 Minner Mrotsboxa
[(9)] -0.0125 .37103 1 0 0 Minner Mrotsboxa
[(9.5)] -0.0125 .47155 1 0 0 Minner Mrotsboxa
[(10)] -0.0125 .57206 1 0 0 Minner Mrotsboxa
[(percent)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotsboxa
[(7/19/1988, 65 knots,   5.1 parms, Delta)] .5 .61803 0 -2 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
[ .002 .01 ] 0 setdash
.002 w
.02372 0 m
.02372 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.18246 0 m
.18246 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.3412 0 m
.3412 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.49993 0 m
.49993 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.65867 0 m
.65867 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.81741 0 m
.81741 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.97615 0 m
.97615 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .06949 m
1 .06949 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .17001 m
1 .17001 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .27052 m
1 .27052 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .37103 m
1 .37103 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .47155 m
1 .47155 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .57206 m
1 .57206 L
s
P
p
.002 w
.02372 0 m
.02372 .00625 L
s
P
[(0)] .02372 0 0 2 0 Minner Mrotshowa
p
.002 w
.18246 0 m
.18246 .00625 L
s
P
[(5)] .18246 0 0 2 0 Minner Mrotshowa
p
.002 w
.3412 0 m
.3412 .00625 L
s
P
[(10)] .3412 0 0 2 0 Minner Mrotshowa
p
.002 w
.49993 0 m
.49993 .00625 L
s
P
[(15)] .49993 0 0 2 0 Minner Mrotshowa
p
.002 w
.65867 0 m
.65867 .00625 L
s
P
[(20)] .65867 0 0 2 0 Minner Mrotshowa
p
.002 w
.81741 0 m
.81741 .00625 L
s
P
[(25)] .81741 0 0 2 0 Minner Mrotshowa
p
.002 w
.97615 0 m
.97615 .00625 L
s
P
[(30)] .97615 0 0 2 0 Minner Mrotshowa
p
.001 w
.05547 0 m
.05547 .00375 L
s
P
p
.001 w
.08722 0 m
.08722 .00375 L
s
P
p
.001 w
.11897 0 m
.11897 .00375 L
s
P
p
.001 w
.15071 0 m
.15071 .00375 L
s
P
p
.001 w
.21421 0 m
.21421 .00375 L
s
P
p
.001 w
.24595 0 m
.24595 .00375 L
s
P
p
.001 w
.2777 0 m
.2777 .00375 L
s
P
p
.001 w
.30945 0 m
.30945 .00375 L
s
P
p
.001 w
.37294 0 m
.37294 .00375 L
s
P
p
.001 w
.40469 0 m
.40469 .00375 L
s
P
p
.001 w
.43644 0 m
.43644 .00375 L
s
P
p
.001 w
.46819 0 m
.46819 .00375 L
s
P
p
.001 w
.53168 0 m
.53168 .00375 L
s
P
p
.001 w
.56343 0 m
.56343 .00375 L
s
P
p
.001 w
.59518 0 m
.59518 .00375 L
s
P
p
.001 w
.62692 0 m
.62692 .00375 L
s
P
p
.001 w
.69042 0 m
.69042 .00375 L
s
P
p
.001 w
.72217 0 m
.72217 .00375 L
s
P
p
.001 w
.75391 0 m
.75391 .00375 L
s
P
p
.001 w
.78566 0 m
.78566 .00375 L
s
P
p
.001 w
.84916 0 m
.84916 .00375 L
s
P
p
.001 w
.8809 0 m
.8809 .00375 L
s
P
p
.001 w
.91265 0 m
.91265 .00375 L
s
P
p
.001 w
.9444 0 m
.9444 .00375 L
s
P
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotshowa
p
.002 w
0 0 m
1 0 L
s
P
p
.002 w
0 .06949 m
.00625 .06949 L
s
P
[(7.5)] -0.0125 .06949 1 0 0 Minner Mrotshowa
p
.002 w
0 .17001 m
.00625 .17001 L
s
P
[(8)] -0.0125 .17001 1 0 0 Minner Mrotshowa
p
.002 w
0 .27052 m
.00625 .27052 L
s
P
[(8.5)] -0.0125 .27052 1 0 0 Minner Mrotshowa
p
.002 w
0 .37103 m
.00625 .37103 L
s
P
[(9)] -0.0125 .37103 1 0 0 Minner Mrotshowa
p
.002 w
0 .47155 m
.00625 .47155 L
s
P
[(9.5)] -0.0125 .47155 1 0 0 Minner Mrotshowa
p
.002 w
0 .57206 m
.00625 .57206 L
s
P
[(10)] -0.0125 .57206 1 0 0 Minner Mrotshowa
p
.001 w
0 .08959 m
.00375 .08959 L
s
P
p
.001 w
0 .1097 m
.00375 .1097 L
s
P
p
.001 w
0 .1298 m
.00375 .1298 L
s
P
p
.001 w
0 .1499 m
.00375 .1499 L
s
P
p
.001 w
0 .19011 m
.00375 .19011 L
s
P
p
.001 w
0 .21021 m
.00375 .21021 L
s
P
p
.001 w
0 .23031 m
.00375 .23031 L
s
P
p
.001 w
0 .25042 m
.00375 .25042 L
s
P
p
.001 w
0 .29062 m
.00375 .29062 L
s
P
p
.001 w
0 .31073 m
.00375 .31073 L
s
P
p
.001 w
0 .33083 m
.00375 .33083 L
s
P
p
.001 w
0 .35093 m
.00375 .35093 L
s
P
p
.001 w
0 .39114 m
.00375 .39114 L
s
P
p
.001 w
0 .41124 m
.00375 .41124 L
s
P
p
.001 w
0 .43134 m
.00375 .43134 L
s
P
p
.001 w
0 .45145 m
.00375 .45145 L
s
P
p
.001 w
0 .49165 m
.00375 .49165 L
s
P
p
.001 w
0 .51176 m
.00375 .51176 L
s
P
p
.001 w
0 .53186 m
.00375 .53186 L
s
P
p
.001 w
0 .55196 m
.00375 .55196 L
s
P
p
.001 w
0 .04939 m
.00375 .04939 L
s
P
p
.001 w
0 .02928 m
.00375 .02928 L
s
P
p
.001 w
0 .00918 m
.00375 .00918 L
s
P
p
.001 w
0 .59217 m
.00375 .59217 L
s
P
p
.001 w
0 .61227 m
.00375 .61227 L
s
P
[(percent)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotshowa
p
.002 w
0 0 m
0 .61803 L
s
P
P
p
[(7/19/1988, 65 knots,   5.1 parms, Delta)] .5 .61803 0 -2 Mshowa
P
p
P
p
p
[ .02 .01 ] 0 setdash
p
.004 w
.02381 .06856 m
.06349 .19539 L
.10317 .28617 L
.14286 .37169 L
.1627 .42689 L
.18254 .4824 L
.19246 .50473 L
.20238 .52271 L
.20734 .53003 L
.2123 .53628 L
.21726 .54148 L
.22222 .54583 L
.2247 .54766 L
.22718 .54922 L
.22966 .55051 L
.23214 .55156 L
.23462 .55245 L
.2371 .55317 L
.23958 .5537 L
.24206 .5541 L
.24702 .55475 L
.2495 .55501 L
.25198 .55523 L
.25446 .5554 L
.25694 .55552 L
.25818 .55557 L
.25942 .5556 L
.26066 .55563 L
.2619 .55564 L
.26314 .55564 L
.26438 .55563 L
.26563 .55561 L
.26687 .55558 L
.26935 .55552 L
.27183 .55543 L
.27679 .5552 L
.28175 .55491 L
.28423 .55479 L
.28547 .55474 L
.28671 .55469 L
.28795 .55466 L
.28919 .55463 L
.29043 .55461 L
.29167 .55459 L
.29291 .55459 L
.29415 .5546 L
.29539 .55461 L
.29663 .55463 L
.29911 .55471 L
.30035 .55476 L
Mistroke
.30159 .55483 L
.34127 .5597 L
.36111 .56273 L
.38095 .56652 L
.39087 .5689 L
.40079 .57173 L
.42063 .57917 L
.44048 .58931 L
.4504 .59441 L
.45536 .59663 L
.46032 .59855 L
.46528 .60015 L
.47024 .60143 L
.47272 .60194 L
.4752 .60236 L
.47768 .60269 L
.48016 .60294 L
.48264 .60313 L
.48388 .6032 L
.48512 .60325 L
.48636 .60329 L
.4876 .60331 L
.48884 .60332 L
.49008 .60331 L
.49132 .60328 L
.49256 .60324 L
.4938 .60318 L
.49504 .60311 L
.49752 .60291 L
.5 .60264 L
.50496 .60188 L
.50992 .60084 L
.51488 .5995 L
.51984 .59786 L
.52976 .59359 L
.53968 .58779 L
.5496 .5801 L
.55952 .56994 L
.57937 .54125 L
.59921 .50098 L
.61905 .45472 L
.65873 .34942 L
.69841 .23615 L
.7381 .13111 L
.75794 .08669 L
.76786 .06752 L
.77778 .05078 L
.7877 .03681 L
.79762 .02594 L
.80258 .02179 L
Mistroke
.80754 .01854 L
.81002 .01728 L
.8125 .01626 L
.81498 .01549 L
.81622 .0152 L
.81746 .01498 L
.8187 .01483 L
.81994 .01474 L
.82118 .01472 L
.82242 .01476 L
.82366 .01488 L
.8249 .01507 L
.82614 .01532 L
.82738 .01565 L
.82986 .01653 L
.83234 .0177 L
.8373 .02097 L
.84226 .02551 L
.84722 .03138 L
.85714 .04736 L
.86706 .06942 L
.87698 .09601 L
.89683 .15674 L
.93651 .32297 L
.97619 .53928 L
Mfstroke
P
P
p
p
.004 w
.02381 .0684 m
.06349 .1348 L
.10317 .18884 L
.14286 .23501 L
.18254 .28299 L
.22222 .33033 L
.2619 .36744 L
.30159 .39423 L
.34127 .41457 L
.38095 .43105 L
.42063 .44515 L
.46032 .45824 L
.5 .47023 L
.51984 .47545 L
.53968 .47998 L
.55952 .48367 L
.56944 .48513 L
.5744 .48574 L
.57937 .48627 L
.58433 .48672 L
.58929 .48708 L
.59177 .48722 L
.59425 .48734 L
.59673 .48743 L
.59797 .48747 L
.59921 .4875 L
.60045 .48753 L
.60169 .48755 L
.60293 .48756 L
.60417 .48757 L
.60541 .48757 L
.60665 .48757 L
.60789 .48756 L
.60913 .48754 L
.61037 .48752 L
.61161 .48749 L
.61409 .48742 L
.61657 .48732 L
.61905 .48719 L
.62401 .48688 L
.62897 .48646 L
.63889 .48534 L
.64881 .48384 L
.65873 .48195 L
.67857 .47708 L
.69841 .47082 L
.7381 .4548 L
.77778 .4355 L
.81746 .41514 L
.85714 .39653 L
Mistroke
.87698 .38894 L
.8869 .38574 L
.89683 .38296 L
.90675 .38061 L
.91667 .37876 L
.92163 .37802 L
.92659 .3774 L
.93155 .37691 L
.93403 .37672 L
.93651 .37656 L
.93899 .37643 L
.94023 .37638 L
.94147 .37633 L
.94271 .3763 L
.94395 .37627 L
.94519 .37625 L
.94643 .37624 L
.94767 .37624 L
.94891 .37625 L
.95015 .37626 L
.95139 .37629 L
.95263 .37632 L
.95387 .37636 L
.95635 .37647 L
.95883 .37661 L
.96131 .37679 L
.96627 .37726 L
.97123 .37787 L
.97619 .37863 L
Mfstroke
P
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell["\<\
The pricing errors are a bit smaller. And although the number of knots chosen \
was 7 fewer, the number of effective parameters dropped by less than 4.\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
Let's use this discount function for an illustrative digression (even though \
it could probably be improved on by using a differenct functional form).  \
Consider a coupon bond with a face value of \\$1 that pays its coupon as a \
continuous stream---an instantaneous coupon bond. Let this bond mature at \
time $\\tau$. The (instantaneous) swap rate is the coupon rate that makes the \
value of this bond equal \\$1 today. The (instantaneous) forward swap rate is \
the coupon rate that makes the forward value of the bond equal the forward \
value of \\$1 at time $t'$.  Here is the formula for the forward swap rate: \
\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
$$s(t',\\tau) = {\\delta(t') - \\delta(\\tau) \\over \\int_{t'}^\\tau \
\\delta(u)\\,du}$$\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "We can plot the instantaneous forward swap surface as follows. First \
extract $\\delta(\\tau)$ from the ",
  StyleBox["DF",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object (otherwise it the plotting routine will run much slower) and then \
run ",
  StyleBox["PlotSwapSurface",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ":"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
delta = Delta[df5];
PlotSwapSurface[delta];\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .82055 
MathPictureStart
%% Graphics3D
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0249355 0.99742 -0.0396341 0.99742 [
[(0)] .06425 .25367 1 .94269 Msboxa
[(10)] .24637 .18201 .92842 1 Msboxa
[(20)] .44296 .10469 .79605 1 Msboxa
[(30)] .65583 .02105 .66367 1 Msboxa
[(0)] .69877 .03239 -1 .38878 Msboxa
[(10)] .79536 .18028 -1 .35504 Msboxa
[(20)] .87922 .30871 -1 .3267 Msboxa
[(30)] .95273 .42129 -1 .30254 Msboxa
[(8)] .03939 .32984 1 -0.38011 Msboxa
[(9)] .02795 .40152 1 -0.36587 Msboxa
[(10)] .01595 .47675 1 -0.35086 Msboxa
[ 0 0 0 0 ]
[ 1 .82055 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
.002 w
.06024 .26735 m
.67932 .02494 L
s
P
p
.002 w
.07332 .26223 m
.07786 .2665 L
s
P
[(0)] .06425 .25367 1 .94269 Mshowa
p
.002 w
.25485 .19115 m
.25909 .19571 L
s
P
[(10)] .24637 .18201 .92842 1 Mshowa
p
.002 w
.45073 .11445 m
.45461 .11932 L
s
P
[(20)] .44296 .10469 .79605 1 Mshowa
p
.002 w
.66272 .03144 m
.66617 .03663 L
s
P
[(30)] .65583 .02105 .66367 1 Mshowa
p
.001 w
.10856 .24843 m
.11125 .25103 L
s
P
p
.001 w
.14431 .23443 m
.14697 .23706 L
s
P
p
.001 w
.18061 .22022 m
.18323 .22289 L
s
P
p
.001 w
.21745 .20579 m
.22003 .2085 L
s
P
p
.001 w
.29283 .17628 m
.29533 .17905 L
s
P
p
.001 w
.33139 .16118 m
.33385 .16399 L
s
P
p
.001 w
.37055 .14584 m
.37297 .14869 L
s
P
p
.001 w
.41032 .13027 m
.4127 .13316 L
s
P
p
.001 w
.49177 .09837 m
.49406 .10134 L
s
P
p
.001 w
.53348 .08204 m
.53571 .08505 L
s
P
p
.001 w
.57586 .06545 m
.57804 .06849 L
s
P
p
.001 w
.61894 .04858 m
.62106 .05166 L
s
P
P
p
p
.002 w
.67932 .02494 m
.94594 .43277 L
s
P
p
.002 w
.68715 .03691 m
.68134 .03917 L
s
P
[(0)] .69877 .03239 -1 .38878 Mshowa
p
.002 w
.78361 .18445 m
.77773 .18654 L
s
P
[(10)] .79536 .18028 -1 .35504 Mshowa
p
.002 w
.86737 .31259 m
.86145 .31452 L
s
P
[(20)] .87922 .30871 -1 .3267 Mshowa
p
.002 w
.9408 .4249 m
.93483 .42671 L
s
P
[(30)] .95273 .42129 -1 .30254 Mshowa
p
.001 w
.70759 .06818 m
.7041 .06951 L
s
P
p
.001 w
.72744 .09853 m
.72393 .09984 L
s
P
p
.001 w
.7467 .128 m
.74319 .12929 L
s
P
p
.001 w
.76542 .15663 m
.7619 .1579 L
s
P
p
.001 w
.80129 .2115 m
.79776 .21274 L
s
P
p
.001 w
.81849 .23781 m
.81495 .23903 L
s
P
p
.001 w
.83523 .26341 m
.83168 .26461 L
s
P
p
.001 w
.85151 .28833 m
.84796 .28951 L
s
P
p
.001 w
.88282 .33622 m
.87926 .33736 L
s
P
p
.001 w
.89787 .35924 m
.89431 .36037 L
s
P
p
.001 w
.91254 .38168 m
.90897 .38279 L
s
P
p
.001 w
.92685 .40356 m
.92327 .40466 L
s
P
P
p
p
.002 w
.06024 .26735 m
.02494 .49015 L
s
P
p
.002 w
.05104 .32541 m
.05687 .32319 L
s
P
[(8)] .03939 .32984 1 -0.38011 Mshowa
p
.002 w
.03966 .39724 m
.04551 .3951 L
s
P
[(9)] .02795 .40152 1 -0.36587 Mshowa
p
.002 w
.02771 .47262 m
.0336 .47055 L
s
P
[(10)] .01595 .47675 1 -0.35086 Mshowa
p
.001 w
.04881 .3395 m
.05231 .33818 L
s
P
p
.001 w
.04655 .35373 m
.05006 .35242 L
s
P
p
.001 w
.04428 .3681 m
.04778 .36679 L
s
P
p
.001 w
.04198 .3826 m
.04549 .38131 L
s
P
p
.001 w
.03732 .41202 m
.04083 .41075 L
s
P
p
.001 w
.03495 .42695 m
.03847 .42569 L
s
P
p
.001 w
.03256 .44203 m
.03608 .44077 L
s
P
p
.001 w
.03015 .45725 m
.03368 .456 L
s
P
p
.001 w
.05325 .31145 m
.05675 .31011 L
s
P
p
.001 w
.05545 .29761 m
.05894 .29627 L
s
P
p
.001 w
.05762 .28391 m
.0611 .28256 L
s
P
p
.001 w
.05977 .27033 m
.06325 .26897 L
s
P
p
.001 w
.02525 .48814 m
.02879 .48691 L
s
P
P
0 0 m
1 0 L
1 .82055 L
0 .82055 L
closepath
clip
newpath
p
.002 w
.06024 .26735 m
.02494 .49015 L
s
.02494 .49015 m
.40296 .79562 L
s
.40296 .79562 m
.41001 .59401 L
s
.41001 .59401 m
.06024 .26735 L
s
.67932 .02494 m
.94594 .43277 L
s
.94594 .43277 m
.97506 .64585 L
s
.97506 .64585 m
.69286 .25814 L
s
.69286 .25814 m
.67932 .02494 L
s
.06024 .26735 m
.02494 .49015 L
s
.02494 .49015 m
.69286 .25814 L
s
.69286 .25814 m
.67932 .02494 L
s
.67932 .02494 m
.06024 .26735 L
s
.41001 .59401 m
.94594 .43277 L
s
.94594 .43277 m
.97506 .64585 L
s
.97506 .64585 m
.40296 .79562 L
s
.40296 .79562 m
.41001 .59401 L
s
P
p
.002 w
.82365 .42073 m .79661 .40959 L .81547 .40544 L closepath p
.63 .65 .851 r
F P
s
P
p
.002 w
.85254 .44971 m .82365 .42073 L .84352 .42546 L closepath p
.508 .494 .777 r
F P
s
P
p
.002 w
.82365 .42073 m .81547 .40544 L .83504 .4068 L p
.477 .527 .83 r
F P
s
P
p
.002 w
.83504 .4068 m .84352 .42546 L .82365 .42073 L p
.477 .527 .83 r
F P
s
P
p
.002 w
.79661 .40959 m .77066 .41264 L .78894 .40179 L closepath p
.762 .84 .911 r
F P
s
P
p
.002 w
.79661 .40959 m .78894 .40179 L .80768 .39522 L p
.624 .74 .93 r
F P
s
P
p
.002 w
.80768 .39522 m .81547 .40544 L .79661 .40959 L p
.624 .74 .93 r
F P
s
P
p
.002 w
.81547 .40544 m .80768 .39522 L .82706 .39371 L p
.502 .624 .904 r
F P
s
P
p
.002 w
.82706 .39371 m .83504 .4068 L .81547 .40544 L p
.502 .624 .904 r
F P
s
P
p
.002 w
.88314 .49147 m .85254 .44971 L .8734 .46091 L closepath p
.45 .424 .738 r
F P
s
P
p
.002 w
.85254 .44971 m .84352 .42546 L .8641 .43428 L p
.427 .43 .761 r
F P
s
P
p
.002 w
.8641 .43428 m .8734 .46091 L .85254 .44971 L p
.427 .43 .761 r
F P
s
P
p
.002 w
.77066 .41264 m .74518 .42609 L .76319 .41036 L closepath p
.835 .972 .885 r
F P
s
P
p
.002 w
.77066 .41264 m .76319 .41036 L .78149 .39785 L p
.751 .936 .946 r
F P
s
P
p
.002 w
.78149 .39785 m .78894 .40179 L .77066 .41264 L p
.751 .936 .946 r
F P
s
P
p
.002 w
.84352 .42546 m .83504 .4068 L .85532 .41268 L p
.422 .467 .801 r
F P
s
P
p
.002 w
.85532 .41268 m .8641 .43428 L .84352 .42546 L p
.422 .467 .801 r
F P
s
P
p
.002 w
.78894 .40179 m .78149 .39785 L .8002 .38926 L p
.654 .862 .978 r
F P
s
P
p
.002 w
.8002 .38926 m .80768 .39522 L .78894 .40179 L p
.654 .862 .978 r
F P
s
P
p
.002 w
.83504 .4068 m .82706 .39371 L .8471 .3967 L p
.42 .532 .865 r
F P
s
P
p
.002 w
.8471 .3967 m .85532 .41268 L .83504 .4068 L p
.42 .532 .865 r
F P
s
P
p
.002 w
.80768 .39522 m .8002 .38926 L .81946 .38532 L p
.522 .743 .976 r
F P
s
P
p
.002 w
.81946 .38532 m .82706 .39371 L .80768 .39522 L p
.522 .743 .976 r
F P
s
P
p
.002 w
.88314 .49147 m .8734 .46091 L .89508 .47598 L p
.388 .373 .721 r
F P
s
P
p
.002 w
.8734 .46091 m .8641 .43428 L .88541 .44655 L p
.392 .394 .742 r
F P
s
P
p
.002 w
.88541 .44655 m .89508 .47598 L .8734 .46091 L p
.392 .394 .742 r
F P
s
P
p
.002 w
.91576 .54592 m .88314 .49147 L .90522 .50937 L closepath p
.407 .374 .71 r
F P
s
P
p
.002 w
.89508 .47598 m .90522 .50937 L .88314 .49147 L p
.388 .373 .721 r
F P
s
P
p
.002 w
.8641 .43428 m .85532 .41268 L .8763 .42212 L p
.388 .424 .776 r
F P
s
P
p
.002 w
.8763 .42212 m .88541 .44655 L .8641 .43428 L p
.388 .424 .776 r
F P
s
P
p
.002 w
.74518 .42609 m .71967 .44672 L .73767 .42769 L closepath p
.823 .999 .789 r
F P
s
P
p
.002 w
.74518 .42609 m .73767 .42769 L .75581 .41094 L p
.778 .998 .838 r
F P
s
P
p
.002 w
.75581 .41094 m .76319 .41036 L .74518 .42609 L p
.778 .998 .838 r
F P
s
P
p
.002 w
.76319 .41036 m .75581 .41094 L .77419 .3971 L p
.725 .985 .896 r
F P
s
P
p
.002 w
.77419 .3971 m .78149 .39785 L .76319 .41036 L p
.725 .985 .896 r
F P
s
P
p
.002 w
.82706 .39371 m .81946 .38532 L .83934 .38572 L p
.416 .625 .939 r
F P
s
P
p
.002 w
.83934 .38572 m .8471 .3967 L .82706 .39371 L p
.416 .625 .939 r
F P
s
P
p
.002 w
.78149 .39785 m .77419 .3971 L .79292 .38684 L p
.64 .943 .958 r
F P
s
P
p
.002 w
.79292 .38684 m .8002 .38926 L .78149 .39785 L p
.64 .943 .958 r
F P
s
P
p
.002 w
.85532 .41268 m .8471 .3967 L .86781 .40334 L p
.374 .469 .829 r
F P
s
P
p
.002 w
.86781 .40334 m .8763 .42212 L .85532 .41268 L p
.374 .469 .829 r
F P
s
P
p
.002 w
.8002 .38926 m .79292 .38684 L .81214 .38086 L p
.513 .848 .997 r
F P
s
P
p
.002 w
.81214 .38086 m .81946 .38532 L .8002 .38926 L p
.513 .848 .997 r
F P
s
P
p
.002 w
.8471 .3967 m .83934 .38572 L .85984 .38976 L p
.354 .536 .895 r
F P
s
P
p
.002 w
.85984 .38976 m .86781 .40334 L .8471 .3967 L p
.354 .536 .895 r
F P
s
P
p
.002 w
.81946 .38532 m .81214 .38086 L .83192 .379 L p
.394 .726 .99 r
F P
s
P
p
.002 w
.83192 .379 m .83934 .38572 L .81946 .38532 L p
.394 .726 .99 r
F P
s
P
p
.002 w
.90522 .50937 m .89508 .47598 L .91752 .49398 L p
.374 .355 .709 r
F P
s
P
p
.002 w
.89508 .47598 m .88541 .44655 L .90744 .46172 L p
.376 .372 .727 r
F P
s
P
p
.002 w
.90744 .46172 m .91752 .49398 L .89508 .47598 L p
.376 .372 .727 r
F P
s
P
p
.002 w
.88541 .44655 m .8763 .42212 L .89798 .43449 L p
.369 .394 .756 r
F P
s
P
p
.002 w
.89798 .43449 m .90744 .46172 L .88541 .44655 L p
.369 .394 .756 r
F P
s
P
p
.002 w
.91576 .54592 m .90522 .50937 L .92812 .53032 L p
.366 .339 .696 r
F P
s
P
p
.002 w
.91752 .49398 m .92812 .53032 L .90522 .50937 L p
.374 .355 .709 r
F P
s
P
p
.002 w
.74843 .41377 m .75581 .41094 L .73767 .42769 L p
.711 .973 .748 r
F P
s
P
p
.002 w
.75581 .41094 m .74843 .41377 L .76694 .3989 L p
.655 .966 .794 r
F P
s
P
p
.002 w
.76694 .3989 m .77419 .3971 L .75581 .41094 L p
.655 .966 .794 r
F P
s
P
p
.002 w
.73013 .43127 m .73767 .42769 L .71967 .44672 L p
.752 .976 .719 r
F P
s
P
p
.002 w
.73767 .42769 m .73013 .43127 L .74843 .41377 L p
.711 .973 .748 r
F P
s
P
p
.002 w
.8763 .42212 m .86781 .40334 L .88918 .41296 L p
.351 .427 .8 r
F P
s
P
p
.002 w
.88918 .41296 m .89798 .43449 L .8763 .42212 L p
.351 .427 .8 r
F P
s
P
p
.002 w
.71967 .44672 m .69379 .47172 L .71194 .45078 L closepath p
.783 .978 .705 r
F P
s
P
p
.002 w
.71967 .44672 m .71194 .45078 L .73013 .43127 L p
.752 .976 .719 r
F P
s
P
p
.002 w
.77419 .3971 m .76694 .3989 L .78575 .38728 L p
.574 .941 .856 r
F P
s
P
p
.002 w
.78575 .38728 m .79292 .38684 L .77419 .3971 L p
.574 .941 .856 r
F P
s
P
p
.002 w
.95079 .61376 m .91576 .54592 L .93919 .57015 L closepath p
.386 .343 .687 r
F P
s
P
p
.002 w
.92812 .53032 m .93919 .57015 L .91576 .54592 L p
.366 .339 .696 r
F P
s
P
p
.002 w
.79292 .38684 m .78575 .38728 L .80499 .37959 L p
.455 .879 .919 r
F P
s
P
p
.002 w
.80499 .37959 m .81214 .38086 L .79292 .38684 L p
.455 .879 .919 r
F P
s
P
p
.002 w
.83934 .38572 m .83192 .379 L .85228 .38069 L p
.318 .618 .956 r
F P
s
P
p
.002 w
.85228 .38069 m .85984 .38976 L .83934 .38572 L p
.318 .618 .956 r
F P
s
P
p
.002 w
.86781 .40334 m .85984 .38976 L .88096 .39679 L p
.323 .475 .857 r
F P
s
P
p
.002 w
.88096 .39679 m .88918 .41296 L .86781 .40334 L p
.323 .475 .857 r
F P
s
P
p
.002 w
.81214 .38086 m .80499 .37959 L .82474 .37579 L p
.337 .785 .957 r
F P
s
P
p
.002 w
.82474 .37579 m .83192 .379 L .81214 .38086 L p
.337 .785 .957 r
F P
s
P
p
.002 w
.85984 .38976 m .85228 .38069 L .87323 .38536 L p
.281 .539 .917 r
F P
s
P
p
.002 w
.87323 .38536 m .88096 .39679 L .85984 .38976 L p
.281 .539 .917 r
F P
s
P
p
.002 w
.741 .41825 m .74843 .41377 L .73013 .43127 L p
.645 .931 .666 r
F P
s
P
p
.002 w
.74843 .41377 m .741 .41825 L .75967 .40261 L p
.58 .915 .689 r
F P
s
P
p
.002 w
.75967 .40261 m .76694 .3989 L .74843 .41377 L p
.58 .915 .689 r
F P
s
P
p
.002 w
.7225 .43624 m .73013 .43127 L .71194 .45078 L p
.696 .944 .657 r
F P
s
P
p
.002 w
.73013 .43127 m .7225 .43624 L .741 .41825 L p
.645 .931 .666 r
F P
s
P
p
.002 w
.83192 .379 m .82474 .37579 L .84503 .37542 L p
.257 .686 .963 r
F P
s
P
p
.002 w
.84503 .37542 m .85228 .38069 L .83192 .379 L p
.257 .686 .963 r
F P
s
P
p
.002 w
.76694 .3989 m .75967 .40261 L .77861 .38993 L p
.493 .888 .728 r
F P
s
P
p
.002 w
.77861 .38993 m .78575 .38728 L .76694 .3989 L p
.493 .888 .728 r
F P
s
P
p
.002 w
.70409 .45598 m .71194 .45078 L .69379 .47172 L p
.737 .956 .659 r
F P
s
P
p
.002 w
.71194 .45078 m .70409 .45598 L .7225 .43624 L p
.696 .944 .657 r
F P
s
P
p
.002 w
.78575 .38728 m .77861 .38993 L .79793 .38084 L p
.376 .838 .78 r
F P
s
P
p
.002 w
.79793 .38084 m .80499 .37959 L .78575 .38728 L p
.376 .838 .78 r
F P
s
P
p
.002 w
.69379 .47172 m .6673 .49853 L .6857 .47693 L closepath p
.757 .964 .671 r
F P
s
P
p
.002 w
.69379 .47172 m .6857 .47693 L .70409 .45598 L p
.737 .956 .659 r
F P
s
P
p
.002 w
.80499 .37959 m .79793 .38084 L .8177 .37539 L p
.261 .77 .834 r
F P
s
P
p
.002 w
.8177 .37539 m .82474 .37579 L .80499 .37959 L p
.261 .77 .834 r
F P
s
P
p
.002 w
.85228 .38069 m .84503 .37542 L .86587 .37796 L p
.216 .601 .949 r
F P
s
P
p
.002 w
.86587 .37796 m .87323 .38536 L .85228 .38069 L p
.216 .601 .949 r
F P
s
P
p
.002 w
.82474 .37579 m .8177 .37539 L .83798 .37323 L p
.178 .698 .879 r
F P
s
P
p
.002 w
.83798 .37323 m .84503 .37542 L .82474 .37579 L p
.178 .698 .879 r
F P
s
P
p
.002 w
.73346 .42373 m .741 .41825 L .7225 .43624 L p
.603 .9 .616 r
F P
s
P
p
.002 w
.741 .41825 m .73346 .42373 L .75233 .40758 L p
.53 .873 .618 r
F P
s
P
p
.002 w
.75233 .40758 m .75967 .40261 L .741 .41825 L p
.53 .873 .618 r
F P
s
P
p
.002 w
.71473 .44196 m .7225 .43624 L .70409 .45598 L p
.663 .923 .624 r
F P
s
P
p
.002 w
.7225 .43624 m .71473 .44196 L .73346 .42373 L p
.603 .9 .616 r
F P
s
P
p
.002 w
.75967 .40261 m .75233 .40758 L .77143 .39409 L p
.437 .836 .633 r
F P
s
P
p
.002 w
.77143 .39409 m .77861 .38993 L .75967 .40261 L p
.437 .836 .633 r
F P
s
P
p
.002 w
.69608 .46169 m .70409 .45598 L .6857 .47693 L p
.712 .943 .641 r
F P
s
P
p
.002 w
.70409 .45598 m .69608 .46169 L .71473 .44196 L p
.663 .923 .624 r
F P
s
P
p
.002 w
.79087 .3839 m .79793 .38084 L .77861 .38993 L p
.32 .784 .661 r
F P
s
P
p
.002 w
.77861 .38993 m .77143 .39409 L .79087 .3839 L p
.32 .784 .661 r
F P
s
P
p
.002 w
.67743 .48239 m .6857 .47693 L .6673 .49853 L p
.752 .961 .666 r
F P
s
P
p
.002 w
.6857 .47693 m .67743 .48239 L .69608 .46169 L p
.712 .943 .641 r
F P
s
P
p
.002 w
.84503 .37542 m .83798 .37323 L .85877 .37389 L p
.134 .632 .908 r
F P
s
P
p
.002 w
.85877 .37389 m .86587 .37796 L .84503 .37542 L p
.134 .632 .908 r
F P
s
P
p
.002 w
.6673 .49853 m .64004 .52415 L .65877 .50345 L closepath p
.755 .971 .699 r
F P
s
P
p
.002 w
.6673 .49853 m .65877 .50345 L .67743 .48239 L p
.752 .961 .666 r
F P
s
P
p
.002 w
.81072 .37709 m .8177 .37539 L .79793 .38084 L p
.204 .723 .701 r
F P
s
P
p
.002 w
.79793 .38084 m .79087 .3839 L .81072 .37709 L p
.204 .723 .701 r
F P
s
P
p
.002 w
.83103 .37339 m .83798 .37323 L .8177 .37539 L p
.116 .668 .751 r
F P
s
P
p
.002 w
.8177 .37539 m .81072 .37709 L .83103 .37339 L p
.116 .668 .751 r
F P
s
P
p
.002 w
.72574 .42945 m .73346 .42373 L .71473 .44196 L p
.597 .896 .609 r
F P
s
P
p
.002 w
.73346 .42373 m .72574 .42945 L .74483 .41304 L p
.519 .861 .597 r
F P
s
P
p
.002 w
.74483 .41304 m .75233 .40758 L .73346 .42373 L p
.519 .861 .597 r
F P
s
P
p
.002 w
.70676 .44768 m .71473 .44196 L .69608 .46169 L p
.662 .925 .63 r
F P
s
P
p
.002 w
.71473 .44196 m .70676 .44768 L .72574 .42945 L p
.597 .896 .609 r
F P
s
P
p
.002 w
.76413 .39902 m .77143 .39409 L .77042 .39481 L p
.423 .817 .593 r
F P
s
P
p
.002 w
.77042 .39481 m .75233 .40758 L p
.76413 .39902 L .423 .817 .593 r
F P
s
P
p
.002 w
.75233 .40758 m .74483 .41304 L .76413 .39902 L p
.423 .817 .593 r
F P
s
P
p
.002 w
.75664 .40395 m .76413 .39902 L .74483 .41304 L p
.448 .833 .606 r
F P
s
P
p
.002 w
.76413 .39902 m .75664 .40395 L .75768 .40334 L p
0 0 0 r
F P
s
P
p
.002 w
.75768 .40334 m .77642 .39232 L p
.76413 .39902 L 0 0 0 r
F P
s
P
p
.002 w
.68785 .46716 m .69608 .46169 L .67743 .48239 L p
.715 .95 .658 r
F P
s
P
p
.002 w
.69608 .46169 m .68785 .46716 L .70676 .44768 L p
.662 .925 .63 r
F P
s
P
p
.002 w
.83798 .37323 m .83103 .37339 L .85183 .37241 L p
.064 .622 .804 r
F P
s
P
p
.002 w
.85183 .37241 m .85877 .37389 L .83798 .37323 L p
.064 .622 .804 r
F P
s
P
p
.002 w
.77143 .39409 m .76413 .39902 L .78372 .38798 L p
0 0 0 r
F P
s
P
p
.002 w
.78372 .38798 m .79087 .3839 L .77143 .39409 L p
0 0 0 r
F P
s
P
p
.002 w
.66894 .48736 m .67743 .48239 L .65877 .50345 L p
.757 .97 .691 r
F P
s
P
p
.002 w
.67743 .48239 m .66894 .48736 L .68785 .46716 L p
.715 .95 .658 r
F P
s
P
p
.002 w
.65 .50769 m .65877 .50345 L .64004 .52415 L p
.785 .985 .729 r
F P
s
P
p
.002 w
.65877 .50345 m .65 .50769 L .66894 .48736 L p
.757 .97 .691 r
F P
s
P
p
.002 w
.79087 .3839 m .78372 .38798 L .8037 .38008 L p
0 0 0 r
F P
s
P
p
.002 w
.8037 .38008 m .81072 .37709 L .79087 .3839 L p
0 0 0 r
F P
s
P
p
.002 w
.64004 .52415 m .61196 .54645 L .631 .52752 L closepath p
.775 .992 .772 r
F P
s
P
p
.002 w
.64004 .52415 m .631 .52752 L .65 .50769 L p
.785 .985 .729 r
F P
s
P
p
.002 w
.81072 .37709 m .8037 .38008 L .82409 .3751 L p
0 0 0 r
F P
s
P
p
.002 w
.82409 .3751 m .83103 .37339 L .81072 .37709 L p
0 0 0 r
F P
s
P
p
.002 w
.84493 .37271 m .85183 .37241 L .83103 .37339 L p
.035 .605 .709 r
F P
s
P
p
.002 w
.83103 .37339 m .82409 .3751 L .84493 .37271 L p
.035 .605 .709 r
F P
s
P
p
.002 w
.71779 .43471 m .72574 .42945 L .70676 .44768 L p
.624 .917 .642 r
F P
s
P
p
.002 w
.72574 .42945 m .71779 .43471 L .73712 .41827 L p
.544 .879 .62 r
F P
s
P
p
.002 w
.73712 .41827 m .74483 .41304 L .72574 .42945 L p
.544 .879 .62 r
F P
s
P
p
.002 w
.69855 .45271 m .70676 .44768 L .68785 .46716 L p
.689 .946 .67 r
F P
s
P
p
.002 w
.70676 .44768 m .69855 .45271 L .71779 .43471 L p
.624 .917 .642 r
F P
s
P
p
.002 w
.74483 .41304 m .73712 .41827 L .75664 .40395 L p
.448 .833 .606 r
F P
s
P
p
.002 w
.67937 .47173 m .68785 .46716 L .66894 .48736 L p
.741 .969 .702 r
F P
s
P
p
.002 w
.68785 .46716 m .67937 .47173 L .69855 .45271 L p
.689 .946 .67 r
F P
s
P
p
.002 w
.77642 .39232 m .78372 .38798 L .77359 .39369 L p
0 0 0 r
F P
s
P
p
.002 w
.78372 .38798 m .77642 .39232 L .79655 .38357 L p
0 0 0 r
F P
s
P
p
.002 w
.77642 .39232 m .76886 .39574 L .77933 .39092 L p
.308 .781 .677 r
F P
s
P
p
.002 w
.77933 .39092 m .78916 .38639 L p
.77642 .39232 L .308 .781 .677 r
F P
s
P
p
.002 w
.77359 .39369 m .76413 .39902 L p
.77642 .39232 L 0 0 0 r
F P
s
P
p
.002 w
.66018 .49124 m .66894 .48736 L .65 .50769 L p
.78 .986 .738 r
F P
s
P
p
.002 w
.66894 .48736 m .66018 .49124 L .67937 .47173 L p
.741 .969 .702 r
F P
s
P
p
.002 w
.64097 .51068 m .65 .50769 L .631 .52752 L p
.804 .996 .776 r
F P
s
P
p
.002 w
.65 .50769 m .64097 .51068 L .66018 .49124 L p
.78 .986 .738 r
F P
s
P
p
.002 w
.62172 .52945 m .631 .52752 L .61196 .54645 L p
.818 1 .814 r
F P
s
P
p
.002 w
.631 .52752 m .62172 .52945 L .64097 .51068 L p
.804 .996 .776 r
F P
s
P
p
.002 w
.79655 .38357 m .8037 .38008 L .78372 .38798 L p
0 0 0 r
F P
s
P
p
.002 w
.61196 .54645 m .58308 .56207 L .60242 .54705 L closepath p
.761 .993 .875 r
F P
s
P
p
.002 w
.61196 .54645 m .60242 .54705 L .62172 .52945 L p
.818 1 .814 r
F P
s
P
p
.002 w
.8037 .38008 m .79655 .38357 L .81706 .37754 L p
0 0 0 r
F P
s
P
p
.002 w
.81706 .37754 m .82409 .3751 L .8037 .38008 L p
0 0 0 r
F P
s
P
p
.002 w
.10572 .33632 m .07988 .29461 L .09485 .30615 L closepath p
.201 .285 .727 r
F P
s
P
p
.002 w
.82409 .3751 m .81706 .37754 L .83799 .37397 L p
0 0 0 r
F P
s
P
p
.002 w
.83799 .37397 m .84493 .37271 L .82409 .3751 L p
0 0 0 r
F P
s
P
p
.002 w
.70953 .43829 m .71779 .43471 L .69855 .45271 L p
.7 .966 .73 r
F P
s
P
p
.002 w
.71779 .43471 m .70953 .43829 L .72912 .42207 L p
.627 .937 .708 r
F P
s
P
p
.002 w
.72912 .42207 m .73712 .41827 L .71779 .43471 L p
.627 .937 .708 r
F P
s
P
p
.002 w
.69003 .45581 m .69855 .45271 L .67937 .47173 L p
.757 .986 .756 r
F P
s
P
p
.002 w
.69855 .45271 m .69003 .45581 L .70953 .43829 L p
.7 .966 .73 r
F P
s
P
p
.002 w
.73712 .41827 m .72912 .42207 L .74887 .4077 L p
.536 .897 .691 r
F P
s
P
p
.002 w
.74887 .4077 m .75664 .40395 L .73712 .41827 L p
.536 .897 .691 r
F P
s
P
p
.002 w
.67057 .47412 m .67937 .47173 L .66018 .49124 L p
.799 .997 .785 r
F P
s
P
p
.002 w
.67937 .47173 m .67057 .47412 L .69003 .45581 L p
.757 .986 .756 r
F P
s
P
p
.002 w
.65112 .49268 m .66018 .49124 L .64097 .51068 L p
.827 1 .814 r
F P
s
P
p
.002 w
.66018 .49124 m .65112 .49268 L .67057 .47412 L p
.799 .997 .785 r
F P
s
P
p
.002 w
.58308 .56207 m .55355 .56742 L .57307 .55727 L closepath p
.724 .91 .959 r
F P
s
P
p
.002 w
.58308 .56207 m .57307 .55727 L .59261 .54414 L p
.813 .954 .909 r
F P
s
P
p
.002 w
.59261 .54414 m .60242 .54705 L .58308 .56207 L p
.813 .954 .909 r
F P
s
P
p
.002 w
.75664 .40395 m .74887 .4077 L .76886 .39574 L p
.424 .843 .68 r
F P
s
P
p
.002 w
.76886 .39574 m .77642 .39232 L .75664 .40395 L p
.424 .843 .68 r
F P
s
P
p
.002 w
.63165 .51094 m .64097 .51068 L .62172 .52945 L p
.841 .994 .844 r
F P
s
P
p
.002 w
.64097 .51068 m .63165 .51094 L .65112 .49268 L p
.827 1 .814 r
F P
s
P
p
.002 w
.60242 .54705 m .59261 .54414 L .61214 .5283 L p
.841 .98 .873 r
F P
s
P
p
.002 w
.61214 .5283 m .62172 .52945 L .60242 .54705 L p
.841 .98 .873 r
F P
s
P
p
.002 w
.62172 .52945 m .61214 .5283 L .63165 .51094 L p
.841 .994 .844 r
F P
s
P
p
.002 w
.78916 .38639 m .79655 .38357 L .77642 .39232 L p
.308 .781 .677 r
F P
s
P
p
.002 w
.13316 .36998 m .10572 .33632 L .12166 .34414 L closepath p
.292 .385 .786 r
F P
s
P
p
.002 w
.10572 .33632 m .09485 .30615 L .11056 .3148 L p
.328 .375 .759 r
F P
s
P
p
.002 w
.11056 .3148 m .12166 .34414 L .10572 .33632 L p
.328 .375 .759 r
F P
s
P
p
.002 w
.80981 .37956 m .81706 .37754 L .79655 .38357 L p
.207 .723 .687 r
F P
s
P
p
.002 w
.79655 .38357 m .78916 .38639 L .80981 .37956 L p
.207 .723 .687 r
F P
s
P
p
.002 w
.55355 .56742 m .52365 .56628 L .54319 .55896 L closepath p
.674 .819 .953 r
F P
s
P
p
.002 w
.55355 .56742 m .54319 .55896 L .56285 .54988 L p
.734 .858 .939 r
F P
s
P
p
.002 w
.56285 .54988 m .57307 .55727 L .55355 .56742 L p
.734 .858 .939 r
F P
s
P
p
.002 w
.57307 .55727 m .56285 .54988 L .58257 .53825 L p
.787 .898 .921 r
F P
s
P
p
.002 w
.58257 .53825 m .59261 .54414 L .57307 .55727 L p
.787 .898 .921 r
F P
s
P
p
.002 w
.83086 .37503 m .83799 .37397 L .81706 .37754 L p
.129 .675 .71 r
F P
s
P
p
.002 w
.81706 .37754 m .80981 .37956 L .83086 .37503 L p
.129 .675 .71 r
F P
s
P
p
.002 w
.59261 .54414 m .58257 .53825 L .6023 .52405 L p
.827 .932 .9 r
F P
s
P
p
.002 w
.6023 .52405 m .61214 .5283 L .59261 .54414 L p
.827 .932 .9 r
F P
s
P
p
.002 w
.68117 .45634 m .69003 .45581 L .67057 .47412 L p
.813 .999 .836 r
F P
s
P
p
.002 w
.69003 .45581 m .68117 .45634 L .70093 .43948 L p
.774 .997 .821 r
F P
s
P
p
.002 w
.70093 .43948 m .70953 .43829 L .69003 .45581 L p
.774 .997 .821 r
F P
s
P
p
.002 w
.70953 .43829 m .70093 .43948 L .72077 .42369 L p
.718 .986 .808 r
F P
s
P
p
.002 w
.72077 .42369 m .72912 .42207 L .70953 .43829 L p
.718 .986 .808 r
F P
s
P
p
.002 w
.66145 .47375 m .67057 .47412 L .65112 .49268 L p
.838 .993 .851 r
F P
s
P
p
.002 w
.67057 .47412 m .66145 .47375 L .68117 .45634 L p
.813 .999 .836 r
F P
s
P
p
.002 w
.52365 .56628 m .49352 .56068 L .51302 .55546 L closepath p
.63 .758 .939 r
F P
s
P
p
.002 w
.52365 .56628 m .51302 .55546 L .53267 .54881 L p
.677 .788 .933 r
F P
s
P
p
.002 w
.53267 .54881 m .54319 .55896 L .52365 .56628 L p
.677 .788 .933 r
F P
s
P
p
.002 w
.16156 .3989 m .13316 .36998 L .14983 .37502 L closepath p
.376 .456 .816 r
F P
s
P
p
.002 w
.13316 .36998 m .12166 .34414 L .1382 .3495 L p
.388 .448 .801 r
F P
s
P
p
.002 w
.1382 .3495 m .14983 .37502 L .13316 .36998 L p
.388 .448 .801 r
F P
s
P
p
.002 w
.61214 .5283 m .6023 .52405 L .62203 .50817 L p
.844 .959 .884 r
F P
s
P
p
.002 w
.62203 .50817 m .63165 .51094 L .61214 .5283 L p
.844 .959 .884 r
F P
s
P
p
.002 w
.64175 .49121 m .65112 .49268 L .63165 .51094 L p
.848 .98 .867 r
F P
s
P
p
.002 w
.65112 .49268 m .64175 .49121 L .66145 .47375 L p
.838 .993 .851 r
F P
s
P
p
.002 w
.63165 .51094 m .62203 .50817 L .64175 .49121 L p
.848 .98 .867 r
F P
s
P
p
.002 w
.72912 .42207 m .72077 .42369 L .74076 .40949 L p
.643 .962 .799 r
F P
s
P
p
.002 w
.74076 .40949 m .74887 .4077 L .72912 .42207 L p
.643 .962 .799 r
F P
s
P
p
.002 w
.74887 .4077 m .74076 .40949 L .76097 .39741 L p
.544 .923 .794 r
F P
s
P
p
.002 w
.76097 .39741 m .76886 .39574 L .74887 .4077 L p
.544 .923 .794 r
F P
s
P
p
.002 w
.12166 .34414 m .11056 .3148 L .12693 .32081 L p
.394 .423 .773 r
F P
s
P
p
.002 w
.12693 .32081 m .1382 .3495 L .12166 .34414 L p
.394 .423 .773 r
F P
s
P
p
.002 w
.49352 .56068 m .4633 .55049 L .48268 .54741 L closepath p
.584 .696 .919 r
F P
s
P
p
.002 w
.49352 .56068 m .48268 .54741 L .50226 .54286 L p
.633 .728 .916 r
F P
s
P
p
.002 w
.50226 .54286 m .51302 .55546 L .49352 .56068 L p
.633 .728 .916 r
F P
s
P
p
.002 w
.54319 .55896 m .53267 .54881 L .55245 .54057 L p
.712 .815 .929 r
F P
s
P
p
.002 w
.55245 .54057 m .56285 .54988 L .54319 .55896 L p
.712 .815 .929 r
F P
s
P
p
.002 w
.76886 .39574 m .76097 .39741 L .78146 .3877 L p
.436 .871 .792 r
F P
s
P
p
.002 w
.78146 .3877 m .78916 .38639 L .76886 .39574 L p
.436 .871 .792 r
F P
s
P
p
.002 w
.4633 .55049 m .43299 .54011 L .45227 .53652 L closepath p
.597 .69 .908 r
F P
s
P
p
.002 w
.4633 .55049 m .45227 .53652 L .47172 .53325 L p
.586 .684 .909 r
F P
s
P
p
.002 w
.47172 .53325 m .48268 .54741 L .4633 .55049 L p
.586 .684 .909 r
F P
s
P
p
.002 w
.18988 .43125 m .16156 .3989 L .17844 .40491 L closepath p
.374 .431 .791 r
F P
s
P
p
.002 w
.16156 .3989 m .14983 .37502 L .16676 .38001 L p
.361 .438 .806 r
F P
s
P
p
.002 w
.16676 .38001 m .17844 .40491 L .16156 .3989 L p
.361 .438 .806 r
F P
s
P
p
.002 w
.43299 .54011 m .40251 .53198 L .42173 .52731 L closepath p
.618 .72 .919 r
F P
s
P
p
.002 w
.43299 .54011 m .42173 .52731 L .44109 .52318 L p
.595 .703 .919 r
F P
s
P
p
.002 w
.44109 .52318 m .45227 .53652 L .43299 .54011 L p
.595 .703 .919 r
F P
s
P
p
.002 w
.56285 .54988 m .55245 .54057 L .57231 .53011 L p
.758 .85 .919 r
F P
s
P
p
.002 w
.57231 .53011 m .58257 .53825 L .56285 .54988 L p
.758 .85 .919 r
F P
s
P
p
.002 w
.78916 .38639 m .78146 .3877 L .80227 .38028 L p
.335 .816 .796 r
F P
s
P
p
.002 w
.80227 .38028 m .80981 .37956 L .78916 .38639 L p
.335 .816 .796 r
F P
s
P
p
.002 w
.51302 .55546 m .50226 .54286 L .52199 .53686 L p
.662 .752 .916 r
F P
s
P
p
.002 w
.52199 .53686 m .53267 .54881 L .51302 .55546 L p
.662 .752 .916 r
F P
s
P
p
.002 w
.14983 .37502 m .1382 .3495 L .15511 .35417 L p
.393 .449 .8 r
F P
s
P
p
.002 w
.15511 .35417 m .16676 .38001 L .14983 .37502 L p
.393 .449 .8 r
F P
s
P
p
.002 w
.58257 .53825 m .57231 .53011 L .59222 .51729 L p
.8 .885 .906 r
F P
s
P
p
.002 w
.59222 .51729 m .6023 .52405 L .58257 .53825 L p
.8 .885 .906 r
F P
s
P
p
.002 w
.40251 .53198 m .37183 .52492 L .39099 .51975 L closepath p
.628 .735 .925 r
F P
s
P
p
.002 w
.40251 .53198 m .39099 .51975 L .4103 .51478 L p
.617 .728 .925 r
F P
s
P
p
.002 w
.4103 .51478 m .42173 .52731 L .40251 .53198 L p
.617 .728 .925 r
F P
s
P
p
.002 w
.48268 .54741 m .47172 .53325 L .49136 .52901 L p
.618 .704 .906 r
F P
s
P
p
.002 w
.49136 .52901 m .50226 .54286 L .48268 .54741 L p
.618 .704 .906 r
F P
s
P
p
.002 w
.6023 .52405 m .59222 .51729 L .61215 .50277 L p
.826 .915 .898 r
F P
s
P
p
.002 w
.61215 .50277 m .62203 .50817 L .6023 .52405 L p
.826 .915 .898 r
F P
s
P
p
.002 w
.62203 .50817 m .61215 .50277 L .63209 .48708 L p
.84 .941 .89 r
F P
s
P
p
.002 w
.63209 .48708 m .64175 .49121 L .62203 .50817 L p
.84 .941 .89 r
F P
s
P
p
.002 w
.80981 .37956 m .80227 .38028 L .82346 .37499 L p
.253 .765 .808 r
F P
s
P
p
.002 w
.82346 .37499 m .83086 .37503 L .80981 .37956 L p
.253 .765 .808 r
F P
s
P
p
.002 w
.21795 .46974 m .18988 .43125 L .20669 .44134 L closepath p
.281 .356 .763 r
F P
s
P
p
.002 w
.18988 .43125 m .17844 .40491 L .19535 .4131 L p
.278 .356 .764 r
F P
s
P
p
.002 w
.19535 .4131 m .20669 .44134 L .18988 .43125 L p
.278 .356 .764 r
F P
s
P
p
.002 w
.1382 .3495 m .12693 .32081 L .14373 .32563 L p
.412 .437 .777 r
F P
s
P
p
.002 w
.14373 .32563 m .15511 .35417 L .1382 .3495 L p
.412 .437 .777 r
F P
s
P
p
.002 w
.64175 .49121 m .63209 .48708 L .65203 .47075 L p
.842 .962 .884 r
F P
s
P
p
.002 w
.65203 .47075 m .66145 .47375 L .64175 .49121 L p
.842 .962 .884 r
F P
s
P
p
.002 w
.66145 .47375 m .65203 .47075 L .672 .4543 L p
.833 .979 .88 r
F P
s
P
p
.002 w
.672 .4543 m .68117 .45634 L .66145 .47375 L p
.833 .979 .88 r
F P
s
P
p
.002 w
.53267 .54881 m .52199 .53686 L .54188 .52936 L p
.693 .776 .915 r
F P
s
P
p
.002 w
.54188 .52936 m .55245 .54057 L .53267 .54881 L p
.693 .776 .915 r
F P
s
P
p
.002 w
.45227 .53652 m .44109 .52318 L .46062 .51946 L p
.587 .694 .916 r
F P
s
P
p
.002 w
.46062 .51946 m .47172 .53325 L .45227 .53652 L p
.587 .694 .916 r
F P
s
P
p
.002 w
.68117 .45634 m .672 .4543 L .69201 .43821 L p
.811 .99 .877 r
F P
s
P
p
.002 w
.69201 .43821 m .70093 .43948 L .68117 .45634 L p
.811 .99 .877 r
F P
s
P
p
.002 w
.37183 .52492 m .34098 .51818 L .36007 .51281 L closepath p
.631 .74 .926 r
F P
s
P
p
.002 w
.37183 .52492 m .36007 .51281 L .37931 .50748 L p
.628 .737 .926 r
F P
s
P
p
.002 w
.37931 .50748 m .39099 .51975 L .37183 .52492 L p
.628 .737 .926 r
F P
s
P
p
.002 w
.70093 .43948 m .69201 .43821 L .71209 .42297 L p
.774 .994 .876 r
F P
s
P
p
.002 w
.71209 .42297 m .72077 .42369 L .70093 .43948 L p
.774 .994 .876 r
F P
s
P
p
.002 w
.72077 .42369 m .71209 .42297 L .73232 .40909 L p
.719 .987 .878 r
F P
s
P
p
.002 w
.73232 .40909 m .74076 .40949 L .72077 .42369 L p
.719 .987 .878 r
F P
s
P
p
.002 w
.17844 .40491 m .16676 .38001 L .18378 .38657 L p
.31 .393 .786 r
F P
s
P
p
.002 w
.18378 .38657 m .19535 .4131 L .17844 .40491 L p
.31 .393 .786 r
F P
s
P
p
.002 w
.42173 .52731 m .4103 .51478 L .42975 .51023 L p
.604 .717 .924 r
F P
s
P
p
.002 w
.42975 .51023 m .44109 .52318 L .42173 .52731 L p
.604 .717 .924 r
F P
s
P
p
.002 w
.74076 .40949 m .73232 .40909 L .75274 .39706 L p
.642 .966 .883 r
F P
s
P
p
.002 w
.75274 .39706 m .76097 .39741 L .74076 .40949 L p
.642 .966 .883 r
F P
s
P
p
.002 w
.50226 .54286 m .49136 .52901 L .51118 .52349 L p
.647 .724 .905 r
F P
s
P
p
.002 w
.51118 .52349 m .52199 .53686 L .50226 .54286 L p
.647 .724 .905 r
F P
s
P
p
.002 w
.34098 .51818 m .30986 .51289 L .32894 .50661 L closepath p
.651 .757 .927 r
F P
s
P
p
.002 w
.34098 .51818 m .32894 .50661 L .34813 .50087 L p
.631 .746 .931 r
F P
s
P
p
.002 w
.34813 .50087 m .36007 .51281 L .34098 .51818 L p
.631 .746 .931 r
F P
s
P
p
.002 w
.55245 .54057 m .54188 .52936 L .56187 .51989 L p
.733 .807 .91 r
F P
s
P
p
.002 w
.56187 .51989 m .57231 .53011 L .55245 .54057 L p
.733 .807 .91 r
F P
s
P
p
.002 w
.39099 .51975 m .37931 .50748 L .39871 .50228 L p
.622 .733 .926 r
F P
s
P
p
.002 w
.39871 .50228 m .4103 .51478 L .39099 .51975 L p
.622 .733 .926 r
F P
s
P
p
.002 w
.76097 .39741 m .75274 .39706 L .77342 .38715 L p
.551 .931 .889 r
F P
s
P
p
.002 w
.77342 .38715 m .78146 .3877 L .76097 .39741 L p
.551 .931 .889 r
F P
s
P
p
.002 w
.47172 .53325 m .46062 .51946 L .48033 .51515 L p
.609 .703 .911 r
F P
s
P
p
.002 w
.48033 .51515 m .49136 .52901 L .47172 .53325 L p
.609 .703 .911 r
F P
s
P
p
.002 w
.16676 .38001 m .15511 .35417 L .17217 .35976 L p
.356 .42 .79 r
F P
s
P
p
.002 w
.17217 .35976 m .18378 .38657 L .16676 .38001 L p
.356 .42 .79 r
F P
s
P
p
.002 w
.57231 .53011 m .56187 .51989 L .58193 .50824 L p
.773 .839 .902 r
F P
s
P
p
.002 w
.58193 .50824 m .59222 .51729 L .57231 .53011 L p
.773 .839 .902 r
F P
s
P
p
.002 w
.30986 .51289 m .27855 .50748 L .29754 .50156 L closepath p
.641 .76 .935 r
F P
s
P
p
.002 w
.30986 .51289 m .29754 .50156 L .31671 .49522 L p
.65 .763 .932 r
F P
s
P
p
.002 w
.31671 .49522 m .32894 .50661 L .30986 .51289 L p
.65 .763 .932 r
F P
s
P
p
.002 w
.44109 .52318 m .42975 .51023 L .44936 .50608 L p
.594 .707 .922 r
F P
s
P
p
.002 w
.44936 .50608 m .46062 .51946 L .44109 .52318 L p
.594 .707 .922 r
F P
s
P
p
.002 w
.52199 .53686 m .51118 .52349 L .53115 .51659 L p
.676 .746 .904 r
F P
s
P
p
.002 w
.53115 .51659 m .54188 .52936 L .52199 .53686 L p
.676 .746 .904 r
F P
s
P
p
.002 w
.36007 .51281 m .34813 .50087 L .36747 .4953 L p
.629 .742 .929 r
F P
s
P
p
.002 w
.36747 .4953 m .37931 .50748 L .36007 .51281 L p
.629 .742 .929 r
F P
s
P
p
.002 w
.24756 .4961 m .21795 .46974 L .23561 .47516 L closepath p
.356 .481 .848 r
F P
s
P
p
.002 w
.21795 .46974 m .20669 .44134 L .22411 .44839 L p
.407 .449 .792 r
F P
s
P
p
.002 w
.22411 .44839 m .23561 .47516 L .21795 .46974 L p
.407 .449 .792 r
F P
s
P
p
.002 w
.59222 .51729 m .58193 .50824 L .60204 .49494 L p
.801 .868 .898 r
F P
s
P
p
.002 w
.60204 .49494 m .61215 .50277 L .59222 .51729 L p
.801 .868 .898 r
F P
s
P
p
.002 w
.78146 .3877 m .77342 .38715 L .79441 .37932 L p
.46 .886 .895 r
F P
s
P
p
.002 w
.79441 .37932 m .80227 .38028 L .78146 .3877 L p
.46 .886 .895 r
F P
s
P
p
.002 w
.15511 .35417 m .14373 .32563 L .16075 .33079 L p
.392 .423 .774 r
F P
s
P
p
.002 w
.16075 .33079 m .17217 .35976 L .15511 .35417 L p
.392 .423 .774 r
F P
s
P
p
.002 w
.61215 .50277 m .60204 .49494 L .62218 .48043 L p
.819 .895 .896 r
F P
s
P
p
.002 w
.62218 .48043 m .63209 .48708 L .61215 .50277 L p
.819 .895 .896 r
F P
s
P
p
.002 w
.4103 .51478 m .39871 .50228 L .41825 .4974 L p
.612 .725 .926 r
F P
s
P
p
.002 w
.41825 .4974 m .42975 .51023 L .4103 .51478 L p
.612 .725 .926 r
F P
s
P
p
.002 w
.20669 .44134 m .19535 .4131 L .21271 .41998 L p
.37 .414 .776 r
F P
s
P
p
.002 w
.21271 .41998 m .22411 .44839 L .20669 .44134 L p
.37 .414 .776 r
F P
s
P
p
.002 w
.49136 .52901 m .48033 .51515 L .50023 .5098 L p
.636 .717 .906 r
F P
s
P
p
.002 w
.50023 .5098 m .51118 .52349 L .49136 .52901 L p
.636 .717 .906 r
F P
s
P
p
.002 w
.63209 .48708 m .62218 .48043 L .64233 .4652 L p
.829 .919 .896 r
F P
s
P
p
.002 w
.64233 .4652 m .65203 .47075 L .63209 .48708 L p
.829 .919 .896 r
F P
s
P
p
.002 w
.27855 .50748 m .24756 .4961 L .26615 .49391 L closepath p
.546 .687 .932 r
F P
s
P
p
.002 w
.27855 .50748 m .26615 .49391 L .28515 .48874 L p
.642 .737 .917 r
F P
s
P
p
.002 w
.28515 .48874 m .29754 .50156 L .27855 .50748 L p
.642 .737 .917 r
F P
s
P
p
.002 w
.65203 .47075 m .64233 .4652 L .66252 .44972 L p
.83 .941 .898 r
F P
s
P
p
.002 w
.66252 .44972 m .672 .4543 L .65203 .47075 L p
.83 .941 .898 r
F P
s
P
p
.002 w
.32894 .50661 m .31671 .49522 L .336 .48919 L p
.637 .755 .934 r
F P
s
P
p
.002 w
.336 .48919 m .34813 .50087 L .32894 .50661 L p
.637 .755 .934 r
F P
s
P
p
.002 w
.54188 .52936 m .53115 .51659 L .55126 .5079 L p
.711 .772 .9 r
F P
s
P
p
.002 w
.55126 .5079 m .56187 .51989 L .54188 .52936 L p
.711 .772 .9 r
F P
s
P
p
.002 w
.672 .4543 m .66252 .44972 L .68276 .43444 L p
.82 .958 .903 r
F P
s
P
p
.002 w
.68276 .43444 m .69201 .43821 L .672 .4543 L p
.82 .958 .903 r
F P
s
P
p
.002 w
.19535 .4131 m .18378 .38657 L .20116 .39255 L p
.36 .417 .785 r
F P
s
P
p
.002 w
.20116 .39255 m .21271 .41998 L .19535 .4131 L p
.36 .417 .785 r
F P
s
P
p
.002 w
.24756 .4961 m .23561 .47516 L .25401 .47518 L p
.562 .618 .866 r
F P
s
P
p
.002 w
.25401 .47518 m .26615 .49391 L .24756 .4961 L p
.562 .618 .866 r
F P
s
P
p
.002 w
.80227 .38028 m .79441 .37932 L .81574 .37344 L p
.38 .839 .902 r
F P
s
P
p
.002 w
.81574 .37344 m .82346 .37499 L .80227 .38028 L p
.38 .839 .902 r
F P
s
P
p
.002 w
.46062 .51946 m .44936 .50608 L .46915 .50154 L p
.608 .71 .917 r
F P
s
P
p
.002 w
.46915 .50154 m .48033 .51515 L .46062 .51946 L p
.608 .71 .917 r
F P
s
P
p
.002 w
.69201 .43821 m .68276 .43444 L .70308 .41983 L p
.798 .97 .91 r
F P
s
P
p
.002 w
.70308 .41983 m .71209 .42297 L .69201 .43821 L p
.798 .97 .91 r
F P
s
P
p
.002 w
.37931 .50748 m .36747 .4953 L .38695 .48986 L p
.625 .738 .928 r
F P
s
P
p
.002 w
.38695 .48986 m .39871 .50228 L .37931 .50748 L p
.625 .738 .928 r
F P
s
P
p
.002 w
.71209 .42297 m .70308 .41983 L .72353 .40635 L p
.761 .974 .921 r
F P
s
P
p
.002 w
.72353 .40635 m .73232 .40909 L .71209 .42297 L p
.761 .974 .921 r
F P
s
P
p
.002 w
.56187 .51989 m .55126 .5079 L .57145 .49725 L p
.748 .801 .895 r
F P
s
P
p
.002 w
.57145 .49725 m .58193 .50824 L .56187 .51989 L p
.748 .801 .895 r
F P
s
P
p
.002 w
.42975 .51023 m .41825 .4974 L .43795 .49288 L p
.602 .716 .925 r
F P
s
P
p
.002 w
.43795 .49288 m .44936 .50608 L .42975 .51023 L p
.602 .716 .925 r
F P
s
P
p
.002 w
.18378 .38657 m .17217 .35976 L .18956 .36495 L p
.379 .429 .786 r
F P
s
P
p
.002 w
.18956 .36495 m .20116 .39255 L .18378 .38657 L p
.379 .429 .786 r
F P
s
P
p
.002 w
.51118 .52349 m .50023 .5098 L .52029 .50325 L p
.662 .735 .903 r
F P
s
P
p
.002 w
.52029 .50325 m .53115 .51659 L .51118 .52349 L p
.662 .735 .903 r
F P
s
P
p
.002 w
.29754 .50156 m .28515 .48874 L .30437 .48272 L p
.65 .746 .92 r
F P
s
P
p
.002 w
.30437 .48272 m .31671 .49522 L .29754 .50156 L p
.65 .746 .92 r
F P
s
P
p
.002 w
.73232 .40909 m .72353 .40635 L .74417 .3945 L p
.705 .968 .935 r
F P
s
P
p
.002 w
.74417 .3945 m .75274 .39706 L .73232 .40909 L p
.705 .968 .935 r
F P
s
P
p
.002 w
.34813 .50087 m .336 .48919 L .35543 .48335 L p
.632 .751 .933 r
F P
s
P
p
.002 w
.35543 .48335 m .36747 .4953 L .34813 .50087 L p
.632 .751 .933 r
F P
s
P
p
.002 w
.23561 .47516 m .22411 .44839 L .24229 .45046 L p
.522 .534 .81 r
F P
s
P
p
.002 w
.24229 .45046 m .25401 .47518 L .23561 .47516 L p
.522 .534 .81 r
F P
s
P
p
.002 w
.58193 .50824 m .57145 .49725 L .59172 .48501 L p
.775 .827 .893 r
F P
s
P
p
.002 w
.59172 .48501 m .60204 .49494 L .58193 .50824 L p
.775 .827 .893 r
F P
s
P
p
.002 w
.17217 .35976 m .16075 .33079 L .17811 .33548 L p
.405 .429 .772 r
F P
s
P
p
.002 w
.17811 .33548 m .18956 .36495 L .17217 .35976 L p
.405 .429 .772 r
F P
s
P
p
.002 w
.39871 .50228 m .38695 .48986 L .40658 .4847 L p
.617 .732 .928 r
F P
s
P
p
.002 w
.40658 .4847 m .41825 .4974 L .39871 .50228 L p
.617 .732 .928 r
F P
s
P
p
.002 w
.48033 .51515 m .46915 .50154 L .48913 .49619 L p
.63 .72 .912 r
F P
s
P
p
.002 w
.48913 .49619 m .50023 .5098 L .48033 .51515 L p
.63 .72 .912 r
F P
s
P
p
.002 w
.75274 .39706 m .74417 .3945 L .76505 .38452 L p
.634 .947 .949 r
F P
s
P
p
.002 w
.76505 .38452 m .77342 .38715 L .75274 .39706 L p
.634 .947 .949 r
F P
s
P
p
.002 w
.26615 .49391 m .25401 .47518 L .27292 .47163 L p
.623 .667 .871 r
F P
s
P
p
.002 w
.27292 .47163 m .28515 .48874 L .26615 .49391 L p
.623 .667 .871 r
F P
s
P
p
.002 w
.60204 .49494 m .59172 .48501 L .61203 .47157 L p
.795 .852 .893 r
F P
s
P
p
.002 w
.61203 .47157 m .62218 .48043 L .60204 .49494 L p
.795 .852 .893 r
F P
s
P
p
.002 w
.53115 .51659 m .52029 .50325 L .54049 .49513 L p
.695 .756 .899 r
F P
s
P
p
.002 w
.54049 .49513 m .55126 .5079 L .53115 .51659 L p
.695 .756 .899 r
F P
s
P
p
.002 w
.22411 .44839 m .21271 .41998 L .23074 .42296 L p
.483 .49 .788 r
F P
s
P
p
.002 w
.23074 .42296 m .24229 .45046 L .22411 .44839 L p
.483 .49 .788 r
F P
s
P
p
.002 w
.31671 .49522 m .30437 .48272 L .32374 .4767 L p
.642 .746 .924 r
F P
s
P
p
.002 w
.32374 .4767 m .336 .48919 L .31671 .49522 L p
.642 .746 .924 r
F P
s
P
p
.002 w
.44936 .50608 m .43795 .49288 L .45782 .4881 L p
.61 .716 .92 r
F P
s
P
p
.002 w
.45782 .4881 m .46915 .50154 L .44936 .50608 L p
.61 .716 .92 r
F P
s
P
p
.002 w
.62218 .48043 m .61203 .47157 L .63238 .45737 L p
.809 .876 .896 r
F P
s
P
p
.002 w
.63238 .45737 m .64233 .4652 L .62218 .48043 L p
.809 .876 .896 r
F P
s
P
p
.002 w
.36747 .4953 m .35543 .48335 L .37501 .47766 L p
.628 .746 .933 r
F P
s
P
p
.002 w
.37501 .47766 m .38695 .48986 L .36747 .4953 L p
.628 .746 .933 r
F P
s
P
p
.002 w
.77342 .38715 m .76505 .38452 L .78621 .37642 L p
.559 .916 .959 r
F P
s
P
p
.002 w
.78621 .37642 m .79441 .37932 L .77342 .38715 L p
.559 .916 .959 r
F P
s
P
p
.002 w
.64233 .4652 m .63238 .45737 L .65278 .44281 L p
.814 .898 .901 r
F P
s
P
p
.002 w
.65278 .44281 m .66252 .44972 L .64233 .4652 L p
.814 .898 .901 r
F P
s
P
p
.002 w
.21271 .41998 m .20116 .39255 L .21912 .39558 L p
.458 .478 .791 r
F P
s
P
p
.002 w
.21912 .39558 m .23074 .42296 L .21271 .41998 L p
.458 .478 .791 r
F P
s
P
p
.002 w
.50023 .5098 m .48913 .49619 L .50928 .48981 L p
.654 .733 .907 r
F P
s
P
p
.002 w
.50928 .48981 m .52029 .50325 L .50023 .5098 L p
.654 .733 .907 r
F P
s
P
p
.002 w
.41825 .4974 m .40658 .4847 L .42637 .47985 L p
.608 .724 .928 r
F P
s
P
p
.002 w
.42637 .47985 m .43795 .49288 L .41825 .4974 L p
.608 .724 .928 r
F P
s
P
p
.002 w
.66252 .44972 m .65278 .44281 L .67323 .42834 L p
.811 .916 .909 r
F P
s
P
p
.002 w
.67323 .42834 m .68276 .43444 L .66252 .44972 L p
.811 .916 .909 r
F P
s
P
p
.002 w
.55126 .5079 m .54049 .49513 L .56081 .48523 L p
.728 .781 .894 r
F P
s
P
p
.002 w
.56081 .48523 m .57145 .49725 L .55126 .5079 L p
.728 .781 .894 r
F P
s
P
p
.002 w
.28515 .48874 m .27292 .47163 L .29212 .46656 L p
.64 .688 .879 r
F P
s
P
p
.002 w
.29212 .46656 m .30437 .48272 L .28515 .48874 L p
.64 .688 .879 r
F P
s
P
p
.002 w
.25401 .47518 m .24229 .45046 L .26103 .44872 L p
.587 .586 .819 r
F P
s
P
p
.002 w
.26103 .44872 m .27292 .47163 L .25401 .47518 L p
.587 .586 .819 r
F P
s
P
p
.002 w
.68276 .43444 m .67323 .42834 L .69377 .41437 L p
.799 .931 .92 r
F P
s
P
p
.002 w
.69377 .41437 m .70308 .41983 L .68276 .43444 L p
.799 .931 .92 r
F P
s
P
p
.002 w
.20116 .39255 m .18956 .36495 L .20748 .36778 L p
.454 .473 .788 r
F P
s
P
p
.002 w
.20748 .36778 m .21912 .39558 L .20116 .39255 L p
.454 .473 .788 r
F P
s
P
p
.002 w
.336 .48919 m .32374 .4767 L .34325 .47077 L p
.637 .744 .926 r
F P
s
P
p
.002 w
.34325 .47077 m .35543 .48335 L .336 .48919 L p
.637 .744 .926 r
F P
s
P
p
.002 w
.79441 .37932 m .78621 .37642 L .80769 .3701 L p
.488 .877 .966 r
F P
s
P
p
.002 w
.80769 .3701 m .81574 .37344 L .79441 .37932 L p
.488 .877 .966 r
F P
s
P
p
.002 w
.46915 .50154 m .45782 .4881 L .47789 .48266 L p
.628 .723 .916 r
F P
s
P
p
.002 w
.47789 .48266 m .48913 .49619 L .46915 .50154 L p
.628 .723 .916 r
F P
s
P
p
.002 w
.70308 .41983 m .69377 .41437 L .71443 .40135 L p
.774 .94 .935 r
F P
s
P
p
.002 w
.71443 .40135 m .72353 .40635 L .70308 .41983 L p
.774 .94 .935 r
F P
s
P
p
.002 w
.38695 .48986 m .37501 .47766 L .39474 .47222 L p
.621 .74 .932 r
F P
s
P
p
.002 w
.39474 .47222 m .40658 .4847 L .38695 .48986 L p
.621 .74 .932 r
F P
s
P
p
.002 w
.57145 .49725 m .56081 .48523 L .58121 .47385 L p
.755 .804 .892 r
F P
s
P
p
.002 w
.58121 .47385 m .59172 .48501 L .57145 .49725 L p
.755 .804 .892 r
F P
s
P
p
.002 w
.18956 .36495 m .17811 .33548 L .19598 .33815 L p
.46 .463 .774 r
F P
s
P
p
.002 w
.19598 .33815 m .20748 .36778 L .18956 .36495 L p
.46 .463 .774 r
F P
s
P
p
.002 w
.52029 .50325 m .50928 .48981 L .52958 .48206 L p
.683 .751 .903 r
F P
s
P
p
.002 w
.52958 .48206 m .54049 .49513 L .52029 .50325 L p
.683 .751 .903 r
F P
s
P
p
.002 w
.72353 .40635 m .71443 .40135 L .73527 .38974 L p
.734 .941 .952 r
F P
s
P
p
.002 w
.73527 .38974 m .74417 .3945 L .72353 .40635 L p
.734 .941 .952 r
F P
s
P
p
.002 w
.24229 .45046 m .23074 .42296 L .24934 .42246 L p
.553 .54 .795 r
F P
s
P
p
.002 w
.24934 .42246 m .26103 .44872 L .24229 .45046 L p
.553 .54 .795 r
F P
s
P
p
.002 w
.30437 .48272 m .29212 .46656 L .31151 .46106 L p
.639 .697 .887 r
F P
s
P
p
.002 w
.31151 .46106 m .32374 .4767 L .30437 .48272 L p
.639 .697 .887 r
F P
s
P
p
.002 w
.43795 .49288 m .42637 .47985 L .44634 .47481 L p
.613 .723 .924 r
F P
s
P
p
.002 w
.44634 .47481 m .45782 .4881 L .43795 .49288 L p
.613 .723 .924 r
F P
s
P
p
.002 w
.59172 .48501 m .58121 .47385 L .60168 .46132 L p
.775 .826 .893 r
F P
s
P
p
.002 w
.60168 .46132 m .61203 .47157 L .59172 .48501 L p
.775 .826 .893 r
F P
s
P
p
.002 w
.35543 .48335 m .34325 .47077 L .36292 .46493 L p
.632 .741 .927 r
F P
s
P
p
.002 w
.36292 .46493 m .37501 .47766 L .35543 .48335 L p
.632 .741 .927 r
F P
s
P
p
.002 w
.27292 .47163 m .26103 .44872 L .28013 .44499 L p
.614 .614 .827 r
F P
s
P
p
.002 w
.28013 .44499 m .29212 .46656 L .27292 .47163 L p
.614 .614 .827 r
F P
s
P
p
.002 w
.23074 .42296 m .21912 .39558 L .23763 .3956 L p
.528 .523 .793 r
F P
s
P
p
.002 w
.23763 .3956 m .24934 .42246 L .23074 .42296 L p
.528 .523 .793 r
F P
s
P
p
.002 w
.74417 .3945 m .73527 .38974 L .75634 .37979 L p
.68 .93 .969 r
F P
s
P
p
.002 w
.75634 .37979 m .76505 .38452 L .74417 .3945 L p
.68 .93 .969 r
F P
s
P
p
.002 w
.61203 .47157 m .60168 .46132 L .62221 .44801 L p
.788 .847 .896 r
F P
s
P
p
.002 w
.62221 .44801 m .63238 .45737 L .61203 .47157 L p
.788 .847 .896 r
F P
s
P
p
.002 w
.48913 .49619 m .47789 .48266 L .49812 .47634 L p
.648 .733 .911 r
F P
s
P
p
.002 w
.49812 .47634 m .50928 .48981 L .48913 .49619 L p
.648 .733 .911 r
F P
s
P
p
.002 w
.40658 .4847 m .39474 .47222 L .41462 .46705 L p
.613 .733 .932 r
F P
s
P
p
.002 w
.41462 .46705 m .42637 .47985 L .40658 .4847 L p
.613 .733 .932 r
F P
s
P
p
.002 w
.54049 .49513 m .52958 .48206 L .55001 .47272 L p
.714 .771 .897 r
F P
s
P
p
.002 w
.55001 .47272 m .56081 .48523 L .54049 .49513 L p
.714 .771 .897 r
F P
s
P
p
.002 w
.63238 .45737 m .62221 .44801 L .6428 .43429 L p
.796 .867 .901 r
F P
s
P
p
.002 w
.6428 .43429 m .65278 .44281 L .63238 .45737 L p
.796 .867 .901 r
F P
s
P
p
.002 w
.32374 .4767 m .31151 .46106 L .33107 .4554 L p
.637 .701 .892 r
F P
s
P
p
.002 w
.33107 .4554 m .34325 .47077 L .32374 .4767 L p
.637 .701 .892 r
F P
s
P
p
.002 w
.21912 .39558 m .20748 .36778 L .22592 .368 L p
.516 .511 .789 r
F P
s
P
p
.002 w
.22592 .368 m .23763 .3956 L .21912 .39558 L p
.516 .511 .789 r
F P
s
P
p
.002 w
.45782 .4881 m .44634 .47481 L .46649 .46923 L p
.628 .728 .919 r
F P
s
P
p
.002 w
.46649 .46923 m .47789 .48266 L .45782 .4881 L p
.628 .728 .919 r
F P
s
P
p
.002 w
.65278 .44281 m .6428 .43429 L .66345 .42056 L p
.796 .885 .909 r
F P
s
P
p
.002 w
.66345 .42056 m .67323 .42834 L .65278 .44281 L p
.796 .885 .909 r
F P
s
P
p
.002 w
.76505 .38452 m .75634 .37979 L .77767 .37152 L p
.621 .909 .983 r
F P
s
P
p
.002 w
.77767 .37152 m .78621 .37642 L .76505 .38452 L p
.621 .909 .983 r
F P
s
P
p
.002 w
.26103 .44872 m .24934 .42246 L .26833 .41982 L p
.586 .568 .8 r
F P
s
P
p
.002 w
.26833 .41982 m .28013 .44499 L .26103 .44872 L p
.586 .568 .8 r
F P
s
P
p
.002 w
.37501 .47766 m .36292 .46493 L .38274 .45929 L p
.626 .737 .927 r
F P
s
P
p
.002 w
.38274 .45929 m .39474 .47222 L .37501 .47766 L p
.626 .737 .927 r
F P
s
P
p
.002 w
.56081 .48523 m .55001 .47272 L .57055 .46202 L p
.738 .791 .895 r
F P
s
P
p
.002 w
.57055 .46202 m .58121 .47385 L .56081 .48523 L p
.738 .791 .895 r
F P
s
P
p
.002 w
.29212 .46656 m .28013 .44499 L .29948 .44039 L p
.622 .628 .836 r
F P
s
P
p
.002 w
.29948 .44039 m .31151 .46106 L .29212 .46656 L p
.622 .628 .836 r
F P
s
P
p
.002 w
.50928 .48981 m .49812 .47634 L .51852 .46882 L p
.674 .748 .906 r
F P
s
P
p
.002 w
.51852 .46882 m .52958 .48206 L .50928 .48981 L p
.674 .748 .906 r
F P
s
P
p
.002 w
.67323 .42834 m .66345 .42056 L .68419 .40721 L p
.788 .899 .921 r
F P
s
P
p
.002 w
.68419 .40721 m .69377 .41437 L .67323 .42834 L p
.788 .899 .921 r
F P
s
P
p
.002 w
.20748 .36778 m .19598 .33815 L .21434 .33851 L p
.511 .495 .775 r
F P
s
P
p
.002 w
.21434 .33851 m .22592 .368 L .20748 .36778 L p
.511 .495 .775 r
F P
s
P
p
.002 w
.42637 .47985 m .41462 .46705 L .43468 .46175 L p
.616 .732 .928 r
F P
s
P
p
.002 w
.43468 .46175 m .44634 .47481 L .42637 .47985 L p
.616 .732 .928 r
F P
s
P
p
.002 w
.69377 .41437 m .68419 .40721 L .70505 .39465 L p
.77 .909 .935 r
F P
s
P
p
.002 w
.70505 .39465 m .71443 .40135 L .69377 .41437 L p
.77 .909 .935 r
F P
s
P
p
.002 w
.58121 .47385 m .57055 .46202 L .59116 .45023 L p
.758 .811 .895 r
F P
s
P
p
.002 w
.59116 .45023 m .60168 .46132 L .58121 .47385 L p
.758 .811 .895 r
F P
s
P
p
.002 w
.24934 .42246 m .23763 .3956 L .25654 .39359 L p
.565 .549 .796 r
F P
s
P
p
.002 w
.25654 .39359 m .26833 .41982 L .24934 .42246 L p
.565 .549 .796 r
F P
s
P
p
.002 w
.34325 .47077 m .33107 .4554 L .35079 .44969 L p
.634 .704 .896 r
F P
s
P
p
.002 w
.35079 .44969 m .36292 .46493 L .34325 .47077 L p
.634 .704 .896 r
F P
s
P
p
.002 w
.78621 .37642 m .77767 .37152 L .79931 .36487 L p
.562 .88 .992 r
F P
s
P
p
.002 w
.79931 .36487 m .80769 .3701 L .78621 .37642 L p
.562 .88 .992 r
F P
s
P
p
.002 w
.47789 .48266 m .46649 .46923 L .48681 .46291 L p
.645 .736 .915 r
F P
s
P
p
.002 w
.48681 .46291 m .49812 .47634 L .47789 .48266 L p
.645 .736 .915 r
F P
s
P
p
.002 w
.71443 .40135 m .70505 .39465 L .72609 .38331 L p
.739 .911 .952 r
F P
s
P
p
.002 w
.72609 .38331 m .73527 .38974 L .71443 .40135 L p
.739 .911 .952 r
F P
s
P
p
.002 w
.52958 .48206 m .51852 .46882 L .53906 .45989 L p
.702 .766 .901 r
F P
s
P
p
.002 w
.53906 .45989 m .55001 .47272 L .52958 .48206 L p
.702 .766 .901 r
F P
s
P
p
.002 w
.39474 .47222 m .38274 .45929 L .40272 .45388 L p
.618 .732 .928 r
F P
s
P
p
.002 w
.40272 .45388 m .41462 .46705 L .39474 .47222 L p
.618 .732 .928 r
F P
s
P
p
.002 w
.60168 .46132 m .59116 .45023 L .61185 .43767 L p
.771 .829 .897 r
F P
s
P
p
.002 w
.61185 .43767 m .62221 .44801 L .60168 .46132 L p
.771 .829 .897 r
F P
s
P
p
.002 w
.28013 .44499 m .26833 .41982 L .28761 .41607 L p
.6 .584 .806 r
F P
s
P
p
.002 w
.28761 .41607 m .29948 .44039 L .28013 .44499 L p
.6 .584 .806 r
F P
s
P
p
.002 w
.31151 .46106 m .29948 .44039 L .31903 .43535 L p
.624 .637 .843 r
F P
s
P
p
.002 w
.31903 .43535 m .33107 .4554 L .31151 .46106 L p
.624 .637 .843 r
F P
s
P
p
.002 w
.23763 .3956 m .22592 .368 L .24476 .36635 L p
.552 .535 .79 r
F P
s
P
p
.002 w
.24476 .36635 m .25654 .39359 L .23763 .3956 L p
.552 .535 .79 r
F P
s
P
p
.002 w
.44634 .47481 m .43468 .46175 L .45492 .45601 L p
.628 .735 .924 r
F P
s
P
p
.002 w
.45492 .45601 m .46649 .46923 L .44634 .47481 L p
.628 .735 .924 r
F P
s
P
p
.002 w
.62221 .44801 m .61185 .43767 L .63261 .42468 L p
.779 .847 .902 r
F P
s
P
p
.002 w
.63261 .42468 m .6428 .43429 L .62221 .44801 L p
.779 .847 .902 r
F P
s
P
p
.002 w
.73527 .38974 m .72609 .38331 L .74734 .37343 L p
.697 .904 .97 r
F P
s
P
p
.002 w
.74734 .37343 m .75634 .37979 L .73527 .38974 L p
.697 .904 .97 r
F P
s
P
p
.002 w
.36292 .46493 m .35079 .44969 L .37068 .44406 L p
.629 .704 .899 r
F P
s
P
p
.002 w
.37068 .44406 m .38274 .45929 L .36292 .46493 L p
.629 .704 .899 r
F P
s
P
p
.002 w
.55001 .47272 m .53906 .45989 L .55972 .44973 L p
.725 .783 .898 r
F P
s
P
p
.002 w
.55972 .44973 m .57055 .46202 L .55001 .47272 L p
.725 .783 .898 r
F P
s
P
p
.002 w
.22592 .368 m .21434 .33851 L .2331 .33711 L p
.543 .516 .776 r
F P
s
P
p
.002 w
.2331 .33711 m .24476 .36635 L .22592 .368 L p
.543 .516 .776 r
F P
s
P
p
.002 w
.49812 .47634 m .48681 .46291 L .50731 .45553 L p
.668 .748 .91 r
F P
s
P
p
.002 w
.50731 .45553 m .51852 .46882 L .49812 .47634 L p
.668 .748 .91 r
F P
s
P
p
.002 w
.26833 .41982 m .25654 .39359 L .27575 .3904 L p
.583 .565 .799 r
F P
s
P
p
.002 w
.27575 .3904 m .28761 .41607 L .26833 .41982 L p
.583 .565 .799 r
F P
s
P
p
.002 w
.6428 .43429 m .63261 .42468 L .65345 .41161 L p
.781 .862 .91 r
F P
s
P
p
.002 w
.65345 .41161 m .66345 .42056 L .6428 .43429 L p
.781 .862 .91 r
F P
s
P
p
.002 w
.41462 .46705 m .40272 .45388 L .42287 .44836 L p
.621 .731 .925 r
F P
s
P
p
.002 w
.42287 .44836 m .43468 .46175 L .41462 .46705 L p
.621 .731 .925 r
F P
s
P
p
.002 w
.33107 .4554 m .31903 .43535 L .33876 .43005 L p
.625 .644 .849 r
F P
s
P
p
.002 w
.33876 .43005 m .35079 .44969 L .33107 .4554 L p
.625 .644 .849 r
F P
s
P
p
.002 w
.75634 .37979 m .74734 .37343 L .76885 .36506 L p
.649 .889 .983 r
F P
s
P
p
.002 w
.76885 .36506 m .77767 .37152 L .75634 .37979 L p
.649 .889 .983 r
F P
s
P
p
.002 w
.29948 .44039 m .28761 .41607 L .30711 .41166 L p
.607 .594 .812 r
F P
s
P
p
.002 w
.30711 .41166 m .31903 .43535 L .29948 .44039 L p
.607 .594 .812 r
F P
s
P
p
.002 w
.66345 .42056 m .65345 .41161 L .67439 .39882 L p
.776 .875 .92 r
F P
s
P
p
.002 w
.67439 .39882 m .68419 .40721 L .66345 .42056 L p
.776 .875 .92 r
F P
s
P
p
.002 w
.57055 .46202 m .55972 .44973 L .58048 .43854 L p
.744 .8 .898 r
F P
s
P
p
.002 w
.58048 .43854 m .59116 .45023 L .57055 .46202 L p
.744 .8 .898 r
F P
s
P
p
.002 w
.46649 .46923 m .45492 .45601 L .47534 .44962 L p
.643 .741 .92 r
F P
s
P
p
.002 w
.47534 .44962 m .48681 .46291 L .46649 .46923 L p
.643 .741 .92 r
F P
s
P
p
.002 w
.25654 .39359 m .24476 .36635 L .26392 .36351 L p
.571 .55 .792 r
F P
s
P
p
.002 w
.26392 .36351 m .27575 .3904 L .25654 .39359 L p
.571 .55 .792 r
F P
s
P
p
.002 w
.51852 .46882 m .50731 .45553 L .52796 .4469 L p
.694 .763 .905 r
F P
s
P
p
.002 w
.52796 .4469 m .53906 .45989 L .51852 .46882 L p
.694 .763 .905 r
F P
s
P
p
.002 w
.68419 .40721 m .67439 .39882 L .69544 .38669 L p
.761 .884 .933 r
F P
s
P
p
.002 w
.69544 .38669 m .70505 .39465 L .68419 .40721 L p
.761 .884 .933 r
F P
s
P
p
.002 w
.38274 .45929 m .37068 .44406 L .39073 .43858 L p
.624 .702 .901 r
F P
s
P
p
.002 w
.39073 .43858 m .40272 .45388 L .38274 .45929 L p
.624 .702 .901 r
F P
s
P
p
.002 w
.59116 .45023 m .58048 .43854 L .60132 .42662 L p
.757 .816 .899 r
F P
s
P
p
.002 w
.60132 .42662 m .61185 .43767 L .59116 .45023 L p
.757 .816 .899 r
F P
s
P
p
.002 w
.28761 .41607 m .27575 .3904 L .29522 .38644 L p
.594 .575 .802 r
F P
s
P
p
.002 w
.29522 .38644 m .30711 .41166 L .28761 .41607 L p
.594 .575 .802 r
F P
s
P
p
.002 w
.77767 .37152 m .76885 .36506 L .79064 .35814 L p
.601 .867 .993 r
F P
s
P
p
.002 w
.79064 .35814 m .79931 .36487 L .77767 .37152 L p
.601 .867 .993 r
F P
s
P
p
.002 w
.43468 .46175 m .42287 .44836 L .4432 .44247 L p
.63 .733 .922 r
F P
s
P
p
.002 w
.4432 .44247 m .45492 .45601 L .43468 .46175 L p
.63 .733 .922 r
F P
s
P
p
.002 w
.24476 .36635 m .2331 .33711 L .2522 .33454 L p
.562 .53 .777 r
F P
s
P
p
.002 w
.2522 .33454 m .26392 .36351 L .24476 .36635 L p
.562 .53 .777 r
F P
s
P
p
.002 w
.70505 .39465 m .69544 .38669 L .71667 .37562 L p
.736 .887 .949 r
F P
s
P
p
.002 w
.71667 .37562 m .72609 .38331 L .70505 .39465 L p
.736 .887 .949 r
F P
s
P
p
.002 w
.35079 .44969 m .33876 .43005 L .35868 .42468 L p
.624 .647 .853 r
F P
s
P
p
.002 w
.35868 .42468 m .37068 .44406 L .35079 .44969 L p
.624 .647 .853 r
F P
s
P
p
.002 w
.31903 .43535 m .30711 .41166 L .32683 .40684 L p
.611 .602 .817 r
F P
s
P
p
.002 w
.32683 .40684 m .33876 .43005 L .31903 .43535 L p
.611 .602 .817 r
F P
s
P
p
.002 w
.61185 .43767 m .60132 .42662 L .62224 .41425 L p
.765 .832 .904 r
F P
s
P
p
.002 w
.62224 .41425 m .63261 .42468 L .61185 .43767 L p
.765 .832 .904 r
F P
s
P
p
.002 w
.53906 .45989 m .52796 .4469 L .54874 .43714 L p
.715 .778 .902 r
F P
s
P
p
.002 w
.54874 .43714 m .55972 .44973 L .53906 .45989 L p
.715 .778 .902 r
F P
s
P
p
.002 w
.48681 .46291 m .47534 .44962 L .49594 .44231 L p
.664 .751 .915 r
F P
s
P
p
.002 w
.49594 .44231 m .50731 .45553 L .48681 .46291 L p
.664 .751 .915 r
F P
s
P
p
.002 w
.40272 .45388 m .39073 .43858 L .41095 .43298 L p
.625 .703 .901 r
F P
s
P
p
.002 w
.41095 .43298 m .42287 .44836 L .40272 .45388 L p
.625 .703 .901 r
F P
s
P
p
.002 w
.27575 .3904 m .26392 .36351 L .28334 .35986 L p
.583 .56 .794 r
F P
s
P
p
.002 w
.28334 .35986 m .29522 .38644 L .27575 .3904 L p
.583 .56 .794 r
F P
s
P
p
.002 w
.72609 .38331 m .71667 .37562 L .7381 .36584 L p
.701 .882 .965 r
F P
s
P
p
.002 w
.7381 .36584 m .74734 .37343 L .72609 .38331 L p
.701 .882 .965 r
F P
s
P
p
.002 w
.63261 .42468 m .62224 .41425 L .64326 .40176 L p
.767 .846 .91 r
F P
s
P
p
.002 w
.64326 .40176 m .65345 .41161 L .63261 .42468 L p
.767 .846 .91 r
F P
s
P
p
.002 w
.55972 .44973 m .54874 .43714 L .56962 .42645 L p
.732 .793 .901 r
F P
s
P
p
.002 w
.56962 .42645 m .58048 .43854 L .55972 .44973 L p
.732 .793 .901 r
F P
s
P
p
.002 w
.45492 .45601 m .4432 .44247 L .46372 .43601 L p
.643 .738 .918 r
F P
s
P
p
.002 w
.46372 .43601 m .47534 .44962 L .45492 .45601 L p
.643 .738 .918 r
F P
s
P
p
.002 w
.30711 .41166 m .29522 .38644 L .31491 .38197 L p
.6 .583 .805 r
F P
s
P
p
.002 w
.31491 .38197 m .32683 .40684 L .30711 .41166 L p
.6 .583 .805 r
F P
s
P
p
.002 w
.37068 .44406 m .35868 .42468 L .37876 .41934 L p
.621 .649 .857 r
F P
s
P
p
.002 w
.37876 .41934 m .39073 .43858 L .37068 .44406 L p
.621 .649 .857 r
F P
s
P
p
.002 w
.65345 .41161 m .64326 .40176 L .66437 .38947 L p
.763 .857 .92 r
F P
s
P
p
.002 w
.66437 .38947 m .67439 .39882 L .65345 .41161 L p
.763 .857 .92 r
F P
s
P
p
.002 w
.50731 .45553 m .49594 .44231 L .5167 .43388 L p
.687 .764 .91 r
F P
s
P
p
.002 w
.5167 .43388 m .52796 .4469 L .50731 .45553 L p
.687 .764 .91 r
F P
s
P
p
.002 w
.33876 .43005 m .32683 .40684 L .34674 .4018 L p
.613 .607 .821 r
F P
s
P
p
.002 w
.34674 .4018 m .35868 .42468 L .33876 .43005 L p
.613 .607 .821 r
F P
s
P
p
.002 w
.26392 .36351 m .2522 .33454 L .27157 .33113 L p
.574 .539 .778 r
F P
s
P
p
.002 w
.27157 .33113 m .28334 .35986 L .26392 .36351 L p
.574 .539 .778 r
F P
s
P
p
.002 w
.74734 .37343 m .7381 .36584 L .75977 .3574 L p
.662 .871 .978 r
F P
s
P
p
.002 w
.75977 .3574 m .76885 .36506 L .74734 .37343 L p
.662 .871 .978 r
F P
s
P
p
.002 w
.58048 .43854 m .56962 .42645 L .59061 .41505 L p
.744 .808 .902 r
F P
s
P
p
.002 w
.59061 .41505 m .60132 .42662 L .58048 .43854 L p
.744 .808 .902 r
F P
s
P
p
.002 w
.42287 .44836 m .41095 .43298 L .43137 .42703 L p
.632 .706 .9 r
F P
s
P
p
.002 w
.43137 .42703 m .4432 .44247 L .42287 .44836 L p
.632 .706 .9 r
F P
s
P
p
.002 w
.67439 .39882 m .66437 .38947 L .68561 .37773 L p
.752 .865 .931 r
F P
s
P
p
.002 w
.68561 .37773 m .69544 .38669 L .67439 .39882 L p
.752 .865 .931 r
F P
s
P
p
.002 w
.29522 .38644 m .28334 .35986 L .303 .35564 L p
.591 .567 .796 r
F P
s
P
p
.002 w
.303 .35564 m .31491 .38197 L .29522 .38644 L p
.591 .567 .796 r
F P
s
P
p
.002 w
.47534 .44962 m .46372 .43601 L .48441 .42873 L p
.66 .747 .914 r
F P
s
P
p
.002 w
.48441 .42873 m .49594 .44231 L .47534 .44962 L p
.66 .747 .914 r
F P
s
P
p
.002 w
.52796 .4469 m .5167 .43388 L .53759 .42444 L p
.706 .777 .907 r
F P
s
P
p
.002 w
.53759 .42444 m .54874 .43714 L .52796 .4469 L p
.706 .777 .907 r
F P
s
P
p
.002 w
.60132 .42662 m .59061 .41505 L .61169 .40321 L p
.752 .821 .906 r
F P
s
P
p
.002 w
.61169 .40321 m .62224 .41425 L .60132 .42662 L p
.752 .821 .906 r
F P
s
P
p
.002 w
.39073 .43858 m .37876 .41934 L .39903 .41382 L p
.622 .652 .859 r
F P
s
P
p
.002 w
.39903 .41382 m .41095 .43298 L .39073 .43858 L p
.622 .652 .859 r
F P
s
P
p
.002 w
.32683 .40684 m .31491 .38197 L .33481 .37719 L p
.603 .588 .808 r
F P
s
P
p
.002 w
.33481 .37719 m .34674 .4018 L .32683 .40684 L p
.603 .588 .808 r
F P
s
P
p
.002 w
.69544 .38669 m .68561 .37773 L .70702 .36691 L p
.73 .868 .945 r
F P
s
P
p
.002 w
.70702 .36691 m .71667 .37562 L .69544 .38669 L p
.73 .868 .945 r
F P
s
P
p
.002 w
.76885 .36506 m .75977 .3574 L .78171 .35027 L p
.621 .853 .987 r
F P
s
P
p
.002 w
.78171 .35027 m .79064 .35814 L .76885 .36506 L p
.621 .853 .987 r
F P
s
P
p
.002 w
.35868 .42468 m .34674 .4018 L .36684 .39667 L p
.612 .61 .824 r
F P
s
P
p
.002 w
.36684 .39667 m .37876 .41934 L .35868 .42468 L p
.612 .61 .824 r
F P
s
P
p
.002 w
.28334 .35986 m .27157 .33113 L .29119 .32712 L p
.582 .547 .78 r
F P
s
P
p
.002 w
.29119 .32712 m .303 .35564 L .28334 .35986 L p
.582 .547 .78 r
F P
s
P
p
.002 w
.62224 .41425 m .61169 .40321 L .63287 .39122 L p
.755 .833 .912 r
F P
s
P
p
.002 w
.63287 .39122 m .64326 .40176 L .62224 .41425 L p
.755 .833 .912 r
F P
s
P
p
.002 w
.4432 .44247 m .43137 .42703 L .45196 .42056 L p
.642 .711 .898 r
F P
s
P
p
.002 w
.45196 .42056 m .46372 .43601 L .4432 .44247 L p
.642 .711 .898 r
F P
s
P
p
.002 w
.54874 .43714 m .53759 .42444 L .55861 .41414 L p
.722 .79 .906 r
F P
s
P
p
.002 w
.55861 .41414 m .56962 .42645 L .54874 .43714 L p
.722 .79 .906 r
F P
s
P
p
.002 w
.71667 .37562 m .70702 .36691 L .72862 .35723 L p
.701 .864 .959 r
F P
s
P
p
.002 w
.72862 .35723 m .7381 .36584 L .71667 .37562 L p
.701 .864 .959 r
F P
s
P
p
.002 w
.49594 .44231 m .48441 .42873 L .50527 .42047 L p
.681 .758 .91 r
F P
s
P
p
.002 w
.50527 .42047 m .5167 .43388 L .49594 .44231 L p
.681 .758 .91 r
F P
s
P
p
.002 w
.31491 .38197 m .303 .35564 L .32288 .35105 L p
.595 .572 .798 r
F P
s
P
p
.002 w
.32288 .35105 m .33481 .37719 L .31491 .38197 L p
.595 .572 .798 r
F P
s
P
p
.002 w
.64326 .40176 m .63287 .39122 L .65416 .37937 L p
.752 .843 .92 r
F P
s
P
p
.002 w
.65416 .37937 m .66437 .38947 L .64326 .40176 L p
.752 .843 .92 r
F P
s
P
p
.002 w
.41095 .43298 m .39903 .41382 L .4195 .40793 L p
.628 .657 .859 r
F P
s
P
p
.002 w
.4195 .40793 m .43137 .42703 L .41095 .43298 L p
.628 .657 .859 r
F P
s
P
p
.002 w
.34674 .4018 m .33481 .37719 L .3549 .37223 L p
.604 .591 .811 r
F P
s
P
p
.002 w
.3549 .37223 m .36684 .39667 L .34674 .4018 L p
.604 .591 .811 r
F P
s
P
p
.002 w
.56962 .42645 m .55861 .41414 L .57973 .40317 L p
.734 .803 .906 r
F P
s
P
p
.002 w
.57973 .40317 m .59061 .41505 L .56962 .42645 L p
.734 .803 .906 r
F P
s
P
p
.002 w
.37876 .41934 m .36684 .39667 L .38713 .3913 L p
.614 .614 .827 r
F P
s
P
p
.002 w
.38713 .3913 m .39903 .41382 L .37876 .41934 L p
.614 .614 .827 r
F P
s
P
p
.002 w
.7381 .36584 m .72862 .35723 L .75046 .34874 L p
.667 .854 .971 r
F P
s
P
p
.002 w
.75046 .34874 m .75977 .3574 L .7381 .36584 L p
.667 .854 .971 r
F P
s
P
p
.002 w
.46372 .43601 m .45196 .42056 L .47274 .41335 L p
.656 .719 .895 r
F P
s
P
p
.002 w
.47274 .41335 m .48441 .42873 L .46372 .43601 L p
.656 .719 .895 r
F P
s
P
p
.002 w
.303 .35564 m .29119 .32712 L .31103 .32268 L p
.588 .552 .782 r
F P
s
P
p
.002 w
.31103 .32268 m .32288 .35105 L .303 .35564 L p
.588 .552 .782 r
F P
s
P
p
.002 w
.66437 .38947 m .65416 .37937 L .67558 .36798 L p
.742 .85 .93 r
F P
s
P
p
.002 w
.67558 .36798 m .68561 .37773 L .66437 .38947 L p
.742 .85 .93 r
F P
s
P
p
.002 w
.5167 .43388 m .50527 .42047 L .52628 .41128 L p
.698 .77 .907 r
F P
s
P
p
.002 w
.52628 .41128 m .53759 .42444 L .5167 .43388 L p
.698 .77 .907 r
F P
s
P
p
.002 w
.59061 .41505 m .57973 .40317 L .60097 .39178 L p
.742 .815 .909 r
F P
s
P
p
.002 w
.60097 .39178 m .61169 .40321 L .59061 .41505 L p
.742 .815 .909 r
F P
s
P
p
.002 w
.33481 .37719 m .32288 .35105 L .34297 .34621 L p
.598 .575 .799 r
F P
s
P
p
.002 w
.34297 .34621 m .3549 .37223 L .33481 .37719 L p
.598 .575 .799 r
F P
s
P
p
.002 w
.43137 .42703 m .4195 .40793 L .44015 .40156 L p
.636 .663 .859 r
F P
s
P
p
.002 w
.44015 .40156 m .45196 .42056 L .43137 .42703 L p
.636 .663 .859 r
F P
s
P
p
.002 w
.68561 .37773 m .67558 .36798 L .69716 .35739 L p
.724 .853 .942 r
F P
s
P
p
.002 w
.69716 .35739 m .70702 .36691 L .68561 .37773 L p
.724 .853 .942 r
F P
s
P
p
.002 w
.36684 .39667 m .3549 .37223 L .3752 .36698 L p
.608 .595 .812 r
F P
s
P
p
.002 w
.3752 .36698 m .38713 .3913 L .36684 .39667 L p
.608 .595 .812 r
F P
s
P
p
.002 w
.75977 .3574 m .75046 .34874 L .77256 .34143 L p
.633 .84 .98 r
F P
s
P
p
.002 w
.77256 .34143 m .78171 .35027 L .75977 .3574 L p
.633 .84 .98 r
F P
s
P
p
.002 w
.39903 .41382 m .38713 .3913 L .40762 .38555 L p
.62 .619 .828 r
F P
s
P
p
.002 w
.40762 .38555 m .4195 .40793 L .39903 .41382 L p
.62 .619 .828 r
F P
s
P
p
.002 w
.53759 .42444 m .52628 .41128 L .54742 .40129 L p
.713 .781 .906 r
F P
s
P
p
.002 w
.54742 .40129 m .55861 .41414 L .53759 .42444 L p
.713 .781 .906 r
F P
s
P
p
.002 w
.61169 .40321 m .60097 .39178 L .62231 .38021 L p
.745 .825 .914 r
F P
s
P
p
.002 w
.62231 .38021 m .63287 .39122 L .61169 .40321 L p
.745 .825 .914 r
F P
s
P
p
.002 w
.48441 .42873 m .47274 .41335 L .4937 .40524 L p
.673 .73 .892 r
F P
s
P
p
.002 w
.4937 .40524 m .50527 .42047 L .48441 .42873 L p
.673 .73 .892 r
F P
s
P
p
.002 w
.32288 .35105 m .31103 .32268 L .33109 .31794 L p
.591 .555 .783 r
F P
s
P
p
.002 w
.33109 .31794 m .34297 .34621 L .32288 .35105 L p
.591 .555 .783 r
F P
s
P
p
.002 w
.70702 .36691 m .69716 .35739 L .71893 .3478 L p
.699 .85 .955 r
F P
s
P
p
.002 w
.71893 .3478 m .72862 .35723 L .70702 .36691 L p
.699 .85 .955 r
F P
s
P
p
.002 w
.63287 .39122 m .62231 .38021 L .64376 .36874 L p
.742 .834 .921 r
F P
s
P
p
.002 w
.64376 .36874 m .65416 .37937 L .63287 .39122 L p
.742 .834 .921 r
F P
s
P
p
.002 w
.55861 .41414 m .54742 .40129 L .56868 .39069 L p
.724 .793 .906 r
F P
s
P
p
.002 w
.56868 .39069 m .57973 .40317 L .55861 .41414 L p
.724 .793 .906 r
F P
s
P
p
.002 w
.45196 .42056 m .44015 .40156 L .46099 .39449 L p
.647 .671 .859 r
F P
s
P
p
.002 w
.46099 .39449 m .47274 .41335 L .45196 .42056 L p
.647 .671 .859 r
F P
s
P
p
.002 w
.3549 .37223 m .34297 .34621 L .36326 .34104 L p
.601 .579 .801 r
F P
s
P
p
.002 w
.36326 .34104 m .3752 .36698 L .3549 .37223 L p
.601 .579 .801 r
F P
s
P
p
.002 w
.50527 .42047 m .4937 .40524 L .51481 .3963 L p
.689 .741 .891 r
F P
s
P
p
.002 w
.51481 .3963 m .52628 .41128 L .50527 .42047 L p
.689 .741 .891 r
F P
s
P
p
.002 w
.38713 .3913 m .3752 .36698 L .3957 .36133 L p
.613 .6 .814 r
F P
s
P
p
.002 w
.3957 .36133 m .40762 .38555 L .38713 .3913 L p
.613 .6 .814 r
F P
s
P
p
.002 w
.4195 .40793 m .40762 .38555 L .4283 .3793 L p
.627 .625 .829 r
F P
s
P
p
.002 w
.4283 .3793 m .44015 .40156 L .4195 .40793 L p
.627 .625 .829 r
F P
s
P
p
.002 w
.65416 .37937 m .64376 .36874 L .66535 .35766 L p
.734 .84 .93 r
F P
s
P
p
.002 w
.66535 .35766 m .67558 .36798 L .65416 .37937 L p
.734 .84 .93 r
F P
s
P
p
.002 w
.72862 .35723 m .71893 .3478 L .74093 .33928 L p
.67 .841 .965 r
F P
s
P
p
.002 w
.74093 .33928 m .75046 .34874 L .72862 .35723 L p
.67 .841 .965 r
F P
s
P
p
.002 w
.57973 .40317 m .56868 .39069 L .59006 .37968 L p
.731 .803 .909 r
F P
s
P
p
.002 w
.59006 .37968 m .60097 .39178 L .57973 .40317 L p
.731 .803 .909 r
F P
s
P
p
.002 w
.34297 .34621 m .33109 .31794 L .35137 .31285 L p
.595 .559 .784 r
F P
s
P
p
.002 w
.35137 .31285 m .36326 .34104 L .34297 .34621 L p
.595 .559 .784 r
F P
s
P
p
.002 w
.47274 .41335 m .46099 .39449 L .48201 .38659 L p
.661 .681 .858 r
F P
s
P
p
.002 w
.48201 .38659 m .4937 .40524 L .47274 .41335 L p
.661 .681 .858 r
F P
s
P
p
.002 w
.52628 .41128 m .51481 .3963 L .53606 .38661 L p
.702 .751 .89 r
F P
s
P
p
.002 w
.53606 .38661 m .54742 .40129 L .52628 .41128 L p
.702 .751 .89 r
F P
s
P
p
.002 w
.67558 .36798 m .66535 .35766 L .6871 .34727 L p
.718 .842 .941 r
F P
s
P
p
.002 w
.6871 .34727 m .69716 .35739 L .67558 .36798 L p
.718 .842 .941 r
F P
s
P
p
.002 w
.3752 .36698 m .36326 .34104 L .38377 .33547 L p
.607 .584 .801 r
F P
s
P
p
.002 w
.38377 .33547 m .3957 .36133 L .3752 .36698 L p
.607 .584 .801 r
F P
s
P
p
.002 w
.60097 .39178 m .59006 .37968 L .61155 .36848 L p
.734 .813 .913 r
F P
s
P
p
.002 w
.61155 .36848 m .62231 .38021 L .60097 .39178 L p
.734 .813 .913 r
F P
s
P
p
.002 w
.44015 .40156 m .4283 .3793 L .44918 .3724 L p
.637 .632 .829 r
F P
s
P
p
.002 w
.44918 .3724 m .46099 .39449 L .44015 .40156 L p
.637 .632 .829 r
F P
s
P
p
.002 w
.40762 .38555 m .3957 .36133 L .41641 .35519 L p
.62 .605 .814 r
F P
s
P
p
.002 w
.41641 .35519 m .4283 .3793 L .40762 .38555 L p
.62 .605 .814 r
F P
s
P
p
.002 w
.75046 .34874 m .74093 .33928 L .76318 .33181 L p
.64 .829 .973 r
F P
s
P
p
.002 w
.76318 .33181 m .77256 .34143 L .75046 .34874 L p
.64 .829 .973 r
F P
s
P
p
.002 w
.54742 .40129 m .53606 .38661 L .55744 .37635 L p
.712 .762 .891 r
F P
s
P
p
.002 w
.55744 .37635 m .56868 .39069 L .54742 .40129 L p
.712 .762 .891 r
F P
s
P
p
.002 w
.69716 .35739 m .6871 .34727 L .70905 .33776 L p
.695 .84 .952 r
F P
s
P
p
.002 w
.70905 .33776 m .71893 .3478 L .69716 .35739 L p
.695 .84 .952 r
F P
s
P
p
.002 w
.4937 .40524 m .48201 .38659 L .5032 .3779 L p
.674 .691 .857 r
F P
s
P
p
.002 w
.5032 .3779 m .51481 .3963 L .4937 .40524 L p
.674 .691 .857 r
F P
s
P
p
.002 w
.62231 .38021 m .61155 .36848 L .63317 .35734 L p
.733 .821 .919 r
F P
s
P
p
.002 w
.63317 .35734 m .64376 .36874 L .62231 .38021 L p
.733 .821 .919 r
F P
s
P
p
.002 w
.36326 .34104 m .35137 .31285 L .37186 .30733 L p
.6 .563 .785 r
F P
s
P
p
.002 w
.37186 .30733 m .38377 .33547 L .36326 .34104 L p
.6 .563 .785 r
F P
s
P
p
.002 w
.3957 .36133 m .38377 .33547 L .40448 .32941 L p
.614 .589 .802 r
F P
s
P
p
.002 w
.40448 .32941 m .41641 .35519 L .3957 .36133 L p
.614 .589 .802 r
F P
s
P
p
.002 w
.46099 .39449 m .44918 .3724 L .47025 .36471 L p
.649 .641 .829 r
F P
s
P
p
.002 w
.47025 .36471 m .48201 .38659 L .46099 .39449 L p
.649 .641 .829 r
F P
s
P
p
.002 w
.4283 .3793 m .41641 .35519 L .43731 .34843 L p
.629 .612 .814 r
F P
s
P
p
.002 w
.43731 .34843 m .44918 .3724 L .4283 .3793 L p
.629 .612 .814 r
F P
s
P
p
.002 w
.56868 .39069 m .55744 .37635 L .57895 .36569 L p
.719 .771 .894 r
F P
s
P
p
.002 w
.57895 .36569 m .59006 .37968 L .56868 .39069 L p
.719 .771 .894 r
F P
s
P
p
.002 w
.64376 .36874 m .63317 .35734 L .65492 .34653 L p
.725 .826 .927 r
F P
s
P
p
.002 w
.65492 .34653 m .66535 .35766 L .64376 .36874 L p
.725 .826 .927 r
F P
s
P
p
.002 w
.71893 .3478 m .70905 .33776 L .73121 .32921 L p
.67 .833 .961 r
F P
s
P
p
.002 w
.73121 .32921 m .74093 .33928 L .71893 .3478 L p
.67 .833 .961 r
F P
s
P
p
.002 w
.51481 .3963 m .5032 .3779 L .52454 .36853 L p
.685 .701 .858 r
F P
s
P
p
.002 w
.52454 .36853 m .53606 .38661 L .51481 .3963 L p
.685 .701 .858 r
F P
s
P
p
.002 w
.38377 .33547 m .37186 .30733 L .39256 .30133 L p
.607 .568 .785 r
F P
s
P
p
.002 w
.39256 .30133 m .40448 .32941 L .38377 .33547 L p
.607 .568 .785 r
F P
s
P
p
.002 w
.59006 .37968 m .57895 .36569 L .60058 .35484 L p
.722 .78 .898 r
F P
s
P
p
.002 w
.60058 .35484 m .61155 .36848 L .59006 .37968 L p
.722 .78 .898 r
F P
s
P
p
.002 w
.66535 .35766 m .65492 .34653 L .67684 .33631 L p
.712 .828 .937 r
F P
s
P
p
.002 w
.67684 .33631 m .6871 .34727 L .66535 .35766 L p
.712 .828 .937 r
F P
s
P
p
.002 w
.48201 .38659 m .47025 .36471 L .49149 .35629 L p
.66 .65 .829 r
F P
s
P
p
.002 w
.49149 .35629 m .5032 .3779 L .48201 .38659 L p
.66 .65 .829 r
F P
s
P
p
.002 w
.41641 .35519 m .40448 .32941 L .4254 .32274 L p
.622 .595 .802 r
F P
s
P
p
.002 w
.4254 .32274 m .43731 .34843 L .41641 .35519 L p
.622 .595 .802 r
F P
s
P
p
.002 w
.44918 .3724 m .43731 .34843 L .45841 .34092 L p
.64 .62 .814 r
F P
s
P
p
.002 w
.45841 .34092 m .47025 .36471 L .44918 .3724 L p
.64 .62 .814 r
F P
s
P
p
.002 w
.53606 .38661 m .52454 .36853 L .54602 .35861 L p
.694 .71 .86 r
F P
s
P
p
.002 w
.54602 .35861 m .55744 .37635 L .53606 .38661 L p
.694 .71 .86 r
F P
s
P
p
.002 w
.74093 .33928 m .73121 .32921 L .75361 .32159 L p
.644 .822 .969 r
F P
s
P
p
.002 w
.75361 .32159 m .76318 .33181 L .74093 .33928 L p
.644 .822 .969 r
F P
s
P
p
.002 w
.61155 .36848 m .60058 .35484 L .62235 .34401 L p
.721 .787 .904 r
F P
s
P
p
.002 w
.62235 .34401 m .63317 .35734 L .61155 .36848 L p
.721 .787 .904 r
F P
s
P
p
.002 w
.6871 .34727 m .67684 .33631 L .69895 .32688 L p
.692 .826 .947 r
F P
s
P
p
.002 w
.69895 .32688 m .70905 .33776 L .6871 .34727 L p
.692 .826 .947 r
F P
s
P
p
.002 w
.5032 .3779 m .49149 .35629 L .51289 .34722 L p
.67 .659 .831 r
F P
s
P
p
.002 w
.51289 .34722 m .52454 .36853 L .5032 .3779 L p
.67 .659 .831 r
F P
s
P
p
.002 w
.40448 .32941 m .39256 .30133 L .41348 .29475 L p
.614 .574 .785 r
F P
s
P
p
.002 w
.41348 .29475 m .4254 .32274 L .40448 .32941 L p
.614 .574 .785 r
F P
s
P
p
.002 w
.55744 .37635 m .54602 .35861 L .56763 .3483 L p
.701 .718 .863 r
F P
s
P
p
.002 w
.56763 .3483 m .57895 .36569 L .55744 .37635 L p
.701 .718 .863 r
F P
s
P
p
.002 w
.47025 .36471 m .45841 .34092 L .47969 .33273 L p
.65 .628 .815 r
F P
s
P
p
.002 w
.47969 .33273 m .49149 .35629 L .47025 .36471 L p
.65 .628 .815 r
F P
s
P
p
.002 w
.43731 .34843 m .4254 .32274 L .44651 .31538 L p
.632 .602 .802 r
F P
s
P
p
.002 w
.44651 .31538 m .45841 .34092 L .43731 .34843 L p
.632 .602 .802 r
F P
s
P
p
.002 w
.63317 .35734 m .62235 .34401 L .64425 .33345 L p
.716 .792 .911 r
F P
s
P
p
.002 w
.64425 .33345 m .65492 .34653 L .63317 .35734 L p
.716 .792 .911 r
F P
s
P
p
.002 w
.70905 .33776 m .69895 .32688 L .72126 .3183 L p
.671 .82 .955 r
F P
s
P
p
.002 w
.72126 .3183 m .73121 .32921 L .70905 .33776 L p
.671 .82 .955 r
F P
s
P
p
.002 w
.52454 .36853 m .51289 .34722 L .53444 .33763 L p
.678 .667 .832 r
F P
s
P
p
.002 w
.53444 .33763 m .54602 .35861 L .52454 .36853 L p
.678 .667 .832 r
F P
s
P
p
.002 w
.57895 .36569 m .56763 .3483 L .58938 .33779 L p
.704 .726 .867 r
F P
s
P
p
.002 w
.58938 .33779 m .60058 .35484 L .57895 .36569 L p
.704 .726 .867 r
F P
s
P
p
.002 w
.65492 .34653 m .64425 .33345 L .66632 .32341 L p
.705 .793 .92 r
F P
s
P
p
.002 w
.66632 .32341 m .67684 .33631 L .65492 .34653 L p
.705 .793 .92 r
F P
s
P
p
.002 w
.4254 .32274 m .41348 .29475 L .4346 .28751 L p
.623 .581 .786 r
F P
s
P
p
.002 w
.4346 .28751 m .44651 .31538 L .4254 .32274 L p
.623 .581 .786 r
F P
s
P
p
.002 w
.49149 .35629 m .47969 .33273 L .50114 .32392 L p
.659 .636 .816 r
F P
s
P
p
.002 w
.50114 .32392 m .51289 .34722 L .49149 .35629 L p
.659 .636 .816 r
F P
s
P
p
.002 w
.45841 .34092 m .44651 .31538 L .46782 .30737 L p
.641 .61 .803 r
F P
s
P
p
.002 w
.46782 .30737 m .47969 .33273 L .45841 .34092 L p
.641 .61 .803 r
F P
s
P
p
.002 w
.60058 .35484 m .58938 .33779 L .61126 .32727 L p
.705 .732 .872 r
F P
s
P
p
.002 w
.61126 .32727 m .62235 .34401 L .60058 .35484 L p
.705 .732 .872 r
F P
s
P
p
.002 w
.73121 .32921 m .72126 .3183 L .74382 .31055 L p
.648 .811 .962 r
F P
s
P
p
.002 w
.74382 .31055 m .75361 .32159 L .73121 .32921 L p
.648 .811 .962 r
F P
s
P
p
.002 w
.54602 .35861 m .53444 .33763 L .55613 .32765 L p
.683 .675 .835 r
F P
s
P
p
.002 w
.55613 .32765 m .56763 .3483 L .54602 .35861 L p
.683 .675 .835 r
F P
s
P
p
.002 w
.67684 .33631 m .66632 .32341 L .68857 .31405 L p
.689 .792 .928 r
F P
s
P
p
.002 w
.68857 .31405 m .69895 .32688 L .67684 .33631 L p
.689 .792 .928 r
F P
s
P
p
.002 w
.51289 .34722 m .50114 .32392 L .52274 .3146 L p
.666 .643 .818 r
F P
s
P
p
.002 w
.52274 .3146 m .53444 .33763 L .51289 .34722 L p
.666 .643 .818 r
F P
s
P
p
.002 w
.44651 .31538 m .4346 .28751 L .45591 .27965 L p
.631 .587 .786 r
F P
s
P
p
.002 w
.45591 .27965 m .46782 .30737 L .44651 .31538 L p
.631 .587 .786 r
F P
s
P
p
.002 w
.47969 .33273 m .46782 .30737 L .4893 .29877 L p
.649 .616 .804 r
F P
s
P
p
.002 w
.4893 .29877 m .50114 .32392 L .47969 .33273 L p
.649 .616 .804 r
F P
s
P
p
.002 w
.62235 .34401 m .61126 .32727 L .63329 .31696 L p
.701 .736 .878 r
F P
s
P
p
.002 w
.63329 .31696 m .64425 .33345 L .62235 .34401 L p
.701 .736 .878 r
F P
s
P
p
.002 w
.56763 .3483 m .55613 .32765 L .57797 .31746 L p
.687 .681 .839 r
F P
s
P
p
.002 w
.57797 .31746 m .58938 .33779 L .56763 .3483 L p
.687 .681 .839 r
F P
s
P
p
.002 w
.69895 .32688 m .68857 .31405 L .71104 .30545 L p
.672 .787 .935 r
F P
s
P
p
.002 w
.71104 .30545 m .72126 .3183 L .69895 .32688 L p
.672 .787 .935 r
F P
s
P
p
.002 w
.53444 .33763 m .52274 .3146 L .5445 .30491 L p
.671 .65 .82 r
F P
s
P
p
.002 w
.5445 .30491 m .55613 .32765 L .53444 .33763 L p
.671 .65 .82 r
F P
s
P
p
.002 w
.64425 .33345 m .63329 .31696 L .65548 .3071 L p
.693 .737 .885 r
F P
s
P
p
.002 w
.65548 .3071 m .66632 .32341 L .64425 .33345 L p
.693 .737 .885 r
F P
s
P
p
.002 w
.50114 .32392 m .4893 .29877 L .51095 .28969 L p
.655 .623 .805 r
F P
s
P
p
.002 w
.51095 .28969 m .52274 .3146 L .50114 .32392 L p
.655 .623 .805 r
F P
s
P
p
.002 w
.58938 .33779 m .57797 .31746 L .59995 .30723 L p
.688 .686 .843 r
F P
s
P
p
.002 w
.59995 .30723 m .61126 .32727 L .58938 .33779 L p
.688 .686 .843 r
F P
s
P
p
.002 w
.46782 .30737 m .45591 .27965 L .4774 .27123 L p
.638 .593 .787 r
F P
s
P
p
.002 w
.4774 .27123 m .4893 .29877 L .46782 .30737 L p
.638 .593 .787 r
F P
s
P
p
.002 w
.72126 .3183 m .71104 .30545 L .73373 .2976 L p
.653 .78 .941 r
F P
s
P
p
.002 w
.73373 .2976 m .74382 .31055 L .72126 .3183 L p
.653 .78 .941 r
F P
s
P
p
.002 w
.55613 .32765 m .5445 .30491 L .5664 .295 L p
.674 .655 .823 r
F P
s
P
p
.002 w
.5664 .295 m .57797 .31746 L .55613 .32765 L p
.674 .655 .823 r
F P
s
P
p
.002 w
.66632 .32341 m .65548 .3071 L .67787 .29784 L p
.681 .736 .892 r
F P
s
P
p
.002 w
.67787 .29784 m .68857 .31405 L .66632 .32341 L p
.681 .736 .892 r
F P
s
P
p
.002 w
.61126 .32727 m .59995 .30723 L .62208 .29717 L p
.685 .689 .848 r
F P
s
P
p
.002 w
.62208 .29717 m .63329 .31696 L .61126 .32727 L p
.685 .689 .848 r
F P
s
P
p
.002 w
.52274 .3146 m .51095 .28969 L .53275 .28024 L p
.66 .628 .807 r
F P
s
P
p
.002 w
.53275 .28024 m .5445 .30491 L .52274 .3146 L p
.66 .628 .807 r
F P
s
P
p
.002 w
.4893 .29877 m .4774 .27123 L .49907 .26235 L p
.644 .599 .788 r
F P
s
P
p
.002 w
.49907 .26235 m .51095 .28969 L .4893 .29877 L p
.644 .599 .788 r
F P
s
P
p
.002 w
.57797 .31746 m .5664 .295 L .58846 .28503 L p
.675 .659 .827 r
F P
s
P
p
.002 w
.58846 .28503 m .59995 .30723 L .57797 .31746 L p
.675 .659 .827 r
F P
s
P
p
.002 w
.63329 .31696 m .62208 .29717 L .64437 .28749 L p
.679 .69 .854 r
F P
s
P
p
.002 w
.64437 .28749 m .65548 .3071 L .63329 .31696 L p
.679 .69 .854 r
F P
s
P
p
.002 w
.68857 .31405 m .67787 .29784 L .70046 .28924 L p
.668 .733 .898 r
F P
s
P
p
.002 w
.70046 .28924 m .71104 .30545 L .68857 .31405 L p
.668 .733 .898 r
F P
s
P
p
.002 w
.5445 .30491 m .53275 .28024 L .55471 .27057 L p
.663 .633 .809 r
F P
s
P
p
.002 w
.55471 .27057 m .5664 .295 L .5445 .30491 L p
.663 .633 .809 r
F P
s
P
p
.002 w
.51095 .28969 m .49907 .26235 L .52091 .25311 L p
.649 .604 .79 r
F P
s
P
p
.002 w
.52091 .25311 m .53275 .28024 L .51095 .28969 L p
.649 .604 .79 r
F P
s
P
p
.002 w
.59995 .30723 m .58846 .28503 L .61067 .27519 L p
.673 .662 .831 r
F P
s
P
p
.002 w
.61067 .27519 m .62208 .29717 L .59995 .30723 L p
.673 .662 .831 r
F P
s
P
p
.002 w
.65548 .3071 m .64437 .28749 L .66686 .27833 L p
.67 .689 .859 r
F P
s
P
p
.002 w
.66686 .27833 m .67787 .29784 L .65548 .3071 L p
.67 .689 .859 r
F P
s
P
p
.002 w
.71104 .30545 m .70046 .28924 L .72327 .28131 L p
.653 .727 .903 r
F P
s
P
p
.002 w
.72327 .28131 m .73373 .2976 L .71104 .30545 L p
.653 .727 .903 r
F P
s
P
p
.002 w
.5664 .295 m .55471 .27057 L .57683 .26082 L p
.664 .637 .812 r
F P
s
P
p
.002 w
.57683 .26082 m .58846 .28503 L .5664 .295 L p
.664 .637 .812 r
F P
s
P
p
.002 w
.53275 .28024 m .52091 .25311 L .5429 .24364 L p
.652 .608 .792 r
F P
s
P
p
.002 w
.5429 .24364 m .55471 .27057 L .53275 .28024 L p
.652 .608 .792 r
F P
s
P
p
.002 w
.62208 .29717 m .61067 .27519 L .63306 .26566 L p
.668 .663 .835 r
F P
s
P
p
.002 w
.63306 .26566 m .64437 .28749 L .62208 .29717 L p
.668 .663 .835 r
F P
s
P
p
.002 w
.67787 .29784 m .66686 .27833 L .68955 .26974 L p
.659 .686 .864 r
F P
s
P
p
.002 w
.68955 .26974 m .70046 .28924 L .67787 .29784 L p
.659 .686 .864 r
F P
s
P
p
.002 w
.58846 .28503 m .57683 .26082 L .59911 .25116 L p
.663 .639 .816 r
F P
s
P
p
.002 w
.59911 .25116 m .61067 .27519 L .58846 .28503 L p
.663 .639 .816 r
F P
s
P
p
.002 w
.55471 .27057 m .5429 .24364 L .56506 .23408 L p
.653 .611 .795 r
F P
s
P
p
.002 w
.56506 .23408 m .57683 .26082 L .55471 .27057 L p
.653 .611 .795 r
F P
s
P
p
.002 w
.64437 .28749 m .63306 .26566 L .65563 .25658 L p
.661 .662 .84 r
F P
s
P
p
.002 w
.65563 .25658 m .66686 .27833 L .64437 .28749 L p
.661 .662 .84 r
F P
s
P
p
.002 w
.61067 .27519 m .59911 .25116 L .62156 .24175 L p
.659 .64 .82 r
F P
s
P
p
.002 w
.62156 .24175 m .63306 .26566 L .61067 .27519 L p
.659 .64 .82 r
F P
s
P
p
.002 w
.70046 .28924 m .68955 .26974 L .71246 .26174 L p
.648 .682 .868 r
F P
s
P
p
.002 w
.71246 .26174 m .72327 .28131 L .70046 .28924 L p
.648 .682 .868 r
F P
s
P
p
.002 w
.57683 .26082 m .56506 .23408 L .58738 .22457 L p
.652 .613 .797 r
F P
s
P
p
.002 w
.58738 .22457 m .59911 .25116 L .57683 .26082 L p
.652 .613 .797 r
F P
s
P
p
.002 w
.66686 .27833 m .65563 .25658 L .6784 .24799 L p
.652 .659 .844 r
F P
s
P
p
.002 w
.6784 .24799 m .68955 .26974 L .66686 .27833 L p
.652 .659 .844 r
F P
s
P
p
.002 w
.63306 .26566 m .62156 .24175 L .6442 .23274 L p
.653 .638 .823 r
F P
s
P
p
.002 w
.6442 .23274 m .65563 .25658 L .63306 .26566 L p
.653 .638 .823 r
F P
s
P
p
.002 w
.59911 .25116 m .58738 .22457 L .60988 .21527 L p
.648 .613 .8 r
F P
s
P
p
.002 w
.60988 .21527 m .62156 .24175 L .59911 .25116 L p
.648 .613 .8 r
F P
s
P
p
.002 w
.68955 .26974 m .6784 .24799 L .7014 .23992 L p
.643 .655 .847 r
F P
s
P
p
.002 w
.7014 .23992 m .71246 .26174 L .68955 .26974 L p
.643 .655 .847 r
F P
s
P
p
.002 w
.65563 .25658 m .6442 .23274 L .66704 .22415 L p
.645 .636 .827 r
F P
s
P
p
.002 w
.66704 .22415 m .6784 .24799 L .65563 .25658 L p
.645 .636 .827 r
F P
s
P
p
.002 w
.62156 .24175 m .60988 .21527 L .63257 .2063 L p
.643 .612 .803 r
F P
s
P
p
.002 w
.63257 .2063 m .6442 .23274 L .62156 .24175 L p
.643 .612 .803 r
F P
s
P
p
.002 w
.6784 .24799 m .66704 .22415 L .6901 .216 L p
.638 .633 .829 r
F P
s
P
p
.002 w
.6901 .216 m .7014 .23992 L .6784 .24799 L p
.638 .633 .829 r
F P
s
P
p
.002 w
.6442 .23274 m .63257 .2063 L .65546 .1977 L p
.637 .61 .806 r
F P
s
P
p
.002 w
.65546 .1977 m .66704 .22415 L .6442 .23274 L p
.637 .61 .806 r
F P
s
P
p
.002 w
.66704 .22415 m .65546 .1977 L .67857 .18948 L p
.631 .607 .808 r
F P
s
P
p
.002 w
.67857 .18948 m .6901 .216 L .66704 .22415 L p
.631 .607 .808 r
F P
s
P
p
.002 w
.67932 .02494 m
.94594 .43277 L
s
.94594 .43277 m
.97506 .64585 L
s
.97506 .64585 m
.69286 .25814 L
s
.69286 .25814 m
.67932 .02494 L
s
.06024 .26735 m
.02494 .49015 L
s
.02494 .49015 m
.69286 .25814 L
s
.69286 .25814 m
.67932 .02494 L
s
.67932 .02494 m
.06024 .26735 L
s
P
p
p
.002 w
.06024 .26735 m
.67932 .02494 L
s
P
p
.002 w
.07332 .26223 m
.07786 .2665 L
s
P
[(0)] .06425 .25367 1 .94269 Mshowa
p
.002 w
.25485 .19115 m
.25909 .19571 L
s
P
[(10)] .24637 .18201 .92842 1 Mshowa
p
.002 w
.45073 .11445 m
.45461 .11932 L
s
P
[(20)] .44296 .10469 .79605 1 Mshowa
p
.002 w
.66272 .03144 m
.66617 .03663 L
s
P
[(30)] .65583 .02105 .66367 1 Mshowa
p
.001 w
.10856 .24843 m
.11125 .25103 L
s
P
p
.001 w
.14431 .23443 m
.14697 .23706 L
s
P
p
.001 w
.18061 .22022 m
.18323 .22289 L
s
P
p
.001 w
.21745 .20579 m
.22003 .2085 L
s
P
p
.001 w
.29283 .17628 m
.29533 .17905 L
s
P
p
.001 w
.33139 .16118 m
.33385 .16399 L
s
P
p
.001 w
.37055 .14584 m
.37297 .14869 L
s
P
p
.001 w
.41032 .13027 m
.4127 .13316 L
s
P
p
.001 w
.49177 .09837 m
.49406 .10134 L
s
P
p
.001 w
.53348 .08204 m
.53571 .08505 L
s
P
p
.001 w
.57586 .06545 m
.57804 .06849 L
s
P
p
.001 w
.61894 .04858 m
.62106 .05166 L
s
P
P
p
P
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 231},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell[TextData[{
  "The spot swap curve (also known as the par coupon yield curve) is on the \
nearest face, while the forward rate curve runs along the diagonal. ",
  StyleBox["PlotSwapSurface",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " calls ",
  StyleBox["TriangularPlot3D",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", a package that is included in the distribution along with ",
  StyleBox["YieldCurve",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
Now let's calculate the actual and fitted yields. First consider how to \
compute the compouning. The default uses the Securities Industry Association \
Standard Securities Calculation Methods. \
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
Options[YieldCalc]
?Compounding\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
{Compounding -> SIACompounding}\
\>", "\<\
{Compounding -> SIACompounding}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
Compounding is an option for YieldCalc that specifies how to
   handle the compounding of yields. The Default is
   Compounding -> SIACompounding, which specifies the use of
   the Securities Industry Association formulas. Currently
   this handles zeros incorrectly. The other valid setting
   is Compounding -> ContinuousCompounding, which specifies
   the use of continuous compounding for bills and zeros.
   Compounding is also a YCSelector. It returns the setting
   used for the constructions of the YC object.\
\>", "Print",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  StyleBox["SIACompounding",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is useful for comparing with yields quoted in the newpaper. On the other \
hand, to compare the yields on bills or strips to the continuously compounded \
",
  StyleBox["ZeroCurve",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", it is useful to use ",
  StyleBox["Compounding -> ContinuousCompounding",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". (As an alternative, one can use ",
  StyleBox["SemiAnnualZeroCurve",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " instead.)"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["Now let's calculate the actual and fitted yields. ", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["yc = YieldCalc[df5]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
YC[{1988, 7, 19}, {6.44687116785863, 6.53054642936328, 
 
   6.738231055218748, 6.877855448879383, 7.084080011810177, 
 
   7.185707601242326, 6.121040576896599, 5.785428490181357, 
 
   6.142974725473192, 5.968559557482515, 6.188056547768726, 
 
   6.528585872651662, 6.516227930148954, 6.698901876180791, 
 
   6.706200043172133, 6.725311580610667, 6.938854268465102, 
 
   6.969117093152933, 7.020640997694263, 7.071321254137334, 
 
   7.070617023310109, 7.122659569108586, 7.142414796771473, 
 
   7.227115050302042, 7.269488857592584, 7.430044217911712, 
 
   7.506300714095854, 7.561493020022001, 7.628836675306289, 
 
   7.66298613703285, 7.778253353331782, 7.374118366143701, 
 
   6.001846359933836, 5.460918140275984, 5.437042283916535, 
 
   4.037253742516086, 4.191435178630737, 8.56288454448779, 
 
   9.16967391519941, 9.56325158171004, 9.9065207420099, 
 
   8.06989820493338, 9.51794104026271, 9.59778039853952, 
 
   9.48946562895964, 9.62668202287639, 8.30565031318679, 
 
   9.56159843063694, 8.56995519876542, 8.62221456453477, 
 
   9.54742123034835, 8.6572188159929, 9.52272132513079, 
 
   8.76655523174762, 9.48415947617891, 8.71388459452211, 
 
   8.88519684686504, 9.49657250551818, 8.93789103754714, 
 
   9.48527038446347, 8.91065889342485, 9.48627598097059, 
 
   8.945905334446, 8.94237628994466, 9.4856520135914, 
 
   9.23060964333318, 9.24645775348017, 9.48475131381674, 
 
   9.25958877054991, 9.21707923194838, 9.47717557186632, 
 
   9.25027533182218, 9.34036542786222, 9.49595911562644, 
 
   9.36156268955585, 9.36304896659469, 9.37736826855743, 
 
   9.48755989450095, 9.37400093636022, 9.37713797779377, 
 
   9.48948658109339, 9.34566709837694, 9.49143998767899, 
 
   9.3812610974669, 9.40344555439331, 9.35662114197621, 
 
   9.37228207298945, 9.39072201630624, 9.35030764812468, 
 
   9.35622735885886, 9.282503017475, 9.34072897054633, 
 
   9.33931443217225, 9.33054450791756, 9.33227981652984, 
 
   9.31226210270532, 9.2091601966765, 7.539912297858696, 
 
   7.863878165021374, 8.08943871788113, 8.33302202419315, 
 
   8.38369183871122, 8.30200859814042, 8.3821543225323, 
 
   6.785006159443616, 8.4795428410665, 7.487160275018318, 
 
   8.50439447339204, 7.440935961827607, 8.60247800249294, 
 
   7.695678979897881, 8.09867387209915, 8.65817592735205, 
 
   8.17698396569344, 8.68821311854651, 8.26368833197406, 
 
   8.72146193145887, 8.31388403325086, 8.75828408548859, 
 
   7.425041818482312, 8.38318797064487, 8.78262095153771, 
 
   7.819135727450923, 8.42734126157664, 7.91672727340535, 
 
   8.45755196267372, 8.83825857677947, 7.939371710539652, 
 
   8.49919374197303, 8.87201494466396, 8.03173031393839, 
 
   8.52901937581653, 7.257691816045374, 8.91135729705454, 
 
   8.20491386054024, 7.756402929200271, 8.56238435063593, 
 
   8.94585245063765, 8.28999404382816, 7.87559742835195, 
 
   8.60640719758274, 8.96688174422947, 8.3451549878421, 
 
   7.93878720943857, 8.62021957608865, 8.99251727609621, 
 
   8.37679829551064, 8.0987902664399, 8.64938010051311, 
 
   7.515525903382717, 8.99827997104389, 8.43118731070203, 
 
   8.18227586043668, 8.66365946344484, 7.811472571488033, 
 
   9.0230834281176, 8.44656791834923, 8.22391261538601, 
 
   8.6955736548306, 7.935764358027592, 9.05295160058825, 
 
   8.49688023523418, 8.29132835692596, 8.71961617979021, 
 
   8.04308186059657, 7.06215520528275, 8.52931770697654, 
 
   7.199845120125204, 8.35973216750261, 7.398322614167542, 
 
   8.76171617722822, 8.14101272913202, 9.08273592205816, 
 
   7.511724574591279, 8.56613093590114, 7.744044177714848, 
 
   8.39268148997195, 8.80362158135648, 7.702188692190579, 
 
   8.20056470784545, 7.77132794076489, 8.60264262025665, 
 
   7.809032154218381, 8.44562065128393, 8.83040164347083, 
 
   7.899507292212341, 8.25913433709476, 9.11903001138754, 
 
   7.943307401922978, 8.61730255426895, 7.960845531024709, 
 
   8.46996812399942, 8.86106269805941, 8.00570057216705, 
 
   8.3215817793836, 9.1319657375864, 8.03544469669332, 
 
   8.64269878463453, 8.09065359341933, 8.50261226880232, 
 
   8.89092745267201, 8.09396040065264, 8.36794501049531, 
 
   9.14412632919457, 8.13276747181384, 8.66815770472041, 
 
   8.19589398470502, 8.51344553158035, 8.92104908481101, 
 
   8.18949255063575, 8.40325862625631, 9.14171643416015, 
 
   8.21188680837245, 8.72673524729204, 8.20768209639422, 
 
   8.57141654356673, 8.95393434059781, 8.24700716970834, 
 
   9.09885917692801, 8.28120211882497, 8.70843668460133, 
 
   8.2966439545584, 8.58961870099215, 8.94784972533604}, 
 
  {7.177286658111648, 7.223899194294484, 7.276570177243958, 
 
   7.334523938428974, 7.399103995760463, 7.462479416427265, 
 
   7.156435192645407, 7.166654369418943, 7.188332851567686, 
 
   7.199788443130015, 7.211648847868237, 7.236528723259435, 
 
   7.249524286696183, 7.262875619848935, 7.290596949342583, 
 
   7.304941308696673, 7.349728843077654, 7.365188720841297, 
 
   7.380887625425409, 7.41294635977802, 7.429278897471981, 
 
   7.445795518821952, 7.47931545478117, 7.496289793516612, 
 
   7.513385486924816, 7.600092964175791, 7.670105363862616, 
 
   7.740026614263079, 7.809438906103014, 7.877942967439472, 
 
   7.945343825856642, 8.19934837903627, 8.61954818315619, 
 
   8.682111176561, 8.77966391188427, 8.99510403024106, 
 
   9.25501891788364, 8.66817067158778, 8.56614797169219, 
 
   8.82398218718789, 8.94332881163469, 8.23402281231723, 
 
   9.23816828280638, 8.99013999073285, 9.07545996068598, 
 
   9.33743627173029, 8.60679758196152, 9.37561583540049, 
 
   8.66298644268776, 8.71904068165846, 9.40360634607771, 
 
   8.75026065923016, 9.40378835905479, 8.79162135980738, 
 
   9.40557955699777, 8.8618284888565, 8.88618196407848, 
 
   9.3930033480982, 8.92166285437932, 9.38396491929077, 
 
   8.92437001180885, 9.39922728583381, 8.9413193825058, 
 
   8.98378382865151, 9.37574903984504, 9.23974572509541, 
 
   9.23357439559723, 9.36986120148308, 9.24774624711617, 
 
   9.23987063257631, 9.37243759839136, 9.26759823315373, 
 
   9.32717356466326, 9.40069456826967, 9.3560326955702, 
 
   9.35864633261629, 9.36724517101776, 9.39258051520129, 
 
   9.36435593380003, 9.36773278894669, 9.38221412406708, 
 
   9.36192106905295, 9.38977959303843, 9.3796250120982, 
 
   9.38971051130617, 9.35700257507336, 9.3789321112793, 
 
   9.39533232550651, 9.35177480368417, 9.34617495642025, 
 
   9.41229993444343, 9.34643483089494, 9.34016610612497, 
 
   9.33467868271164, 9.33155698514879, 9.33349062031542, 
 
   9.32932358775562, 7.499318298940004, 7.794752421936312, 
 
   8.07768966409212, 8.30766458321142, 8.35728350428025, 
 
   8.42028193785263, 8.46147220800903, 7.516422225999009, 
 
   8.49194715975611, 7.776516616124295, 8.54242529947942, 
 
   7.800492915632139, 8.54124292964666, 7.964128888897911, 
 
   8.04211485086517, 8.61522243186423, 8.17981645568799, 
 
   8.65041209288604, 8.21829852523715, 8.67521799285622, 
 
   8.30407531441815, 8.70391657052836, 7.49931829894001, 
 
   8.33579212459613, 8.73362099188577, 7.726644117984351, 
 
   8.40527968981574, 7.793647140462065, 8.41410159326655, 
 
   8.79567173203992, 7.954059876339456, 8.45863725136471, 
 
   8.8368032219696, 8.07498010902492, 8.48204198252385, 
 
   7.399863217600195, 8.87550217271803, 8.18628942910059, 
 
   7.782760090263365, 8.52939518566257, 8.91607949142641, 
 
   8.22801111953616, 7.742702169285681, 8.53908006313226, 
 
   8.93445675873273, 8.3094712479177, 7.889083403217436, 
 
   8.57877280978326, 8.97151772189247, 8.36567512486525, 
 
   8.06223567773942, 8.59892066563841, 7.499318298940019, 
 
   9.00108251285541, 8.42006160763772, 8.17701551005946, 
 
   8.6457865407974, 7.727399841494275, 9.03223483149895, 
 
   8.43718036160023, 8.23808705058479, 8.66737724637215, 
 
   7.795836811116497, 9.0600840911867, 8.481303592734, 
 
   8.30690640812594, 8.71096079643781, 7.95948991204682, 
 
   7.382617342196736, 8.51635560787445, 7.39986321760046, 
 
   8.35944048772438, 7.424771205422218, 8.74174784057247, 
 
   8.08395759954117, 9.09903855744365, 7.534139469284117, 
 
   8.56951588262458, 7.782760090263328, 8.40651459982039, 
 
   8.80490707668445, 7.673508443065662, 8.19388763058421, 
 
   7.66190068998503, 8.56529649639054, 7.744391001784393, 
 
   8.44067965807805, 8.81720957396184, 7.782091155564256, 
 
   8.23477242049047, 9.11078496543933, 7.847894229825123, 
 
   8.6029059641995, 7.88953405428303, 8.47181118750585, 
 
   8.85617059254401, 7.920372116207853, 8.31374615217024, 
 
   9.1271761525124, 7.989484051699757, 8.62386493913699, 
 
   8.06338028348467, 8.49479198862787, 8.88165364582986, 
 
   8.05395718127101, 8.36998677966269, 9.13839348439273, 
 
   8.10019700157657, 8.66131867051721, 8.17773095992007, 
 
   8.53231646543895, 8.93397822778047, 8.16623305917976, 
 
   8.42491288896646, 9.1623752721339, 8.17023514642542, 
 
   8.68720020942327, 8.23784362019627, 8.5553256149286, 
 
   8.95433062320979, 8.21509037164971, 9.15237356871093, 
 
   8.25982687800937, 8.71841336401073, 8.3054536143051, 
 
   8.57919052772754, 8.9846602847759}, 
 
  {-73.04154902530175, -69.33527649312046, 
 
   -53.83391220252101, -45.66684895495908, 
 
   -31.50239839502857, -27.67718151849383, 
 
   -103.5394615748808, -138.1225879237587, 
 
   -104.5358126094494, -123.12288856475, -102.359230009951, 
 
   -70.79428506077728, -73.32963565472284, 
 
   -56.39737436681447, -58.43969061704506, 
 
   -57.96297280860064, -41.0874574612552, 
 
   -39.60716276883644, -36.02466277311463, 
 
   -34.1625105640686, -35.86618741618714, 
 
   -32.31359497133654, -33.69006580096973, 
 
   -26.91747432145695, -24.3896629332232, 
 
   -17.00487462640794, -16.3804649766762, 
 
   -17.85335942410784, -18.06022307967252, 
 
   -21.49568304066216, -16.70904725248601, 
 
   -82.5230012892566, -261.7701823222357, 
 
   -322.1193036285014, -334.2621627967732, 
 
   -495.7850287724972, -506.3583739252903, 
 
   -10.52861270999905, 60.35259435072185, 
 
   73.92693945221432, 96.3191930375208, -16.41246073838509, 
 
   27.97727574563336, 60.76404078066755, 41.40056682736564, 
 
   28.92457511461011, -30.11472687747272, 
 
   18.59825952364513, -9.30312439223382, -9.68261171236904, 
 
   14.38148842706433, -9.30418432372591, 11.89329660759988, 
 
   -2.506612805976616, 7.857991918113961, 
 
   -14.79438943343876, -0.0985117213438969, 
 
   10.35691574199831, 1.62281831678186, 10.13054651727021, 
 
   -1.371111838400729, 8.70486951367742, 
 
   0.4585951940194733, -4.140753870685466, 
 
   10.9902973746367, -0.91360817622288, 1.288335788294148, 
 
   11.4890112333665, 1.184252343373338, -2.279140062792884, 
 
   10.47379734749505, -1.732290133155878, 
 
   1.319186319895493, 9.5264547356777, 0.5529993985648574, 
 
   0.4402633978402904, 1.012309753967244, 9.49793792996623, 
 
   0.964500256018397, 0.940518884707409, 10.72724570263137, 
 
   -1.625397067600787, 10.16603946405574, 
 
   0.1636085368698303, 1.373504308714146, 
 
   -0.03814330971572133, -0.6650038289850003, 
 
   -0.4610309200272056, -0.1467155559497613, 
 
   1.005240243860506, -12.9796916968429, 
 
   -0.5705860348609093, -0.0851673952723786, 
 
   -0.413417479408551, 0.07228313810472287, 
 
   -2.122851761010303, -12.01633910791173, 
 
   4.059399891869209, 6.912574308506159, 1.174905378900704, 
 
   2.535744098172721, 2.640833443097001, 
 
   -11.82733397122124, -7.93178854767298, 
 
   -73.14160665553926, -1.240431868960989, 
 
   -28.9356341105977, -3.803082608737895, 
 
   -35.95569538045318, 6.123507284627472, 
 
   -26.84499090000294, 5.655902123398349, 
 
   4.295349548782035, -0.2832489994554876, 
 
   3.780102566046395, 4.538980673691029, 4.624393860264853, 
 
   0.980871883271206, 5.436751496023007, 
 
   -7.427648045769786, 4.739584604874203, 
 
   4.899995965194926, 9.24916094665722, 2.206157176089185, 
 
   12.30801329432847, 4.345036940716796, 4.258684473955121, 
 
   -1.468816579980459, 4.055649060831179, 
 
   3.521172269435979, -4.324979508653826, 4.69773932926767, 
 
   -14.21714015548208, 3.585512433650884, 
 
   1.862443143964931, -2.635716106309349, 
 
   3.298916497336357, 2.977295921123968, 6.198292429199804, 
 
   13.28952590662684, 6.732713445047978, 3.242498549673911, 
 
   3.568373992439433, 4.970380622113347, 4.144676630538769, 
 
   2.099955420374044, 1.112317064539603, 3.6554588700481, 
 
   5.045943487469807, 1.62076044426982, 
 
   -0.2802541811524506, 1.112570306430882, 
 
   0.5260350377220391, 1.787292264744167, 8.40727299937587, 
 
   -0.915140338135423, 0.938755674899916, 
 
   -1.417443519877715, 2.819640845844518, 
 
   13.99275469110952, -0.7132490598451114, 
 
   1.557664250017687, -1.557805119997546, 
 
   0.865538335239435, 8.35919485497545, -32.04621369139859, 
 
   1.296209910209356, -20.00180974752554, 
 
   0.02916797782308578, -2.644859125467569, 
 
   1.996833665574726, 5.705512959085191, 
 
   -1.630263538548959, -2.241489469283753, 
 
   -0.3384946723446802, -3.871591254847928, 
 
   -1.383310984843789, -0.1285495327971731, 
 
   2.868024912491673, 0.6677077261239007, 
 
   10.94272507798601, 3.734612386611324, 6.464115243398805, 
 
   0.494099320588326, 1.31920695089871, 11.74161366480853, 
 
   2.436191660428655, 0.824504594821285, 9.54131720978548, 
 
   1.439659006944893, 7.131147674167871, 
 
   -0.1843063506424158, 0.4892105515397205, 
 
   8.53284559591927, 0.783562721336395, 0.4789585074004066, 
 
   4.596064499356701, 1.883384549753408, 2.727330993465671, 
 
   0.7820280174446026, 0.927380684215073, 
 
   4.000321938163154, -0.2041769167380636, 
 
   0.5732844801837445, 3.257047023727289, 
 
   0.6839034203204975, 1.816302478494691, 
 
   -1.887093385859728, -1.292914296946002, 
 
   2.325949145599182, -2.165426271014326, 
 
   -2.065883797375179, 4.165166194703396, 
 
   3.953503786876489, -3.016152380205384, 
 
   1.609092863813366, -0.0396282611978549, 
 
   3.191679805863679, -5.351439178292062, 
 
   2.137524081560294, -0.997667940940339, 
 
   -0.880965974669756, 1.042817326460721, 
 
   -3.681055943985712}, {0.04155504166666667, 
 
   0.1185199444444444, 0.1946356527777778, 0.269895175, 
 
   0.3468199111111111, 0.4176345833333334, 
 
   0.002777305555555555, 0.02219368888888889, 
 
   0.06088255555555557, 0.0801700972222222, 0.099385, 
 
   0.1376408333333333, 0.1567164333333333, 
 
   0.1756855111111111, 0.2135635666666667, 
 
   0.2324204583333333, 0.2885490777777778, 
 
   0.3071694694444444, 0.325711, 0.3626762111111111, 
 
   0.381112425, 0.3994154111111111, 0.435987, 
 
   0.4540400944444445, 0.4721102222222222, 
 
   0.5616521083333334, 0.6323754111111111, 
 
   0.7022819416666667, 0.7712316111111112, 
 
   0.839503019444445, 0.9062703, 1.487874161298511, 
 
   3.661166852476839, 4.099196713535762, 4.688076595386082, 
 
   5.873768219550401, 8.51680691084916, 3.822949490349042, 
 
   3.049269132760217, 4.469526860183771, 4.931287209387496, 
 
   1.664647860031613, 7.219280634207075, 5.170672517849584, 
 
   5.728424351920875, 7.863325819983352, 3.445954266053583, 
 
   8.03506150802448, 3.757296532643033, 4.04089397709261, 
 
   8.12327356074801, 4.199108454632817, 8.11945849871945, 
 
   4.320230942849308, 8.17775774543052, 4.637771046551227, 
 
   4.664208378903078, 8.07899105020717, 4.764389282584821, 
 
   7.948673263899886, 4.744631840560709, 8.244659505726, 
 
   4.905895759321145, 5.06534155296511, 7.990300879784217, 
 
   7.006302425200485, 6.931809180583702, 7.977789787592712, 
 
   6.973862279856911, 6.852766571779547, 8.06356157035992, 
 
   7.033636377021031, 7.48498938493428, 8.64362113315392, 
 
   7.65489074931439, 7.719366612442365, 7.723382278123644, 
 
   8.55728145446822, 7.691908584148137, 7.748520429570549, 
 
   8.54731128569712, 7.670275730764842, 8.65447832525504, 
 
   7.946658308184079, 8.81507147181066, 9.51400769643548, 
 
   8.01074861424297, 8.21490962977975, 9.63234600485264, 
 
   9.74356624539707, 8.57171398982045, 9.85763084192722, 
 
   10.2392211989816, 10.22321384152147, 10.03134351047815, 
 
   10.04156089666995, 10.1344563348931, 0.3196274117923905, 
 
   0.7826782000448961, 1.222288579343281, 
 
   1.840736351386401, 2.009547471756114, 2.352050782834996, 
 
   2.513848915804091, 0.245937006657295, 2.695252068438737, 
 
   0.4703500341529791, 2.847326209832509, 
 
   0.7069065248286215, 3.030772853620483, 
 
   0.920105206791626, 1.148029897498641, 3.463353473572165, 
 
   1.36052976558052, 3.603520825200956, 1.575939459191527, 
 
   3.796790030023919, 1.77110889177138, 3.857128646422602, 
 
   0.3213522085422959, 1.96615588369424, 4.00923431599712, 
 
   0.5499020390933082, 2.147205505113497, 
 
   0.7803588453006624, 2.328437871810787, 4.20989772147622, 
 
   0.995965807287683, 2.47338543450573, 4.365267350475671, 
 
   1.214923392771401, 2.685382381061793, 
 
   0.2005933175336271, 4.557535708809813, 
 
   1.430701583653417, 0.433999355407792, 2.866876883843253, 
 
   4.708737095055649, 1.63563485930027, 0.6666266782365258, 
 
   3.037409978570095, 4.836861552243523, 1.850514550172554, 
 
   0.892981019753478, 3.244931547518816, 5.017563301443761, 
 
   2.055532504893026, 1.117639956298995, 3.442117017847298, 
 
   0.3195837929673496, 5.234096124953198, 
 
   2.258161924442953, 1.33856386236322, 3.649060114014663, 
 
   0.5500375574059627, 5.419263305844087, 
 
   2.475496669379745, 1.564410746679352, 3.898890567711826, 
 
   0.7844988698942506, 5.724088683500906, 
 
   2.683303461797633, 1.776525049778835, 4.069426413083345, 
 
   1.012472365069093, 0.1158970893033911, 
 
   2.909538817786684, 0.1977312579178477, 
 
   1.995008089642541, 0.2810114536558008, 
 
   4.248218225960693, 1.240209326523133, 5.994173192485564, 
 
   0.3567667401424653, 3.095383928807628, 
 
   0.4330518048172153, 2.203583922513569, 
 
   4.414107360793543, 0.5131754643634784, 
 
   1.460003927292405, 0.5905006202244778, 
 
   3.284991897535492, 0.6683740267968861, 
 
   2.406714670207112, 4.578682500745057, 
 
   0.7480695845848699, 1.666433467746553, 
 
   6.055690377245262, 0.820786412786172, 3.391733365884741, 
 
   0.89668418510055, 2.576211322597796, 4.640009603082039, 
 
   0.974368378510659, 1.874615893249559, 6.147721667667318, 
 
   1.048390237867832, 3.560151482364549, 1.120432062357415, 
 
   2.734267025205459, 4.666794151129003, 1.199259580368375, 
 
   2.080041980847595, 6.223608589874324, 1.27023279700495, 
 
   3.733570446807222, 1.341200013959821, 2.943452251535351, 
 
   4.872128548558545, 1.419534672208175, 2.291845005990017, 
 
   6.436313115154544, 1.493091012524336, 3.937424303692423, 
 
   1.563545849416774, 3.141308242296187, 5.038071076322249, 
 
   1.635345673725165, 6.421891804209486, 1.698258201120595, 
 
   4.029836010809989, 1.768261221626748, 3.303723924086356, 
 
   5.120338147192653}, {0.001726821487918403, 
 
   0.0140469772311142, 0.03788303733223167, 
 
   0.07284340548828061, 0.120284050743119, 
 
   0.1744186451960069, 7.713426148919753*10^-6, 
 
   0.0004925598264967902, 0.003706685570975309, 
 
   0.006427244488620563, 0.009877378225, 
 
   0.01894499900069444, 0.02456004047672112, 
 
   0.03086539881437234, 0.04560939700738779, 
 
   0.05401926945187673, 0.083260570286406, 
 
   0.0943530829587815, 0.106087655521, 0.1315340341059112, 
 
   0.1452466804893806, 0.1595326706330579, 0.190084664169, 
 
   0.20615240736312, 0.222888061926716, 0.3154530907952783, 
 
   0.3998986605779468, 0.4931999255911034, 
 
   0.5947981979770401, 0.7047653196563393, 
 
   0.82132585666209, 1.476562722269404, 7.91368760249415, 
 
   9.83702353895295, 12.83367494230551, 19.72360580477777, 
 
   42.49310023620407, 8.86031767985242, 5.704324176451276, 
 
   12.45994458044353, 15.24180436556862, 1.823103944168497, 
 
   35.69968629487396, 16.94513023653118, 21.01384209330493, 
 
   43.4324715947116, 7.234491037445947, 46.0165152479417, 
 
   8.64945848722894, 10.09710627989962, 47.89341650277489, 
 
   10.91860804883259, 48.17598584596911, 11.63559497136388, 
 
   49.38280511461916, 13.4444712103264, 13.82174453299368, 
 
   49.00164366250616, 14.53124570120509, 47.9531544401818, 
 
   14.69774184034777, 51.27748780330389, 15.43652771162114, 
 
   16.73726291348765, 49.11261750483278, 34.99483434097816, 
 
   34.65642141767406, 49.42711855517053, 35.27485212043882, 
 
   34.48512060107889, 50.80456549501683, 36.27754993342402, 
 
   41.03024139809524, 58.45989596653169, 42.8746337150165, 
 
   43.74766898751844, 44.08766337294601, 58.15605988144156, 
 
   44.04730118826164, 45.11931160760933, 58.64590796933531, 
 
   44.5905348831076, 60.23714515530142, 47.62240500707007, 
 
   62.42342477154916, 76.5595334970922, 48.79604304742986, 
 
   51.22829547951013, 78.80202752728901, 80.7762009262152, 
 
   55.59201394724391, 82.8339770655017, 89.3156911929352, 
 
   89.515083543331, 86.7025831046679, 87.0585080270773, 
 
   88.9754111406192, 0.1006036988348557, 
 
   0.4945993166285664, 1.05363574211274, 2.209900271485824, 
 
   2.619237612957018, 3.559097623520756, 4.06343860017434, 
 
   0.05820782856605159, 4.642668217853904, 
 
   0.2208091493606404, 5.192130273746849, 
 
   0.418980804590653, 5.850418437045794, 
 
   0.6511430204168392, 0.944836967256903, 
 
   7.496477483489132, 1.276616263346109, 8.15366457747731, 
 
   1.657726577734561, 9.00942512732499, 2.060216040269714, 
 
   9.44800668229647, 0.1011831674679113, 2.501215601961261, 
 
   10.21810705424948, 0.2826881661879565, 
 
   2.964402351560014, 0.4921784474219711, 
 
   3.460835639495601, 11.44659935519561, 
 
   0.7424114055833967, 3.923009547355197, 
 
   12.31576573750298, 1.044811473564499, 4.564403047043953, 
 
   0.03890550993891148, 13.36434068288437, 
 
   1.39574834164075, 0.1873040297495987, 5.185819285699601, 
 
   14.28456214224507, 1.777949655221947, 
 
   0.3813036659915722, 5.808093137960014, 
 
   15.11951241030318, 2.226010905969399, 
 
   0.6164722121048069, 6.581110913182124, 
 
   16.23799196003884, 2.703066388978156, 0.900089763735402, 
 
   7.361574062858103, 0.1005976912458481, 
 
   17.56091119862652, 3.224750813059908, 1.234809867596668, 
 
   8.22571749264239, 0.2830302968578656, 18.7868696709685, 
 
   3.820763798812791, 1.625078208644771, 9.27130853663451, 
 
   0.496394989136493, 20.66398097084849, 4.448171721424561, 
 
   2.049783832443254, 10.10952830917067, 
 
   0.7595974836006335, 0.01312162400470235, 
 
   5.165028250542409, 0.03835471045937707, 
 
   2.53071099990717, 0.07788643915463922, 
 
   11.00685152181669, 1.07536023505842, 22.7978646958901, 
 
   0.1261999029656604, 5.833376354358571, 
 
   0.1869050579967361, 3.047680601622716, 
 
   11.90099156926603, 0.2549826231035562, 
 
   1.435260486527306, 0.3158171272019457, 
 
   6.549426842966273, 0.3834559312761484, 
 
   3.596795975150932, 12.81808937040764, 
 
   0.4593452827898308, 1.824793038108986, 
 
   23.82994874796695, 0.5348849185692655, 
 
   7.071135739775228, 0.6200148080683079, 
 
   4.122846213736702, 13.35694090018424, 
 
   0.7129905657618285, 2.266075638997103, 
 
   24.70953227460945, 0.806150470754722, 7.792833396545304, 
 
   0.903345851291448, 4.651576360370714, 13.74242710781433, 
 
   1.015780618466151, 2.74779049161266, 25.49150454582217, 
 
   1.123631344308068, 8.57115545974449, 1.238088052148131, 
 
   5.347843985034697, 14.91533040026254, 1.368350151386055, 
 
   3.291261973178658, 27.14920113432341, 1.494938775312991, 
 
   9.48266086080024, 1.623907438922168, 6.054006789066827, 
 
   15.94356238489095, 1.762092914348646, 27.39349095339164, 
 
   1.89050867922857, 10.0561369405233, 2.036751610687871, 
 
   6.712513071033305, 16.62988853348355}, SIACompounding, 
 
  Hold[df5]]\
\>", "\<\
-YC[{1988, 7, 19}]-\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  StyleBox["YieldCalc",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " calculates yield to maturity for each of the bonds in the sample (",
  StyleBox["Yield[yc]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ") and the difference between that yield and the predicted yield (",
  StyleBox["YieldError[yc]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "), among other things. ",
  StyleBox["YieldCalc",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " can be run on a ",
  StyleBox["CP",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object, in which case it will not attempt to calculate predicted yields."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["MaturityPlot[Yield[yc]];", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0237221 0.0319428 -0.390164 0.100286 [
[(0)] .02372 0 0 2 0 Minner Mrotsboxa
[(5)] .18344 0 0 2 0 Minner Mrotsboxa
[(10)] .34315 0 0 2 0 Minner Mrotsboxa
[(15)] .50286 0 0 2 0 Minner Mrotsboxa
[(20)] .66258 0 0 2 0 Minner Mrotsboxa
[(25)] .82229 0 0 2 0 Minner Mrotsboxa
[(30)] .98201 0 0 2 0 Minner Mrotsboxa
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotsboxa
[(4)] -0.0125 .01098 1 0 0 Minner Mrotsboxa
[(5)] -0.0125 .11126 1 0 0 Minner Mrotsboxa
[(6)] -0.0125 .21155 1 0 0 Minner Mrotsboxa
[(7)] -0.0125 .31184 1 0 0 Minner Mrotsboxa
[(8)] -0.0125 .41212 1 0 0 Minner Mrotsboxa
[(9)] -0.0125 .51241 1 0 0 Minner Mrotsboxa
[(10)] -0.0125 .61269 1 0 0 Minner Mrotsboxa
[(Yield)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotsboxa
[(7/19/1988)] .5 .61803 0 -2 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
.002 w
.02372 0 m
.02372 .00625 L
s
P
[(0)] .02372 0 0 2 0 Minner Mrotshowa
p
.002 w
.18344 0 m
.18344 .00625 L
s
P
[(5)] .18344 0 0 2 0 Minner Mrotshowa
p
.002 w
.34315 0 m
.34315 .00625 L
s
P
[(10)] .34315 0 0 2 0 Minner Mrotshowa
p
.002 w
.50286 0 m
.50286 .00625 L
s
P
[(15)] .50286 0 0 2 0 Minner Mrotshowa
p
.002 w
.66258 0 m
.66258 .00625 L
s
P
[(20)] .66258 0 0 2 0 Minner Mrotshowa
p
.002 w
.82229 0 m
.82229 .00625 L
s
P
[(25)] .82229 0 0 2 0 Minner Mrotshowa
p
.002 w
.98201 0 m
.98201 .00625 L
s
P
[(30)] .98201 0 0 2 0 Minner Mrotshowa
p
.001 w
.05566 0 m
.05566 .00375 L
s
P
p
.001 w
.08761 0 m
.08761 .00375 L
s
P
p
.001 w
.11955 0 m
.11955 .00375 L
s
P
p
.001 w
.15149 0 m
.15149 .00375 L
s
P
p
.001 w
.21538 0 m
.21538 .00375 L
s
P
p
.001 w
.24732 0 m
.24732 .00375 L
s
P
p
.001 w
.27926 0 m
.27926 .00375 L
s
P
p
.001 w
.31121 0 m
.31121 .00375 L
s
P
p
.001 w
.37509 0 m
.37509 .00375 L
s
P
p
.001 w
.40704 0 m
.40704 .00375 L
s
P
p
.001 w
.43898 0 m
.43898 .00375 L
s
P
p
.001 w
.47092 0 m
.47092 .00375 L
s
P
p
.001 w
.53481 0 m
.53481 .00375 L
s
P
p
.001 w
.56675 0 m
.56675 .00375 L
s
P
p
.001 w
.59869 0 m
.59869 .00375 L
s
P
p
.001 w
.63064 0 m
.63064 .00375 L
s
P
p
.001 w
.69452 0 m
.69452 .00375 L
s
P
p
.001 w
.72646 0 m
.72646 .00375 L
s
P
p
.001 w
.75841 0 m
.75841 .00375 L
s
P
p
.001 w
.79035 0 m
.79035 .00375 L
s
P
p
.001 w
.85423 0 m
.85423 .00375 L
s
P
p
.001 w
.88618 0 m
.88618 .00375 L
s
P
p
.001 w
.91812 0 m
.91812 .00375 L
s
P
p
.001 w
.95006 0 m
.95006 .00375 L
s
P
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotshowa
p
.002 w
0 0 m
1 0 L
s
P
p
.002 w
0 .01098 m
.00625 .01098 L
s
P
[(4)] -0.0125 .01098 1 0 0 Minner Mrotshowa
p
.002 w
0 .11126 m
.00625 .11126 L
s
P
[(5)] -0.0125 .11126 1 0 0 Minner Mrotshowa
p
.002 w
0 .21155 m
.00625 .21155 L
s
P
[(6)] -0.0125 .21155 1 0 0 Minner Mrotshowa
p
.002 w
0 .31184 m
.00625 .31184 L
s
P
[(7)] -0.0125 .31184 1 0 0 Minner Mrotshowa
p
.002 w
0 .41212 m
.00625 .41212 L
s
P
[(8)] -0.0125 .41212 1 0 0 Minner Mrotshowa
p
.002 w
0 .51241 m
.00625 .51241 L
s
P
[(9)] -0.0125 .51241 1 0 0 Minner Mrotshowa
p
.002 w
0 .61269 m
.00625 .61269 L
s
P
[(10)] -0.0125 .61269 1 0 0 Minner Mrotshowa
p
.001 w
0 .03104 m
.00375 .03104 L
s
P
p
.001 w
0 .05109 m
.00375 .05109 L
s
P
p
.001 w
0 .07115 m
.00375 .07115 L
s
P
p
.001 w
0 .09121 m
.00375 .09121 L
s
P
p
.001 w
0 .13132 m
.00375 .13132 L
s
P
p
.001 w
0 .15138 m
.00375 .15138 L
s
P
p
.001 w
0 .17144 m
.00375 .17144 L
s
P
p
.001 w
0 .19149 m
.00375 .19149 L
s
P
p
.001 w
0 .23161 m
.00375 .23161 L
s
P
p
.001 w
0 .25166 m
.00375 .25166 L
s
P
p
.001 w
0 .27172 m
.00375 .27172 L
s
P
p
.001 w
0 .29178 m
.00375 .29178 L
s
P
p
.001 w
0 .33189 m
.00375 .33189 L
s
P
p
.001 w
0 .35195 m
.00375 .35195 L
s
P
p
.001 w
0 .37201 m
.00375 .37201 L
s
P
p
.001 w
0 .39206 m
.00375 .39206 L
s
P
p
.001 w
0 .43218 m
.00375 .43218 L
s
P
p
.001 w
0 .45224 m
.00375 .45224 L
s
P
p
.001 w
0 .47229 m
.00375 .47229 L
s
P
p
.001 w
0 .49235 m
.00375 .49235 L
s
P
p
.001 w
0 .53246 m
.00375 .53246 L
s
P
p
.001 w
0 .55252 m
.00375 .55252 L
s
P
p
.001 w
0 .57258 m
.00375 .57258 L
s
P
p
.001 w
0 .59264 m
.00375 .59264 L
s
P
[(Yield)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotshowa
p
.002 w
0 0 m
0 .61803 L
s
P
P
p
[(7/19/1988)] .5 .61803 0 -2 Mshowa
P
p
P
p
.008 w
.02381 .22369 Mdot
.02442 .19003 Mdot
.02503 .25637 Mdot
.02565 .22589 Mdot
.02626 .2084 Mdot
.02687 .23041 Mdot
.0274 .31807 Mdot
.02748 .26476 Mdot
.02809 .26456 Mdot
.02871 .26332 Mdot
.02932 .28164 Mdot
.02993 .28558 Mdot
.03002 .33188 Mdot
.03002 .33768 Mdot
.03054 .28237 Mdot
.03116 .28429 Mdot
.03151 .29028 Mdot
.03238 .29959 Mdot
.03273 .35178 Mdot
.03299 .3057 Mdot
.0336 .30874 Mdot
.03404 .35446 Mdot
.03404 .36354 Mdot
.03404 .36598 Mdot
.03422 .31391 Mdot
.03492 .32027 Mdot
.03535 .36315 Mdot
.03544 .31899 Mdot
.03605 .31892 Mdot
.03667 .32414 Mdot
.03728 .33046 Mdot
.03789 .32612 Mdot
.03833 .38645 Mdot
.03833 .38769 Mdot
.0385 .33461 Mdot
.03911 .33886 Mdot
.03955 .36069 Mdot
.04078 .38226 Mdot
.04209 .39322 Mdot
.04209 .39398 Mdot
.04217 .35496 Mdot
.04322 .38919 Mdot
.04462 .36261 Mdot
.04594 .39297 Mdot
.04594 .39965 Mdot
.04707 .36815 Mdot
.04742 .35606 Mdot
.04865 .40204 Mdot
.04952 .3749 Mdot
.04987 .39847 Mdot
.04987 .40377 Mdot
.04987 .40568 Mdot
.05127 .40644 Mdot
.05197 .37832 Mdot
.05389 .40598 Mdot
.05389 .4082 Mdot
.05442 .38988 Mdot
.05538 .3816 Mdot
.05661 .41269 Mdot
.05792 .40604 Mdot
.05792 .41644 Mdot
.05932 .41568 Mdot
.06211 .42121 Mdot
.06211 .42203 Mdot
.06334 .42202 Mdot
.06465 .42154 Mdot
.06596 .4153 Mdot
.06596 .42109 Mdot
.06596 .42626 Mdot
.06727 .42544 Mdot
.07016 .4304 Mdot
.07016 .43177 Mdot
.07138 .42987 Mdot
.0727 .43113 Mdot
.07401 .34936 Mdot
.07401 .43224 Mdot
.07401 .43267 Mdot
.07515 .43337 Mdot
.07803 .43295 Mdot
.07803 .43458 Mdot
.07926 .43857 Mdot
.08048 .43689 Mdot
.08179 .41913 Mdot
.08179 .43811 Mdot
.08179 .4412 Mdot
.08319 .44032 Mdot
.08599 .44134 Mdot
.08599 .44187 Mdot
.08721 .4436 Mdot
.08984 .44437 Mdot
.08984 .44552 Mdot
.08984 .44674 Mdot
.09395 .4482 Mdot
.09517 .45055 Mdot
.09788 .44902 Mdot
.09788 .44991 Mdot
.09788 .4506 Mdot
.10191 .4515 Mdot
.10322 .45498 Mdot
.10593 .45256 Mdot
.10593 .45536 Mdot
.10986 .45681 Mdot
.11109 .45801 Mdot
.11371 .44241 Mdot
.11371 .45691 Mdot
.11782 .45925 Mdot
.11905 .46218 Mdot
.12176 .45045 Mdot
.12176 .46195 Mdot
.12578 .46253 Mdot
.12709 .46518 Mdot
.1298 .46021 Mdot
.1298 .46521 Mdot
.13383 .46361 Mdot
.13514 .46852 Mdot
.13811 .46271 Mdot
.13811 .4689 Mdot
.13811 .52942 Mdot
.14179 .46943 Mdot
.1431 .47294 Mdot
.14572 .47254 Mdot
.14572 .47256 Mdot
.14974 .47125 Mdot
.15106 .47432 Mdot
.15394 .21174 Mdot
.15394 .44277 Mdot
.15394 .47403 Mdot
.1591 .47725 Mdot
.1619 .47658 Mdot
.1619 .47813 Mdot
.16715 .47868 Mdot
.16995 .15749 Mdot
.16995 .46857 Mdot
.16995 .46928 Mdot
.16995 .47913 Mdot
.16995 .48114 Mdot
.17502 .48188 Mdot
.17782 .48447 Mdot
.17782 .485 Mdot
.18298 .48429 Mdot
.18578 .47452 Mdot
.18578 .48317 Mdot
.18578 .48817 Mdot
.19102 .48851 Mdot
.19373 .15509 Mdot
.19373 .47803 Mdot
.19373 .49061 Mdot
.19933 .49271 Mdot
.20178 .489 Mdot
.20694 .4954 Mdot
.20965 .49619 Mdot
.20965 .56889 Mdot
.2149 .49847 Mdot
.21761 .48371 Mdot
.21761 .49957 Mdot
.22312 .50147 Mdot
.22565 .50089 Mdot
.22565 .50352 Mdot
.23116 .50449 Mdot
.2337 .01472 Mdot
.2337 .50618 Mdot
.2337 .50698 Mdot
.2337 .60332 Mdot
.23904 .50779 Mdot
.24148 .50345 Mdot
.24148 .50698 Mdot
.24148 .50909 Mdot
.24699 .50718 Mdot
.24953 .51166 Mdot
.24953 .57236 Mdot
.25758 .50663 Mdot
.25758 .51224 Mdot
.26562 .51472 Mdot
.27349 .51772 Mdot
.28154 .56149 Mdot
.28958 .52071 Mdot
.30541 .52434 Mdot
.31346 .52564 Mdot
.32168 .52686 Mdot
.32973 .52662 Mdot
.33733 .52232 Mdot
.35351 .03018 Mdot
.40126 .56435 Mdot
.4254 .53553 Mdot
.43318 .53712 Mdot
.44123 .53844 Mdot
.44928 .53418 Mdot
.45732 .53751 Mdot
.45732 .57526 Mdot
.4812 .54654 Mdot
.4812 .56873 Mdot
.48951 .54867 Mdot
.49703 .54882 Mdot
.50507 .55025 Mdot
.50507 .56731 Mdot
.51329 .54991 Mdot
.51329 .56483 Mdot
.52921 .55023 Mdot
.52921 .56096 Mdot
.53717 .54707 Mdot
.54513 .55064 Mdot
.54513 .56221 Mdot
.55317 .56107 Mdot
.56104 .54974 Mdot
.56104 .56117 Mdot
.569 .55159 Mdot
.57705 .56111 Mdot
.58509 .54074 Mdot
.59288 .56102 Mdot
.60897 .56026 Mdot
.64089 .56215 Mdot
.66485 .5613 Mdot
.68873 .5615 Mdot
.69695 .56169 Mdot
.70491 .55287 Mdot
.87273 .54817 Mdot
.88856 .54754 Mdot
.89652 .54813 Mdot
.90457 .54658 Mdot
.91244 .54644 Mdot
.92844 .54556 Mdot
.94427 .54573 Mdot
.95232 .54372 Mdot
.97619 .53338 Mdot
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell["\<\
Although yields are ploted in percent, yield errors are ploted in basis \
points:\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["MaturityPlot[YieldError[yc]];", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0237221 0.0319428 0.509249 0.000976648 [
[(0)] .02372 0 0 2 0 Minner Mrotsboxa
[(5)] .18344 0 0 2 0 Minner Mrotsboxa
[(10)] .34315 0 0 2 0 Minner Mrotsboxa
[(15)] .50286 0 0 2 0 Minner Mrotsboxa
[(20)] .66258 0 0 2 0 Minner Mrotsboxa
[(25)] .82229 0 0 2 0 Minner Mrotsboxa
[(30)] .98201 0 0 2 0 Minner Mrotsboxa
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotsboxa
[(-500)] -0.0125 .02092 1 0 0 Minner Mrotsboxa
[(-400)] -0.0125 .11859 1 0 0 Minner Mrotsboxa
[(-300)] -0.0125 .21625 1 0 0 Minner Mrotsboxa
[(-200)] -0.0125 .31392 1 0 0 Minner Mrotsboxa
[(-100)] -0.0125 .41158 1 0 0 Minner Mrotsboxa
[(0)] -0.0125 .50925 1 0 0 Minner Mrotsboxa
[(100)] -0.0125 .60691 1 0 0 Minner Mrotsboxa
[(YieldError)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotsboxa
[(7/19/1988)] .5 .61803 0 -2 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
.002 w
.02372 0 m
.02372 .00625 L
s
P
[(0)] .02372 0 0 2 0 Minner Mrotshowa
p
.002 w
.18344 0 m
.18344 .00625 L
s
P
[(5)] .18344 0 0 2 0 Minner Mrotshowa
p
.002 w
.34315 0 m
.34315 .00625 L
s
P
[(10)] .34315 0 0 2 0 Minner Mrotshowa
p
.002 w
.50286 0 m
.50286 .00625 L
s
P
[(15)] .50286 0 0 2 0 Minner Mrotshowa
p
.002 w
.66258 0 m
.66258 .00625 L
s
P
[(20)] .66258 0 0 2 0 Minner Mrotshowa
p
.002 w
.82229 0 m
.82229 .00625 L
s
P
[(25)] .82229 0 0 2 0 Minner Mrotshowa
p
.002 w
.98201 0 m
.98201 .00625 L
s
P
[(30)] .98201 0 0 2 0 Minner Mrotshowa
p
.001 w
.05566 0 m
.05566 .00375 L
s
P
p
.001 w
.08761 0 m
.08761 .00375 L
s
P
p
.001 w
.11955 0 m
.11955 .00375 L
s
P
p
.001 w
.15149 0 m
.15149 .00375 L
s
P
p
.001 w
.21538 0 m
.21538 .00375 L
s
P
p
.001 w
.24732 0 m
.24732 .00375 L
s
P
p
.001 w
.27926 0 m
.27926 .00375 L
s
P
p
.001 w
.31121 0 m
.31121 .00375 L
s
P
p
.001 w
.37509 0 m
.37509 .00375 L
s
P
p
.001 w
.40704 0 m
.40704 .00375 L
s
P
p
.001 w
.43898 0 m
.43898 .00375 L
s
P
p
.001 w
.47092 0 m
.47092 .00375 L
s
P
p
.001 w
.53481 0 m
.53481 .00375 L
s
P
p
.001 w
.56675 0 m
.56675 .00375 L
s
P
p
.001 w
.59869 0 m
.59869 .00375 L
s
P
p
.001 w
.63064 0 m
.63064 .00375 L
s
P
p
.001 w
.69452 0 m
.69452 .00375 L
s
P
p
.001 w
.72646 0 m
.72646 .00375 L
s
P
p
.001 w
.75841 0 m
.75841 .00375 L
s
P
p
.001 w
.79035 0 m
.79035 .00375 L
s
P
p
.001 w
.85423 0 m
.85423 .00375 L
s
P
p
.001 w
.88618 0 m
.88618 .00375 L
s
P
p
.001 w
.91812 0 m
.91812 .00375 L
s
P
p
.001 w
.95006 0 m
.95006 .00375 L
s
P
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotshowa
p
.002 w
0 0 m
1 0 L
s
P
p
.002 w
0 .02092 m
.00625 .02092 L
s
P
[(-500)] -0.0125 .02092 1 0 0 Minner Mrotshowa
p
.002 w
0 .11859 m
.00625 .11859 L
s
P
[(-400)] -0.0125 .11859 1 0 0 Minner Mrotshowa
p
.002 w
0 .21625 m
.00625 .21625 L
s
P
[(-300)] -0.0125 .21625 1 0 0 Minner Mrotshowa
p
.002 w
0 .31392 m
.00625 .31392 L
s
P
[(-200)] -0.0125 .31392 1 0 0 Minner Mrotshowa
p
.002 w
0 .41158 m
.00625 .41158 L
s
P
[(-100)] -0.0125 .41158 1 0 0 Minner Mrotshowa
p
.002 w
0 .50925 m
.00625 .50925 L
s
P
[(0)] -0.0125 .50925 1 0 0 Minner Mrotshowa
p
.002 w
0 .60691 m
.00625 .60691 L
s
P
[(100)] -0.0125 .60691 1 0 0 Minner Mrotshowa
p
.001 w
0 .04046 m
.00375 .04046 L
s
P
p
.001 w
0 .05999 m
.00375 .05999 L
s
P
p
.001 w
0 .07952 m
.00375 .07952 L
s
P
p
.001 w
0 .09906 m
.00375 .09906 L
s
P
p
.001 w
0 .13812 m
.00375 .13812 L
s
P
p
.001 w
0 .15766 m
.00375 .15766 L
s
P
p
.001 w
0 .17719 m
.00375 .17719 L
s
P
p
.001 w
0 .19672 m
.00375 .19672 L
s
P
p
.001 w
0 .23579 m
.00375 .23579 L
s
P
p
.001 w
0 .25532 m
.00375 .25532 L
s
P
p
.001 w
0 .27485 m
.00375 .27485 L
s
P
p
.001 w
0 .29439 m
.00375 .29439 L
s
P
p
.001 w
0 .33345 m
.00375 .33345 L
s
P
p
.001 w
0 .35299 m
.00375 .35299 L
s
P
p
.001 w
0 .37252 m
.00375 .37252 L
s
P
p
.001 w
0 .39205 m
.00375 .39205 L
s
P
p
.001 w
0 .43112 m
.00375 .43112 L
s
P
p
.001 w
0 .45065 m
.00375 .45065 L
s
P
p
.001 w
0 .47018 m
.00375 .47018 L
s
P
p
.001 w
0 .48972 m
.00375 .48972 L
s
P
p
.001 w
0 .52878 m
.00375 .52878 L
s
P
p
.001 w
0 .54831 m
.00375 .54831 L
s
P
p
.001 w
0 .56785 m
.00375 .56785 L
s
P
p
.001 w
0 .58738 m
.00375 .58738 L
s
P
p
.001 w
0 .00139 m
.00375 .00139 L
s
P
[(YieldError)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotshowa
p
.002 w
0 0 m
0 .61803 L
s
P
P
p
[(7/19/1988)] .5 .61803 0 -2 Mshowa
P
p
P
p
.008 w
.02381 .40813 Mdot
.02442 .37435 Mdot
.02503 .43791 Mdot
.02565 .40715 Mdot
.02626 .389 Mdot
.02687 .40928 Mdot
.0274 .47795 Mdot
.02748 .44153 Mdot
.02809 .44011 Mdot
.02871 .43763 Mdot
.02932 .45417 Mdot
.02993 .45667 Mdot
.03002 .48971 Mdot
.03002 .49536 Mdot
.03054 .45217 Mdot
.03116 .45264 Mdot
.03151 .43782 Mdot
.03238 .46465 Mdot
.03273 .50667 Mdot
.03299 .46912 Mdot
.0336 .47057 Mdot
.03404 .50199 Mdot
.03404 .51083 Mdot
.03404 .51321 Mdot
.03422 .47407 Mdot
.03492 .47848 Mdot
.03535 .50706 Mdot
.03544 .47588 Mdot
.03605 .47422 Mdot
.03667 .47769 Mdot
.03728 .48222 Mdot
.03789 .47635 Mdot
.03833 .50547 Mdot
.03833 .50667 Mdot
.0385 .48296 Mdot
.03911 .48543 Mdot
.03955 .48099 Mdot
.04078 .51205 Mdot
.04209 .51746 Mdot
.04209 .51828 Mdot
.04217 .49264 Mdot
.04322 .51994 Mdot
.04462 .49325 Mdot
.04594 .51556 Mdot
.04594 .52223 Mdot
.04707 .49181 Mdot
.04742 .47413 Mdot
.04865 .52072 Mdot
.04952 .49161 Mdot
.04987 .516 Mdot
.04987 .52127 Mdot
.04987 .52291 Mdot
.05127 .51857 Mdot
.05197 .48826 Mdot
.05389 .5141 Mdot
.05389 .51621 Mdot
.05442 .49293 Mdot
.05538 .48303 Mdot
.05661 .51758 Mdot
.05792 .50781 Mdot
.05792 .51741 Mdot
.05932 .51374 Mdot
.06211 .51191 Mdot
.06211 .51282 Mdot
.06334 .51477 Mdot
.06465 .51316 Mdot
.06596 .50502 Mdot
.06596 .5104 Mdot
.06596 .51482 Mdot
.06727 .51243 Mdot
.07016 .50976 Mdot
.07016 .51102 Mdot
.07138 .50897 Mdot
.0727 .51152 Mdot
.07401 .42865 Mdot
.07401 .5099 Mdot
.07401 .51107 Mdot
.07515 .51332 Mdot
.07803 .5063 Mdot
.07803 .50786 Mdot
.07926 .51368 Mdot
.08048 .51237 Mdot
.08179 .49322 Mdot
.08179 .51163 Mdot
.08179 .5153 Mdot
.08319 .51134 Mdot
.08599 .50773 Mdot
.08599 .50839 Mdot
.08721 .51021 Mdot
.08984 .51001 Mdot
.08984 .51173 Mdot
.08984 .51273 Mdot
.09395 .50928 Mdot
.09517 .51388 Mdot
.09788 .50905 Mdot
.09788 .51034 Mdot
.09788 .51183 Mdot
.10191 .5079 Mdot
.10322 .5114 Mdot
.10593 .50713 Mdot
.10593 .51034 Mdot
.10986 .50973 Mdot
.11109 .51349 Mdot
.11371 .4977 Mdot
.11371 .51017 Mdot
.11782 .50907 Mdot
.11905 .51321 Mdot
.12176 .5015 Mdot
.12176 .51077 Mdot
.12578 .51001 Mdot
.12709 .51384 Mdot
.1298 .50804 Mdot
.1298 .51051 Mdot
.13383 .50741 Mdot
.13514 .51247 Mdot
.13811 .50553 Mdot
.13811 .50892 Mdot
.13811 .56819 Mdot
.14179 .51082 Mdot
.1431 .51582 Mdot
.14572 .5129 Mdot
.14572 .51523 Mdot
.14974 .51027 Mdot
.15106 .5133 Mdot
.15394 .25359 Mdot
.15394 .47984 Mdot
.15394 .51065 Mdot
.1591 .51418 Mdot
.1619 .51109 Mdot
.1619 .51344 Mdot
.16715 .51099 Mdot
.16995 .19465 Mdot
.16995 .49897 Mdot
.16995 .50016 Mdot
.16995 .50992 Mdot
.16995 .51294 Mdot
.17502 .512 Mdot
.17782 .51311 Mdot
.17782 .51377 Mdot
.18298 .51009 Mdot
.18578 .49979 Mdot
.18578 .50827 Mdot
.18578 .51456 Mdot
.19102 .5112 Mdot
.19373 .18279 Mdot
.19373 .50016 Mdot
.19373 .51403 Mdot
.19933 .50912 Mdot
.20178 .5068 Mdot
.20694 .51054 Mdot
.20965 .51341 Mdot
.20965 .58145 Mdot
.2149 .50973 Mdot
.21761 .4948 Mdot
.21761 .51269 Mdot
.22312 .51015 Mdot
.22565 .50915 Mdot
.22565 .51275 Mdot
.23116 .50799 Mdot
.2337 .02504 Mdot
.2337 .51083 Mdot
.2337 .51216 Mdot
.2337 .60332 Mdot
.23904 .50921 Mdot
.24148 .50791 Mdot
.24148 .5097 Mdot
.24148 .51242 Mdot
.24699 .50565 Mdot
.24953 .5113 Mdot
.24953 .56859 Mdot
.25758 .5052 Mdot
.25758 .50898 Mdot
.26562 .50836 Mdot
.27349 .50855 Mdot
.28154 .54968 Mdot
.28958 .50766 Mdot
.30541 .51005 Mdot
.31346 .50972 Mdot
.32168 .50981 Mdot
.32973 .50723 Mdot
.33733 .50402 Mdot
.35351 .01472 Mdot
.40126 .53657 Mdot
.4254 .50836 Mdot
.43318 .51051 Mdot
.44123 .51041 Mdot
.44928 .50702 Mdot
.45732 .50756 Mdot
.45732 .5375 Mdot
.4812 .51054 Mdot
.4812 .52741 Mdot
.48951 .50979 Mdot
.49703 .50968 Mdot
.50507 .51024 Mdot
.50507 .52329 Mdot
.51329 .51019 Mdot
.51329 .52086 Mdot
.52921 .51017 Mdot
.52921 .51692 Mdot
.53717 .50766 Mdot
.54513 .50941 Mdot
.54513 .51936 Mdot
.55317 .51914 Mdot
.56104 .5086 Mdot
.56104 .51775 Mdot
.569 .5088 Mdot
.57705 .51998 Mdot
.58509 .49657 Mdot
.59288 .52047 Mdot
.60897 .51948 Mdot
.64089 .51855 Mdot
.66485 .51853 Mdot
.68873 .51973 Mdot
.69695 .51918 Mdot
.70491 .51059 Mdot
.87273 .50921 Mdot
.88856 .50911 Mdot
.89652 .51023 Mdot
.90457 .50869 Mdot
.91244 .50917 Mdot
.92844 .50885 Mdot
.94427 .50932 Mdot
.95232 .50718 Mdot
.97619 .49751 Mdot
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell[TextData[{
  "As we see, the outliers compress the image of the rest of the data. If one \
is not interested in pricing the securities with special features, one can \
omit them from the data file or create another ",
  StyleBox["MD",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object that excludes them (see below)."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
So far we have restricted ourselves to splining the discount function \
directly, largely for speed considerations. With a large number of \
securities, the nonlinear routines run quite slowly. So to speed things up, \
we will spline only Treasury bills.\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
pos = Flatten @ Position[CouponRate[md], 0|0.];
mdbills = MDSubset[md, pos];
cpbills = ConstructPayments[mdbills];\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
For comparison purposes, we will spline the discount function first with no \
penalty:\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
dfbills1 = DiscountFunction[cpbills,
\tFunctionalForm -> Delta, MaximumMaturity -> 1,
\tLambdaValue -> 0]\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
DF[{1988, 7, 19}, InterpolatingFunction[{0, 366}, 
 
   {{0, 0, {1, 0}, {0}}, 
 
    {15, 0, {0.997443116037325, -0.0001706548150588634, 
 
      -(1.305894759161097*10^-8), 3.47258376524226*10^-8}, 
 
     {0, 0, 15}}, {36, 15, 
 
     {0.99381474916186, -0.0001876928959098425, 
 
      -(7.101676613183415*10^-7), 
 
      -(2.899991139317727*10^-8)}, {0, 0, 21}}, 
 
    {57, 36, {0.9896159682493, -0.0001967375710195951, 
 
      1.525893903583764*10^-7, 3.504179026032733*10^-8}, 
 
     {0, 0, 21}}, {78, 57, 
 
     {0.985662463964733, -0.0001899689910298818, 
 
      -(8.12801067127425*10^-8), -(2.308921648899873*10^-8)}
 
      , {0, 0, 21}}, {106, 78, 
 
     {0.98001130442612, -0.0002096910947938444, 
 
      -(2.808560148148007*10^-7), 5.094600681548027*10^-9}, 
 
     {0, 0, 28}}, {120, 106, 
 
     {0.977100552336291, -0.0002024155003497521, 
 
      3.925259435343953*10^-7, 1.895475497383095*10^-8}, 
 
     {0, 0, 14}}, {141, 120, 
 
     {0.972977661445994, -0.0001979691549765517, 
 
      -(7.814368301809625*10^-8), 
 
      -(1.752467133777899*10^-8)}, {0, 0, 21}}, 
 
    {162, 141, {0.968673153123202, -0.0002096220530424554, 
 
      -(2.212126782299262*10^-7), 5.355931020967973*10^-9}, 
 
     {0, 0, 21}}, {211, 162, 
 
     {0.958310460473163, -0.0002098782961129182, 
 
      3.275974198504526*10^-8, 1.443855803830567*10^-9}, 
 
     {0, 0, 49}}, {295, 211, 
 
     {0.941249313703277, -0.0001982648069425285, 
 
      5.766765684716843*10^-8, -(2.728632114605146*10^-10)}\\
 
      , {0, 0, 84}}, {366, 295, 
 
     {0.925833696452673, -0.0002573015427383973, 
 
      -(5.659179297405857*10^-7), 
 
      -(4.230035752290086*10^-9)}, {0, 0, 71}}}], 0, 13, 
 
  Delta, {-0.01221160373249575, -0.01882791003220063, 
 
   -0.00874249742018662, -0.003362248905261822, 
 
   -0.00833608226109561, -0.01433422419466979, 
 
   -0.0006530158601236736, 0.006532761733510029, 
 
   -0.0003468964547721498, 0.0141092721391658, 
 
   0.00352508381402572, -0.001064679468356644, 
 
   0.01720317506996594, -0.004342053321295225, 
 
   0.001553603526701863, 0.006231305343860072, 
 
   -0.003330442612025308, 0.004980390907164179, 
 
   0.003244766370883667, 0.0006405434175320579, 
 
   0.007533855400566836, -0.002085251822435907, 
 
   0.01868468767976594, -0.001496582623317977, 
 
   -0.004109364808272176, -0.004146047316339719, 
 
   0.002297090894046505, 0.006304004579988032, 
 
   -0.00853137032771656, 0.003671687263391732, 
 
   -0.0004589609078635704}, 
 
  {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 
 
   17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 
 
   31}, 0.0001168377596829914, 0.006222305168032177, 
 
  {-0.939240419753879, -1.499674556741843, 
 
   -0.7148664955171498, -0.2590186656718787, 
 
   -0.6685227286734868, -1.203913336333714, 
 
   -0.05757946626607451, 0.4486087547787333, 
 
   -0.02631605146669823, 1.134083818861967, 
 
   0.2726717473113407, -0.0865167234234559, 
 
   1.343138021111412, -0.3504527567673203, 
 
   0.1214177336985955, 0.4913636701499445, 
 
   -0.2595322905060946, 0.4058735663170852, 
 
   0.2510277000302896, 0.05288749779261623, 
 
   0.6006920191768969, -0.1681979716092984, 
 
   1.553913380388874, -0.12027900177985, 
 
   -0.3243346262857676, -0.3039480608939633, 
 
   0.1763940589697827, 0.4736803634534338, 
 
   -0.6380919532968955, 0.2453320810782049, 
 
   -0.03003376717456423}, 
 
  {0, 15, 36, 57, 78, 106, 120, 141, 162, 211, 295, 366}, 
 
  {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
 
   {0, 0.0003180309268343071, -0.0002118808911886135, 
 
    0.0001045990456436456, -0.00005105270825140931, 
 
    0.00003197614171716587, -0.00001327140890289068, 
 
    6.103265888276118*10^-6, -(2.743827357726276*10^-6), 
 
    1.736012767732012*10^-6, -(1.260705075256942*10^-6), 
 
    1.089776842768214*10^-6, -(7.49086616248132*10^-7), 
 
    6.028098493344497*10^-7}, 
 
   {0, -0.0002118808911886135, 0.0003290092246087627, 
 
    -0.0001896325806005411, 0.0000962640039395103, 
 
    -0.00006077357320888121, 0.00002530180408092213, 
 
    -0.0000116447378208644, 5.235796839930093*10^-6, 
 
    -(3.312736658244803*10^-6), 2.40574922648379*10^-6, 
 
    -(2.079577979863584*10^-6), 1.429453394856638*10^-6, 
 
    -(1.150319245886172*10^-6)}, 
 
   {0, 0.0001045990456436456, -0.0001896325806005411, 
 
    0.0002239494363664545, -0.0001284735056916896, 
 
    0.0000829624040935561, -0.00003483983498162834, 
 
    0.00001606857577647613, -(7.227574560108554*10^-6), 
 
    4.573185574004106*10^-6, -(3.321162800160081*10^-6), 
 
    2.870894160467806*10^-6, -(1.97338979834223*10^-6), 
 
    1.588040213211176*10^-6}, 
 
   {0, -0.00005105270825140931, 0.0000962640039395103, 
 
    -0.0001284735056916896, 0.0001712219392271303, 
 
    -0.0001222392173215095, 0.00005319016817354129, 
 
    -0.00002474119554373858, 0.00001114498798364892, 
 
    -(7.053319802356553*10^-6), 5.122668347480582*10^-6, 
 
    -(4.42824692077462*10^-6), 3.043905189198374*10^-6, 
 
    -(2.449518113892703*10^-6)}, 
 
   {0, 0.00003197614171716587, -0.00006077357320888121, 
 
    0.0000829624040935561, -0.0001222392173215095, 
 
    0.0001960291226234681, -0.0001019897495236949, 
 
    0.00004926391540026791, -0.00002233426106516906, 
 
    0.00001414696357577439, -0.00001027781170029682, 
 
    8.88532089009351*10^-6, -(6.107843052174166*10^-6), 
 
    4.915201749258546*10^-6}, 
 
   {0, -0.00001327140890289068, 0.00002530180408092213, 
 
    -0.00003483983498162834, 0.00005319016817354129, 
 
    -0.0001019897495236949, 0.0001813760624498361, 
 
    -0.0001002590419489425, 0.00004640834017374852, 
 
    -0.00002947775330114319, 0.00002143686632583594, 
 
    -0.00001853750052080685, 0.00001274427547286395, 
 
    -0.00001025607504726781}, 
 
   {0, 6.103265888276118*10^-6, -0.0000116447378208644, 
 
    0.00001606857577647613, -0.00002474119554373858, 
 
    0.00004926391540026791, -0.0001002590419489425, 
 
    0.0001449005530114352, -0.00007314221581547951, 
 
    0.00004696984745835456, -0.00003428959408787582, 
 
    0.00002968308739106735, -0.00002041570498898573, 
 
    0.00001643157629008606}, 
 
   {0, -(2.743827357726276*10^-6), 5.235796839930093*10^-6, 
 
    -(7.227574560108554*10^-6), 0.00001114498798364892, 
 
    -0.00002233426106516906, 0.00004640834017374852, 
 
    -0.00007314221581547951, 0.0001201214728309251, 
 
    -0.0000845407207504059, 0.00006362013387751019, 
 
    -0.00005552113735330677, 0.00003831573641321171, 
 
    -0.00003086486969616626}, 
 
   {0, 1.736012767732012*10^-6, -(3.312736658244803*10^-6), 
 
    4.573185574004106*10^-6, -(7.053319802356553*10^-6), 
 
    0.00001414696357577439, -0.00002947775330114319, 
 
    0.00004696984745835456, -0.0000845407207504059, 
 
    0.0001390213964738526, -0.0001240198612735991, 
 
    0.0001126659993592115, -0.00007901823878696293, 
 
    0.00006391139415058396}, 
 
   {0, -(1.260705075256942*10^-6), 2.40574922648379*10^-6, 
 
    -(3.321162800160081*10^-6), 5.122668347480582*10^-6, 
 
    -0.00001027781170029682, 0.00002143686632583594, 
 
    -0.00003428959408787582, 0.00006362013387751019, 
 
    -0.0001240198612735991, 0.0003653890788376187, 
 
    -0.000382402196727178, 0.0002820883538036736, 
 
    -0.0002309564567849527}, 
 
   {0, 1.089776842768214*10^-6, -(2.079577979863584*10^-6), 
 
    2.870894160467806*10^-6, -(4.42824692077462*10^-6), 
 
    8.88532089009351*10^-6, -0.00001853750052080685, 
 
    0.00002968308739106735, -0.00005552113735330677, 
 
    0.0001126659993592115, -0.000382402196727178, 
 
    0.0006688245258112731, -0.0005666614017636197, 
 
    0.0004784168186796633}, 
 
   {0, -(7.49086616248132*10^-7), 1.429453394856638*10^-6, 
 
    -(1.97338979834223*10^-6), 3.043905189198374*10^-6, 
 
    -(6.107843052174166*10^-6), 0.00001274427547286395, 
 
    -0.00002041570498898573, 0.00003831573641321171, 
 
    -0.00007901823878696293, 0.0002820883538036736, 
 
    -0.0005666614017636197, 0.0007648665398105487, 
 
    -0.0007060770392995491}, 
 
   {0, 6.028098493344497*10^-7, -(1.150319245886172*10^-6), 
 
    1.588040213211176*10^-6, -(2.449518113892703*10^-6), 
 
    4.915201749258546*10^-6, -0.00001025607504726781, 
 
    0.00001643157629008606, -0.00003086486969616626, 
 
    0.00006391139415058396, -0.0002309564567849527, 
 
    0.0004784168186796633, -0.0007060770392995491, 
 
    0.001068813951071736}}, 
 
  {1, 0.999187751334203, 0.997048484402402, 
 
   0.994008666534544, 0.989485363602384, 0.985330169101648, 
 
   0.981007921942067, 0.976563776035851, 0.973043247227818, 
 
   0.966753977805628, 0.955719866447863, 0.942039383871624, 
 
   0.931923166297482, 0.925833696452673}, True, 0, True, 2, 
 
  Automatic, 3, 0.002607818602118641, Hold[cpbills]]\
\>", "\<\
-DF[{1988, 7, 19}]-\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["PlotGCVFunction[];", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.04329 0.0147151 39.2844 [
[(0)] .02381 0 0 2 Msboxa
[(5)] .24026 0 0 2 Msboxa
[(10)] .45671 0 0 2 Msboxa
[(15)] .67316 0 0 2 Msboxa
[(20)] .88961 0 0 2 Msboxa
[(0)] -0.0125 .01472 1 0 Msboxa
[(0.002)] -0.0125 .09328 1 0 Msboxa
[(0.004)] -0.0125 .17185 1 0 Msboxa
[(0.006)] -0.0125 .25042 1 0 Msboxa
[(0.008)] -0.0125 .32899 1 0 Msboxa
[(0.01)] -0.0125 .40756 1 0 Msboxa
[(0.012)] -0.0125 .48613 1 0 Msboxa
[(0.014)] -0.0125 .5647 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
.002 w
.02381 0 m
.02381 .00625 L
s
P
[(0)] .02381 0 0 2 Mshowa
p
.002 w
.24026 0 m
.24026 .00625 L
s
P
[(5)] .24026 0 0 2 Mshowa
p
.002 w
.45671 0 m
.45671 .00625 L
s
P
[(10)] .45671 0 0 2 Mshowa
p
.002 w
.67316 0 m
.67316 .00625 L
s
P
[(15)] .67316 0 0 2 Mshowa
p
.002 w
.88961 0 m
.88961 .00625 L
s
P
[(20)] .88961 0 0 2 Mshowa
p
.001 w
.0671 0 m
.0671 .00375 L
s
P
p
.001 w
.11039 0 m
.11039 .00375 L
s
P
p
.001 w
.15368 0 m
.15368 .00375 L
s
P
p
.001 w
.19697 0 m
.19697 .00375 L
s
P
p
.001 w
.28355 0 m
.28355 .00375 L
s
P
p
.001 w
.32684 0 m
.32684 .00375 L
s
P
p
.001 w
.37013 0 m
.37013 .00375 L
s
P
p
.001 w
.41342 0 m
.41342 .00375 L
s
P
p
.001 w
.5 0 m
.5 .00375 L
s
P
p
.001 w
.54329 0 m
.54329 .00375 L
s
P
p
.001 w
.58658 0 m
.58658 .00375 L
s
P
p
.001 w
.62987 0 m
.62987 .00375 L
s
P
p
.001 w
.71645 0 m
.71645 .00375 L
s
P
p
.001 w
.75974 0 m
.75974 .00375 L
s
P
p
.001 w
.80303 0 m
.80303 .00375 L
s
P
p
.001 w
.84632 0 m
.84632 .00375 L
s
P
p
.001 w
.9329 0 m
.9329 .00375 L
s
P
p
.001 w
.97619 0 m
.97619 .00375 L
s
P
p
.002 w
0 0 m
1 0 L
s
P
p
.002 w
0 .01472 m
.00625 .01472 L
s
P
[(0)] -0.0125 .01472 1 0 Mshowa
p
.002 w
0 .09328 m
.00625 .09328 L
s
P
[(0.002)] -0.0125 .09328 1 0 Mshowa
p
.002 w
0 .17185 m
.00625 .17185 L
s
P
[(0.004)] -0.0125 .17185 1 0 Mshowa
p
.002 w
0 .25042 m
.00625 .25042 L
s
P
[(0.006)] -0.0125 .25042 1 0 Mshowa
p
.002 w
0 .32899 m
.00625 .32899 L
s
P
[(0.008)] -0.0125 .32899 1 0 Mshowa
p
.002 w
0 .40756 m
.00625 .40756 L
s
P
[(0.01)] -0.0125 .40756 1 0 Mshowa
p
.002 w
0 .48613 m
.00625 .48613 L
s
P
[(0.012)] -0.0125 .48613 1 0 Mshowa
p
.002 w
0 .5647 m
.00625 .5647 L
s
P
[(0.014)] -0.0125 .5647 1 0 Mshowa
p
.001 w
0 .03043 m
.00375 .03043 L
s
P
p
.001 w
0 .04614 m
.00375 .04614 L
s
P
p
.001 w
0 .06186 m
.00375 .06186 L
s
P
p
.001 w
0 .07757 m
.00375 .07757 L
s
P
p
.001 w
0 .109 m
.00375 .109 L
s
P
p
.001 w
0 .12471 m
.00375 .12471 L
s
P
p
.001 w
0 .14043 m
.00375 .14043 L
s
P
p
.001 w
0 .15614 m
.00375 .15614 L
s
P
p
.001 w
0 .18757 m
.00375 .18757 L
s
P
p
.001 w
0 .20328 m
.00375 .20328 L
s
P
p
.001 w
0 .21899 m
.00375 .21899 L
s
P
p
.001 w
0 .23471 m
.00375 .23471 L
s
P
p
.001 w
0 .26614 m
.00375 .26614 L
s
P
p
.001 w
0 .28185 m
.00375 .28185 L
s
P
p
.001 w
0 .29756 m
.00375 .29756 L
s
P
p
.001 w
0 .31328 m
.00375 .31328 L
s
P
p
.001 w
0 .3447 m
.00375 .3447 L
s
P
p
.001 w
0 .36042 m
.00375 .36042 L
s
P
p
.001 w
0 .37613 m
.00375 .37613 L
s
P
p
.001 w
0 .39185 m
.00375 .39185 L
s
P
p
.001 w
0 .42327 m
.00375 .42327 L
s
P
p
.001 w
0 .43899 m
.00375 .43899 L
s
P
p
.001 w
0 .4547 m
.00375 .4547 L
s
P
p
.001 w
0 .47041 m
.00375 .47041 L
s
P
p
.001 w
0 .50184 m
.00375 .50184 L
s
P
p
.001 w
0 .51756 m
.00375 .51756 L
s
P
p
.001 w
0 .53327 m
.00375 .53327 L
s
P
p
.001 w
0 .54898 m
.00375 .54898 L
s
P
p
.001 w
0 .58041 m
.00375 .58041 L
s
P
p
.001 w
0 .59612 m
.00375 .59612 L
s
P
p
.001 w
0 .61184 m
.00375 .61184 L
s
P
p
.002 w
0 0 m
0 .61803 L
s
P
P
p
P
p
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.004 w
.0671 .11716 m
.11039 .11713 L
.15368 .11686 L
.19697 .11423 L
.24026 .09618 L
.28355 .0578 L
.32684 .03373 L
.37013 .02735 L
.41342 .02928 L
.45671 .05094 L
.5 .14956 L
.54329 .25041 L
.58658 .2687 L
.62987 .27067 L
.67316 .27087 L
.71645 .27089 L
.75974 .27089 L
.80303 .2709 L
.84632 .27189 L
.88961 .37045 L
s
.8907 .61803 m
.88961 .37045 L
s
s
s
1 0 0 r
p
.015 w
.0671 .11716 Mdot
.11039 .11713 Mdot
.15368 .11686 Mdot
.19697 .11423 Mdot
.24026 .09618 Mdot
.28355 .0578 Mdot
.32684 .03373 Mdot
.37013 .02735 Mdot
.41342 .02928 Mdot
.45671 .05094 Mdot
.5 .14956 Mdot
.54329 .25041 Mdot
.58658 .2687 Mdot
.62987 .27067 Mdot
.67316 .27087 Mdot
.71645 .27089 Mdot
.75974 .27089 Mdot
.80303 .2709 Mdot
.84632 .27189 Mdot
.88961 .37045 Mdot
.9329 10.1794 Mdot
.97619 945.407 Mdot
P
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell[CellGroupData[{Cell["\<\
dfbills2 = DiscountFunction[cpbills,
\tFunctionalForm -> Delta, MaximumMaturity -> 1,
\tLambda0 -> {10^8, 10^9}];\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
DF[{1988, 7, 19}, InterpolatingFunction[{0, 366}, 
 
   {{0, 0, {1, 0}, {0}}, 
 
    {15, 0, {0.997406364614296, -0.0001744743556198365, 
 
      -(1.043553270833893*10^-7), 
 
      -(3.571928452205699*10^-9)}, {0, 0, 15}}, 
 
    {36, 15, {0.993656682844813, -0.0001834034936410711, 
 
      -(2.30819947798087*10^-7), -(1.735373665045474*10^-9)}
 
      , {0, 0, 21}}, {57, 36, 
 
     {0.989707102192498, -0.0001918063015013196, 
 
      -(1.776682068325827*10^-7), 2.133204474558*10^-9}, 
 
     {0, 0, 21}}, {78, 57, 
 
     {0.985622420694276, -0.0001971230361446225, 
 
      -(1.244949224825304*10^-7), 1.994283424864227*10^-10}\\
 
      , {0, 0, 21}}, {106, 78, 
 
     {0.980028264093194, -0.0002017592552911521, 
 
      -(7.028386105888782*10^-8), 8.93269039836609*10^-10}, 
 
     {0, 0, 28}}, {120, 106, 
 
     {0.977198011991871, -0.0002023302699670736, 
 
      -(1.210039906287871*10^-8), 1.184711745718826*10^-9}, 
 
     {0, 0, 14}}, {141, 120, 
 
     {0.972941052742049, -0.0002035706926276082, 
 
      -(4.087368561840222*10^-8), 
 
      -(1.079982166526701*10^-9)}, {0, 0, 21}}, 
 
    {162, 141, {0.968630924324928, -0.0002072566262286864, 
 
      -(9.5829328074823*10^-8), -(7.684765942502865*10^-10)}
 
      , {0, 0, 21}}, {211, 162, 
 
     {0.958369105330261, -0.0002082749803780683, 
 
      2.346728702278399*10^-8, 1.381985954858278*10^-9}, 
 
     {0, 0, 49}}, {295, 211, 
 
     {0.941200386649381, -0.0002042780698611468, 
 
      1.326645768674508*10^-9, -(5.34868768109995*10^-10)}, 
 
     {0, 0, 84}}, {366, 295, 
 
     {0.926498704055163, -0.0002095459508879849, 
 
      -(3.492956136267757*10^-8), 6.107584077322904*10^-11}\\
 
      , {0, 0, 71}}}], 1.6481191051471*10^8, 
 
  5.687335535601268, Delta, 
 
  {-0.00853646142959974, -0.01004942015165966, 
 
   -0.01113185039295672, -0.0001054144881891261, 
 
   -0.01481039574493081, -0.00864978965093144, 
 
   0.0002148149110894337, 0.00948285033453544, 
 
   0.00832898866492826, 0.02869933434762117, 
 
   0.01933171551874579, -0.002850393914215488, 
 
   0.00808978075019695, -0.01270158111799446, 
 
   0.005557930572436476, 0.0132029206893236, 
 
   -0.005026409319370373, -0.002113731821523856, 
 
   -0.006501199187113116, -0.000930391208072479, 
 
   0.01119472579509306, 0.003831980606051388, 
 
   0.02290756750721812, 0.00108844369816552, 
 
   -0.003117812922184271, -0.01001053302606181, 
 
   -0.006515316594487786, -0.00007298173800052155, 
 
   -0.003638664938051761, 0.02212427971852549, 
 
   -0.01268037227696084}, 
 
  {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 
 
   17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 
 
   31}, 0.000155791091041474, 0.00882251783987855, 
 
  {-0.6718190362186734, -0.7653240469073845, 
 
   -0.846900792404221, -0.00801556293895335, 
 
   -1.129552342346634, -0.6553786992064558, 
 
   0.01720878758876827, 0.7552565928471844, 
 
   0.6471967865301711, 2.206886684166462, 
 
   1.476553782900499, -0.2168750962691332, 
 
   0.6154227778609439, -0.966487085419669, 
 
   0.4226451919387724, 1.004042267197452, 
 
   -0.3822007899616286, -0.1608518125436029, 
 
   -0.4952475135077498, -0.0709757162355484, 
 
   0.853404428518131, 0.2915593986278583, 
 
   1.725433310213144, 0.0815168373107352, 
 
   -0.232234727816779, -0.7146857259561879, 
 
   -0.4632390174474095, -0.005171924162359442, 
 
   -0.2559699457811249, 1.554848888242749, 
 
   -0.7800117533282328}, 
 
  {0, 15, 36, 57, 78, 106, 120, 141, 162, 211, 295, 366}, 
 
  {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
 
   {0, 4.940678297443369*10^-7, 1.672651209739775*10^-6, 
 
    2.225629706933137*10^-6, 1.375531461355897*10^-6, 
 
    3.027096997104745*10^-7, -(1.043715889715874*10^-7), 
 
    -(1.715675785468755*10^-7), -(1.060609684552121*10^-7), 
 
    1.207785178759449*10^-8, 3.295093655251232*10^-8, 
 
    -(1.142518358984983*10^-8), 
 
    -(4.146716409260004*10^-10), 5.813793432176927*10^-10}, 
 
   {0, 1.672651209739775*10^-6, 5.686832325363125*10^-6, 
 
    7.589202666640383*10^-6, 4.667133619256898*10^-6, 
 
    1.026839390225181*10^-6, -(3.571370544063646*10^-7), 
 
    -(5.828784287739276*10^-7), -(3.597653387151834*10^-7), 
 
    4.138547270695167*10^-8, 1.116673291591559*10^-7, 
 
    -(3.874019304217107*10^-8), -(1.412267963263684*10^-9), 
 
    1.979616277295404*10^-9}, 
 
   {0, 2.225629706933137*10^-6, 7.589202666640382*10^-6, 
 
    0.00001246217748420019, 9.83572621175165*10^-6, 
 
    3.675627449639243*10^-6, 2.767159835984672*10^-7, 
 
    -(8.39576134268203*10^-7), -(7.63332497354033*10^-7), 
 
    -(1.054947861732856*10^-7), 2.215523710965158*10^-7, 
 
    -(6.155143543861663*10^-8), -(2.907795067079517*10^-9), 
 
    -(9.76260383882464*10^-12)}, 
 
   {0, 1.375531461355898*10^-6, 4.667133619256897*10^-6, 
 
    9.83572621175165*10^-6, 0.00001271917420702479, 
 
    9.24681591224882*10^-6, 3.646537337495486*10^-6, 
 
    3.580667087831263*10^-7, -(8.54683381010811*10^-7), 
 
    -(7.881958600623804*10^-7), 1.70433575143819*10^-7, 
 
    1.23858438217546*10^-8, -(2.746237478501515*10^-9), 
 
    -(1.49812901630267*10^-8)}, 
 
   {0, 3.027096997104746*10^-7, 1.02683939022518*10^-6, 
 
    3.675627449639243*10^-6, 9.24681591224881*10^-6, 
 
    0.00001370106835160716, 9.81863688529454*10^-6, 
 
    4.336611608056943*10^-6, 5.180770285569682*10^-7, 
 
    -(1.984145625410027*10^-6), -(4.622316499627859*10^-7), 
 
    3.115791458331515*10^-7, 8.26917062576228*10^-9, 
 
    -(4.464406871327686*10^-8)}, 
 
   {0, -(1.043715889715872*10^-7), 
 
    -(3.571370544063643*10^-7), 2.767159835984672*10^-7, 
 
    3.646537337495486*10^-6, 9.81863688529454*10^-6, 
 
    0.00001302799538868149, 9.75892393535726*10^-6, 
 
    4.488364802538226*10^-6, -(2.095110410368115*10^-6), 
 
    -(2.03508916090437*10^-6), 7.947781044659653*10^-7, 
 
    4.008322474795886*10^-8, -(4.541272367597477*10^-8)}, 
 
   {0, -(1.715675785468756*10^-7), 
 
    -(5.828784287739276*10^-7), -(8.39576134268204*10^-7), 
 
    3.580667087831266*10^-7, 4.336611608056943*10^-6, 
 
    9.75892393535726*10^-6, 0.00001172989565464466, 
 
    9.35080705349434*10^-6, 1.587374864639623*10^-6, 
 
    -(3.406916140411397*10^-6), 7.615314617797895*10^-7, 
 
    1.043848891867664*10^-7, 8.00005968351009*10^-8}, 
 
   {0, -(1.060609684552122*10^-7), 
 
    -(3.597653387151837*10^-7), -(7.63332497354033*10^-7), 
 
    -(8.54683381010811*10^-7), 5.180770285569682*10^-7, 
 
    4.488364802538226*10^-6, 9.35080705349434*10^-6, 
 
    0.00001214415511460777, 8.28927830965502*10^-6, 
 
    -(2.984641830936814*10^-6), -(1.481607531486263*10^-7), 
 
    9.74075851446233*10^-8, 3.111901262563948*10^-7}, 
 
   {0, 1.207785178759421*10^-8, 4.138547270695188*10^-8, 
 
    -(1.054947861732843*10^-7), -(7.881958600623805*10^-7), 
 
    -(1.984145625410027*10^-6), -(2.095110410368116*10^-6), 
 
    1.587374864639622*10^-6, 8.28927830965502*10^-6, 
 
    0.00002189408849724766, 7.067977424963907*10^-6, 
 
    -(5.020091709866418*10^-6), 6.153122317227934*10^-8, 
 
    6.674652416035995*10^-7}, 
 
   {0, 3.295093655251222*10^-8, 1.116673291591555*10^-7, 
 
    2.215523710965161*10^-7, 1.704335751438199*10^-7, 
 
    -(4.62231649962785*10^-7), -(2.035089160904369*10^-6), 
 
    -(3.406916140411396*10^-6), -(2.984641830936816*10^-6), 
 
    7.067977424963909*10^-6, 0.00006726699878282563, 
 
    -0.00001449288193650251, -(5.791218087770134*10^-6), 
 
    -(2.099822111190748*10^-6)}, 
 
   {0, -(1.14251835898497*10^-8), 
 
    -(3.874019304217086*10^-8), -(6.155143543861747*10^-8), 
 
    1.23858438217546*10^-8, 3.115791458331528*10^-7, 
 
    7.947781044659666*10^-7, 7.615314617797903*10^-7, 
 
    -(1.481607531486263*10^-7), -(5.020091709866418*10^-6), 
 
    -0.00001449288193650251, 0.0000904342095895274, 
 
    -(1.364692417437816*10^-6), -0.00003707121302871648}, 
 
   {0, -(4.146716409259472*10^-10), 
 
    -(1.412267963264319*10^-9), -(2.907795067079517*10^-9), 
 
    -(2.746237478500668*10^-9), 8.26917062576143*10^-9, 
 
    4.008322474795886*10^-8, 1.043848891867673*10^-7, 
 
    9.74075851446249*10^-8, 6.153122317227934*10^-8, 
 
    -(5.791218087770137*10^-6), -(1.36469241743782*10^-6), 
 
    0.00005168582451576197, 0.00007173028219076825}, 
 
   {0, 5.813793432178517*10^-10, 1.979616277295404*10^-9, 
 
    -(9.76260383776585*10^-12), -(1.498129016302585*10^-8), 
 
    -(4.464406871327516*10^-8), -(4.541272367597646*10^-8), 
 
    8.00005968351026*10^-8, 3.111901262563965*10^-7, 
 
    6.674652416036011*10^-7, -(2.099822111190742*10^-6), 
 
    -0.00003707121302871648, 0.00007173028219076825, 
 
    0.000119452180474698}}, 
 
  {1, 0.999139263101455, 0.997073998999712, 
 
   0.993695970475643, 0.98972663421669, 0.985186047101021, 
 
   0.980975722868737, 0.976725468443208, 0.972950395078783, 
 
   0.966734933943231, 0.955814125289614, 0.942172273052315, 
 
   0.931457958226178, 0.926498704055163}, True, 0, False, 
 
  2, Automatic, 3, 0.0003174009528607147, Hold[cpbills]]\
\>", "\<\
-DF[{1988, 7, 19}]-\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["\<\
dfbills3 = DiscountFunction[cpbills, DropNumber -> 0,
\tFunctionalForm -> DLogDelta, MaximumMaturity -> 1,
\tLambdaValue -> 0];\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["lambdastar = 0", "Print",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
DF[{1988, 7, 19}, InterpolatingFunction[{0, 366}, 
 
   {{0, 0, {0.0000836712239233903, 0}, {0}}, 
 
    {15, 0, {0.0001661817575334792, 
 
      -(5.107160487641287*10^-6), 
 
      -(7.071908485542588*10^-7), 8.62822225339537*10^-8}, 
 
     {0, 0, 15}}, {36, 15, 
 
     {0.0001985991664461284, 2.517487112819327*10^-6, 
 
      4.637147495817843*10^-8, -(1.287311939278259*10^-8)}, 
 
     {0, 0, 21}}, {57, 36, 
 
     {0.0001940084126632299, -(9.87551516528769*10^-7), 
 
      -(3.661639016826693*10^-8), 4.460658145761645*10^-9}, 
 
     {0, 0, 21}}, {78, 57, 
 
     {0.0001936589315276561, 7.269711077997087*10^-7, 
 
      3.541014603031229*10^-8, -(5.154115443432237*10^-10)}\\
 
      , {0, 0, 21}}, {106, 78, 
 
     {0.000213744613862863, 9.67307668353467*10^-9, 
 
      -(2.527402575008671*10^-8), 
 
      -(8.90366595521278*10^-10)}, {0, 0, 28}}, 
 
    {120, 106, {0.0002074996910567463, 
 
      -(6.546838315894022*10^-7), 
 
      -(1.490127977619817*10^-8), 1.260821808874429*10^-9}, 
 
     {0, 0, 14}}, {141, 120, 
 
     {0.0002013300902032608, 3.702413761719528*10^-7, 
 
      3.16205663335521*10^-8, 6.873890663216182*10^-10}, 
 
     {0, 0, 21}}, {162, 141, 
 
     {0.0002184324668882142, 7.355434456827305*10^-7, 
 
      -(3.755021146521673*10^-9), -(1.18597042554351*10^-9)}
 
      , {0, 0, 21}}, {211, 162, 
 
     {0.0002152817774131317, -(2.596266429189046*10^-7), 
 
      -(3.986262402308993*10^-9), 2.517769151084356*10^-10}\\
 
      , {0, 0, 49}}, {295, 211, 
 
     {0.0002128249918967496, -(2.699567948253482*10^-7), 
 
      -(2.865587478592275*10^-9), 
 
      -(6.676424950355139*10^-11)}, {0, 0, 84}}, 
 
    {366, 295, {0.0003671733309146274, 
 
      7.663313088242231*10^-6, 7.731539183640558*10^-8, 
 
      6.041491286851843*10^-10}, {0, 0, 71}}}], 0, 
 
  14.00000000000023, DLogDelta, 
 
  {-0.00467750557702118, -0.01542850730082534, 
 
   -0.00970212777569657, -0.003271577852800078, 
 
   -0.007919033199129899, -0.01691594759135739, 
 
   -0.006964976093172482, 0.004708330602056777, 
 
   -0.001372620805852875, 0.007818919153734783, 
 
   0.001659725231178299, 0.002394402379977123, 
 
   0.01706037562175311, -0.006111986161570826, 
 
   0.001855785131681387, 0.00645971966508796, 
 
   -0.002966050455043501, 0.004589821359843427, 
 
   0.002144452441527278, 0.002320299770943279, 
 
   0.00902668498181924, -0.002825459498012606, 
 
   0.01631656622713251, -0.001639556845759671, 
 
   -0.001240031667592234, 0.0001867981870447011, 
 
   -0.002468786399248302, 0.00351501097240714, 
 
   -0.002128657111356347, 0.0005487241462134307, 
 
   -0.00003452236133227871}, 
 
  {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 
 
   17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 
 
   31}, 0.0000972040917821574, 0.005363643953812002, 
 
  {-0.368795868167189, -1.343962803535624, 
 
   -0.839406447097076, -0.2787728488674067, 
 
   -0.7036086766903864, -1.522050440675348, 
 
   -0.6212023881290051, 0.357646113166592, 
 
   -0.1171019773710905, 0.6533658778419835, 
 
   0.1432011041308268, 0.2044459792734315, 
 
   1.512040647676351, -0.5336141172997009, 
 
   0.1641857006847537, 0.5424346278709304, 
 
   -0.2608831246741212, 0.3938480535677463, 
 
   0.1902151387041636, 0.2029320091579365, 
 
   0.811566350521686, -0.2582506865807593, 
 
   1.467215698558616, -0.1477004253878531, 
 
   -0.1078264158851606, 0.01460628807074495, 
 
   -0.197363029234126, 0.2879295894591812, 
 
   -0.1593078122594495, 0.03946195610183562, 
 
   -0.002475984306747269}, 
 
  {0, 15, 36, 57, 78, 106, 120, 141, 162, 211, 295, 366}, 
 
  {{0.00004223381229831822, -0.00003248273886355708, 
 
    0.00001553829318686173, -(6.882947612557803*10^-6), 
 
    3.425929806687838*10^-6, -(2.43463552657649*10^-6), 
 
    1.422810522919178*10^-6, -(6.763667365003823*10^-7), 
 
    2.813405465675374*10^-7, -(1.66308651253396*10^-7), 
 
    1.665582239318665*10^-7, -(1.903962918558462*10^-7), 
 
    2.051206317345215*10^-7, -(3.945656392275012*10^-7)}, 
 
   {-0.00003248273886355712, 0.00002928160522983702, 
 
    -0.00001709592691648385, 8.12156589723374*10^-6, 
 
    -(4.158280257276202*10^-6), 2.983762180159935*10^-6, 
 
    -(1.750403164490621*10^-6), 8.32889281828933*10^-7, 
 
    -(3.465682317579589*10^-7), 2.048843936397364*10^-7, 
 
    -(2.051964553842247*10^-7), 2.34565746061637*10^-7, 
 
    -(2.527065919539572*10^-7), 4.861014904724298*10^-7}, 
 
   {0.00001553829318686173, -0.00001709592691648382, 
 
    0.00001307052554318159, -(7.50370727147073*10^-6), 
 
    4.156009232593612*10^-6, -(3.061683538565671*10^-6), 
 
    1.814619192065678*10^-6, -(8.65642275892849*10^-7), 
 
    3.605312872429083*10^-7, -(2.131881951195156*10^-7), 
 
    2.135256362327838*10^-7, -(2.44090550298693*10^-7), 
 
    2.629698283337216*10^-7, -(5.058450939140348*10^-7)}, 
 
   {-(6.882947612557805*10^-6), 8.12156589723374*10^-6, 
 
    -(7.503707271470736*10^-6), 5.48974800196099*10^-6, 
 
    -(3.662590228647599*10^-6), 2.866251405268684*10^-6, 
 
    -(1.737691189721488*10^-6), 8.33557790448007*10^-7, 
 
    -(3.478703725000568*10^-7), 2.058046150716558*10^-7, 
 
    -(2.061570721079393*10^-7), 2.356745337706917*10^-7, 
 
    -(2.539066526576526*10^-7), 4.884143183864084*10^-7}, 
 
   {3.425929806687841*10^-6, -(4.1582802572762*10^-6), 
 
    4.156009232593613*10^-6, -(3.662590228647598*10^-6), 
 
    3.233961480811136*10^-6, -(3.063554435739864*10^-6), 
 
    1.98531668736318*10^-6, -(9.67759406086553*10^-7), 
 
    4.062329928879155*10^-7, -(2.406782735262834*10^-7), 
 
    2.411799861829977*10^-7, -(2.757365288460289*10^-7), 
 
    2.970805764279461*10^-7, -(5.714737237580949*10^-7)}, 
 
   {-(2.434635526576499*10^-6), 2.983762180159944*10^-6, 
 
    -(3.061683538565679*10^-6), 2.866251405268689*10^-6, 
 
    -(3.063554435739868*10^-6), 3.811407813756897*10^-6, 
 
    -(2.972159147764499*10^-6), 1.520089423703781*10^-6, 
 
    -(6.493661892386636*10^-7), 3.863826666449425*10^-7, 
 
    -(3.876170401246115*10^-7), 4.432726908719398*10^-7, 
 
    -(4.77645988864765*10^-7), 9.18863678728083*10^-7}, 
 
   {1.422810522919178*10^-6, -(1.75040316449062*10^-6), 
 
    1.81461919206568*10^-6, -(1.737691189721486*10^-6), 
 
    1.985316687363181*10^-6, -(2.972159147764497*10^-6), 
 
    2.966518960484409*10^-6, -(1.890806078407832*10^-6), 
 
    8.73693407674915*10^-7, -(5.295613667352903*10^-7), 
 
    5.337585507006598*10^-7, -(6.110833531131572*10^-7), 
 
    6.588237825623739*10^-7, -(1.267684787944829*10^-6)}, 
 
   {-(6.763667365003787*10^-7), 8.32889281828928*10^-7, 
 
    -(8.65642275892848*10^-7), 8.33557790448003*10^-7, 
 
    -(9.67759406086552*10^-7), 1.52008942370378*10^-6, 
 
    -(1.890806078407832*10^-6), 1.752543390584838*10^-6, 
 
    -(1.09704995311725*10^-6), 7.074695196037779*10^-7, 
 
    -(7.239485208415287*10^-7), 8.31792926520746*10^-7, 
 
    -(8.9830906065455*10^-7), 1.729718843630719*10^-6}, 
 
   {2.813405465675358*10^-7, -(3.465682317579574*10^-7), 
 
    3.605312872429087*10^-7, -(3.478703725000565*10^-7), 
 
    4.062329928879162*10^-7, -(6.493661892386639*10^-7), 
 
    8.73693407674915*10^-7, -(1.097049953117251*10^-6), 
 
    1.102852118807317*10^-6, -(9.53338639168383*10^-7), 
 
    1.038940933494123*10^-6, -(1.211110871464444*10^-6), 
 
    1.316975705128989*10^-6, -(2.543088402468602*10^-6)}, 
 
   {-(1.66308651253402*10^-7), 2.048843936397406*10^-7, 
 
    -(2.131881951195179*10^-7), 2.05804615071657*10^-7, 
 
    -(2.406782735262853*10^-7), 3.863826666449439*10^-7, 
 
    -(5.295613667352923*10^-7), 7.074695196037793*10^-7, 
 
    -(9.53338639168384*10^-7), 1.205662632975791*10^-6, 
 
    -(1.52257149500108*10^-6), 1.837325415865525*10^-6, 
 
    -(2.031154838449741*10^-6), 3.949096595752875*10^-6}, 
 
   {1.665582239318622*10^-7, -(2.051964553842214*10^-7), 
 
    2.135256362327798*10^-7, -(2.061570721079368*10^-7), 
 
    2.411799861829974*10^-7, -(3.876170401246098*10^-7), 
 
    5.337585507006586*10^-7, -(7.239485208415281*10^-7), 
 
    1.038940933494122*10^-6, -(1.522571495001078*10^-6), 
 
    2.259656784912101*10^-6, -(3.057793961191708*10^-6), 
 
    3.583680709304308*10^-6, -(7.141857619389668*10^-6)}, 
 
   {-(1.903962918558418*10^-7), 2.345657460616354*10^-7, 
 
    -(2.440905502986863*10^-7), 2.356745337706882*10^-7, 
 
    -(2.757365288460292*10^-7), 4.432726908719387*10^-7, 
 
    -(6.110833531131558*10^-7), 8.31792926520745*10^-7, 
 
    -(1.211110871464443*10^-6), 1.837325415865522*10^-6, 
 
    -(3.057793961191708*10^-6), 4.7399351820066*10^-6, 
 
    -(6.171764963540804*10^-6), 0.00001289585024854992}, 
 
   {2.051206317345189*10^-7, -(2.527065919539567*10^-7), 
 
    2.629698283337134*10^-7, -(2.53906652657648*10^-7), 
 
    2.970805764279472*10^-7, -(4.776459888647647*10^-7), 
 
    6.588237825623727*10^-7, -(8.98309060654549*10^-7), 
 
    1.316975705128987*10^-6, -(2.031154838449736*10^-6), 
 
    3.583680709304307*10^-6, -(6.171764963540806*10^-6), 
 
    9.17521774434847*10^-6, -0.00002103661103713942}, 
 
   {-(3.945656392275056*10^-7), 4.86101490472437*10^-7, 
 
    -(5.058450939140311*10^-7), 4.884143183864069*10^-7, 
 
    -(5.714737237581021*10^-7), 9.18863678728086*10^-7, 
 
    -(1.26768478794483*10^-6), 1.729718843630719*10^-6, 
 
    -(2.543088402468601*10^-6), 3.94909659575287*10^-6, 
 
    -(7.141857619389667*10^-6), 0.00001289585024854992, 
 
    -0.00002103661103713942, 0.00005458205315439909}}, 
 
  {0.0000836712239233903, 0.0002612815491190206, 
 
   0.0000943279751654167, 0.000231521879192796, 
 
   0.000185620970321999, 0.0001905362427404309, 
 
   0.0002202595001204984, 0.000205702573345996, 
 
   0.0001945598969044938, 0.0002351280562763666, 
 
   0.0002007954934794692, 0.0002308406881348646, 
 
   0.0001858082544928946, 0.0003671733309146274}, False, 0, 
 
  True, 2, Automatic, 3, 0.00569183959657872, Hold[cpbills]]\
\>", "\<\
-DF[{1988, 7, 19}]-\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["PlotGCVFunction[];", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.04329 0.0147151 103.413 [
[(0)] .02381 0 0 2 Msboxa
[(5)] .24026 0 0 2 Msboxa
[(10)] .45671 0 0 2 Msboxa
[(15)] .67316 0 0 2 Msboxa
[(20)] .88961 0 0 2 Msboxa
[(0)] -0.0125 .01472 1 0 Msboxa
[(0.001)] -0.0125 .11813 1 0 Msboxa
[(0.002)] -0.0125 .22154 1 0 Msboxa
[(0.003)] -0.0125 .32495 1 0 Msboxa
[(0.004)] -0.0125 .42837 1 0 Msboxa
[(0.005)] -0.0125 .53178 1 0 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
p
.002 w
.02381 0 m
.02381 .00625 L
s
P
[(0)] .02381 0 0 2 Mshowa
p
.002 w
.24026 0 m
.24026 .00625 L
s
P
[(5)] .24026 0 0 2 Mshowa
p
.002 w
.45671 0 m
.45671 .00625 L
s
P
[(10)] .45671 0 0 2 Mshowa
p
.002 w
.67316 0 m
.67316 .00625 L
s
P
[(15)] .67316 0 0 2 Mshowa
p
.002 w
.88961 0 m
.88961 .00625 L
s
P
[(20)] .88961 0 0 2 Mshowa
p
.001 w
.0671 0 m
.0671 .00375 L
s
P
p
.001 w
.11039 0 m
.11039 .00375 L
s
P
p
.001 w
.15368 0 m
.15368 .00375 L
s
P
p
.001 w
.19697 0 m
.19697 .00375 L
s
P
p
.001 w
.28355 0 m
.28355 .00375 L
s
P
p
.001 w
.32684 0 m
.32684 .00375 L
s
P
p
.001 w
.37013 0 m
.37013 .00375 L
s
P
p
.001 w
.41342 0 m
.41342 .00375 L
s
P
p
.001 w
.5 0 m
.5 .00375 L
s
P
p
.001 w
.54329 0 m
.54329 .00375 L
s
P
p
.001 w
.58658 0 m
.58658 .00375 L
s
P
p
.001 w
.62987 0 m
.62987 .00375 L
s
P
p
.001 w
.71645 0 m
.71645 .00375 L
s
P
p
.001 w
.75974 0 m
.75974 .00375 L
s
P
p
.001 w
.80303 0 m
.80303 .00375 L
s
P
p
.001 w
.84632 0 m
.84632 .00375 L
s
P
p
.001 w
.9329 0 m
.9329 .00375 L
s
P
p
.001 w
.97619 0 m
.97619 .00375 L
s
P
p
.002 w
0 0 m
1 0 L
s
P
p
.002 w
0 .01472 m
.00625 .01472 L
s
P
[(0)] -0.0125 .01472 1 0 Mshowa
p
.002 w
0 .11813 m
.00625 .11813 L
s
P
[(0.001)] -0.0125 .11813 1 0 Mshowa
p
.002 w
0 .22154 m
.00625 .22154 L
s
P
[(0.002)] -0.0125 .22154 1 0 Mshowa
p
.002 w
0 .32495 m
.00625 .32495 L
s
P
[(0.003)] -0.0125 .32495 1 0 Mshowa
p
.002 w
0 .42837 m
.00625 .42837 L
s
P
[(0.004)] -0.0125 .42837 1 0 Mshowa
p
.002 w
0 .53178 m
.00625 .53178 L
s
P
[(0.005)] -0.0125 .53178 1 0 Mshowa
p
.001 w
0 .0354 m
.00375 .0354 L
s
P
p
.001 w
0 .05608 m
.00375 .05608 L
s
P
p
.001 w
0 .07676 m
.00375 .07676 L
s
P
p
.001 w
0 .09745 m
.00375 .09745 L
s
P
p
.001 w
0 .13881 m
.00375 .13881 L
s
P
p
.001 w
0 .15949 m
.00375 .15949 L
s
P
p
.001 w
0 .18018 m
.00375 .18018 L
s
P
p
.001 w
0 .20086 m
.00375 .20086 L
s
P
p
.001 w
0 .24222 m
.00375 .24222 L
s
P
p
.001 w
0 .26291 m
.00375 .26291 L
s
P
p
.001 w
0 .28359 m
.00375 .28359 L
s
P
p
.001 w
0 .30427 m
.00375 .30427 L
s
P
p
.001 w
0 .34564 m
.00375 .34564 L
s
P
p
.001 w
0 .36632 m
.00375 .36632 L
s
P
p
.001 w
0 .387 m
.00375 .387 L
s
P
p
.001 w
0 .40768 m
.00375 .40768 L
s
P
p
.001 w
0 .44905 m
.00375 .44905 L
s
P
p
.001 w
0 .46973 m
.00375 .46973 L
s
P
p
.001 w
0 .49041 m
.00375 .49041 L
s
P
p
.001 w
0 .5111 m
.00375 .5111 L
s
P
p
.001 w
0 .55246 m
.00375 .55246 L
s
P
p
.001 w
0 .57314 m
.00375 .57314 L
s
P
p
.001 w
0 .59383 m
.00375 .59383 L
s
P
p
.001 w
0 .61451 m
.00375 .61451 L
s
P
p
.002 w
0 0 m
0 .61803 L
s
P
P
p
P
p
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.004 w
.0671 .60332 m
.11039 .60327 L
.15368 .60273 L
.19697 .59748 L
.24026 .55204 L
.28355 .37568 L
.32684 .24339 L
.37013 .13336 L
.41342 .07179 L
.45671 .05212 L
.5 .04445 L
.54329 .04538 L
.58658 .05117 L
.62987 .09814 L
.67316 .13273 L
.71645 .13801 L
.75974 .13857 L
.80303 .13863 L
.84632 .13863 L
.88961 .13863 L
.9329 .13863 L
.97619 .13863 L
s
1 0 0 r
p
.015 w
.0671 .60332 Mdot
.11039 .60327 Mdot
.15368 .60273 Mdot
.19697 .59748 Mdot
.24026 .55204 Mdot
.28355 .37568 Mdot
.32684 .24339 Mdot
.37013 .13336 Mdot
.41342 .07179 Mdot
.45671 .05212 Mdot
.5 .04445 Mdot
.54329 .04538 Mdot
.58658 .05117 Mdot
.62987 .09814 Mdot
.67316 .13273 Mdot
.71645 .13801 Mdot
.75974 .13857 Mdot
.80303 .13863 Mdot
.84632 .13863 Mdot
.88961 .13863 Mdot
.9329 .13863 Mdot
.97619 .13863 Mdot
P
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282, 174},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell[CellGroupData[{Cell["\<\
dfbills4 = DiscountFunction[cpbills,
\tFunctionalForm -> DLogDelta, MaximumMaturity -> 1,
\tLambda0 -> {10^11, 10^12}];\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
lambda = 1.*10^11
lambda = 8.*10^10
lambda = 1.12360679775*10^11
lambda = 1.11014526840039*10^11
lambda = 1.200000000000047*10^11
lambda = 1.076393202250034*10^11
lambda = 1.152786404500073*10^11
                       11
lambdastar = 1.15279 10\
\>", "Print",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
DF[{1988, 7, 19}, InterpolatingFunction[{0, 366}, 
 
   {{0, 0, {0.0001657700452952142, 0}, {0}}, 
 
    {15, 0, {0.0001742586574694371, 
 
      5.634291450859389*10^-7, -(1.652222130391789*10^-10), 
 
      -(5.315363368196541*10^-12)}, {0, 0, 15}}, 
 
    {36, 15, {0.0001855768427064093, 
 
      4.951693210446725*10^-7, -(2.085327653138638*10^-9), 
 
      -(4.381845213278093*10^-11)}, {0, 0, 21}}, 
 
    {57, 36, {0.000194549455074028, 
 
      3.545789428197907*10^-7, -(3.461348227671384*10^-9), 
 
      -(1.085316856534328*10^-11)}, {0, 0, 21}}, 
 
    {78, 57, {0.0002005634386553607, 
 
      2.274571860977834*10^-7, -(2.805856401993631*10^-9), 
 
      2.103353251309464*10^-11}, {0, 0, 21}}, 
 
    {106, 78, {0.0002054096492483429, 
 
      1.305187391906597*10^-7, -(1.520007519953851*10^-9), 
 
      1.507401248687261*10^-11}, {0, 0, 28}}, 
 
    {120, 106, {0.0002071280922600151, 
 
      1.225714022186443*10^-7, -(1.246622760771363*10^-11), 
 
      3.876674795407838*10^-11}, {0, 0, 14}}, 
 
    {141, 120, {0.000210130212932778, 
 
      1.725959442672588*10^-7, 1.41132461870642*10^-9, 
 
      2.097753273708669*10^-11}, {0, 0, 21}}, 
 
    {162, 141, {0.0002140026931133165, 
 
      1.691306568982425*10^-7, -(7.27293391554365*10^-10), 
 
      -(6.140824280332665*10^-11)}, {0, 0, 21}}, 
 
    {211, 162, {0.000217507817062755, 
 
      -(2.483543028988299*10^-8), 
 
      -(1.966705553370557*10^-9), 5.118462964643169*10^-13}\\
 
      , {0, 0, 49}}, {295, 211, 
 
     {0.0002151678200905227, 1.291960472697793*10^-7, 
 
      1.8696804057389*10^-9, 2.268634220584934*10^-11}, 
 
     {0, 0, 84}}, {366, 295, 
 
     {0.0002351890331652437, 3.195260013733914*10^-7, 
 
      5.286913356059876*10^-10, -(2.286367475650902*10^-11)}
 
      , {0, 0, 71}}}], 1.152786404500073*10^11, 
 
  5.611196838927617, DLogDelta, 
 
  {-0.01319634320194041, -0.01272384908398294, 
 
   -0.01173796353604928, -0.0001095767171932493, 
 
   -0.01522062456588458, -0.006938975708507656, 
 
   -0.0003960149168733551, 0.00594008396369361, 
 
   0.00366963671389442, 0.02467761497324261, 
 
   0.01610681064212827, -0.005133725240526133, 
 
   0.006241377466622567, -0.01394678113673819, 
 
   0.005457281121735491, 0.013358980004881, 
 
   -0.005319758016028687, -0.002702585975029592, 
 
   -0.007205413323561061, -0.000909305271321159, 
 
   0.0118134878862719, 0.00504619376296489, 
 
   0.02484883222570033, 0.002818525351614198, 
 
   -0.001962832615873822, -0.01190876595855173, 
 
   -0.005651892280326365, 0.002001502598758975, 
 
   -0.005043320135953877, 0.01779900611926166, 
 
   -0.00918497901011506}, 
 
  {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 
 
   17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 
 
   31}, 0.0001428205002361394, 0.0086797432104912, 
 
  {-1.076895469704281, -1.006980500921195, 
 
   -0.935571389722851, -0.00872246608668515, 
 
   -1.21491672693942, -0.5501793252900045, 
 
   -0.03313094589681386, 0.4923714992037633, 
 
   0.2951114542366573, 1.964508880151253, 
 
   1.275808392239852, -0.4069443753805484, 
 
   0.4959419949876783, -1.110440578547638, 
 
   0.4350266251090409, 1.06439091281885, 
 
   -0.4235349365648392, -0.2153095200133042, 
 
   -0.574588308340852, -0.07259649203653939, 
 
   0.942437285130424, 0.4016721611007127, 
 
   1.959035770119528, 0.2205873922754302, 
 
   -0.1523409197165353, -0.893313465567502, 
 
   -0.4171781324457823, 0.1462211390043117, 
 
   -0.3636342487602977, 1.282025349260843, 
 
   -0.5683151770261343}, 
 
  {0, 15, 36, 57, 78, 106, 120, 141, 162, 211, 295, 366}, 
 
  {{3.922869274471652*10^-8, 3.294244488198481*10^-8, 
 
    1.773350586836198*10^-8, -(4.148652407388567*10^-9), 
 
    -(1.641723961171214*10^-8), -(1.462585790122128*10^-8), 
 
    -(5.108553007672365*10^-9), 1.828707840820542*10^-9, 
 
    4.775114628392724*10^-9, 3.604627997730375*10^-9, 
 
    -(3.371055699661138*10^-9), 2.57229589088243*10^-10, 
 
    7.840333221918997*10^-10, 1.195786568703268*10^-9}, 
 
   {3.294244488198481*10^-8, 2.779573688034324*10^-8, 
 
    1.535218781318707*10^-8, -(2.685449753791385*10^-9), 
 
    -(1.329568756645218*10^-8), -(1.256483863256171*10^-8), 
 
    -(4.984195111986305*10^-9), 9.06440960840544*10^-10, 
 
    3.650048741057596*10^-9, 3.131495328288156*10^-9, 
 
    -(2.633106556091145*10^-9), 5.654061265865708*10^-11, 
 
    6.652966192026487*10^-10, 1.049805904168348*10^-9}, 
 
   {1.773350586836198*10^-8, 1.535218781318706*10^-8, 
 
    9.62158901831782*10^-9, 8.89315875948758*10^-10, 
 
    -(5.755654650887458*10^-9), -(7.621623213681385*10^-9), 
 
    -(4.711380243996454*10^-9), -(1.326425025198088*10^-9), 
 
    9.44503931665265*10^-10, 2.006824521156481*10^-9, 
 
    -(8.61668813305916*10^-10), 
 
    -(4.306056848108917*10^-10), 3.826402902785442*10^-10, 
 
    7.042140622177583*10^-10}, 
 
   {-(4.148652407388571*10^-9), -(2.68544975379138*10^-9), 
 
    8.89315875948757*10^-10, 5.460876329561661*10^-9, 
 
    5.233337913754311*10^-9, 1.883529486378106*10^-10, 
 
    -(3.821083631399194*10^-9), -(4.469572906446768*10^-9), 
 
    -(3.192957080851827*10^-9), 4.22190135645028*10^-11, 
 
    1.896637274132715*10^-9, -(1.08271201024777*10^-9), 
 
    -(1.039226835915541*10^-10), 7.109127756829564*10^-11}, 
 
   {-(1.641723961171214*10^-8), -(1.329568756645217*10^-8), 
 
    -(5.755654650887459*10^-9), 5.233337913754314*10^-9, 
 
    1.096785808373167*10^-8, 7.528834455800814*10^-9, 
 
    -(3.656254442205164*10^-10), -(5.1326184772892*10^-9), 
 
    -(6.205427991957922*10^-9), -(2.963626593770383*10^-9), 
 
    4.085937487452574*10^-9, -(8.49491710212213*10^-10), 
 
    -(8.12675489137659*10^-10), -(1.096007182362264*10^-9)}\\
 
    , {-(1.462585790122128*10^-8), 
 
    -(1.256483863256171*10^-8), -(7.621623213681382*10^-9), 
 
    1.883529486378123*10^-10, 7.528834455800818*10^-9, 
 
    1.061840169589896*10^-8, 5.709078876643914*10^-9, 
 
    -(1.035455241675877*10^-9), -(5.508738904197939*10^-9), 
 
    -(6.347051702958773*10^-9), 3.642611000941493*10^-9, 
 
    1.170581797911703*10^-9, -(1.525539881265241*10^-9), 
 
    -(2.651044750562054*10^-9)}, 
 
   {-(5.108553007672361*10^-9), -(4.984195111986304*10^-9), 
 
    -(4.711380243996456*10^-9), -(3.821083631399195*10^-9), 
 
    -(3.656254442205181*10^-10), 5.709078876643915*10^-9, 
 
    8.33064980346182*10^-9, 5.552630283967059*10^-9, 
 
    7.281332104823307*10^-10, -(5.834773142451401*10^-9), 
 
    -(1.894901852972432*10^-9), 4.065131683066338*10^-9, 
 
    -(1.072871755150657*10^-9), -(2.586669060094763*10^-9)}\\
 
    , {1.828707840820544*10^-9, 9.06440960840544*10^-10, 
 
    -(1.326425025198088*10^-9), -(4.469572906446768*10^-9), 
 
    -(5.1326184772892*10^-9), -(1.03545524167588*10^-9), 
 
    5.552630283967057*10^-9, 8.83471919991503*10^-9, 
 
    7.811839691466758*10^-9, 9.37319491221975*10^-11, 
 
    -(8.88523613512784*10^-9), 3.805023327204399*10^-9, 
 
    1.331505350917717*10^-9, 1.163421725115161*10^-9}, 
 
   {4.775114628392731*10^-9, 3.650048741057599*10^-9, 
 
    9.44503931665268*10^-10, -(3.192957080851828*10^-9), 
 
    -(6.205427991957926*10^-9), -(5.508738904197945*10^-9), 
 
    7.281332104823303*10^-10, 7.811839691466761*10^-9, 
 
    1.171237413342918*10^-8, 8.07408789808581*10^-9, 
 
    -(1.283855216759367*10^-8), 
 
    -(4.358883841993031*10^-10), 4.461837931564647*10^-9, 
 
    6.954197196743915*10^-9}, 
 
   {3.604627997730371*10^-9, 3.131495328288154*10^-9, 
 
    2.006824521156478*10^-9, 4.221901356450446*10^-11, 
 
    -(2.96362659377038*10^-9), -(6.347051702958774*10^-9), 
 
    -(5.834773142451403*10^-9), 9.37319491221954*10^-11, 
 
    8.0740878980858*10^-9, 1.864782059402424*10^-8, 
 
    -(5.717531602622326*10^-9), -(1.775886779882065*10^-8), 
 
    9.94986439267933*10^-9, 1.923427674788701*10^-8}, 
 
   {-(3.371055699661138*10^-9), -(2.633106556091146*10^-9), 
 
    -(8.61668813305916*10^-10), 1.896637274132716*10^-9, 
 
    4.085937487452574*10^-9, 3.642611000941493*10^-9, 
 
    -(1.894901852972432*10^-9), -(8.88523613512784*10^-9), 
 
    -(1.283855216759366*10^-8), -(5.71753160262232*10^-9), 
 
    4.104614965654538*10^-8, -(2.597925668258203*10^-8), 
 
    -(1.675410898709608*10^-8), -(1.550603015628743*10^-8)}\\
 
    , {2.572295890882443*10^-10, 5.654061265865361*10^-11, 
 
    -(4.306056848108934*10^-10), 
 
    -(1.082712010247771*10^-9), -(8.49491710212216*10^-10), 
 
    1.170581797911701*10^-9, 4.065131683066343*10^-9, 
 
    3.805023327204398*10^-9, -(4.358883841993016*10^-10), 
 
    -(1.775886779882066*10^-8), -(2.597925668258205*10^-8), 
 
    9.05636247939437*10^-8, -(5.051754786735377*10^-8), 
 
    -(9.99395124002248*10^-8)}, 
 
   {7.840333221918982*10^-10, 6.652966192026505*10^-10, 
 
    3.826402902785445*10^-10, -(1.039226835915538*10^-10), 
 
    -(8.12675489137659*10^-10), -(1.525539881265243*10^-9), 
 
    -(1.072871755150656*10^-9), 1.331505350917721*10^-9, 
 
    4.461837931564654*10^-9, 9.94986439267934*10^-9, 
 
    -(1.675410898709611*10^-8), -(5.05175478673537*10^-8), 
 
    1.205777604019076*10^-7, 1.851856523484418*10^-7}, 
 
   {1.195786568703261*10^-9, 1.049805904168349*10^-9, 
 
    7.042140622177597*10^-10, 7.109127756830135*10^-11, 
 
    -(1.09600718236226*10^-9), -(2.65104475056206*10^-9), 
 
    -(2.586669060094766*10^-9), 1.163421725115163*10^-9, 
 
    6.954197196743924*10^-9, 1.923427674788703*10^-8, 
 
    -(1.550603015628743*10^-8), -(9.99395124002248*10^-8), 
 
    1.851856523484418*10^-7, 2.922495406785015*10^-7}}, 
 
  {0.0001657700452952142, 0.0001686059945688106, 
 
   0.0001754112357892829, 0.0001860186534331546, 
 
   0.0001950917769948569, 0.0002015575459245557, 
 
   0.0002049440253277085, 0.0002073621259106378, 
 
   0.0002098579905702688, 0.0002162730311172489, 
 
   0.0002198819799924454, 0.0002071026082481056, 
 
   0.0002276269177994068, 0.0002351890331652437}, False, 0, 
 
  False, 2, Automatic, 3, 0.0002873736978661365, 
 
  Hold[cpbills]]\
\>", "\<\
-DF[{1988, 7, 19}]-\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  "Note that when the functional form is either ",
  StyleBox["LogDelta",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " or ",
  StyleBox["DLogDelta",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", the option ",
  StyleBox["ShowLambdaProgress",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " controls whether the adaptive seach routine reports its progess or not. \
For long calculations involving many securities and many knot points, it is \
useful to see how $\\lambda$ evolved and where it ended if the the \
calculation was ultimately aborted by the user or if it bombed on its own. "
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
Now let's compare the graphs of the two estimation techniques:\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
ShowGraph[dfbills2];
ShowGraph[dfbills4];\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.29787 
MathPictureStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0309017 0.952381 [
[ 0 0 0 0 ]
[ 1 1.29787 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
0 g
p
P
0 0 m
1 0 L
1 1.29787 L
0 1.29787 L
closepath
clip
newpath
p
p
% Start of sub-graphic
p
0.0238095 0.0309017 0.97619 0.619505 MathSubStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0212003 0.953033 -1.39903 0.233095 [
[(0)] .0212 0 0 2 0 Minner Mrotsboxa
[(0.2)] .21181 0 0 2 0 Minner Mrotsboxa
[(0.4)] .40241 0 0 2 0 Minner Mrotsboxa
[(0.6)] .59302 0 0 2 0 Minner Mrotsboxa
[(0.8)] .78363 0 0 2 0 Minner Mrotsboxa
[(1)] .97423 0 0 2 0 Minner Mrotsboxa
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotsboxa
[(6.5)] -0.0125 .11609 1 0 0 Minner Mrotsboxa
[(7)] -0.0125 .23263 1 0 0 Minner Mrotsboxa
[(7.5)] -0.0125 .34918 1 0 0 Minner Mrotsboxa
[(8)] -0.0125 .46573 1 0 0 Minner Mrotsboxa
[(8.5)] -0.0125 .58228 1 0 0 Minner Mrotsboxa
[(percent)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotsboxa
[(7/19/1988, 10 knots,   5.6 parms, DLogDelta)] .5 .61803 0 -2 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
p
p
[ .002 .01 ] 0 setdash
.002 w
.0212 0 m
.0212 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.21181 0 m
.21181 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.40241 0 m
.40241 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.59302 0 m
.59302 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.78363 0 m
.78363 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.97423 0 m
.97423 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .11609 m
1 .11609 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .23263 m
1 .23263 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .34918 m
1 .34918 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .46573 m
1 .46573 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .58228 m
1 .58228 L
s
P
p
.002 w
.0212 0 m
.0212 .00625 L
s
P
[(0)] .0212 0 0 2 0 Minner Mrotshowa
p
.002 w
.21181 0 m
.21181 .00625 L
s
P
[(0.2)] .21181 0 0 2 0 Minner Mrotshowa
p
.002 w
.40241 0 m
.40241 .00625 L
s
P
[(0.4)] .40241 0 0 2 0 Minner Mrotshowa
p
.002 w
.59302 0 m
.59302 .00625 L
s
P
[(0.6)] .59302 0 0 2 0 Minner Mrotshowa
p
.002 w
.78363 0 m
.78363 .00625 L
s
P
[(0.8)] .78363 0 0 2 0 Minner Mrotshowa
p
.002 w
.97423 0 m
.97423 .00625 L
s
P
[(1)] .97423 0 0 2 0 Minner Mrotshowa
p
.001 w
.05932 0 m
.05932 .00375 L
s
P
p
.001 w
.09744 0 m
.09744 .00375 L
s
P
p
.001 w
.13556 0 m
.13556 .00375 L
s
P
p
.001 w
.17369 0 m
.17369 .00375 L
s
P
p
.001 w
.24993 0 m
.24993 .00375 L
s
P
p
.001 w
.28805 0 m
.28805 .00375 L
s
P
p
.001 w
.32617 0 m
.32617 .00375 L
s
P
p
.001 w
.36429 0 m
.36429 .00375 L
s
P
p
.001 w
.44053 0 m
.44053 .00375 L
s
P
p
.001 w
.47866 0 m
.47866 .00375 L
s
P
p
.001 w
.51678 0 m
.51678 .00375 L
s
P
p
.001 w
.5549 0 m
.5549 .00375 L
s
P
p
.001 w
.63114 0 m
.63114 .00375 L
s
P
p
.001 w
.66926 0 m
.66926 .00375 L
s
P
p
.001 w
.70738 0 m
.70738 .00375 L
s
P
p
.001 w
.74551 0 m
.74551 .00375 L
s
P
p
.001 w
.82175 0 m
.82175 .00375 L
s
P
p
.001 w
.85987 0 m
.85987 .00375 L
s
P
p
.001 w
.89799 0 m
.89799 .00375 L
s
P
p
.001 w
.93611 0 m
.93611 .00375 L
s
P
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotshowa
p
.002 w
0 0 m
1 0 L
s
P
p
.002 w
0 .11609 m
.00625 .11609 L
s
P
[(6.5)] -0.0125 .11609 1 0 0 Minner Mrotshowa
p
.002 w
0 .23263 m
.00625 .23263 L
s
P
[(7)] -0.0125 .23263 1 0 0 Minner Mrotshowa
p
.002 w
0 .34918 m
.00625 .34918 L
s
P
[(7.5)] -0.0125 .34918 1 0 0 Minner Mrotshowa
p
.002 w
0 .46573 m
.00625 .46573 L
s
P
[(8)] -0.0125 .46573 1 0 0 Minner Mrotshowa
p
.002 w
0 .58228 m
.00625 .58228 L
s
P
[(8.5)] -0.0125 .58228 1 0 0 Minner Mrotshowa
p
.001 w
0 .02285 m
.00375 .02285 L
s
P
p
.001 w
0 .04616 m
.00375 .04616 L
s
P
p
.001 w
0 .06947 m
.00375 .06947 L
s
P
p
.001 w
0 .09278 m
.00375 .09278 L
s
P
p
.001 w
0 .1394 m
.00375 .1394 L
s
P
p
.001 w
0 .1627 m
.00375 .1627 L
s
P
p
.001 w
0 .18601 m
.00375 .18601 L
s
P
p
.001 w
0 .20932 m
.00375 .20932 L
s
P
p
.001 w
0 .25594 m
.00375 .25594 L
s
P
p
.001 w
0 .27925 m
.00375 .27925 L
s
P
p
.001 w
0 .30256 m
.00375 .30256 L
s
P
p
.001 w
0 .32587 m
.00375 .32587 L
s
P
p
.001 w
0 .37249 m
.00375 .37249 L
s
P
p
.001 w
0 .3958 m
.00375 .3958 L
s
P
p
.001 w
0 .41911 m
.00375 .41911 L
s
P
p
.001 w
0 .44242 m
.00375 .44242 L
s
P
p
.001 w
0 .48904 m
.00375 .48904 L
s
P
p
.001 w
0 .51235 m
.00375 .51235 L
s
P
p
.001 w
0 .53566 m
.00375 .53566 L
s
P
p
.001 w
0 .55897 m
.00375 .55897 L
s
P
p
.001 w
0 .60558 m
.00375 .60558 L
s
P
[(percent)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotshowa
p
.002 w
0 0 m
0 .61803 L
s
P
P
p
[(7/19/1988, 10 knots,   5.6 parms, DLogDelta)] .5 .61803 0 -2 Mshowa
P
p
P
p
p
p
.002 w
.02381 .01713 m
.06349 .09036 L
.10317 .16111 L
.14286 .22272 L
.18254 .2712 L
.22222 .30665 L
.2619 .3324 L
.30159 .35139 L
.34127 .36723 L
.38095 .38553 L
.42063 .40911 L
.44048 .42102 L
.46032 .43132 L
.48016 .43965 L
.49008 .44307 L
.5 .44601 L
.50992 .44845 L
.51984 .45041 L
.5248 .45121 L
.52976 .45189 L
.53472 .45244 L
.53968 .45288 L
.54216 .45305 L
.54464 .45319 L
.54712 .4533 L
.54836 .45335 L
.5496 .45339 L
.55084 .45342 L
.55208 .45344 L
.55332 .45345 L
.55456 .45346 L
.5558 .45346 L
.55704 .45345 L
.55828 .45344 L
.55952 .45341 L
.56076 .45338 L
.562 .45334 L
.56448 .45325 L
.56696 .45312 L
.56944 .45296 L
.5744 .45255 L
.57937 .45203 L
.58929 .45068 L
.59921 .44896 L
.61905 .44467 L
.65873 .43452 L
.67857 .42968 L
.68849 .42753 L
.69841 .42566 L
.70833 .42413 L
Mistroke
.71329 .4235 L
.71825 .42299 L
.72321 .42259 L
.72569 .42244 L
.72817 .42231 L
.72941 .42226 L
.73065 .42222 L
.73189 .42219 L
.73313 .42217 L
.73438 .42216 L
.73562 .42215 L
.73686 .42215 L
.7381 .42217 L
.73934 .42219 L
.74058 .42222 L
.74182 .42226 L
.74306 .42231 L
.74554 .42244 L
.74802 .42261 L
.75298 .42307 L
.75794 .4237 L
.7629 .42452 L
.76786 .42552 L
.77778 .42811 L
.7877 .43155 L
.79762 .43589 L
.81746 .44716 L
.85714 .47829 L
.89683 .51743 L
.93651 .56048 L
.97619 .60332 L
Mfstroke
P
P
p
p
.004 w
.02381 .01472 m
.06349 .05138 L
.10317 .0876 L
.14286 .12195 L
.18254 .15299 L
.22222 .18001 L
.2619 .20312 L
.30159 .22283 L
.34127 .23975 L
.38095 .25478 L
.42063 .26891 L
.46032 .28263 L
.5 .29562 L
.53968 .30744 L
.57937 .31779 L
.61905 .32649 L
.65873 .33353 L
.69841 .33917 L
.7381 .34383 L
.77778 .34804 L
.81746 .35245 L
.85714 .35765 L
.89683 .36398 L
.93651 .37156 L
.97619 .3803 L
s
P
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
MathSubEnd
P
% End of sub-graphic
P
p
% Start of sub-graphic
p
0.0238095 0.678366 0.97619 1.26697 MathSubStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0212003 0.953033 -1.86164 0.29839 [
[(0)] .0212 0 0 2 0 Minner Mrotsboxa
[(0.2)] .21181 0 0 2 0 Minner Mrotsboxa
[(0.4)] .40241 0 0 2 0 Minner Mrotsboxa
[(0.6)] .59302 0 0 2 0 Minner Mrotsboxa
[(0.8)] .78363 0 0 2 0 Minner Mrotsboxa
[(1)] .97423 0 0 2 0 Minner Mrotsboxa
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotsboxa
[(6.5)] -0.0125 .0779 1 0 0 Minner Mrotsboxa
[(7)] -0.0125 .22709 1 0 0 Minner Mrotsboxa
[(7.5)] -0.0125 .37629 1 0 0 Minner Mrotsboxa
[(8)] -0.0125 .52548 1 0 0 Minner Mrotsboxa
[(percent)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotsboxa
[(7/19/1988, 10 knots,   5.7 parms, Delta)] .5 .61803 0 -2 Msboxa
[ -0.001 -0.001 0 0 ]
[ 1.001 .61903 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
[ ] 0 setdash
p
p
[ .002 .01 ] 0 setdash
.002 w
.0212 0 m
.0212 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.21181 0 m
.21181 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.40241 0 m
.40241 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.59302 0 m
.59302 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.78363 0 m
.78363 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
.97423 0 m
.97423 .61803 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .0779 m
1 .0779 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .22709 m
1 .22709 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .37629 m
1 .37629 L
s
P
p
[ .002 .01 ] 0 setdash
.002 w
0 .52548 m
1 .52548 L
s
P
p
.002 w
.0212 0 m
.0212 .00625 L
s
P
[(0)] .0212 0 0 2 0 Minner Mrotshowa
p
.002 w
.21181 0 m
.21181 .00625 L
s
P
[(0.2)] .21181 0 0 2 0 Minner Mrotshowa
p
.002 w
.40241 0 m
.40241 .00625 L
s
P
[(0.4)] .40241 0 0 2 0 Minner Mrotshowa
p
.002 w
.59302 0 m
.59302 .00625 L
s
P
[(0.6)] .59302 0 0 2 0 Minner Mrotshowa
p
.002 w
.78363 0 m
.78363 .00625 L
s
P
[(0.8)] .78363 0 0 2 0 Minner Mrotshowa
p
.002 w
.97423 0 m
.97423 .00625 L
s
P
[(1)] .97423 0 0 2 0 Minner Mrotshowa
p
.001 w
.05932 0 m
.05932 .00375 L
s
P
p
.001 w
.09744 0 m
.09744 .00375 L
s
P
p
.001 w
.13556 0 m
.13556 .00375 L
s
P
p
.001 w
.17369 0 m
.17369 .00375 L
s
P
p
.001 w
.24993 0 m
.24993 .00375 L
s
P
p
.001 w
.28805 0 m
.28805 .00375 L
s
P
p
.001 w
.32617 0 m
.32617 .00375 L
s
P
p
.001 w
.36429 0 m
.36429 .00375 L
s
P
p
.001 w
.44053 0 m
.44053 .00375 L
s
P
p
.001 w
.47866 0 m
.47866 .00375 L
s
P
p
.001 w
.51678 0 m
.51678 .00375 L
s
P
p
.001 w
.5549 0 m
.5549 .00375 L
s
P
p
.001 w
.63114 0 m
.63114 .00375 L
s
P
p
.001 w
.66926 0 m
.66926 .00375 L
s
P
p
.001 w
.70738 0 m
.70738 .00375 L
s
P
p
.001 w
.74551 0 m
.74551 .00375 L
s
P
p
.001 w
.82175 0 m
.82175 .00375 L
s
P
p
.001 w
.85987 0 m
.85987 .00375 L
s
P
p
.001 w
.89799 0 m
.89799 .00375 L
s
P
p
.001 w
.93611 0 m
.93611 .00375 L
s
P
[(years)] .5 0 0 2 0 0 -1 Mouter Mrotshowa
p
.002 w
0 0 m
1 0 L
s
P
p
.002 w
0 .0779 m
.00625 .0779 L
s
P
[(6.5)] -0.0125 .0779 1 0 0 Minner Mrotshowa
p
.002 w
0 .22709 m
.00625 .22709 L
s
P
[(7)] -0.0125 .22709 1 0 0 Minner Mrotshowa
p
.002 w
0 .37629 m
.00625 .37629 L
s
P
[(7.5)] -0.0125 .37629 1 0 0 Minner Mrotshowa
p
.002 w
0 .52548 m
.00625 .52548 L
s
P
[(8)] -0.0125 .52548 1 0 0 Minner Mrotshowa
p
.001 w
0 .10774 m
.00375 .10774 L
s
P
p
.001 w
0 .13758 m
.00375 .13758 L
s
P
p
.001 w
0 .16742 m
.00375 .16742 L
s
P
p
.001 w
0 .19725 m
.00375 .19725 L
s
P
p
.001 w
0 .25693 m
.00375 .25693 L
s
P
p
.001 w
0 .28677 m
.00375 .28677 L
s
P
p
.001 w
0 .31661 m
.00375 .31661 L
s
P
p
.001 w
0 .34645 m
.00375 .34645 L
s
P
p
.001 w
0 .40613 m
.00375 .40613 L
s
P
p
.001 w
0 .43597 m
.00375 .43597 L
s
P
p
.001 w
0 .46581 m
.00375 .46581 L
s
P
p
.001 w
0 .49565 m
.00375 .49565 L
s
P
p
.001 w
0 .04806 m
.00375 .04806 L
s
P
p
.001 w
0 .01822 m
.00375 .01822 L
s
P
p
.001 w
0 .55532 m
.00375 .55532 L
s
P
p
.001 w
0 .58516 m
.00375 .58516 L
s
P
p
.001 w
0 .615 m
.00375 .615 L
s
P
[(percent)] -0.0125 .30902 1 0 90 -1 0 Mouter Mrotshowa
p
.002 w
0 0 m
0 .61803 L
s
P
P
p
[(7/19/1988, 10 knots,   5.7 parms, Delta)] .5 .61803 0 -2 Mshowa
P
p
P
p
p
p
.002 w
.02381 .01492 m
.02629 .0155 L
.02877 .01629 L
.03373 .0185 L
.03869 .02156 L
.04365 .02546 L
.05357 .03583 L
.06349 .04951 L
.08333 .08271 L
.10317 .12264 L
.12302 .16819 L
.14286 .20847 L
.18254 .26653 L
.22222 .31511 L
.2619 .35635 L
.28175 .37195 L
.29167 .37848 L
.30159 .38415 L
.31151 .38873 L
.31647 .3906 L
.32143 .39218 L
.33135 .39449 L
.33631 .39526 L
.34127 .39619 L
.35119 .39885 L
.36111 .40256 L
.37103 .40732 L
.38095 .41314 L
.42063 .44556 L
.44048 .46643 L
.46032 .48752 L
.47024 .49614 L
.48016 .50342 L
.49008 .50937 L
.5 .51397 L
.50496 .51577 L
.50992 .51723 L
.5124 .51783 L
.51488 .51835 L
.51736 .51879 L
.51984 .51914 L
.52232 .5194 L
.52356 .5195 L
.5248 .51958 L
.52604 .51964 L
.52728 .51968 L
.52852 .51969 L
.52976 .51969 L
.531 .51966 L
.53224 .51961 L
Mistroke
.53348 .51954 L
.53472 .51945 L
.5372 .51921 L
.53968 .51888 L
.54464 .51796 L
.5496 .5167 L
.55952 .51316 L
.56944 .50824 L
.57937 .5025 L
.59921 .49245 L
.61905 .48447 L
.62897 .48126 L
.63889 .47857 L
.64881 .47641 L
.65377 .47552 L
.65873 .47476 L
.66369 .47414 L
.66865 .47365 L
.67113 .47345 L
.67361 .47329 L
.67609 .47316 L
.67733 .47311 L
.67857 .47306 L
.67981 .47303 L
.68105 .473 L
.68229 .47298 L
.68353 .47297 L
.68477 .47296 L
.68601 .47297 L
.68725 .47298 L
.68849 .473 L
.68973 .47303 L
.69097 .47307 L
.69345 .47317 L
.69593 .47331 L
.69841 .47348 L
.70337 .47391 L
.70833 .47449 L
.71825 .47603 L
.72817 .47811 L
.7381 .48073 L
.75794 .48758 L
.77778 .49662 L
.81746 .51913 L
.85714 .54138 L
.89683 .56283 L
.93651 .58348 L
.97619 .60332 L
Mfstroke
P
P
p
p
.004 w
.02381 .01472 m
.02629 .01495 L
.02877 .01525 L
.03373 .01607 L
.03869 .01717 L
.04365 .01856 L
.05357 .02218 L
.06349 .02692 L
.08333 .03926 L
.10317 .05447 L
.14286 .09122 L
.18254 .12771 L
.22222 .15995 L
.2619 .18912 L
.30159 .21491 L
.34127 .2368 L
.38095 .25516 L
.42063 .27236 L
.46032 .28991 L
.5 .30753 L
.53968 .32366 L
.57937 .33707 L
.61905 .3474 L
.65873 .35559 L
.69841 .36249 L
.7381 .3688 L
.77778 .37505 L
.81746 .38166 L
.85714 .38872 L
.89683 .39612 L
.93651 .4038 L
.97619 .41168 L
s
P
P
P
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
MathSubEnd
P
% End of sub-graphic
P
P
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{291, 378},
  ImageMargins->{{34, Inherited}, {Inherited, Inherited}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCacheValid->False]}, Open]],

Cell["\<\
Very similar looking results in this case. In other cases, the results may \
differ substantially.\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["Two ways to use the program", "Section",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["Interactive", "Subsection",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "The program was designed to facilitate an interactive examination of how \
well different splining methods fit the term structure. To that end, we have \
built in flexibility as to the number and placement of knot points, the \
functional form to be splined, and the size of the penalty and technique of \
choosing it. The way in which we have stored the output at each stage (in \
self-contained, insulated objects) allows the user to have multiple estimates \
of the term structure available at once, making comparisons easy. For \
example, one can run ",
  StyleBox["DiscountFunction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " with different settings on the same data as we have done above. All of \
this is easily accomplished by giving different objects different names."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["Estimation loop", "Subsection",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Some users may wish to run the same settings on many different days of \
data and save some of the results to a file. Here are some suggestions as to \
how to do that. In a separate subdirectory (for convenience) create your data \
files, one per day. Name them in such a way that they are in calendar order \
(for convenience). Use ",
  StyleBox["SetDirectory",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " to change the directory. Use ",
  StyleBox["FileNames",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " to get a list of the file names. Write a function that (",
  StyleBox["i",
    FontSlant->"Italic"],
  ") estimates the term structure according to the settings you have chosen \
and (",
  StyleBox["ii",
    FontSlant->"Italic"],
  ") extracts the information you want and appends it to a file. Map that \
function onto the list of filenames."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
Here is an example of such a function. It takes two arguments, an input file \
name (where to find the data) and an output file name (where to append the \
results).\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
EstimationLoop[in_String, out_String] :=
  Module[{md, cp, df, qd, zc, fc, zeros, forwards},
  (* estimate the term structure *)
  md = ReadData[in]; (* or use a version of MassageData *)
  cp = ConstructPayments[md];
  df = DiscountFunction[cp]; (* use your settings *)
  (* extract the zero and forward curves *)
  zc = ZeroCurve[df];
  fc = ForwardCurve[df];
  (* read off zero and forward rates every 6 months *)
  zeros = Table[{m, zc[m Years]}, {m, .5, 30, .5}];
  forwards = Table[{m, fc[m Years]}, {m, 0, 30, .5}];
  (* get the instantaneous zero rate from the forward curve *)
  PrependTo[zeros, First @ forwards];
  (* prepare data for file *)
  qd = DateToString @ QuoteDate[md];
  zeros = Join[{qd, \"z\"}, #]& /@ zeros;
  forwards = Join[{qd, \"f\"}, #]& /@ forwards;
  (* append data to output file *)
  AppendDelimited[out, #]& /@ {zeros, forwards}]\
\>", "Input",
  Evaluatable->False,
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Here is the usage. ",
  StyleBox["EsitmationLoop[#, \"output.dat\"]& /@ fn",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", where ",
  StyleBox["fn",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is the list of input data file names. A typical row in ",
  StyleBox["output.dat",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " will look like:"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  StyleBox[" \"",
    FontFamily->"courier"],
  StyleBox["9/29/1993\",\"z\",7.5,5.21428",
    FontWeight->"Plain"]
}], "Input",
  Editable->False,
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "This file can be read back into ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " using"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[StyleBox["data = ToExpression @ ReadDelimited[\"output.dat\"];",
  
  FontFamily->"courier"]], "Input",
  Evaluatable->False,
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "The dates can be converted to ",
  StyleBox["{yyyy, mm, dd}",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " format with ",
  StyleBox["StringToDate",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". The file can also be searched using ",
  StyleBox["FindList",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". (",
  StyleBox["StringToDate",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["DateToString",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["AppendDelimited",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["WriteDelimited",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " are defined the package ",
  StyleBox["CommaDelimited",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", which is distributed with ",
  StyleBox["YieldCurve",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ".) "
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]]}, Open]],

Cell[CellGroupData[{Cell[TextData[{
  "The structure of ",
  StyleBox["MD",
    FontFamily->"Courier"],
  ", ",
  StyleBox["CP",
    FontFamily->"Courier"],
  ", and ",
  StyleBox["DF",
    FontFamily->"Courier"],
  " objects"
}], "Section",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "In this section we describe the structure of ",
  StyleBox["MD",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["CP",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", and ",
  StyleBox["DF",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " objects. These objects are merely ``wrappers'' that contain (and hide) \
the data and results. This section will serve two purposes. First, it will \
help those who wish to write their own version of ",
  StyleBox["ReadData",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " (or its older cousin, ",
  StyleBox["MassageData",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "), and second it describes the overall structure of the ",
  StyleBox["YieldCurve",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " package."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell[TextData[{
  StyleBox["MD",
    FontFamily->"Courier"],
  " objects"
}], "Subsection",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
The structure of the flat file can be seen as follows. We select a few rows \
from the file:\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
pos = {1, 30, 60, -1};
ReadList[\"19880719.dat\", String][[ pos ]]\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
{\"7/19/1988,912794PY4,8/4/1988,0.5,0,99.7321\", \"7/19/1988,9\\
 
   12794SC9,6/8/1989,0.5,0,93.5669\", \"7/19/1988,912810CM8,2\\
 
   /15/2005,25,11.75,118.719\", \"7/19/1988,912827WK4,7/15/19\\
 
   95,7,8.875,99.625\"}\
\>", "\<\
{7/19/1988,912794PY4,8/4/1988,0.5,0,99.7321, 
 
  7/19/1988,912794SC9,6/8/1989,0.5,0,93.5669, 
 
  7/19/1988,912810CM8,2/15/2005,25,11.75,118.719, 
 
  7/19/1988,912827WK4,7/15/1995,7,8.875,99.625}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  "The fields in the file are quote date, identification number, maturity \
date, original term to maturity (in years), coupon rate, and price per \
hundred dollars of face value (in decimal form). The price does not include \
accrued interest for coupon bonds. Fields are separated by commas  with no \
intervening spaces. The identification number and the term are both required, \
but neither need be meaningful nor unique. (In fact, in the data above, the \
term for all bills is set to 0.5, which is not true since some bills have an \
original term to maturity of 1 year.) The identification number is useful for \
",
  StyleBox["IDNumberDropList",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and the term is useful for ",
  StyleBox["DropByTerm",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", both options of ",
  StyleBox["DiscountFunction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "We can easily create a mini-",
  StyleBox["MD",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object of these five securities to see what it looks like internally:"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["?MDSubset", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
MDSubset[md, poslist] creates an MD object that contains a
   subset of bonds from another MD object according to the
   list of positions given.\
\>", "Print",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["\<\
mdsub = MDSubset[md, pos];
InputForm[mdsub]\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
InputForm[MD[{1988, 7, 19}, 
 
   {{1988, 8, 4}, {1989, 6, 8}, {2005, 2, 15}, 
 
    {1995, 7, 15}}, {0.5, 0.5, 25, 7}, 
 
   {0, 0, 11.75, 8.875}, 
 
   {99.7321, 93.5669, 118.719, 99.625}, 
 
   {\"912794PY4\", \"912794SC9\", \"912810CM8\", \"912827WK4\"}, {}]
 
   ]\
\>", "\<\
MD[{1988, 7, 19}, {{1988, 8, 4}, {1989, 6, 8}, 
   {2005, 2, 15}, {1995, 7, 15}}, {0.5, 0.5, 25, 7}, 
  {0, 0, 11.75, 8.875}, {99.7321, 93.5669, 118.719, 
   99.625}, {\"912794PY4\", \"912794SC9\", \"912810CM8\", \"91282\\
    7WK4\"}, {}]\
\>"], "Output",
  LineSpacing->{1, 0},
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  "The ",
  StyleBox["MD",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object contains seven items, each of which is accessible via an ",
  StyleBox["MDSelector",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ": ",
  StyleBox["QuoteDate",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["MaturityDate",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Term",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["CouponRate",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Price",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["IDNumber",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", and ",
  StyleBox["Description",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". An additional ",
  StyleBox["MDSelector",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["SettlementDate",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is not contained in the ",
  StyleBox["MD",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object; instead, it is calculated on demand (see below). "
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "The user may write a different ",
  StyleBox["ReadData",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " function to read data in a different form. What is required for the \
program to work properly is to have the new ",
  StyleBox["ReadData",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " production an ",
  StyleBox["MD",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object with the correct internal structure. Here is the relevant ",
  StyleBox["Mathematica",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " code the sets up the ",
  StyleBox["MDSelectors",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and formats the ",
  StyleBox["MD",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " objects:"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
MD /: QuoteDate[md_MD]      := md[[1]]
MD /: MaturityDate[md_MD]   := md[[2]]
MD /: Term[md_MD]           := md[[3]]
MD /: CouponRate[md_MD]     := md[[4]]
MD /: Price[md_MD]          := md[[5]]
MD /: IDNumber[md_MD]          := md[[6]]
MD /: Description[md_MD]    := md[[7]]
MD /: NumberOfSecurities[md_MD] := Length[Price[md]]
MD /: SettlementDate[md_MD] := FirstBusinessDay @ 
     DaysPlus[QuoteDate[md], BusinessDaysToSettlement]
MD /: WhenceMD[md_MD]       := md

Format[md_MD] := StringForm[\"-MD[`1`]-\", QuoteDate[md]]

MDSelectors = Sort @ {QuoteDate, MaturityDate, Term, 
\tCouponRate, Price, IDNumber, Description, 
\tSettlementDate}\
\>", "Input",
  Evaluatable->False,
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Here is an earlier version of ",
  StyleBox["ReadData",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["MassageData",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", that uses symbolic switches to identify the structure of the data file."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  StyleBox["MassageData[OpenMarket]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " reads the file ",
  StyleBox["openmkt.dat",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". That file has a different structure. In particular, security price \
quotes are given; bills are quoted as banker's discounts and coupon \
securities are quoted in 32nds."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["First @ ReadList[\"openmkt.dat\",String]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
\"{1993,10,27} 912810DP0 {2015,2,15} 11.250 10957 159.13\"\
\>", "\<\
{1993,10,27} 912810DP0 {2015,2,15} 11.250 10957 159.13\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["mdopen = MassageData[OpenMarket]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
MD[{1993, 10, 27}, {{2015, 2, 15}, {2005, 5, 15}, 
 
   {2005, 8, 15}, {2015, 8, 15}, {2015, 11, 15}, 
 
   {2006, 2, 15}, {2016, 2, 15}, {2016, 5, 15}, 
 
   {2016, 11, 15}, {2017, 5, 15}, {2017, 8, 15}, 
 
   {2018, 5, 15}, {2018, 11, 15}, {2019, 2, 15}, 
 
   {2019, 8, 15}, {2020, 2, 15}, {2020, 5, 15}, 
 
   {2020, 8, 15}, {2021, 2, 15}, {2021, 5, 15}, 
 
   {2021, 8, 15}, {2021, 11, 15}, {2022, 8, 15}, 
 
   {2022, 11, 15}, {2023, 2, 15}, {2023, 8, 15}, 
 
   {1996, 3, 31}, {1998, 4, 15}, {1996, 4, 30}, 
 
   {1994, 5, 15}, {2001, 5, 15}, {1996, 5, 31}, 
 
   {1996, 6, 30}, {1998, 7, 15}, {1996, 7, 31}, 
 
   {1994, 8, 15}, {2001, 8, 15}, {1996, 8, 31}, 
 
   {1996, 9, 30}, {1998, 10, 15}, {1996, 10, 31}, 
 
   {1994, 11, 15}, {2001, 11, 15}, {1993, 11, 30}, 
 
   {1996, 11, 30}, {1993, 12, 31}, {1996, 12, 31}, 
 
   {1999, 1, 15}, {1994, 1, 31}, {1997, 1, 31}, 
 
   {1995, 2, 15}, {1994, 2, 28}, {1997, 2, 28}, 
 
   {1994, 3, 31}, {1997, 3, 31}, {1999, 4, 15}, 
 
   {1994, 4, 30}, {1997, 4, 30}, {1995, 5, 15}, 
 
   {2002, 5, 15}, {1994, 5, 31}, {1997, 5, 31}, 
 
   {1994, 6, 30}, {1997, 6, 30}, {1999, 7, 15}, 
 
   {1994, 7, 31}, {1997, 7, 31}, {1995, 8, 15}, 
 
   {2002, 8, 15}, {1994, 8, 31}, {1997, 8, 31}, 
 
   {1994, 9, 30}, {1997, 9, 30}, {1999, 10, 15}, 
 
   {1994, 10, 31}, {1997, 10, 31}, {1995, 11, 15}, 
 
   {1994, 11, 30}, {1997, 11, 30}, {1994, 12, 31}, 
 
   {1997, 12, 31}, {2000, 1, 15}, {1998, 1, 31}, 
 
   {1995, 1, 31}, {1996, 2, 15}, {2003, 2, 15}, 
 
   {1995, 2, 28}, {1998, 2, 28}, {1995, 3, 31}, 
 
   {1998, 3, 31}, {2000, 4, 15}, {1995, 4, 30}, 
 
   {1998, 4, 30}, {1996, 5, 15}, {1995, 5, 31}, 
 
   {1998, 5, 31}, {1995, 6, 30}, {1998, 6, 30}, 
 
   {1995, 7, 31}, {1998, 7, 31}, {1996, 8, 15}, 
 
   {2003, 8, 15}, {1995, 8, 31}, {1998, 8, 31}, 
 
   {1995, 9, 30}, {1998, 9, 30}, {1995, 2, 15}, 
 
   {1995, 5, 15}, {1995, 8, 15}, {1995, 11, 15}, 
 
   {1996, 2, 15}, {1996, 5, 15}, {1996, 11, 15}, 
 
   {1994, 1, 15}, {1994, 4, 15}, {1997, 5, 15}, 
 
   {1994, 7, 15}, {1997, 8, 15}, {1994, 10, 15}, 
 
   {1997, 11, 15}, {1995, 1, 15}, {1998, 2, 15}, 
 
   {1995, 4, 15}, {1998, 5, 15}, {1995, 7, 15}, 
 
   {1998, 8, 15}, {1995, 10, 15}, {1998, 11, 15}, 
 
   {1994, 2, 15}, {1996, 1, 15}, {1999, 2, 15}, 
 
   {1994, 5, 15}, {1996, 4, 15}, {1999, 5, 15}, 
 
   {1994, 8, 15}, {1996, 7, 15}, {1999, 8, 15}, 
 
   {1994, 11, 15}, {1996, 10, 15}, {1999, 11, 15}, 
 
   {1995, 2, 15}, {1993, 12, 31}, {1997, 1, 15}, 
 
   {2000, 2, 15}, {1995, 5, 15}, {1994, 3, 31}, 
 
   {1997, 4, 15}, {2000, 5, 15}, {1995, 8, 15}, 
 
   {1994, 6, 30}, {1997, 7, 15}, {2000, 8, 15}, 
 
   {1995, 11, 15}, {1994, 9, 30}, {1997, 10, 15}, 
 
   {2000, 11, 15}, {1996, 2, 15}, {1994, 12, 31}, 
 
   {1998, 1, 15}, {1996, 1, 31}, {1994, 2, 15}, 
 
   {2001, 2, 15}, {1996, 2, 29}}, 
 
  {30, 20, 20, 30, 30, 20, 30, 30, 30, 30, 30, 30, 30, 30, 
 
   30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 5, 7, 5, 
 
   3, 10, 5, 5, 7, 5, 3, 10, 5, 5, 7, 5, 3, 10, 2, 5, 2, 5, 
 
   7, 2, 5, 3, 2, 5, 2, 5, 7, 2, 5, 3, 10, 2, 5, 2, 5, 7, 
 
   2, 5, 3, 10, 2, 5, 2, 5, 7, 2, 5, 3, 2, 5, 2, 5, 7, 5, 
 
   2, 3, 10, 2, 5, 2, 5, 7, 2, 5, 3, 2, 5, 2, 5, 2, 5, 3, 
 
   10, 2, 5, 2, 5, 10, 10, 10, 10, 10, 10, 10, 7, 7, 10, 7, 
 
   10, 7, 10, 7, 10, 7, 10, 7, 10, 7, 10, 5, 7, 10, 5, 7, 
 
   10, 5, 7, 10, 5, 7, 10, 5, 4, 7, 10, 5, 4, 7, 10, 5, 4, 
 
   7, 10, 5, 4, 7, 10, 5, 4, 7, 5, 3, 10, 5}, 
 
  {11.25, 12., 10.75, 10.625, 9.875, 9.375, 9.25, 7.25, 
 
   7.5, 8.75, 8.875, 9.125, 9., 8.875, 8.125, 8.5, 8.75, 
 
   8.75, 7.875, 8.125, 8.125, 8., 7.25, 7.625, 7.125, 6.25, 
 
   7.75, 7.875, 7.625, 7., 8., 7.625, 7.875, 8.25, 7.875, 
 
   6.875, 7.875, 7.25, 7., 7.125, 6.875, 6., 7.5, 5.5, 6.5, 
 
   5., 6.125, 6.375, 4.875, 6.25, 5.5, 5.375, 6.75, 5.75, 
 
   6.875, 7., 5.375, 6.875, 5.875, 7.5, 5.125, 6.75, 5., 
 
   6.375, 6.375, 4.25, 5.5, 4.625, 6.375, 4.25, 5.625, 4., 
 
   5.5, 6., 4.25, 5.75, 5.125, 4.625, 6., 4.625, 6., 6.375, 
 
   5.625, 4.25, 4.625, 6.25, 3.875, 5.125, 3.875, 5.125, 
 
   5.5, 3.875, 5.125, 4.25, 4.125, 5.375, 4.125, 5.125, 
 
   4.25, 5.25, 4.375, 5.75, 3.875, 4.75, 3.875, 4.75, 
 
   11.25, 11.25, 10.5, 9.5, 8.875, 7.375, 7.25, 7., 7., 
 
   8.5, 8., 8.625, 9.5, 8.875, 8.625, 8.125, 8.375, 9., 
 
   8.875, 9.25, 8.625, 8.875, 8.875, 9.25, 8.875, 9.5, 
 
   9.375, 9.125, 8.625, 7.875, 8., 8.25, 8., 7.875, 7.75, 
 
   7.625, 8., 8.5, 8.5, 8.5, 8.5, 8.875, 8.5, 8.5, 8.5, 
 
   8.75, 8.5, 8.5, 8.75, 8.5, 7.875, 7.625, 7.875, 7.5, 
 
   6.875, 7.75, 7.5}, {159.40625, 153.9375, 143.78125, 
 
   152.3125, 143.46875, 132.65625, 136., 111.9375, 
 
   114.96875, 130.53125, 132.15625, 135.71875, 134.40625, 
 
   132.84375, 123.375, 128.4374999999999, 131.875, 
 
   131.90625, 120.71875, 124.1875, 124.21875, 122.78125, 
 
   113.25, 118.4375, 112.34375, 103.46875, 108.28125, 
 
   112.5625, 108.1875, 101.96875, 116.96875, 108.40625, 
 
   109.59375, 114.5, 109.375, 102.65625, 116.5, 107.90625, 
 
   107.53125, 110.09375, 107.21875, 102.5, 114.25, 
 
   100.1875, 106.25, 100.3125, 105.28125, 106.84375, 
 
   100.40625, 105.65625, 102.25, 100.71875, 107.25, 
 
   101.03125, 107.78125, 109.875, 101.03125, 107.8125, 
 
   103.09375, 114.65625, 101.03125, 107.46875, 101.0625, 
 
   106.375, 106.875, 100.59375, 103.4375, 101.28125, 
 
   106.84375, 100.65625, 103.78125, 100.46875, 103.375, 
 
   105.03125, 100.71875, 104.1875, 102.25, 101.125, 
 
   105.125, 101.1875, 105.15625, 106.9375, 103.65625, 
 
   100.71875, 101.1875, 105.78125, 100.21875, 101.71875, 
 
   100.1875, 101.65625, 102.71875, 100.125, 101.59375, 
 
   100.25, 100.4375, 102.5625, 100.4375, 101.53125, 
 
   100.625, 102., 100.40625, 102.46875, 99.90625, 99.84375, 
 
   99.875, 99.8125, 109.5, 111.0625, 111.375, 110.84375, 
 
   110.375, 107.65625, 108.3125, 100.75, 101.6875, 
 
   113.0625, 103.1875, 114.0625, 105.59375, 115.5625, 
 
   105.8125, 113.28125, 106.46875, 117.34375, 108.21875, 
 
   118.96875, 108.75, 117.9375, 101.65625, 110.90625, 
 
   118.4375, 103.3125, 112.15625, 120.21875, 104.0625, 
 
   109.28125, 115.0625, 104.78125, 110.1875, 114.6875, 
 
   105.09375, 100.75, 110.75, 118.3125, 107., 102.1875, 
 
   112.875, 120.75, 107.90625, 103.375, 113.4375, 
 
   120.46875, 108.8125, 104.53125, 114.90625, 119.3125, 
 
   108.1875, 104.59375, 112.15625, 107.3125, 101.0625, 
 
   115.1875, 107.5}, {\"912810DP0\", \"912810DQ8\", \"912810DR\\
 
    6\", \"912810DS4\", \"912810DT2\", \"912810DU9\", \"912810DV7\"\\
 
    , \"912810DW5\", \"912810DX3\", \"912810DY1\", \"912810DZ8\"\\
 
    , \"912810EA2\", \"912810EB0\", \"912810EC8\", \"912810ED6\"\\
 
    , \"912810EE4\", \"912810EF1\", \"912810EG9\", \"912810EH7\"\\
 
    , \"912810EJ3\", \"912810EK0\", \"912810EL8\", \"912810EM6\"\\
 
    , \"912810EN4\", \"912810EP9\", \"912810EQ7\", \"912827A36\"\\
 
    , \"912827A44\", \"912827A69\", \"912827A77\", \"912827A85\"\\
 
    , \"912827B27\", \"912827B43\", \"912827B50\", \"912827B76\"\\
 
    , \"912827B84\", \"912827B92\", \"912827C34\", \"912827C59\"\\
 
    , \"912827C67\", \"912827C83\", \"912827C91\", \"912827D25\"\\
 
    , \"912827D33\", \"912827D41\", \"912827D58\", \"912827D66\"\\
 
    , \"912827D74\", \"912827D82\", \"912827D90\", \"912827E24\"\\
 
    , \"912827E40\", \"912827E57\", \"912827E65\", \"912827E73\"\\
 
    , \"912827E81\", \"912827E99\", \"912827F23\", \"912827F31\"\\
 
    , \"912827F49\", \"912827F56\", \"912827F64\", \"912827F72\"\\
 
    , \"912827F80\", \"912827F98\", \"912827G22\", \"912827G30\"\\
 
    , \"912827G48\", \"912827G55\", \"912827G63\", \"912827G71\"\\
 
    , \"912827G89\", \"912827G97\", \"912827H21\", \"912827H39\"\\
 
    , \"912827H47\", \"912827H54\", \"912827H70\", \"912827H88\"\\
 
    , \"912827H96\", \"912827J29\", \"912827J37\", \"912827J45\"\\
 
    , \"912827J52\", \"912827J60\", \"912827J78\", \"912827J86\"\\
 
    , \"912827J94\", \"912827K27\", \"912827K35\", \"912827K43\"\\
 
    , \"912827K50\", \"912827K68\", \"912827K76\", \"912827K92\"\\
 
    , \"912827L26\", \"912827L34\", \"912827L42\", \"912827L59\"\\
 
    , \"912827L67\", \"912827L75\", \"912827L83\", \"912827L91\"\\
 
    , \"912827M25\", \"912827M33\", \"912827M41\", \"912827RW4\"\\
 
    , \"912827SE3\", \"912827SP8\", \"912827SY9\", \"912827TF9\"\\
 
    , \"912827TQ5\", \"912827UF7\", \"912827UL4\", \"912827UT7\"\\
 
    , \"912827UW0\", \"912827VB5\", \"912827VE9\", \"912827VK5\"\\
 
    , \"912827VN9\", \"912827VT6\", \"912827VW9\", \"912827WB4\"\\
 
    , \"912827WE8\", \"912827WK4\", \"912827WN8\", \"912827WT5\"\\
 
    , \"912827WW8\", \"912827WY4\", \"912827XB3\", \"912827XE7\"\\
 
    , \"912827XG2\", \"912827XK3\", \"912827XN7\", \"912827XQ0\"\\
 
    , \"912827XT4\", \"912827XW7\", \"912827XY3\", \"912827YB2\"\\
 
    , \"912827YE6\", \"912827YG1\", \"912827YJ5\", \"912827YK2\"\\
 
    , \"912827YN6\", \"912827YQ9\", \"912827YS5\", \"912827YT3\"\\
 
    , \"912827YW6\", \"912827YY2\", \"912827ZA3\", \"912827ZB1\"\\
 
    , \"912827ZE5\", \"912827ZG0\", \"912827ZJ4\", \"912827ZK1\"\\
 
    , \"912827ZN5\", \"912827ZQ8\", \"912827ZS4\", \"912827ZT2\"\\
 
    , \"912827ZV7\", \"912827ZW5\", \"912827ZX3\", \"912827ZZ8\"}, 
 
  {}]\
\>", "\<\
-MD[{1993, 10, 27}]-\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  StyleBox["MassageData[Strips]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " looks for the flat file ",
  StyleBox["strips.dat",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " in the working directory and assumes it has the following structure:"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["Take[ReadList[\"strips.dat\", String], 3]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
{\"9/30/1994,11/15/1994,99.46875\", \"9/30/1994,11/15/1994,99.\\
 
   46875\", \"9/30/1994,2/15/1995,98.09375\"}\
\>", "\<\
{9/30/1994,11/15/1994,99.46875, 
 
  9/30/1994,11/15/1994,99.46875, 
 
  9/30/1994,2/15/1995,98.09375}\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  "Note the following aspects of structure of the data file. There are only \
three fields per record: quote date, maturity date, and price as a decimal. \
The missing information, which is not necessary to run ",
  StyleBox["DiscountFunction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", is simply ``made up'' by ",
  StyleBox["MassageData",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ".\nIn this case, the terms-to-maturity are arbitrarily set to ",
  StyleBox["-1",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and the IDNumbers are set to ",
  StyleBox["\"N/A\"",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". The ",
  StyleBox["Description",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is set to ",
  StyleBox["{\"Strips\"}",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". The term of ",
  StyleBox["-1",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is used by ",
  StyleBox["ConstructPayments",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " to distinguish strips from bills for the purposes of the Security \
Industry Associations standard securities yield calculations, which treat \
bills and strips separately.\\footnote{$^7$}{Bills and strips are treated the \
same when yields are calculated continuously.}"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "If one wanted to distinguish between principal and coupon STRIPs, one \
could add a fourth column to ",
  StyleBox["strips.dat",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ": an ",
  StyleBox["IDNumber",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " that took on two values, say ",
  StyleBox["\"P\"",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " for principal and ",
  StyleBox["\"C\"",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " for coupon. One would write a new version of ",
  StyleBox["MassageData",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " (or ",
  StyleBox["ReadData",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ") to put this ",
  StyleBox["IDNumber",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " in the proper place in the ",
  StyleBox["MD",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object. Then one could use the option ",
  StyleBox["IDNumberDropList -> \"P\"",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " to specify that only principal should be used in the estimation."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Note that STRIPs data can be read correctly with ",
  StyleBox["ReadData",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", and STRIPs data can be mixed with other security types as well. "
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell[TextData[{
  StyleBox["CP",
    FontFamily->"Courier"],
  " objects"
}], "Subsection",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Here is the ",
  StyleBox["InputForm",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " of the ",
  StyleBox["CP",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object created by ",
  StyleBox["ConstructPayments ",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "from ",
  StyleBox["mdsub",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ":"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
cpsub = ConstructPayments[mdsub];
InputForm[cpsub]\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
InputForm[CP[{1988, 7, 19}, 
 
   {{100}, {100}, {5.875, 5.875, 5.875, 5.875, 5.875, 
 
     5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
     5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
     5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
     5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
 
     105.875}, {4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
     4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
 
     4.4375, 4.4375, 104.4375}}, 
 
   {{15}, {323}, {26, 210, 391, 575, 756, 940, 1121, 1308, 
 
     1489, 1672, 1853, 2036, 2217, 2401, 2582, 2766, 2948, 
 
     3135, 3313, 3499, 3680, 3863, 4044, 4227, 4409, 4593, 
 
     4774, 4958, 5139, 5326, 5504, 5690, 5871, 6054}, 
 
    {181, 362, 545, 726, 909, 1090, 1274, 1456, 1640, 1821, 
 
     2008, 2186, 2372, 2553}}, {0, 0, 155, 5}, 
 
   {2, 2, 2, 2}, False, Hold[mdsub]]]\
\>", "\<\
CP[{1988, 7, 19}, {{100}, {100}, 
   {5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
    5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 5.875, 
    5.875, 5.875, 5.875, 5.875, 5.875, 105.875}, 
   {4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
    4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 4.4375, 
    4.4375, 104.4375}}, {{15}, {323}, 
   {26, 210, 391, 575, 756, 940, 1121, 1308, 1489, 1672, 
    1853, 2036, 2217, 2401, 2582, 2766, 2948, 3135, 3313, 
    3499, 3680, 3863, 4044, 4227, 4409, 4593, 4774, 4958, 
    5139, 5326, 5504, 5690, 5871, 6054}, 
   {181, 362, 545, 726, 909, 1090, 1274, 1456, 1640, 1821, 
    2008, 2186, 2372, 2553}}, {0, 0, 155, 5}, {2, 2, 2, 2}, 
  False, Hold[mdsub]]\
\>"], "Output",
  LineSpacing->{1, 0},
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  "There are seven items contained in the object, all of which can be \
extracted (",
  StyleBox["i.e.",
    FontSlant->"Italic"],
  ", selected) by the ",
  StyleBox["CPSelectors",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ": ",
  StyleBox["QuoteDate",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Payments",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["DaysToPayments",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["AccruedDays",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["PaymentsPerYear",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["FullPrice",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", and ",
  StyleBox["WhenceMD",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". The other ",
  StyleBox["CPSelectors ",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "calculate what they return on demand."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell[TextData[{
  StyleBox["DF",
    FontFamily->"Courier"],
  " objects"
}], "Subsection",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Here is the ",
  StyleBox["InputForm",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " of the ",
  StyleBox["DF",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object created by ",
  StyleBox["DiscountFunction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " from the ",
  StyleBox["CP",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " object given above:"
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[CellGroupData[{Cell["\<\
dfsub = DiscountFunction[cpsub, MaximumMaturity -> 16.6, 
\tKnots -> 0, FunctionalForm -> Delta, LambdaValue -> 0]\
\>", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
DF[{1988, 7, 19}, InterpolatingFunction[{0, 6064}, 
 
   {{0, 0, {1, 0}, {0}}, 
 
    {6064, 0, {0.2117929536117513, -0.00001535208454085584, 
 
      1.890324679160249*10^-8, 1.195289841813016*10^-12}, 
 
     {0, 0, 6064}}}], 0, 3, Delta, 
 
  {0.03298646745656697, -0.001870960109926045, 
 
   -(2.131225215862287*10^-6), 0.00003979571182810559}, 
 
  {1, 2, 3, 4}, 0.001091609115236887, 0.00872483862588425, 
 
  {0.996796948422626, -0.04007415733109442, 
 
   -0.00004561217393878103, 0.000851702378932987}, 
 
  {0, 6064}, {{0, 0, 0, 0}, 
 
   {0, 0.005933603944256982, -0.01007417955412868, 
 
    0.001370051508054565}, 
 
   {0, -0.01007417955412868, 0.01787216815776105, 
 
    -0.002587323349510649}, 
 
   {0, 0.001370051508054565, -0.002587323349510649, 
 
    0.0004489482759882174}}, 
 
  {1, 0.5944046122549978, 0.2428246338303346, 
 
   0.2117929536117513}, True, 0, True, 2, Automatic, 3, 
 
  0.001091609116883774, Hold[cpsub]]\
\>", "\<\
-DF[{1988, 7, 19}]-\
\>"], "Output",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[CellGroupData[{Cell["InputForm[dfsub]", "Input",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[OutputFormData["\<\
InputForm[DF[{1988, 7, 19}, 
 
   InterpolatingFunction[{0, 6064}, 
 
    {{0, 0, {1, 0}, {0}}, 
 
     {6064, 0, {0.2117929536117513, 
 
       -0.00001535208454085584, 1.890324679160249*10^-8, 
 
       1.195289841813016*10^-12}, {0, 0, 6064}}}], 0, 3, 
 
   Delta, {0.03298646745656697, -0.001870960109926045, 
 
    -(2.131225215862287*10^-6), 0.00003979571182810559}, 
 
   {1, 2, 3, 4}, 0.001091609115236887, 0.00872483862588425, 
 
   {0.996796948422626, -0.04007415733109442, 
 
    -0.00004561217393878103, 0.000851702378932987}, 
 
   {0, 6064}, {{0, 0, 0, 0}, 
 
    {0, 0.005933603944256982, -0.01007417955412868, 
 
     0.001370051508054565}, 
 
    {0, -0.01007417955412868, 0.01787216815776105, 
 
     -0.002587323349510649}, 
 
    {0, 0.001370051508054565, -0.002587323349510649, 
 
     0.0004489482759882174}}, 
 
   {1, 0.5944046122549978, 0.2428246338303346, 
 
    0.2117929536117513}, True, 0, True, 2, Automatic, 3, 
 
   0.001091609116883774, Hold[cpsub]]]\
\>", "\<\
DF[{1988, 7, 19}, InterpolatingFunction[{0, 6064}, 
   {{0, 0, {1, 0}, {0}}, 
    {6064, 0, {0.2117929536117513, 
      -0.00001535208454085584, 1.890324679160249*10^-8, 
      1.195289841813016*10^-12}, {0, 0, 6064}}}], 0, 3, 
  Delta, {0.03298646745656697, -0.001870960109926045, 
   -(2.131225215862287*10^-6), 0.00003979571182810559}, 
  {1, 2, 3, 4}, 0.001091609115236887, 0.00872483862588425, 
  {0.996796948422626, -0.04007415733109442, 
   -0.00004561217393878103, 0.000851702378932987}, 
  {0, 6064}, {{0, 0, 0, 0}, 
   {0, 0.005933603944256982, -0.01007417955412868, 
    0.001370051508054565}, 
   {0, -0.01007417955412868, 0.01787216815776105, 
    -0.002587323349510649}, 
   {0, 0.001370051508054565, -0.002587323349510649, 
    0.0004489482759882174}}, 
  {1, 0.5944046122549978, 0.2428246338303346, 
   0.2117929536117513}, True, 0, True, 2, Automatic, 3, 
  0.001091609116883774, Hold[cpsub]]\
\>"], "Output",
  LineSpacing->{1, 0},
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]],

Cell[TextData[{
  "There are nineteen items contained in the object, all of which can be \
extracted (",
  StyleBox["i.e. ",
    FontSlant->"Italic"],
  "selected) by the ",
  StyleBox["DFSelectors",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ":  ",
  StyleBox["QuoteDate",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["FinalSpline",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Lambda",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["EffectiveParameters",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["FunctionalForm",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["PriceError",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["KeepList",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["ResidualVariance",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["AverageAbsoluteError",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Significance",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Knots",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["BetaHatCovarianceMatrix",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["BetaHat",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Restriction",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["DropByTerm",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["FixedLambda",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["TuningParameter",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["ObservationWeights",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["SplineOrder",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", ",
  StyleBox["GCVValue",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", and ",
  StyleBox["WhenceCP",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". The remaining ",
  StyleBox["DFSelectors ",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "calculate what they return on demand. "
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]]}, Open]],

Cell[CellGroupData[{Cell["References", "Section",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Adams, K. J., and D. R. Van Deventer. ``Fitting Yield Curves and Forward \
Rate Curves with Maximum Smoothness.'' ",
  StyleBox["Journal of Fixed Income",
    FontSlant->"Italic"],
  ", June (1994): 52--62."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
Bliss, R. R., Jr. ``Testing Term Structure Estimation Methods.'' Working \
Paper, Indiana University (1993).\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Chambers, D., W. Carleton and D. Waldman. ``A New Approach to Estimation \
of Term Structure of Interest Rates.'' ",
  StyleBox["Journal of Financial and Quantitative Studies",
    FontSlant->"Italic"],
  ", 19 No. 3 (1984): 233--252. "
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Chow, G. C. ",
  StyleBox["Econometrics",
    FontSlant->"Italic"],
  " McGraw-Hill, New York (1983)."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Coleman, T. S., L. Fisher and R. Ibbotson. ``Estimating the Term Structure \
of Interest From Data That Include the Prices of Coupon Bonds.'' ",
  StyleBox["The Journal of Fixed Income",
    FontSlant->"Italic"],
  ", September (1992): 85--116."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "de Boor, C., ",
  StyleBox["A Practical Guide to Splines",
    FontSlant->"Italic"],
  ", Springer-Verlag (1978)."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
Fisher, M.,  D. Nychka, and D. Zervos. ``Fitting the Term Structure of \
Interest Rates with Smoothing Splines,'' Finance and Economics Discussion \
Series, 95-1, Federal Reserve Board (1995). \
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell["\<\
Gilles, C. ``Forward Rates and Expected Future Short Rates.'' Working paper, \
Federal Reserve Board (1994).\
\>", "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Jordan, J.V. ``Tax Effects in Term Structure Estimation.'' ",
  StyleBox["Journal of Finance",
    FontSlant->"Italic"],
  ", 39 No. 2 (1984): 393--406."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "McCulloch, J. H. ``Measuring the Term Structure of Interest Rates.' '",
  StyleBox["Journal of Business",
    FontSlant->"Italic"],
  ",  44 (1971): 19--31."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "McCulloch, J. H. ``The Tax-Adjusted Yield Curve.''  ",
  StyleBox["Journal of Finance",
    FontSlant->"Italic"],
  ", 30 (1975): 811--830."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Nelson, C. and A. Siegel. ``Parsimonious Modeling of Yield Curves.'' ",
  StyleBox["Journal of Business",
    FontSlant->"Italic"],
  ", 60 No. 4 (1987): 473--489."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Shea, G. ``Pitfalls in Smoothing Interest Rate Term Structure Data: \
Equilibrium Models and Spline Approximations.'' ",
  StyleBox["Journal of Financial and Quantitative Studies",
    FontSlant->"Italic"],
  ", 19 No. 3 (1984): 253--269."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Shea, G. ``Interest Rate Term Structure Estimation with Exponential \
Splines: A Note.'' ",
  StyleBox["Journal of Finance",
    FontSlant->"Italic"],
  ", 40 No. 1 (1985): 319--325."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Vasicek, O. and G. Fong. ``Term Structure Estimation Using Exponential \
Splines.'' ",
  StyleBox["Journal of Finance",
    FontSlant->"Italic"],
  ", 38 (1982): 339--348."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}],

Cell[TextData[{
  "Wahba, G. ",
  StyleBox["Spline Models for Observational Data",
    FontSlant->"Italic"],
  " SIAM: Philadelphia (1990)."
}], "Text",
  ImageRegion->{{0, 1}, {0, 1}}]}, Open]]}, Open]]
},
FrontEndVersion->"4.1 for Microsoft Windows",
ScreenRectangle->{{0, 1280}, {0, 979}},
WindowToolbars->{},
CellGrouping->Manual,
WindowSize->{496, 768},
WindowMargins->{{91, Automatic}, {69, Automatic}},
PrivateNotebookOptions->{"ColorPalette"->{RGBColor, -1}},
ShowCellLabel->True,
ShowCellTags->False,
RenderingOptions->{"ObjectDithering"->True,
"RasterDithering"->False},
Magnification->1.5
]


(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1727, 52, 60, 1, 70, "Title"],
Cell[1790, 55, 211, 7, 70, "Subtitle"],
Cell[2004, 64, 84, 1, 70, "Subsubtitle"],

Cell[CellGroupData[{
Cell[2111, 67, 96, 3, 70, "Section"],
Cell[2210, 72, 775, 15, 70, "Text"],
Cell[2988, 89, 866, 16, 70, "Text"],
Cell[3857, 107, 674, 17, 70, "Text"],
Cell[4534, 126, 347, 6, 70, "Text"],
Cell[4884, 134, 196, 4, 70, "Text"],
Cell[5083, 140, 207, 4, 70, "Text"],
Cell[5293, 146, 157, 4, 70, "Text"],
Cell[5453, 152, 271, 5, 70, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5756, 159, 78, 1, 70, "Section"],
Cell[5837, 162, 1179, 17, 70, "Text"],
Cell[7019, 181, 1488, 22, 70, "Text"],
Cell[8510, 205, 906, 26, 70, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[9448, 233, 84, 1, 70, "Subsection"],
Cell[9535, 236, 2577, 96, 70, "Text"],
Cell[12115, 334, 1176, 41, 70, "Text"],
Cell[13294, 377, 382, 9, 70, "Text"],
Cell[13679, 388, 781, 22, 70, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[14492, 412, 66, 1, 70, "Section"],
Cell[14561, 415, 696, 11, 70, "Text"],
Cell[15260, 428, 990, 15, 70, "Text"],
Cell[16253, 445, 68, 1, 70, "Text"],

Cell[CellGroupData[{
Cell[16344, 448, 278, 4, 70, "Input"],

Cell[CellGroupData[{
Cell[16645, 454, 110, 4, 70, "Input"],
Cell[16758, 460, 159, 5, 70, "Output"]
}, Closed]]
}, Closed]],
Cell[16938, 467, 250, 5, 70, "Text"],
Cell[17191, 474, 450, 16, 70, "Input"],
Cell[17644, 492, 236, 8, 70, "Text"],

Cell[CellGroupData[{
Cell[17903, 502, 95, 1, 70, "Input"],
Cell[18001, 505, 147, 5, 70, "Output"]
}, Closed]],
Cell[18160, 512, 318, 12, 70, "Text"],

Cell[CellGroupData[{
Cell[18501, 526, 67, 1, 70, "Input"],
Cell[18571, 529, 806, 26, 70, "Output"]
}, Closed]],
Cell[19389, 557, 267, 11, 70, "Text"],

Cell[CellGroupData[{
Cell[19679, 570, 62, 1, 70, "Input"],
Cell[19744, 573, 639, 13, 70, "Print"]
}, Closed]],
Cell[20395, 588, 145, 4, 70, "Text"],

Cell[CellGroupData[{
Cell[20563, 594, 95, 1, 70, "Input"],
Cell[20661, 597, 459, 19, 70, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[21152, 618, 80, 1, 70, "Input"],
Cell[21235, 621, 5401, 502, 70, 5258, 498, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[26648, 1125, 390, 15, 70, "Text"],

Cell[CellGroupData[{
Cell[27061, 1142, 66, 1, 70, "Input"],
Cell[27130, 1145, 808, 29, 70, "Output"]
}, Closed]],
Cell[27950, 1176, 359, 12, 70, "Text"],

Cell[CellGroupData[{
Cell[28332, 1190, 75, 1, 70, "Input"],
Cell[28410, 1193, 6546, 558, 70, 6403, 554, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[34968, 1753, 145, 6, 70, "Text"],
Cell[35116, 1761, 130, 3, 70, "Text"],
Cell[35249, 1766, 140, 4, 70, "Text"],

Cell[CellGroupData[{
Cell[35412, 1772, 110, 4, 70, "Input"],
Cell[35525, 1778, 2580, 113, 70, "Output"],
Cell[38108, 1893, 190, 7, 70, "Output"]
}, Closed]],
Cell[38310, 1902, 382, 8, 70, "Text"],

Cell[CellGroupData[{
Cell[38715, 1912, 255, 7, 70, "Input"],
Cell[38973, 1921, 41069, 3329, 70, 40926, 3325, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[80054, 5252, 445, 11, 70, "Text"],

Cell[CellGroupData[{
Cell[80522, 5265, 419, 6, 70, "Input"],
Cell[80944, 5273, 11815, 1163, 70, 11672, 1159, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[92791, 6438, 229, 8, 70, "Text"],

Cell[CellGroupData[{
Cell[93043, 6448, 158, 4, 70, "Input"],
Cell[93204, 6454, 5558, 503, 70, 5415, 499, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[98803, 6959, 806, 16, 70, "Text"],

Cell[CellGroupData[{
Cell[99632, 6977, 151, 5, 70, "Input"],
Cell[99786, 6984, 14123, 805, 70, 13980, 801, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{
Cell[113959, 7791, 77, 1, 70, "Section"],
Cell[114039, 7794, 405, 7, 70, "Text"],
Cell[114447, 7803, 145, 4, 70, "Text"],
Cell[114595, 7809, 52, 1, 70, "Text"],
Cell[114650, 7812, 205, 5, 70, "Text"],
Cell[114858, 7819, 752, 18, 70, "Text"],
Cell[115613, 7839, 196, 4, 70, "Text"],

Cell[CellGroupData[{
Cell[115832, 7845, 137, 4, 70, "Input"],
Cell[115972, 7851, 4952, 447, 70, 4809, 443, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[120936, 8300, 726, 25, 70, "Text"],

Cell[CellGroupData[{
Cell[121685, 8327, 76, 1, 70, "Input"],
Cell[121764, 8330, 6169, 497, 70, 6026, 493, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[127945, 8829, 274, 7, 70, "Text"],

Cell[CellGroupData[{
Cell[128242, 8838, 101, 1, 70, "Input"],
Cell[128346, 8841, 517, 21, 70, "Output"]
}, Closed]],
Cell[128875, 8864, 166, 4, 70, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[129073, 8870, 70, 1, 70, "Section"],
Cell[129146, 8873, 259, 5, 70, "Text"],
Cell[129408, 8880, 324, 6, 70, "Text"],
Cell[129735, 8888, 136, 4, 70, "Text"],
Cell[129874, 8894, 176, 4, 70, "Text"],
Cell[130053, 8900, 237, 5, 70, "Text"],
Cell[130293, 8907, 323, 6, 70, "Text"],
Cell[130619, 8915, 193, 4, 70, "Text"],
Cell[130815, 8921, 94, 1, 70, "Text"],
Cell[130912, 8924, 152, 4, 70, "Text"],
Cell[131067, 8930, 339, 6, 70, "Text"],
Cell[131409, 8938, 736, 13, 70, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[132177, 8953, 69, 1, 70, "Section"],
Cell[132249, 8956, 562, 9, 70, "Text"],
Cell[132814, 8967, 100, 3, 70, "Text"],
Cell[132917, 8972, 152, 4, 70, "Text"],
Cell[133072, 8978, 127, 3, 70, "Text"],
Cell[133202, 8983, 316, 6, 70, "Text"],
Cell[133521, 8991, 504, 8, 70, "Text"],

Cell[CellGroupData[{
Cell[134048, 9001, 99, 4, 70, "Input"],
Cell[134150, 9007, 225, 9, 70, "Output"]
}, Closed]],
Cell[134387, 9018, 108, 1, 70, "Text"],
Cell[134498, 9021, 131, 3, 70, "Text"],
Cell[134632, 9026, 272, 5, 70, "Text"],
Cell[134907, 9033, 171, 4, 70, "Text"],
Cell[135081, 9039, 225, 5, 70, "Text"],
Cell[135309, 9046, 67, 1, 70, "Text"],
Cell[135379, 9049, 217, 5, 70, "Text"],
Cell[135599, 9056, 758, 14, 70, "Text"],
Cell[136360, 9072, 210, 5, 70, "Text"],
Cell[136573, 9079, 445, 8, 70, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[137050, 9089, 80, 1, 70, "Section"],
Cell[137133, 9092, 250, 5, 70, "Text"],
Cell[137386, 9099, 261, 6, 70, "Text"],
Cell[137650, 9107, 810, 18, 70, "Text"],
Cell[138463, 9127, 523, 9, 70, "Text"],
Cell[138989, 9138, 233, 5, 70, "Text"],
Cell[139225, 9145, 98, 1, 70, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[139355, 9148, 79, 1, 70, "Section"],
Cell[139437, 9151, 913, 21, 70, "Text"],
Cell[140353, 9174, 1184, 22, 70, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[141569, 9198, 66, 1, 70, "Section"],
Cell[141638, 9201, 301, 7, 70, "Text"],

Cell[CellGroupData[{
Cell[141962, 9210, 81, 1, 70, "Input"],
Cell[142046, 9213, 12354, 417, 70, "Output"]
}, Closed]],
Cell[154412, 9632, 778, 27, 70, "Text"],

Cell[CellGroupData[{
Cell[155213, 9661, 92, 1, 70, "Input"],
Cell[155308, 9664, 175, 5, 70, "Output"]
}, Closed]],
Cell[155495, 9671, 489, 14, 70, "Text"],

Cell[CellGroupData[{
Cell[156007, 9687, 67, 1, 70, "Input"],
Cell[156077, 9690, 125, 5, 70, "Output"]
}, Closed]],
Cell[156214, 9697, 240, 8, 70, "Text"],
Cell[156457, 9707, 284, 7, 70, "Text"],

Cell[CellGroupData[{
Cell[156764, 9716, 61, 1, 70, "Input"],
Cell[156828, 9719, 325, 13, 70, "Output"]
}, Closed]],
Cell[157165, 9734, 71, 1, 70, "Text"],

Cell[CellGroupData[{
Cell[157259, 9737, 103, 4, 70, "Input"],
Cell[157365, 9743, 93, 5, 70, "Output"],
Cell[157461, 9750, 157, 5, 70, "Output"]
}, Closed]],
Cell[157630, 9757, 351, 8, 70, "Text"],

Cell[CellGroupData[{
Cell[158004, 9767, 76, 1, 70, "Input"],
Cell[158083, 9770, 171, 5, 70, "Output"]
}, Closed]],
Cell[158266, 9777, 431, 9, 70, "Text"],

Cell[CellGroupData[{
Cell[158720, 9788, 76, 1, 70, "Input"],
Cell[158799, 9791, 42626, 1499, 70, "Output"]
}, Closed]],
Cell[201437, 11292, 1051, 35, 70, "Text"],

Cell[CellGroupData[{
Cell[202511, 11329, 61, 1, 70, "Input"],
Cell[202575, 11332, 417, 13, 70, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[203024, 11347, 134, 5, 70, "Input"],
Cell[203161, 11354, 18215, 649, 70, "Output"],
Cell[221379, 12005, 22838, 799, 70, "Output"],
Cell[244220, 12806, 1627, 53, 70, "Output"]
}, Closed]],
Cell[245859, 12861, 156, 4, 70, "Text"],
Cell[246018, 12867, 1341, 41, 70, "Text"],
Cell[247362, 12910, 195, 7, 70, "Text"],

Cell[CellGroupData[{
Cell[247580, 12919, 75, 1, 70, "Input"],
Cell[247658, 12922, 1135, 42, 70, "Output"]
}, Closed]],
Cell[248805, 12966, 312, 11, 70, "Text"],
Cell[249120, 12979, 101, 1, 70, "Text"],

Cell[CellGroupData[{
Cell[249244, 12982, 67, 1, 70, "Input"],
Cell[249314, 12985, 876, 17, 70, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[250222, 13004, 133, 4, 70, "Input"],
Cell[250358, 13010, 209054, 7049, 70, "Output"]
}, Closed]],
Cell[459424, 20061, 450, 15, 70, "Text"],

Cell[CellGroupData[{
Cell[459897, 20078, 61, 1, 70, "Input"],
Cell[459961, 20081, 991, 33, 70, "Output"]
}, Closed]],
Cell[460964, 20116, 109, 1, 70, "Text"],

Cell[CellGroupData[{
Cell[461096, 20119, 75, 1, 70, "Input"],
Cell[461174, 20122, 111, 5, 70, "Output"]
}, Closed]],
Cell[461297, 20129, 178, 4, 70, "Text"],
Cell[461478, 20135, 140, 4, 70, "Text"],

Cell[CellGroupData[{
Cell[461641, 20141, 74, 1, 70, "Input"],
Cell[461718, 20144, 1662, 60, 70, "Output"]
}, Closed]],
Cell[463392, 20206, 628, 12, 70, "Text"],

Cell[CellGroupData[{
Cell[464043, 20220, 104, 1, 70, "Input"],
Cell[464150, 20223, 17722, 1374, 70, 17579, 1370, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[481884, 21599, 174, 4, 70, "Text"],

Cell[CellGroupData[{
Cell[482081, 21605, 86, 1, 70, "Input"],
Cell[482170, 21608, 5441, 477, 70, 5298, 473, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[487623, 22087, 245, 5, 70, "Text"],

Cell[CellGroupData[{
Cell[487891, 22094, 87, 1, 70, "Input"],
Cell[487981, 22097, 5098, 503, 70, 4955, 499, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[493091, 22602, 203, 4, 70, "Text"],

Cell[CellGroupData[{
Cell[493317, 22608, 257, 4, 70, "Input"],
Cell[493577, 22614, 202, 8, 70, "Output"]
}, Closed]],
Cell[493791, 22624, 261, 5, 70, "Text"],

Cell[CellGroupData[{
Cell[494075, 22631, 104, 1, 70, "Input"],
Cell[494182, 22634, 7030, 631, 70, 6887, 627, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[501224, 23267, 206, 4, 70, "Text"],

Cell[CellGroupData[{
Cell[501453, 23273, 80, 1, 70, "Input"],
Cell[501536, 23276, 9216, 733, 70, 9073, 729, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[510764, 24011, 1364, 27, 70, "Text"],
Cell[512131, 24040, 184, 5, 70, "Input"],

Cell[CellGroupData[{
Cell[512338, 24047, 85, 1, 70, "Input"],
Cell[512426, 24050, 5283, 502, 70, 5140, 498, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[517721, 24554, 119, 3, 70, "Text"],

Cell[CellGroupData[{
Cell[517863, 24559, 257, 7, 70, "Input"],
Cell[518123, 24568, 1677, 60, 70, "Output"],
Cell[519803, 24630, 10851, 927, 70, 10708, 923, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[530666, 25559, 221, 5, 70, "Text"],
Cell[530890, 25566, 665, 13, 70, "Text"],

Cell[CellGroupData[{
Cell[531578, 25581, 277, 8, 70, "Input"],
Cell[531858, 25591, 1676, 60, 70, "Output"],
Cell[533537, 25653, 8646, 767, 70, 8503, 763, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[542195, 26422, 208, 4, 70, "Text"],
Cell[542406, 26428, 680, 11, 70, "Text"],
Cell[543089, 26441, 147, 4, 70, "Text"],
Cell[543239, 26447, 414, 13, 70, "Text"],

Cell[CellGroupData[{
Cell[543676, 26462, 101, 4, 70, "Input"],
Cell[543780, 26468, 80653, 6882, 70, 80510, 6878, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[624445, 33352, 536, 16, 70, "Text"],
Cell[624984, 33370, 251, 5, 70, "Text"],

Cell[CellGroupData[{
Cell[625258, 33377, 89, 4, 70, "Input"],
Cell[625350, 33383, 149, 5, 70, "Output"],
Cell[625502, 33390, 575, 11, 70, "Print"]
}, Closed]],
Cell[626089, 33403, 651, 20, 70, "Text"],
Cell[626743, 33425, 99, 1, 70, "Text"],

Cell[CellGroupData[{
Cell[626865, 33428, 69, 1, 70, "Input"],
Cell[626937, 33431, 23684, 847, 70, "Output"]
}, Closed]],
Cell[650633, 34280, 712, 22, 70, "Text"],

Cell[CellGroupData[{
Cell[651368, 34304, 74, 1, 70, "Input"],
Cell[651445, 34307, 8800, 683, 70, 8657, 679, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[660257, 34992, 138, 4, 70, "Text"],

Cell[CellGroupData[{
Cell[660418, 34998, 79, 1, 70, "Input"],
Cell[660500, 35001, 8878, 689, 70, 8735, 685, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[669390, 35692, 380, 9, 70, "Text"],
Cell[669773, 35703, 311, 6, 70, "Text"],
Cell[670087, 35711, 172, 5, 70, "Input"],
Cell[670262, 35718, 143, 4, 70, "Text"],

Cell[CellGroupData[{
Cell[670428, 35724, 163, 5, 70, "Input"],
Cell[670594, 35731, 8844, 341, 70, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[679470, 36074, 68, 1, 70, "Input"],
Cell[679541, 36077, 4740, 501, 70, 4597, 497, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[684313, 36580, 171, 5, 70, "Input"],
Cell[684487, 36587, 9081, 343, 70, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[693600, 36932, 185, 5, 70, "Input"],
Cell[693788, 36939, 64, 1, 70, "Print"],
Cell[693855, 36942, 9965, 363, 70, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[703852, 37307, 68, 1, 70, "Input"],
Cell[703923, 37310, 4240, 440, 70, 4097, 436, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[708195, 37752, 177, 5, 70, "Input"],
Cell[708375, 37759, 303, 11, 70, "Print"],
Cell[708681, 37772, 10065, 371, 70, "Output"]
}, Closed]],
Cell[718758, 38145, 678, 18, 70, "Text"],
Cell[719439, 38165, 119, 3, 70, "Text"],

Cell[CellGroupData[{
Cell[719581, 38170, 99, 4, 70, "Input"],
Cell[719683, 38176, 13000, 1187, 70, 12857, 1183, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[732695, 39365, 155, 4, 70, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[732882, 39371, 79, 1, 70, "Section"],

Cell[CellGroupData[{
Cell[732984, 39374, 66, 1, 70, "Subsection"],
Cell[733053, 39377, 855, 15, 70, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[733940, 39394, 70, 1, 70, "Subsection"],
Cell[734013, 39397, 938, 23, 70, "Text"],
Cell[734954, 39422, 222, 5, 70, "Text"],
Cell[735179, 39429, 944, 23, 70, "Input",
  Evaluatable->False],
Cell[736126, 39454, 435, 15, 70, "Text"],
Cell[736564, 39471, 195, 7, 70, "Input"],
Cell[736762, 39480, 161, 6, 70, "Text"],
Cell[736926, 39488, 172, 4, 70, "Input",
  Evaluatable->False],
Cell[737101, 39494, 982, 39, 70, "Text"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[738124, 39535, 254, 12, 70, "Section"],
Cell[738381, 39549, 880, 29, 70, "Text"],

Cell[CellGroupData[{
Cell[739284, 39580, 124, 5, 70, "Subsection"],
Cell[739411, 39587, 149, 4, 70, "Text"],

Cell[CellGroupData[{
Cell[739583, 39593, 124, 4, 70, "Input"],
Cell[739710, 39599, 506, 17, 70, "Output"]
}, Closed]],
Cell[740228, 39618, 991, 23, 70, "Text"],
Cell[741222, 39643, 238, 7, 70, "Text"],

Cell[CellGroupData[{
Cell[741483, 39652, 59, 1, 70, "Input"],
Cell[741545, 39655, 203, 5, 70, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[741780, 39662, 101, 4, 70, "Input"],
Cell[741884, 39668, 618, 22, 70, "Output"]
}, Closed]],
Cell[742514, 39692, 1226, 51, 70, "Text"],
Cell[743743, 39745, 820, 28, 70, "Text"],
Cell[744566, 39775, 725, 20, 70, "Input",
  Evaluatable->False],
Cell[745294, 39797, 335, 11, 70, "Text"],
Cell[745632, 39810, 427, 12, 70, "Text"],

Cell[CellGroupData[{
Cell[746082, 39824, 90, 1, 70, "Input"],
Cell[746175, 39827, 199, 5, 70, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[746406, 39834, 82, 1, 70, "Input"],
Cell[746491, 39837, 9304, 317, 70, "Output"]
}, Closed]],
Cell[755807, 40156, 328, 10, 70, "Text"],

Cell[CellGroupData[{
Cell[756158, 40168, 91, 1, 70, "Input"],
Cell[756252, 40171, 301, 11, 70, "Output"]
}, Closed]],
Cell[756565, 40184, 1368, 40, 70, "Text"],
Cell[757936, 40226, 1156, 40, 70, "Text"],
Cell[759095, 40268, 261, 7, 70, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[759388, 40277, 124, 5, 70, "Subsection"],
Cell[759515, 40284, 436, 19, 70, "Text"],

Cell[CellGroupData[{
Cell[759974, 40305, 108, 4, 70, "Input"],
Cell[760085, 40311, 1775, 51, 70, "Output"]
}, Closed]],
Cell[761872, 40364, 1036, 43, 70, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[762940, 40409, 124, 5, 70, "Subsection"],
Cell[763067, 40416, 455, 19, 70, "Text"],

Cell[CellGroupData[{
Cell[763545, 40437, 172, 4, 70, "Input"],
Cell[763720, 40443, 1052, 43, 70, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[764804, 40488, 66, 1, 70, "Input"],
Cell[764873, 40491, 2002, 65, 70, "Output"]
}, Closed]],
Cell[766887, 40558, 2272, 99, 70, "Text"]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{
Cell[769200, 40659, 62, 1, 70, "Section"],
Cell[769265, 40662, 271, 7, 70, "Text"],
Cell[769539, 40671, 165, 4, 70, "Text"],
Cell[769707, 40677, 300, 7, 70, "Text"],
Cell[770010, 40686, 166, 6, 70, "Text"],
Cell[770179, 40694, 309, 7, 70, "Text"],
Cell[770491, 40703, 178, 6, 70, "Text"],
Cell[770672, 40711, 250, 5, 70, "Text"],
Cell[770925, 40718, 165, 4, 70, "Text"],
Cell[771093, 40724, 217, 6, 70, "Text"],
Cell[771313, 40732, 221, 6, 70, "Text"],
Cell[771537, 40740, 204, 6, 70, "Text"],
Cell[771744, 40748, 228, 6, 70, "Text"],
Cell[771975, 40756, 303, 7, 70, "Text"],
Cell[772281, 40765, 247, 7, 70, "Text"],
Cell[772531, 40774, 236, 7, 70, "Text"],
Cell[772770, 40783, 185, 6, 70, "Text"]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)
